/*
 * Copyright (C) 2021 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "MessageArgumentDescriptions.h"

#if ENABLE(IPC_TESTING_API) || !LOG_DISABLED

#include "AudioMediaStreamTrackRendererInternalUnitIdentifier.h"
#include "ContentWorldShared.h"
#include "DataTaskIdentifier.h"
#include "GeolocationIdentifier.h"
#include "GraphicsContextGLIdentifier.h"
#include "IPCConnectionTesterIdentifier.h"
#include "IPCStreamTesterIdentifier.h"
#include "IdentifierTypes.h"
#include "JSIPCBinding.h"
#include "LegacyCustomProtocolID.h"
#include "LibWebRTCResolverIdentifier.h"
#include "MDNSRegisterIdentifier.h"
#include "MarkSurfacesAsVolatileRequestIdentifier.h"
#include "MediaRecorderIdentifier.h"
#include "NetworkResourceLoadIdentifier.h"
#include "PDFPluginIdentifier.h"
#include "PlaybackSessionContextIdentifier.h"
#include "QuotaIncreaseRequestIdentifier.h"
#include "RemoteAudioDestinationIdentifier.h"
#include "RemoteAudioHardwareListenerIdentifier.h"
#include "RemoteAudioSessionIdentifier.h"
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
#include "RemoteCDMIdentifier.h"
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
#include "RemoteCDMInstanceIdentifier.h"
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
#include "RemoteCDMInstanceSessionIdentifier.h"
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
#include "RemoteLegacyCDMIdentifier.h"
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
#include "RemoteLegacyCDMSessionIdentifier.h"
#endif
#include "RemoteMediaResourceIdentifier.h"
#include "RemoteRemoteCommandListenerIdentifier.h"
#include "RemoteSerializedImageBufferIdentifier.h"
#include "RemoteVideoFrameIdentifier.h"
#include "RenderingBackendIdentifier.h"
#include "RenderingUpdateID.h"
#include "SampleBufferDisplayLayerIdentifier.h"
#include "StorageAreaIdentifier.h"
#include "StorageAreaImplIdentifier.h"
#include "StorageAreaMapIdentifier.h"
#include "StorageNamespaceIdentifier.h"
#include "TrackPrivateRemoteIdentifier.h"
#include "TransactionID.h"
#include "UserContentControllerIdentifier.h"
#include "VideoDecoderIdentifier.h"
#include "VideoEncoderIdentifier.h"
#include "WebExtensionContextIdentifier.h"
#include "WebExtensionControllerIdentifier.h"
#include "WebGPUIdentifier.h"
#include "WebPageProxyIdentifier.h"
#include "WebURLSchemeHandlerIdentifier.h"
#include <WebCore/BroadcastChannelIdentifier.h>
#include <WebCore/DOMCacheIdentifier.h>
#include <WebCore/DictationContext.h>
#include <WebCore/DisplayList.h>
#include <WebCore/ElementIdentifier.h>
#include <WebCore/FetchIdentifier.h>
#include <WebCore/FileSystemHandleIdentifier.h>
#include <WebCore/FileSystemSyncAccessHandleIdentifier.h>
#include <WebCore/ImageDecoderIdentifier.h>
#include <WebCore/LayerHostingContextIdentifier.h>
#include <WebCore/LibWebRTCSocketIdentifier.h>
#include <WebCore/MediaKeySystemRequestIdentifier.h>
#include <WebCore/MediaPlayerIdentifier.h>
#include <WebCore/MediaSessionIdentifier.h>
#include <WebCore/PageIdentifier.h>
#include <WebCore/PlaybackTargetClientContextIdentifier.h>
#include <WebCore/PortIdentifier.h>
#include <WebCore/ProcessIdentifier.h>
#include <WebCore/PushSubscriptionIdentifier.h>
#include <WebCore/RealtimeMediaSourceIdentifier.h>
#include <WebCore/RenderingResourceIdentifier.h>
#include <WebCore/ResourceLoaderIdentifier.h>
#include <WebCore/SecurityOriginData.h>
#include <WebCore/ServiceWorkerIdentifier.h>
#include <WebCore/ServiceWorkerTypes.h>
#include <WebCore/SharedWorkerIdentifier.h>
#include <WebCore/SleepDisablerIdentifier.h>
#include <WebCore/SpeechRecognitionConnectionClientIdentifier.h>
#include <WebCore/TextCheckingRequestIdentifier.h>
#include <WebCore/TextManipulationToken.h>
#include <WebCore/UserMediaRequestIdentifier.h>
#include <WebCore/WebSocketIdentifier.h>
#if ENABLE(GPU_PROCESS)
#include "GPUConnectionToWebProcessMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "GPUProcessMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteDisplayListRecorderMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(WEBGL)
#include "RemoteGraphicsContextGLMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteRenderingBackendMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteAdapterMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteBindGroupMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteBindGroupLayoutMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteBufferMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteCommandBufferMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteCommandEncoderMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteCompositorIntegrationMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteComputePassEncoderMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteComputePipelineMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteDeviceMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteExternalTextureMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteGPUMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemotePipelineLayoutMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemotePresentationContextMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteQuerySetMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteQueueMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteRenderBundleMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteRenderBundleEncoderMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteRenderPassEncoderMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteRenderPipelineMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteSamplerMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteShaderModuleMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteTextureMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteTextureViewMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(WEB_AUDIO)
#include "RemoteAudioDestinationManagerMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && USE(AUDIO_SESSION)
#include "RemoteAudioSessionProxyMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
#include "RemoteCDMFactoryProxyMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
#include "RemoteCDMInstanceProxyMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
#include "RemoteCDMInstanceSessionProxyMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
#include "RemoteCDMProxyMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
#include "RemoteLegacyCDMFactoryProxyMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
#include "RemoteLegacyCDMProxyMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
#include "RemoteLegacyCDMSessionProxyMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteMediaEngineConfigurationFactoryProxyMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
#include "RemoteMediaPlayerManagerProxyMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
#include "RemoteMediaPlayerProxyMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
#include "RemoteMediaResourceManagerMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
#include "RemoteMediaSourceProxyMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteRemoteCommandListenerProxyMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
#include "RemoteSourceBufferProxyMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
#include "RemoteVideoFrameObjectHeapMessages.h" // NOLINT
#endif
#if USE(LIBWEBRTC) && PLATFORM(COCOA) && ENABLE(GPU_PROCESS)
#include "LibWebRTCCodecsProxyMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM) && PLATFORM(COCOA)
#include "RemoteAudioMediaStreamTrackRendererInternalUnitManagerMessages.h" // NOLINT
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
#include "RemoteMediaRecorderMessages.h" // NOLINT
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
#include "RemoteMediaRecorderManagerMessages.h" // NOLINT
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
#include "RemoteSampleBufferDisplayLayerMessages.h" // NOLINT
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
#include "RemoteSampleBufferDisplayLayerManagerMessages.h" // NOLINT
#endif
#include "WebCookieManagerMessages.h" // NOLINT
#include "NetworkBroadcastChannelRegistryMessages.h" // NOLINT
#include "NetworkConnectionToWebProcessMessages.h" // NOLINT
#if ENABLE(CONTENT_EXTENSIONS)
#include "NetworkContentRuleListManagerMessages.h" // NOLINT
#endif
#include "NetworkProcessMessages.h" // NOLINT
#include "NetworkResourceLoaderMessages.h" // NOLINT
#include "NetworkSocketChannelMessages.h" // NOLINT
#include "NetworkSocketStreamMessages.h" // NOLINT
#if ENABLE(SERVICE_WORKER)
#include "ServiceWorkerDownloadTaskMessages.h" // NOLINT
#endif
#if ENABLE(SERVICE_WORKER)
#include "ServiceWorkerFetchTaskMessages.h" // NOLINT
#endif
#if ENABLE(SERVICE_WORKER)
#include "WebSWServerConnectionMessages.h" // NOLINT
#endif
#if ENABLE(SERVICE_WORKER)
#include "WebSWServerToContextConnectionMessages.h" // NOLINT
#endif
#include "WebSharedWorkerServerConnectionMessages.h" // NOLINT
#include "WebSharedWorkerServerToContextConnectionMessages.h" // NOLINT
#include "CacheStorageEngineConnectionMessages.h" // NOLINT
#include "NetworkStorageManagerMessages.h" // NOLINT
#if ENABLE(WEB_RTC)
#include "NetworkMDNSRegisterMessages.h" // NOLINT
#endif
#if USE(LIBWEBRTC)
#include "NetworkRTCMonitorMessages.h" // NOLINT
#endif
#if USE(LIBWEBRTC)
#include "NetworkRTCProviderMessages.h" // NOLINT
#endif
#if ENABLE(WEB_RTC)
#include "RTCDataChannelRemoteManagerProxyMessages.h" // NOLINT
#endif
#include "AuxiliaryProcessMessages.h" // NOLINT
#if ENABLE(IPC_TESTING_API)
#include "IPCConnectionTesterMessages.h" // NOLINT
#endif
#if ENABLE(IPC_TESTING_API)
#include "IPCStreamTesterMessages.h" // NOLINT
#endif
#if ENABLE(IPC_TESTING_API)
#include "IPCStreamTesterProxyMessages.h" // NOLINT
#endif
#if ENABLE(IPC_TESTING_API)
#include "IPCTesterMessages.h" // NOLINT
#endif
#if ENABLE(IPC_TESTING_API)
#include "IPCTesterReceiverMessages.h" // NOLINT
#endif
#include "WebConnectionMessages.h" // NOLINT
#include "AuthenticationManagerMessages.h" // NOLINT
#include "NotificationManagerMessageHandlerMessages.h" // NOLINT
#include "DrawingAreaProxyMessages.h" // NOLINT
#if ENABLE(MEDIA_SESSION_COORDINATOR)
#include "RemoteMediaSessionCoordinatorProxyMessages.h" // NOLINT
#endif
#if ENABLE(MEDIA_STREAM)
#include "SpeechRecognitionRemoteRealtimeMediaSourceManagerMessages.h" // NOLINT
#endif
#include "SpeechRecognitionServerMessages.h" // NOLINT
#include "VisitedLinkStoreMessages.h" // NOLINT
#include "WebFrameProxyMessages.h" // NOLINT
#if ENABLE(FULLSCREEN_API)
#include "WebFullScreenManagerProxyMessages.h" // NOLINT
#endif
#include "WebGeolocationManagerProxyMessages.h" // NOLINT
#include "WebLockRegistryProxyMessages.h" // NOLINT
#include "WebPageProxyMessages.h" // NOLINT
#include "WebPasteboardProxyMessages.h" // NOLINT
#include "WebPermissionControllerProxyMessages.h" // NOLINT
#include "WebProcessPoolMessages.h" // NOLINT
#include "WebProcessProxyMessages.h" // NOLINT
#include "WebScreenOrientationManagerProxyMessages.h" // NOLINT
#include "WebAutomationSessionMessages.h" // NOLINT
#include "DownloadProxyMessages.h" // NOLINT
#if ENABLE(GPU_PROCESS)
#include "GPUProcessProxyMessages.h" // NOLINT
#endif
#include "RemoteWebInspectorUIProxyMessages.h" // NOLINT
#include "WebInspectorUIProxyMessages.h" // NOLINT
#include "NetworkProcessProxyMessages.h" // NOLINT
#include "WebUserContentControllerProxyMessages.h" // NOLINT
#if ENABLE(WEBXR)
#include "PlatformXRSystemMessages.h" // NOLINT
#endif
#include "WebProcessMessages.h" // NOLINT
#include "WebAutomationSessionProxyMessages.h" // NOLINT
#include "WebIDBConnectionToServerMessages.h" // NOLINT
#if ENABLE(FULLSCREEN_API)
#include "WebFullScreenManagerMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "GPUProcessConnectionMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(WEBGL)
#include "RemoteGraphicsContextGLProxyMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteRenderingBackendProxyMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteGPUProxyMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteAudioHardwareListenerMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
#include "MediaPlayerPrivateRemoteMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
#include "MediaSourcePrivateRemoteMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && USE(AUDIO_SESSION)
#include "RemoteAudioSessionMessages.h" // NOLINT
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS)
#include "RemoteAudioSourceProviderManagerMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
#include "RemoteCDMInstanceMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
#include "RemoteCDMInstanceSessionMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
#include "RemoteLegacyCDMSessionMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS)
#include "RemoteRemoteCommandListenerMessages.h" // NOLINT
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
#include "SourceBufferPrivateRemoteMessages.h" // NOLINT
#endif
#if USE(LIBWEBRTC) && PLATFORM(COCOA) && ENABLE(GPU_PROCESS)
#include "LibWebRTCCodecsMessages.h" // NOLINT
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
#include "RemoteVideoFrameObjectHeapProxyProcessorMessages.h" // NOLINT
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
#include "SampleBufferDisplayLayerMessages.h" // NOLINT
#endif
#include "WebGeolocationManagerMessages.h" // NOLINT
#include "RemoteWebInspectorUIMessages.h" // NOLINT
#include "WebInspectorMessages.h" // NOLINT
#include "WebInspectorInterruptDispatcherMessages.h" // NOLINT
#include "WebInspectorUIMessages.h" // NOLINT
#if ENABLE(MEDIA_SESSION_COORDINATOR)
#include "RemoteMediaSessionCoordinatorMessages.h" // NOLINT
#endif
#if USE(LIBWEBRTC)
#include "LibWebRTCNetworkMessages.h" // NOLINT
#endif
#if ENABLE(WEB_RTC)
#include "RTCDataChannelRemoteManagerMessages.h" // NOLINT
#endif
#if ENABLE(WEB_RTC)
#include "WebMDNSRegisterMessages.h" // NOLINT
#endif
#if USE(LIBWEBRTC)
#include "WebRTCMonitorMessages.h" // NOLINT
#endif
#if USE(LIBWEBRTC)
#include "WebRTCResolverMessages.h" // NOLINT
#endif
#include "NetworkProcessConnectionMessages.h" // NOLINT
#include "WebResourceLoaderMessages.h" // NOLINT
#include "WebSocketChannelMessages.h" // NOLINT
#include "WebSocketStreamMessages.h" // NOLINT
#include "WebNotificationManagerMessages.h" // NOLINT
#if ENABLE(MEDIA_STREAM)
#include "SpeechRecognitionRealtimeMediaSourceManagerMessages.h" // NOLINT
#endif
#include "WebSharedWorkerContextManagerConnectionMessages.h" // NOLINT
#include "WebSharedWorkerObjectConnectionMessages.h" // NOLINT
#if ENABLE(SERVICE_WORKER)
#include "WebSWClientConnectionMessages.h" // NOLINT
#endif
#if ENABLE(SERVICE_WORKER)
#include "WebSWContextManagerConnectionMessages.h" // NOLINT
#endif
#include "WebUserContentControllerMessages.h" // NOLINT
#include "RemoteWebLockRegistryMessages.h" // NOLINT
#include "WebBroadcastChannelRegistryMessages.h" // NOLINT
#include "WebFileSystemStorageConnectionMessages.h" // NOLINT
#include "WebPermissionControllerMessages.h" // NOLINT
#include "WebScreenOrientationManagerMessages.h" // NOLINT
#include "WebSpeechRecognitionConnectionMessages.h" // NOLINT
#include "DrawingAreaMessages.h" // NOLINT
#include "EventDispatcherMessages.h" // NOLINT
#include "VisitedLinkTableControllerMessages.h" // NOLINT
#include "WebFrameMessages.h" // NOLINT
#include "WebPageMessages.h" // NOLINT
#include "StorageAreaMapMessages.h" // NOLINT
#if ENABLE(WEBXR)
#include "PlatformXRSystemProxyMessages.h" // NOLINT
#endif
#include "ViewGestureControllerMessages.h" // NOLINT
#include "GtkSettingsManagerProxyMessages.h" // NOLINT
#include "ViewGestureGeometryCollectorMessages.h" // NOLINT
#if ENABLE(MEDIA_STREAM)
#include "UserMediaCaptureManagerMessages.h" // NOLINT
#endif

namespace IPC {

#if ENABLE(IPC_TESTING_API)

std::optional<JSC::JSValue> jsValueForArguments(JSC::JSGlobalObject* globalObject, MessageName name, Decoder& decoder)
{
    switch (name) {
#if ENABLE(GPU_PROCESS)
    case MessageName::GPUConnectionToWebProcess_CreateRenderingBackend:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_CreateRenderingBackend>(globalObject, decoder);
    case MessageName::GPUConnectionToWebProcess_ReleaseRenderingBackend:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_ReleaseRenderingBackend>(globalObject, decoder);
    case MessageName::GPUConnectionToWebProcess_ReleaseSerializedImageBuffer:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_ReleaseSerializedImageBuffer>(globalObject, decoder);
#if ENABLE(WEBGL)
    case MessageName::GPUConnectionToWebProcess_CreateGraphicsContextGL:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_CreateGraphicsContextGL>(globalObject, decoder);
    case MessageName::GPUConnectionToWebProcess_ReleaseGraphicsContextGL:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_ReleaseGraphicsContextGL>(globalObject, decoder);
#endif
    case MessageName::GPUConnectionToWebProcess_CreateRemoteGPU:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_CreateRemoteGPU>(globalObject, decoder);
    case MessageName::GPUConnectionToWebProcess_ClearNowPlayingInfo:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_ClearNowPlayingInfo>(globalObject, decoder);
    case MessageName::GPUConnectionToWebProcess_SetNowPlayingInfo:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_SetNowPlayingInfo>(globalObject, decoder);
#if USE(AUDIO_SESSION)
    case MessageName::GPUConnectionToWebProcess_EnsureAudioSession:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_EnsureAudioSession>(globalObject, decoder);
#endif
#if PLATFORM(IOS_FAMILY)
    case MessageName::GPUConnectionToWebProcess_EnsureMediaSessionHelper:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_EnsureMediaSessionHelper>(globalObject, decoder);
#endif
#if ENABLE(VP9)
    case MessageName::GPUConnectionToWebProcess_EnableVP9Decoders:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_EnableVP9Decoders>(globalObject, decoder);
#endif
#if HAVE(VISIBILITY_PROPAGATION_VIEW)
    case MessageName::GPUConnectionToWebProcess_CreateVisibilityPropagationContextForPage:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_CreateVisibilityPropagationContextForPage>(globalObject, decoder);
    case MessageName::GPUConnectionToWebProcess_DestroyVisibilityPropagationContextForPage:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_DestroyVisibilityPropagationContextForPage>(globalObject, decoder);
#endif
    case MessageName::GPUConnectionToWebProcess_SetMediaOverridesForTesting:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_SetMediaOverridesForTesting>(globalObject, decoder);
    case MessageName::GPUConnectionToWebProcess_CreateAudioHardwareListener:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_CreateAudioHardwareListener>(globalObject, decoder);
    case MessageName::GPUConnectionToWebProcess_ReleaseAudioHardwareListener:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_ReleaseAudioHardwareListener>(globalObject, decoder);
    case MessageName::GPUConnectionToWebProcess_CreateRemoteCommandListener:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_CreateRemoteCommandListener>(globalObject, decoder);
    case MessageName::GPUConnectionToWebProcess_ReleaseRemoteCommandListener:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_ReleaseRemoteCommandListener>(globalObject, decoder);
    case MessageName::GPUConnectionToWebProcess_SetUserPreferredLanguages:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_SetUserPreferredLanguages>(globalObject, decoder);
    case MessageName::GPUConnectionToWebProcess_ConfigureLoggingChannel:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_ConfigureLoggingChannel>(globalObject, decoder);
#if USE(GRAPHICS_LAYER_WC)
    case MessageName::GPUConnectionToWebProcess_CreateWCLayerTreeHost:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_CreateWCLayerTreeHost>(globalObject, decoder);
    case MessageName::GPUConnectionToWebProcess_ReleaseWCLayerTreeHost:
        return jsValueForDecodedMessage<MessageName::GPUConnectionToWebProcess_ReleaseWCLayerTreeHost>(globalObject, decoder);
#endif
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::GPUProcess_InitializeGPUProcess:
        return jsValueForDecodedMessage<MessageName::GPUProcess_InitializeGPUProcess>(globalObject, decoder);
    case MessageName::GPUProcess_CreateGPUConnectionToWebProcess:
        return jsValueForDecodedMessage<MessageName::GPUProcess_CreateGPUConnectionToWebProcess>(globalObject, decoder);
    case MessageName::GPUProcess_UpdateGPUProcessPreferences:
        return jsValueForDecodedMessage<MessageName::GPUProcess_UpdateGPUProcessPreferences>(globalObject, decoder);
    case MessageName::GPUProcess_PrepareToSuspend:
        return jsValueForDecodedMessage<MessageName::GPUProcess_PrepareToSuspend>(globalObject, decoder);
    case MessageName::GPUProcess_ProcessDidResume:
        return jsValueForDecodedMessage<MessageName::GPUProcess_ProcessDidResume>(globalObject, decoder);
    case MessageName::GPUProcess_AddSession:
        return jsValueForDecodedMessage<MessageName::GPUProcess_AddSession>(globalObject, decoder);
    case MessageName::GPUProcess_RemoveSession:
        return jsValueForDecodedMessage<MessageName::GPUProcess_RemoveSession>(globalObject, decoder);
#if ENABLE(MEDIA_STREAM)
    case MessageName::GPUProcess_SetMockCaptureDevicesEnabled:
        return jsValueForDecodedMessage<MessageName::GPUProcess_SetMockCaptureDevicesEnabled>(globalObject, decoder);
    case MessageName::GPUProcess_SetOrientationForMediaCapture:
        return jsValueForDecodedMessage<MessageName::GPUProcess_SetOrientationForMediaCapture>(globalObject, decoder);
    case MessageName::GPUProcess_UpdateSandboxAccess:
        return jsValueForDecodedMessage<MessageName::GPUProcess_UpdateSandboxAccess>(globalObject, decoder);
    case MessageName::GPUProcess_UpdateCaptureAccess:
        return jsValueForDecodedMessage<MessageName::GPUProcess_UpdateCaptureAccess>(globalObject, decoder);
    case MessageName::GPUProcess_UpdateCaptureOrigin:
        return jsValueForDecodedMessage<MessageName::GPUProcess_UpdateCaptureOrigin>(globalObject, decoder);
    case MessageName::GPUProcess_AddMockMediaDevice:
        return jsValueForDecodedMessage<MessageName::GPUProcess_AddMockMediaDevice>(globalObject, decoder);
    case MessageName::GPUProcess_ClearMockMediaDevices:
        return jsValueForDecodedMessage<MessageName::GPUProcess_ClearMockMediaDevices>(globalObject, decoder);
    case MessageName::GPUProcess_RemoveMockMediaDevice:
        return jsValueForDecodedMessage<MessageName::GPUProcess_RemoveMockMediaDevice>(globalObject, decoder);
    case MessageName::GPUProcess_SetMockMediaDeviceIsEphemeral:
        return jsValueForDecodedMessage<MessageName::GPUProcess_SetMockMediaDeviceIsEphemeral>(globalObject, decoder);
    case MessageName::GPUProcess_ResetMockMediaDevices:
        return jsValueForDecodedMessage<MessageName::GPUProcess_ResetMockMediaDevices>(globalObject, decoder);
    case MessageName::GPUProcess_SetMockCaptureDevicesInterrupted:
        return jsValueForDecodedMessage<MessageName::GPUProcess_SetMockCaptureDevicesInterrupted>(globalObject, decoder);
    case MessageName::GPUProcess_TriggerMockMicrophoneConfigurationChange:
        return jsValueForDecodedMessage<MessageName::GPUProcess_TriggerMockMicrophoneConfigurationChange>(globalObject, decoder);
#endif
#if PLATFORM(MAC)
    case MessageName::GPUProcess_DisplayConfigurationChanged:
        return jsValueForDecodedMessage<MessageName::GPUProcess_DisplayConfigurationChanged>(globalObject, decoder);
    case MessageName::GPUProcess_SetScreenProperties:
        return jsValueForDecodedMessage<MessageName::GPUProcess_SetScreenProperties>(globalObject, decoder);
#endif
#if HAVE(SC_CONTENT_SHARING_SESSION)
    case MessageName::GPUProcess_ShowWindowPicker:
        return jsValueForDecodedMessage<MessageName::GPUProcess_ShowWindowPicker>(globalObject, decoder);
    case MessageName::GPUProcess_ShowScreenPicker:
        return jsValueForDecodedMessage<MessageName::GPUProcess_ShowScreenPicker>(globalObject, decoder);
#endif
#if ENABLE(VIDEO)
    case MessageName::GPUProcess_RequestBitmapImageForCurrentTime:
        return jsValueForDecodedMessage<MessageName::GPUProcess_RequestBitmapImageForCurrentTime>(globalObject, decoder);
#endif
#if ENABLE(CFPREFS_DIRECT_MODE)
    case MessageName::GPUProcess_NotifyPreferencesChanged:
        return jsValueForDecodedMessage<MessageName::GPUProcess_NotifyPreferencesChanged>(globalObject, decoder);
#endif
#if PLATFORM(MAC)
    case MessageName::GPUProcess_OpenDirectoryCacheInvalidated:
        return jsValueForDecodedMessage<MessageName::GPUProcess_OpenDirectoryCacheInvalidated>(globalObject, decoder);
#endif
#if HAVE(AUDIO_COMPONENT_SERVER_REGISTRATIONS)
    case MessageName::GPUProcess_ConsumeAudioComponentRegistrations:
        return jsValueForDecodedMessage<MessageName::GPUProcess_ConsumeAudioComponentRegistrations>(globalObject, decoder);
#endif
#if HAVE(POWERLOG_TASK_MODE_QUERY)
    case MessageName::GPUProcess_EnablePowerLogging:
        return jsValueForDecodedMessage<MessageName::GPUProcess_EnablePowerLogging>(globalObject, decoder);
#endif
    case MessageName::GPUProcess_WebProcessConnectionCountForTesting:
        return jsValueForDecodedMessage<MessageName::GPUProcess_WebProcessConnectionCountForTesting>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteDisplayListRecorder_Save:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_Save>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_Restore:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_Restore>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_Translate:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_Translate>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_Rotate:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_Rotate>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_Scale:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_Scale>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_SetCTM:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_SetCTM>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_ConcatenateCTM:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_ConcatenateCTM>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_SetInlineFillColor:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_SetInlineFillColor>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_SetInlineStrokeColor:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_SetInlineStrokeColor>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_SetStrokeThickness:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_SetStrokeThickness>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_SetState:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_SetState>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_SetLineCap:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_SetLineCap>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_SetLineDash:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_SetLineDash>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_SetLineJoin:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_SetLineJoin>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_SetMiterLimit:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_SetMiterLimit>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_ClearShadow:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_ClearShadow>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_Clip:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_Clip>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_ClipOut:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_ClipOut>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_ClipToImageBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_ClipToImageBuffer>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_ClipOutToPath:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_ClipOutToPath>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_ClipPath:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_ClipPath>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_DrawGlyphs:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_DrawGlyphs>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_DrawDecomposedGlyphs:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_DrawDecomposedGlyphs>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_DrawFilteredImageBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_DrawFilteredImageBuffer>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_DrawImageBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_DrawImageBuffer>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_DrawNativeImage:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_DrawNativeImage>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_DrawSystemImage:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_DrawSystemImage>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_DrawPattern:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_DrawPattern>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_BeginTransparencyLayer:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_BeginTransparencyLayer>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_EndTransparencyLayer:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_EndTransparencyLayer>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_DrawRect:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_DrawRect>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_DrawLine:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_DrawLine>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_DrawLinesForText:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_DrawLinesForText>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_DrawDotsForDocumentMarker:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_DrawDotsForDocumentMarker>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_DrawEllipse:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_DrawEllipse>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_DrawPath:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_DrawPath>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_DrawFocusRingPath:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_DrawFocusRingPath>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_DrawFocusRingRects:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_DrawFocusRingRects>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_FillRect:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_FillRect>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_FillRectWithColor:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_FillRectWithColor>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_FillRectWithGradient:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_FillRectWithGradient>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_FillCompositedRect:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_FillCompositedRect>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_FillRoundedRect:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_FillRoundedRect>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_FillRectWithRoundedHole:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_FillRectWithRoundedHole>(globalObject, decoder);
#if ENABLE(INLINE_PATH_DATA)
    case MessageName::RemoteDisplayListRecorder_FillLine:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_FillLine>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_FillArc:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_FillArc>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_FillQuadCurve:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_FillQuadCurve>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_FillBezierCurve:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_FillBezierCurve>(globalObject, decoder);
#endif
    case MessageName::RemoteDisplayListRecorder_FillPath:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_FillPath>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_FillEllipse:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_FillEllipse>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_ConvertToLuminanceMask:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_ConvertToLuminanceMask>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_TransformToColorSpace:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_TransformToColorSpace>(globalObject, decoder);
#if ENABLE(VIDEO)
    case MessageName::RemoteDisplayListRecorder_PaintFrameForMedia:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_PaintFrameForMedia>(globalObject, decoder);
#endif
    case MessageName::RemoteDisplayListRecorder_StrokeRect:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_StrokeRect>(globalObject, decoder);
#if ENABLE(INLINE_PATH_DATA)
    case MessageName::RemoteDisplayListRecorder_StrokeLine:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_StrokeLine>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_StrokeLineWithColorAndThickness:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_StrokeLineWithColorAndThickness>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_StrokeArc:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_StrokeArc>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_StrokeQuadCurve:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_StrokeQuadCurve>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_StrokeBezierCurve:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_StrokeBezierCurve>(globalObject, decoder);
#endif
    case MessageName::RemoteDisplayListRecorder_StrokePath:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_StrokePath>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_StrokeEllipse:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_StrokeEllipse>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_ClearRect:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_ClearRect>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_DrawControlPart:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_DrawControlPart>(globalObject, decoder);
#if USE(CG)
    case MessageName::RemoteDisplayListRecorder_ApplyStrokePattern:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_ApplyStrokePattern>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_ApplyFillPattern:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_ApplyFillPattern>(globalObject, decoder);
#endif
    case MessageName::RemoteDisplayListRecorder_ApplyDeviceScaleFactor:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_ApplyDeviceScaleFactor>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_FlushContext:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_FlushContext>(globalObject, decoder);
#if PLATFORM(COCOA) && ENABLE(VIDEO)
    case MessageName::RemoteDisplayListRecorder_PaintVideoFrame:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_PaintVideoFrame>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_SetSharedVideoFrameSemaphore:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_SetSharedVideoFrameSemaphore>(globalObject, decoder);
    case MessageName::RemoteDisplayListRecorder_SetSharedVideoFrameMemory:
        return jsValueForDecodedMessage<MessageName::RemoteDisplayListRecorder_SetSharedVideoFrameMemory>(globalObject, decoder);
#endif
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(WEBGL)
    case MessageName::RemoteGraphicsContextGL_Reshape:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Reshape>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::RemoteGraphicsContextGL_PrepareForDisplay:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_PrepareForDisplay>(globalObject, decoder);
#endif
#if USE(GRAPHICS_LAYER_WC)
    case MessageName::RemoteGraphicsContextGL_PrepareForDisplay:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_PrepareForDisplay>(globalObject, decoder);
#endif
#if USE(GBM)
    case MessageName::RemoteGraphicsContextGL_PrepareForDisplay:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_PrepareForDisplay>(globalObject, decoder);
#endif
#if !PLATFORM(COCOA) && !USE(GRAPHICS_LAYER_WC) && !USE(GBM)
    case MessageName::RemoteGraphicsContextGL_PrepareForDisplay:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_PrepareForDisplay>(globalObject, decoder);
#endif
    case MessageName::RemoteGraphicsContextGL_EnsureExtensionEnabled:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_EnsureExtensionEnabled>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_MarkContextChanged:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_MarkContextChanged>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_SynthesizeGLError:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_SynthesizeGLError>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetError:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetError>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_PaintRenderingResultsToCanvas:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_PaintRenderingResultsToCanvas>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_PaintCompositedResultsToCanvas:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_PaintCompositedResultsToCanvas>(globalObject, decoder);
#if ENABLE(VIDEO) && PLATFORM(COCOA)
    case MessageName::RemoteGraphicsContextGL_CopyTextureFromVideoFrame:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CopyTextureFromVideoFrame>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_SetSharedVideoFrameSemaphore:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_SetSharedVideoFrameSemaphore>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_SetSharedVideoFrameMemory:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_SetSharedVideoFrameMemory>(globalObject, decoder);
#endif
#if ENABLE(MEDIA_STREAM)
    case MessageName::RemoteGraphicsContextGL_PaintCompositedResultsToVideoFrame:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_PaintCompositedResultsToVideoFrame>(globalObject, decoder);
#endif
    case MessageName::RemoteGraphicsContextGL_SimulateEventForTesting:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_SimulateEventForTesting>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_ReadnPixels0:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_ReadnPixels0>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_ReadnPixels1:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_ReadnPixels1>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_ReadnPixels2:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_ReadnPixels2>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_MultiDrawArraysANGLE:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_MultiDrawArraysANGLE>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_MultiDrawArraysInstancedANGLE:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_MultiDrawArraysInstancedANGLE>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_MultiDrawElementsANGLE:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_MultiDrawElementsANGLE>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_MultiDrawElementsInstancedANGLE:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_MultiDrawElementsInstancedANGLE>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_MultiDrawArraysInstancedBaseInstanceANGLE:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_MultiDrawArraysInstancedBaseInstanceANGLE>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_MultiDrawElementsInstancedBaseVertexBaseInstanceANGLE:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_MultiDrawElementsInstancedBaseVertexBaseInstanceANGLE>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_MoveErrorsToSyntheticErrorList:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_MoveErrorsToSyntheticErrorList>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_ActiveTexture:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_ActiveTexture>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_AttachShader:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_AttachShader>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BindAttribLocation:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BindAttribLocation>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BindBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BindBuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BindFramebuffer:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BindFramebuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BindRenderbuffer:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BindRenderbuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BindTexture:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BindTexture>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BlendColor:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BlendColor>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BlendEquation:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BlendEquation>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BlendEquationSeparate:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BlendEquationSeparate>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BlendFunc:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BlendFunc>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BlendFuncSeparate:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BlendFuncSeparate>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CheckFramebufferStatus:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CheckFramebufferStatus>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Clear:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Clear>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_ClearColor:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_ClearColor>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_ClearDepth:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_ClearDepth>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_ClearStencil:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_ClearStencil>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_ColorMask:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_ColorMask>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CompileShader:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CompileShader>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CopyTexImage2D:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CopyTexImage2D>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CopyTexSubImage2D:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CopyTexSubImage2D>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CreateBuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateFramebuffer:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CreateFramebuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateProgram:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CreateProgram>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateRenderbuffer:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CreateRenderbuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateShader:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CreateShader>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateTexture:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CreateTexture>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CullFace:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CullFace>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DeleteBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DeleteBuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DeleteFramebuffer:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DeleteFramebuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DeleteProgram:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DeleteProgram>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DeleteRenderbuffer:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DeleteRenderbuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DeleteShader:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DeleteShader>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DeleteTexture:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DeleteTexture>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DepthFunc:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DepthFunc>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DepthMask:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DepthMask>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DepthRange:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DepthRange>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DetachShader:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DetachShader>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Disable:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Disable>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DisableVertexAttribArray:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DisableVertexAttribArray>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DrawArrays:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DrawArrays>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DrawElements:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DrawElements>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Enable:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Enable>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_EnableVertexAttribArray:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_EnableVertexAttribArray>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Finish:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Finish>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Flush:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Flush>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_FramebufferRenderbuffer:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_FramebufferRenderbuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_FramebufferTexture2D:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_FramebufferTexture2D>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_FrontFace:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_FrontFace>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GenerateMipmap:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GenerateMipmap>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetActiveAttrib:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetActiveAttrib>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetActiveUniform:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetActiveUniform>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetAttribLocation:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetAttribLocation>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetBufferParameteri:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetBufferParameteri>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetString:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetString>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetFloatv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetFloatv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetIntegerv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetIntegerv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetIntegeri_v:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetIntegeri_v>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetInteger64:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetInteger64>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetInteger64i:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetInteger64i>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetProgrami:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetProgrami>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetBooleanv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetBooleanv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetFramebufferAttachmentParameteri:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetFramebufferAttachmentParameteri>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetProgramInfoLog:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetProgramInfoLog>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetRenderbufferParameteri:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetRenderbufferParameteri>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetShaderi:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetShaderi>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetShaderInfoLog:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetShaderInfoLog>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetShaderPrecisionFormat:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetShaderPrecisionFormat>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetShaderSource:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetShaderSource>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetTexParameterf:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetTexParameterf>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetTexParameteri:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetTexParameteri>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetUniformfv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetUniformfv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetUniformiv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetUniformiv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetUniformuiv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetUniformuiv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetUniformLocation:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetUniformLocation>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetVertexAttribOffset:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetVertexAttribOffset>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Hint:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Hint>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_IsBuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsEnabled:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_IsEnabled>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsFramebuffer:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_IsFramebuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsProgram:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_IsProgram>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsRenderbuffer:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_IsRenderbuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsShader:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_IsShader>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsTexture:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_IsTexture>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_LineWidth:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_LineWidth>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_LinkProgram:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_LinkProgram>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_PixelStorei:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_PixelStorei>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_PolygonOffset:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_PolygonOffset>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_RenderbufferStorage:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_RenderbufferStorage>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_SampleCoverage:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_SampleCoverage>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Scissor:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Scissor>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_ShaderSource:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_ShaderSource>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_StencilFunc:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_StencilFunc>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_StencilFuncSeparate:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_StencilFuncSeparate>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_StencilMask:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_StencilMask>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_StencilMaskSeparate:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_StencilMaskSeparate>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_StencilOp:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_StencilOp>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_StencilOpSeparate:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_StencilOpSeparate>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_TexParameterf:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_TexParameterf>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_TexParameteri:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_TexParameteri>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform1f:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform1f>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform1fv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform1fv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform1i:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform1i>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform1iv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform1iv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform2f:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform2f>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform2fv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform2fv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform2i:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform2i>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform2iv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform2iv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform3f:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform3f>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform3fv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform3fv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform3i:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform3i>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform3iv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform3iv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform4f:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform4f>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform4fv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform4fv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform4i:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform4i>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform4iv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform4iv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_UniformMatrix2fv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_UniformMatrix2fv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_UniformMatrix3fv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_UniformMatrix3fv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_UniformMatrix4fv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_UniformMatrix4fv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_UseProgram:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_UseProgram>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_ValidateProgram:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_ValidateProgram>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_VertexAttrib1f:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_VertexAttrib1f>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_VertexAttrib1fv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_VertexAttrib1fv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_VertexAttrib2f:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_VertexAttrib2f>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_VertexAttrib2fv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_VertexAttrib2fv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_VertexAttrib3f:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_VertexAttrib3f>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_VertexAttrib3fv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_VertexAttrib3fv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_VertexAttrib4f:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_VertexAttrib4f>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_VertexAttrib4fv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_VertexAttrib4fv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_VertexAttribPointer:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_VertexAttribPointer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Viewport:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Viewport>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BufferData0:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BufferData0>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BufferData1:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BufferData1>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BufferSubData:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BufferSubData>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_TexImage2D0:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_TexImage2D0>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_TexImage2D1:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_TexImage2D1>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_TexSubImage2D0:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_TexSubImage2D0>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_TexSubImage2D1:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_TexSubImage2D1>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CompressedTexImage2D0:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CompressedTexImage2D0>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CompressedTexImage2D1:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CompressedTexImage2D1>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CompressedTexSubImage2D0:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CompressedTexSubImage2D0>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CompressedTexSubImage2D1:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CompressedTexSubImage2D1>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DrawArraysInstanced:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DrawArraysInstanced>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DrawElementsInstanced:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DrawElementsInstanced>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_VertexAttribDivisor:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_VertexAttribDivisor>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateVertexArray:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CreateVertexArray>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DeleteVertexArray:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DeleteVertexArray>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsVertexArray:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_IsVertexArray>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BindVertexArray:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BindVertexArray>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CopyBufferSubData:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CopyBufferSubData>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetBufferSubData:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetBufferSubData>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BlitFramebuffer:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BlitFramebuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_FramebufferTextureLayer:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_FramebufferTextureLayer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_InvalidateFramebuffer:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_InvalidateFramebuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_InvalidateSubFramebuffer:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_InvalidateSubFramebuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_ReadBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_ReadBuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_RenderbufferStorageMultisample:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_RenderbufferStorageMultisample>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_TexStorage2D:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_TexStorage2D>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_TexStorage3D:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_TexStorage3D>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_TexImage3D0:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_TexImage3D0>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_TexImage3D1:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_TexImage3D1>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_TexSubImage3D0:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_TexSubImage3D0>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_TexSubImage3D1:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_TexSubImage3D1>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CopyTexSubImage3D:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CopyTexSubImage3D>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CompressedTexImage3D0:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CompressedTexImage3D0>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CompressedTexImage3D1:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CompressedTexImage3D1>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CompressedTexSubImage3D0:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CompressedTexSubImage3D0>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CompressedTexSubImage3D1:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CompressedTexSubImage3D1>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetFragDataLocation:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetFragDataLocation>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform1ui:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform1ui>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform2ui:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform2ui>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform3ui:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform3ui>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform4ui:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform4ui>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform1uiv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform1uiv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform2uiv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform2uiv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform3uiv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform3uiv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_Uniform4uiv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_Uniform4uiv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_UniformMatrix2x3fv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_UniformMatrix2x3fv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_UniformMatrix3x2fv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_UniformMatrix3x2fv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_UniformMatrix2x4fv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_UniformMatrix2x4fv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_UniformMatrix4x2fv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_UniformMatrix4x2fv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_UniformMatrix3x4fv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_UniformMatrix3x4fv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_UniformMatrix4x3fv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_UniformMatrix4x3fv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_VertexAttribI4i:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_VertexAttribI4i>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_VertexAttribI4iv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_VertexAttribI4iv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_VertexAttribI4ui:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_VertexAttribI4ui>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_VertexAttribI4uiv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_VertexAttribI4uiv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_VertexAttribIPointer:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_VertexAttribIPointer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DrawRangeElements:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DrawRangeElements>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DrawBuffers:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DrawBuffers>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_ClearBufferiv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_ClearBufferiv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_ClearBufferuiv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_ClearBufferuiv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_ClearBufferfv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_ClearBufferfv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_ClearBufferfi:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_ClearBufferfi>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateQuery:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CreateQuery>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DeleteQuery:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DeleteQuery>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsQuery:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_IsQuery>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BeginQuery:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BeginQuery>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_EndQuery:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_EndQuery>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetQuery:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetQuery>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetQueryObjectui:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetQueryObjectui>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateSampler:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CreateSampler>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DeleteSampler:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DeleteSampler>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsSampler:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_IsSampler>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BindSampler:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BindSampler>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_SamplerParameteri:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_SamplerParameteri>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_SamplerParameterf:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_SamplerParameterf>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetSamplerParameterf:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetSamplerParameterf>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetSamplerParameteri:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetSamplerParameteri>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_FenceSync:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_FenceSync>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsSync:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_IsSync>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DeleteSync:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DeleteSync>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_ClientWaitSync:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_ClientWaitSync>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_WaitSync:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_WaitSync>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetSynci:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetSynci>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateTransformFeedback:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CreateTransformFeedback>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DeleteTransformFeedback:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DeleteTransformFeedback>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsTransformFeedback:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_IsTransformFeedback>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BindTransformFeedback:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BindTransformFeedback>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BeginTransformFeedback:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BeginTransformFeedback>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_EndTransformFeedback:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_EndTransformFeedback>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_TransformFeedbackVaryings:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_TransformFeedbackVaryings>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetTransformFeedbackVarying:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetTransformFeedbackVarying>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_PauseTransformFeedback:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_PauseTransformFeedback>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_ResumeTransformFeedback:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_ResumeTransformFeedback>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BindBufferBase:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BindBufferBase>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BindBufferRange:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BindBufferRange>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetUniformIndices:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetUniformIndices>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetActiveUniforms:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetActiveUniforms>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetUniformBlockIndex:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetUniformBlockIndex>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetActiveUniformBlockName:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetActiveUniformBlockName>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_UniformBlockBinding:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_UniformBlockBinding>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetActiveUniformBlockiv:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetActiveUniformBlockiv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetTranslatedShaderSourceANGLE:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetTranslatedShaderSourceANGLE>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DrawBuffersEXT:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DrawBuffersEXT>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateQueryEXT:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_CreateQueryEXT>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DeleteQueryEXT:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DeleteQueryEXT>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsQueryEXT:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_IsQueryEXT>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BeginQueryEXT:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BeginQueryEXT>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_EndQueryEXT:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_EndQueryEXT>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_QueryCounterEXT:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_QueryCounterEXT>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetQueryiEXT:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetQueryiEXT>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetQueryObjectiEXT:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetQueryObjectiEXT>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetQueryObjectui64EXT:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetQueryObjectui64EXT>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetInteger64EXT:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetInteger64EXT>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_EnableiOES:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_EnableiOES>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DisableiOES:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DisableiOES>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BlendEquationiOES:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BlendEquationiOES>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BlendEquationSeparateiOES:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BlendEquationSeparateiOES>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BlendFunciOES:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BlendFunciOES>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_BlendFuncSeparateiOES:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_BlendFuncSeparateiOES>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_ColorMaskiOES:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_ColorMaskiOES>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DrawArraysInstancedBaseInstanceANGLE:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DrawArraysInstancedBaseInstanceANGLE>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_DrawElementsInstancedBaseVertexBaseInstanceANGLE:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_DrawElementsInstancedBaseVertexBaseInstanceANGLE>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_ProvokingVertexANGLE:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_ProvokingVertexANGLE>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetInternalformativ:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_GetInternalformativ>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_SetDrawingBufferColorSpace:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_SetDrawingBufferColorSpace>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_PaintRenderingResultsToPixelBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGL_PaintRenderingResultsToPixelBuffer>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteRenderingBackend_CreateImageBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackend_CreateImageBuffer>(globalObject, decoder);
    case MessageName::RemoteRenderingBackend_GetPixelBufferForImageBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackend_GetPixelBufferForImageBuffer>(globalObject, decoder);
    case MessageName::RemoteRenderingBackend_GetPixelBufferForImageBufferWithNewMemory:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackend_GetPixelBufferForImageBufferWithNewMemory>(globalObject, decoder);
    case MessageName::RemoteRenderingBackend_DestroyGetPixelBufferSharedMemory:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackend_DestroyGetPixelBufferSharedMemory>(globalObject, decoder);
    case MessageName::RemoteRenderingBackend_PutPixelBufferForImageBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackend_PutPixelBufferForImageBuffer>(globalObject, decoder);
    case MessageName::RemoteRenderingBackend_GetShareableBitmapForImageBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackend_GetShareableBitmapForImageBuffer>(globalObject, decoder);
    case MessageName::RemoteRenderingBackend_GetFilteredImageForImageBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackend_GetFilteredImageForImageBuffer>(globalObject, decoder);
    case MessageName::RemoteRenderingBackend_CacheNativeImage:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackend_CacheNativeImage>(globalObject, decoder);
    case MessageName::RemoteRenderingBackend_CacheFont:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackend_CacheFont>(globalObject, decoder);
    case MessageName::RemoteRenderingBackend_CacheDecomposedGlyphs:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackend_CacheDecomposedGlyphs>(globalObject, decoder);
    case MessageName::RemoteRenderingBackend_ReleaseAllResources:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackend_ReleaseAllResources>(globalObject, decoder);
    case MessageName::RemoteRenderingBackend_ReleaseResource:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackend_ReleaseResource>(globalObject, decoder);
    case MessageName::RemoteRenderingBackend_PrepareBuffersForDisplay:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackend_PrepareBuffersForDisplay>(globalObject, decoder);
    case MessageName::RemoteRenderingBackend_MarkSurfacesVolatile:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackend_MarkSurfacesVolatile>(globalObject, decoder);
    case MessageName::RemoteRenderingBackend_FinalizeRenderingUpdate:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackend_FinalizeRenderingUpdate>(globalObject, decoder);
    case MessageName::RemoteRenderingBackend_MoveToSerializedBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackend_MoveToSerializedBuffer>(globalObject, decoder);
    case MessageName::RemoteRenderingBackend_MoveToImageBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackend_MoveToImageBuffer>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteAdapter_RequestDevice:
        return jsValueForDecodedMessage<MessageName::RemoteAdapter_RequestDevice>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteBindGroup_SetLabel:
        return jsValueForDecodedMessage<MessageName::RemoteBindGroup_SetLabel>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteBindGroupLayout_SetLabel:
        return jsValueForDecodedMessage<MessageName::RemoteBindGroupLayout_SetLabel>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteBuffer_GetMappedRange:
        return jsValueForDecodedMessage<MessageName::RemoteBuffer_GetMappedRange>(globalObject, decoder);
    case MessageName::RemoteBuffer_MapAsync:
        return jsValueForDecodedMessage<MessageName::RemoteBuffer_MapAsync>(globalObject, decoder);
    case MessageName::RemoteBuffer_Unmap:
        return jsValueForDecodedMessage<MessageName::RemoteBuffer_Unmap>(globalObject, decoder);
    case MessageName::RemoteBuffer_Destroy:
        return jsValueForDecodedMessage<MessageName::RemoteBuffer_Destroy>(globalObject, decoder);
    case MessageName::RemoteBuffer_SetLabel:
        return jsValueForDecodedMessage<MessageName::RemoteBuffer_SetLabel>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteCommandBuffer_SetLabel:
        return jsValueForDecodedMessage<MessageName::RemoteCommandBuffer_SetLabel>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteCommandEncoder_BeginRenderPass:
        return jsValueForDecodedMessage<MessageName::RemoteCommandEncoder_BeginRenderPass>(globalObject, decoder);
    case MessageName::RemoteCommandEncoder_BeginComputePass:
        return jsValueForDecodedMessage<MessageName::RemoteCommandEncoder_BeginComputePass>(globalObject, decoder);
    case MessageName::RemoteCommandEncoder_CopyBufferToBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteCommandEncoder_CopyBufferToBuffer>(globalObject, decoder);
    case MessageName::RemoteCommandEncoder_CopyBufferToTexture:
        return jsValueForDecodedMessage<MessageName::RemoteCommandEncoder_CopyBufferToTexture>(globalObject, decoder);
    case MessageName::RemoteCommandEncoder_CopyTextureToBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteCommandEncoder_CopyTextureToBuffer>(globalObject, decoder);
    case MessageName::RemoteCommandEncoder_CopyTextureToTexture:
        return jsValueForDecodedMessage<MessageName::RemoteCommandEncoder_CopyTextureToTexture>(globalObject, decoder);
    case MessageName::RemoteCommandEncoder_ClearBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteCommandEncoder_ClearBuffer>(globalObject, decoder);
    case MessageName::RemoteCommandEncoder_PushDebugGroup:
        return jsValueForDecodedMessage<MessageName::RemoteCommandEncoder_PushDebugGroup>(globalObject, decoder);
    case MessageName::RemoteCommandEncoder_PopDebugGroup:
        return jsValueForDecodedMessage<MessageName::RemoteCommandEncoder_PopDebugGroup>(globalObject, decoder);
    case MessageName::RemoteCommandEncoder_InsertDebugMarker:
        return jsValueForDecodedMessage<MessageName::RemoteCommandEncoder_InsertDebugMarker>(globalObject, decoder);
    case MessageName::RemoteCommandEncoder_WriteTimestamp:
        return jsValueForDecodedMessage<MessageName::RemoteCommandEncoder_WriteTimestamp>(globalObject, decoder);
    case MessageName::RemoteCommandEncoder_ResolveQuerySet:
        return jsValueForDecodedMessage<MessageName::RemoteCommandEncoder_ResolveQuerySet>(globalObject, decoder);
    case MessageName::RemoteCommandEncoder_Finish:
        return jsValueForDecodedMessage<MessageName::RemoteCommandEncoder_Finish>(globalObject, decoder);
    case MessageName::RemoteCommandEncoder_SetLabel:
        return jsValueForDecodedMessage<MessageName::RemoteCommandEncoder_SetLabel>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
#if PLATFORM(COCOA)
    case MessageName::RemoteCompositorIntegration_RecreateRenderBuffers:
        return jsValueForDecodedMessage<MessageName::RemoteCompositorIntegration_RecreateRenderBuffers>(globalObject, decoder);
#endif
    case MessageName::RemoteCompositorIntegration_PrepareForDisplay:
        return jsValueForDecodedMessage<MessageName::RemoteCompositorIntegration_PrepareForDisplay>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteComputePassEncoder_SetPipeline:
        return jsValueForDecodedMessage<MessageName::RemoteComputePassEncoder_SetPipeline>(globalObject, decoder);
    case MessageName::RemoteComputePassEncoder_Dispatch:
        return jsValueForDecodedMessage<MessageName::RemoteComputePassEncoder_Dispatch>(globalObject, decoder);
    case MessageName::RemoteComputePassEncoder_DispatchIndirect:
        return jsValueForDecodedMessage<MessageName::RemoteComputePassEncoder_DispatchIndirect>(globalObject, decoder);
    case MessageName::RemoteComputePassEncoder_End:
        return jsValueForDecodedMessage<MessageName::RemoteComputePassEncoder_End>(globalObject, decoder);
    case MessageName::RemoteComputePassEncoder_SetBindGroup:
        return jsValueForDecodedMessage<MessageName::RemoteComputePassEncoder_SetBindGroup>(globalObject, decoder);
    case MessageName::RemoteComputePassEncoder_PushDebugGroup:
        return jsValueForDecodedMessage<MessageName::RemoteComputePassEncoder_PushDebugGroup>(globalObject, decoder);
    case MessageName::RemoteComputePassEncoder_PopDebugGroup:
        return jsValueForDecodedMessage<MessageName::RemoteComputePassEncoder_PopDebugGroup>(globalObject, decoder);
    case MessageName::RemoteComputePassEncoder_InsertDebugMarker:
        return jsValueForDecodedMessage<MessageName::RemoteComputePassEncoder_InsertDebugMarker>(globalObject, decoder);
    case MessageName::RemoteComputePassEncoder_SetLabel:
        return jsValueForDecodedMessage<MessageName::RemoteComputePassEncoder_SetLabel>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteComputePipeline_GetBindGroupLayout:
        return jsValueForDecodedMessage<MessageName::RemoteComputePipeline_GetBindGroupLayout>(globalObject, decoder);
    case MessageName::RemoteComputePipeline_SetLabel:
        return jsValueForDecodedMessage<MessageName::RemoteComputePipeline_SetLabel>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteDevice_Destroy:
        return jsValueForDecodedMessage<MessageName::RemoteDevice_Destroy>(globalObject, decoder);
    case MessageName::RemoteDevice_CreateBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteDevice_CreateBuffer>(globalObject, decoder);
    case MessageName::RemoteDevice_CreateTexture:
        return jsValueForDecodedMessage<MessageName::RemoteDevice_CreateTexture>(globalObject, decoder);
    case MessageName::RemoteDevice_CreateSampler:
        return jsValueForDecodedMessage<MessageName::RemoteDevice_CreateSampler>(globalObject, decoder);
    case MessageName::RemoteDevice_ImportExternalTexture:
        return jsValueForDecodedMessage<MessageName::RemoteDevice_ImportExternalTexture>(globalObject, decoder);
    case MessageName::RemoteDevice_CreateBindGroupLayout:
        return jsValueForDecodedMessage<MessageName::RemoteDevice_CreateBindGroupLayout>(globalObject, decoder);
    case MessageName::RemoteDevice_CreatePipelineLayout:
        return jsValueForDecodedMessage<MessageName::RemoteDevice_CreatePipelineLayout>(globalObject, decoder);
    case MessageName::RemoteDevice_CreateBindGroup:
        return jsValueForDecodedMessage<MessageName::RemoteDevice_CreateBindGroup>(globalObject, decoder);
    case MessageName::RemoteDevice_CreateShaderModule:
        return jsValueForDecodedMessage<MessageName::RemoteDevice_CreateShaderModule>(globalObject, decoder);
    case MessageName::RemoteDevice_CreateComputePipeline:
        return jsValueForDecodedMessage<MessageName::RemoteDevice_CreateComputePipeline>(globalObject, decoder);
    case MessageName::RemoteDevice_CreateRenderPipeline:
        return jsValueForDecodedMessage<MessageName::RemoteDevice_CreateRenderPipeline>(globalObject, decoder);
    case MessageName::RemoteDevice_CreateComputePipelineAsync:
        return jsValueForDecodedMessage<MessageName::RemoteDevice_CreateComputePipelineAsync>(globalObject, decoder);
    case MessageName::RemoteDevice_CreateRenderPipelineAsync:
        return jsValueForDecodedMessage<MessageName::RemoteDevice_CreateRenderPipelineAsync>(globalObject, decoder);
    case MessageName::RemoteDevice_CreateCommandEncoder:
        return jsValueForDecodedMessage<MessageName::RemoteDevice_CreateCommandEncoder>(globalObject, decoder);
    case MessageName::RemoteDevice_CreateRenderBundleEncoder:
        return jsValueForDecodedMessage<MessageName::RemoteDevice_CreateRenderBundleEncoder>(globalObject, decoder);
    case MessageName::RemoteDevice_CreateQuerySet:
        return jsValueForDecodedMessage<MessageName::RemoteDevice_CreateQuerySet>(globalObject, decoder);
    case MessageName::RemoteDevice_PushErrorScope:
        return jsValueForDecodedMessage<MessageName::RemoteDevice_PushErrorScope>(globalObject, decoder);
    case MessageName::RemoteDevice_PopErrorScope:
        return jsValueForDecodedMessage<MessageName::RemoteDevice_PopErrorScope>(globalObject, decoder);
    case MessageName::RemoteDevice_SetLabel:
        return jsValueForDecodedMessage<MessageName::RemoteDevice_SetLabel>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteExternalTexture_SetLabel:
        return jsValueForDecodedMessage<MessageName::RemoteExternalTexture_SetLabel>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteGPU_RequestAdapter:
        return jsValueForDecodedMessage<MessageName::RemoteGPU_RequestAdapter>(globalObject, decoder);
    case MessageName::RemoteGPU_CreatePresentationContext:
        return jsValueForDecodedMessage<MessageName::RemoteGPU_CreatePresentationContext>(globalObject, decoder);
    case MessageName::RemoteGPU_CreateCompositorIntegration:
        return jsValueForDecodedMessage<MessageName::RemoteGPU_CreateCompositorIntegration>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemotePipelineLayout_SetLabel:
        return jsValueForDecodedMessage<MessageName::RemotePipelineLayout_SetLabel>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemotePresentationContext_Configure:
        return jsValueForDecodedMessage<MessageName::RemotePresentationContext_Configure>(globalObject, decoder);
    case MessageName::RemotePresentationContext_Unconfigure:
        return jsValueForDecodedMessage<MessageName::RemotePresentationContext_Unconfigure>(globalObject, decoder);
    case MessageName::RemotePresentationContext_GetCurrentTexture:
        return jsValueForDecodedMessage<MessageName::RemotePresentationContext_GetCurrentTexture>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteQuerySet_Destroy:
        return jsValueForDecodedMessage<MessageName::RemoteQuerySet_Destroy>(globalObject, decoder);
    case MessageName::RemoteQuerySet_SetLabel:
        return jsValueForDecodedMessage<MessageName::RemoteQuerySet_SetLabel>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteQueue_Submit:
        return jsValueForDecodedMessage<MessageName::RemoteQueue_Submit>(globalObject, decoder);
    case MessageName::RemoteQueue_OnSubmittedWorkDone:
        return jsValueForDecodedMessage<MessageName::RemoteQueue_OnSubmittedWorkDone>(globalObject, decoder);
    case MessageName::RemoteQueue_WriteBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteQueue_WriteBuffer>(globalObject, decoder);
    case MessageName::RemoteQueue_WriteTexture:
        return jsValueForDecodedMessage<MessageName::RemoteQueue_WriteTexture>(globalObject, decoder);
    case MessageName::RemoteQueue_CopyExternalImageToTexture:
        return jsValueForDecodedMessage<MessageName::RemoteQueue_CopyExternalImageToTexture>(globalObject, decoder);
    case MessageName::RemoteQueue_SetLabel:
        return jsValueForDecodedMessage<MessageName::RemoteQueue_SetLabel>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteRenderBundle_SetLabel:
        return jsValueForDecodedMessage<MessageName::RemoteRenderBundle_SetLabel>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteRenderBundleEncoder_SetPipeline:
        return jsValueForDecodedMessage<MessageName::RemoteRenderBundleEncoder_SetPipeline>(globalObject, decoder);
    case MessageName::RemoteRenderBundleEncoder_SetIndexBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteRenderBundleEncoder_SetIndexBuffer>(globalObject, decoder);
    case MessageName::RemoteRenderBundleEncoder_SetVertexBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteRenderBundleEncoder_SetVertexBuffer>(globalObject, decoder);
    case MessageName::RemoteRenderBundleEncoder_Draw:
        return jsValueForDecodedMessage<MessageName::RemoteRenderBundleEncoder_Draw>(globalObject, decoder);
    case MessageName::RemoteRenderBundleEncoder_DrawIndexed:
        return jsValueForDecodedMessage<MessageName::RemoteRenderBundleEncoder_DrawIndexed>(globalObject, decoder);
    case MessageName::RemoteRenderBundleEncoder_DrawIndirect:
        return jsValueForDecodedMessage<MessageName::RemoteRenderBundleEncoder_DrawIndirect>(globalObject, decoder);
    case MessageName::RemoteRenderBundleEncoder_DrawIndexedIndirect:
        return jsValueForDecodedMessage<MessageName::RemoteRenderBundleEncoder_DrawIndexedIndirect>(globalObject, decoder);
    case MessageName::RemoteRenderBundleEncoder_SetBindGroup:
        return jsValueForDecodedMessage<MessageName::RemoteRenderBundleEncoder_SetBindGroup>(globalObject, decoder);
    case MessageName::RemoteRenderBundleEncoder_PushDebugGroup:
        return jsValueForDecodedMessage<MessageName::RemoteRenderBundleEncoder_PushDebugGroup>(globalObject, decoder);
    case MessageName::RemoteRenderBundleEncoder_PopDebugGroup:
        return jsValueForDecodedMessage<MessageName::RemoteRenderBundleEncoder_PopDebugGroup>(globalObject, decoder);
    case MessageName::RemoteRenderBundleEncoder_InsertDebugMarker:
        return jsValueForDecodedMessage<MessageName::RemoteRenderBundleEncoder_InsertDebugMarker>(globalObject, decoder);
    case MessageName::RemoteRenderBundleEncoder_Finish:
        return jsValueForDecodedMessage<MessageName::RemoteRenderBundleEncoder_Finish>(globalObject, decoder);
    case MessageName::RemoteRenderBundleEncoder_SetLabel:
        return jsValueForDecodedMessage<MessageName::RemoteRenderBundleEncoder_SetLabel>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteRenderPassEncoder_SetPipeline:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPassEncoder_SetPipeline>(globalObject, decoder);
    case MessageName::RemoteRenderPassEncoder_SetIndexBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPassEncoder_SetIndexBuffer>(globalObject, decoder);
    case MessageName::RemoteRenderPassEncoder_SetVertexBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPassEncoder_SetVertexBuffer>(globalObject, decoder);
    case MessageName::RemoteRenderPassEncoder_Draw:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPassEncoder_Draw>(globalObject, decoder);
    case MessageName::RemoteRenderPassEncoder_DrawIndexed:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPassEncoder_DrawIndexed>(globalObject, decoder);
    case MessageName::RemoteRenderPassEncoder_DrawIndirect:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPassEncoder_DrawIndirect>(globalObject, decoder);
    case MessageName::RemoteRenderPassEncoder_DrawIndexedIndirect:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPassEncoder_DrawIndexedIndirect>(globalObject, decoder);
    case MessageName::RemoteRenderPassEncoder_SetBindGroup:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPassEncoder_SetBindGroup>(globalObject, decoder);
    case MessageName::RemoteRenderPassEncoder_PushDebugGroup:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPassEncoder_PushDebugGroup>(globalObject, decoder);
    case MessageName::RemoteRenderPassEncoder_PopDebugGroup:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPassEncoder_PopDebugGroup>(globalObject, decoder);
    case MessageName::RemoteRenderPassEncoder_InsertDebugMarker:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPassEncoder_InsertDebugMarker>(globalObject, decoder);
    case MessageName::RemoteRenderPassEncoder_SetViewport:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPassEncoder_SetViewport>(globalObject, decoder);
    case MessageName::RemoteRenderPassEncoder_SetScissorRect:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPassEncoder_SetScissorRect>(globalObject, decoder);
    case MessageName::RemoteRenderPassEncoder_SetBlendConstant:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPassEncoder_SetBlendConstant>(globalObject, decoder);
    case MessageName::RemoteRenderPassEncoder_SetStencilReference:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPassEncoder_SetStencilReference>(globalObject, decoder);
    case MessageName::RemoteRenderPassEncoder_BeginOcclusionQuery:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPassEncoder_BeginOcclusionQuery>(globalObject, decoder);
    case MessageName::RemoteRenderPassEncoder_EndOcclusionQuery:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPassEncoder_EndOcclusionQuery>(globalObject, decoder);
    case MessageName::RemoteRenderPassEncoder_ExecuteBundles:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPassEncoder_ExecuteBundles>(globalObject, decoder);
    case MessageName::RemoteRenderPassEncoder_End:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPassEncoder_End>(globalObject, decoder);
    case MessageName::RemoteRenderPassEncoder_SetLabel:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPassEncoder_SetLabel>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteRenderPipeline_GetBindGroupLayout:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPipeline_GetBindGroupLayout>(globalObject, decoder);
    case MessageName::RemoteRenderPipeline_SetLabel:
        return jsValueForDecodedMessage<MessageName::RemoteRenderPipeline_SetLabel>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteSampler_SetLabel:
        return jsValueForDecodedMessage<MessageName::RemoteSampler_SetLabel>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteShaderModule_CompilationInfo:
        return jsValueForDecodedMessage<MessageName::RemoteShaderModule_CompilationInfo>(globalObject, decoder);
    case MessageName::RemoteShaderModule_SetLabel:
        return jsValueForDecodedMessage<MessageName::RemoteShaderModule_SetLabel>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteTexture_CreateView:
        return jsValueForDecodedMessage<MessageName::RemoteTexture_CreateView>(globalObject, decoder);
    case MessageName::RemoteTexture_Destroy:
        return jsValueForDecodedMessage<MessageName::RemoteTexture_Destroy>(globalObject, decoder);
    case MessageName::RemoteTexture_SetLabel:
        return jsValueForDecodedMessage<MessageName::RemoteTexture_SetLabel>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteTextureView_SetLabel:
        return jsValueForDecodedMessage<MessageName::RemoteTextureView_SetLabel>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(WEB_AUDIO)
    case MessageName::RemoteAudioDestinationManager_CreateAudioDestination:
        return jsValueForDecodedMessage<MessageName::RemoteAudioDestinationManager_CreateAudioDestination>(globalObject, decoder);
    case MessageName::RemoteAudioDestinationManager_DeleteAudioDestination:
        return jsValueForDecodedMessage<MessageName::RemoteAudioDestinationManager_DeleteAudioDestination>(globalObject, decoder);
    case MessageName::RemoteAudioDestinationManager_StartAudioDestination:
        return jsValueForDecodedMessage<MessageName::RemoteAudioDestinationManager_StartAudioDestination>(globalObject, decoder);
    case MessageName::RemoteAudioDestinationManager_StopAudioDestination:
        return jsValueForDecodedMessage<MessageName::RemoteAudioDestinationManager_StopAudioDestination>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::RemoteAudioDestinationManager_AudioSamplesStorageChanged:
        return jsValueForDecodedMessage<MessageName::RemoteAudioDestinationManager_AudioSamplesStorageChanged>(globalObject, decoder);
#endif
#endif
#if ENABLE(GPU_PROCESS) && USE(AUDIO_SESSION)
    case MessageName::RemoteAudioSessionProxy_SetCategory:
        return jsValueForDecodedMessage<MessageName::RemoteAudioSessionProxy_SetCategory>(globalObject, decoder);
    case MessageName::RemoteAudioSessionProxy_SetPreferredBufferSize:
        return jsValueForDecodedMessage<MessageName::RemoteAudioSessionProxy_SetPreferredBufferSize>(globalObject, decoder);
    case MessageName::RemoteAudioSessionProxy_TryToSetActive:
        return jsValueForDecodedMessage<MessageName::RemoteAudioSessionProxy_TryToSetActive>(globalObject, decoder);
    case MessageName::RemoteAudioSessionProxy_SetIsPlayingToBluetoothOverride:
        return jsValueForDecodedMessage<MessageName::RemoteAudioSessionProxy_SetIsPlayingToBluetoothOverride>(globalObject, decoder);
    case MessageName::RemoteAudioSessionProxy_TriggerBeginInterruptionForTesting:
        return jsValueForDecodedMessage<MessageName::RemoteAudioSessionProxy_TriggerBeginInterruptionForTesting>(globalObject, decoder);
    case MessageName::RemoteAudioSessionProxy_TriggerEndInterruptionForTesting:
        return jsValueForDecodedMessage<MessageName::RemoteAudioSessionProxy_TriggerEndInterruptionForTesting>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteCDMFactoryProxy_CreateCDM:
        return jsValueForDecodedMessage<MessageName::RemoteCDMFactoryProxy_CreateCDM>(globalObject, decoder);
    case MessageName::RemoteCDMFactoryProxy_SupportsKeySystem:
        return jsValueForDecodedMessage<MessageName::RemoteCDMFactoryProxy_SupportsKeySystem>(globalObject, decoder);
    case MessageName::RemoteCDMFactoryProxy_RemoveInstance:
        return jsValueForDecodedMessage<MessageName::RemoteCDMFactoryProxy_RemoveInstance>(globalObject, decoder);
    case MessageName::RemoteCDMFactoryProxy_RemoveSession:
        return jsValueForDecodedMessage<MessageName::RemoteCDMFactoryProxy_RemoveSession>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteCDMInstanceProxy_CreateSession:
        return jsValueForDecodedMessage<MessageName::RemoteCDMInstanceProxy_CreateSession>(globalObject, decoder);
    case MessageName::RemoteCDMInstanceProxy_InitializeWithConfiguration:
        return jsValueForDecodedMessage<MessageName::RemoteCDMInstanceProxy_InitializeWithConfiguration>(globalObject, decoder);
    case MessageName::RemoteCDMInstanceProxy_SetServerCertificate:
        return jsValueForDecodedMessage<MessageName::RemoteCDMInstanceProxy_SetServerCertificate>(globalObject, decoder);
    case MessageName::RemoteCDMInstanceProxy_SetStorageDirectory:
        return jsValueForDecodedMessage<MessageName::RemoteCDMInstanceProxy_SetStorageDirectory>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteCDMInstanceSessionProxy_SetLogIdentifier:
        return jsValueForDecodedMessage<MessageName::RemoteCDMInstanceSessionProxy_SetLogIdentifier>(globalObject, decoder);
    case MessageName::RemoteCDMInstanceSessionProxy_RequestLicense:
        return jsValueForDecodedMessage<MessageName::RemoteCDMInstanceSessionProxy_RequestLicense>(globalObject, decoder);
    case MessageName::RemoteCDMInstanceSessionProxy_UpdateLicense:
        return jsValueForDecodedMessage<MessageName::RemoteCDMInstanceSessionProxy_UpdateLicense>(globalObject, decoder);
    case MessageName::RemoteCDMInstanceSessionProxy_LoadSession:
        return jsValueForDecodedMessage<MessageName::RemoteCDMInstanceSessionProxy_LoadSession>(globalObject, decoder);
    case MessageName::RemoteCDMInstanceSessionProxy_CloseSession:
        return jsValueForDecodedMessage<MessageName::RemoteCDMInstanceSessionProxy_CloseSession>(globalObject, decoder);
    case MessageName::RemoteCDMInstanceSessionProxy_RemoveSessionData:
        return jsValueForDecodedMessage<MessageName::RemoteCDMInstanceSessionProxy_RemoveSessionData>(globalObject, decoder);
    case MessageName::RemoteCDMInstanceSessionProxy_StoreRecordOfKeyUsage:
        return jsValueForDecodedMessage<MessageName::RemoteCDMInstanceSessionProxy_StoreRecordOfKeyUsage>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteCDMProxy_GetSupportedConfiguration:
        return jsValueForDecodedMessage<MessageName::RemoteCDMProxy_GetSupportedConfiguration>(globalObject, decoder);
    case MessageName::RemoteCDMProxy_CreateInstance:
        return jsValueForDecodedMessage<MessageName::RemoteCDMProxy_CreateInstance>(globalObject, decoder);
    case MessageName::RemoteCDMProxy_LoadAndInitialize:
        return jsValueForDecodedMessage<MessageName::RemoteCDMProxy_LoadAndInitialize>(globalObject, decoder);
    case MessageName::RemoteCDMProxy_SetLogIdentifier:
        return jsValueForDecodedMessage<MessageName::RemoteCDMProxy_SetLogIdentifier>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
    case MessageName::RemoteLegacyCDMFactoryProxy_CreateCDM:
        return jsValueForDecodedMessage<MessageName::RemoteLegacyCDMFactoryProxy_CreateCDM>(globalObject, decoder);
    case MessageName::RemoteLegacyCDMFactoryProxy_SupportsKeySystem:
        return jsValueForDecodedMessage<MessageName::RemoteLegacyCDMFactoryProxy_SupportsKeySystem>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
    case MessageName::RemoteLegacyCDMProxy_SupportsMIMEType:
        return jsValueForDecodedMessage<MessageName::RemoteLegacyCDMProxy_SupportsMIMEType>(globalObject, decoder);
    case MessageName::RemoteLegacyCDMProxy_CreateSession:
        return jsValueForDecodedMessage<MessageName::RemoteLegacyCDMProxy_CreateSession>(globalObject, decoder);
    case MessageName::RemoteLegacyCDMProxy_SetPlayerId:
        return jsValueForDecodedMessage<MessageName::RemoteLegacyCDMProxy_SetPlayerId>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
    case MessageName::RemoteLegacyCDMSessionProxy_GenerateKeyRequest:
        return jsValueForDecodedMessage<MessageName::RemoteLegacyCDMSessionProxy_GenerateKeyRequest>(globalObject, decoder);
    case MessageName::RemoteLegacyCDMSessionProxy_ReleaseKeys:
        return jsValueForDecodedMessage<MessageName::RemoteLegacyCDMSessionProxy_ReleaseKeys>(globalObject, decoder);
    case MessageName::RemoteLegacyCDMSessionProxy_Update:
        return jsValueForDecodedMessage<MessageName::RemoteLegacyCDMSessionProxy_Update>(globalObject, decoder);
    case MessageName::RemoteLegacyCDMSessionProxy_CachedKeyForKeyID:
        return jsValueForDecodedMessage<MessageName::RemoteLegacyCDMSessionProxy_CachedKeyForKeyID>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteMediaEngineConfigurationFactoryProxy_CreateDecodingConfiguration:
        return jsValueForDecodedMessage<MessageName::RemoteMediaEngineConfigurationFactoryProxy_CreateDecodingConfiguration>(globalObject, decoder);
    case MessageName::RemoteMediaEngineConfigurationFactoryProxy_CreateEncodingConfiguration:
        return jsValueForDecodedMessage<MessageName::RemoteMediaEngineConfigurationFactoryProxy_CreateEncodingConfiguration>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
    case MessageName::RemoteMediaPlayerManagerProxy_CreateMediaPlayer:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerManagerProxy_CreateMediaPlayer>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerManagerProxy_DeleteMediaPlayer:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerManagerProxy_DeleteMediaPlayer>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerManagerProxy_GetSupportedTypes:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerManagerProxy_GetSupportedTypes>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerManagerProxy_SupportsTypeAndCodecs:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerManagerProxy_SupportsTypeAndCodecs>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerManagerProxy_OriginsInMediaCache:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerManagerProxy_OriginsInMediaCache>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerManagerProxy_ClearMediaCache:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerManagerProxy_ClearMediaCache>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerManagerProxy_ClearMediaCacheForOrigins:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerManagerProxy_ClearMediaCacheForOrigins>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerManagerProxy_SupportsKeySystem:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerManagerProxy_SupportsKeySystem>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
    case MessageName::RemoteMediaPlayerProxy_PrepareForPlayback:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_PrepareForPlayback>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_Load:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_Load>(globalObject, decoder);
#if ENABLE(MEDIA_SOURCE)
    case MessageName::RemoteMediaPlayerProxy_LoadMediaSource:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_LoadMediaSource>(globalObject, decoder);
#endif
    case MessageName::RemoteMediaPlayerProxy_CancelLoad:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_CancelLoad>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_PrepareToPlay:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_PrepareToPlay>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_Play:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_Play>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_Pause:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_Pause>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_SetVolume:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetVolume>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_SetMuted:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetMuted>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_Seek:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_Seek>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_SeekWithTolerance:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SeekWithTolerance>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_SetPreload:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetPreload>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_SetPrivateBrowsingMode:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetPrivateBrowsingMode>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_SetPreservesPitch:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetPreservesPitch>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_SetPitchCorrectionAlgorithm:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetPitchCorrectionAlgorithm>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_PrepareForRendering:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_PrepareForRendering>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_SetPageIsVisible:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetPageIsVisible>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_SetShouldMaintainAspectRatio:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetShouldMaintainAspectRatio>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_AcceleratedRenderingStateChanged:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_AcceleratedRenderingStateChanged>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_SetShouldDisableSleep:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetShouldDisableSleep>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_SetRate:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetRate>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_DidLoadingProgress:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_DidLoadingProgress>(globalObject, decoder);
#if ENABLE(VIDEO_PRESENTATION_MODE)
    case MessageName::RemoteMediaPlayerProxy_UpdateVideoFullscreenInlineImage:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_UpdateVideoFullscreenInlineImage>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_SetVideoFullscreenMode:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetVideoFullscreenMode>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_SetVideoFullscreenGravity:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetVideoFullscreenGravity>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_VideoFullscreenStandbyChanged:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_VideoFullscreenStandbyChanged>(globalObject, decoder);
#endif
    case MessageName::RemoteMediaPlayerProxy_SetBufferingPolicy:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetBufferingPolicy>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_SetPresentationSize:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetPresentationSize>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::RemoteMediaPlayerProxy_SetVideoInlineSizeFenced:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetVideoInlineSizeFenced>(globalObject, decoder);
#endif
#if ENABLE(WIRELESS_PLAYBACK_TARGET)
    case MessageName::RemoteMediaPlayerProxy_SetWirelessVideoPlaybackDisabled:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetWirelessVideoPlaybackDisabled>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_SetWirelessPlaybackTarget:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetWirelessPlaybackTarget>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_SetShouldPlayToPlaybackTarget:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetShouldPlayToPlaybackTarget>(globalObject, decoder);
#endif
#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
    case MessageName::RemoteMediaPlayerProxy_SetLegacyCDMSession:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetLegacyCDMSession>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_KeyAdded:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_KeyAdded>(globalObject, decoder);
#endif
#if ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteMediaPlayerProxy_CdmInstanceAttached:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_CdmInstanceAttached>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_CdmInstanceDetached:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_CdmInstanceDetached>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_AttemptToDecryptWithInstance:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_AttemptToDecryptWithInstance>(globalObject, decoder);
#endif
#if ENABLE(LEGACY_ENCRYPTED_MEDIA) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteMediaPlayerProxy_SetShouldContinueAfterKeyNeeded:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetShouldContinueAfterKeyNeeded>(globalObject, decoder);
#endif
    case MessageName::RemoteMediaPlayerProxy_BeginSimulatedHDCPError:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_BeginSimulatedHDCPError>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_EndSimulatedHDCPError:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_EndSimulatedHDCPError>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_NotifyActiveSourceBuffersChanged:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_NotifyActiveSourceBuffersChanged>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_ApplicationWillResignActive:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_ApplicationWillResignActive>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_ApplicationDidBecomeActive:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_ApplicationDidBecomeActive>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_NotifyTrackModeChanged:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_NotifyTrackModeChanged>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_TracksChanged:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_TracksChanged>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_AudioTrackSetEnabled:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_AudioTrackSetEnabled>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_VideoTrackSetSelected:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_VideoTrackSetSelected>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_TextTrackSetMode:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_TextTrackSetMode>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_PerformTaskAtMediaTime:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_PerformTaskAtMediaTime>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_IsCrossOrigin:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_IsCrossOrigin>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_SetVideoPlaybackMetricsUpdateInterval:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetVideoPlaybackMetricsUpdateInterval>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_SetPreferredDynamicRangeMode:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetPreferredDynamicRangeMode>(globalObject, decoder);
#if PLATFORM(IOS_FAMILY)
    case MessageName::RemoteMediaPlayerProxy_ErrorLog:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_ErrorLog>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_AccessLog:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_AccessLog>(globalObject, decoder);
#endif
#if ENABLE(WEB_AUDIO)
    case MessageName::RemoteMediaPlayerProxy_CreateAudioSourceProvider:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_CreateAudioSourceProvider>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_SetShouldEnableAudioSourceProvider:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetShouldEnableAudioSourceProvider>(globalObject, decoder);
#endif
#if PLATFORM(COCOA)
    case MessageName::RemoteMediaPlayerProxy_NativeImageForCurrentTime:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_NativeImageForCurrentTime>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_ColorSpace:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_ColorSpace>(globalObject, decoder);
#endif
#if (PLATFORM(COCOA) && !HAVE(AVSAMPLEBUFFERDISPLAYLAYER_COPYDISPLAYEDPIXELBUFFER))
    case MessageName::RemoteMediaPlayerProxy_WillBeAskedToPaintGL:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_WillBeAskedToPaintGL>(globalObject, decoder);
#endif
    case MessageName::RemoteMediaPlayerProxy_VideoFrameForCurrentTimeIfChanged:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_VideoFrameForCurrentTimeIfChanged>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_PlayAtHostTime:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_PlayAtHostTime>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_PauseAtHostTime:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_PauseAtHostTime>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_StartVideoFrameMetadataGathering:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_StartVideoFrameMetadataGathering>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_StopVideoFrameMetadataGathering:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_StopVideoFrameMetadataGathering>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_PlayerContentBoxRectChanged:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_PlayerContentBoxRectChanged>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_SetShouldDisableHDR:
        return jsValueForDecodedMessage<MessageName::RemoteMediaPlayerProxy_SetShouldDisableHDR>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
    case MessageName::RemoteMediaResourceManager_ResponseReceived:
        return jsValueForDecodedMessage<MessageName::RemoteMediaResourceManager_ResponseReceived>(globalObject, decoder);
    case MessageName::RemoteMediaResourceManager_RedirectReceived:
        return jsValueForDecodedMessage<MessageName::RemoteMediaResourceManager_RedirectReceived>(globalObject, decoder);
    case MessageName::RemoteMediaResourceManager_DataSent:
        return jsValueForDecodedMessage<MessageName::RemoteMediaResourceManager_DataSent>(globalObject, decoder);
    case MessageName::RemoteMediaResourceManager_DataReceived:
        return jsValueForDecodedMessage<MessageName::RemoteMediaResourceManager_DataReceived>(globalObject, decoder);
    case MessageName::RemoteMediaResourceManager_AccessControlCheckFailed:
        return jsValueForDecodedMessage<MessageName::RemoteMediaResourceManager_AccessControlCheckFailed>(globalObject, decoder);
    case MessageName::RemoteMediaResourceManager_LoadFailed:
        return jsValueForDecodedMessage<MessageName::RemoteMediaResourceManager_LoadFailed>(globalObject, decoder);
    case MessageName::RemoteMediaResourceManager_LoadFinished:
        return jsValueForDecodedMessage<MessageName::RemoteMediaResourceManager_LoadFinished>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
    case MessageName::RemoteMediaSourceProxy_AddSourceBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSourceProxy_AddSourceBuffer>(globalObject, decoder);
    case MessageName::RemoteMediaSourceProxy_DurationChanged:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSourceProxy_DurationChanged>(globalObject, decoder);
    case MessageName::RemoteMediaSourceProxy_BufferedChanged:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSourceProxy_BufferedChanged>(globalObject, decoder);
    case MessageName::RemoteMediaSourceProxy_SetReadyState:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSourceProxy_SetReadyState>(globalObject, decoder);
    case MessageName::RemoteMediaSourceProxy_SetIsSeeking:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSourceProxy_SetIsSeeking>(globalObject, decoder);
    case MessageName::RemoteMediaSourceProxy_WaitForSeekCompleted:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSourceProxy_WaitForSeekCompleted>(globalObject, decoder);
    case MessageName::RemoteMediaSourceProxy_SeekCompleted:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSourceProxy_SeekCompleted>(globalObject, decoder);
    case MessageName::RemoteMediaSourceProxy_SetTimeFudgeFactor:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSourceProxy_SetTimeFudgeFactor>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteRemoteCommandListenerProxy_UpdateSupportedCommands:
        return jsValueForDecodedMessage<MessageName::RemoteRemoteCommandListenerProxy_UpdateSupportedCommands>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
    case MessageName::RemoteSourceBufferProxy_SetActive:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_SetActive>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_CanSwitchToType:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_CanSwitchToType>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_SetMode:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_SetMode>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_Append:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_Append>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_Abort:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_Abort>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_ResetParserState:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_ResetParserState>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_RemovedFromMediaSource:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_RemovedFromMediaSource>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_SetMediaSourceEnded:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_SetMediaSourceEnded>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_SetReadyState:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_SetReadyState>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_StartChangingType:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_StartChangingType>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_UpdateBufferedFromTrackBuffers:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_UpdateBufferedFromTrackBuffers>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_AddTrackBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_AddTrackBuffer>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_ResetTrackBuffers:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_ResetTrackBuffers>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_ClearTrackBuffers:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_ClearTrackBuffers>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_SetAllTrackBuffersNeedRandomAccess:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_SetAllTrackBuffersNeedRandomAccess>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_RemoveCodedFrames:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_RemoveCodedFrames>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_EvictCodedFrames:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_EvictCodedFrames>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_ReenqueueMediaIfNeeded:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_ReenqueueMediaIfNeeded>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_SetGroupStartTimestamp:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_SetGroupStartTimestamp>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_SetGroupStartTimestampToEndTimestamp:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_SetGroupStartTimestampToEndTimestamp>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_SetShouldGenerateTimestamps:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_SetShouldGenerateTimestamps>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_ResetTimestampOffsetInTrackBuffers:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_ResetTimestampOffsetInTrackBuffers>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_SetTimestampOffset:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_SetTimestampOffset>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_SetAppendWindowStart:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_SetAppendWindowStart>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_SetAppendWindowEnd:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_SetAppendWindowEnd>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_SeekToTime:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_SeekToTime>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_UpdateTrackIds:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_UpdateTrackIds>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_BufferedSamplesForTrackId:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_BufferedSamplesForTrackId>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_EnqueuedSamplesForTrackID:
        return jsValueForDecodedMessage<MessageName::RemoteSourceBufferProxy_EnqueuedSamplesForTrackID>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
#if PLATFORM(COCOA)
    case MessageName::RemoteVideoFrameObjectHeap_GetVideoFrameBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteVideoFrameObjectHeap_GetVideoFrameBuffer>(globalObject, decoder);
    case MessageName::RemoteVideoFrameObjectHeap_PixelBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteVideoFrameObjectHeap_PixelBuffer>(globalObject, decoder);
    case MessageName::RemoteVideoFrameObjectHeap_SetSharedVideoFrameSemaphore:
        return jsValueForDecodedMessage<MessageName::RemoteVideoFrameObjectHeap_SetSharedVideoFrameSemaphore>(globalObject, decoder);
    case MessageName::RemoteVideoFrameObjectHeap_SetSharedVideoFrameMemory:
        return jsValueForDecodedMessage<MessageName::RemoteVideoFrameObjectHeap_SetSharedVideoFrameMemory>(globalObject, decoder);
    case MessageName::RemoteVideoFrameObjectHeap_ConvertFrameBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteVideoFrameObjectHeap_ConvertFrameBuffer>(globalObject, decoder);
#endif
    case MessageName::RemoteVideoFrameObjectHeap_ReleaseVideoFrame:
        return jsValueForDecodedMessage<MessageName::RemoteVideoFrameObjectHeap_ReleaseVideoFrame>(globalObject, decoder);
#endif
#if USE(LIBWEBRTC) && PLATFORM(COCOA) && ENABLE(GPU_PROCESS)
    case MessageName::LibWebRTCCodecsProxy_CreateDecoder:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecsProxy_CreateDecoder>(globalObject, decoder);
    case MessageName::LibWebRTCCodecsProxy_ReleaseDecoder:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecsProxy_ReleaseDecoder>(globalObject, decoder);
    case MessageName::LibWebRTCCodecsProxy_FlushDecoder:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecsProxy_FlushDecoder>(globalObject, decoder);
    case MessageName::LibWebRTCCodecsProxy_SetDecoderFormatDescription:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecsProxy_SetDecoderFormatDescription>(globalObject, decoder);
    case MessageName::LibWebRTCCodecsProxy_DecodeFrame:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecsProxy_DecodeFrame>(globalObject, decoder);
    case MessageName::LibWebRTCCodecsProxy_SetFrameSize:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecsProxy_SetFrameSize>(globalObject, decoder);
    case MessageName::LibWebRTCCodecsProxy_CreateEncoder:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecsProxy_CreateEncoder>(globalObject, decoder);
    case MessageName::LibWebRTCCodecsProxy_ReleaseEncoder:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecsProxy_ReleaseEncoder>(globalObject, decoder);
    case MessageName::LibWebRTCCodecsProxy_InitializeEncoder:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecsProxy_InitializeEncoder>(globalObject, decoder);
    case MessageName::LibWebRTCCodecsProxy_EncodeFrame:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecsProxy_EncodeFrame>(globalObject, decoder);
    case MessageName::LibWebRTCCodecsProxy_FlushEncoder:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecsProxy_FlushEncoder>(globalObject, decoder);
    case MessageName::LibWebRTCCodecsProxy_SetEncodeRates:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecsProxy_SetEncodeRates>(globalObject, decoder);
    case MessageName::LibWebRTCCodecsProxy_SetSharedVideoFrameSemaphore:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecsProxy_SetSharedVideoFrameSemaphore>(globalObject, decoder);
    case MessageName::LibWebRTCCodecsProxy_SetSharedVideoFrameMemory:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecsProxy_SetSharedVideoFrameMemory>(globalObject, decoder);
    case MessageName::LibWebRTCCodecsProxy_SetRTCLoggingLevel:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecsProxy_SetRTCLoggingLevel>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM) && PLATFORM(COCOA)
    case MessageName::RemoteAudioMediaStreamTrackRendererInternalUnitManager_CreateUnit:
        return jsValueForDecodedMessage<MessageName::RemoteAudioMediaStreamTrackRendererInternalUnitManager_CreateUnit>(globalObject, decoder);
    case MessageName::RemoteAudioMediaStreamTrackRendererInternalUnitManager_DeleteUnit:
        return jsValueForDecodedMessage<MessageName::RemoteAudioMediaStreamTrackRendererInternalUnitManager_DeleteUnit>(globalObject, decoder);
    case MessageName::RemoteAudioMediaStreamTrackRendererInternalUnitManager_StartUnit:
        return jsValueForDecodedMessage<MessageName::RemoteAudioMediaStreamTrackRendererInternalUnitManager_StartUnit>(globalObject, decoder);
    case MessageName::RemoteAudioMediaStreamTrackRendererInternalUnitManager_StopUnit:
        return jsValueForDecodedMessage<MessageName::RemoteAudioMediaStreamTrackRendererInternalUnitManager_StopUnit>(globalObject, decoder);
    case MessageName::RemoteAudioMediaStreamTrackRendererInternalUnitManager_SetAudioOutputDevice:
        return jsValueForDecodedMessage<MessageName::RemoteAudioMediaStreamTrackRendererInternalUnitManager_SetAudioOutputDevice>(globalObject, decoder);
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
    case MessageName::RemoteMediaRecorder_AudioSamplesStorageChanged:
        return jsValueForDecodedMessage<MessageName::RemoteMediaRecorder_AudioSamplesStorageChanged>(globalObject, decoder);
    case MessageName::RemoteMediaRecorder_AudioSamplesAvailable:
        return jsValueForDecodedMessage<MessageName::RemoteMediaRecorder_AudioSamplesAvailable>(globalObject, decoder);
    case MessageName::RemoteMediaRecorder_VideoFrameAvailable:
        return jsValueForDecodedMessage<MessageName::RemoteMediaRecorder_VideoFrameAvailable>(globalObject, decoder);
    case MessageName::RemoteMediaRecorder_FetchData:
        return jsValueForDecodedMessage<MessageName::RemoteMediaRecorder_FetchData>(globalObject, decoder);
    case MessageName::RemoteMediaRecorder_StopRecording:
        return jsValueForDecodedMessage<MessageName::RemoteMediaRecorder_StopRecording>(globalObject, decoder);
    case MessageName::RemoteMediaRecorder_Pause:
        return jsValueForDecodedMessage<MessageName::RemoteMediaRecorder_Pause>(globalObject, decoder);
    case MessageName::RemoteMediaRecorder_Resume:
        return jsValueForDecodedMessage<MessageName::RemoteMediaRecorder_Resume>(globalObject, decoder);
    case MessageName::RemoteMediaRecorder_SetSharedVideoFrameSemaphore:
        return jsValueForDecodedMessage<MessageName::RemoteMediaRecorder_SetSharedVideoFrameSemaphore>(globalObject, decoder);
    case MessageName::RemoteMediaRecorder_SetSharedVideoFrameMemory:
        return jsValueForDecodedMessage<MessageName::RemoteMediaRecorder_SetSharedVideoFrameMemory>(globalObject, decoder);
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
    case MessageName::RemoteMediaRecorderManager_CreateRecorder:
        return jsValueForDecodedMessage<MessageName::RemoteMediaRecorderManager_CreateRecorder>(globalObject, decoder);
    case MessageName::RemoteMediaRecorderManager_ReleaseRecorder:
        return jsValueForDecodedMessage<MessageName::RemoteMediaRecorderManager_ReleaseRecorder>(globalObject, decoder);
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
#if !RELEASE_LOG_DISABLED
    case MessageName::RemoteSampleBufferDisplayLayer_SetLogIdentifier:
        return jsValueForDecodedMessage<MessageName::RemoteSampleBufferDisplayLayer_SetLogIdentifier>(globalObject, decoder);
#endif
    case MessageName::RemoteSampleBufferDisplayLayer_UpdateDisplayMode:
        return jsValueForDecodedMessage<MessageName::RemoteSampleBufferDisplayLayer_UpdateDisplayMode>(globalObject, decoder);
    case MessageName::RemoteSampleBufferDisplayLayer_UpdateAffineTransform:
        return jsValueForDecodedMessage<MessageName::RemoteSampleBufferDisplayLayer_UpdateAffineTransform>(globalObject, decoder);
    case MessageName::RemoteSampleBufferDisplayLayer_UpdateBoundsAndPosition:
        return jsValueForDecodedMessage<MessageName::RemoteSampleBufferDisplayLayer_UpdateBoundsAndPosition>(globalObject, decoder);
    case MessageName::RemoteSampleBufferDisplayLayer_Flush:
        return jsValueForDecodedMessage<MessageName::RemoteSampleBufferDisplayLayer_Flush>(globalObject, decoder);
    case MessageName::RemoteSampleBufferDisplayLayer_FlushAndRemoveImage:
        return jsValueForDecodedMessage<MessageName::RemoteSampleBufferDisplayLayer_FlushAndRemoveImage>(globalObject, decoder);
    case MessageName::RemoteSampleBufferDisplayLayer_EnqueueVideoFrame:
        return jsValueForDecodedMessage<MessageName::RemoteSampleBufferDisplayLayer_EnqueueVideoFrame>(globalObject, decoder);
    case MessageName::RemoteSampleBufferDisplayLayer_ClearVideoFrames:
        return jsValueForDecodedMessage<MessageName::RemoteSampleBufferDisplayLayer_ClearVideoFrames>(globalObject, decoder);
    case MessageName::RemoteSampleBufferDisplayLayer_Play:
        return jsValueForDecodedMessage<MessageName::RemoteSampleBufferDisplayLayer_Play>(globalObject, decoder);
    case MessageName::RemoteSampleBufferDisplayLayer_Pause:
        return jsValueForDecodedMessage<MessageName::RemoteSampleBufferDisplayLayer_Pause>(globalObject, decoder);
    case MessageName::RemoteSampleBufferDisplayLayer_SetSharedVideoFrameSemaphore:
        return jsValueForDecodedMessage<MessageName::RemoteSampleBufferDisplayLayer_SetSharedVideoFrameSemaphore>(globalObject, decoder);
    case MessageName::RemoteSampleBufferDisplayLayer_SetSharedVideoFrameMemory:
        return jsValueForDecodedMessage<MessageName::RemoteSampleBufferDisplayLayer_SetSharedVideoFrameMemory>(globalObject, decoder);
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
    case MessageName::RemoteSampleBufferDisplayLayerManager_CreateLayer:
        return jsValueForDecodedMessage<MessageName::RemoteSampleBufferDisplayLayerManager_CreateLayer>(globalObject, decoder);
    case MessageName::RemoteSampleBufferDisplayLayerManager_ReleaseLayer:
        return jsValueForDecodedMessage<MessageName::RemoteSampleBufferDisplayLayerManager_ReleaseLayer>(globalObject, decoder);
#endif
    case MessageName::WebCookieManager_GetHostnamesWithCookies:
        return jsValueForDecodedMessage<MessageName::WebCookieManager_GetHostnamesWithCookies>(globalObject, decoder);
    case MessageName::WebCookieManager_DeleteCookiesForHostnames:
        return jsValueForDecodedMessage<MessageName::WebCookieManager_DeleteCookiesForHostnames>(globalObject, decoder);
    case MessageName::WebCookieManager_DeleteAllCookies:
        return jsValueForDecodedMessage<MessageName::WebCookieManager_DeleteAllCookies>(globalObject, decoder);
    case MessageName::WebCookieManager_SetCookie:
        return jsValueForDecodedMessage<MessageName::WebCookieManager_SetCookie>(globalObject, decoder);
    case MessageName::WebCookieManager_SetCookies:
        return jsValueForDecodedMessage<MessageName::WebCookieManager_SetCookies>(globalObject, decoder);
    case MessageName::WebCookieManager_GetAllCookies:
        return jsValueForDecodedMessage<MessageName::WebCookieManager_GetAllCookies>(globalObject, decoder);
    case MessageName::WebCookieManager_GetCookies:
        return jsValueForDecodedMessage<MessageName::WebCookieManager_GetCookies>(globalObject, decoder);
    case MessageName::WebCookieManager_DeleteCookie:
        return jsValueForDecodedMessage<MessageName::WebCookieManager_DeleteCookie>(globalObject, decoder);
    case MessageName::WebCookieManager_DeleteAllCookiesModifiedSince:
        return jsValueForDecodedMessage<MessageName::WebCookieManager_DeleteAllCookiesModifiedSince>(globalObject, decoder);
    case MessageName::WebCookieManager_SetHTTPCookieAcceptPolicy:
        return jsValueForDecodedMessage<MessageName::WebCookieManager_SetHTTPCookieAcceptPolicy>(globalObject, decoder);
    case MessageName::WebCookieManager_GetHTTPCookieAcceptPolicy:
        return jsValueForDecodedMessage<MessageName::WebCookieManager_GetHTTPCookieAcceptPolicy>(globalObject, decoder);
    case MessageName::WebCookieManager_StartObservingCookieChanges:
        return jsValueForDecodedMessage<MessageName::WebCookieManager_StartObservingCookieChanges>(globalObject, decoder);
    case MessageName::WebCookieManager_StopObservingCookieChanges:
        return jsValueForDecodedMessage<MessageName::WebCookieManager_StopObservingCookieChanges>(globalObject, decoder);
#if USE(SOUP)
    case MessageName::WebCookieManager_SetCookiePersistentStorage:
        return jsValueForDecodedMessage<MessageName::WebCookieManager_SetCookiePersistentStorage>(globalObject, decoder);
#endif
    case MessageName::NetworkBroadcastChannelRegistry_RegisterChannel:
        return jsValueForDecodedMessage<MessageName::NetworkBroadcastChannelRegistry_RegisterChannel>(globalObject, decoder);
    case MessageName::NetworkBroadcastChannelRegistry_UnregisterChannel:
        return jsValueForDecodedMessage<MessageName::NetworkBroadcastChannelRegistry_UnregisterChannel>(globalObject, decoder);
    case MessageName::NetworkBroadcastChannelRegistry_PostMessage:
        return jsValueForDecodedMessage<MessageName::NetworkBroadcastChannelRegistry_PostMessage>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_ScheduleResourceLoad:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_ScheduleResourceLoad>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_PerformSynchronousLoad:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_PerformSynchronousLoad>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_TestProcessIncomingSyncMessagesWhenWaitingForSyncReply:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_TestProcessIncomingSyncMessagesWhenWaitingForSyncReply>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_LoadPing:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_LoadPing>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_RemoveLoadIdentifier:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_RemoveLoadIdentifier>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_PageLoadCompleted:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_PageLoadCompleted>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_BrowsingContextRemoved:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_BrowsingContextRemoved>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_PrefetchDNS:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_PrefetchDNS>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_SendH2Ping:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_SendH2Ping>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_PreconnectTo:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_PreconnectTo>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_IsResourceLoadFinished:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_IsResourceLoadFinished>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_StartDownload:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_StartDownload>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_ConvertMainResourceLoadToDownload:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_ConvertMainResourceLoadToDownload>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_CookiesForDOM:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_CookiesForDOM>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_SetCookiesFromDOM:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_SetCookiesFromDOM>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_CookieRequestHeaderFieldValue:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_CookieRequestHeaderFieldValue>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_GetRawCookies:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_GetRawCookies>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_SetRawCookie:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_SetRawCookie>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_DeleteCookie:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_DeleteCookie>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_DomCookiesForHost:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_DomCookiesForHost>(globalObject, decoder);
#if HAVE(COOKIE_CHANGE_LISTENER_API)
    case MessageName::NetworkConnectionToWebProcess_UnsubscribeFromCookieChangeNotifications:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_UnsubscribeFromCookieChangeNotifications>(globalObject, decoder);
#endif
    case MessageName::NetworkConnectionToWebProcess_RegisterFileBlobURL:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_RegisterFileBlobURL>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_RegisterBlobURL:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_RegisterBlobURL>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_RegisterBlobURLFromURL:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_RegisterBlobURLFromURL>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_RegisterBlobURLOptionallyFileBacked:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_RegisterBlobURLOptionallyFileBacked>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_RegisterBlobURLForSlice:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_RegisterBlobURLForSlice>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_UnregisterBlobURL:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_UnregisterBlobURL>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_BlobSize:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_BlobSize>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_WriteBlobsToTemporaryFilesForIndexedDB:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_WriteBlobsToTemporaryFilesForIndexedDB>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_RegisterBlobURLHandle:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_RegisterBlobURLHandle>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_UnregisterBlobURLHandle:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_UnregisterBlobURLHandle>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_SetCaptureExtraNetworkLoadMetricsEnabled:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_SetCaptureExtraNetworkLoadMetricsEnabled>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_CreateSocketStream:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_CreateSocketStream>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_CreateSocketChannel:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_CreateSocketChannel>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_ClearPageSpecificData:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_ClearPageSpecificData>(globalObject, decoder);
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::NetworkConnectionToWebProcess_RemoveStorageAccessForFrame:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_RemoveStorageAccessForFrame>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_LogUserInteraction:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_LogUserInteraction>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_ResourceLoadStatisticsUpdated:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_ResourceLoadStatisticsUpdated>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_HasStorageAccess:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_HasStorageAccess>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_RequestStorageAccess:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_RequestStorageAccess>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_RequestStorageAccessUnderOpener:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_RequestStorageAccessUnderOpener>(globalObject, decoder);
#endif
    case MessageName::NetworkConnectionToWebProcess_AddOriginAccessAllowListEntry:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_AddOriginAccessAllowListEntry>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_RemoveOriginAccessAllowListEntry:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_RemoveOriginAccessAllowListEntry>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_ResetOriginAccessAllowLists:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_ResetOriginAccessAllowLists>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_GetNetworkLoadInformationResponse:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_GetNetworkLoadInformationResponse>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_GetNetworkLoadIntermediateInformation:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_GetNetworkLoadIntermediateInformation>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_TakeNetworkLoadInformationMetrics:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_TakeNetworkLoadInformationMetrics>(globalObject, decoder);
#if ENABLE(SERVICE_WORKER)
    case MessageName::NetworkConnectionToWebProcess_EstablishSWContextConnection:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_EstablishSWContextConnection>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_CloseSWContextConnection:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_CloseSWContextConnection>(globalObject, decoder);
#endif
    case MessageName::NetworkConnectionToWebProcess_EstablishSharedWorkerContextConnection:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_EstablishSharedWorkerContextConnection>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_CloseSharedWorkerContextConnection:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_CloseSharedWorkerContextConnection>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_CreateRTCProvider:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_CreateRTCProvider>(globalObject, decoder);
#if ENABLE(WEB_RTC)
    case MessageName::NetworkConnectionToWebProcess_ConnectToRTCDataChannelRemoteSource:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_ConnectToRTCDataChannelRemoteSource>(globalObject, decoder);
#endif
    case MessageName::NetworkConnectionToWebProcess_UpdateQuotaBasedOnSpaceUsageForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_UpdateQuotaBasedOnSpaceUsageForTesting>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_CreateNewMessagePortChannel:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_CreateNewMessagePortChannel>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_EntangleLocalPortInThisProcessToRemote:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_EntangleLocalPortInThisProcessToRemote>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_MessagePortDisentangled:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_MessagePortDisentangled>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_MessagePortClosed:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_MessagePortClosed>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_TakeAllMessagesForPort:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_TakeAllMessagesForPort>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_PostMessageToRemote:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_PostMessageToRemote>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_DidDeliverMessagePortMessages:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_DidDeliverMessagePortMessages>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_RegisterURLSchemesAsCORSEnabled:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_RegisterURLSchemesAsCORSEnabled>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_SetCORSDisablingPatterns:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_SetCORSDisablingPatterns>(globalObject, decoder);
#if PLATFORM(MAC)
    case MessageName::NetworkConnectionToWebProcess_GetProcessDisplayName:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_GetProcessDisplayName>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_UpdateActivePages:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_UpdateActivePages>(globalObject, decoder);
#endif
    case MessageName::NetworkConnectionToWebProcess_SetResourceLoadSchedulingMode:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_SetResourceLoadSchedulingMode>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_PrioritizeResourceLoads:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_PrioritizeResourceLoads>(globalObject, decoder);
#if ENABLE(CONTENT_FILTERING_IN_NETWORKING_PROCESS)
    case MessageName::NetworkConnectionToWebProcess_InstallMockContentFilter:
        return jsValueForDecodedMessage<MessageName::NetworkConnectionToWebProcess_InstallMockContentFilter>(globalObject, decoder);
#endif
#if ENABLE(CONTENT_EXTENSIONS)
    case MessageName::NetworkContentRuleListManager_Remove:
        return jsValueForDecodedMessage<MessageName::NetworkContentRuleListManager_Remove>(globalObject, decoder);
    case MessageName::NetworkContentRuleListManager_AddContentRuleLists:
        return jsValueForDecodedMessage<MessageName::NetworkContentRuleListManager_AddContentRuleLists>(globalObject, decoder);
    case MessageName::NetworkContentRuleListManager_RemoveContentRuleList:
        return jsValueForDecodedMessage<MessageName::NetworkContentRuleListManager_RemoveContentRuleList>(globalObject, decoder);
    case MessageName::NetworkContentRuleListManager_RemoveAllContentRuleLists:
        return jsValueForDecodedMessage<MessageName::NetworkContentRuleListManager_RemoveAllContentRuleLists>(globalObject, decoder);
#endif
    case MessageName::NetworkProcess_InitializeNetworkProcess:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_InitializeNetworkProcess>(globalObject, decoder);
    case MessageName::NetworkProcess_CreateNetworkConnectionToWebProcess:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_CreateNetworkConnectionToWebProcess>(globalObject, decoder);
    case MessageName::NetworkProcess_AddAllowedFirstPartyForCookies:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_AddAllowedFirstPartyForCookies>(globalObject, decoder);
#if USE(SOUP)
    case MessageName::NetworkProcess_SetIgnoreTLSErrors:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetIgnoreTLSErrors>(globalObject, decoder);
    case MessageName::NetworkProcess_UserPreferredLanguagesChanged:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_UserPreferredLanguagesChanged>(globalObject, decoder);
    case MessageName::NetworkProcess_SetNetworkProxySettings:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetNetworkProxySettings>(globalObject, decoder);
    case MessageName::NetworkProcess_PrefetchDNS:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_PrefetchDNS>(globalObject, decoder);
    case MessageName::NetworkProcess_SetPersistentCredentialStorageEnabled:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetPersistentCredentialStorageEnabled>(globalObject, decoder);
#endif
#if USE(CURL)
    case MessageName::NetworkProcess_SetNetworkProxySettings:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetNetworkProxySettings>(globalObject, decoder);
#endif
    case MessageName::NetworkProcess_AddWebsiteDataStore:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_AddWebsiteDataStore>(globalObject, decoder);
    case MessageName::NetworkProcess_DestroySession:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_DestroySession>(globalObject, decoder);
    case MessageName::NetworkProcess_FetchWebsiteData:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_FetchWebsiteData>(globalObject, decoder);
    case MessageName::NetworkProcess_DeleteWebsiteData:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_DeleteWebsiteData>(globalObject, decoder);
    case MessageName::NetworkProcess_DeleteWebsiteDataForOrigins:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_DeleteWebsiteDataForOrigins>(globalObject, decoder);
    case MessageName::NetworkProcess_RenameOriginInWebsiteData:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_RenameOriginInWebsiteData>(globalObject, decoder);
    case MessageName::NetworkProcess_WebsiteDataOriginDirectoryForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_WebsiteDataOriginDirectoryForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_DownloadRequest:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_DownloadRequest>(globalObject, decoder);
    case MessageName::NetworkProcess_ResumeDownload:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ResumeDownload>(globalObject, decoder);
    case MessageName::NetworkProcess_CancelDownload:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_CancelDownload>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::NetworkProcess_PublishDownloadProgress:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_PublishDownloadProgress>(globalObject, decoder);
#endif
    case MessageName::NetworkProcess_DataTaskWithRequest:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_DataTaskWithRequest>(globalObject, decoder);
    case MessageName::NetworkProcess_CancelDataTask:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_CancelDataTask>(globalObject, decoder);
    case MessageName::NetworkProcess_ApplicationDidEnterBackground:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ApplicationDidEnterBackground>(globalObject, decoder);
    case MessageName::NetworkProcess_ApplicationWillEnterForeground:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ApplicationWillEnterForeground>(globalObject, decoder);
    case MessageName::NetworkProcess_ContinueWillSendRequest:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ContinueWillSendRequest>(globalObject, decoder);
    case MessageName::NetworkProcess_SetAllowsAnySSLCertificateForWebSocket:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetAllowsAnySSLCertificateForWebSocket>(globalObject, decoder);
    case MessageName::NetworkProcess_FlushCookies:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_FlushCookies>(globalObject, decoder);
    case MessageName::NetworkProcess_AllowSpecificHTTPSCertificateForHost:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_AllowSpecificHTTPSCertificateForHost>(globalObject, decoder);
    case MessageName::NetworkProcess_AllowTLSCertificateChainForLocalPCMTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_AllowTLSCertificateChainForLocalPCMTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetCacheModel:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetCacheModel>(globalObject, decoder);
    case MessageName::NetworkProcess_SetCacheModelSynchronouslyForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetCacheModelSynchronouslyForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_ProcessWillSuspendImminentlyForTestingSync:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ProcessWillSuspendImminentlyForTestingSync>(globalObject, decoder);
    case MessageName::NetworkProcess_PrepareToSuspend:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_PrepareToSuspend>(globalObject, decoder);
    case MessageName::NetworkProcess_ProcessDidResume:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ProcessDidResume>(globalObject, decoder);
    case MessageName::NetworkProcess_NotifyMediaStreamingActivity:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_NotifyMediaStreamingActivity>(globalObject, decoder);
    case MessageName::NetworkProcess_PreconnectTo:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_PreconnectTo>(globalObject, decoder);
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::NetworkProcess_ClearPrevalentResource:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ClearPrevalentResource>(globalObject, decoder);
    case MessageName::NetworkProcess_ClearUserInteraction:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ClearUserInteraction>(globalObject, decoder);
    case MessageName::NetworkProcess_DumpResourceLoadStatistics:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_DumpResourceLoadStatistics>(globalObject, decoder);
    case MessageName::NetworkProcess_SetTrackingPreventionEnabled:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetTrackingPreventionEnabled>(globalObject, decoder);
    case MessageName::NetworkProcess_SetResourceLoadStatisticsLogTestingEvent:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetResourceLoadStatisticsLogTestingEvent>(globalObject, decoder);
    case MessageName::NetworkProcess_UpdatePrevalentDomainsToBlockCookiesFor:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_UpdatePrevalentDomainsToBlockCookiesFor>(globalObject, decoder);
    case MessageName::NetworkProcess_IsGrandfathered:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_IsGrandfathered>(globalObject, decoder);
    case MessageName::NetworkProcess_IsPrevalentResource:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_IsPrevalentResource>(globalObject, decoder);
    case MessageName::NetworkProcess_IsVeryPrevalentResource:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_IsVeryPrevalentResource>(globalObject, decoder);
    case MessageName::NetworkProcess_SetLastSeen:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetLastSeen>(globalObject, decoder);
    case MessageName::NetworkProcess_MergeStatisticForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_MergeStatisticForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_InsertExpiredStatisticForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_InsertExpiredStatisticForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetPrevalentResource:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetPrevalentResource>(globalObject, decoder);
    case MessageName::NetworkProcess_SetPrevalentResourceForDebugMode:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetPrevalentResourceForDebugMode>(globalObject, decoder);
    case MessageName::NetworkProcess_IsResourceLoadStatisticsEphemeral:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_IsResourceLoadStatisticsEphemeral>(globalObject, decoder);
    case MessageName::NetworkProcess_HadUserInteraction:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_HadUserInteraction>(globalObject, decoder);
    case MessageName::NetworkProcess_IsRelationshipOnlyInDatabaseOnce:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_IsRelationshipOnlyInDatabaseOnce>(globalObject, decoder);
    case MessageName::NetworkProcess_HasLocalStorage:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_HasLocalStorage>(globalObject, decoder);
    case MessageName::NetworkProcess_GetAllStorageAccessEntries:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_GetAllStorageAccessEntries>(globalObject, decoder);
    case MessageName::NetworkProcess_IsRegisteredAsRedirectingTo:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_IsRegisteredAsRedirectingTo>(globalObject, decoder);
    case MessageName::NetworkProcess_IsRegisteredAsSubFrameUnder:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_IsRegisteredAsSubFrameUnder>(globalObject, decoder);
    case MessageName::NetworkProcess_IsRegisteredAsSubresourceUnder:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_IsRegisteredAsSubresourceUnder>(globalObject, decoder);
    case MessageName::NetworkProcess_DomainIDExistsInDatabase:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_DomainIDExistsInDatabase>(globalObject, decoder);
    case MessageName::NetworkProcess_LogFrameNavigation:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_LogFrameNavigation>(globalObject, decoder);
    case MessageName::NetworkProcess_LogUserInteraction:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_LogUserInteraction>(globalObject, decoder);
    case MessageName::NetworkProcess_ResetParametersToDefaultValues:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ResetParametersToDefaultValues>(globalObject, decoder);
    case MessageName::NetworkProcess_ScheduleClearInMemoryAndPersistent:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ScheduleClearInMemoryAndPersistent>(globalObject, decoder);
    case MessageName::NetworkProcess_ScheduleCookieBlockingUpdate:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ScheduleCookieBlockingUpdate>(globalObject, decoder);
    case MessageName::NetworkProcess_ScheduleStatisticsAndDataRecordsProcessing:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ScheduleStatisticsAndDataRecordsProcessing>(globalObject, decoder);
    case MessageName::NetworkProcess_StatisticsDatabaseHasAllTables:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_StatisticsDatabaseHasAllTables>(globalObject, decoder);
    case MessageName::NetworkProcess_SetCacheMaxAgeCapForPrevalentResources:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetCacheMaxAgeCapForPrevalentResources>(globalObject, decoder);
    case MessageName::NetworkProcess_SetGrandfathered:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetGrandfathered>(globalObject, decoder);
    case MessageName::NetworkProcess_GetResourceLoadStatisticsDataSummary:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_GetResourceLoadStatisticsDataSummary>(globalObject, decoder);
    case MessageName::NetworkProcess_SetGrandfatheringTime:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetGrandfatheringTime>(globalObject, decoder);
    case MessageName::NetworkProcess_SetMaxStatisticsEntries:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetMaxStatisticsEntries>(globalObject, decoder);
    case MessageName::NetworkProcess_SetMinimumTimeBetweenDataRecordsRemoval:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetMinimumTimeBetweenDataRecordsRemoval>(globalObject, decoder);
    case MessageName::NetworkProcess_SetPruneEntriesDownTo:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetPruneEntriesDownTo>(globalObject, decoder);
    case MessageName::NetworkProcess_SetShouldClassifyResourcesBeforeDataRecordsRemoval:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetShouldClassifyResourcesBeforeDataRecordsRemoval>(globalObject, decoder);
    case MessageName::NetworkProcess_SetNotifyPagesWhenDataRecordsWereScanned:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetNotifyPagesWhenDataRecordsWereScanned>(globalObject, decoder);
    case MessageName::NetworkProcess_SetResourceLoadStatisticsTimeAdvanceForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetResourceLoadStatisticsTimeAdvanceForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetIsRunningResourceLoadStatisticsTest:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetIsRunningResourceLoadStatisticsTest>(globalObject, decoder);
    case MessageName::NetworkProcess_SetResourceLoadStatisticsDebugMode:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetResourceLoadStatisticsDebugMode>(globalObject, decoder);
    case MessageName::NetworkProcess_SetVeryPrevalentResource:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetVeryPrevalentResource>(globalObject, decoder);
    case MessageName::NetworkProcess_SetSubframeUnderTopFrameDomain:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetSubframeUnderTopFrameDomain>(globalObject, decoder);
    case MessageName::NetworkProcess_SetSubresourceUnderTopFrameDomain:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetSubresourceUnderTopFrameDomain>(globalObject, decoder);
    case MessageName::NetworkProcess_SetSubresourceUniqueRedirectTo:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetSubresourceUniqueRedirectTo>(globalObject, decoder);
    case MessageName::NetworkProcess_SetSubresourceUniqueRedirectFrom:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetSubresourceUniqueRedirectFrom>(globalObject, decoder);
    case MessageName::NetworkProcess_SetTimeToLiveUserInteraction:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetTimeToLiveUserInteraction>(globalObject, decoder);
    case MessageName::NetworkProcess_SetTopFrameUniqueRedirectTo:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetTopFrameUniqueRedirectTo>(globalObject, decoder);
    case MessageName::NetworkProcess_SetTopFrameUniqueRedirectFrom:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetTopFrameUniqueRedirectFrom>(globalObject, decoder);
    case MessageName::NetworkProcess_ResetCacheMaxAgeCapForPrevalentResources:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ResetCacheMaxAgeCapForPrevalentResources>(globalObject, decoder);
    case MessageName::NetworkProcess_DidCommitCrossSiteLoadWithDataTransfer:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_DidCommitCrossSiteLoadWithDataTransfer>(globalObject, decoder);
    case MessageName::NetworkProcess_SetCrossSiteLoadWithLinkDecorationForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetCrossSiteLoadWithLinkDecorationForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_ResetCrossSiteLoadsWithLinkDecorationForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ResetCrossSiteLoadsWithLinkDecorationForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_DeleteCookiesForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_DeleteCookiesForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_HasIsolatedSession:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_HasIsolatedSession>(globalObject, decoder);
    case MessageName::NetworkProcess_CloseITPDatabase:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_CloseITPDatabase>(globalObject, decoder);
#endif
#if (ENABLE(TRACKING_PREVENTION) && ENABLE(APP_BOUND_DOMAINS))
    case MessageName::NetworkProcess_SetAppBoundDomainsForResourceLoadStatistics:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetAppBoundDomainsForResourceLoadStatistics>(globalObject, decoder);
#endif
#if (ENABLE(TRACKING_PREVENTION) && ENABLE(MANAGED_DOMAINS))
    case MessageName::NetworkProcess_SetManagedDomainsForResourceLoadStatistics:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetManagedDomainsForResourceLoadStatistics>(globalObject, decoder);
#endif
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::NetworkProcess_SetShouldDowngradeReferrerForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetShouldDowngradeReferrerForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetThirdPartyCookieBlockingMode:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetThirdPartyCookieBlockingMode>(globalObject, decoder);
    case MessageName::NetworkProcess_SetShouldEnbleSameSiteStrictEnforcementForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetShouldEnbleSameSiteStrictEnforcementForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetFirstPartyWebsiteDataRemovalModeForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetFirstPartyWebsiteDataRemovalModeForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetToSameSiteStrictCookiesForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetToSameSiteStrictCookiesForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetFirstPartyHostCNAMEDomainForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetFirstPartyHostCNAMEDomainForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetThirdPartyCNAMEDomainForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetThirdPartyCNAMEDomainForTesting>(globalObject, decoder);
#endif
    case MessageName::NetworkProcess_SetPrivateClickMeasurementDebugMode:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetPrivateClickMeasurementDebugMode>(globalObject, decoder);
    case MessageName::NetworkProcess_SetSessionIsControlledByAutomation:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetSessionIsControlledByAutomation>(globalObject, decoder);
    case MessageName::NetworkProcess_RegisterURLSchemeAsSecure:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_RegisterURLSchemeAsSecure>(globalObject, decoder);
    case MessageName::NetworkProcess_RegisterURLSchemeAsBypassingContentSecurityPolicy:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_RegisterURLSchemeAsBypassingContentSecurityPolicy>(globalObject, decoder);
    case MessageName::NetworkProcess_RegisterURLSchemeAsLocal:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_RegisterURLSchemeAsLocal>(globalObject, decoder);
    case MessageName::NetworkProcess_RegisterURLSchemeAsNoAccess:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_RegisterURLSchemeAsNoAccess>(globalObject, decoder);
    case MessageName::NetworkProcess_SyncLocalStorage:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SyncLocalStorage>(globalObject, decoder);
    case MessageName::NetworkProcess_StorePrivateClickMeasurement:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_StorePrivateClickMeasurement>(globalObject, decoder);
    case MessageName::NetworkProcess_DumpPrivateClickMeasurement:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_DumpPrivateClickMeasurement>(globalObject, decoder);
    case MessageName::NetworkProcess_ClearPrivateClickMeasurement:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ClearPrivateClickMeasurement>(globalObject, decoder);
    case MessageName::NetworkProcess_SetPrivateClickMeasurementOverrideTimerForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetPrivateClickMeasurementOverrideTimerForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_MarkAttributedPrivateClickMeasurementsAsExpiredForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_MarkAttributedPrivateClickMeasurementsAsExpiredForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetPrivateClickMeasurementEphemeralMeasurementForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetPrivateClickMeasurementEphemeralMeasurementForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SimulatePrivateClickMeasurementSessionRestart:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SimulatePrivateClickMeasurementSessionRestart>(globalObject, decoder);
    case MessageName::NetworkProcess_SetPrivateClickMeasurementTokenPublicKeyURLForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetPrivateClickMeasurementTokenPublicKeyURLForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetPrivateClickMeasurementTokenSignatureURLForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetPrivateClickMeasurementTokenSignatureURLForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetPrivateClickMeasurementAttributionReportURLsForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetPrivateClickMeasurementAttributionReportURLsForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_MarkPrivateClickMeasurementsAsExpiredForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_MarkPrivateClickMeasurementsAsExpiredForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetPCMFraudPreventionValuesForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetPCMFraudPreventionValuesForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetPrivateClickMeasurementAppBundleIDForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetPrivateClickMeasurementAppBundleIDForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_ClosePCMDatabase:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ClosePCMDatabase>(globalObject, decoder);
    case MessageName::NetworkProcess_SetServiceWorkerFetchTimeoutForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetServiceWorkerFetchTimeoutForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_ResetServiceWorkerFetchTimeoutForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ResetServiceWorkerFetchTimeoutForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_ResetQuota:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ResetQuota>(globalObject, decoder);
#if PLATFORM(IOS_FAMILY)
    case MessageName::NetworkProcess_SetBackupExclusionPeriodForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetBackupExclusionPeriodForTesting>(globalObject, decoder);
#endif
    case MessageName::NetworkProcess_ResetStoragePersistedState:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ResetStoragePersistedState>(globalObject, decoder);
    case MessageName::NetworkProcess_CloneSessionStorageForWebPage:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_CloneSessionStorageForWebPage>(globalObject, decoder);
    case MessageName::NetworkProcess_DidIncreaseQuota:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_DidIncreaseQuota>(globalObject, decoder);
#if ENABLE(APP_BOUND_DOMAINS)
    case MessageName::NetworkProcess_HasAppBoundSession:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_HasAppBoundSession>(globalObject, decoder);
    case MessageName::NetworkProcess_ClearAppBoundSession:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ClearAppBoundSession>(globalObject, decoder);
#endif
#if PLATFORM(IOS_FAMILY) && !PLATFORM(MACCATALYST)
    case MessageName::NetworkProcess_DisableServiceWorkerEntitlement:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_DisableServiceWorkerEntitlement>(globalObject, decoder);
    case MessageName::NetworkProcess_ClearServiceWorkerEntitlementOverride:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ClearServiceWorkerEntitlementOverride>(globalObject, decoder);
#endif
    case MessageName::NetworkProcess_UpdateBundleIdentifier:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_UpdateBundleIdentifier>(globalObject, decoder);
    case MessageName::NetworkProcess_ClearBundleIdentifier:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ClearBundleIdentifier>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::NetworkProcess_AppPrivacyReportTestingData:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_AppPrivacyReportTestingData>(globalObject, decoder);
    case MessageName::NetworkProcess_ClearAppPrivacyReportTestingData:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ClearAppPrivacyReportTestingData>(globalObject, decoder);
#endif
    case MessageName::NetworkProcess_AddWebPageNetworkParameters:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_AddWebPageNetworkParameters>(globalObject, decoder);
    case MessageName::NetworkProcess_RemoveWebPageNetworkParameters:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_RemoveWebPageNetworkParameters>(globalObject, decoder);
    case MessageName::NetworkProcess_CountNonDefaultSessionSets:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_CountNonDefaultSessionSets>(globalObject, decoder);
#if ENABLE(CFPREFS_DIRECT_MODE)
    case MessageName::NetworkProcess_NotifyPreferencesChanged:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_NotifyPreferencesChanged>(globalObject, decoder);
#endif
    case MessageName::NetworkProcess_TerminateRemoteWorkerContextConnectionWhenPossible:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_TerminateRemoteWorkerContextConnectionWhenPossible>(globalObject, decoder);
#if ENABLE(SERVICE_WORKER)
    case MessageName::NetworkProcess_GetPendingPushMessages:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_GetPendingPushMessages>(globalObject, decoder);
    case MessageName::NetworkProcess_ProcessPushMessage:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ProcessPushMessage>(globalObject, decoder);
    case MessageName::NetworkProcess_ProcessNotificationEvent:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_ProcessNotificationEvent>(globalObject, decoder);
#endif
    case MessageName::NetworkProcess_SetPushAndNotificationsEnabledForOrigin:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetPushAndNotificationsEnabledForOrigin>(globalObject, decoder);
    case MessageName::NetworkProcess_DeletePushAndNotificationRegistration:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_DeletePushAndNotificationRegistration>(globalObject, decoder);
    case MessageName::NetworkProcess_GetOriginsWithPushAndNotificationPermissions:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_GetOriginsWithPushAndNotificationPermissions>(globalObject, decoder);
    case MessageName::NetworkProcess_HasPushSubscriptionForTesting:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_HasPushSubscriptionForTesting>(globalObject, decoder);
#if ENABLE(INSPECTOR_NETWORK_THROTTLING)
    case MessageName::NetworkProcess_SetEmulatedConditions:
        return jsValueForDecodedMessage<MessageName::NetworkProcess_SetEmulatedConditions>(globalObject, decoder);
#endif
    case MessageName::NetworkResourceLoader_ContinueWillSendRequest:
        return jsValueForDecodedMessage<MessageName::NetworkResourceLoader_ContinueWillSendRequest>(globalObject, decoder);
    case MessageName::NetworkResourceLoader_ContinueDidReceiveResponse:
        return jsValueForDecodedMessage<MessageName::NetworkResourceLoader_ContinueDidReceiveResponse>(globalObject, decoder);
    case MessageName::NetworkSocketChannel_SendString:
        return jsValueForDecodedMessage<MessageName::NetworkSocketChannel_SendString>(globalObject, decoder);
    case MessageName::NetworkSocketChannel_SendData:
        return jsValueForDecodedMessage<MessageName::NetworkSocketChannel_SendData>(globalObject, decoder);
    case MessageName::NetworkSocketChannel_Close:
        return jsValueForDecodedMessage<MessageName::NetworkSocketChannel_Close>(globalObject, decoder);
    case MessageName::NetworkSocketStream_SendData:
        return jsValueForDecodedMessage<MessageName::NetworkSocketStream_SendData>(globalObject, decoder);
    case MessageName::NetworkSocketStream_SendHandshake:
        return jsValueForDecodedMessage<MessageName::NetworkSocketStream_SendHandshake>(globalObject, decoder);
    case MessageName::NetworkSocketStream_Close:
        return jsValueForDecodedMessage<MessageName::NetworkSocketStream_Close>(globalObject, decoder);
#if ENABLE(SERVICE_WORKER)
    case MessageName::ServiceWorkerDownloadTask_DidFail:
        return jsValueForDecodedMessage<MessageName::ServiceWorkerDownloadTask_DidFail>(globalObject, decoder);
    case MessageName::ServiceWorkerDownloadTask_DidReceiveData:
        return jsValueForDecodedMessage<MessageName::ServiceWorkerDownloadTask_DidReceiveData>(globalObject, decoder);
    case MessageName::ServiceWorkerDownloadTask_DidReceiveFormData:
        return jsValueForDecodedMessage<MessageName::ServiceWorkerDownloadTask_DidReceiveFormData>(globalObject, decoder);
    case MessageName::ServiceWorkerDownloadTask_DidFinish:
        return jsValueForDecodedMessage<MessageName::ServiceWorkerDownloadTask_DidFinish>(globalObject, decoder);
#endif
#if ENABLE(SERVICE_WORKER)
    case MessageName::ServiceWorkerFetchTask_DidNotHandle:
        return jsValueForDecodedMessage<MessageName::ServiceWorkerFetchTask_DidNotHandle>(globalObject, decoder);
    case MessageName::ServiceWorkerFetchTask_DidFail:
        return jsValueForDecodedMessage<MessageName::ServiceWorkerFetchTask_DidFail>(globalObject, decoder);
    case MessageName::ServiceWorkerFetchTask_DidReceiveRedirectResponse:
        return jsValueForDecodedMessage<MessageName::ServiceWorkerFetchTask_DidReceiveRedirectResponse>(globalObject, decoder);
    case MessageName::ServiceWorkerFetchTask_DidReceiveResponse:
        return jsValueForDecodedMessage<MessageName::ServiceWorkerFetchTask_DidReceiveResponse>(globalObject, decoder);
    case MessageName::ServiceWorkerFetchTask_DidReceiveData:
        return jsValueForDecodedMessage<MessageName::ServiceWorkerFetchTask_DidReceiveData>(globalObject, decoder);
    case MessageName::ServiceWorkerFetchTask_DidReceiveFormData:
        return jsValueForDecodedMessage<MessageName::ServiceWorkerFetchTask_DidReceiveFormData>(globalObject, decoder);
    case MessageName::ServiceWorkerFetchTask_DidFinish:
        return jsValueForDecodedMessage<MessageName::ServiceWorkerFetchTask_DidFinish>(globalObject, decoder);
    case MessageName::ServiceWorkerFetchTask_UsePreload:
        return jsValueForDecodedMessage<MessageName::ServiceWorkerFetchTask_UsePreload>(globalObject, decoder);
#endif
#if ENABLE(SERVICE_WORKER)
    case MessageName::WebSWServerConnection_ScheduleJobInServer:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_ScheduleJobInServer>(globalObject, decoder);
    case MessageName::WebSWServerConnection_ScheduleUnregisterJobInServer:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_ScheduleUnregisterJobInServer>(globalObject, decoder);
    case MessageName::WebSWServerConnection_FinishFetchingScriptInServer:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_FinishFetchingScriptInServer>(globalObject, decoder);
    case MessageName::WebSWServerConnection_AddServiceWorkerRegistrationInServer:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_AddServiceWorkerRegistrationInServer>(globalObject, decoder);
    case MessageName::WebSWServerConnection_RemoveServiceWorkerRegistrationInServer:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_RemoveServiceWorkerRegistrationInServer>(globalObject, decoder);
    case MessageName::WebSWServerConnection_PostMessageToServiceWorker:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_PostMessageToServiceWorker>(globalObject, decoder);
    case MessageName::WebSWServerConnection_DidResolveRegistrationPromise:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_DidResolveRegistrationPromise>(globalObject, decoder);
    case MessageName::WebSWServerConnection_MatchRegistration:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_MatchRegistration>(globalObject, decoder);
    case MessageName::WebSWServerConnection_WhenRegistrationReady:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_WhenRegistrationReady>(globalObject, decoder);
    case MessageName::WebSWServerConnection_GetRegistrations:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_GetRegistrations>(globalObject, decoder);
    case MessageName::WebSWServerConnection_RegisterServiceWorkerClient:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_RegisterServiceWorkerClient>(globalObject, decoder);
    case MessageName::WebSWServerConnection_UnregisterServiceWorkerClient:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_UnregisterServiceWorkerClient>(globalObject, decoder);
    case MessageName::WebSWServerConnection_TerminateWorkerFromClient:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_TerminateWorkerFromClient>(globalObject, decoder);
    case MessageName::WebSWServerConnection_WhenServiceWorkerIsTerminatedForTesting:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_WhenServiceWorkerIsTerminatedForTesting>(globalObject, decoder);
    case MessageName::WebSWServerConnection_SetThrottleState:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_SetThrottleState>(globalObject, decoder);
    case MessageName::WebSWServerConnection_StoreRegistrationsOnDisk:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_StoreRegistrationsOnDisk>(globalObject, decoder);
    case MessageName::WebSWServerConnection_SubscribeToPushService:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_SubscribeToPushService>(globalObject, decoder);
    case MessageName::WebSWServerConnection_UnsubscribeFromPushService:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_UnsubscribeFromPushService>(globalObject, decoder);
    case MessageName::WebSWServerConnection_GetPushSubscription:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_GetPushSubscription>(globalObject, decoder);
    case MessageName::WebSWServerConnection_GetPushPermissionState:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_GetPushPermissionState>(globalObject, decoder);
    case MessageName::WebSWServerConnection_EnableNavigationPreload:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_EnableNavigationPreload>(globalObject, decoder);
    case MessageName::WebSWServerConnection_DisableNavigationPreload:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_DisableNavigationPreload>(globalObject, decoder);
    case MessageName::WebSWServerConnection_SetNavigationPreloadHeaderValue:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_SetNavigationPreloadHeaderValue>(globalObject, decoder);
    case MessageName::WebSWServerConnection_GetNavigationPreloadState:
        return jsValueForDecodedMessage<MessageName::WebSWServerConnection_GetNavigationPreloadState>(globalObject, decoder);
#endif
#if ENABLE(SERVICE_WORKER)
    case MessageName::WebSWServerToContextConnection_ScriptContextFailedToStart:
        return jsValueForDecodedMessage<MessageName::WebSWServerToContextConnection_ScriptContextFailedToStart>(globalObject, decoder);
    case MessageName::WebSWServerToContextConnection_ScriptContextStarted:
        return jsValueForDecodedMessage<MessageName::WebSWServerToContextConnection_ScriptContextStarted>(globalObject, decoder);
    case MessageName::WebSWServerToContextConnection_DidFinishInstall:
        return jsValueForDecodedMessage<MessageName::WebSWServerToContextConnection_DidFinishInstall>(globalObject, decoder);
    case MessageName::WebSWServerToContextConnection_DidFinishActivation:
        return jsValueForDecodedMessage<MessageName::WebSWServerToContextConnection_DidFinishActivation>(globalObject, decoder);
    case MessageName::WebSWServerToContextConnection_SetServiceWorkerHasPendingEvents:
        return jsValueForDecodedMessage<MessageName::WebSWServerToContextConnection_SetServiceWorkerHasPendingEvents>(globalObject, decoder);
    case MessageName::WebSWServerToContextConnection_SkipWaiting:
        return jsValueForDecodedMessage<MessageName::WebSWServerToContextConnection_SkipWaiting>(globalObject, decoder);
    case MessageName::WebSWServerToContextConnection_WorkerTerminated:
        return jsValueForDecodedMessage<MessageName::WebSWServerToContextConnection_WorkerTerminated>(globalObject, decoder);
    case MessageName::WebSWServerToContextConnection_FindClientByVisibleIdentifier:
        return jsValueForDecodedMessage<MessageName::WebSWServerToContextConnection_FindClientByVisibleIdentifier>(globalObject, decoder);
    case MessageName::WebSWServerToContextConnection_MatchAll:
        return jsValueForDecodedMessage<MessageName::WebSWServerToContextConnection_MatchAll>(globalObject, decoder);
    case MessageName::WebSWServerToContextConnection_Claim:
        return jsValueForDecodedMessage<MessageName::WebSWServerToContextConnection_Claim>(globalObject, decoder);
    case MessageName::WebSWServerToContextConnection_Focus:
        return jsValueForDecodedMessage<MessageName::WebSWServerToContextConnection_Focus>(globalObject, decoder);
    case MessageName::WebSWServerToContextConnection_Navigate:
        return jsValueForDecodedMessage<MessageName::WebSWServerToContextConnection_Navigate>(globalObject, decoder);
    case MessageName::WebSWServerToContextConnection_SetScriptResource:
        return jsValueForDecodedMessage<MessageName::WebSWServerToContextConnection_SetScriptResource>(globalObject, decoder);
    case MessageName::WebSWServerToContextConnection_PostMessageToServiceWorkerClient:
        return jsValueForDecodedMessage<MessageName::WebSWServerToContextConnection_PostMessageToServiceWorkerClient>(globalObject, decoder);
    case MessageName::WebSWServerToContextConnection_DidFailHeartBeatCheck:
        return jsValueForDecodedMessage<MessageName::WebSWServerToContextConnection_DidFailHeartBeatCheck>(globalObject, decoder);
    case MessageName::WebSWServerToContextConnection_SetAsInspected:
        return jsValueForDecodedMessage<MessageName::WebSWServerToContextConnection_SetAsInspected>(globalObject, decoder);
    case MessageName::WebSWServerToContextConnection_OpenWindow:
        return jsValueForDecodedMessage<MessageName::WebSWServerToContextConnection_OpenWindow>(globalObject, decoder);
#endif
    case MessageName::WebSharedWorkerServerConnection_RequestSharedWorker:
        return jsValueForDecodedMessage<MessageName::WebSharedWorkerServerConnection_RequestSharedWorker>(globalObject, decoder);
    case MessageName::WebSharedWorkerServerConnection_SharedWorkerObjectIsGoingAway:
        return jsValueForDecodedMessage<MessageName::WebSharedWorkerServerConnection_SharedWorkerObjectIsGoingAway>(globalObject, decoder);
    case MessageName::WebSharedWorkerServerConnection_SuspendForBackForwardCache:
        return jsValueForDecodedMessage<MessageName::WebSharedWorkerServerConnection_SuspendForBackForwardCache>(globalObject, decoder);
    case MessageName::WebSharedWorkerServerConnection_ResumeForBackForwardCache:
        return jsValueForDecodedMessage<MessageName::WebSharedWorkerServerConnection_ResumeForBackForwardCache>(globalObject, decoder);
    case MessageName::WebSharedWorkerServerToContextConnection_PostExceptionToWorkerObject:
        return jsValueForDecodedMessage<MessageName::WebSharedWorkerServerToContextConnection_PostExceptionToWorkerObject>(globalObject, decoder);
    case MessageName::WebSharedWorkerServerToContextConnection_SharedWorkerTerminated:
        return jsValueForDecodedMessage<MessageName::WebSharedWorkerServerToContextConnection_SharedWorkerTerminated>(globalObject, decoder);
    case MessageName::CacheStorageEngineConnection_Reference:
        return jsValueForDecodedMessage<MessageName::CacheStorageEngineConnection_Reference>(globalObject, decoder);
    case MessageName::CacheStorageEngineConnection_Dereference:
        return jsValueForDecodedMessage<MessageName::CacheStorageEngineConnection_Dereference>(globalObject, decoder);
    case MessageName::CacheStorageEngineConnection_Open:
        return jsValueForDecodedMessage<MessageName::CacheStorageEngineConnection_Open>(globalObject, decoder);
    case MessageName::CacheStorageEngineConnection_Remove:
        return jsValueForDecodedMessage<MessageName::CacheStorageEngineConnection_Remove>(globalObject, decoder);
    case MessageName::CacheStorageEngineConnection_Caches:
        return jsValueForDecodedMessage<MessageName::CacheStorageEngineConnection_Caches>(globalObject, decoder);
    case MessageName::CacheStorageEngineConnection_RetrieveRecords:
        return jsValueForDecodedMessage<MessageName::CacheStorageEngineConnection_RetrieveRecords>(globalObject, decoder);
    case MessageName::CacheStorageEngineConnection_DeleteMatchingRecords:
        return jsValueForDecodedMessage<MessageName::CacheStorageEngineConnection_DeleteMatchingRecords>(globalObject, decoder);
    case MessageName::CacheStorageEngineConnection_PutRecords:
        return jsValueForDecodedMessage<MessageName::CacheStorageEngineConnection_PutRecords>(globalObject, decoder);
    case MessageName::CacheStorageEngineConnection_ClearMemoryRepresentation:
        return jsValueForDecodedMessage<MessageName::CacheStorageEngineConnection_ClearMemoryRepresentation>(globalObject, decoder);
    case MessageName::CacheStorageEngineConnection_EngineRepresentation:
        return jsValueForDecodedMessage<MessageName::CacheStorageEngineConnection_EngineRepresentation>(globalObject, decoder);
    case MessageName::NetworkStorageManager_Persisted:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_Persisted>(globalObject, decoder);
    case MessageName::NetworkStorageManager_Persist:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_Persist>(globalObject, decoder);
    case MessageName::NetworkStorageManager_Estimate:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_Estimate>(globalObject, decoder);
    case MessageName::NetworkStorageManager_FileSystemGetDirectory:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_FileSystemGetDirectory>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CloseHandle:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_CloseHandle>(globalObject, decoder);
    case MessageName::NetworkStorageManager_IsSameEntry:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_IsSameEntry>(globalObject, decoder);
    case MessageName::NetworkStorageManager_GetFileHandle:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_GetFileHandle>(globalObject, decoder);
    case MessageName::NetworkStorageManager_GetDirectoryHandle:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_GetDirectoryHandle>(globalObject, decoder);
    case MessageName::NetworkStorageManager_RemoveEntry:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_RemoveEntry>(globalObject, decoder);
    case MessageName::NetworkStorageManager_Resolve:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_Resolve>(globalObject, decoder);
    case MessageName::NetworkStorageManager_Move:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_Move>(globalObject, decoder);
    case MessageName::NetworkStorageManager_GetFile:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_GetFile>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CreateSyncAccessHandle:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_CreateSyncAccessHandle>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CloseSyncAccessHandle:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_CloseSyncAccessHandle>(globalObject, decoder);
    case MessageName::NetworkStorageManager_RequestNewCapacityForSyncAccessHandle:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_RequestNewCapacityForSyncAccessHandle>(globalObject, decoder);
    case MessageName::NetworkStorageManager_GetHandleNames:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_GetHandleNames>(globalObject, decoder);
    case MessageName::NetworkStorageManager_GetHandle:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_GetHandle>(globalObject, decoder);
    case MessageName::NetworkStorageManager_ConnectToStorageArea:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_ConnectToStorageArea>(globalObject, decoder);
    case MessageName::NetworkStorageManager_ConnectToStorageAreaSync:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_ConnectToStorageAreaSync>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CancelConnectToStorageArea:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_CancelConnectToStorageArea>(globalObject, decoder);
    case MessageName::NetworkStorageManager_DisconnectFromStorageArea:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_DisconnectFromStorageArea>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CloneSessionStorageNamespace:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_CloneSessionStorageNamespace>(globalObject, decoder);
    case MessageName::NetworkStorageManager_SetItem:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_SetItem>(globalObject, decoder);
    case MessageName::NetworkStorageManager_RemoveItem:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_RemoveItem>(globalObject, decoder);
    case MessageName::NetworkStorageManager_Clear:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_Clear>(globalObject, decoder);
    case MessageName::NetworkStorageManager_OpenDatabase:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_OpenDatabase>(globalObject, decoder);
    case MessageName::NetworkStorageManager_OpenDBRequestCancelled:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_OpenDBRequestCancelled>(globalObject, decoder);
    case MessageName::NetworkStorageManager_DeleteDatabase:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_DeleteDatabase>(globalObject, decoder);
    case MessageName::NetworkStorageManager_EstablishTransaction:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_EstablishTransaction>(globalObject, decoder);
    case MessageName::NetworkStorageManager_DatabaseConnectionPendingClose:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_DatabaseConnectionPendingClose>(globalObject, decoder);
    case MessageName::NetworkStorageManager_DatabaseConnectionClosed:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_DatabaseConnectionClosed>(globalObject, decoder);
    case MessageName::NetworkStorageManager_AbortOpenAndUpgradeNeeded:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_AbortOpenAndUpgradeNeeded>(globalObject, decoder);
    case MessageName::NetworkStorageManager_DidFireVersionChangeEvent:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_DidFireVersionChangeEvent>(globalObject, decoder);
    case MessageName::NetworkStorageManager_DidFinishHandlingVersionChangeTransaction:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_DidFinishHandlingVersionChangeTransaction>(globalObject, decoder);
    case MessageName::NetworkStorageManager_AbortTransaction:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_AbortTransaction>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CommitTransaction:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_CommitTransaction>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CreateObjectStore:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_CreateObjectStore>(globalObject, decoder);
    case MessageName::NetworkStorageManager_DeleteObjectStore:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_DeleteObjectStore>(globalObject, decoder);
    case MessageName::NetworkStorageManager_RenameObjectStore:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_RenameObjectStore>(globalObject, decoder);
    case MessageName::NetworkStorageManager_ClearObjectStore:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_ClearObjectStore>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CreateIndex:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_CreateIndex>(globalObject, decoder);
    case MessageName::NetworkStorageManager_DeleteIndex:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_DeleteIndex>(globalObject, decoder);
    case MessageName::NetworkStorageManager_RenameIndex:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_RenameIndex>(globalObject, decoder);
    case MessageName::NetworkStorageManager_PutOrAdd:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_PutOrAdd>(globalObject, decoder);
    case MessageName::NetworkStorageManager_GetRecord:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_GetRecord>(globalObject, decoder);
    case MessageName::NetworkStorageManager_GetAllRecords:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_GetAllRecords>(globalObject, decoder);
    case MessageName::NetworkStorageManager_GetCount:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_GetCount>(globalObject, decoder);
    case MessageName::NetworkStorageManager_DeleteRecord:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_DeleteRecord>(globalObject, decoder);
    case MessageName::NetworkStorageManager_OpenCursor:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_OpenCursor>(globalObject, decoder);
    case MessageName::NetworkStorageManager_IterateCursor:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_IterateCursor>(globalObject, decoder);
    case MessageName::NetworkStorageManager_GetAllDatabaseNamesAndVersions:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_GetAllDatabaseNamesAndVersions>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CacheStorageOpenCache:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_CacheStorageOpenCache>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CacheStorageRemoveCache:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_CacheStorageRemoveCache>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CacheStorageAllCaches:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_CacheStorageAllCaches>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CacheStorageReference:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_CacheStorageReference>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CacheStorageDereference:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_CacheStorageDereference>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CacheStorageRetrieveRecords:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_CacheStorageRetrieveRecords>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CacheStorageRemoveRecords:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_CacheStorageRemoveRecords>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CacheStoragePutRecords:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_CacheStoragePutRecords>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CacheStorageClearMemoryRepresentation:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_CacheStorageClearMemoryRepresentation>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CacheStorageRepresentation:
        return jsValueForDecodedMessage<MessageName::NetworkStorageManager_CacheStorageRepresentation>(globalObject, decoder);
#if ENABLE(WEB_RTC)
    case MessageName::NetworkMDNSRegister_UnregisterMDNSNames:
        return jsValueForDecodedMessage<MessageName::NetworkMDNSRegister_UnregisterMDNSNames>(globalObject, decoder);
    case MessageName::NetworkMDNSRegister_RegisterMDNSName:
        return jsValueForDecodedMessage<MessageName::NetworkMDNSRegister_RegisterMDNSName>(globalObject, decoder);
#endif
#if USE(LIBWEBRTC)
    case MessageName::NetworkRTCMonitor_StartUpdatingIfNeeded:
        return jsValueForDecodedMessage<MessageName::NetworkRTCMonitor_StartUpdatingIfNeeded>(globalObject, decoder);
    case MessageName::NetworkRTCMonitor_StopUpdating:
        return jsValueForDecodedMessage<MessageName::NetworkRTCMonitor_StopUpdating>(globalObject, decoder);
#endif
#if USE(LIBWEBRTC)
    case MessageName::NetworkRTCProvider_CreateUDPSocket:
        return jsValueForDecodedMessage<MessageName::NetworkRTCProvider_CreateUDPSocket>(globalObject, decoder);
    case MessageName::NetworkRTCProvider_CreateClientTCPSocket:
        return jsValueForDecodedMessage<MessageName::NetworkRTCProvider_CreateClientTCPSocket>(globalObject, decoder);
    case MessageName::NetworkRTCProvider_WrapNewTCPConnection:
        return jsValueForDecodedMessage<MessageName::NetworkRTCProvider_WrapNewTCPConnection>(globalObject, decoder);
    case MessageName::NetworkRTCProvider_SetPlatformTCPSocketsEnabled:
        return jsValueForDecodedMessage<MessageName::NetworkRTCProvider_SetPlatformTCPSocketsEnabled>(globalObject, decoder);
    case MessageName::NetworkRTCProvider_SetPlatformUDPSocketsEnabled:
        return jsValueForDecodedMessage<MessageName::NetworkRTCProvider_SetPlatformUDPSocketsEnabled>(globalObject, decoder);
    case MessageName::NetworkRTCProvider_CreateResolver:
        return jsValueForDecodedMessage<MessageName::NetworkRTCProvider_CreateResolver>(globalObject, decoder);
    case MessageName::NetworkRTCProvider_StopResolver:
        return jsValueForDecodedMessage<MessageName::NetworkRTCProvider_StopResolver>(globalObject, decoder);
    case MessageName::NetworkRTCProvider_SendToSocket:
        return jsValueForDecodedMessage<MessageName::NetworkRTCProvider_SendToSocket>(globalObject, decoder);
    case MessageName::NetworkRTCProvider_CloseSocket:
        return jsValueForDecodedMessage<MessageName::NetworkRTCProvider_CloseSocket>(globalObject, decoder);
    case MessageName::NetworkRTCProvider_SetSocketOption:
        return jsValueForDecodedMessage<MessageName::NetworkRTCProvider_SetSocketOption>(globalObject, decoder);
#endif
#if ENABLE(WEB_RTC)
    case MessageName::RTCDataChannelRemoteManagerProxy_SendData:
        return jsValueForDecodedMessage<MessageName::RTCDataChannelRemoteManagerProxy_SendData>(globalObject, decoder);
    case MessageName::RTCDataChannelRemoteManagerProxy_Close:
        return jsValueForDecodedMessage<MessageName::RTCDataChannelRemoteManagerProxy_Close>(globalObject, decoder);
    case MessageName::RTCDataChannelRemoteManagerProxy_ChangeReadyState:
        return jsValueForDecodedMessage<MessageName::RTCDataChannelRemoteManagerProxy_ChangeReadyState>(globalObject, decoder);
    case MessageName::RTCDataChannelRemoteManagerProxy_ReceiveData:
        return jsValueForDecodedMessage<MessageName::RTCDataChannelRemoteManagerProxy_ReceiveData>(globalObject, decoder);
    case MessageName::RTCDataChannelRemoteManagerProxy_DetectError:
        return jsValueForDecodedMessage<MessageName::RTCDataChannelRemoteManagerProxy_DetectError>(globalObject, decoder);
    case MessageName::RTCDataChannelRemoteManagerProxy_BufferedAmountIsDecreasing:
        return jsValueForDecodedMessage<MessageName::RTCDataChannelRemoteManagerProxy_BufferedAmountIsDecreasing>(globalObject, decoder);
#endif
    case MessageName::AuxiliaryProcess_ShutDown:
        return jsValueForDecodedMessage<MessageName::AuxiliaryProcess_ShutDown>(globalObject, decoder);
    case MessageName::AuxiliaryProcess_SetProcessSuppressionEnabled:
        return jsValueForDecodedMessage<MessageName::AuxiliaryProcess_SetProcessSuppressionEnabled>(globalObject, decoder);
    case MessageName::AuxiliaryProcess_MainThreadPing:
        return jsValueForDecodedMessage<MessageName::AuxiliaryProcess_MainThreadPing>(globalObject, decoder);
#if OS(LINUX)
    case MessageName::AuxiliaryProcess_DidReceiveMemoryPressureEvent:
        return jsValueForDecodedMessage<MessageName::AuxiliaryProcess_DidReceiveMemoryPressureEvent>(globalObject, decoder);
#endif
#if ENABLE(IPC_TESTING_API)
    case MessageName::IPCConnectionTester_AsyncMessage:
        return jsValueForDecodedMessage<MessageName::IPCConnectionTester_AsyncMessage>(globalObject, decoder);
    case MessageName::IPCConnectionTester_SyncMessage:
        return jsValueForDecodedMessage<MessageName::IPCConnectionTester_SyncMessage>(globalObject, decoder);
#endif
#if ENABLE(IPC_TESTING_API)
    case MessageName::IPCStreamTester_SyncMessageReturningSharedMemory1:
        return jsValueForDecodedMessage<MessageName::IPCStreamTester_SyncMessageReturningSharedMemory1>(globalObject, decoder);
    case MessageName::IPCStreamTester_SyncCrashOnZero:
        return jsValueForDecodedMessage<MessageName::IPCStreamTester_SyncCrashOnZero>(globalObject, decoder);
    case MessageName::IPCStreamTester_CheckAutoreleasePool:
        return jsValueForDecodedMessage<MessageName::IPCStreamTester_CheckAutoreleasePool>(globalObject, decoder);
    case MessageName::IPCStreamTester_AsyncMessage:
        return jsValueForDecodedMessage<MessageName::IPCStreamTester_AsyncMessage>(globalObject, decoder);
#endif
#if ENABLE(IPC_TESTING_API)
    case MessageName::IPCStreamTesterProxy_WasCreated:
        return jsValueForDecodedMessage<MessageName::IPCStreamTesterProxy_WasCreated>(globalObject, decoder);
#endif
#if ENABLE(IPC_TESTING_API)
    case MessageName::IPCTester_StartMessageTesting:
        return jsValueForDecodedMessage<MessageName::IPCTester_StartMessageTesting>(globalObject, decoder);
    case MessageName::IPCTester_StopMessageTesting:
        return jsValueForDecodedMessage<MessageName::IPCTester_StopMessageTesting>(globalObject, decoder);
    case MessageName::IPCTester_CreateStreamTester:
        return jsValueForDecodedMessage<MessageName::IPCTester_CreateStreamTester>(globalObject, decoder);
    case MessageName::IPCTester_ReleaseStreamTester:
        return jsValueForDecodedMessage<MessageName::IPCTester_ReleaseStreamTester>(globalObject, decoder);
    case MessageName::IPCTester_CreateConnectionTester:
        return jsValueForDecodedMessage<MessageName::IPCTester_CreateConnectionTester>(globalObject, decoder);
    case MessageName::IPCTester_CreateConnectionTesterAndSendAsyncMessages:
        return jsValueForDecodedMessage<MessageName::IPCTester_CreateConnectionTesterAndSendAsyncMessages>(globalObject, decoder);
    case MessageName::IPCTester_ReleaseConnectionTester:
        return jsValueForDecodedMessage<MessageName::IPCTester_ReleaseConnectionTester>(globalObject, decoder);
    case MessageName::IPCTester_SendSameSemaphoreBack:
        return jsValueForDecodedMessage<MessageName::IPCTester_SendSameSemaphoreBack>(globalObject, decoder);
    case MessageName::IPCTester_SendSemaphoreBackAndSignalProtocol:
        return jsValueForDecodedMessage<MessageName::IPCTester_SendSemaphoreBackAndSignalProtocol>(globalObject, decoder);
    case MessageName::IPCTester_SendAsyncMessageToReceiver:
        return jsValueForDecodedMessage<MessageName::IPCTester_SendAsyncMessageToReceiver>(globalObject, decoder);
#endif
#if ENABLE(IPC_TESTING_API)
    case MessageName::IPCTesterReceiver_AsyncMessage:
        return jsValueForDecodedMessage<MessageName::IPCTesterReceiver_AsyncMessage>(globalObject, decoder);
#endif
    case MessageName::WebConnection_HandleMessage:
        return jsValueForDecodedMessage<MessageName::WebConnection_HandleMessage>(globalObject, decoder);
    case MessageName::AuthenticationManager_CompleteAuthenticationChallenge:
        return jsValueForDecodedMessage<MessageName::AuthenticationManager_CompleteAuthenticationChallenge>(globalObject, decoder);
    case MessageName::NotificationManagerMessageHandler_RequestSystemNotificationPermission:
        return jsValueForDecodedMessage<MessageName::NotificationManagerMessageHandler_RequestSystemNotificationPermission>(globalObject, decoder);
    case MessageName::NotificationManagerMessageHandler_ShowNotification:
        return jsValueForDecodedMessage<MessageName::NotificationManagerMessageHandler_ShowNotification>(globalObject, decoder);
    case MessageName::NotificationManagerMessageHandler_CancelNotification:
        return jsValueForDecodedMessage<MessageName::NotificationManagerMessageHandler_CancelNotification>(globalObject, decoder);
    case MessageName::NotificationManagerMessageHandler_ClearNotifications:
        return jsValueForDecodedMessage<MessageName::NotificationManagerMessageHandler_ClearNotifications>(globalObject, decoder);
    case MessageName::NotificationManagerMessageHandler_DidDestroyNotification:
        return jsValueForDecodedMessage<MessageName::NotificationManagerMessageHandler_DidDestroyNotification>(globalObject, decoder);
    case MessageName::DrawingAreaProxy_EnterAcceleratedCompositingMode:
        return jsValueForDecodedMessage<MessageName::DrawingAreaProxy_EnterAcceleratedCompositingMode>(globalObject, decoder);
    case MessageName::DrawingAreaProxy_UpdateAcceleratedCompositingMode:
        return jsValueForDecodedMessage<MessageName::DrawingAreaProxy_UpdateAcceleratedCompositingMode>(globalObject, decoder);
    case MessageName::DrawingAreaProxy_DidFirstLayerFlush:
        return jsValueForDecodedMessage<MessageName::DrawingAreaProxy_DidFirstLayerFlush>(globalObject, decoder);
    case MessageName::DrawingAreaProxy_DispatchPresentationCallbacksAfterFlushingLayers:
        return jsValueForDecodedMessage<MessageName::DrawingAreaProxy_DispatchPresentationCallbacksAfterFlushingLayers>(globalObject, decoder);
#if USE(COORDINATED_GRAPHICS) || USE(TEXTURE_MAPPER)
    case MessageName::DrawingAreaProxy_Update:
        return jsValueForDecodedMessage<MessageName::DrawingAreaProxy_Update>(globalObject, decoder);
    case MessageName::DrawingAreaProxy_DidUpdateBackingStoreState:
        return jsValueForDecodedMessage<MessageName::DrawingAreaProxy_DidUpdateBackingStoreState>(globalObject, decoder);
    case MessageName::DrawingAreaProxy_ExitAcceleratedCompositingMode:
        return jsValueForDecodedMessage<MessageName::DrawingAreaProxy_ExitAcceleratedCompositingMode>(globalObject, decoder);
#endif
#if ENABLE(MEDIA_SESSION_COORDINATOR)
    case MessageName::RemoteMediaSessionCoordinatorProxy_Join:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSessionCoordinatorProxy_Join>(globalObject, decoder);
    case MessageName::RemoteMediaSessionCoordinatorProxy_Leave:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSessionCoordinatorProxy_Leave>(globalObject, decoder);
    case MessageName::RemoteMediaSessionCoordinatorProxy_CoordinateSeekTo:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSessionCoordinatorProxy_CoordinateSeekTo>(globalObject, decoder);
    case MessageName::RemoteMediaSessionCoordinatorProxy_CoordinatePlay:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSessionCoordinatorProxy_CoordinatePlay>(globalObject, decoder);
    case MessageName::RemoteMediaSessionCoordinatorProxy_CoordinatePause:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSessionCoordinatorProxy_CoordinatePause>(globalObject, decoder);
    case MessageName::RemoteMediaSessionCoordinatorProxy_CoordinateSetTrack:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSessionCoordinatorProxy_CoordinateSetTrack>(globalObject, decoder);
    case MessageName::RemoteMediaSessionCoordinatorProxy_PositionStateChanged:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSessionCoordinatorProxy_PositionStateChanged>(globalObject, decoder);
    case MessageName::RemoteMediaSessionCoordinatorProxy_ReadyStateChanged:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSessionCoordinatorProxy_ReadyStateChanged>(globalObject, decoder);
    case MessageName::RemoteMediaSessionCoordinatorProxy_PlaybackStateChanged:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSessionCoordinatorProxy_PlaybackStateChanged>(globalObject, decoder);
    case MessageName::RemoteMediaSessionCoordinatorProxy_TrackIdentifierChanged:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSessionCoordinatorProxy_TrackIdentifierChanged>(globalObject, decoder);
#endif
#if ENABLE(MEDIA_STREAM)
    case MessageName::SpeechRecognitionRemoteRealtimeMediaSourceManager_RemoteAudioSamplesAvailable:
        return jsValueForDecodedMessage<MessageName::SpeechRecognitionRemoteRealtimeMediaSourceManager_RemoteAudioSamplesAvailable>(globalObject, decoder);
    case MessageName::SpeechRecognitionRemoteRealtimeMediaSourceManager_RemoteCaptureFailed:
        return jsValueForDecodedMessage<MessageName::SpeechRecognitionRemoteRealtimeMediaSourceManager_RemoteCaptureFailed>(globalObject, decoder);
    case MessageName::SpeechRecognitionRemoteRealtimeMediaSourceManager_RemoteSourceStopped:
        return jsValueForDecodedMessage<MessageName::SpeechRecognitionRemoteRealtimeMediaSourceManager_RemoteSourceStopped>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::SpeechRecognitionRemoteRealtimeMediaSourceManager_SetStorage:
        return jsValueForDecodedMessage<MessageName::SpeechRecognitionRemoteRealtimeMediaSourceManager_SetStorage>(globalObject, decoder);
#endif
#endif
    case MessageName::SpeechRecognitionServer_Start:
        return jsValueForDecodedMessage<MessageName::SpeechRecognitionServer_Start>(globalObject, decoder);
    case MessageName::SpeechRecognitionServer_Stop:
        return jsValueForDecodedMessage<MessageName::SpeechRecognitionServer_Stop>(globalObject, decoder);
    case MessageName::SpeechRecognitionServer_Abort:
        return jsValueForDecodedMessage<MessageName::SpeechRecognitionServer_Abort>(globalObject, decoder);
    case MessageName::SpeechRecognitionServer_Invalidate:
        return jsValueForDecodedMessage<MessageName::SpeechRecognitionServer_Invalidate>(globalObject, decoder);
    case MessageName::VisitedLinkStore_AddVisitedLinkHashFromPage:
        return jsValueForDecodedMessage<MessageName::VisitedLinkStore_AddVisitedLinkHashFromPage>(globalObject, decoder);
    case MessageName::WebFrameProxy_DidCreateSubframe:
        return jsValueForDecodedMessage<MessageName::WebFrameProxy_DidCreateSubframe>(globalObject, decoder);
#if ENABLE(FULLSCREEN_API)
    case MessageName::WebFullScreenManagerProxy_SupportsFullScreen:
        return jsValueForDecodedMessage<MessageName::WebFullScreenManagerProxy_SupportsFullScreen>(globalObject, decoder);
    case MessageName::WebFullScreenManagerProxy_EnterFullScreen:
        return jsValueForDecodedMessage<MessageName::WebFullScreenManagerProxy_EnterFullScreen>(globalObject, decoder);
    case MessageName::WebFullScreenManagerProxy_ExitFullScreen:
        return jsValueForDecodedMessage<MessageName::WebFullScreenManagerProxy_ExitFullScreen>(globalObject, decoder);
    case MessageName::WebFullScreenManagerProxy_BeganEnterFullScreen:
        return jsValueForDecodedMessage<MessageName::WebFullScreenManagerProxy_BeganEnterFullScreen>(globalObject, decoder);
    case MessageName::WebFullScreenManagerProxy_BeganExitFullScreen:
        return jsValueForDecodedMessage<MessageName::WebFullScreenManagerProxy_BeganExitFullScreen>(globalObject, decoder);
    case MessageName::WebFullScreenManagerProxy_Close:
        return jsValueForDecodedMessage<MessageName::WebFullScreenManagerProxy_Close>(globalObject, decoder);
#endif
    case MessageName::WebGeolocationManagerProxy_StartUpdating:
        return jsValueForDecodedMessage<MessageName::WebGeolocationManagerProxy_StartUpdating>(globalObject, decoder);
    case MessageName::WebGeolocationManagerProxy_StopUpdating:
        return jsValueForDecodedMessage<MessageName::WebGeolocationManagerProxy_StopUpdating>(globalObject, decoder);
    case MessageName::WebGeolocationManagerProxy_SetEnableHighAccuracy:
        return jsValueForDecodedMessage<MessageName::WebGeolocationManagerProxy_SetEnableHighAccuracy>(globalObject, decoder);
    case MessageName::WebLockRegistryProxy_RequestLock:
        return jsValueForDecodedMessage<MessageName::WebLockRegistryProxy_RequestLock>(globalObject, decoder);
    case MessageName::WebLockRegistryProxy_ReleaseLock:
        return jsValueForDecodedMessage<MessageName::WebLockRegistryProxy_ReleaseLock>(globalObject, decoder);
    case MessageName::WebLockRegistryProxy_AbortLockRequest:
        return jsValueForDecodedMessage<MessageName::WebLockRegistryProxy_AbortLockRequest>(globalObject, decoder);
    case MessageName::WebLockRegistryProxy_Snapshot:
        return jsValueForDecodedMessage<MessageName::WebLockRegistryProxy_Snapshot>(globalObject, decoder);
    case MessageName::WebLockRegistryProxy_ClientIsGoingAway:
        return jsValueForDecodedMessage<MessageName::WebLockRegistryProxy_ClientIsGoingAway>(globalObject, decoder);
    case MessageName::WebPageProxy_CreateNewPage:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_CreateNewPage>(globalObject, decoder);
    case MessageName::WebPageProxy_ShowPage:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ShowPage>(globalObject, decoder);
    case MessageName::WebPageProxy_ClosePage:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ClosePage>(globalObject, decoder);
    case MessageName::WebPageProxy_RunJavaScriptAlert:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RunJavaScriptAlert>(globalObject, decoder);
    case MessageName::WebPageProxy_RunJavaScriptConfirm:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RunJavaScriptConfirm>(globalObject, decoder);
    case MessageName::WebPageProxy_RunJavaScriptPrompt:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RunJavaScriptPrompt>(globalObject, decoder);
    case MessageName::WebPageProxy_MouseDidMoveOverElement:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_MouseDidMoveOverElement>(globalObject, decoder);
    case MessageName::WebPageProxy_DidChangeViewportProperties:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidChangeViewportProperties>(globalObject, decoder);
    case MessageName::WebPageProxy_DidReceiveEvent:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidReceiveEvent>(globalObject, decoder);
    case MessageName::WebPageProxy_SetCursor:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetCursor>(globalObject, decoder);
    case MessageName::WebPageProxy_SetCursorHiddenUntilMouseMoves:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetCursorHiddenUntilMouseMoves>(globalObject, decoder);
    case MessageName::WebPageProxy_SetStatusText:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetStatusText>(globalObject, decoder);
    case MessageName::WebPageProxy_SetFocus:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetFocus>(globalObject, decoder);
    case MessageName::WebPageProxy_TakeFocus:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_TakeFocus>(globalObject, decoder);
    case MessageName::WebPageProxy_FocusFromServiceWorker:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_FocusFromServiceWorker>(globalObject, decoder);
    case MessageName::WebPageProxy_FocusedFrameChanged:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_FocusedFrameChanged>(globalObject, decoder);
    case MessageName::WebPageProxy_SetRenderTreeSize:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetRenderTreeSize>(globalObject, decoder);
    case MessageName::WebPageProxy_SetToolbarsAreVisible:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetToolbarsAreVisible>(globalObject, decoder);
    case MessageName::WebPageProxy_GetToolbarsAreVisible:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_GetToolbarsAreVisible>(globalObject, decoder);
    case MessageName::WebPageProxy_SetMenuBarIsVisible:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetMenuBarIsVisible>(globalObject, decoder);
    case MessageName::WebPageProxy_GetMenuBarIsVisible:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_GetMenuBarIsVisible>(globalObject, decoder);
    case MessageName::WebPageProxy_SetStatusBarIsVisible:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetStatusBarIsVisible>(globalObject, decoder);
    case MessageName::WebPageProxy_GetStatusBarIsVisible:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_GetStatusBarIsVisible>(globalObject, decoder);
    case MessageName::WebPageProxy_SetIsResizable:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetIsResizable>(globalObject, decoder);
    case MessageName::WebPageProxy_SetWindowFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetWindowFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_GetWindowFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_GetWindowFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_ScreenToRootView:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ScreenToRootView>(globalObject, decoder);
    case MessageName::WebPageProxy_RootViewToScreen:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RootViewToScreen>(globalObject, decoder);
    case MessageName::WebPageProxy_AccessibilityScreenToRootView:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_AccessibilityScreenToRootView>(globalObject, decoder);
    case MessageName::WebPageProxy_RootViewToAccessibilityScreen:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RootViewToAccessibilityScreen>(globalObject, decoder);
#if PLATFORM(COCOA) || PLATFORM(GTK)
    case MessageName::WebPageProxy_ShowValidationMessage:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ShowValidationMessage>(globalObject, decoder);
    case MessageName::WebPageProxy_HideValidationMessage:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_HideValidationMessage>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_RunBeforeUnloadConfirmPanel:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RunBeforeUnloadConfirmPanel>(globalObject, decoder);
    case MessageName::WebPageProxy_PageDidScroll:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_PageDidScroll>(globalObject, decoder);
    case MessageName::WebPageProxy_SetHasActiveAnimatedScrolls:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetHasActiveAnimatedScrolls>(globalObject, decoder);
    case MessageName::WebPageProxy_RunOpenPanel:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RunOpenPanel>(globalObject, decoder);
    case MessageName::WebPageProxy_ShowShareSheet:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ShowShareSheet>(globalObject, decoder);
    case MessageName::WebPageProxy_ShowContactPicker:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ShowContactPicker>(globalObject, decoder);
    case MessageName::WebPageProxy_PrintFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_PrintFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_RunModal:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RunModal>(globalObject, decoder);
    case MessageName::WebPageProxy_NotifyScrollerThumbIsVisibleInRect:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_NotifyScrollerThumbIsVisibleInRect>(globalObject, decoder);
    case MessageName::WebPageProxy_RecommendedScrollbarStyleDidChange:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RecommendedScrollbarStyleDidChange>(globalObject, decoder);
    case MessageName::WebPageProxy_DidChangeScrollbarsForMainFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidChangeScrollbarsForMainFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_DidChangeScrollOffsetPinningForMainFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidChangeScrollOffsetPinningForMainFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_DidChangePageCount:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidChangePageCount>(globalObject, decoder);
    case MessageName::WebPageProxy_ThemeColorChanged:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ThemeColorChanged>(globalObject, decoder);
    case MessageName::WebPageProxy_PageExtendedBackgroundColorDidChange:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_PageExtendedBackgroundColorDidChange>(globalObject, decoder);
    case MessageName::WebPageProxy_SampledPageTopColorChanged:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SampledPageTopColorChanged>(globalObject, decoder);
    case MessageName::WebPageProxy_SetCanShortCircuitHorizontalWheelEvents:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetCanShortCircuitHorizontalWheelEvents>(globalObject, decoder);
    case MessageName::WebPageProxy_DidChangeContentSize:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidChangeContentSize>(globalObject, decoder);
    case MessageName::WebPageProxy_DidChangeIntrinsicContentSize:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidChangeIntrinsicContentSize>(globalObject, decoder);
#if ENABLE(INPUT_TYPE_COLOR)
    case MessageName::WebPageProxy_ShowColorPicker:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ShowColorPicker>(globalObject, decoder);
    case MessageName::WebPageProxy_SetColorPickerColor:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetColorPickerColor>(globalObject, decoder);
    case MessageName::WebPageProxy_EndColorPicker:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_EndColorPicker>(globalObject, decoder);
#endif
#if ENABLE(DATALIST_ELEMENT)
    case MessageName::WebPageProxy_ShowDataListSuggestions:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ShowDataListSuggestions>(globalObject, decoder);
    case MessageName::WebPageProxy_HandleKeydownInDataList:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_HandleKeydownInDataList>(globalObject, decoder);
    case MessageName::WebPageProxy_EndDataListSuggestions:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_EndDataListSuggestions>(globalObject, decoder);
#endif
#if ENABLE(DATE_AND_TIME_INPUT_TYPES)
    case MessageName::WebPageProxy_ShowDateTimePicker:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ShowDateTimePicker>(globalObject, decoder);
    case MessageName::WebPageProxy_EndDateTimePicker:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_EndDateTimePicker>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_DecidePolicyForResponse:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DecidePolicyForResponse>(globalObject, decoder);
    case MessageName::WebPageProxy_DecidePolicyForNavigationActionAsync:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DecidePolicyForNavigationActionAsync>(globalObject, decoder);
    case MessageName::WebPageProxy_DecidePolicyForNavigationActionSync:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DecidePolicyForNavigationActionSync>(globalObject, decoder);
    case MessageName::WebPageProxy_DecidePolicyForNewWindowAction:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DecidePolicyForNewWindowAction>(globalObject, decoder);
    case MessageName::WebPageProxy_DidChangeProgress:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidChangeProgress>(globalObject, decoder);
    case MessageName::WebPageProxy_DidFinishProgress:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidFinishProgress>(globalObject, decoder);
    case MessageName::WebPageProxy_DidStartProgress:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidStartProgress>(globalObject, decoder);
    case MessageName::WebPageProxy_SetNetworkRequestsInProgress:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetNetworkRequestsInProgress>(globalObject, decoder);
    case MessageName::WebPageProxy_DidCreateMainFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidCreateMainFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_DidStartProvisionalLoadForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidStartProvisionalLoadForFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_DidReceiveServerRedirectForProvisionalLoadForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidReceiveServerRedirectForProvisionalLoadForFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_WillPerformClientRedirectForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_WillPerformClientRedirectForFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_DidCancelClientRedirectForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidCancelClientRedirectForFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_DidChangeProvisionalURLForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidChangeProvisionalURLForFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_DidFailProvisionalLoadForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidFailProvisionalLoadForFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_DidCommitLoadForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidCommitLoadForFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_DidFailLoadForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidFailLoadForFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_DidFinishDocumentLoadForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidFinishDocumentLoadForFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_DidFinishLoadForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidFinishLoadForFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_DidFirstLayoutForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidFirstLayoutForFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_DidFirstVisuallyNonEmptyLayoutForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidFirstVisuallyNonEmptyLayoutForFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_DidReachLayoutMilestone:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidReachLayoutMilestone>(globalObject, decoder);
    case MessageName::WebPageProxy_DidReceiveTitleForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidReceiveTitleForFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_DidDisplayInsecureContentForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidDisplayInsecureContentForFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_DidRunInsecureContentForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidRunInsecureContentForFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_DidSameDocumentNavigationForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidSameDocumentNavigationForFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_DidSameDocumentNavigationForFrameViaJSHistoryAPI:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidSameDocumentNavigationForFrameViaJSHistoryAPI>(globalObject, decoder);
    case MessageName::WebPageProxy_DidChangeMainDocument:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidChangeMainDocument>(globalObject, decoder);
    case MessageName::WebPageProxy_DidExplicitOpenForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidExplicitOpenForFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_DidDestroyNavigation:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidDestroyNavigation>(globalObject, decoder);
    case MessageName::WebPageProxy_UpdateRemoteFrameSize:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_UpdateRemoteFrameSize>(globalObject, decoder);
    case MessageName::WebPageProxy_MainFramePluginHandlesPageScaleGestureDidChange:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_MainFramePluginHandlesPageScaleGestureDidChange>(globalObject, decoder);
#if PLATFORM(GTK) || PLATFORM(WPE)
    case MessageName::WebPageProxy_DidInitiateLoadForResource:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidInitiateLoadForResource>(globalObject, decoder);
    case MessageName::WebPageProxy_DidSendRequestForResource:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidSendRequestForResource>(globalObject, decoder);
    case MessageName::WebPageProxy_DidReceiveResponseForResource:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidReceiveResponseForResource>(globalObject, decoder);
    case MessageName::WebPageProxy_DidFinishLoadForResource:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidFinishLoadForResource>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_DidNavigateWithNavigationData:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidNavigateWithNavigationData>(globalObject, decoder);
    case MessageName::WebPageProxy_DidPerformClientRedirect:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidPerformClientRedirect>(globalObject, decoder);
    case MessageName::WebPageProxy_DidPerformServerRedirect:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidPerformServerRedirect>(globalObject, decoder);
    case MessageName::WebPageProxy_DidUpdateHistoryTitle:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidUpdateHistoryTitle>(globalObject, decoder);
    case MessageName::WebPageProxy_DidFinishLoadingDataForCustomContentProvider:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidFinishLoadingDataForCustomContentProvider>(globalObject, decoder);
    case MessageName::WebPageProxy_WillSubmitForm:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_WillSubmitForm>(globalObject, decoder);
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebPageProxy_InterpretKeyEvent:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_InterpretKeyEvent>(globalObject, decoder);
    case MessageName::WebPageProxy_DidReceivePositionInformation:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidReceivePositionInformation>(globalObject, decoder);
    case MessageName::WebPageProxy_SaveImageToLibrary:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SaveImageToLibrary>(globalObject, decoder);
    case MessageName::WebPageProxy_ShowPlaybackTargetPicker:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ShowPlaybackTargetPicker>(globalObject, decoder);
    case MessageName::WebPageProxy_CommitPotentialTapFailed:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_CommitPotentialTapFailed>(globalObject, decoder);
    case MessageName::WebPageProxy_DidNotHandleTapAsClick:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidNotHandleTapAsClick>(globalObject, decoder);
    case MessageName::WebPageProxy_DidCompleteSyntheticClick:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidCompleteSyntheticClick>(globalObject, decoder);
    case MessageName::WebPageProxy_DisableDoubleTapGesturesDuringTapIfNecessary:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DisableDoubleTapGesturesDuringTapIfNecessary>(globalObject, decoder);
    case MessageName::WebPageProxy_HandleSmartMagnificationInformationForPotentialTap:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_HandleSmartMagnificationInformationForPotentialTap>(globalObject, decoder);
#endif
#if ENABLE(DATA_DETECTION)
    case MessageName::WebPageProxy_SetDataDetectionResult:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetDataDetectionResult>(globalObject, decoder);
    case MessageName::WebPageProxy_HandleClickForDataDetectionResult:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_HandleClickForDataDetectionResult>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_PageScaleFactorDidChange:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_PageScaleFactorDidChange>(globalObject, decoder);
    case MessageName::WebPageProxy_PluginScaleFactorDidChange:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_PluginScaleFactorDidChange>(globalObject, decoder);
    case MessageName::WebPageProxy_PluginZoomFactorDidChange:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_PluginZoomFactorDidChange>(globalObject, decoder);
#if PLATFORM(GTK) || PLATFORM(WPE)
    case MessageName::WebPageProxy_BindAccessibilityTree:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_BindAccessibilityTree>(globalObject, decoder);
    case MessageName::WebPageProxy_SetInputMethodState:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetInputMethodState>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_BackForwardAddItem:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_BackForwardAddItem>(globalObject, decoder);
    case MessageName::WebPageProxy_BackForwardGoToItem:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_BackForwardGoToItem>(globalObject, decoder);
    case MessageName::WebPageProxy_BackForwardItemAtIndex:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_BackForwardItemAtIndex>(globalObject, decoder);
    case MessageName::WebPageProxy_BackForwardListContainsItem:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_BackForwardListContainsItem>(globalObject, decoder);
    case MessageName::WebPageProxy_BackForwardListCounts:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_BackForwardListCounts>(globalObject, decoder);
    case MessageName::WebPageProxy_BackForwardClear:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_BackForwardClear>(globalObject, decoder);
    case MessageName::WebPageProxy_WillGoToBackForwardListItem:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_WillGoToBackForwardListItem>(globalObject, decoder);
    case MessageName::WebPageProxy_RegisterEditCommandForUndo:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RegisterEditCommandForUndo>(globalObject, decoder);
    case MessageName::WebPageProxy_ClearAllEditCommands:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ClearAllEditCommands>(globalObject, decoder);
    case MessageName::WebPageProxy_RegisterInsertionUndoGrouping:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RegisterInsertionUndoGrouping>(globalObject, decoder);
    case MessageName::WebPageProxy_CanUndoRedo:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_CanUndoRedo>(globalObject, decoder);
    case MessageName::WebPageProxy_ExecuteUndoRedo:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ExecuteUndoRedo>(globalObject, decoder);
    case MessageName::WebPageProxy_LogDiagnosticMessageFromWebProcess:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_LogDiagnosticMessageFromWebProcess>(globalObject, decoder);
    case MessageName::WebPageProxy_LogDiagnosticMessageWithResultFromWebProcess:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_LogDiagnosticMessageWithResultFromWebProcess>(globalObject, decoder);
    case MessageName::WebPageProxy_LogDiagnosticMessageWithValueFromWebProcess:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_LogDiagnosticMessageWithValueFromWebProcess>(globalObject, decoder);
    case MessageName::WebPageProxy_LogDiagnosticMessageWithEnhancedPrivacyFromWebProcess:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_LogDiagnosticMessageWithEnhancedPrivacyFromWebProcess>(globalObject, decoder);
    case MessageName::WebPageProxy_LogDiagnosticMessageWithValueDictionaryFromWebProcess:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_LogDiagnosticMessageWithValueDictionaryFromWebProcess>(globalObject, decoder);
    case MessageName::WebPageProxy_LogDiagnosticMessageWithDomainFromWebProcess:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_LogDiagnosticMessageWithDomainFromWebProcess>(globalObject, decoder);
    case MessageName::WebPageProxy_LogScrollingEvent:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_LogScrollingEvent>(globalObject, decoder);
    case MessageName::WebPageProxy_EditorStateChanged:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_EditorStateChanged>(globalObject, decoder);
    case MessageName::WebPageProxy_CompositionWasCanceled:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_CompositionWasCanceled>(globalObject, decoder);
    case MessageName::WebPageProxy_SetHasHadSelectionChangesFromUserInteraction:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetHasHadSelectionChangesFromUserInteraction>(globalObject, decoder);
#if ENABLE(IMAGE_ANALYSIS)
    case MessageName::WebPageProxy_RequestTextRecognition:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RequestTextRecognition>(globalObject, decoder);
#endif
#if HAVE(TRANSLATION_UI_SERVICES) && ENABLE(CONTEXT_MENUS)
    case MessageName::WebPageProxy_HandleContextMenuTranslation:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_HandleContextMenuTranslation>(globalObject, decoder);
#endif
#if ENABLE(MEDIA_CONTROLS_CONTEXT_MENUS) && USE(UICONTEXTMENU)
    case MessageName::WebPageProxy_ShowMediaControlsContextMenu:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ShowMediaControlsContextMenu>(globalObject, decoder);
#endif
#if HAVE(TOUCH_BAR)
    case MessageName::WebPageProxy_SetIsTouchBarUpdateSupressedForHiddenContentEditable:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetIsTouchBarUpdateSupressedForHiddenContentEditable>(globalObject, decoder);
    case MessageName::WebPageProxy_SetIsNeverRichlyEditableForTouchBar:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetIsNeverRichlyEditableForTouchBar>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_RequestDOMPasteAccess:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RequestDOMPasteAccess>(globalObject, decoder);
    case MessageName::WebPageProxy_DidCountStringMatches:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidCountStringMatches>(globalObject, decoder);
    case MessageName::WebPageProxy_SetTextIndicator:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetTextIndicator>(globalObject, decoder);
    case MessageName::WebPageProxy_ClearTextIndicator:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ClearTextIndicator>(globalObject, decoder);
    case MessageName::WebPageProxy_DidFindString:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidFindString>(globalObject, decoder);
    case MessageName::WebPageProxy_DidFailToFindString:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidFailToFindString>(globalObject, decoder);
    case MessageName::WebPageProxy_DidFindStringMatches:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidFindStringMatches>(globalObject, decoder);
    case MessageName::WebPageProxy_DidGetImageForFindMatch:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidGetImageForFindMatch>(globalObject, decoder);
    case MessageName::WebPageProxy_ShowPopupMenu:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ShowPopupMenu>(globalObject, decoder);
    case MessageName::WebPageProxy_HidePopupMenu:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_HidePopupMenu>(globalObject, decoder);
#if ENABLE(CONTEXT_MENUS)
    case MessageName::WebPageProxy_ShowContextMenu:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ShowContextMenu>(globalObject, decoder);
#endif
#if ENABLE(SERVICE_WORKER)
    case MessageName::WebPageProxy_DidFinishServiceWorkerPageRegistration:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidFinishServiceWorkerPageRegistration>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_ExceededDatabaseQuota:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ExceededDatabaseQuota>(globalObject, decoder);
    case MessageName::WebPageProxy_ReachedApplicationCacheOriginQuota:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ReachedApplicationCacheOriginQuota>(globalObject, decoder);
    case MessageName::WebPageProxy_RequestGeolocationPermissionForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RequestGeolocationPermissionForFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_RevokeGeolocationAuthorizationToken:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RevokeGeolocationAuthorizationToken>(globalObject, decoder);
#if ENABLE(MEDIA_STREAM)
    case MessageName::WebPageProxy_RequestUserMediaPermissionForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RequestUserMediaPermissionForFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_EnumerateMediaDevicesForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_EnumerateMediaDevicesForFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_BeginMonitoringCaptureDevices:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_BeginMonitoringCaptureDevices>(globalObject, decoder);
#endif
#if ENABLE(ENCRYPTED_MEDIA)
    case MessageName::WebPageProxy_RequestMediaKeySystemPermissionForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RequestMediaKeySystemPermissionForFrame>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_RequestNotificationPermission:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RequestNotificationPermission>(globalObject, decoder);
#if USE(UNIFIED_TEXT_CHECKING)
    case MessageName::WebPageProxy_CheckTextOfParagraph:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_CheckTextOfParagraph>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_CheckSpellingOfString:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_CheckSpellingOfString>(globalObject, decoder);
    case MessageName::WebPageProxy_CheckGrammarOfString:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_CheckGrammarOfString>(globalObject, decoder);
    case MessageName::WebPageProxy_SpellingUIIsShowing:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SpellingUIIsShowing>(globalObject, decoder);
    case MessageName::WebPageProxy_UpdateSpellingUIWithMisspelledWord:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_UpdateSpellingUIWithMisspelledWord>(globalObject, decoder);
    case MessageName::WebPageProxy_UpdateSpellingUIWithGrammarString:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_UpdateSpellingUIWithGrammarString>(globalObject, decoder);
    case MessageName::WebPageProxy_GetGuessesForWord:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_GetGuessesForWord>(globalObject, decoder);
    case MessageName::WebPageProxy_LearnWord:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_LearnWord>(globalObject, decoder);
    case MessageName::WebPageProxy_IgnoreWord:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_IgnoreWord>(globalObject, decoder);
    case MessageName::WebPageProxy_RequestCheckingOfString:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RequestCheckingOfString>(globalObject, decoder);
#if ENABLE(DRAG_SUPPORT)
    case MessageName::WebPageProxy_DidPerformDragControllerAction:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidPerformDragControllerAction>(globalObject, decoder);
    case MessageName::WebPageProxy_DidEndDragging:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidEndDragging>(globalObject, decoder);
#endif
#if PLATFORM(COCOA) && ENABLE(DRAG_SUPPORT)
    case MessageName::WebPageProxy_StartDrag:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_StartDrag>(globalObject, decoder);
    case MessageName::WebPageProxy_SetPromisedDataForImage:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetPromisedDataForImage>(globalObject, decoder);
#endif
#if PLATFORM(GTK) && ENABLE(DRAG_SUPPORT)
    case MessageName::WebPageProxy_StartDrag:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_StartDrag>(globalObject, decoder);
#endif
#if ENABLE(DRAG_SUPPORT)
    case MessageName::WebPageProxy_DidPerformDragOperation:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidPerformDragOperation>(globalObject, decoder);
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(DRAG_SUPPORT)
    case MessageName::WebPageProxy_DidHandleDragStartRequest:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidHandleDragStartRequest>(globalObject, decoder);
    case MessageName::WebPageProxy_DidHandleAdditionalDragItemsRequest:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidHandleAdditionalDragItemsRequest>(globalObject, decoder);
    case MessageName::WebPageProxy_WillReceiveEditDragSnapshot:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_WillReceiveEditDragSnapshot>(globalObject, decoder);
    case MessageName::WebPageProxy_DidReceiveEditDragSnapshot:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidReceiveEditDragSnapshot>(globalObject, decoder);
#endif
#if PLATFORM(COCOA)
    case MessageName::WebPageProxy_DidPerformDictionaryLookup:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidPerformDictionaryLookup>(globalObject, decoder);
    case MessageName::WebPageProxy_ExecuteSavedCommandBySelector:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ExecuteSavedCommandBySelector>(globalObject, decoder);
    case MessageName::WebPageProxy_RegisterWebProcessAccessibilityToken:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RegisterWebProcessAccessibilityToken>(globalObject, decoder);
    case MessageName::WebPageProxy_GetIsSpeaking:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_GetIsSpeaking>(globalObject, decoder);
    case MessageName::WebPageProxy_Speak:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_Speak>(globalObject, decoder);
    case MessageName::WebPageProxy_StopSpeaking:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_StopSpeaking>(globalObject, decoder);
    case MessageName::WebPageProxy_MakeFirstResponder:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_MakeFirstResponder>(globalObject, decoder);
    case MessageName::WebPageProxy_AssistiveTechnologyMakeFirstResponder:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_AssistiveTechnologyMakeFirstResponder>(globalObject, decoder);
    case MessageName::WebPageProxy_SearchWithSpotlight:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SearchWithSpotlight>(globalObject, decoder);
    case MessageName::WebPageProxy_SearchTheWeb:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SearchTheWeb>(globalObject, decoder);
#endif
#if HAVE(TOUCH_BAR)
    case MessageName::WebPageProxy_TouchBarMenuDataChanged:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_TouchBarMenuDataChanged>(globalObject, decoder);
    case MessageName::WebPageProxy_TouchBarMenuItemDataAdded:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_TouchBarMenuItemDataAdded>(globalObject, decoder);
    case MessageName::WebPageProxy_TouchBarMenuItemDataRemoved:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_TouchBarMenuItemDataRemoved>(globalObject, decoder);
#endif
#if USE(APPKIT)
    case MessageName::WebPageProxy_SubstitutionsPanelIsShowing:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SubstitutionsPanelIsShowing>(globalObject, decoder);
#endif
#if USE(AUTOMATIC_TEXT_REPLACEMENT)
    case MessageName::WebPageProxy_toggleSmartInsertDelete:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_toggleSmartInsertDelete>(globalObject, decoder);
    case MessageName::WebPageProxy_toggleAutomaticQuoteSubstitution:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_toggleAutomaticQuoteSubstitution>(globalObject, decoder);
    case MessageName::WebPageProxy_toggleAutomaticLinkDetection:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_toggleAutomaticLinkDetection>(globalObject, decoder);
    case MessageName::WebPageProxy_toggleAutomaticDashSubstitution:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_toggleAutomaticDashSubstitution>(globalObject, decoder);
    case MessageName::WebPageProxy_toggleAutomaticTextReplacement:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_toggleAutomaticTextReplacement>(globalObject, decoder);
#endif
#if PLATFORM(MAC)
    case MessageName::WebPageProxy_ShowCorrectionPanel:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ShowCorrectionPanel>(globalObject, decoder);
    case MessageName::WebPageProxy_DismissCorrectionPanel:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DismissCorrectionPanel>(globalObject, decoder);
    case MessageName::WebPageProxy_DismissCorrectionPanelSoon:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DismissCorrectionPanelSoon>(globalObject, decoder);
    case MessageName::WebPageProxy_RecordAutocorrectionResponse:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RecordAutocorrectionResponse>(globalObject, decoder);
    case MessageName::WebPageProxy_SetEditableElementIsFocused:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetEditableElementIsFocused>(globalObject, decoder);
    case MessageName::WebPageProxy_SetCaretDecorationVisibility:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetCaretDecorationVisibility>(globalObject, decoder);
    case MessageName::WebPageProxy_HandleAcceptsFirstMouse:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_HandleAcceptsFirstMouse>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_DidUpdateRenderingAfterCommittingLoad:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidUpdateRenderingAfterCommittingLoad>(globalObject, decoder);
#if USE(DICTATION_ALTERNATIVES)
    case MessageName::WebPageProxy_ShowDictationAlternativeUI:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ShowDictationAlternativeUI>(globalObject, decoder);
    case MessageName::WebPageProxy_RemoveDictationAlternatives:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RemoveDictationAlternatives>(globalObject, decoder);
    case MessageName::WebPageProxy_DictationAlternatives:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DictationAlternatives>(globalObject, decoder);
#endif
#if ENABLE(VIDEO)
    case MessageName::WebPageProxy_BeginTextRecognitionForVideoInElementFullScreen:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_BeginTextRecognitionForVideoInElementFullScreen>(globalObject, decoder);
    case MessageName::WebPageProxy_CancelTextRecognitionForVideoInElementFullScreen:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_CancelTextRecognitionForVideoInElementFullScreen>(globalObject, decoder);
#endif
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebPageProxy_CouldNotRestorePageState:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_CouldNotRestorePageState>(globalObject, decoder);
    case MessageName::WebPageProxy_RestorePageState:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RestorePageState>(globalObject, decoder);
    case MessageName::WebPageProxy_RestorePageCenterAndScale:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RestorePageCenterAndScale>(globalObject, decoder);
    case MessageName::WebPageProxy_DidGetTapHighlightGeometries:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidGetTapHighlightGeometries>(globalObject, decoder);
    case MessageName::WebPageProxy_ElementDidFocus:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ElementDidFocus>(globalObject, decoder);
    case MessageName::WebPageProxy_ElementDidBlur:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ElementDidBlur>(globalObject, decoder);
    case MessageName::WebPageProxy_UpdateInputContextAfterBlurringAndRefocusingElement:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_UpdateInputContextAfterBlurringAndRefocusingElement>(globalObject, decoder);
    case MessageName::WebPageProxy_FocusedElementDidChangeInputMode:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_FocusedElementDidChangeInputMode>(globalObject, decoder);
    case MessageName::WebPageProxy_ScrollingNodeScrollWillStartScroll:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ScrollingNodeScrollWillStartScroll>(globalObject, decoder);
    case MessageName::WebPageProxy_ScrollingNodeScrollDidEndScroll:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ScrollingNodeScrollDidEndScroll>(globalObject, decoder);
    case MessageName::WebPageProxy_ShowInspectorHighlight:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ShowInspectorHighlight>(globalObject, decoder);
    case MessageName::WebPageProxy_HideInspectorHighlight:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_HideInspectorHighlight>(globalObject, decoder);
    case MessageName::WebPageProxy_ShowInspectorIndication:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ShowInspectorIndication>(globalObject, decoder);
    case MessageName::WebPageProxy_HideInspectorIndication:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_HideInspectorIndication>(globalObject, decoder);
    case MessageName::WebPageProxy_EnableInspectorNodeSearch:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_EnableInspectorNodeSearch>(globalObject, decoder);
    case MessageName::WebPageProxy_DisableInspectorNodeSearch:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DisableInspectorNodeSearch>(globalObject, decoder);
    case MessageName::WebPageProxy_UpdateStringForFind:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_UpdateStringForFind>(globalObject, decoder);
    case MessageName::WebPageProxy_HandleAutocorrectionContext:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_HandleAutocorrectionContext>(globalObject, decoder);
    case MessageName::WebPageProxy_ShowDataDetectorsUIForPositionInformation:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ShowDataDetectorsUIForPositionInformation>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_DidChangeInspectorFrontendCount:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidChangeInspectorFrontendCount>(globalObject, decoder);
    case MessageName::WebPageProxy_CreateInspectorTarget:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_CreateInspectorTarget>(globalObject, decoder);
    case MessageName::WebPageProxy_DestroyInspectorTarget:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DestroyInspectorTarget>(globalObject, decoder);
    case MessageName::WebPageProxy_SendMessageToInspectorFrontend:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SendMessageToInspectorFrontend>(globalObject, decoder);
    case MessageName::WebPageProxy_SaveRecentSearches:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SaveRecentSearches>(globalObject, decoder);
    case MessageName::WebPageProxy_LoadRecentSearches:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_LoadRecentSearches>(globalObject, decoder);
#if ENABLE(PDFKIT_PLUGIN) && !ENABLE(UI_PROCESS_PDF_HUD)
    case MessageName::WebPageProxy_SavePDFToFileInDownloadsFolder:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SavePDFToFileInDownloadsFolder>(globalObject, decoder);
    case MessageName::WebPageProxy_SavePDFToTemporaryFolderAndOpenWithNativeApplication:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SavePDFToTemporaryFolderAndOpenWithNativeApplication>(globalObject, decoder);
    case MessageName::WebPageProxy_OpenPDFFromTemporaryFolderWithNativeApplication:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_OpenPDFFromTemporaryFolderWithNativeApplication>(globalObject, decoder);
#endif
#if ENABLE(PDFKIT_PLUGIN)
    case MessageName::WebPageProxy_ShowPDFContextMenu:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ShowPDFContextMenu>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_DidUpdateActivityState:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidUpdateActivityState>(globalObject, decoder);
#if ENABLE(WEB_CRYPTO)
    case MessageName::WebPageProxy_WrapCryptoKey:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_WrapCryptoKey>(globalObject, decoder);
    case MessageName::WebPageProxy_UnwrapCryptoKey:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_UnwrapCryptoKey>(globalObject, decoder);
#endif
#if (ENABLE(TELEPHONE_NUMBER_DETECTION) && PLATFORM(MAC))
    case MessageName::WebPageProxy_ShowTelephoneNumberMenu:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ShowTelephoneNumberMenu>(globalObject, decoder);
#endif
#if USE(QUICK_LOOK)
    case MessageName::WebPageProxy_DidStartLoadForQuickLookDocumentInMainFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidStartLoadForQuickLookDocumentInMainFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_DidFinishLoadForQuickLookDocumentInMainFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidFinishLoadForQuickLookDocumentInMainFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_RequestPasswordForQuickLookDocumentInMainFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RequestPasswordForQuickLookDocumentInMainFrame>(globalObject, decoder);
#endif
#if ENABLE(CONTENT_FILTERING)
    case MessageName::WebPageProxy_ContentFilterDidBlockLoadForFrame:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ContentFilterDidBlockLoadForFrame>(globalObject, decoder);
#endif
#if ENABLE(ACCESSIBILITY_ANIMATION_CONTROL)
    case MessageName::WebPageProxy_IsAnyAnimationAllowedToPlayDidChange:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_IsAnyAnimationAllowedToPlayDidChange>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_IsPlayingMediaDidChange:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_IsPlayingMediaDidChange>(globalObject, decoder);
    case MessageName::WebPageProxy_HandleAutoplayEvent:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_HandleAutoplayEvent>(globalObject, decoder);
#if PLATFORM(MAC)
    case MessageName::WebPageProxy_DidPerformImmediateActionHitTest:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidPerformImmediateActionHitTest>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_HandleMessage:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_HandleMessage>(globalObject, decoder);
    case MessageName::WebPageProxy_HandleSynchronousMessage:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_HandleSynchronousMessage>(globalObject, decoder);
    case MessageName::WebPageProxy_HandleAutoFillButtonClick:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_HandleAutoFillButtonClick>(globalObject, decoder);
    case MessageName::WebPageProxy_DidResignInputElementStrongPasswordAppearance:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidResignInputElementStrongPasswordAppearance>(globalObject, decoder);
#if ENABLE(CONTENT_EXTENSIONS)
    case MessageName::WebPageProxy_ContentRuleListNotification:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ContentRuleListNotification>(globalObject, decoder);
#endif
#if ENABLE(WIRELESS_PLAYBACK_TARGET) && !PLATFORM(IOS_FAMILY)
    case MessageName::WebPageProxy_AddPlaybackTargetPickerClient:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_AddPlaybackTargetPickerClient>(globalObject, decoder);
    case MessageName::WebPageProxy_RemovePlaybackTargetPickerClient:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RemovePlaybackTargetPickerClient>(globalObject, decoder);
    case MessageName::WebPageProxy_ShowPlaybackTargetPicker:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ShowPlaybackTargetPicker>(globalObject, decoder);
    case MessageName::WebPageProxy_PlaybackTargetPickerClientStateDidChange:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_PlaybackTargetPickerClientStateDidChange>(globalObject, decoder);
    case MessageName::WebPageProxy_SetMockMediaPlaybackTargetPickerEnabled:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetMockMediaPlaybackTargetPickerEnabled>(globalObject, decoder);
    case MessageName::WebPageProxy_SetMockMediaPlaybackTargetPickerState:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetMockMediaPlaybackTargetPickerState>(globalObject, decoder);
    case MessageName::WebPageProxy_MockMediaPlaybackTargetPickerDismissPopup:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_MockMediaPlaybackTargetPickerDismissPopup>(globalObject, decoder);
#endif
#if ENABLE(VIDEO_PRESENTATION_MODE)
    case MessageName::WebPageProxy_SetMockVideoPresentationModeEnabled:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetMockVideoPresentationModeEnabled>(globalObject, decoder);
#endif
#if ENABLE(POINTER_LOCK)
    case MessageName::WebPageProxy_RequestPointerLock:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RequestPointerLock>(globalObject, decoder);
    case MessageName::WebPageProxy_RequestPointerUnlock:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RequestPointerUnlock>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_DidFailToSuspendAfterProcessSwap:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidFailToSuspendAfterProcessSwap>(globalObject, decoder);
    case MessageName::WebPageProxy_DidSuspendAfterProcessSwap:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidSuspendAfterProcessSwap>(globalObject, decoder);
    case MessageName::WebPageProxy_ImageOrMediaDocumentSizeChanged:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ImageOrMediaDocumentSizeChanged>(globalObject, decoder);
    case MessageName::WebPageProxy_UseFixedLayoutDidChange:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_UseFixedLayoutDidChange>(globalObject, decoder);
    case MessageName::WebPageProxy_FixedLayoutSizeDidChange:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_FixedLayoutSizeDidChange>(globalObject, decoder);
    case MessageName::WebPageProxy_DidRestoreScrollPosition:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidRestoreScrollPosition>(globalObject, decoder);
    case MessageName::WebPageProxy_RequestScrollToRect:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RequestScrollToRect>(globalObject, decoder);
    case MessageName::WebPageProxy_GetLoadDecisionForIcon:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_GetLoadDecisionForIcon>(globalObject, decoder);
#if PLATFORM(MAC)
    case MessageName::WebPageProxy_DidHandleAcceptedCandidate:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidHandleAcceptedCandidate>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_SetIsUsingHighPerformanceWebGL:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetIsUsingHighPerformanceWebGL>(globalObject, decoder);
    case MessageName::WebPageProxy_StartURLSchemeTask:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_StartURLSchemeTask>(globalObject, decoder);
    case MessageName::WebPageProxy_StopURLSchemeTask:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_StopURLSchemeTask>(globalObject, decoder);
    case MessageName::WebPageProxy_LoadSynchronousURLSchemeTask:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_LoadSynchronousURLSchemeTask>(globalObject, decoder);
#if ENABLE(DEVICE_ORIENTATION)
    case MessageName::WebPageProxy_ShouldAllowDeviceOrientationAndMotionAccess:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ShouldAllowDeviceOrientationAndMotionAccess>(globalObject, decoder);
#endif
#if ENABLE(ATTACHMENT_ELEMENT)
    case MessageName::WebPageProxy_RegisterAttachmentIdentifierFromData:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RegisterAttachmentIdentifierFromData>(globalObject, decoder);
    case MessageName::WebPageProxy_RegisterAttachmentIdentifierFromFilePath:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RegisterAttachmentIdentifierFromFilePath>(globalObject, decoder);
    case MessageName::WebPageProxy_RegisterAttachmentIdentifier:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RegisterAttachmentIdentifier>(globalObject, decoder);
    case MessageName::WebPageProxy_RegisterAttachmentsFromSerializedData:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RegisterAttachmentsFromSerializedData>(globalObject, decoder);
    case MessageName::WebPageProxy_CloneAttachmentData:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_CloneAttachmentData>(globalObject, decoder);
    case MessageName::WebPageProxy_DidInsertAttachmentWithIdentifier:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidInsertAttachmentWithIdentifier>(globalObject, decoder);
    case MessageName::WebPageProxy_DidRemoveAttachmentWithIdentifier:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidRemoveAttachmentWithIdentifier>(globalObject, decoder);
    case MessageName::WebPageProxy_SerializedAttachmentDataForIdentifiers:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SerializedAttachmentDataForIdentifiers>(globalObject, decoder);
#endif
#if (ENABLE(ATTACHMENT_ELEMENT) && PLATFORM(IOS_FAMILY))
    case MessageName::WebPageProxy_WritePromisedAttachmentToPasteboard:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_WritePromisedAttachmentToPasteboard>(globalObject, decoder);
#endif
#if ENABLE(ATTACHMENT_ELEMENT)
    case MessageName::WebPageProxy_RequestAttachmentIcon:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RequestAttachmentIcon>(globalObject, decoder);
#endif
#if ENABLE(APP_HIGHLIGHTS)
    case MessageName::WebPageProxy_StoreAppHighlight:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_StoreAppHighlight>(globalObject, decoder);
#endif
#if ENABLE(SPEECH_SYNTHESIS)
    case MessageName::WebPageProxy_SpeechSynthesisVoiceList:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SpeechSynthesisVoiceList>(globalObject, decoder);
    case MessageName::WebPageProxy_SpeechSynthesisSpeak:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SpeechSynthesisSpeak>(globalObject, decoder);
    case MessageName::WebPageProxy_SpeechSynthesisSetFinishedCallback:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SpeechSynthesisSetFinishedCallback>(globalObject, decoder);
    case MessageName::WebPageProxy_SpeechSynthesisCancel:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SpeechSynthesisCancel>(globalObject, decoder);
    case MessageName::WebPageProxy_SpeechSynthesisPause:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SpeechSynthesisPause>(globalObject, decoder);
    case MessageName::WebPageProxy_SpeechSynthesisResume:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SpeechSynthesisResume>(globalObject, decoder);
    case MessageName::WebPageProxy_SpeechSynthesisResetState:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SpeechSynthesisResetState>(globalObject, decoder);
#endif
#if ENABLE(UI_PROCESS_PDF_HUD)
    case MessageName::WebPageProxy_CreatePDFHUD:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_CreatePDFHUD>(globalObject, decoder);
    case MessageName::WebPageProxy_UpdatePDFHUDLocation:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_UpdatePDFHUDLocation>(globalObject, decoder);
    case MessageName::WebPageProxy_RemovePDFHUD:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RemovePDFHUD>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_ConfigureLoggingChannel:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ConfigureLoggingChannel>(globalObject, decoder);
#if PLATFORM(GTK)
    case MessageName::WebPageProxy_ShowEmojiPicker:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ShowEmojiPicker>(globalObject, decoder);
#endif
#if HAVE(VISIBILITY_PROPAGATION_VIEW)
    case MessageName::WebPageProxy_DidCreateContextInWebProcessForVisibilityPropagation:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidCreateContextInWebProcessForVisibilityPropagation>(globalObject, decoder);
#endif
#if ENABLE(WEB_AUTHN)
    case MessageName::WebPageProxy_SetMockWebAuthenticationConfiguration:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetMockWebAuthenticationConfiguration>(globalObject, decoder);
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    case MessageName::WebPageProxy_SendMessageToWebView:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SendMessageToWebView>(globalObject, decoder);
    case MessageName::WebPageProxy_SendMessageToWebViewWithReply:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SendMessageToWebViewWithReply>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_DidFindTextManipulationItems:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidFindTextManipulationItems>(globalObject, decoder);
#if ENABLE(MEDIA_USAGE)
    case MessageName::WebPageProxy_AddMediaUsageManagerSession:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_AddMediaUsageManagerSession>(globalObject, decoder);
    case MessageName::WebPageProxy_UpdateMediaUsageManagerSessionState:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_UpdateMediaUsageManagerSessionState>(globalObject, decoder);
    case MessageName::WebPageProxy_RemoveMediaUsageManagerSession:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RemoveMediaUsageManagerSession>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_SetHasExecutedAppBoundBehaviorBeforeNavigation:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_SetHasExecutedAppBoundBehaviorBeforeNavigation>(globalObject, decoder);
#if PLATFORM(MAC)
    case MessageName::WebPageProxy_ChangeUniversalAccessZoomFocus:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ChangeUniversalAccessZoomFocus>(globalObject, decoder);
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW_IOS)
    case MessageName::WebPageProxy_TakeModelElementFullscreen:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_TakeModelElementFullscreen>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementSetInteractionEnabled:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ModelElementSetInteractionEnabled>(globalObject, decoder);
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW_MAC)
    case MessageName::WebPageProxy_ModelElementCreateRemotePreview:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ModelElementCreateRemotePreview>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementLoadRemotePreview:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ModelElementLoadRemotePreview>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementDestroyRemotePreview:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ModelElementDestroyRemotePreview>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementSizeDidChange:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ModelElementSizeDidChange>(globalObject, decoder);
    case MessageName::WebPageProxy_HandleMouseDownForModelElement:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_HandleMouseDownForModelElement>(globalObject, decoder);
    case MessageName::WebPageProxy_HandleMouseMoveForModelElement:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_HandleMouseMoveForModelElement>(globalObject, decoder);
    case MessageName::WebPageProxy_HandleMouseUpForModelElement:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_HandleMouseUpForModelElement>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelInlinePreviewUUIDs:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ModelInlinePreviewUUIDs>(globalObject, decoder);
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW)
    case MessageName::WebPageProxy_ModelElementGetCamera:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ModelElementGetCamera>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementSetCamera:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ModelElementSetCamera>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementIsPlayingAnimation:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ModelElementIsPlayingAnimation>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementSetAnimationIsPlaying:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ModelElementSetAnimationIsPlaying>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementIsLoopingAnimation:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ModelElementIsLoopingAnimation>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementSetIsLoopingAnimation:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ModelElementSetIsLoopingAnimation>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementAnimationDuration:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ModelElementAnimationDuration>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementAnimationCurrentTime:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ModelElementAnimationCurrentTime>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementSetAnimationCurrentTime:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ModelElementSetAnimationCurrentTime>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementHasAudio:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ModelElementHasAudio>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementIsMuted:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ModelElementIsMuted>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementSetIsMuted:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ModelElementSetIsMuted>(globalObject, decoder);
#endif
#if ENABLE(APPLE_PAY_AMS_UI)
    case MessageName::WebPageProxy_StartApplePayAMSUISession:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_StartApplePayAMSUISession>(globalObject, decoder);
    case MessageName::WebPageProxy_AbortApplePayAMSUISession:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_AbortApplePayAMSUISession>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_RequestCookieConsent:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_RequestCookieConsent>(globalObject, decoder);
    case MessageName::WebPageProxy_ClassifyModalContainerControls:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_ClassifyModalContainerControls>(globalObject, decoder);
    case MessageName::WebPageProxy_DecidePolicyForModalContainer:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DecidePolicyForModalContainer>(globalObject, decoder);
    case MessageName::WebPageProxy_DidChangeLookalikeCharacters:
        return jsValueForDecodedMessage<MessageName::WebPageProxy_DidChangeLookalikeCharacters>(globalObject, decoder);
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebPasteboardProxy_WriteURLToPasteboard:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_WriteURLToPasteboard>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_WriteWebContentToPasteboard:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_WriteWebContentToPasteboard>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_WriteImageToPasteboard:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_WriteImageToPasteboard>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_WriteStringToPasteboard:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_WriteStringToPasteboard>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_UpdateSupportedTypeIdentifiers:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_UpdateSupportedTypeIdentifiers>(globalObject, decoder);
#endif
    case MessageName::WebPasteboardProxy_WriteCustomData:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_WriteCustomData>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_TypesSafeForDOMToReadAndWrite:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_TypesSafeForDOMToReadAndWrite>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_AllPasteboardItemInfo:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_AllPasteboardItemInfo>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_InformationForItemAtIndex:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_InformationForItemAtIndex>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_GetPasteboardItemsCount:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_GetPasteboardItemsCount>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_ReadStringFromPasteboard:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_ReadStringFromPasteboard>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_ReadURLFromPasteboard:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_ReadURLFromPasteboard>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_ReadBufferFromPasteboard:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_ReadBufferFromPasteboard>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_ContainsStringSafeForDOMToReadForType:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_ContainsStringSafeForDOMToReadForType>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::WebPasteboardProxy_GetNumberOfFiles:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_GetNumberOfFiles>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_GetPasteboardTypes:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_GetPasteboardTypes>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_GetPasteboardPathnamesForType:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_GetPasteboardPathnamesForType>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_GetPasteboardStringForType:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_GetPasteboardStringForType>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_GetPasteboardStringsForType:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_GetPasteboardStringsForType>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_GetPasteboardBufferForType:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_GetPasteboardBufferForType>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_GetPasteboardChangeCount:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_GetPasteboardChangeCount>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_GetPasteboardColor:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_GetPasteboardColor>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_GetPasteboardURL:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_GetPasteboardURL>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_AddPasteboardTypes:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_AddPasteboardTypes>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_SetPasteboardTypes:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_SetPasteboardTypes>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_SetPasteboardURL:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_SetPasteboardURL>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_SetPasteboardColor:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_SetPasteboardColor>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_SetPasteboardStringForType:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_SetPasteboardStringForType>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_SetPasteboardBufferForType:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_SetPasteboardBufferForType>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_ContainsURLStringSuitableForLoading:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_ContainsURLStringSuitableForLoading>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_URLStringSuitableForLoading:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_URLStringSuitableForLoading>(globalObject, decoder);
#endif
#if (PLATFORM(COCOA) && ENABLE(IPC_TESTING_API))
    case MessageName::WebPasteboardProxy_TestIPCSharedMemory:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_TestIPCSharedMemory>(globalObject, decoder);
#endif
#if PLATFORM(GTK)
    case MessageName::WebPasteboardProxy_GetTypes:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_GetTypes>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_ReadText:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_ReadText>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_ReadFilePaths:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_ReadFilePaths>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_ReadBuffer:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_ReadBuffer>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_WriteToClipboard:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_WriteToClipboard>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_ClearClipboard:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_ClearClipboard>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_GetPasteboardChangeCount:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_GetPasteboardChangeCount>(globalObject, decoder);
#endif
#if USE(LIBWPE)
    case MessageName::WebPasteboardProxy_GetPasteboardTypes:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_GetPasteboardTypes>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_WriteWebContentToPasteboard:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_WriteWebContentToPasteboard>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_WriteStringToPasteboard:
        return jsValueForDecodedMessage<MessageName::WebPasteboardProxy_WriteStringToPasteboard>(globalObject, decoder);
#endif
    case MessageName::WebPermissionControllerProxy_Query:
        return jsValueForDecodedMessage<MessageName::WebPermissionControllerProxy_Query>(globalObject, decoder);
    case MessageName::WebProcessPool_HandleMessage:
        return jsValueForDecodedMessage<MessageName::WebProcessPool_HandleMessage>(globalObject, decoder);
    case MessageName::WebProcessPool_HandleSynchronousMessage:
        return jsValueForDecodedMessage<MessageName::WebProcessPool_HandleSynchronousMessage>(globalObject, decoder);
#if ENABLE(GAMEPAD)
    case MessageName::WebProcessPool_StartedUsingGamepads:
        return jsValueForDecodedMessage<MessageName::WebProcessPool_StartedUsingGamepads>(globalObject, decoder);
    case MessageName::WebProcessPool_StoppedUsingGamepads:
        return jsValueForDecodedMessage<MessageName::WebProcessPool_StoppedUsingGamepads>(globalObject, decoder);
    case MessageName::WebProcessPool_PlayGamepadEffect:
        return jsValueForDecodedMessage<MessageName::WebProcessPool_PlayGamepadEffect>(globalObject, decoder);
    case MessageName::WebProcessPool_StopGamepadEffects:
        return jsValueForDecodedMessage<MessageName::WebProcessPool_StopGamepadEffects>(globalObject, decoder);
#endif
    case MessageName::WebProcessPool_ReportWebContentCPUTime:
        return jsValueForDecodedMessage<MessageName::WebProcessPool_ReportWebContentCPUTime>(globalObject, decoder);
    case MessageName::WebProcessProxy_UpdateBackForwardItem:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_UpdateBackForwardItem>(globalObject, decoder);
    case MessageName::WebProcessProxy_DidDestroyFrame:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_DidDestroyFrame>(globalObject, decoder);
    case MessageName::WebProcessProxy_DidDestroyUserGestureToken:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_DidDestroyUserGestureToken>(globalObject, decoder);
    case MessageName::WebProcessProxy_ShouldTerminate:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_ShouldTerminate>(globalObject, decoder);
    case MessageName::WebProcessProxy_EnableSuddenTermination:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_EnableSuddenTermination>(globalObject, decoder);
    case MessageName::WebProcessProxy_DisableSuddenTermination:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_DisableSuddenTermination>(globalObject, decoder);
    case MessageName::WebProcessProxy_GetNetworkProcessConnection:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_GetNetworkProcessConnection>(globalObject, decoder);
#if ENABLE(GPU_PROCESS)
    case MessageName::WebProcessProxy_CreateGPUProcessConnection:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_CreateGPUProcessConnection>(globalObject, decoder);
#endif
    case MessageName::WebProcessProxy_SetIsHoldingLockedFiles:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_SetIsHoldingLockedFiles>(globalObject, decoder);
    case MessageName::WebProcessProxy_DidExceedActiveMemoryLimit:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_DidExceedActiveMemoryLimit>(globalObject, decoder);
    case MessageName::WebProcessProxy_DidExceedInactiveMemoryLimit:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_DidExceedInactiveMemoryLimit>(globalObject, decoder);
    case MessageName::WebProcessProxy_DidExceedCPULimit:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_DidExceedCPULimit>(globalObject, decoder);
    case MessageName::WebProcessProxy_StopResponsivenessTimer:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_StopResponsivenessTimer>(globalObject, decoder);
    case MessageName::WebProcessProxy_DidReceiveBackgroundResponsivenessPing:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_DidReceiveBackgroundResponsivenessPing>(globalObject, decoder);
    case MessageName::WebProcessProxy_MemoryPressureStatusChanged:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_MemoryPressureStatusChanged>(globalObject, decoder);
    case MessageName::WebProcessProxy_DidCollectPrewarmInformation:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_DidCollectPrewarmInformation>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::WebProcessProxy_CacheMediaMIMETypes:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_CacheMediaMIMETypes>(globalObject, decoder);
#endif
#if PLATFORM(MAC)
    case MessageName::WebProcessProxy_RequestHighPerformanceGPU:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_RequestHighPerformanceGPU>(globalObject, decoder);
    case MessageName::WebProcessProxy_ReleaseHighPerformanceGPU:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_ReleaseHighPerformanceGPU>(globalObject, decoder);
#endif
#if HAVE(CVDISPLAYLINK)
    case MessageName::WebProcessProxy_StartDisplayLink:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_StartDisplayLink>(globalObject, decoder);
    case MessageName::WebProcessProxy_StopDisplayLink:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_StopDisplayLink>(globalObject, decoder);
    case MessageName::WebProcessProxy_SetDisplayLinkPreferredFramesPerSecond:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_SetDisplayLinkPreferredFramesPerSecond>(globalObject, decoder);
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    case MessageName::WebProcessProxy_SendMessageToWebContext:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_SendMessageToWebContext>(globalObject, decoder);
    case MessageName::WebProcessProxy_SendMessageToWebContextWithReply:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_SendMessageToWebContextWithReply>(globalObject, decoder);
#endif
    case MessageName::WebProcessProxy_DidCreateSleepDisabler:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_DidCreateSleepDisabler>(globalObject, decoder);
    case MessageName::WebProcessProxy_DidDestroySleepDisabler:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_DidDestroySleepDisabler>(globalObject, decoder);
    case MessageName::WebProcessProxy_CreateSpeechRecognitionServer:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_CreateSpeechRecognitionServer>(globalObject, decoder);
    case MessageName::WebProcessProxy_DestroySpeechRecognitionServer:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_DestroySpeechRecognitionServer>(globalObject, decoder);
    case MessageName::WebProcessProxy_SystemBeep:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_SystemBeep>(globalObject, decoder);
#if PLATFORM(MAC)
    case MessageName::WebProcessProxy_IsAXAuthenticated:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_IsAXAuthenticated>(globalObject, decoder);
#endif
#if HAVE(MEDIA_ACCESSIBILITY_FRAMEWORK)
    case MessageName::WebProcessProxy_SetCaptionDisplayMode:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_SetCaptionDisplayMode>(globalObject, decoder);
    case MessageName::WebProcessProxy_SetCaptionLanguage:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_SetCaptionLanguage>(globalObject, decoder);
#endif
    case MessageName::WebProcessProxy_GetNotifications:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_GetNotifications>(globalObject, decoder);
    case MessageName::WebProcessProxy_SetAppBadge:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_SetAppBadge>(globalObject, decoder);
    case MessageName::WebProcessProxy_SetClientBadge:
        return jsValueForDecodedMessage<MessageName::WebProcessProxy_SetClientBadge>(globalObject, decoder);
    case MessageName::WebScreenOrientationManagerProxy_CurrentOrientation:
        return jsValueForDecodedMessage<MessageName::WebScreenOrientationManagerProxy_CurrentOrientation>(globalObject, decoder);
    case MessageName::WebScreenOrientationManagerProxy_Lock:
        return jsValueForDecodedMessage<MessageName::WebScreenOrientationManagerProxy_Lock>(globalObject, decoder);
    case MessageName::WebScreenOrientationManagerProxy_Unlock:
        return jsValueForDecodedMessage<MessageName::WebScreenOrientationManagerProxy_Unlock>(globalObject, decoder);
    case MessageName::WebScreenOrientationManagerProxy_SetShouldSendChangeNotification:
        return jsValueForDecodedMessage<MessageName::WebScreenOrientationManagerProxy_SetShouldSendChangeNotification>(globalObject, decoder);
    case MessageName::WebAutomationSession_DidEvaluateJavaScriptFunction:
        return jsValueForDecodedMessage<MessageName::WebAutomationSession_DidEvaluateJavaScriptFunction>(globalObject, decoder);
    case MessageName::WebAutomationSession_DidTakeScreenshot:
        return jsValueForDecodedMessage<MessageName::WebAutomationSession_DidTakeScreenshot>(globalObject, decoder);
    case MessageName::DownloadProxy_DidStart:
        return jsValueForDecodedMessage<MessageName::DownloadProxy_DidStart>(globalObject, decoder);
    case MessageName::DownloadProxy_DidReceiveAuthenticationChallenge:
        return jsValueForDecodedMessage<MessageName::DownloadProxy_DidReceiveAuthenticationChallenge>(globalObject, decoder);
    case MessageName::DownloadProxy_WillSendRequest:
        return jsValueForDecodedMessage<MessageName::DownloadProxy_WillSendRequest>(globalObject, decoder);
    case MessageName::DownloadProxy_DecideDestinationWithSuggestedFilename:
        return jsValueForDecodedMessage<MessageName::DownloadProxy_DecideDestinationWithSuggestedFilename>(globalObject, decoder);
    case MessageName::DownloadProxy_DidReceiveData:
        return jsValueForDecodedMessage<MessageName::DownloadProxy_DidReceiveData>(globalObject, decoder);
    case MessageName::DownloadProxy_DidCreateDestination:
        return jsValueForDecodedMessage<MessageName::DownloadProxy_DidCreateDestination>(globalObject, decoder);
    case MessageName::DownloadProxy_DidFinish:
        return jsValueForDecodedMessage<MessageName::DownloadProxy_DidFinish>(globalObject, decoder);
    case MessageName::DownloadProxy_DidFail:
        return jsValueForDecodedMessage<MessageName::DownloadProxy_DidFail>(globalObject, decoder);
#if ENABLE(GPU_PROCESS)
#if HAVE(VISIBILITY_PROPAGATION_VIEW)
    case MessageName::GPUProcessProxy_DidCreateContextForVisibilityPropagation:
        return jsValueForDecodedMessage<MessageName::GPUProcessProxy_DidCreateContextForVisibilityPropagation>(globalObject, decoder);
#endif
#if ENABLE(VP9)
    case MessageName::GPUProcessProxy_SetHasVP9HardwareDecoder:
        return jsValueForDecodedMessage<MessageName::GPUProcessProxy_SetHasVP9HardwareDecoder>(globalObject, decoder);
    case MessageName::GPUProcessProxy_SetHasVP9ExtensionSupport:
        return jsValueForDecodedMessage<MessageName::GPUProcessProxy_SetHasVP9ExtensionSupport>(globalObject, decoder);
#endif
#if ENABLE(MEDIA_STREAM) && PLATFORM(IOS_FAMILY)
    case MessageName::GPUProcessProxy_StatusBarWasTapped:
        return jsValueForDecodedMessage<MessageName::GPUProcessProxy_StatusBarWasTapped>(globalObject, decoder);
#endif
    case MessageName::GPUProcessProxy_ProcessIsReadyToExit:
        return jsValueForDecodedMessage<MessageName::GPUProcessProxy_ProcessIsReadyToExit>(globalObject, decoder);
    case MessageName::GPUProcessProxy_TerminateWebProcess:
        return jsValueForDecodedMessage<MessageName::GPUProcessProxy_TerminateWebProcess>(globalObject, decoder);
#endif
    case MessageName::RemoteWebInspectorUIProxy_FrontendLoaded:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUIProxy_FrontendLoaded>(globalObject, decoder);
    case MessageName::RemoteWebInspectorUIProxy_FrontendDidClose:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUIProxy_FrontendDidClose>(globalObject, decoder);
    case MessageName::RemoteWebInspectorUIProxy_Reopen:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUIProxy_Reopen>(globalObject, decoder);
    case MessageName::RemoteWebInspectorUIProxy_ResetState:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUIProxy_ResetState>(globalObject, decoder);
    case MessageName::RemoteWebInspectorUIProxy_BringToFront:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUIProxy_BringToFront>(globalObject, decoder);
    case MessageName::RemoteWebInspectorUIProxy_Save:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUIProxy_Save>(globalObject, decoder);
    case MessageName::RemoteWebInspectorUIProxy_Load:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUIProxy_Load>(globalObject, decoder);
    case MessageName::RemoteWebInspectorUIProxy_PickColorFromScreen:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUIProxy_PickColorFromScreen>(globalObject, decoder);
    case MessageName::RemoteWebInspectorUIProxy_SetForcedAppearance:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUIProxy_SetForcedAppearance>(globalObject, decoder);
    case MessageName::RemoteWebInspectorUIProxy_SetSheetRect:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUIProxy_SetSheetRect>(globalObject, decoder);
    case MessageName::RemoteWebInspectorUIProxy_StartWindowDrag:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUIProxy_StartWindowDrag>(globalObject, decoder);
    case MessageName::RemoteWebInspectorUIProxy_OpenURLExternally:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUIProxy_OpenURLExternally>(globalObject, decoder);
    case MessageName::RemoteWebInspectorUIProxy_RevealFileExternally:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUIProxy_RevealFileExternally>(globalObject, decoder);
    case MessageName::RemoteWebInspectorUIProxy_ShowCertificate:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUIProxy_ShowCertificate>(globalObject, decoder);
    case MessageName::RemoteWebInspectorUIProxy_SendMessageToBackend:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUIProxy_SendMessageToBackend>(globalObject, decoder);
    case MessageName::RemoteWebInspectorUIProxy_SetInspectorPageDeveloperExtrasEnabled:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUIProxy_SetInspectorPageDeveloperExtrasEnabled>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_OpenLocalInspectorFrontend:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_OpenLocalInspectorFrontend>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_SetFrontendConnection:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_SetFrontendConnection>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_SendMessageToBackend:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_SendMessageToBackend>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_FrontendLoaded:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_FrontendLoaded>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_DidClose:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_DidClose>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_BringToFront:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_BringToFront>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_Reopen:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_Reopen>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_ResetState:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_ResetState>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_SetForcedAppearance:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_SetForcedAppearance>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_OpenURLExternally:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_OpenURLExternally>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_RevealFileExternally:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_RevealFileExternally>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_InspectedURLChanged:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_InspectedURLChanged>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_ShowCertificate:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_ShowCertificate>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_ElementSelectionChanged:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_ElementSelectionChanged>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_TimelineRecordingChanged:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_TimelineRecordingChanged>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_SetDeveloperPreferenceOverride:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_SetDeveloperPreferenceOverride>(globalObject, decoder);
#if ENABLE(INSPECTOR_NETWORK_THROTTLING)
    case MessageName::WebInspectorUIProxy_SetEmulatedConditions:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_SetEmulatedConditions>(globalObject, decoder);
#endif
    case MessageName::WebInspectorUIProxy_Save:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_Save>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_Load:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_Load>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_PickColorFromScreen:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_PickColorFromScreen>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_AttachBottom:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_AttachBottom>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_AttachRight:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_AttachRight>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_AttachLeft:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_AttachLeft>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_Detach:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_Detach>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_AttachAvailabilityChanged:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_AttachAvailabilityChanged>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_SetAttachedWindowHeight:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_SetAttachedWindowHeight>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_SetAttachedWindowWidth:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_SetAttachedWindowWidth>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_SetSheetRect:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_SetSheetRect>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_StartWindowDrag:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_StartWindowDrag>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_SetInspectorPageDeveloperExtrasEnabled:
        return jsValueForDecodedMessage<MessageName::WebInspectorUIProxy_SetInspectorPageDeveloperExtrasEnabled>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_DidReceiveAuthenticationChallenge:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_DidReceiveAuthenticationChallenge>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_NegotiatedLegacyTLS:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_NegotiatedLegacyTLS>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_DidNegotiateModernTLS:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_DidNegotiateModernTLS>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_DidFailLoadDueToNetworkConnectionIntegrity:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_DidFailLoadDueToNetworkConnectionIntegrity>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_TestProcessIncomingSyncMessagesWhenWaitingForSyncReply:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_TestProcessIncomingSyncMessagesWhenWaitingForSyncReply>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_TerminateUnresponsiveServiceWorkerProcesses:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_TerminateUnresponsiveServiceWorkerProcesses>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_LogDiagnosticMessage:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_LogDiagnosticMessage>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_LogDiagnosticMessageWithResult:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_LogDiagnosticMessageWithResult>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_LogDiagnosticMessageWithValue:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_LogDiagnosticMessageWithValue>(globalObject, decoder);
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::NetworkProcessProxy_LogTestingEvent:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_LogTestingEvent>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_NotifyResourceLoadStatisticsProcessed:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_NotifyResourceLoadStatisticsProcessed>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_NotifyWebsiteDataDeletionForRegistrableDomainsFinished:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_NotifyWebsiteDataDeletionForRegistrableDomainsFinished>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_NotifyWebsiteDataScanForRegistrableDomainsFinished:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_NotifyWebsiteDataScanForRegistrableDomainsFinished>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_RequestStorageAccessConfirm:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_RequestStorageAccessConfirm>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_DeleteWebsiteDataInUIProcessForRegistrableDomains:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_DeleteWebsiteDataInUIProcessForRegistrableDomains>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_DidCommitCrossSiteLoadWithDataTransferFromPrevalentResource:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_DidCommitCrossSiteLoadWithDataTransferFromPrevalentResource>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_SetDomainsWithUserInteraction:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_SetDomainsWithUserInteraction>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_SetDomainsWithCrossPageStorageAccess:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_SetDomainsWithCrossPageStorageAccess>(globalObject, decoder);
#endif
#if ENABLE(CONTENT_EXTENSIONS)
    case MessageName::NetworkProcessProxy_ContentExtensionRules:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_ContentExtensionRules>(globalObject, decoder);
#endif
    case MessageName::NetworkProcessProxy_TerminateWebProcess:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_TerminateWebProcess>(globalObject, decoder);
#if ENABLE(SERVICE_WORKER)
    case MessageName::NetworkProcessProxy_StartServiceWorkerBackgroundProcessing:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_StartServiceWorkerBackgroundProcessing>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_EndServiceWorkerBackgroundProcessing:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_EndServiceWorkerBackgroundProcessing>(globalObject, decoder);
#endif
    case MessageName::NetworkProcessProxy_EstablishRemoteWorkerContextConnectionToNetworkProcess:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_EstablishRemoteWorkerContextConnectionToNetworkProcess>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_RemoteWorkerContextConnectionNoLongerNeeded:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_RemoteWorkerContextConnectionNoLongerNeeded>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_RegisterRemoteWorkerClientProcess:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_RegisterRemoteWorkerClientProcess>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_UnregisterRemoteWorkerClientProcess:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_UnregisterRemoteWorkerClientProcess>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_SetWebProcessHasUploads:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_SetWebProcessHasUploads>(globalObject, decoder);
#if ENABLE(APP_BOUND_DOMAINS)
    case MessageName::NetworkProcessProxy_GetAppBoundDomains:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_GetAppBoundDomains>(globalObject, decoder);
#endif
    case MessageName::NetworkProcessProxy_RequestStorageSpace:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_RequestStorageSpace>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_IncreaseQuota:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_IncreaseQuota>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_ResourceLoadDidSendRequest:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_ResourceLoadDidSendRequest>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_ResourceLoadDidPerformHTTPRedirection:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_ResourceLoadDidPerformHTTPRedirection>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_ResourceLoadDidReceiveChallenge:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_ResourceLoadDidReceiveChallenge>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_ResourceLoadDidReceiveResponse:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_ResourceLoadDidReceiveResponse>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_ResourceLoadDidCompleteWithError:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_ResourceLoadDidCompleteWithError>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_TriggerBrowsingContextGroupSwitchForNavigation:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_TriggerBrowsingContextGroupSwitchForNavigation>(globalObject, decoder);
#if USE(SOUP)
    case MessageName::NetworkProcessProxy_DidExceedMemoryLimit:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_DidExceedMemoryLimit>(globalObject, decoder);
#endif
#if ENABLE(APPLE_PAY_REMOTE_UI_USES_SCENE)
    case MessageName::NetworkProcessProxy_GetWindowSceneIdentifierForPaymentPresentation:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_GetWindowSceneIdentifierForPaymentPresentation>(globalObject, decoder);
#endif
    case MessageName::NetworkProcessProxy_DataTaskReceivedChallenge:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_DataTaskReceivedChallenge>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_DataTaskWillPerformHTTPRedirection:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_DataTaskWillPerformHTTPRedirection>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_DataTaskDidReceiveResponse:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_DataTaskDidReceiveResponse>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_DataTaskDidReceiveData:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_DataTaskDidReceiveData>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_DataTaskDidCompleteWithError:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_DataTaskDidCompleteWithError>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_OpenWindowFromServiceWorker:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_OpenWindowFromServiceWorker>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_NavigateServiceWorkerClient:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_NavigateServiceWorkerClient>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_CookiesDidChange:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_CookiesDidChange>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_DeleteWebsiteDataInWebProcessesForOrigin:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_DeleteWebsiteDataInWebProcessesForOrigin>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_ReloadExecutionContextsForOrigin:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_ReloadExecutionContextsForOrigin>(globalObject, decoder);
#if ENABLE(NETWORK_ISSUE_REPORTING)
    case MessageName::NetworkProcessProxy_ReportNetworkIssue:
        return jsValueForDecodedMessage<MessageName::NetworkProcessProxy_ReportNetworkIssue>(globalObject, decoder);
#endif
    case MessageName::WebUserContentControllerProxy_DidPostMessage:
        return jsValueForDecodedMessage<MessageName::WebUserContentControllerProxy_DidPostMessage>(globalObject, decoder);
#if ENABLE(WEBXR)
    case MessageName::PlatformXRSystem_EnumerateImmersiveXRDevices:
        return jsValueForDecodedMessage<MessageName::PlatformXRSystem_EnumerateImmersiveXRDevices>(globalObject, decoder);
    case MessageName::PlatformXRSystem_RequestPermissionOnSessionFeatures:
        return jsValueForDecodedMessage<MessageName::PlatformXRSystem_RequestPermissionOnSessionFeatures>(globalObject, decoder);
    case MessageName::PlatformXRSystem_InitializeTrackingAndRendering:
        return jsValueForDecodedMessage<MessageName::PlatformXRSystem_InitializeTrackingAndRendering>(globalObject, decoder);
    case MessageName::PlatformXRSystem_ShutDownTrackingAndRendering:
        return jsValueForDecodedMessage<MessageName::PlatformXRSystem_ShutDownTrackingAndRendering>(globalObject, decoder);
    case MessageName::PlatformXRSystem_RequestFrame:
        return jsValueForDecodedMessage<MessageName::PlatformXRSystem_RequestFrame>(globalObject, decoder);
    case MessageName::PlatformXRSystem_SubmitFrame:
        return jsValueForDecodedMessage<MessageName::PlatformXRSystem_SubmitFrame>(globalObject, decoder);
#endif
    case MessageName::WebProcess_InitializeWebProcess:
        return jsValueForDecodedMessage<MessageName::WebProcess_InitializeWebProcess>(globalObject, decoder);
    case MessageName::WebProcess_SetWebsiteDataStoreParameters:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetWebsiteDataStoreParameters>(globalObject, decoder);
    case MessageName::WebProcess_CreateWebPage:
        return jsValueForDecodedMessage<MessageName::WebProcess_CreateWebPage>(globalObject, decoder);
    case MessageName::WebProcess_PrewarmWithDomainInformation:
        return jsValueForDecodedMessage<MessageName::WebProcess_PrewarmWithDomainInformation>(globalObject, decoder);
    case MessageName::WebProcess_SetCacheModel:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetCacheModel>(globalObject, decoder);
    case MessageName::WebProcess_RegisterURLSchemeAsEmptyDocument:
        return jsValueForDecodedMessage<MessageName::WebProcess_RegisterURLSchemeAsEmptyDocument>(globalObject, decoder);
    case MessageName::WebProcess_RegisterURLSchemeAsSecure:
        return jsValueForDecodedMessage<MessageName::WebProcess_RegisterURLSchemeAsSecure>(globalObject, decoder);
    case MessageName::WebProcess_RegisterURLSchemeAsBypassingContentSecurityPolicy:
        return jsValueForDecodedMessage<MessageName::WebProcess_RegisterURLSchemeAsBypassingContentSecurityPolicy>(globalObject, decoder);
    case MessageName::WebProcess_SetDomainRelaxationForbiddenForURLScheme:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetDomainRelaxationForbiddenForURLScheme>(globalObject, decoder);
    case MessageName::WebProcess_RegisterURLSchemeAsLocal:
        return jsValueForDecodedMessage<MessageName::WebProcess_RegisterURLSchemeAsLocal>(globalObject, decoder);
    case MessageName::WebProcess_RegisterURLSchemeAsNoAccess:
        return jsValueForDecodedMessage<MessageName::WebProcess_RegisterURLSchemeAsNoAccess>(globalObject, decoder);
    case MessageName::WebProcess_RegisterURLSchemeAsDisplayIsolated:
        return jsValueForDecodedMessage<MessageName::WebProcess_RegisterURLSchemeAsDisplayIsolated>(globalObject, decoder);
    case MessageName::WebProcess_RegisterURLSchemeAsCORSEnabled:
        return jsValueForDecodedMessage<MessageName::WebProcess_RegisterURLSchemeAsCORSEnabled>(globalObject, decoder);
    case MessageName::WebProcess_RegisterURLSchemeAsCachePartitioned:
        return jsValueForDecodedMessage<MessageName::WebProcess_RegisterURLSchemeAsCachePartitioned>(globalObject, decoder);
    case MessageName::WebProcess_RegisterURLSchemeAsCanDisplayOnlyIfCanRequest:
        return jsValueForDecodedMessage<MessageName::WebProcess_RegisterURLSchemeAsCanDisplayOnlyIfCanRequest>(globalObject, decoder);
    case MessageName::WebProcess_SetDefaultRequestTimeoutInterval:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetDefaultRequestTimeoutInterval>(globalObject, decoder);
    case MessageName::WebProcess_SetAlwaysUsesComplexTextCodePath:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetAlwaysUsesComplexTextCodePath>(globalObject, decoder);
    case MessageName::WebProcess_SetDisableFontSubpixelAntialiasingForTesting:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetDisableFontSubpixelAntialiasingForTesting>(globalObject, decoder);
    case MessageName::WebProcess_SetTrackingPreventionEnabled:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetTrackingPreventionEnabled>(globalObject, decoder);
    case MessageName::WebProcess_ClearResourceLoadStatistics:
        return jsValueForDecodedMessage<MessageName::WebProcess_ClearResourceLoadStatistics>(globalObject, decoder);
    case MessageName::WebProcess_UserPreferredLanguagesChanged:
        return jsValueForDecodedMessage<MessageName::WebProcess_UserPreferredLanguagesChanged>(globalObject, decoder);
    case MessageName::WebProcess_FullKeyboardAccessModeChanged:
        return jsValueForDecodedMessage<MessageName::WebProcess_FullKeyboardAccessModeChanged>(globalObject, decoder);
#if HAVE(MOUSE_DEVICE_OBSERVATION)
    case MessageName::WebProcess_SetHasMouseDevice:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetHasMouseDevice>(globalObject, decoder);
#endif
#if HAVE(STYLUS_DEVICE_OBSERVATION)
    case MessageName::WebProcess_SetHasStylusDevice:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetHasStylusDevice>(globalObject, decoder);
#endif
    case MessageName::WebProcess_StartMemorySampler:
        return jsValueForDecodedMessage<MessageName::WebProcess_StartMemorySampler>(globalObject, decoder);
    case MessageName::WebProcess_StopMemorySampler:
        return jsValueForDecodedMessage<MessageName::WebProcess_StopMemorySampler>(globalObject, decoder);
    case MessageName::WebProcess_SetTextCheckerState:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetTextCheckerState>(globalObject, decoder);
    case MessageName::WebProcess_SetEnhancedAccessibility:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetEnhancedAccessibility>(globalObject, decoder);
    case MessageName::WebProcess_GarbageCollectJavaScriptObjects:
        return jsValueForDecodedMessage<MessageName::WebProcess_GarbageCollectJavaScriptObjects>(globalObject, decoder);
    case MessageName::WebProcess_SetJavaScriptGarbageCollectorTimerEnabled:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetJavaScriptGarbageCollectorTimerEnabled>(globalObject, decoder);
    case MessageName::WebProcess_SetInjectedBundleParameter:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetInjectedBundleParameter>(globalObject, decoder);
    case MessageName::WebProcess_SetInjectedBundleParameters:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetInjectedBundleParameters>(globalObject, decoder);
    case MessageName::WebProcess_HandleInjectedBundleMessage:
        return jsValueForDecodedMessage<MessageName::WebProcess_HandleInjectedBundleMessage>(globalObject, decoder);
    case MessageName::WebProcess_FetchWebsiteData:
        return jsValueForDecodedMessage<MessageName::WebProcess_FetchWebsiteData>(globalObject, decoder);
    case MessageName::WebProcess_DeleteWebsiteData:
        return jsValueForDecodedMessage<MessageName::WebProcess_DeleteWebsiteData>(globalObject, decoder);
    case MessageName::WebProcess_DeleteWebsiteDataForOrigins:
        return jsValueForDecodedMessage<MessageName::WebProcess_DeleteWebsiteDataForOrigins>(globalObject, decoder);
    case MessageName::WebProcess_DeleteWebsiteDataForOrigin:
        return jsValueForDecodedMessage<MessageName::WebProcess_DeleteWebsiteDataForOrigin>(globalObject, decoder);
    case MessageName::WebProcess_ReloadExecutionContextsForOrigin:
        return jsValueForDecodedMessage<MessageName::WebProcess_ReloadExecutionContextsForOrigin>(globalObject, decoder);
    case MessageName::WebProcess_DeleteAllCookies:
        return jsValueForDecodedMessage<MessageName::WebProcess_DeleteAllCookies>(globalObject, decoder);
    case MessageName::WebProcess_SetHiddenPageDOMTimerThrottlingIncreaseLimit:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetHiddenPageDOMTimerThrottlingIncreaseLimit>(globalObject, decoder);
    case MessageName::WebProcess_SetMemoryCacheDisabled:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetMemoryCacheDisabled>(globalObject, decoder);
#if ENABLE(SERVICE_CONTROLS)
    case MessageName::WebProcess_SetEnabledServices:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetEnabledServices>(globalObject, decoder);
#endif
    case MessageName::WebProcess_EnsureAutomationSessionProxy:
        return jsValueForDecodedMessage<MessageName::WebProcess_EnsureAutomationSessionProxy>(globalObject, decoder);
    case MessageName::WebProcess_DestroyAutomationSessionProxy:
        return jsValueForDecodedMessage<MessageName::WebProcess_DestroyAutomationSessionProxy>(globalObject, decoder);
    case MessageName::WebProcess_PrepareToSuspend:
        return jsValueForDecodedMessage<MessageName::WebProcess_PrepareToSuspend>(globalObject, decoder);
    case MessageName::WebProcess_ProcessDidResume:
        return jsValueForDecodedMessage<MessageName::WebProcess_ProcessDidResume>(globalObject, decoder);
    case MessageName::WebProcess_BackgroundResponsivenessPing:
        return jsValueForDecodedMessage<MessageName::WebProcess_BackgroundResponsivenessPing>(globalObject, decoder);
#if ENABLE(GAMEPAD)
    case MessageName::WebProcess_SetInitialGamepads:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetInitialGamepads>(globalObject, decoder);
    case MessageName::WebProcess_GamepadConnected:
        return jsValueForDecodedMessage<MessageName::WebProcess_GamepadConnected>(globalObject, decoder);
    case MessageName::WebProcess_GamepadDisconnected:
        return jsValueForDecodedMessage<MessageName::WebProcess_GamepadDisconnected>(globalObject, decoder);
#endif
    case MessageName::WebProcess_EstablishRemoteWorkerContextConnectionToNetworkProcess:
        return jsValueForDecodedMessage<MessageName::WebProcess_EstablishRemoteWorkerContextConnectionToNetworkProcess>(globalObject, decoder);
    case MessageName::WebProcess_SetHasSuspendedPageProxy:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetHasSuspendedPageProxy>(globalObject, decoder);
    case MessageName::WebProcess_SetIsInProcessCache:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetIsInProcessCache>(globalObject, decoder);
    case MessageName::WebProcess_MarkIsNoLongerPrewarmed:
        return jsValueForDecodedMessage<MessageName::WebProcess_MarkIsNoLongerPrewarmed>(globalObject, decoder);
    case MessageName::WebProcess_GetActivePagesOriginsForTesting:
        return jsValueForDecodedMessage<MessageName::WebProcess_GetActivePagesOriginsForTesting>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::WebProcess_SetScreenProperties:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetScreenProperties>(globalObject, decoder);
#endif
#if PLATFORM(MAC)
    case MessageName::WebProcess_ScrollerStylePreferenceChanged:
        return jsValueForDecodedMessage<MessageName::WebProcess_ScrollerStylePreferenceChanged>(globalObject, decoder);
    case MessageName::WebProcess_DisplayConfigurationChanged:
        return jsValueForDecodedMessage<MessageName::WebProcess_DisplayConfigurationChanged>(globalObject, decoder);
#endif
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebProcess_UserInterfaceIdiomDidChange:
        return jsValueForDecodedMessage<MessageName::WebProcess_UserInterfaceIdiomDidChange>(globalObject, decoder);
#endif
#if PLATFORM(IOS_FAMILY) && !PLATFORM(MACCATALYST)
    case MessageName::WebProcess_BacklightLevelDidChange:
        return jsValueForDecodedMessage<MessageName::WebProcess_BacklightLevelDidChange>(globalObject, decoder);
#endif
#if PLATFORM(MAC) || PLATFORM(MACCATALYST)
    case MessageName::WebProcess_ColorPreferencesDidChange:
        return jsValueForDecodedMessage<MessageName::WebProcess_ColorPreferencesDidChange>(globalObject, decoder);
#endif
    case MessageName::WebProcess_IsJITEnabled:
        return jsValueForDecodedMessage<MessageName::WebProcess_IsJITEnabled>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::WebProcess_SetMediaMIMETypes:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetMediaMIMETypes>(globalObject, decoder);
#endif
#if (PLATFORM(COCOA) && ENABLE(REMOTE_INSPECTOR))
    case MessageName::WebProcess_EnableRemoteWebInspector:
        return jsValueForDecodedMessage<MessageName::WebProcess_EnableRemoteWebInspector>(globalObject, decoder);
#endif
#if ENABLE(MEDIA_STREAM)
    case MessageName::WebProcess_AddMockMediaDevice:
        return jsValueForDecodedMessage<MessageName::WebProcess_AddMockMediaDevice>(globalObject, decoder);
    case MessageName::WebProcess_ClearMockMediaDevices:
        return jsValueForDecodedMessage<MessageName::WebProcess_ClearMockMediaDevices>(globalObject, decoder);
    case MessageName::WebProcess_RemoveMockMediaDevice:
        return jsValueForDecodedMessage<MessageName::WebProcess_RemoveMockMediaDevice>(globalObject, decoder);
    case MessageName::WebProcess_SetMockMediaDeviceIsEphemeral:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetMockMediaDeviceIsEphemeral>(globalObject, decoder);
    case MessageName::WebProcess_ResetMockMediaDevices:
        return jsValueForDecodedMessage<MessageName::WebProcess_ResetMockMediaDevices>(globalObject, decoder);
#endif
#if (ENABLE(MEDIA_STREAM) && ENABLE(SANDBOX_EXTENSIONS))
    case MessageName::WebProcess_GrantUserMediaDeviceSandboxExtensions:
        return jsValueForDecodedMessage<MessageName::WebProcess_GrantUserMediaDeviceSandboxExtensions>(globalObject, decoder);
    case MessageName::WebProcess_RevokeUserMediaDeviceSandboxExtensions:
        return jsValueForDecodedMessage<MessageName::WebProcess_RevokeUserMediaDeviceSandboxExtensions>(globalObject, decoder);
#endif
    case MessageName::WebProcess_ClearCurrentModifierStateForTesting:
        return jsValueForDecodedMessage<MessageName::WebProcess_ClearCurrentModifierStateForTesting>(globalObject, decoder);
    case MessageName::WebProcess_SetBackForwardCacheCapacity:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetBackForwardCacheCapacity>(globalObject, decoder);
    case MessageName::WebProcess_ClearCachedPage:
        return jsValueForDecodedMessage<MessageName::WebProcess_ClearCachedPage>(globalObject, decoder);
#if PLATFORM(GTK) || PLATFORM(WPE)
    case MessageName::WebProcess_SendMessageToWebProcessExtension:
        return jsValueForDecodedMessage<MessageName::WebProcess_SendMessageToWebProcessExtension>(globalObject, decoder);
#endif
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::WebProcess_SeedResourceLoadStatisticsForTesting:
        return jsValueForDecodedMessage<MessageName::WebProcess_SeedResourceLoadStatisticsForTesting>(globalObject, decoder);
    case MessageName::WebProcess_SetThirdPartyCookieBlockingMode:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetThirdPartyCookieBlockingMode>(globalObject, decoder);
    case MessageName::WebProcess_SetDomainsWithUserInteraction:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetDomainsWithUserInteraction>(globalObject, decoder);
    case MessageName::WebProcess_SetDomainsWithCrossPageStorageAccess:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetDomainsWithCrossPageStorageAccess>(globalObject, decoder);
    case MessageName::WebProcess_SendResourceLoadStatisticsDataImmediately:
        return jsValueForDecodedMessage<MessageName::WebProcess_SendResourceLoadStatisticsDataImmediately>(globalObject, decoder);
#endif
    case MessageName::WebProcess_GrantAccessToAssetServices:
        return jsValueForDecodedMessage<MessageName::WebProcess_GrantAccessToAssetServices>(globalObject, decoder);
    case MessageName::WebProcess_RevokeAccessToAssetServices:
        return jsValueForDecodedMessage<MessageName::WebProcess_RevokeAccessToAssetServices>(globalObject, decoder);
    case MessageName::WebProcess_SwitchFromStaticFontRegistryToUserFontRegistry:
        return jsValueForDecodedMessage<MessageName::WebProcess_SwitchFromStaticFontRegistryToUserFontRegistry>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::WebProcess_DisableURLSchemeCheckInDataDetectors:
        return jsValueForDecodedMessage<MessageName::WebProcess_DisableURLSchemeCheckInDataDetectors>(globalObject, decoder);
    case MessageName::WebProcess_UnblockServicesRequiredByAccessibility:
        return jsValueForDecodedMessage<MessageName::WebProcess_UnblockServicesRequiredByAccessibility>(globalObject, decoder);
#endif
#if (PLATFORM(COCOA) && ENABLE(CFPREFS_DIRECT_MODE))
    case MessageName::WebProcess_NotifyPreferencesChanged:
        return jsValueForDecodedMessage<MessageName::WebProcess_NotifyPreferencesChanged>(globalObject, decoder);
#endif
#if PLATFORM(COCOA)
    case MessageName::WebProcess_PowerSourceDidChange:
        return jsValueForDecodedMessage<MessageName::WebProcess_PowerSourceDidChange>(globalObject, decoder);
#endif
#if PLATFORM(GTK) && !USE(GTK4)
    case MessageName::WebProcess_SetUseSystemAppearanceForScrollbars:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetUseSystemAppearanceForScrollbars>(globalObject, decoder);
#endif
#if PLATFORM(COCOA)
    case MessageName::WebProcess_DidWriteToPasteboardAsynchronously:
        return jsValueForDecodedMessage<MessageName::WebProcess_DidWriteToPasteboardAsynchronously>(globalObject, decoder);
#endif
#if HAVE(CVDISPLAYLINK)
    case MessageName::WebProcess_DisplayWasRefreshed:
        return jsValueForDecodedMessage<MessageName::WebProcess_DisplayWasRefreshed>(globalObject, decoder);
#endif
#if PLATFORM(MAC)
    case MessageName::WebProcess_SystemWillPowerOn:
        return jsValueForDecodedMessage<MessageName::WebProcess_SystemWillPowerOn>(globalObject, decoder);
    case MessageName::WebProcess_SystemWillSleep:
        return jsValueForDecodedMessage<MessageName::WebProcess_SystemWillSleep>(globalObject, decoder);
    case MessageName::WebProcess_SystemDidWake:
        return jsValueForDecodedMessage<MessageName::WebProcess_SystemDidWake>(globalObject, decoder);
#endif
#if HAVE(AUDIO_COMPONENT_SERVER_REGISTRATIONS)
    case MessageName::WebProcess_ConsumeAudioComponentRegistrations:
        return jsValueForDecodedMessage<MessageName::WebProcess_ConsumeAudioComponentRegistrations>(globalObject, decoder);
#endif
#if PLATFORM(COCOA)
    case MessageName::WebProcess_AccessibilityPreferencesDidChange:
        return jsValueForDecodedMessage<MessageName::WebProcess_AccessibilityPreferencesDidChange>(globalObject, decoder);
#endif
#if HAVE(MEDIA_ACCESSIBILITY_FRAMEWORK)
    case MessageName::WebProcess_SetMediaAccessibilityPreferences:
        return jsValueForDecodedMessage<MessageName::WebProcess_SetMediaAccessibilityPreferences>(globalObject, decoder);
#endif
#if PLATFORM(MAC)
    case MessageName::WebProcess_OpenDirectoryCacheInvalidated:
        return jsValueForDecodedMessage<MessageName::WebProcess_OpenDirectoryCacheInvalidated>(globalObject, decoder);
#endif
    case MessageName::WebAutomationSessionProxy_EvaluateJavaScriptFunction:
        return jsValueForDecodedMessage<MessageName::WebAutomationSessionProxy_EvaluateJavaScriptFunction>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_ResolveChildFrameWithOrdinal:
        return jsValueForDecodedMessage<MessageName::WebAutomationSessionProxy_ResolveChildFrameWithOrdinal>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_ResolveChildFrameWithNodeHandle:
        return jsValueForDecodedMessage<MessageName::WebAutomationSessionProxy_ResolveChildFrameWithNodeHandle>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_ResolveChildFrameWithName:
        return jsValueForDecodedMessage<MessageName::WebAutomationSessionProxy_ResolveChildFrameWithName>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_ResolveParentFrame:
        return jsValueForDecodedMessage<MessageName::WebAutomationSessionProxy_ResolveParentFrame>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_ComputeElementLayout:
        return jsValueForDecodedMessage<MessageName::WebAutomationSessionProxy_ComputeElementLayout>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_GetComputedRole:
        return jsValueForDecodedMessage<MessageName::WebAutomationSessionProxy_GetComputedRole>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_GetComputedLabel:
        return jsValueForDecodedMessage<MessageName::WebAutomationSessionProxy_GetComputedLabel>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_SelectOptionElement:
        return jsValueForDecodedMessage<MessageName::WebAutomationSessionProxy_SelectOptionElement>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_SetFilesForInputFileUpload:
        return jsValueForDecodedMessage<MessageName::WebAutomationSessionProxy_SetFilesForInputFileUpload>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_TakeScreenshot:
        return jsValueForDecodedMessage<MessageName::WebAutomationSessionProxy_TakeScreenshot>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_SnapshotRectForScreenshot:
        return jsValueForDecodedMessage<MessageName::WebAutomationSessionProxy_SnapshotRectForScreenshot>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_GetCookiesForFrame:
        return jsValueForDecodedMessage<MessageName::WebAutomationSessionProxy_GetCookiesForFrame>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_DeleteCookie:
        return jsValueForDecodedMessage<MessageName::WebAutomationSessionProxy_DeleteCookie>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_DidDeleteDatabase:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_DidDeleteDatabase>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_DidOpenDatabase:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_DidOpenDatabase>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_DidAbortTransaction:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_DidAbortTransaction>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_DidCommitTransaction:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_DidCommitTransaction>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_DidCreateObjectStore:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_DidCreateObjectStore>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_DidDeleteObjectStore:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_DidDeleteObjectStore>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_DidRenameObjectStore:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_DidRenameObjectStore>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_DidClearObjectStore:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_DidClearObjectStore>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_DidCreateIndex:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_DidCreateIndex>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_DidDeleteIndex:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_DidDeleteIndex>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_DidRenameIndex:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_DidRenameIndex>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_DidPutOrAdd:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_DidPutOrAdd>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_DidGetRecord:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_DidGetRecord>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_DidGetAllRecords:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_DidGetAllRecords>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_DidGetCount:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_DidGetCount>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_DidDeleteRecord:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_DidDeleteRecord>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_DidOpenCursor:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_DidOpenCursor>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_DidIterateCursor:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_DidIterateCursor>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_FireVersionChangeEvent:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_FireVersionChangeEvent>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_DidStartTransaction:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_DidStartTransaction>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_DidCloseFromServer:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_DidCloseFromServer>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_NotifyOpenDBRequestBlocked:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_NotifyOpenDBRequestBlocked>(globalObject, decoder);
    case MessageName::WebIDBConnectionToServer_DidGetAllDatabaseNamesAndVersions:
        return jsValueForDecodedMessage<MessageName::WebIDBConnectionToServer_DidGetAllDatabaseNamesAndVersions>(globalObject, decoder);
#if ENABLE(FULLSCREEN_API)
    case MessageName::WebFullScreenManager_RequestRestoreFullScreen:
        return jsValueForDecodedMessage<MessageName::WebFullScreenManager_RequestRestoreFullScreen>(globalObject, decoder);
    case MessageName::WebFullScreenManager_RequestExitFullScreen:
        return jsValueForDecodedMessage<MessageName::WebFullScreenManager_RequestExitFullScreen>(globalObject, decoder);
    case MessageName::WebFullScreenManager_WillEnterFullScreen:
        return jsValueForDecodedMessage<MessageName::WebFullScreenManager_WillEnterFullScreen>(globalObject, decoder);
    case MessageName::WebFullScreenManager_DidEnterFullScreen:
        return jsValueForDecodedMessage<MessageName::WebFullScreenManager_DidEnterFullScreen>(globalObject, decoder);
    case MessageName::WebFullScreenManager_WillExitFullScreen:
        return jsValueForDecodedMessage<MessageName::WebFullScreenManager_WillExitFullScreen>(globalObject, decoder);
    case MessageName::WebFullScreenManager_DidExitFullScreen:
        return jsValueForDecodedMessage<MessageName::WebFullScreenManager_DidExitFullScreen>(globalObject, decoder);
    case MessageName::WebFullScreenManager_SetAnimatingFullScreen:
        return jsValueForDecodedMessage<MessageName::WebFullScreenManager_SetAnimatingFullScreen>(globalObject, decoder);
    case MessageName::WebFullScreenManager_SaveScrollPosition:
        return jsValueForDecodedMessage<MessageName::WebFullScreenManager_SaveScrollPosition>(globalObject, decoder);
    case MessageName::WebFullScreenManager_RestoreScrollPosition:
        return jsValueForDecodedMessage<MessageName::WebFullScreenManager_RestoreScrollPosition>(globalObject, decoder);
    case MessageName::WebFullScreenManager_SetFullscreenInsets:
        return jsValueForDecodedMessage<MessageName::WebFullScreenManager_SetFullscreenInsets>(globalObject, decoder);
    case MessageName::WebFullScreenManager_SetFullscreenAutoHideDuration:
        return jsValueForDecodedMessage<MessageName::WebFullScreenManager_SetFullscreenAutoHideDuration>(globalObject, decoder);
    case MessageName::WebFullScreenManager_SetFullscreenControlsHidden:
        return jsValueForDecodedMessage<MessageName::WebFullScreenManager_SetFullscreenControlsHidden>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::GPUProcessConnection_DidInitialize:
        return jsValueForDecodedMessage<MessageName::GPUProcessConnection_DidInitialize>(globalObject, decoder);
    case MessageName::GPUProcessConnection_DidReceiveRemoteCommand:
        return jsValueForDecodedMessage<MessageName::GPUProcessConnection_DidReceiveRemoteCommand>(globalObject, decoder);
#if ENABLE(ROUTING_ARBITRATION)
    case MessageName::GPUProcessConnection_BeginRoutingArbitrationWithCategory:
        return jsValueForDecodedMessage<MessageName::GPUProcessConnection_BeginRoutingArbitrationWithCategory>(globalObject, decoder);
    case MessageName::GPUProcessConnection_EndRoutingArbitration:
        return jsValueForDecodedMessage<MessageName::GPUProcessConnection_EndRoutingArbitration>(globalObject, decoder);
#endif
#if ENABLE(MEDIA_STREAM) && PLATFORM(COCOA)
    case MessageName::GPUProcessConnection_ResetAudioMediaStreamTrackRendererInternalUnit:
        return jsValueForDecodedMessage<MessageName::GPUProcessConnection_ResetAudioMediaStreamTrackRendererInternalUnit>(globalObject, decoder);
#endif
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(WEBGL)
    case MessageName::RemoteGraphicsContextGLProxy_WasCreated:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGLProxy_WasCreated>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGLProxy_WasLost:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGLProxy_WasLost>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGLProxy_WasChanged:
        return jsValueForDecodedMessage<MessageName::RemoteGraphicsContextGLProxy_WasChanged>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteRenderingBackendProxy_DidCreateImageBufferBackend:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackendProxy_DidCreateImageBufferBackend>(globalObject, decoder);
    case MessageName::RemoteRenderingBackendProxy_DidFlush:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackendProxy_DidFlush>(globalObject, decoder);
    case MessageName::RemoteRenderingBackendProxy_DidFinalizeRenderingUpdate:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackendProxy_DidFinalizeRenderingUpdate>(globalObject, decoder);
    case MessageName::RemoteRenderingBackendProxy_DidInitialize:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackendProxy_DidInitialize>(globalObject, decoder);
    case MessageName::RemoteRenderingBackendProxy_DidMarkLayersAsVolatile:
        return jsValueForDecodedMessage<MessageName::RemoteRenderingBackendProxy_DidMarkLayersAsVolatile>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteGPUProxy_WasCreated:
        return jsValueForDecodedMessage<MessageName::RemoteGPUProxy_WasCreated>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteAudioHardwareListener_AudioHardwareDidBecomeActive:
        return jsValueForDecodedMessage<MessageName::RemoteAudioHardwareListener_AudioHardwareDidBecomeActive>(globalObject, decoder);
    case MessageName::RemoteAudioHardwareListener_AudioHardwareDidBecomeInactive:
        return jsValueForDecodedMessage<MessageName::RemoteAudioHardwareListener_AudioHardwareDidBecomeInactive>(globalObject, decoder);
    case MessageName::RemoteAudioHardwareListener_AudioOutputDeviceChanged:
        return jsValueForDecodedMessage<MessageName::RemoteAudioHardwareListener_AudioOutputDeviceChanged>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
    case MessageName::MediaPlayerPrivateRemote_NetworkStateChanged:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_NetworkStateChanged>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_ReadyStateChanged:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_ReadyStateChanged>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_FirstVideoFrameAvailable:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_FirstVideoFrameAvailable>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_VolumeChanged:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_VolumeChanged>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_MuteChanged:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_MuteChanged>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_TimeChanged:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_TimeChanged>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_DurationChanged:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_DurationChanged>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_RateChanged:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_RateChanged>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_PlaybackStateChanged:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_PlaybackStateChanged>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_EngineFailedToLoad:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_EngineFailedToLoad>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_UpdateCachedState:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_UpdateCachedState>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_CharacteristicChanged:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_CharacteristicChanged>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_SizeChanged:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_SizeChanged>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_RenderingModeChanged:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_RenderingModeChanged>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_CurrentTimeChanged:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_CurrentTimeChanged>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_AddRemoteAudioTrack:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_AddRemoteAudioTrack>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_RemoveRemoteAudioTrack:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_RemoveRemoteAudioTrack>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_RemoteAudioTrackConfigurationChanged:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_RemoteAudioTrackConfigurationChanged>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_AddRemoteVideoTrack:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_AddRemoteVideoTrack>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_RemoveRemoteVideoTrack:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_RemoveRemoteVideoTrack>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_RemoteVideoTrackConfigurationChanged:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_RemoteVideoTrackConfigurationChanged>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_AddRemoteTextTrack:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_AddRemoteTextTrack>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_RemoveRemoteTextTrack:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_RemoveRemoteTextTrack>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_RemoteTextTrackConfigurationChanged:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_RemoteTextTrackConfigurationChanged>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_ParseWebVTTFileHeader:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_ParseWebVTTFileHeader>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_ParseWebVTTCueData:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_ParseWebVTTCueData>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_ParseWebVTTCueDataStruct:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_ParseWebVTTCueDataStruct>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_AddDataCue:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_AddDataCue>(globalObject, decoder);
#if ENABLE(DATACUE_VALUE)
    case MessageName::MediaPlayerPrivateRemote_AddDataCueWithType:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_AddDataCueWithType>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_UpdateDataCue:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_UpdateDataCue>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_RemoveDataCue:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_RemoveDataCue>(globalObject, decoder);
#endif
    case MessageName::MediaPlayerPrivateRemote_AddGenericCue:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_AddGenericCue>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_UpdateGenericCue:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_UpdateGenericCue>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_RemoveGenericCue:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_RemoveGenericCue>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_RequestResource:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_RequestResource>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_RemoveResource:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_RemoveResource>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_SendH2Ping:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_SendH2Ping>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_ResourceNotSupported:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_ResourceNotSupported>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_ActiveSourceBuffersChanged:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_ActiveSourceBuffersChanged>(globalObject, decoder);
#if ENABLE(ENCRYPTED_MEDIA)
    case MessageName::MediaPlayerPrivateRemote_WaitingForKeyChanged:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_WaitingForKeyChanged>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_InitializationDataEncountered:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_InitializationDataEncountered>(globalObject, decoder);
#endif
#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
    case MessageName::MediaPlayerPrivateRemote_MediaPlayerKeyNeeded:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_MediaPlayerKeyNeeded>(globalObject, decoder);
#endif
#if ENABLE(WIRELESS_PLAYBACK_TARGET)
    case MessageName::MediaPlayerPrivateRemote_CurrentPlaybackTargetIsWirelessChanged:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_CurrentPlaybackTargetIsWirelessChanged>(globalObject, decoder);
#endif
#if PLATFORM(IOS_FAMILY)
    case MessageName::MediaPlayerPrivateRemote_GetRawCookies:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_GetRawCookies>(globalObject, decoder);
#endif
#if PLATFORM(COCOA)
    case MessageName::MediaPlayerPrivateRemote_PushVideoFrameMetadata:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_PushVideoFrameMetadata>(globalObject, decoder);
    case MessageName::MediaPlayerPrivateRemote_LayerHostingContextIdChanged:
        return jsValueForDecodedMessage<MessageName::MediaPlayerPrivateRemote_LayerHostingContextIdChanged>(globalObject, decoder);
#endif
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
    case MessageName::MediaSourcePrivateRemote_SeekToTime:
        return jsValueForDecodedMessage<MessageName::MediaSourcePrivateRemote_SeekToTime>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && USE(AUDIO_SESSION)
    case MessageName::RemoteAudioSession_ConfigurationChanged:
        return jsValueForDecodedMessage<MessageName::RemoteAudioSession_ConfigurationChanged>(globalObject, decoder);
    case MessageName::RemoteAudioSession_BeginInterruption:
        return jsValueForDecodedMessage<MessageName::RemoteAudioSession_BeginInterruption>(globalObject, decoder);
    case MessageName::RemoteAudioSession_EndInterruption:
        return jsValueForDecodedMessage<MessageName::RemoteAudioSession_EndInterruption>(globalObject, decoder);
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS)
    case MessageName::RemoteAudioSourceProviderManager_AudioStorageChanged:
        return jsValueForDecodedMessage<MessageName::RemoteAudioSourceProviderManager_AudioStorageChanged>(globalObject, decoder);
    case MessageName::RemoteAudioSourceProviderManager_AudioSamplesAvailable:
        return jsValueForDecodedMessage<MessageName::RemoteAudioSourceProviderManager_AudioSamplesAvailable>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteCDMInstance_UnrequestedInitializationDataReceived:
        return jsValueForDecodedMessage<MessageName::RemoteCDMInstance_UnrequestedInitializationDataReceived>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteCDMInstanceSession_UpdateKeyStatuses:
        return jsValueForDecodedMessage<MessageName::RemoteCDMInstanceSession_UpdateKeyStatuses>(globalObject, decoder);
    case MessageName::RemoteCDMInstanceSession_SendMessage:
        return jsValueForDecodedMessage<MessageName::RemoteCDMInstanceSession_SendMessage>(globalObject, decoder);
    case MessageName::RemoteCDMInstanceSession_SessionIdChanged:
        return jsValueForDecodedMessage<MessageName::RemoteCDMInstanceSession_SessionIdChanged>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
    case MessageName::RemoteLegacyCDMSession_SendMessage:
        return jsValueForDecodedMessage<MessageName::RemoteLegacyCDMSession_SendMessage>(globalObject, decoder);
    case MessageName::RemoteLegacyCDMSession_SendError:
        return jsValueForDecodedMessage<MessageName::RemoteLegacyCDMSession_SendError>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteRemoteCommandListener_DidReceiveRemoteControlCommand:
        return jsValueForDecodedMessage<MessageName::RemoteRemoteCommandListener_DidReceiveRemoteControlCommand>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateDidReceiveInitializationSegment:
        return jsValueForDecodedMessage<MessageName::SourceBufferPrivateRemote_SourceBufferPrivateDidReceiveInitializationSegment>(globalObject, decoder);
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateStreamEndedWithDecodeError:
        return jsValueForDecodedMessage<MessageName::SourceBufferPrivateRemote_SourceBufferPrivateStreamEndedWithDecodeError>(globalObject, decoder);
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateAppendError:
        return jsValueForDecodedMessage<MessageName::SourceBufferPrivateRemote_SourceBufferPrivateAppendError>(globalObject, decoder);
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateAppendComplete:
        return jsValueForDecodedMessage<MessageName::SourceBufferPrivateRemote_SourceBufferPrivateAppendComplete>(globalObject, decoder);
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateHighestPresentationTimestampChanged:
        return jsValueForDecodedMessage<MessageName::SourceBufferPrivateRemote_SourceBufferPrivateHighestPresentationTimestampChanged>(globalObject, decoder);
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateDurationChanged:
        return jsValueForDecodedMessage<MessageName::SourceBufferPrivateRemote_SourceBufferPrivateDurationChanged>(globalObject, decoder);
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateDidParseSample:
        return jsValueForDecodedMessage<MessageName::SourceBufferPrivateRemote_SourceBufferPrivateDidParseSample>(globalObject, decoder);
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateDidDropSample:
        return jsValueForDecodedMessage<MessageName::SourceBufferPrivateRemote_SourceBufferPrivateDidDropSample>(globalObject, decoder);
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateBufferedDirtyChanged:
        return jsValueForDecodedMessage<MessageName::SourceBufferPrivateRemote_SourceBufferPrivateBufferedDirtyChanged>(globalObject, decoder);
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateDidReceiveRenderingError:
        return jsValueForDecodedMessage<MessageName::SourceBufferPrivateRemote_SourceBufferPrivateDidReceiveRenderingError>(globalObject, decoder);
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateReportExtraMemoryCost:
        return jsValueForDecodedMessage<MessageName::SourceBufferPrivateRemote_SourceBufferPrivateReportExtraMemoryCost>(globalObject, decoder);
#endif
#if USE(LIBWEBRTC) && PLATFORM(COCOA) && ENABLE(GPU_PROCESS)
    case MessageName::LibWebRTCCodecs_FailedDecoding:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecs_FailedDecoding>(globalObject, decoder);
    case MessageName::LibWebRTCCodecs_FlushDecoderCompleted:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecs_FlushDecoderCompleted>(globalObject, decoder);
    case MessageName::LibWebRTCCodecs_CompletedDecoding:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecs_CompletedDecoding>(globalObject, decoder);
    case MessageName::LibWebRTCCodecs_CompletedDecodingCV:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecs_CompletedDecodingCV>(globalObject, decoder);
    case MessageName::LibWebRTCCodecs_CompletedEncoding:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecs_CompletedEncoding>(globalObject, decoder);
    case MessageName::LibWebRTCCodecs_FlushEncoderCompleted:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecs_FlushEncoderCompleted>(globalObject, decoder);
    case MessageName::LibWebRTCCodecs_SetEncodingDescription:
        return jsValueForDecodedMessage<MessageName::LibWebRTCCodecs_SetEncodingDescription>(globalObject, decoder);
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
    case MessageName::RemoteVideoFrameObjectHeapProxyProcessor_SetSharedVideoFrameSemaphore:
        return jsValueForDecodedMessage<MessageName::RemoteVideoFrameObjectHeapProxyProcessor_SetSharedVideoFrameSemaphore>(globalObject, decoder);
    case MessageName::RemoteVideoFrameObjectHeapProxyProcessor_SetSharedVideoFrameMemory:
        return jsValueForDecodedMessage<MessageName::RemoteVideoFrameObjectHeapProxyProcessor_SetSharedVideoFrameMemory>(globalObject, decoder);
    case MessageName::RemoteVideoFrameObjectHeapProxyProcessor_NewVideoFrameBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteVideoFrameObjectHeapProxyProcessor_NewVideoFrameBuffer>(globalObject, decoder);
    case MessageName::RemoteVideoFrameObjectHeapProxyProcessor_NewConvertedVideoFrameBuffer:
        return jsValueForDecodedMessage<MessageName::RemoteVideoFrameObjectHeapProxyProcessor_NewConvertedVideoFrameBuffer>(globalObject, decoder);
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
    case MessageName::SampleBufferDisplayLayer_SetDidFail:
        return jsValueForDecodedMessage<MessageName::SampleBufferDisplayLayer_SetDidFail>(globalObject, decoder);
#endif
    case MessageName::WebGeolocationManager_DidChangePosition:
        return jsValueForDecodedMessage<MessageName::WebGeolocationManager_DidChangePosition>(globalObject, decoder);
    case MessageName::WebGeolocationManager_DidFailToDeterminePosition:
        return jsValueForDecodedMessage<MessageName::WebGeolocationManager_DidFailToDeterminePosition>(globalObject, decoder);
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebGeolocationManager_ResetPermissions:
        return jsValueForDecodedMessage<MessageName::WebGeolocationManager_ResetPermissions>(globalObject, decoder);
#endif
    case MessageName::RemoteWebInspectorUI_Initialize:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUI_Initialize>(globalObject, decoder);
    case MessageName::RemoteWebInspectorUI_UpdateFindString:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUI_UpdateFindString>(globalObject, decoder);
#if ENABLE(INSPECTOR_TELEMETRY)
    case MessageName::RemoteWebInspectorUI_SetDiagnosticLoggingAvailable:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUI_SetDiagnosticLoggingAvailable>(globalObject, decoder);
#endif
    case MessageName::RemoteWebInspectorUI_ShowConsole:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUI_ShowConsole>(globalObject, decoder);
    case MessageName::RemoteWebInspectorUI_ShowResources:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUI_ShowResources>(globalObject, decoder);
    case MessageName::RemoteWebInspectorUI_SendMessageToFrontend:
        return jsValueForDecodedMessage<MessageName::RemoteWebInspectorUI_SendMessageToFrontend>(globalObject, decoder);
    case MessageName::WebInspector_Show:
        return jsValueForDecodedMessage<MessageName::WebInspector_Show>(globalObject, decoder);
    case MessageName::WebInspector_Close:
        return jsValueForDecodedMessage<MessageName::WebInspector_Close>(globalObject, decoder);
    case MessageName::WebInspector_SetAttached:
        return jsValueForDecodedMessage<MessageName::WebInspector_SetAttached>(globalObject, decoder);
    case MessageName::WebInspector_ShowConsole:
        return jsValueForDecodedMessage<MessageName::WebInspector_ShowConsole>(globalObject, decoder);
    case MessageName::WebInspector_ShowResources:
        return jsValueForDecodedMessage<MessageName::WebInspector_ShowResources>(globalObject, decoder);
    case MessageName::WebInspector_ShowMainResourceForFrame:
        return jsValueForDecodedMessage<MessageName::WebInspector_ShowMainResourceForFrame>(globalObject, decoder);
    case MessageName::WebInspector_StartPageProfiling:
        return jsValueForDecodedMessage<MessageName::WebInspector_StartPageProfiling>(globalObject, decoder);
    case MessageName::WebInspector_StopPageProfiling:
        return jsValueForDecodedMessage<MessageName::WebInspector_StopPageProfiling>(globalObject, decoder);
    case MessageName::WebInspector_StartElementSelection:
        return jsValueForDecodedMessage<MessageName::WebInspector_StartElementSelection>(globalObject, decoder);
    case MessageName::WebInspector_StopElementSelection:
        return jsValueForDecodedMessage<MessageName::WebInspector_StopElementSelection>(globalObject, decoder);
    case MessageName::WebInspector_SetFrontendConnection:
        return jsValueForDecodedMessage<MessageName::WebInspector_SetFrontendConnection>(globalObject, decoder);
    case MessageName::WebInspectorInterruptDispatcher_NotifyNeedDebuggerBreak:
        return jsValueForDecodedMessage<MessageName::WebInspectorInterruptDispatcher_NotifyNeedDebuggerBreak>(globalObject, decoder);
    case MessageName::WebInspectorUI_EstablishConnection:
        return jsValueForDecodedMessage<MessageName::WebInspectorUI_EstablishConnection>(globalObject, decoder);
    case MessageName::WebInspectorUI_UpdateConnection:
        return jsValueForDecodedMessage<MessageName::WebInspectorUI_UpdateConnection>(globalObject, decoder);
    case MessageName::WebInspectorUI_AttachedBottom:
        return jsValueForDecodedMessage<MessageName::WebInspectorUI_AttachedBottom>(globalObject, decoder);
    case MessageName::WebInspectorUI_AttachedRight:
        return jsValueForDecodedMessage<MessageName::WebInspectorUI_AttachedRight>(globalObject, decoder);
    case MessageName::WebInspectorUI_AttachedLeft:
        return jsValueForDecodedMessage<MessageName::WebInspectorUI_AttachedLeft>(globalObject, decoder);
    case MessageName::WebInspectorUI_Detached:
        return jsValueForDecodedMessage<MessageName::WebInspectorUI_Detached>(globalObject, decoder);
    case MessageName::WebInspectorUI_SetDockingUnavailable:
        return jsValueForDecodedMessage<MessageName::WebInspectorUI_SetDockingUnavailable>(globalObject, decoder);
    case MessageName::WebInspectorUI_SetIsVisible:
        return jsValueForDecodedMessage<MessageName::WebInspectorUI_SetIsVisible>(globalObject, decoder);
    case MessageName::WebInspectorUI_UpdateFindString:
        return jsValueForDecodedMessage<MessageName::WebInspectorUI_UpdateFindString>(globalObject, decoder);
#if ENABLE(INSPECTOR_TELEMETRY)
    case MessageName::WebInspectorUI_SetDiagnosticLoggingAvailable:
        return jsValueForDecodedMessage<MessageName::WebInspectorUI_SetDiagnosticLoggingAvailable>(globalObject, decoder);
#endif
    case MessageName::WebInspectorUI_ShowConsole:
        return jsValueForDecodedMessage<MessageName::WebInspectorUI_ShowConsole>(globalObject, decoder);
    case MessageName::WebInspectorUI_ShowResources:
        return jsValueForDecodedMessage<MessageName::WebInspectorUI_ShowResources>(globalObject, decoder);
    case MessageName::WebInspectorUI_ShowMainResourceForFrame:
        return jsValueForDecodedMessage<MessageName::WebInspectorUI_ShowMainResourceForFrame>(globalObject, decoder);
    case MessageName::WebInspectorUI_StartPageProfiling:
        return jsValueForDecodedMessage<MessageName::WebInspectorUI_StartPageProfiling>(globalObject, decoder);
    case MessageName::WebInspectorUI_StopPageProfiling:
        return jsValueForDecodedMessage<MessageName::WebInspectorUI_StopPageProfiling>(globalObject, decoder);
    case MessageName::WebInspectorUI_StartElementSelection:
        return jsValueForDecodedMessage<MessageName::WebInspectorUI_StartElementSelection>(globalObject, decoder);
    case MessageName::WebInspectorUI_StopElementSelection:
        return jsValueForDecodedMessage<MessageName::WebInspectorUI_StopElementSelection>(globalObject, decoder);
    case MessageName::WebInspectorUI_SendMessageToFrontend:
        return jsValueForDecodedMessage<MessageName::WebInspectorUI_SendMessageToFrontend>(globalObject, decoder);
    case MessageName::WebInspectorUI_EvaluateInFrontendForTesting:
        return jsValueForDecodedMessage<MessageName::WebInspectorUI_EvaluateInFrontendForTesting>(globalObject, decoder);
#if ENABLE(MEDIA_SESSION_COORDINATOR)
    case MessageName::RemoteMediaSessionCoordinator_SeekSessionToTime:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSessionCoordinator_SeekSessionToTime>(globalObject, decoder);
    case MessageName::RemoteMediaSessionCoordinator_PlaySession:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSessionCoordinator_PlaySession>(globalObject, decoder);
    case MessageName::RemoteMediaSessionCoordinator_PauseSession:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSessionCoordinator_PauseSession>(globalObject, decoder);
    case MessageName::RemoteMediaSessionCoordinator_SetSessionTrack:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSessionCoordinator_SetSessionTrack>(globalObject, decoder);
    case MessageName::RemoteMediaSessionCoordinator_CoordinatorStateChanged:
        return jsValueForDecodedMessage<MessageName::RemoteMediaSessionCoordinator_CoordinatorStateChanged>(globalObject, decoder);
#endif
#if USE(LIBWEBRTC)
    case MessageName::LibWebRTCNetwork_SignalReadPacket:
        return jsValueForDecodedMessage<MessageName::LibWebRTCNetwork_SignalReadPacket>(globalObject, decoder);
    case MessageName::LibWebRTCNetwork_SignalSentPacket:
        return jsValueForDecodedMessage<MessageName::LibWebRTCNetwork_SignalSentPacket>(globalObject, decoder);
    case MessageName::LibWebRTCNetwork_SignalAddressReady:
        return jsValueForDecodedMessage<MessageName::LibWebRTCNetwork_SignalAddressReady>(globalObject, decoder);
    case MessageName::LibWebRTCNetwork_SignalConnect:
        return jsValueForDecodedMessage<MessageName::LibWebRTCNetwork_SignalConnect>(globalObject, decoder);
    case MessageName::LibWebRTCNetwork_SignalClose:
        return jsValueForDecodedMessage<MessageName::LibWebRTCNetwork_SignalClose>(globalObject, decoder);
#endif
#if ENABLE(WEB_RTC)
    case MessageName::RTCDataChannelRemoteManager_SendData:
        return jsValueForDecodedMessage<MessageName::RTCDataChannelRemoteManager_SendData>(globalObject, decoder);
    case MessageName::RTCDataChannelRemoteManager_Close:
        return jsValueForDecodedMessage<MessageName::RTCDataChannelRemoteManager_Close>(globalObject, decoder);
    case MessageName::RTCDataChannelRemoteManager_ChangeReadyState:
        return jsValueForDecodedMessage<MessageName::RTCDataChannelRemoteManager_ChangeReadyState>(globalObject, decoder);
    case MessageName::RTCDataChannelRemoteManager_ReceiveData:
        return jsValueForDecodedMessage<MessageName::RTCDataChannelRemoteManager_ReceiveData>(globalObject, decoder);
    case MessageName::RTCDataChannelRemoteManager_DetectError:
        return jsValueForDecodedMessage<MessageName::RTCDataChannelRemoteManager_DetectError>(globalObject, decoder);
    case MessageName::RTCDataChannelRemoteManager_BufferedAmountIsDecreasing:
        return jsValueForDecodedMessage<MessageName::RTCDataChannelRemoteManager_BufferedAmountIsDecreasing>(globalObject, decoder);
#endif
#if ENABLE(WEB_RTC)
    case MessageName::WebMDNSRegister_FinishedRegisteringMDNSName:
        return jsValueForDecodedMessage<MessageName::WebMDNSRegister_FinishedRegisteringMDNSName>(globalObject, decoder);
#endif
#if USE(LIBWEBRTC)
    case MessageName::WebRTCMonitor_NetworksChanged:
        return jsValueForDecodedMessage<MessageName::WebRTCMonitor_NetworksChanged>(globalObject, decoder);
#endif
#if USE(LIBWEBRTC)
    case MessageName::WebRTCResolver_SetResolvedAddress:
        return jsValueForDecodedMessage<MessageName::WebRTCResolver_SetResolvedAddress>(globalObject, decoder);
    case MessageName::WebRTCResolver_ResolvedAddressError:
        return jsValueForDecodedMessage<MessageName::WebRTCResolver_ResolvedAddressError>(globalObject, decoder);
#endif
#if ENABLE(SHAREABLE_RESOURCE)
    case MessageName::NetworkProcessConnection_DidCacheResource:
        return jsValueForDecodedMessage<MessageName::NetworkProcessConnection_DidCacheResource>(globalObject, decoder);
#endif
    case MessageName::NetworkProcessConnection_DidFinishPingLoad:
        return jsValueForDecodedMessage<MessageName::NetworkProcessConnection_DidFinishPingLoad>(globalObject, decoder);
    case MessageName::NetworkProcessConnection_DidFinishPreconnection:
        return jsValueForDecodedMessage<MessageName::NetworkProcessConnection_DidFinishPreconnection>(globalObject, decoder);
    case MessageName::NetworkProcessConnection_SetOnLineState:
        return jsValueForDecodedMessage<MessageName::NetworkProcessConnection_SetOnLineState>(globalObject, decoder);
    case MessageName::NetworkProcessConnection_CookieAcceptPolicyChanged:
        return jsValueForDecodedMessage<MessageName::NetworkProcessConnection_CookieAcceptPolicyChanged>(globalObject, decoder);
#if HAVE(COOKIE_CHANGE_LISTENER_API)
    case MessageName::NetworkProcessConnection_CookiesAdded:
        return jsValueForDecodedMessage<MessageName::NetworkProcessConnection_CookiesAdded>(globalObject, decoder);
    case MessageName::NetworkProcessConnection_CookiesDeleted:
        return jsValueForDecodedMessage<MessageName::NetworkProcessConnection_CookiesDeleted>(globalObject, decoder);
    case MessageName::NetworkProcessConnection_AllCookiesDeleted:
        return jsValueForDecodedMessage<MessageName::NetworkProcessConnection_AllCookiesDeleted>(globalObject, decoder);
#endif
#if ENABLE(WEB_RTC)
    case MessageName::NetworkProcessConnection_ConnectToRTCDataChannelRemoteSource:
        return jsValueForDecodedMessage<MessageName::NetworkProcessConnection_ConnectToRTCDataChannelRemoteSource>(globalObject, decoder);
#endif
    case MessageName::NetworkProcessConnection_MessagesAvailableForPort:
        return jsValueForDecodedMessage<MessageName::NetworkProcessConnection_MessagesAvailableForPort>(globalObject, decoder);
    case MessageName::NetworkProcessConnection_BroadcastConsoleMessage:
        return jsValueForDecodedMessage<MessageName::NetworkProcessConnection_BroadcastConsoleMessage>(globalObject, decoder);
    case MessageName::WebResourceLoader_WillSendRequest:
        return jsValueForDecodedMessage<MessageName::WebResourceLoader_WillSendRequest>(globalObject, decoder);
    case MessageName::WebResourceLoader_SetWorkerStart:
        return jsValueForDecodedMessage<MessageName::WebResourceLoader_SetWorkerStart>(globalObject, decoder);
    case MessageName::WebResourceLoader_DidSendData:
        return jsValueForDecodedMessage<MessageName::WebResourceLoader_DidSendData>(globalObject, decoder);
    case MessageName::WebResourceLoader_DidReceiveResponse:
        return jsValueForDecodedMessage<MessageName::WebResourceLoader_DidReceiveResponse>(globalObject, decoder);
    case MessageName::WebResourceLoader_DidReceiveData:
        return jsValueForDecodedMessage<MessageName::WebResourceLoader_DidReceiveData>(globalObject, decoder);
    case MessageName::WebResourceLoader_DidFinishResourceLoad:
        return jsValueForDecodedMessage<MessageName::WebResourceLoader_DidFinishResourceLoad>(globalObject, decoder);
    case MessageName::WebResourceLoader_DidFailResourceLoad:
        return jsValueForDecodedMessage<MessageName::WebResourceLoader_DidFailResourceLoad>(globalObject, decoder);
    case MessageName::WebResourceLoader_DidFailServiceWorkerLoad:
        return jsValueForDecodedMessage<MessageName::WebResourceLoader_DidFailServiceWorkerLoad>(globalObject, decoder);
    case MessageName::WebResourceLoader_ServiceWorkerDidNotHandle:
        return jsValueForDecodedMessage<MessageName::WebResourceLoader_ServiceWorkerDidNotHandle>(globalObject, decoder);
    case MessageName::WebResourceLoader_DidBlockAuthenticationChallenge:
        return jsValueForDecodedMessage<MessageName::WebResourceLoader_DidBlockAuthenticationChallenge>(globalObject, decoder);
    case MessageName::WebResourceLoader_StopLoadingAfterXFrameOptionsOrContentSecurityPolicyDenied:
        return jsValueForDecodedMessage<MessageName::WebResourceLoader_StopLoadingAfterXFrameOptionsOrContentSecurityPolicyDenied>(globalObject, decoder);
#if ENABLE(SHAREABLE_RESOURCE)
    case MessageName::WebResourceLoader_DidReceiveResource:
        return jsValueForDecodedMessage<MessageName::WebResourceLoader_DidReceiveResource>(globalObject, decoder);
#endif
#if ENABLE(CONTENT_FILTERING_IN_NETWORKING_PROCESS)
    case MessageName::WebResourceLoader_ContentFilterDidBlockLoad:
        return jsValueForDecodedMessage<MessageName::WebResourceLoader_ContentFilterDidBlockLoad>(globalObject, decoder);
#endif
    case MessageName::WebSocketChannel_DidConnect:
        return jsValueForDecodedMessage<MessageName::WebSocketChannel_DidConnect>(globalObject, decoder);
    case MessageName::WebSocketChannel_DidClose:
        return jsValueForDecodedMessage<MessageName::WebSocketChannel_DidClose>(globalObject, decoder);
    case MessageName::WebSocketChannel_DidReceiveText:
        return jsValueForDecodedMessage<MessageName::WebSocketChannel_DidReceiveText>(globalObject, decoder);
    case MessageName::WebSocketChannel_DidReceiveBinaryData:
        return jsValueForDecodedMessage<MessageName::WebSocketChannel_DidReceiveBinaryData>(globalObject, decoder);
    case MessageName::WebSocketChannel_DidReceiveMessageError:
        return jsValueForDecodedMessage<MessageName::WebSocketChannel_DidReceiveMessageError>(globalObject, decoder);
    case MessageName::WebSocketChannel_DidSendHandshakeRequest:
        return jsValueForDecodedMessage<MessageName::WebSocketChannel_DidSendHandshakeRequest>(globalObject, decoder);
    case MessageName::WebSocketChannel_DidReceiveHandshakeResponse:
        return jsValueForDecodedMessage<MessageName::WebSocketChannel_DidReceiveHandshakeResponse>(globalObject, decoder);
    case MessageName::WebSocketStream_DidOpenSocketStream:
        return jsValueForDecodedMessage<MessageName::WebSocketStream_DidOpenSocketStream>(globalObject, decoder);
    case MessageName::WebSocketStream_DidCloseSocketStream:
        return jsValueForDecodedMessage<MessageName::WebSocketStream_DidCloseSocketStream>(globalObject, decoder);
    case MessageName::WebSocketStream_DidReceiveSocketStreamData:
        return jsValueForDecodedMessage<MessageName::WebSocketStream_DidReceiveSocketStreamData>(globalObject, decoder);
    case MessageName::WebSocketStream_DidFailToReceiveSocketStreamData:
        return jsValueForDecodedMessage<MessageName::WebSocketStream_DidFailToReceiveSocketStreamData>(globalObject, decoder);
    case MessageName::WebSocketStream_DidUpdateBufferedAmount:
        return jsValueForDecodedMessage<MessageName::WebSocketStream_DidUpdateBufferedAmount>(globalObject, decoder);
    case MessageName::WebSocketStream_DidFailSocketStream:
        return jsValueForDecodedMessage<MessageName::WebSocketStream_DidFailSocketStream>(globalObject, decoder);
    case MessageName::WebSocketStream_DidSendData:
        return jsValueForDecodedMessage<MessageName::WebSocketStream_DidSendData>(globalObject, decoder);
    case MessageName::WebSocketStream_DidSendHandshake:
        return jsValueForDecodedMessage<MessageName::WebSocketStream_DidSendHandshake>(globalObject, decoder);
    case MessageName::WebNotificationManager_DidShowNotification:
        return jsValueForDecodedMessage<MessageName::WebNotificationManager_DidShowNotification>(globalObject, decoder);
    case MessageName::WebNotificationManager_DidClickNotification:
        return jsValueForDecodedMessage<MessageName::WebNotificationManager_DidClickNotification>(globalObject, decoder);
    case MessageName::WebNotificationManager_DidCloseNotifications:
        return jsValueForDecodedMessage<MessageName::WebNotificationManager_DidCloseNotifications>(globalObject, decoder);
    case MessageName::WebNotificationManager_DidUpdateNotificationDecision:
        return jsValueForDecodedMessage<MessageName::WebNotificationManager_DidUpdateNotificationDecision>(globalObject, decoder);
    case MessageName::WebNotificationManager_DidRemoveNotificationDecisions:
        return jsValueForDecodedMessage<MessageName::WebNotificationManager_DidRemoveNotificationDecisions>(globalObject, decoder);
#if ENABLE(MEDIA_STREAM)
    case MessageName::SpeechRecognitionRealtimeMediaSourceManager_CreateSource:
        return jsValueForDecodedMessage<MessageName::SpeechRecognitionRealtimeMediaSourceManager_CreateSource>(globalObject, decoder);
    case MessageName::SpeechRecognitionRealtimeMediaSourceManager_DeleteSource:
        return jsValueForDecodedMessage<MessageName::SpeechRecognitionRealtimeMediaSourceManager_DeleteSource>(globalObject, decoder);
    case MessageName::SpeechRecognitionRealtimeMediaSourceManager_Start:
        return jsValueForDecodedMessage<MessageName::SpeechRecognitionRealtimeMediaSourceManager_Start>(globalObject, decoder);
    case MessageName::SpeechRecognitionRealtimeMediaSourceManager_Stop:
        return jsValueForDecodedMessage<MessageName::SpeechRecognitionRealtimeMediaSourceManager_Stop>(globalObject, decoder);
#if ENABLE(SANDBOX_EXTENSIONS)
    case MessageName::SpeechRecognitionRealtimeMediaSourceManager_GrantSandboxExtensions:
        return jsValueForDecodedMessage<MessageName::SpeechRecognitionRealtimeMediaSourceManager_GrantSandboxExtensions>(globalObject, decoder);
    case MessageName::SpeechRecognitionRealtimeMediaSourceManager_RevokeSandboxExtensions:
        return jsValueForDecodedMessage<MessageName::SpeechRecognitionRealtimeMediaSourceManager_RevokeSandboxExtensions>(globalObject, decoder);
#endif
#endif
    case MessageName::WebSharedWorkerContextManagerConnection_LaunchSharedWorker:
        return jsValueForDecodedMessage<MessageName::WebSharedWorkerContextManagerConnection_LaunchSharedWorker>(globalObject, decoder);
    case MessageName::WebSharedWorkerContextManagerConnection_PostConnectEvent:
        return jsValueForDecodedMessage<MessageName::WebSharedWorkerContextManagerConnection_PostConnectEvent>(globalObject, decoder);
    case MessageName::WebSharedWorkerContextManagerConnection_TerminateSharedWorker:
        return jsValueForDecodedMessage<MessageName::WebSharedWorkerContextManagerConnection_TerminateSharedWorker>(globalObject, decoder);
    case MessageName::WebSharedWorkerContextManagerConnection_UpdatePreferencesStore:
        return jsValueForDecodedMessage<MessageName::WebSharedWorkerContextManagerConnection_UpdatePreferencesStore>(globalObject, decoder);
    case MessageName::WebSharedWorkerContextManagerConnection_SetUserAgent:
        return jsValueForDecodedMessage<MessageName::WebSharedWorkerContextManagerConnection_SetUserAgent>(globalObject, decoder);
    case MessageName::WebSharedWorkerContextManagerConnection_Close:
        return jsValueForDecodedMessage<MessageName::WebSharedWorkerContextManagerConnection_Close>(globalObject, decoder);
    case MessageName::WebSharedWorkerContextManagerConnection_SuspendSharedWorker:
        return jsValueForDecodedMessage<MessageName::WebSharedWorkerContextManagerConnection_SuspendSharedWorker>(globalObject, decoder);
    case MessageName::WebSharedWorkerContextManagerConnection_ResumeSharedWorker:
        return jsValueForDecodedMessage<MessageName::WebSharedWorkerContextManagerConnection_ResumeSharedWorker>(globalObject, decoder);
    case MessageName::WebSharedWorkerObjectConnection_FetchScriptInClient:
        return jsValueForDecodedMessage<MessageName::WebSharedWorkerObjectConnection_FetchScriptInClient>(globalObject, decoder);
    case MessageName::WebSharedWorkerObjectConnection_NotifyWorkerObjectOfLoadCompletion:
        return jsValueForDecodedMessage<MessageName::WebSharedWorkerObjectConnection_NotifyWorkerObjectOfLoadCompletion>(globalObject, decoder);
    case MessageName::WebSharedWorkerObjectConnection_PostExceptionToWorkerObject:
        return jsValueForDecodedMessage<MessageName::WebSharedWorkerObjectConnection_PostExceptionToWorkerObject>(globalObject, decoder);
#if ENABLE(SERVICE_WORKER)
    case MessageName::WebSWClientConnection_JobRejectedInServer:
        return jsValueForDecodedMessage<MessageName::WebSWClientConnection_JobRejectedInServer>(globalObject, decoder);
    case MessageName::WebSWClientConnection_RegistrationJobResolvedInServer:
        return jsValueForDecodedMessage<MessageName::WebSWClientConnection_RegistrationJobResolvedInServer>(globalObject, decoder);
    case MessageName::WebSWClientConnection_StartScriptFetchForServer:
        return jsValueForDecodedMessage<MessageName::WebSWClientConnection_StartScriptFetchForServer>(globalObject, decoder);
    case MessageName::WebSWClientConnection_UpdateRegistrationState:
        return jsValueForDecodedMessage<MessageName::WebSWClientConnection_UpdateRegistrationState>(globalObject, decoder);
    case MessageName::WebSWClientConnection_UpdateWorkerState:
        return jsValueForDecodedMessage<MessageName::WebSWClientConnection_UpdateWorkerState>(globalObject, decoder);
    case MessageName::WebSWClientConnection_FireUpdateFoundEvent:
        return jsValueForDecodedMessage<MessageName::WebSWClientConnection_FireUpdateFoundEvent>(globalObject, decoder);
    case MessageName::WebSWClientConnection_SetRegistrationLastUpdateTime:
        return jsValueForDecodedMessage<MessageName::WebSWClientConnection_SetRegistrationLastUpdateTime>(globalObject, decoder);
    case MessageName::WebSWClientConnection_SetRegistrationUpdateViaCache:
        return jsValueForDecodedMessage<MessageName::WebSWClientConnection_SetRegistrationUpdateViaCache>(globalObject, decoder);
    case MessageName::WebSWClientConnection_NotifyClientsOfControllerChange:
        return jsValueForDecodedMessage<MessageName::WebSWClientConnection_NotifyClientsOfControllerChange>(globalObject, decoder);
    case MessageName::WebSWClientConnection_SetSWOriginTableIsImported:
        return jsValueForDecodedMessage<MessageName::WebSWClientConnection_SetSWOriginTableIsImported>(globalObject, decoder);
    case MessageName::WebSWClientConnection_SetSWOriginTableSharedMemory:
        return jsValueForDecodedMessage<MessageName::WebSWClientConnection_SetSWOriginTableSharedMemory>(globalObject, decoder);
    case MessageName::WebSWClientConnection_PostMessageToServiceWorkerClient:
        return jsValueForDecodedMessage<MessageName::WebSWClientConnection_PostMessageToServiceWorkerClient>(globalObject, decoder);
    case MessageName::WebSWClientConnection_SetServiceWorkerClientIsControlled:
        return jsValueForDecodedMessage<MessageName::WebSWClientConnection_SetServiceWorkerClientIsControlled>(globalObject, decoder);
    case MessageName::WebSWClientConnection_FocusServiceWorkerClient:
        return jsValueForDecodedMessage<MessageName::WebSWClientConnection_FocusServiceWorkerClient>(globalObject, decoder);
#endif
#if ENABLE(SERVICE_WORKER)
    case MessageName::WebSWContextManagerConnection_InstallServiceWorker:
        return jsValueForDecodedMessage<MessageName::WebSWContextManagerConnection_InstallServiceWorker>(globalObject, decoder);
    case MessageName::WebSWContextManagerConnection_UpdateAppInitiatedValue:
        return jsValueForDecodedMessage<MessageName::WebSWContextManagerConnection_UpdateAppInitiatedValue>(globalObject, decoder);
    case MessageName::WebSWContextManagerConnection_StartFetch:
        return jsValueForDecodedMessage<MessageName::WebSWContextManagerConnection_StartFetch>(globalObject, decoder);
    case MessageName::WebSWContextManagerConnection_CancelFetch:
        return jsValueForDecodedMessage<MessageName::WebSWContextManagerConnection_CancelFetch>(globalObject, decoder);
    case MessageName::WebSWContextManagerConnection_ContinueDidReceiveFetchResponse:
        return jsValueForDecodedMessage<MessageName::WebSWContextManagerConnection_ContinueDidReceiveFetchResponse>(globalObject, decoder);
    case MessageName::WebSWContextManagerConnection_PostMessageToServiceWorker:
        return jsValueForDecodedMessage<MessageName::WebSWContextManagerConnection_PostMessageToServiceWorker>(globalObject, decoder);
    case MessageName::WebSWContextManagerConnection_FireInstallEvent:
        return jsValueForDecodedMessage<MessageName::WebSWContextManagerConnection_FireInstallEvent>(globalObject, decoder);
    case MessageName::WebSWContextManagerConnection_FireActivateEvent:
        return jsValueForDecodedMessage<MessageName::WebSWContextManagerConnection_FireActivateEvent>(globalObject, decoder);
    case MessageName::WebSWContextManagerConnection_FirePushEvent:
        return jsValueForDecodedMessage<MessageName::WebSWContextManagerConnection_FirePushEvent>(globalObject, decoder);
    case MessageName::WebSWContextManagerConnection_FireNotificationEvent:
        return jsValueForDecodedMessage<MessageName::WebSWContextManagerConnection_FireNotificationEvent>(globalObject, decoder);
    case MessageName::WebSWContextManagerConnection_TerminateWorker:
        return jsValueForDecodedMessage<MessageName::WebSWContextManagerConnection_TerminateWorker>(globalObject, decoder);
#if ENABLE(SHAREABLE_RESOURCE) && PLATFORM(COCOA)
    case MessageName::WebSWContextManagerConnection_DidSaveScriptsToDisk:
        return jsValueForDecodedMessage<MessageName::WebSWContextManagerConnection_DidSaveScriptsToDisk>(globalObject, decoder);
#endif
    case MessageName::WebSWContextManagerConnection_MatchAllCompleted:
        return jsValueForDecodedMessage<MessageName::WebSWContextManagerConnection_MatchAllCompleted>(globalObject, decoder);
    case MessageName::WebSWContextManagerConnection_SkipWaitingCompleted:
        return jsValueForDecodedMessage<MessageName::WebSWContextManagerConnection_SkipWaitingCompleted>(globalObject, decoder);
    case MessageName::WebSWContextManagerConnection_SetUserAgent:
        return jsValueForDecodedMessage<MessageName::WebSWContextManagerConnection_SetUserAgent>(globalObject, decoder);
    case MessageName::WebSWContextManagerConnection_UpdatePreferencesStore:
        return jsValueForDecodedMessage<MessageName::WebSWContextManagerConnection_UpdatePreferencesStore>(globalObject, decoder);
    case MessageName::WebSWContextManagerConnection_Close:
        return jsValueForDecodedMessage<MessageName::WebSWContextManagerConnection_Close>(globalObject, decoder);
    case MessageName::WebSWContextManagerConnection_SetThrottleState:
        return jsValueForDecodedMessage<MessageName::WebSWContextManagerConnection_SetThrottleState>(globalObject, decoder);
    case MessageName::WebSWContextManagerConnection_ConvertFetchToDownload:
        return jsValueForDecodedMessage<MessageName::WebSWContextManagerConnection_ConvertFetchToDownload>(globalObject, decoder);
    case MessageName::WebSWContextManagerConnection_NavigationPreloadIsReady:
        return jsValueForDecodedMessage<MessageName::WebSWContextManagerConnection_NavigationPreloadIsReady>(globalObject, decoder);
    case MessageName::WebSWContextManagerConnection_NavigationPreloadFailed:
        return jsValueForDecodedMessage<MessageName::WebSWContextManagerConnection_NavigationPreloadFailed>(globalObject, decoder);
#endif
    case MessageName::WebUserContentController_AddContentWorlds:
        return jsValueForDecodedMessage<MessageName::WebUserContentController_AddContentWorlds>(globalObject, decoder);
    case MessageName::WebUserContentController_RemoveContentWorlds:
        return jsValueForDecodedMessage<MessageName::WebUserContentController_RemoveContentWorlds>(globalObject, decoder);
    case MessageName::WebUserContentController_AddUserScripts:
        return jsValueForDecodedMessage<MessageName::WebUserContentController_AddUserScripts>(globalObject, decoder);
    case MessageName::WebUserContentController_RemoveUserScript:
        return jsValueForDecodedMessage<MessageName::WebUserContentController_RemoveUserScript>(globalObject, decoder);
    case MessageName::WebUserContentController_RemoveAllUserScripts:
        return jsValueForDecodedMessage<MessageName::WebUserContentController_RemoveAllUserScripts>(globalObject, decoder);
    case MessageName::WebUserContentController_AddUserStyleSheets:
        return jsValueForDecodedMessage<MessageName::WebUserContentController_AddUserStyleSheets>(globalObject, decoder);
    case MessageName::WebUserContentController_RemoveUserStyleSheet:
        return jsValueForDecodedMessage<MessageName::WebUserContentController_RemoveUserStyleSheet>(globalObject, decoder);
    case MessageName::WebUserContentController_RemoveAllUserStyleSheets:
        return jsValueForDecodedMessage<MessageName::WebUserContentController_RemoveAllUserStyleSheets>(globalObject, decoder);
    case MessageName::WebUserContentController_AddUserScriptMessageHandlers:
        return jsValueForDecodedMessage<MessageName::WebUserContentController_AddUserScriptMessageHandlers>(globalObject, decoder);
    case MessageName::WebUserContentController_RemoveUserScriptMessageHandler:
        return jsValueForDecodedMessage<MessageName::WebUserContentController_RemoveUserScriptMessageHandler>(globalObject, decoder);
    case MessageName::WebUserContentController_RemoveAllUserScriptMessageHandlersForWorlds:
        return jsValueForDecodedMessage<MessageName::WebUserContentController_RemoveAllUserScriptMessageHandlersForWorlds>(globalObject, decoder);
    case MessageName::WebUserContentController_RemoveAllUserScriptMessageHandlers:
        return jsValueForDecodedMessage<MessageName::WebUserContentController_RemoveAllUserScriptMessageHandlers>(globalObject, decoder);
#if ENABLE(CONTENT_EXTENSIONS)
    case MessageName::WebUserContentController_AddContentRuleLists:
        return jsValueForDecodedMessage<MessageName::WebUserContentController_AddContentRuleLists>(globalObject, decoder);
    case MessageName::WebUserContentController_RemoveContentRuleList:
        return jsValueForDecodedMessage<MessageName::WebUserContentController_RemoveContentRuleList>(globalObject, decoder);
    case MessageName::WebUserContentController_RemoveAllContentRuleLists:
        return jsValueForDecodedMessage<MessageName::WebUserContentController_RemoveAllContentRuleLists>(globalObject, decoder);
#endif
    case MessageName::RemoteWebLockRegistry_DidCompleteLockRequest:
        return jsValueForDecodedMessage<MessageName::RemoteWebLockRegistry_DidCompleteLockRequest>(globalObject, decoder);
    case MessageName::RemoteWebLockRegistry_DidStealLock:
        return jsValueForDecodedMessage<MessageName::RemoteWebLockRegistry_DidStealLock>(globalObject, decoder);
    case MessageName::WebBroadcastChannelRegistry_PostMessageToRemote:
        return jsValueForDecodedMessage<MessageName::WebBroadcastChannelRegistry_PostMessageToRemote>(globalObject, decoder);
    case MessageName::WebFileSystemStorageConnection_InvalidateAccessHandle:
        return jsValueForDecodedMessage<MessageName::WebFileSystemStorageConnection_InvalidateAccessHandle>(globalObject, decoder);
    case MessageName::WebPermissionController_permissionChanged:
        return jsValueForDecodedMessage<MessageName::WebPermissionController_permissionChanged>(globalObject, decoder);
    case MessageName::WebScreenOrientationManager_OrientationDidChange:
        return jsValueForDecodedMessage<MessageName::WebScreenOrientationManager_OrientationDidChange>(globalObject, decoder);
    case MessageName::WebSpeechRecognitionConnection_DidReceiveUpdate:
        return jsValueForDecodedMessage<MessageName::WebSpeechRecognitionConnection_DidReceiveUpdate>(globalObject, decoder);
#if USE(COORDINATED_GRAPHICS) || USE(TEXTURE_MAPPER)
    case MessageName::DrawingArea_UpdateBackingStoreState:
        return jsValueForDecodedMessage<MessageName::DrawingArea_UpdateBackingStoreState>(globalObject, decoder);
    case MessageName::DrawingArea_TargetRefreshRateDidChange:
        return jsValueForDecodedMessage<MessageName::DrawingArea_TargetRefreshRateDidChange>(globalObject, decoder);
#endif
    case MessageName::DrawingArea_DisplayDidRefresh:
        return jsValueForDecodedMessage<MessageName::DrawingArea_DisplayDidRefresh>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::DrawingArea_UpdateGeometry:
        return jsValueForDecodedMessage<MessageName::DrawingArea_UpdateGeometry>(globalObject, decoder);
    case MessageName::DrawingArea_SetDeviceScaleFactor:
        return jsValueForDecodedMessage<MessageName::DrawingArea_SetDeviceScaleFactor>(globalObject, decoder);
    case MessageName::DrawingArea_SetColorSpace:
        return jsValueForDecodedMessage<MessageName::DrawingArea_SetColorSpace>(globalObject, decoder);
    case MessageName::DrawingArea_SetViewExposedRect:
        return jsValueForDecodedMessage<MessageName::DrawingArea_SetViewExposedRect>(globalObject, decoder);
    case MessageName::DrawingArea_AcceleratedAnimationDidStart:
        return jsValueForDecodedMessage<MessageName::DrawingArea_AcceleratedAnimationDidStart>(globalObject, decoder);
    case MessageName::DrawingArea_AcceleratedAnimationDidEnd:
        return jsValueForDecodedMessage<MessageName::DrawingArea_AcceleratedAnimationDidEnd>(globalObject, decoder);
    case MessageName::DrawingArea_DispatchAfterEnsuringDrawing:
        return jsValueForDecodedMessage<MessageName::DrawingArea_DispatchAfterEnsuringDrawing>(globalObject, decoder);
#endif
#if USE(GRAPHICS_LAYER_WC)
    case MessageName::DrawingArea_UpdateGeometry:
        return jsValueForDecodedMessage<MessageName::DrawingArea_UpdateGeometry>(globalObject, decoder);
#endif
#if PLATFORM(COCOA) || PLATFORM(GTK)
    case MessageName::DrawingArea_AdjustTransientZoom:
        return jsValueForDecodedMessage<MessageName::DrawingArea_AdjustTransientZoom>(globalObject, decoder);
    case MessageName::DrawingArea_CommitTransientZoom:
        return jsValueForDecodedMessage<MessageName::DrawingArea_CommitTransientZoom>(globalObject, decoder);
#endif
    case MessageName::EventDispatcher_WheelEvent:
        return jsValueForDecodedMessage<MessageName::EventDispatcher_WheelEvent>(globalObject, decoder);
#if ENABLE(IOS_TOUCH_EVENTS)
    case MessageName::EventDispatcher_TouchEvent:
        return jsValueForDecodedMessage<MessageName::EventDispatcher_TouchEvent>(globalObject, decoder);
    case MessageName::EventDispatcher_TouchEventWithoutCallback:
        return jsValueForDecodedMessage<MessageName::EventDispatcher_TouchEventWithoutCallback>(globalObject, decoder);
#endif
#if ENABLE(MAC_GESTURE_EVENTS)
    case MessageName::EventDispatcher_GestureEvent:
        return jsValueForDecodedMessage<MessageName::EventDispatcher_GestureEvent>(globalObject, decoder);
#endif
#if HAVE(CVDISPLAYLINK)
    case MessageName::EventDispatcher_DisplayWasRefreshed:
        return jsValueForDecodedMessage<MessageName::EventDispatcher_DisplayWasRefreshed>(globalObject, decoder);
#endif
#if ENABLE(MOMENTUM_EVENT_DISPATCHER)
    case MessageName::EventDispatcher_SetScrollingAccelerationCurve:
        return jsValueForDecodedMessage<MessageName::EventDispatcher_SetScrollingAccelerationCurve>(globalObject, decoder);
#endif
    case MessageName::EventDispatcher_PageScreenDidChange:
        return jsValueForDecodedMessage<MessageName::EventDispatcher_PageScreenDidChange>(globalObject, decoder);
    case MessageName::VisitedLinkTableController_SetVisitedLinkTable:
        return jsValueForDecodedMessage<MessageName::VisitedLinkTableController_SetVisitedLinkTable>(globalObject, decoder);
    case MessageName::VisitedLinkTableController_VisitedLinkStateChanged:
        return jsValueForDecodedMessage<MessageName::VisitedLinkTableController_VisitedLinkStateChanged>(globalObject, decoder);
    case MessageName::VisitedLinkTableController_AllVisitedLinkStateChanged:
        return jsValueForDecodedMessage<MessageName::VisitedLinkTableController_AllVisitedLinkStateChanged>(globalObject, decoder);
    case MessageName::VisitedLinkTableController_RemoveAllVisitedLinks:
        return jsValueForDecodedMessage<MessageName::VisitedLinkTableController_RemoveAllVisitedLinks>(globalObject, decoder);
    case MessageName::WebFrame_GetFrameInfo:
        return jsValueForDecodedMessage<MessageName::WebFrame_GetFrameInfo>(globalObject, decoder);
    case MessageName::WebFrame_ContinueWillSubmitForm:
        return jsValueForDecodedMessage<MessageName::WebFrame_ContinueWillSubmitForm>(globalObject, decoder);
    case MessageName::WebFrame_DidCommitLoadInAnotherProcess:
        return jsValueForDecodedMessage<MessageName::WebFrame_DidCommitLoadInAnotherProcess>(globalObject, decoder);
    case MessageName::WebFrame_DidFinishLoadInAnotherProcess:
        return jsValueForDecodedMessage<MessageName::WebFrame_DidFinishLoadInAnotherProcess>(globalObject, decoder);
    case MessageName::WebPage_SetInitialFocus:
        return jsValueForDecodedMessage<MessageName::WebPage_SetInitialFocus>(globalObject, decoder);
    case MessageName::WebPage_SetActivityState:
        return jsValueForDecodedMessage<MessageName::WebPage_SetActivityState>(globalObject, decoder);
    case MessageName::WebPage_SetLayerHostingMode:
        return jsValueForDecodedMessage<MessageName::WebPage_SetLayerHostingMode>(globalObject, decoder);
    case MessageName::WebPage_SetBackgroundColor:
        return jsValueForDecodedMessage<MessageName::WebPage_SetBackgroundColor>(globalObject, decoder);
    case MessageName::WebPage_AddConsoleMessage:
        return jsValueForDecodedMessage<MessageName::WebPage_AddConsoleMessage>(globalObject, decoder);
    case MessageName::WebPage_EnqueueSecurityPolicyViolationEvent:
        return jsValueForDecodedMessage<MessageName::WebPage_EnqueueSecurityPolicyViolationEvent>(globalObject, decoder);
    case MessageName::WebPage_SendReportToEndpoints:
        return jsValueForDecodedMessage<MessageName::WebPage_SendReportToEndpoints>(globalObject, decoder);
    case MessageName::WebPage_NotifyReportObservers:
        return jsValueForDecodedMessage<MessageName::WebPage_NotifyReportObservers>(globalObject, decoder);
    case MessageName::WebPage_TestProcessIncomingSyncMessagesWhenWaitingForSyncReply:
        return jsValueForDecodedMessage<MessageName::WebPage_TestProcessIncomingSyncMessagesWhenWaitingForSyncReply>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::WebPage_SetTopContentInsetFenced:
        return jsValueForDecodedMessage<MessageName::WebPage_SetTopContentInsetFenced>(globalObject, decoder);
#endif
    case MessageName::WebPage_SetTopContentInset:
        return jsValueForDecodedMessage<MessageName::WebPage_SetTopContentInset>(globalObject, decoder);
    case MessageName::WebPage_SetUnderlayColor:
        return jsValueForDecodedMessage<MessageName::WebPage_SetUnderlayColor>(globalObject, decoder);
    case MessageName::WebPage_SetUnderPageBackgroundColorOverride:
        return jsValueForDecodedMessage<MessageName::WebPage_SetUnderPageBackgroundColorOverride>(globalObject, decoder);
    case MessageName::WebPage_ViewWillStartLiveResize:
        return jsValueForDecodedMessage<MessageName::WebPage_ViewWillStartLiveResize>(globalObject, decoder);
    case MessageName::WebPage_ViewWillEndLiveResize:
        return jsValueForDecodedMessage<MessageName::WebPage_ViewWillEndLiveResize>(globalObject, decoder);
    case MessageName::WebPage_ExecuteEditCommandWithCallback:
        return jsValueForDecodedMessage<MessageName::WebPage_ExecuteEditCommandWithCallback>(globalObject, decoder);
    case MessageName::WebPage_KeyEvent:
        return jsValueForDecodedMessage<MessageName::WebPage_KeyEvent>(globalObject, decoder);
    case MessageName::WebPage_MouseEvent:
        return jsValueForDecodedMessage<MessageName::WebPage_MouseEvent>(globalObject, decoder);
#if ENABLE(NOTIFICATIONS)
    case MessageName::WebPage_ClearNotificationPermissionState:
        return jsValueForDecodedMessage<MessageName::WebPage_ClearNotificationPermissionState>(globalObject, decoder);
#endif
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebPage_SetViewportConfigurationViewLayoutSize:
        return jsValueForDecodedMessage<MessageName::WebPage_SetViewportConfigurationViewLayoutSize>(globalObject, decoder);
    case MessageName::WebPage_SetDeviceOrientation:
        return jsValueForDecodedMessage<MessageName::WebPage_SetDeviceOrientation>(globalObject, decoder);
    case MessageName::WebPage_SetOverrideViewportArguments:
        return jsValueForDecodedMessage<MessageName::WebPage_SetOverrideViewportArguments>(globalObject, decoder);
    case MessageName::WebPage_DynamicViewportSizeUpdate:
        return jsValueForDecodedMessage<MessageName::WebPage_DynamicViewportSizeUpdate>(globalObject, decoder);
    case MessageName::WebPage_SetScreenIsBeingCaptured:
        return jsValueForDecodedMessage<MessageName::WebPage_SetScreenIsBeingCaptured>(globalObject, decoder);
    case MessageName::WebPage_AttemptSyntheticClick:
        return jsValueForDecodedMessage<MessageName::WebPage_AttemptSyntheticClick>(globalObject, decoder);
    case MessageName::WebPage_PotentialTapAtPosition:
        return jsValueForDecodedMessage<MessageName::WebPage_PotentialTapAtPosition>(globalObject, decoder);
    case MessageName::WebPage_CommitPotentialTap:
        return jsValueForDecodedMessage<MessageName::WebPage_CommitPotentialTap>(globalObject, decoder);
    case MessageName::WebPage_CancelPotentialTap:
        return jsValueForDecodedMessage<MessageName::WebPage_CancelPotentialTap>(globalObject, decoder);
    case MessageName::WebPage_TapHighlightAtPosition:
        return jsValueForDecodedMessage<MessageName::WebPage_TapHighlightAtPosition>(globalObject, decoder);
    case MessageName::WebPage_DidRecognizeLongPress:
        return jsValueForDecodedMessage<MessageName::WebPage_DidRecognizeLongPress>(globalObject, decoder);
    case MessageName::WebPage_HandleDoubleTapForDoubleClickAtPoint:
        return jsValueForDecodedMessage<MessageName::WebPage_HandleDoubleTapForDoubleClickAtPoint>(globalObject, decoder);
    case MessageName::WebPage_InspectorNodeSearchMovedToPosition:
        return jsValueForDecodedMessage<MessageName::WebPage_InspectorNodeSearchMovedToPosition>(globalObject, decoder);
    case MessageName::WebPage_InspectorNodeSearchEndedAtPosition:
        return jsValueForDecodedMessage<MessageName::WebPage_InspectorNodeSearchEndedAtPosition>(globalObject, decoder);
    case MessageName::WebPage_BlurFocusedElement:
        return jsValueForDecodedMessage<MessageName::WebPage_BlurFocusedElement>(globalObject, decoder);
    case MessageName::WebPage_SelectWithGesture:
        return jsValueForDecodedMessage<MessageName::WebPage_SelectWithGesture>(globalObject, decoder);
    case MessageName::WebPage_UpdateSelectionWithTouches:
        return jsValueForDecodedMessage<MessageName::WebPage_UpdateSelectionWithTouches>(globalObject, decoder);
    case MessageName::WebPage_SelectWithTwoTouches:
        return jsValueForDecodedMessage<MessageName::WebPage_SelectWithTwoTouches>(globalObject, decoder);
    case MessageName::WebPage_ExtendSelection:
        return jsValueForDecodedMessage<MessageName::WebPage_ExtendSelection>(globalObject, decoder);
    case MessageName::WebPage_SelectWordBackward:
        return jsValueForDecodedMessage<MessageName::WebPage_SelectWordBackward>(globalObject, decoder);
    case MessageName::WebPage_ExtendSelectionForReplacement:
        return jsValueForDecodedMessage<MessageName::WebPage_ExtendSelectionForReplacement>(globalObject, decoder);
    case MessageName::WebPage_MoveSelectionByOffset:
        return jsValueForDecodedMessage<MessageName::WebPage_MoveSelectionByOffset>(globalObject, decoder);
    case MessageName::WebPage_SelectTextWithGranularityAtPoint:
        return jsValueForDecodedMessage<MessageName::WebPage_SelectTextWithGranularityAtPoint>(globalObject, decoder);
    case MessageName::WebPage_SelectPositionAtBoundaryWithDirection:
        return jsValueForDecodedMessage<MessageName::WebPage_SelectPositionAtBoundaryWithDirection>(globalObject, decoder);
    case MessageName::WebPage_MoveSelectionAtBoundaryWithDirection:
        return jsValueForDecodedMessage<MessageName::WebPage_MoveSelectionAtBoundaryWithDirection>(globalObject, decoder);
    case MessageName::WebPage_SelectPositionAtPoint:
        return jsValueForDecodedMessage<MessageName::WebPage_SelectPositionAtPoint>(globalObject, decoder);
    case MessageName::WebPage_BeginSelectionInDirection:
        return jsValueForDecodedMessage<MessageName::WebPage_BeginSelectionInDirection>(globalObject, decoder);
    case MessageName::WebPage_UpdateSelectionWithExtentPoint:
        return jsValueForDecodedMessage<MessageName::WebPage_UpdateSelectionWithExtentPoint>(globalObject, decoder);
    case MessageName::WebPage_UpdateSelectionWithExtentPointAndBoundary:
        return jsValueForDecodedMessage<MessageName::WebPage_UpdateSelectionWithExtentPointAndBoundary>(globalObject, decoder);
    case MessageName::WebPage_RequestDictationContext:
        return jsValueForDecodedMessage<MessageName::WebPage_RequestDictationContext>(globalObject, decoder);
#endif
#if (PLATFORM(IOS_FAMILY) && ENABLE(REVEAL))
    case MessageName::WebPage_RequestRVItemInCurrentSelectedRange:
        return jsValueForDecodedMessage<MessageName::WebPage_RequestRVItemInCurrentSelectedRange>(globalObject, decoder);
    case MessageName::WebPage_PrepareSelectionForContextMenuWithLocationInView:
        return jsValueForDecodedMessage<MessageName::WebPage_PrepareSelectionForContextMenuWithLocationInView>(globalObject, decoder);
#endif
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebPage_WillInsertFinalDictationResult:
        return jsValueForDecodedMessage<MessageName::WebPage_WillInsertFinalDictationResult>(globalObject, decoder);
    case MessageName::WebPage_DidInsertFinalDictationResult:
        return jsValueForDecodedMessage<MessageName::WebPage_DidInsertFinalDictationResult>(globalObject, decoder);
    case MessageName::WebPage_ReplaceDictatedText:
        return jsValueForDecodedMessage<MessageName::WebPage_ReplaceDictatedText>(globalObject, decoder);
    case MessageName::WebPage_ReplaceSelectedText:
        return jsValueForDecodedMessage<MessageName::WebPage_ReplaceSelectedText>(globalObject, decoder);
    case MessageName::WebPage_RequestAutocorrectionData:
        return jsValueForDecodedMessage<MessageName::WebPage_RequestAutocorrectionData>(globalObject, decoder);
    case MessageName::WebPage_ApplyAutocorrection:
        return jsValueForDecodedMessage<MessageName::WebPage_ApplyAutocorrection>(globalObject, decoder);
    case MessageName::WebPage_SyncApplyAutocorrection:
        return jsValueForDecodedMessage<MessageName::WebPage_SyncApplyAutocorrection>(globalObject, decoder);
    case MessageName::WebPage_HandleAutocorrectionContextRequest:
        return jsValueForDecodedMessage<MessageName::WebPage_HandleAutocorrectionContextRequest>(globalObject, decoder);
    case MessageName::WebPage_RequestEvasionRectsAboveSelection:
        return jsValueForDecodedMessage<MessageName::WebPage_RequestEvasionRectsAboveSelection>(globalObject, decoder);
    case MessageName::WebPage_RequestPositionInformation:
        return jsValueForDecodedMessage<MessageName::WebPage_RequestPositionInformation>(globalObject, decoder);
    case MessageName::WebPage_StartInteractionWithElementContextOrPosition:
        return jsValueForDecodedMessage<MessageName::WebPage_StartInteractionWithElementContextOrPosition>(globalObject, decoder);
    case MessageName::WebPage_StopInteraction:
        return jsValueForDecodedMessage<MessageName::WebPage_StopInteraction>(globalObject, decoder);
    case MessageName::WebPage_PerformActionOnElement:
        return jsValueForDecodedMessage<MessageName::WebPage_PerformActionOnElement>(globalObject, decoder);
    case MessageName::WebPage_FocusNextFocusedElement:
        return jsValueForDecodedMessage<MessageName::WebPage_FocusNextFocusedElement>(globalObject, decoder);
    case MessageName::WebPage_AutofillLoginCredentials:
        return jsValueForDecodedMessage<MessageName::WebPage_AutofillLoginCredentials>(globalObject, decoder);
    case MessageName::WebPage_SetFocusedElementValue:
        return jsValueForDecodedMessage<MessageName::WebPage_SetFocusedElementValue>(globalObject, decoder);
    case MessageName::WebPage_SetFocusedElementSelectedIndex:
        return jsValueForDecodedMessage<MessageName::WebPage_SetFocusedElementSelectedIndex>(globalObject, decoder);
    case MessageName::WebPage_ApplicationWillResignActive:
        return jsValueForDecodedMessage<MessageName::WebPage_ApplicationWillResignActive>(globalObject, decoder);
    case MessageName::WebPage_ApplicationDidEnterBackground:
        return jsValueForDecodedMessage<MessageName::WebPage_ApplicationDidEnterBackground>(globalObject, decoder);
    case MessageName::WebPage_ApplicationDidFinishSnapshottingAfterEnteringBackground:
        return jsValueForDecodedMessage<MessageName::WebPage_ApplicationDidFinishSnapshottingAfterEnteringBackground>(globalObject, decoder);
    case MessageName::WebPage_ApplicationWillEnterForeground:
        return jsValueForDecodedMessage<MessageName::WebPage_ApplicationWillEnterForeground>(globalObject, decoder);
    case MessageName::WebPage_ApplicationDidBecomeActive:
        return jsValueForDecodedMessage<MessageName::WebPage_ApplicationDidBecomeActive>(globalObject, decoder);
    case MessageName::WebPage_ApplicationDidEnterBackgroundForMedia:
        return jsValueForDecodedMessage<MessageName::WebPage_ApplicationDidEnterBackgroundForMedia>(globalObject, decoder);
    case MessageName::WebPage_ApplicationWillEnterForegroundForMedia:
        return jsValueForDecodedMessage<MessageName::WebPage_ApplicationWillEnterForegroundForMedia>(globalObject, decoder);
    case MessageName::WebPage_ContentSizeCategoryDidChange:
        return jsValueForDecodedMessage<MessageName::WebPage_ContentSizeCategoryDidChange>(globalObject, decoder);
    case MessageName::WebPage_GetSelectionContext:
        return jsValueForDecodedMessage<MessageName::WebPage_GetSelectionContext>(globalObject, decoder);
    case MessageName::WebPage_SetAllowsMediaDocumentInlinePlayback:
        return jsValueForDecodedMessage<MessageName::WebPage_SetAllowsMediaDocumentInlinePlayback>(globalObject, decoder);
    case MessageName::WebPage_HandleTwoFingerTapAtPoint:
        return jsValueForDecodedMessage<MessageName::WebPage_HandleTwoFingerTapAtPoint>(globalObject, decoder);
    case MessageName::WebPage_SetForceAlwaysUserScalable:
        return jsValueForDecodedMessage<MessageName::WebPage_SetForceAlwaysUserScalable>(globalObject, decoder);
    case MessageName::WebPage_GetRectsForGranularityWithSelectionOffset:
        return jsValueForDecodedMessage<MessageName::WebPage_GetRectsForGranularityWithSelectionOffset>(globalObject, decoder);
    case MessageName::WebPage_GetRectsAtSelectionOffsetWithText:
        return jsValueForDecodedMessage<MessageName::WebPage_GetRectsAtSelectionOffsetWithText>(globalObject, decoder);
    case MessageName::WebPage_StoreSelectionForAccessibility:
        return jsValueForDecodedMessage<MessageName::WebPage_StoreSelectionForAccessibility>(globalObject, decoder);
    case MessageName::WebPage_StartAutoscrollAtPosition:
        return jsValueForDecodedMessage<MessageName::WebPage_StartAutoscrollAtPosition>(globalObject, decoder);
    case MessageName::WebPage_CancelAutoscroll:
        return jsValueForDecodedMessage<MessageName::WebPage_CancelAutoscroll>(globalObject, decoder);
    case MessageName::WebPage_RequestFocusedElementInformation:
        return jsValueForDecodedMessage<MessageName::WebPage_RequestFocusedElementInformation>(globalObject, decoder);
    case MessageName::WebPage_HardwareKeyboardAvailabilityChanged:
        return jsValueForDecodedMessage<MessageName::WebPage_HardwareKeyboardAvailabilityChanged>(globalObject, decoder);
    case MessageName::WebPage_SetIsShowingInputViewForFocusedElement:
        return jsValueForDecodedMessage<MessageName::WebPage_SetIsShowingInputViewForFocusedElement>(globalObject, decoder);
    case MessageName::WebPage_UpdateSelectionWithDelta:
        return jsValueForDecodedMessage<MessageName::WebPage_UpdateSelectionWithDelta>(globalObject, decoder);
    case MessageName::WebPage_RequestDocumentEditingContext:
        return jsValueForDecodedMessage<MessageName::WebPage_RequestDocumentEditingContext>(globalObject, decoder);
    case MessageName::WebPage_GenerateSyntheticEditingCommand:
        return jsValueForDecodedMessage<MessageName::WebPage_GenerateSyntheticEditingCommand>(globalObject, decoder);
    case MessageName::WebPage_SetShouldRevealCurrentSelectionAfterInsertion:
        return jsValueForDecodedMessage<MessageName::WebPage_SetShouldRevealCurrentSelectionAfterInsertion>(globalObject, decoder);
    case MessageName::WebPage_InsertTextPlaceholder:
        return jsValueForDecodedMessage<MessageName::WebPage_InsertTextPlaceholder>(globalObject, decoder);
    case MessageName::WebPage_RemoveTextPlaceholder:
        return jsValueForDecodedMessage<MessageName::WebPage_RemoveTextPlaceholder>(globalObject, decoder);
    case MessageName::WebPage_TextInputContextsInRect:
        return jsValueForDecodedMessage<MessageName::WebPage_TextInputContextsInRect>(globalObject, decoder);
    case MessageName::WebPage_FocusTextInputContextAndPlaceCaret:
        return jsValueForDecodedMessage<MessageName::WebPage_FocusTextInputContextAndPlaceCaret>(globalObject, decoder);
    case MessageName::WebPage_ClearServiceWorkerEntitlementOverride:
        return jsValueForDecodedMessage<MessageName::WebPage_ClearServiceWorkerEntitlementOverride>(globalObject, decoder);
#endif
    case MessageName::WebPage_RequestImageBitmap:
        return jsValueForDecodedMessage<MessageName::WebPage_RequestImageBitmap>(globalObject, decoder);
    case MessageName::WebPage_SetControlledByAutomation:
        return jsValueForDecodedMessage<MessageName::WebPage_SetControlledByAutomation>(globalObject, decoder);
    case MessageName::WebPage_ConnectInspector:
        return jsValueForDecodedMessage<MessageName::WebPage_ConnectInspector>(globalObject, decoder);
    case MessageName::WebPage_DisconnectInspector:
        return jsValueForDecodedMessage<MessageName::WebPage_DisconnectInspector>(globalObject, decoder);
    case MessageName::WebPage_SendMessageToTargetBackend:
        return jsValueForDecodedMessage<MessageName::WebPage_SendMessageToTargetBackend>(globalObject, decoder);
#if ENABLE(REMOTE_INSPECTOR)
    case MessageName::WebPage_SetIndicating:
        return jsValueForDecodedMessage<MessageName::WebPage_SetIndicating>(globalObject, decoder);
#endif
#if ENABLE(IOS_TOUCH_EVENTS)
    case MessageName::WebPage_ResetPotentialTapSecurityOrigin:
        return jsValueForDecodedMessage<MessageName::WebPage_ResetPotentialTapSecurityOrigin>(globalObject, decoder);
#endif
#if !ENABLE(IOS_TOUCH_EVENTS) && ENABLE(TOUCH_EVENTS)
    case MessageName::WebPage_TouchEvent:
        return jsValueForDecodedMessage<MessageName::WebPage_TouchEvent>(globalObject, decoder);
#endif
    case MessageName::WebPage_CancelPointer:
        return jsValueForDecodedMessage<MessageName::WebPage_CancelPointer>(globalObject, decoder);
    case MessageName::WebPage_TouchWithIdentifierWasRemoved:
        return jsValueForDecodedMessage<MessageName::WebPage_TouchWithIdentifierWasRemoved>(globalObject, decoder);
#if ENABLE(INPUT_TYPE_COLOR)
    case MessageName::WebPage_DidEndColorPicker:
        return jsValueForDecodedMessage<MessageName::WebPage_DidEndColorPicker>(globalObject, decoder);
    case MessageName::WebPage_DidChooseColor:
        return jsValueForDecodedMessage<MessageName::WebPage_DidChooseColor>(globalObject, decoder);
#endif
#if ENABLE(DATALIST_ELEMENT)
    case MessageName::WebPage_DidSelectDataListOption:
        return jsValueForDecodedMessage<MessageName::WebPage_DidSelectDataListOption>(globalObject, decoder);
    case MessageName::WebPage_DidCloseSuggestions:
        return jsValueForDecodedMessage<MessageName::WebPage_DidCloseSuggestions>(globalObject, decoder);
#endif
#if ENABLE(DATE_AND_TIME_INPUT_TYPES)
    case MessageName::WebPage_DidChooseDate:
        return jsValueForDecodedMessage<MessageName::WebPage_DidChooseDate>(globalObject, decoder);
    case MessageName::WebPage_DidEndDateTimePicker:
        return jsValueForDecodedMessage<MessageName::WebPage_DidEndDateTimePicker>(globalObject, decoder);
#endif
    case MessageName::WebPage_ScrollBy:
        return jsValueForDecodedMessage<MessageName::WebPage_ScrollBy>(globalObject, decoder);
    case MessageName::WebPage_CenterSelectionInVisibleArea:
        return jsValueForDecodedMessage<MessageName::WebPage_CenterSelectionInVisibleArea>(globalObject, decoder);
    case MessageName::WebPage_GoToBackForwardItem:
        return jsValueForDecodedMessage<MessageName::WebPage_GoToBackForwardItem>(globalObject, decoder);
    case MessageName::WebPage_TryRestoreScrollPosition:
        return jsValueForDecodedMessage<MessageName::WebPage_TryRestoreScrollPosition>(globalObject, decoder);
    case MessageName::WebPage_LoadURLInFrame:
        return jsValueForDecodedMessage<MessageName::WebPage_LoadURLInFrame>(globalObject, decoder);
    case MessageName::WebPage_LoadDataInFrame:
        return jsValueForDecodedMessage<MessageName::WebPage_LoadDataInFrame>(globalObject, decoder);
    case MessageName::WebPage_LoadRequest:
        return jsValueForDecodedMessage<MessageName::WebPage_LoadRequest>(globalObject, decoder);
    case MessageName::WebPage_LoadRequestWaitingForProcessLaunch:
        return jsValueForDecodedMessage<MessageName::WebPage_LoadRequestWaitingForProcessLaunch>(globalObject, decoder);
    case MessageName::WebPage_LoadData:
        return jsValueForDecodedMessage<MessageName::WebPage_LoadData>(globalObject, decoder);
    case MessageName::WebPage_LoadSimulatedRequestAndResponse:
        return jsValueForDecodedMessage<MessageName::WebPage_LoadSimulatedRequestAndResponse>(globalObject, decoder);
    case MessageName::WebPage_LoadAlternateHTML:
        return jsValueForDecodedMessage<MessageName::WebPage_LoadAlternateHTML>(globalObject, decoder);
    case MessageName::WebPage_NavigateToPDFLinkWithSimulatedClick:
        return jsValueForDecodedMessage<MessageName::WebPage_NavigateToPDFLinkWithSimulatedClick>(globalObject, decoder);
    case MessageName::WebPage_GetPDFFirstPageSize:
        return jsValueForDecodedMessage<MessageName::WebPage_GetPDFFirstPageSize>(globalObject, decoder);
    case MessageName::WebPage_Reload:
        return jsValueForDecodedMessage<MessageName::WebPage_Reload>(globalObject, decoder);
    case MessageName::WebPage_StopLoading:
        return jsValueForDecodedMessage<MessageName::WebPage_StopLoading>(globalObject, decoder);
    case MessageName::WebPage_StopLoadingDueToProcessSwap:
        return jsValueForDecodedMessage<MessageName::WebPage_StopLoadingDueToProcessSwap>(globalObject, decoder);
    case MessageName::WebPage_RestoreSession:
        return jsValueForDecodedMessage<MessageName::WebPage_RestoreSession>(globalObject, decoder);
    case MessageName::WebPage_UpdateBackForwardListForReattach:
        return jsValueForDecodedMessage<MessageName::WebPage_UpdateBackForwardListForReattach>(globalObject, decoder);
    case MessageName::WebPage_SetCurrentHistoryItemForReattach:
        return jsValueForDecodedMessage<MessageName::WebPage_SetCurrentHistoryItemForReattach>(globalObject, decoder);
    case MessageName::WebPage_DidRemoveBackForwardItem:
        return jsValueForDecodedMessage<MessageName::WebPage_DidRemoveBackForwardItem>(globalObject, decoder);
    case MessageName::WebPage_UpdateWebsitePolicies:
        return jsValueForDecodedMessage<MessageName::WebPage_UpdateWebsitePolicies>(globalObject, decoder);
    case MessageName::WebPage_NotifyUserScripts:
        return jsValueForDecodedMessage<MessageName::WebPage_NotifyUserScripts>(globalObject, decoder);
#if ENABLE(CONTENT_FILTERING_IN_NETWORKING_PROCESS)
    case MessageName::WebPage_DidReceivePolicyDecision:
        return jsValueForDecodedMessage<MessageName::WebPage_DidReceivePolicyDecision>(globalObject, decoder);
#endif
#if !ENABLE(CONTENT_FILTERING_IN_NETWORKING_PROCESS)
    case MessageName::WebPage_DidReceivePolicyDecision:
        return jsValueForDecodedMessage<MessageName::WebPage_DidReceivePolicyDecision>(globalObject, decoder);
#endif
    case MessageName::WebPage_ClearSelection:
        return jsValueForDecodedMessage<MessageName::WebPage_ClearSelection>(globalObject, decoder);
    case MessageName::WebPage_RestoreSelectionInFocusedEditableElement:
        return jsValueForDecodedMessage<MessageName::WebPage_RestoreSelectionInFocusedEditableElement>(globalObject, decoder);
    case MessageName::WebPage_GetContentsAsString:
        return jsValueForDecodedMessage<MessageName::WebPage_GetContentsAsString>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::WebPage_GetContentsAsAttributedString:
        return jsValueForDecodedMessage<MessageName::WebPage_GetContentsAsAttributedString>(globalObject, decoder);
#endif
#if ENABLE(MHTML)
    case MessageName::WebPage_GetContentsAsMHTMLData:
        return jsValueForDecodedMessage<MessageName::WebPage_GetContentsAsMHTMLData>(globalObject, decoder);
#endif
    case MessageName::WebPage_GetMainResourceDataOfFrame:
        return jsValueForDecodedMessage<MessageName::WebPage_GetMainResourceDataOfFrame>(globalObject, decoder);
    case MessageName::WebPage_GetResourceDataFromFrame:
        return jsValueForDecodedMessage<MessageName::WebPage_GetResourceDataFromFrame>(globalObject, decoder);
    case MessageName::WebPage_GetRenderTreeExternalRepresentation:
        return jsValueForDecodedMessage<MessageName::WebPage_GetRenderTreeExternalRepresentation>(globalObject, decoder);
    case MessageName::WebPage_GetSelectionOrContentsAsString:
        return jsValueForDecodedMessage<MessageName::WebPage_GetSelectionOrContentsAsString>(globalObject, decoder);
    case MessageName::WebPage_GetSelectionAsWebArchiveData:
        return jsValueForDecodedMessage<MessageName::WebPage_GetSelectionAsWebArchiveData>(globalObject, decoder);
    case MessageName::WebPage_GetSourceForFrame:
        return jsValueForDecodedMessage<MessageName::WebPage_GetSourceForFrame>(globalObject, decoder);
    case MessageName::WebPage_GetWebArchiveOfFrame:
        return jsValueForDecodedMessage<MessageName::WebPage_GetWebArchiveOfFrame>(globalObject, decoder);
    case MessageName::WebPage_RunJavaScriptInFrameInScriptWorld:
        return jsValueForDecodedMessage<MessageName::WebPage_RunJavaScriptInFrameInScriptWorld>(globalObject, decoder);
    case MessageName::WebPage_GetAccessibilityTreeData:
        return jsValueForDecodedMessage<MessageName::WebPage_GetAccessibilityTreeData>(globalObject, decoder);
    case MessageName::WebPage_ForceRepaint:
        return jsValueForDecodedMessage<MessageName::WebPage_ForceRepaint>(globalObject, decoder);
    case MessageName::WebPage_SelectAll:
        return jsValueForDecodedMessage<MessageName::WebPage_SelectAll>(globalObject, decoder);
    case MessageName::WebPage_ScheduleFullEditorStateUpdate:
        return jsValueForDecodedMessage<MessageName::WebPage_ScheduleFullEditorStateUpdate>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::WebPage_PerformDictionaryLookupOfCurrentSelection:
        return jsValueForDecodedMessage<MessageName::WebPage_PerformDictionaryLookupOfCurrentSelection>(globalObject, decoder);
    case MessageName::WebPage_PerformDictionaryLookupAtLocation:
        return jsValueForDecodedMessage<MessageName::WebPage_PerformDictionaryLookupAtLocation>(globalObject, decoder);
#endif
#if ENABLE(DATA_DETECTION)
    case MessageName::WebPage_DetectDataInAllFrames:
        return jsValueForDecodedMessage<MessageName::WebPage_DetectDataInAllFrames>(globalObject, decoder);
    case MessageName::WebPage_RemoveDataDetectedLinks:
        return jsValueForDecodedMessage<MessageName::WebPage_RemoveDataDetectedLinks>(globalObject, decoder);
#endif
    case MessageName::WebPage_ChangeFont:
        return jsValueForDecodedMessage<MessageName::WebPage_ChangeFont>(globalObject, decoder);
    case MessageName::WebPage_ChangeFontAttributes:
        return jsValueForDecodedMessage<MessageName::WebPage_ChangeFontAttributes>(globalObject, decoder);
    case MessageName::WebPage_PreferencesDidChange:
        return jsValueForDecodedMessage<MessageName::WebPage_PreferencesDidChange>(globalObject, decoder);
    case MessageName::WebPage_PreferencesDidChangeDuringDOMPrintOperation:
        return jsValueForDecodedMessage<MessageName::WebPage_PreferencesDidChangeDuringDOMPrintOperation>(globalObject, decoder);
    case MessageName::WebPage_SetUserAgent:
        return jsValueForDecodedMessage<MessageName::WebPage_SetUserAgent>(globalObject, decoder);
    case MessageName::WebPage_SetCustomTextEncodingName:
        return jsValueForDecodedMessage<MessageName::WebPage_SetCustomTextEncodingName>(globalObject, decoder);
    case MessageName::WebPage_SuspendActiveDOMObjectsAndAnimations:
        return jsValueForDecodedMessage<MessageName::WebPage_SuspendActiveDOMObjectsAndAnimations>(globalObject, decoder);
    case MessageName::WebPage_ResumeActiveDOMObjectsAndAnimations:
        return jsValueForDecodedMessage<MessageName::WebPage_ResumeActiveDOMObjectsAndAnimations>(globalObject, decoder);
    case MessageName::WebPage_Suspend:
        return jsValueForDecodedMessage<MessageName::WebPage_Suspend>(globalObject, decoder);
    case MessageName::WebPage_Resume:
        return jsValueForDecodedMessage<MessageName::WebPage_Resume>(globalObject, decoder);
    case MessageName::WebPage_Close:
        return jsValueForDecodedMessage<MessageName::WebPage_Close>(globalObject, decoder);
    case MessageName::WebPage_TryClose:
        return jsValueForDecodedMessage<MessageName::WebPage_TryClose>(globalObject, decoder);
    case MessageName::WebPage_SetEditable:
        return jsValueForDecodedMessage<MessageName::WebPage_SetEditable>(globalObject, decoder);
    case MessageName::WebPage_ValidateCommand:
        return jsValueForDecodedMessage<MessageName::WebPage_ValidateCommand>(globalObject, decoder);
    case MessageName::WebPage_ExecuteEditCommand:
        return jsValueForDecodedMessage<MessageName::WebPage_ExecuteEditCommand>(globalObject, decoder);
    case MessageName::WebPage_IncreaseListLevel:
        return jsValueForDecodedMessage<MessageName::WebPage_IncreaseListLevel>(globalObject, decoder);
    case MessageName::WebPage_DecreaseListLevel:
        return jsValueForDecodedMessage<MessageName::WebPage_DecreaseListLevel>(globalObject, decoder);
    case MessageName::WebPage_ChangeListType:
        return jsValueForDecodedMessage<MessageName::WebPage_ChangeListType>(globalObject, decoder);
    case MessageName::WebPage_SetBaseWritingDirection:
        return jsValueForDecodedMessage<MessageName::WebPage_SetBaseWritingDirection>(globalObject, decoder);
    case MessageName::WebPage_SetNeedsFontAttributes:
        return jsValueForDecodedMessage<MessageName::WebPage_SetNeedsFontAttributes>(globalObject, decoder);
    case MessageName::WebPage_RequestFontAttributesAtSelectionStart:
        return jsValueForDecodedMessage<MessageName::WebPage_RequestFontAttributesAtSelectionStart>(globalObject, decoder);
    case MessageName::WebPage_DidRemoveEditCommand:
        return jsValueForDecodedMessage<MessageName::WebPage_DidRemoveEditCommand>(globalObject, decoder);
    case MessageName::WebPage_ReapplyEditCommand:
        return jsValueForDecodedMessage<MessageName::WebPage_ReapplyEditCommand>(globalObject, decoder);
    case MessageName::WebPage_UnapplyEditCommand:
        return jsValueForDecodedMessage<MessageName::WebPage_UnapplyEditCommand>(globalObject, decoder);
    case MessageName::WebPage_SetPageAndTextZoomFactors:
        return jsValueForDecodedMessage<MessageName::WebPage_SetPageAndTextZoomFactors>(globalObject, decoder);
    case MessageName::WebPage_SetPageZoomFactor:
        return jsValueForDecodedMessage<MessageName::WebPage_SetPageZoomFactor>(globalObject, decoder);
    case MessageName::WebPage_SetTextZoomFactor:
        return jsValueForDecodedMessage<MessageName::WebPage_SetTextZoomFactor>(globalObject, decoder);
    case MessageName::WebPage_WindowScreenDidChange:
        return jsValueForDecodedMessage<MessageName::WebPage_WindowScreenDidChange>(globalObject, decoder);
    case MessageName::WebPage_AccessibilitySettingsDidChange:
        return jsValueForDecodedMessage<MessageName::WebPage_AccessibilitySettingsDidChange>(globalObject, decoder);
    case MessageName::WebPage_ScalePage:
        return jsValueForDecodedMessage<MessageName::WebPage_ScalePage>(globalObject, decoder);
    case MessageName::WebPage_ScalePageInViewCoordinates:
        return jsValueForDecodedMessage<MessageName::WebPage_ScalePageInViewCoordinates>(globalObject, decoder);
    case MessageName::WebPage_ScaleView:
        return jsValueForDecodedMessage<MessageName::WebPage_ScaleView>(globalObject, decoder);
    case MessageName::WebPage_SetUseFixedLayout:
        return jsValueForDecodedMessage<MessageName::WebPage_SetUseFixedLayout>(globalObject, decoder);
    case MessageName::WebPage_SetFixedLayoutSize:
        return jsValueForDecodedMessage<MessageName::WebPage_SetFixedLayoutSize>(globalObject, decoder);
    case MessageName::WebPage_SetDefaultUnobscuredSize:
        return jsValueForDecodedMessage<MessageName::WebPage_SetDefaultUnobscuredSize>(globalObject, decoder);
    case MessageName::WebPage_SetMinimumUnobscuredSize:
        return jsValueForDecodedMessage<MessageName::WebPage_SetMinimumUnobscuredSize>(globalObject, decoder);
    case MessageName::WebPage_SetMaximumUnobscuredSize:
        return jsValueForDecodedMessage<MessageName::WebPage_SetMaximumUnobscuredSize>(globalObject, decoder);
    case MessageName::WebPage_ListenForLayoutMilestones:
        return jsValueForDecodedMessage<MessageName::WebPage_ListenForLayoutMilestones>(globalObject, decoder);
    case MessageName::WebPage_SetSuppressScrollbarAnimations:
        return jsValueForDecodedMessage<MessageName::WebPage_SetSuppressScrollbarAnimations>(globalObject, decoder);
    case MessageName::WebPage_SetEnableVerticalRubberBanding:
        return jsValueForDecodedMessage<MessageName::WebPage_SetEnableVerticalRubberBanding>(globalObject, decoder);
    case MessageName::WebPage_SetEnableHorizontalRubberBanding:
        return jsValueForDecodedMessage<MessageName::WebPage_SetEnableHorizontalRubberBanding>(globalObject, decoder);
    case MessageName::WebPage_SetBackgroundExtendsBeyondPage:
        return jsValueForDecodedMessage<MessageName::WebPage_SetBackgroundExtendsBeyondPage>(globalObject, decoder);
    case MessageName::WebPage_SetPaginationMode:
        return jsValueForDecodedMessage<MessageName::WebPage_SetPaginationMode>(globalObject, decoder);
    case MessageName::WebPage_SetPaginationBehavesLikeColumns:
        return jsValueForDecodedMessage<MessageName::WebPage_SetPaginationBehavesLikeColumns>(globalObject, decoder);
    case MessageName::WebPage_SetPageLength:
        return jsValueForDecodedMessage<MessageName::WebPage_SetPageLength>(globalObject, decoder);
    case MessageName::WebPage_SetGapBetweenPages:
        return jsValueForDecodedMessage<MessageName::WebPage_SetGapBetweenPages>(globalObject, decoder);
    case MessageName::WebPage_PostInjectedBundleMessage:
        return jsValueForDecodedMessage<MessageName::WebPage_PostInjectedBundleMessage>(globalObject, decoder);
    case MessageName::WebPage_FindString:
        return jsValueForDecodedMessage<MessageName::WebPage_FindString>(globalObject, decoder);
    case MessageName::WebPage_FindStringMatches:
        return jsValueForDecodedMessage<MessageName::WebPage_FindStringMatches>(globalObject, decoder);
    case MessageName::WebPage_GetImageForFindMatch:
        return jsValueForDecodedMessage<MessageName::WebPage_GetImageForFindMatch>(globalObject, decoder);
    case MessageName::WebPage_SelectFindMatch:
        return jsValueForDecodedMessage<MessageName::WebPage_SelectFindMatch>(globalObject, decoder);
    case MessageName::WebPage_IndicateFindMatch:
        return jsValueForDecodedMessage<MessageName::WebPage_IndicateFindMatch>(globalObject, decoder);
    case MessageName::WebPage_HideFindUI:
        return jsValueForDecodedMessage<MessageName::WebPage_HideFindUI>(globalObject, decoder);
    case MessageName::WebPage_CountStringMatches:
        return jsValueForDecodedMessage<MessageName::WebPage_CountStringMatches>(globalObject, decoder);
    case MessageName::WebPage_ReplaceMatches:
        return jsValueForDecodedMessage<MessageName::WebPage_ReplaceMatches>(globalObject, decoder);
    case MessageName::WebPage_FindRectsForStringMatches:
        return jsValueForDecodedMessage<MessageName::WebPage_FindRectsForStringMatches>(globalObject, decoder);
    case MessageName::WebPage_HideFindIndicator:
        return jsValueForDecodedMessage<MessageName::WebPage_HideFindIndicator>(globalObject, decoder);
    case MessageName::WebPage_FindTextRangesForStringMatches:
        return jsValueForDecodedMessage<MessageName::WebPage_FindTextRangesForStringMatches>(globalObject, decoder);
    case MessageName::WebPage_ReplaceFoundTextRangeWithString:
        return jsValueForDecodedMessage<MessageName::WebPage_ReplaceFoundTextRangeWithString>(globalObject, decoder);
    case MessageName::WebPage_DecorateTextRangeWithStyle:
        return jsValueForDecodedMessage<MessageName::WebPage_DecorateTextRangeWithStyle>(globalObject, decoder);
    case MessageName::WebPage_ScrollTextRangeToVisible:
        return jsValueForDecodedMessage<MessageName::WebPage_ScrollTextRangeToVisible>(globalObject, decoder);
    case MessageName::WebPage_ClearAllDecoratedFoundText:
        return jsValueForDecodedMessage<MessageName::WebPage_ClearAllDecoratedFoundText>(globalObject, decoder);
    case MessageName::WebPage_DidBeginTextSearchOperation:
        return jsValueForDecodedMessage<MessageName::WebPage_DidBeginTextSearchOperation>(globalObject, decoder);
    case MessageName::WebPage_DidEndTextSearchOperation:
        return jsValueForDecodedMessage<MessageName::WebPage_DidEndTextSearchOperation>(globalObject, decoder);
    case MessageName::WebPage_RequestRectForFoundTextRange:
        return jsValueForDecodedMessage<MessageName::WebPage_RequestRectForFoundTextRange>(globalObject, decoder);
    case MessageName::WebPage_AddLayerForFindOverlay:
        return jsValueForDecodedMessage<MessageName::WebPage_AddLayerForFindOverlay>(globalObject, decoder);
    case MessageName::WebPage_RemoveLayerForFindOverlay:
        return jsValueForDecodedMessage<MessageName::WebPage_RemoveLayerForFindOverlay>(globalObject, decoder);
    case MessageName::WebPage_AddMIMETypeWithCustomContentProvider:
        return jsValueForDecodedMessage<MessageName::WebPage_AddMIMETypeWithCustomContentProvider>(globalObject, decoder);
#if PLATFORM(GTK) && ENABLE(DRAG_SUPPORT)
    case MessageName::WebPage_PerformDragControllerAction:
        return jsValueForDecodedMessage<MessageName::WebPage_PerformDragControllerAction>(globalObject, decoder);
#endif
#if !PLATFORM(GTK) && ENABLE(DRAG_SUPPORT)
    case MessageName::WebPage_PerformDragControllerAction:
        return jsValueForDecodedMessage<MessageName::WebPage_PerformDragControllerAction>(globalObject, decoder);
#endif
#if ENABLE(DRAG_SUPPORT)
    case MessageName::WebPage_DidStartDrag:
        return jsValueForDecodedMessage<MessageName::WebPage_DidStartDrag>(globalObject, decoder);
    case MessageName::WebPage_DragEnded:
        return jsValueForDecodedMessage<MessageName::WebPage_DragEnded>(globalObject, decoder);
    case MessageName::WebPage_DragCancelled:
        return jsValueForDecodedMessage<MessageName::WebPage_DragCancelled>(globalObject, decoder);
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(DRAG_SUPPORT)
    case MessageName::WebPage_RequestDragStart:
        return jsValueForDecodedMessage<MessageName::WebPage_RequestDragStart>(globalObject, decoder);
    case MessageName::WebPage_RequestAdditionalItemsForDragSession:
        return jsValueForDecodedMessage<MessageName::WebPage_RequestAdditionalItemsForDragSession>(globalObject, decoder);
    case MessageName::WebPage_InsertDroppedImagePlaceholders:
        return jsValueForDecodedMessage<MessageName::WebPage_InsertDroppedImagePlaceholders>(globalObject, decoder);
    case MessageName::WebPage_DidConcludeDrop:
        return jsValueForDecodedMessage<MessageName::WebPage_DidConcludeDrop>(globalObject, decoder);
#endif
    case MessageName::WebPage_DidChangeSelectedIndexForActivePopupMenu:
        return jsValueForDecodedMessage<MessageName::WebPage_DidChangeSelectedIndexForActivePopupMenu>(globalObject, decoder);
    case MessageName::WebPage_SetTextForActivePopupMenu:
        return jsValueForDecodedMessage<MessageName::WebPage_SetTextForActivePopupMenu>(globalObject, decoder);
#if PLATFORM(GTK)
    case MessageName::WebPage_FailedToShowPopupMenu:
        return jsValueForDecodedMessage<MessageName::WebPage_FailedToShowPopupMenu>(globalObject, decoder);
#endif
#if ENABLE(CONTEXT_MENUS)
    case MessageName::WebPage_DidShowContextMenu:
        return jsValueForDecodedMessage<MessageName::WebPage_DidShowContextMenu>(globalObject, decoder);
    case MessageName::WebPage_DidDismissContextMenu:
        return jsValueForDecodedMessage<MessageName::WebPage_DidDismissContextMenu>(globalObject, decoder);
    case MessageName::WebPage_DidSelectItemFromActiveContextMenu:
        return jsValueForDecodedMessage<MessageName::WebPage_DidSelectItemFromActiveContextMenu>(globalObject, decoder);
    case MessageName::WebPage_ContextMenuForKeyEvent:
        return jsValueForDecodedMessage<MessageName::WebPage_ContextMenuForKeyEvent>(globalObject, decoder);
#endif
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebPage_DidChooseFilesForOpenPanelWithDisplayStringAndIcon:
        return jsValueForDecodedMessage<MessageName::WebPage_DidChooseFilesForOpenPanelWithDisplayStringAndIcon>(globalObject, decoder);
#endif
    case MessageName::WebPage_DidChooseFilesForOpenPanel:
        return jsValueForDecodedMessage<MessageName::WebPage_DidChooseFilesForOpenPanel>(globalObject, decoder);
    case MessageName::WebPage_DidCancelForOpenPanel:
        return jsValueForDecodedMessage<MessageName::WebPage_DidCancelForOpenPanel>(globalObject, decoder);
#if ENABLE(SANDBOX_EXTENSIONS)
    case MessageName::WebPage_ExtendSandboxForFilesFromOpenPanel:
        return jsValueForDecodedMessage<MessageName::WebPage_ExtendSandboxForFilesFromOpenPanel>(globalObject, decoder);
#endif
    case MessageName::WebPage_AdvanceToNextMisspelling:
        return jsValueForDecodedMessage<MessageName::WebPage_AdvanceToNextMisspelling>(globalObject, decoder);
    case MessageName::WebPage_ChangeSpellingToWord:
        return jsValueForDecodedMessage<MessageName::WebPage_ChangeSpellingToWord>(globalObject, decoder);
    case MessageName::WebPage_DidFinishCheckingText:
        return jsValueForDecodedMessage<MessageName::WebPage_DidFinishCheckingText>(globalObject, decoder);
    case MessageName::WebPage_DidCancelCheckingText:
        return jsValueForDecodedMessage<MessageName::WebPage_DidCancelCheckingText>(globalObject, decoder);
#if USE(APPKIT)
    case MessageName::WebPage_UppercaseWord:
        return jsValueForDecodedMessage<MessageName::WebPage_UppercaseWord>(globalObject, decoder);
    case MessageName::WebPage_LowercaseWord:
        return jsValueForDecodedMessage<MessageName::WebPage_LowercaseWord>(globalObject, decoder);
    case MessageName::WebPage_CapitalizeWord:
        return jsValueForDecodedMessage<MessageName::WebPage_CapitalizeWord>(globalObject, decoder);
#endif
#if PLATFORM(COCOA)
    case MessageName::WebPage_SetSmartInsertDeleteEnabled:
        return jsValueForDecodedMessage<MessageName::WebPage_SetSmartInsertDeleteEnabled>(globalObject, decoder);
#endif
#if ENABLE(GEOLOCATION)
    case MessageName::WebPage_DidReceiveGeolocationPermissionDecision:
        return jsValueForDecodedMessage<MessageName::WebPage_DidReceiveGeolocationPermissionDecision>(globalObject, decoder);
#endif
#if ENABLE(MEDIA_STREAM)
    case MessageName::WebPage_UserMediaAccessWasGranted:
        return jsValueForDecodedMessage<MessageName::WebPage_UserMediaAccessWasGranted>(globalObject, decoder);
    case MessageName::WebPage_UserMediaAccessWasDenied:
        return jsValueForDecodedMessage<MessageName::WebPage_UserMediaAccessWasDenied>(globalObject, decoder);
    case MessageName::WebPage_CaptureDevicesChanged:
        return jsValueForDecodedMessage<MessageName::WebPage_CaptureDevicesChanged>(globalObject, decoder);
#endif
#if (ENABLE(MEDIA_STREAM) && USE(GSTREAMER))
    case MessageName::WebPage_SetOrientationForMediaCapture:
        return jsValueForDecodedMessage<MessageName::WebPage_SetOrientationForMediaCapture>(globalObject, decoder);
    case MessageName::WebPage_SetMockCaptureDevicesInterrupted:
        return jsValueForDecodedMessage<MessageName::WebPage_SetMockCaptureDevicesInterrupted>(globalObject, decoder);
#endif
#if ENABLE(ENCRYPTED_MEDIA)
    case MessageName::WebPage_MediaKeySystemWasGranted:
        return jsValueForDecodedMessage<MessageName::WebPage_MediaKeySystemWasGranted>(globalObject, decoder);
    case MessageName::WebPage_MediaKeySystemWasDenied:
        return jsValueForDecodedMessage<MessageName::WebPage_MediaKeySystemWasDenied>(globalObject, decoder);
#endif
    case MessageName::WebPage_RequestMediaPlaybackState:
        return jsValueForDecodedMessage<MessageName::WebPage_RequestMediaPlaybackState>(globalObject, decoder);
    case MessageName::WebPage_PauseAllMediaPlayback:
        return jsValueForDecodedMessage<MessageName::WebPage_PauseAllMediaPlayback>(globalObject, decoder);
    case MessageName::WebPage_SuspendAllMediaPlayback:
        return jsValueForDecodedMessage<MessageName::WebPage_SuspendAllMediaPlayback>(globalObject, decoder);
    case MessageName::WebPage_ResumeAllMediaPlayback:
        return jsValueForDecodedMessage<MessageName::WebPage_ResumeAllMediaPlayback>(globalObject, decoder);
    case MessageName::WebPage_FreezeLayerTreeDueToSwipeAnimation:
        return jsValueForDecodedMessage<MessageName::WebPage_FreezeLayerTreeDueToSwipeAnimation>(globalObject, decoder);
    case MessageName::WebPage_UnfreezeLayerTreeDueToSwipeAnimation:
        return jsValueForDecodedMessage<MessageName::WebPage_UnfreezeLayerTreeDueToSwipeAnimation>(globalObject, decoder);
    case MessageName::WebPage_IsLayerTreeFrozen:
        return jsValueForDecodedMessage<MessageName::WebPage_IsLayerTreeFrozen>(globalObject, decoder);
    case MessageName::WebPage_BeginPrinting:
        return jsValueForDecodedMessage<MessageName::WebPage_BeginPrinting>(globalObject, decoder);
    case MessageName::WebPage_BeginPrintingDuringDOMPrintOperation:
        return jsValueForDecodedMessage<MessageName::WebPage_BeginPrintingDuringDOMPrintOperation>(globalObject, decoder);
    case MessageName::WebPage_EndPrinting:
        return jsValueForDecodedMessage<MessageName::WebPage_EndPrinting>(globalObject, decoder);
    case MessageName::WebPage_EndPrintingDuringDOMPrintOperation:
        return jsValueForDecodedMessage<MessageName::WebPage_EndPrintingDuringDOMPrintOperation>(globalObject, decoder);
    case MessageName::WebPage_ComputePagesForPrinting:
        return jsValueForDecodedMessage<MessageName::WebPage_ComputePagesForPrinting>(globalObject, decoder);
    case MessageName::WebPage_ComputePagesForPrintingDuringDOMPrintOperation:
        return jsValueForDecodedMessage<MessageName::WebPage_ComputePagesForPrintingDuringDOMPrintOperation>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::WebPage_DrawRectToImage:
        return jsValueForDecodedMessage<MessageName::WebPage_DrawRectToImage>(globalObject, decoder);
    case MessageName::WebPage_DrawRectToImageDuringDOMPrintOperation:
        return jsValueForDecodedMessage<MessageName::WebPage_DrawRectToImageDuringDOMPrintOperation>(globalObject, decoder);
    case MessageName::WebPage_DrawPagesToPDF:
        return jsValueForDecodedMessage<MessageName::WebPage_DrawPagesToPDF>(globalObject, decoder);
    case MessageName::WebPage_DrawPagesToPDFDuringDOMPrintOperation:
        return jsValueForDecodedMessage<MessageName::WebPage_DrawPagesToPDFDuringDOMPrintOperation>(globalObject, decoder);
#endif
#if (PLATFORM(COCOA) && PLATFORM(IOS_FAMILY))
    case MessageName::WebPage_ComputePagesForPrintingiOS:
        return jsValueForDecodedMessage<MessageName::WebPage_ComputePagesForPrintingiOS>(globalObject, decoder);
    case MessageName::WebPage_DrawToPDFiOS:
        return jsValueForDecodedMessage<MessageName::WebPage_DrawToPDFiOS>(globalObject, decoder);
#endif
#if PLATFORM(COCOA)
    case MessageName::WebPage_DrawToPDF:
        return jsValueForDecodedMessage<MessageName::WebPage_DrawToPDF>(globalObject, decoder);
#endif
#if PLATFORM(GTK)
    case MessageName::WebPage_DrawPagesForPrinting:
        return jsValueForDecodedMessage<MessageName::WebPage_DrawPagesForPrinting>(globalObject, decoder);
    case MessageName::WebPage_DrawPagesForPrintingDuringDOMPrintOperation:
        return jsValueForDecodedMessage<MessageName::WebPage_DrawPagesForPrintingDuringDOMPrintOperation>(globalObject, decoder);
#endif
    case MessageName::WebPage_SetMediaVolume:
        return jsValueForDecodedMessage<MessageName::WebPage_SetMediaVolume>(globalObject, decoder);
    case MessageName::WebPage_SetMuted:
        return jsValueForDecodedMessage<MessageName::WebPage_SetMuted>(globalObject, decoder);
    case MessageName::WebPage_SetMayStartMediaWhenInWindow:
        return jsValueForDecodedMessage<MessageName::WebPage_SetMayStartMediaWhenInWindow>(globalObject, decoder);
    case MessageName::WebPage_StopMediaCapture:
        return jsValueForDecodedMessage<MessageName::WebPage_StopMediaCapture>(globalObject, decoder);
    case MessageName::WebPage_SetCanRunBeforeUnloadConfirmPanel:
        return jsValueForDecodedMessage<MessageName::WebPage_SetCanRunBeforeUnloadConfirmPanel>(globalObject, decoder);
    case MessageName::WebPage_SetCanRunModal:
        return jsValueForDecodedMessage<MessageName::WebPage_SetCanRunModal>(globalObject, decoder);
#if PLATFORM(GTK) || PLATFORM(WPE)
    case MessageName::WebPage_CancelComposition:
        return jsValueForDecodedMessage<MessageName::WebPage_CancelComposition>(globalObject, decoder);
    case MessageName::WebPage_DeleteSurrounding:
        return jsValueForDecodedMessage<MessageName::WebPage_DeleteSurrounding>(globalObject, decoder);
#endif
#if PLATFORM(GTK)
    case MessageName::WebPage_CollapseSelectionInFrame:
        return jsValueForDecodedMessage<MessageName::WebPage_CollapseSelectionInFrame>(globalObject, decoder);
#endif
#if PLATFORM(COCOA)
    case MessageName::WebPage_WindowAndViewFramesChanged:
        return jsValueForDecodedMessage<MessageName::WebPage_WindowAndViewFramesChanged>(globalObject, decoder);
    case MessageName::WebPage_SetMainFrameIsScrollable:
        return jsValueForDecodedMessage<MessageName::WebPage_SetMainFrameIsScrollable>(globalObject, decoder);
    case MessageName::WebPage_RegisterUIProcessAccessibilityTokens:
        return jsValueForDecodedMessage<MessageName::WebPage_RegisterUIProcessAccessibilityTokens>(globalObject, decoder);
    case MessageName::WebPage_GetStringSelectionForPasteboard:
        return jsValueForDecodedMessage<MessageName::WebPage_GetStringSelectionForPasteboard>(globalObject, decoder);
    case MessageName::WebPage_GetDataSelectionForPasteboard:
        return jsValueForDecodedMessage<MessageName::WebPage_GetDataSelectionForPasteboard>(globalObject, decoder);
    case MessageName::WebPage_ReadSelectionFromPasteboard:
        return jsValueForDecodedMessage<MessageName::WebPage_ReadSelectionFromPasteboard>(globalObject, decoder);
    case MessageName::WebPage_ReplaceSelectionWithPasteboardData:
        return jsValueForDecodedMessage<MessageName::WebPage_ReplaceSelectionWithPasteboardData>(globalObject, decoder);
    case MessageName::WebPage_ShouldDelayWindowOrderingEvent:
        return jsValueForDecodedMessage<MessageName::WebPage_ShouldDelayWindowOrderingEvent>(globalObject, decoder);
    case MessageName::WebPage_SetTextAsync:
        return jsValueForDecodedMessage<MessageName::WebPage_SetTextAsync>(globalObject, decoder);
    case MessageName::WebPage_InsertTextAsync:
        return jsValueForDecodedMessage<MessageName::WebPage_InsertTextAsync>(globalObject, decoder);
    case MessageName::WebPage_InsertDictatedTextAsync:
        return jsValueForDecodedMessage<MessageName::WebPage_InsertDictatedTextAsync>(globalObject, decoder);
    case MessageName::WebPage_AddDictationAlternative:
        return jsValueForDecodedMessage<MessageName::WebPage_AddDictationAlternative>(globalObject, decoder);
    case MessageName::WebPage_DictationAlternativesAtSelection:
        return jsValueForDecodedMessage<MessageName::WebPage_DictationAlternativesAtSelection>(globalObject, decoder);
    case MessageName::WebPage_ClearDictationAlternatives:
        return jsValueForDecodedMessage<MessageName::WebPage_ClearDictationAlternatives>(globalObject, decoder);
    case MessageName::WebPage_HasMarkedText:
        return jsValueForDecodedMessage<MessageName::WebPage_HasMarkedText>(globalObject, decoder);
    case MessageName::WebPage_GetMarkedRangeAsync:
        return jsValueForDecodedMessage<MessageName::WebPage_GetMarkedRangeAsync>(globalObject, decoder);
    case MessageName::WebPage_GetSelectedRangeAsync:
        return jsValueForDecodedMessage<MessageName::WebPage_GetSelectedRangeAsync>(globalObject, decoder);
    case MessageName::WebPage_CharacterIndexForPointAsync:
        return jsValueForDecodedMessage<MessageName::WebPage_CharacterIndexForPointAsync>(globalObject, decoder);
    case MessageName::WebPage_FirstRectForCharacterRangeAsync:
        return jsValueForDecodedMessage<MessageName::WebPage_FirstRectForCharacterRangeAsync>(globalObject, decoder);
    case MessageName::WebPage_SetCompositionAsync:
        return jsValueForDecodedMessage<MessageName::WebPage_SetCompositionAsync>(globalObject, decoder);
    case MessageName::WebPage_ConfirmCompositionAsync:
        return jsValueForDecodedMessage<MessageName::WebPage_ConfirmCompositionAsync>(globalObject, decoder);
#endif
#if PLATFORM(MAC)
    case MessageName::WebPage_AttributedSubstringForCharacterRangeAsync:
        return jsValueForDecodedMessage<MessageName::WebPage_AttributedSubstringForCharacterRangeAsync>(globalObject, decoder);
    case MessageName::WebPage_RequestAcceptsFirstMouse:
        return jsValueForDecodedMessage<MessageName::WebPage_RequestAcceptsFirstMouse>(globalObject, decoder);
#endif
#if ENABLE(IMAGE_ANALYSIS_ENHANCEMENTS)
    case MessageName::WebPage_ReplaceImageForRemoveBackground:
        return jsValueForDecodedMessage<MessageName::WebPage_ReplaceImageForRemoveBackground>(globalObject, decoder);
    case MessageName::WebPage_ShouldAllowRemoveBackground:
        return jsValueForDecodedMessage<MessageName::WebPage_ShouldAllowRemoveBackground>(globalObject, decoder);
#endif
    case MessageName::WebPage_SetAlwaysShowsHorizontalScroller:
        return jsValueForDecodedMessage<MessageName::WebPage_SetAlwaysShowsHorizontalScroller>(globalObject, decoder);
    case MessageName::WebPage_SetAlwaysShowsVerticalScroller:
        return jsValueForDecodedMessage<MessageName::WebPage_SetAlwaysShowsVerticalScroller>(globalObject, decoder);
    case MessageName::WebPage_SetMinimumSizeForAutoLayout:
        return jsValueForDecodedMessage<MessageName::WebPage_SetMinimumSizeForAutoLayout>(globalObject, decoder);
    case MessageName::WebPage_SetSizeToContentAutoSizeMaximumSize:
        return jsValueForDecodedMessage<MessageName::WebPage_SetSizeToContentAutoSizeMaximumSize>(globalObject, decoder);
    case MessageName::WebPage_SetAutoSizingShouldExpandToViewHeight:
        return jsValueForDecodedMessage<MessageName::WebPage_SetAutoSizingShouldExpandToViewHeight>(globalObject, decoder);
    case MessageName::WebPage_SetViewportSizeForCSSViewportUnits:
        return jsValueForDecodedMessage<MessageName::WebPage_SetViewportSizeForCSSViewportUnits>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::WebPage_HandleAlternativeTextUIResult:
        return jsValueForDecodedMessage<MessageName::WebPage_HandleAlternativeTextUIResult>(globalObject, decoder);
#endif
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebPage_WillStartUserTriggeredZooming:
        return jsValueForDecodedMessage<MessageName::WebPage_WillStartUserTriggeredZooming>(globalObject, decoder);
#endif
    case MessageName::WebPage_SetScrollPinningBehavior:
        return jsValueForDecodedMessage<MessageName::WebPage_SetScrollPinningBehavior>(globalObject, decoder);
    case MessageName::WebPage_SetScrollbarOverlayStyle:
        return jsValueForDecodedMessage<MessageName::WebPage_SetScrollbarOverlayStyle>(globalObject, decoder);
    case MessageName::WebPage_GetBytecodeProfile:
        return jsValueForDecodedMessage<MessageName::WebPage_GetBytecodeProfile>(globalObject, decoder);
    case MessageName::WebPage_GetSamplingProfilerOutput:
        return jsValueForDecodedMessage<MessageName::WebPage_GetSamplingProfilerOutput>(globalObject, decoder);
    case MessageName::WebPage_TakeSnapshot:
        return jsValueForDecodedMessage<MessageName::WebPage_TakeSnapshot>(globalObject, decoder);
#if PLATFORM(MAC)
    case MessageName::WebPage_PerformImmediateActionHitTestAtLocation:
        return jsValueForDecodedMessage<MessageName::WebPage_PerformImmediateActionHitTestAtLocation>(globalObject, decoder);
    case MessageName::WebPage_ImmediateActionDidUpdate:
        return jsValueForDecodedMessage<MessageName::WebPage_ImmediateActionDidUpdate>(globalObject, decoder);
    case MessageName::WebPage_ImmediateActionDidCancel:
        return jsValueForDecodedMessage<MessageName::WebPage_ImmediateActionDidCancel>(globalObject, decoder);
    case MessageName::WebPage_ImmediateActionDidComplete:
        return jsValueForDecodedMessage<MessageName::WebPage_ImmediateActionDidComplete>(globalObject, decoder);
    case MessageName::WebPage_DataDetectorsDidPresentUI:
        return jsValueForDecodedMessage<MessageName::WebPage_DataDetectorsDidPresentUI>(globalObject, decoder);
    case MessageName::WebPage_DataDetectorsDidChangeUI:
        return jsValueForDecodedMessage<MessageName::WebPage_DataDetectorsDidChangeUI>(globalObject, decoder);
    case MessageName::WebPage_DataDetectorsDidHideUI:
        return jsValueForDecodedMessage<MessageName::WebPage_DataDetectorsDidHideUI>(globalObject, decoder);
    case MessageName::WebPage_HandleAcceptedCandidate:
        return jsValueForDecodedMessage<MessageName::WebPage_HandleAcceptedCandidate>(globalObject, decoder);
    case MessageName::WebPage_SetUseSystemAppearance:
        return jsValueForDecodedMessage<MessageName::WebPage_SetUseSystemAppearance>(globalObject, decoder);
    case MessageName::WebPage_SemanticContextDidChange:
        return jsValueForDecodedMessage<MessageName::WebPage_SemanticContextDidChange>(globalObject, decoder);
    case MessageName::WebPage_SetHeaderBannerHeightForTesting:
        return jsValueForDecodedMessage<MessageName::WebPage_SetHeaderBannerHeightForTesting>(globalObject, decoder);
    case MessageName::WebPage_SetFooterBannerHeightForTesting:
        return jsValueForDecodedMessage<MessageName::WebPage_SetFooterBannerHeightForTesting>(globalObject, decoder);
    case MessageName::WebPage_DidEndMagnificationGesture:
        return jsValueForDecodedMessage<MessageName::WebPage_DidEndMagnificationGesture>(globalObject, decoder);
#endif
    case MessageName::WebPage_EffectiveAppearanceDidChange:
        return jsValueForDecodedMessage<MessageName::WebPage_EffectiveAppearanceDidChange>(globalObject, decoder);
#if HAVE(APP_ACCENT_COLORS)
    case MessageName::WebPage_SetAccentColor:
        return jsValueForDecodedMessage<MessageName::WebPage_SetAccentColor>(globalObject, decoder);
#endif
#if PLATFORM(COCOA)
    case MessageName::WebPage_RequestActiveNowPlayingSessionInfo:
        return jsValueForDecodedMessage<MessageName::WebPage_RequestActiveNowPlayingSessionInfo>(globalObject, decoder);
#endif
    case MessageName::WebPage_SetShouldDispatchFakeMouseMoveEvents:
        return jsValueForDecodedMessage<MessageName::WebPage_SetShouldDispatchFakeMouseMoveEvents>(globalObject, decoder);
#if ENABLE(WIRELESS_PLAYBACK_TARGET) && !PLATFORM(IOS_FAMILY)
    case MessageName::WebPage_PlaybackTargetSelected:
        return jsValueForDecodedMessage<MessageName::WebPage_PlaybackTargetSelected>(globalObject, decoder);
    case MessageName::WebPage_PlaybackTargetAvailabilityDidChange:
        return jsValueForDecodedMessage<MessageName::WebPage_PlaybackTargetAvailabilityDidChange>(globalObject, decoder);
    case MessageName::WebPage_SetShouldPlayToPlaybackTarget:
        return jsValueForDecodedMessage<MessageName::WebPage_SetShouldPlayToPlaybackTarget>(globalObject, decoder);
    case MessageName::WebPage_PlaybackTargetPickerWasDismissed:
        return jsValueForDecodedMessage<MessageName::WebPage_PlaybackTargetPickerWasDismissed>(globalObject, decoder);
#endif
#if ENABLE(POINTER_LOCK)
    case MessageName::WebPage_DidAcquirePointerLock:
        return jsValueForDecodedMessage<MessageName::WebPage_DidAcquirePointerLock>(globalObject, decoder);
    case MessageName::WebPage_DidNotAcquirePointerLock:
        return jsValueForDecodedMessage<MessageName::WebPage_DidNotAcquirePointerLock>(globalObject, decoder);
    case MessageName::WebPage_DidLosePointerLock:
        return jsValueForDecodedMessage<MessageName::WebPage_DidLosePointerLock>(globalObject, decoder);
#endif
    case MessageName::WebPage_ClearWheelEventTestMonitor:
        return jsValueForDecodedMessage<MessageName::WebPage_ClearWheelEventTestMonitor>(globalObject, decoder);
    case MessageName::WebPage_SetShouldScaleViewToFitDocument:
        return jsValueForDecodedMessage<MessageName::WebPage_SetShouldScaleViewToFitDocument>(globalObject, decoder);
    case MessageName::WebPage_SetUserInterfaceLayoutDirection:
        return jsValueForDecodedMessage<MessageName::WebPage_SetUserInterfaceLayoutDirection>(globalObject, decoder);
    case MessageName::WebPage_DidGetLoadDecisionForIcon:
        return jsValueForDecodedMessage<MessageName::WebPage_DidGetLoadDecisionForIcon>(globalObject, decoder);
    case MessageName::WebPage_SetUseIconLoadingClient:
        return jsValueForDecodedMessage<MessageName::WebPage_SetUseIconLoadingClient>(globalObject, decoder);
#if ENABLE(GAMEPAD)
    case MessageName::WebPage_GamepadActivity:
        return jsValueForDecodedMessage<MessageName::WebPage_GamepadActivity>(globalObject, decoder);
#endif
    case MessageName::WebPage_RegisterURLSchemeHandler:
        return jsValueForDecodedMessage<MessageName::WebPage_RegisterURLSchemeHandler>(globalObject, decoder);
    case MessageName::WebPage_URLSchemeTaskWillPerformRedirection:
        return jsValueForDecodedMessage<MessageName::WebPage_URLSchemeTaskWillPerformRedirection>(globalObject, decoder);
    case MessageName::WebPage_URLSchemeTaskDidPerformRedirection:
        return jsValueForDecodedMessage<MessageName::WebPage_URLSchemeTaskDidPerformRedirection>(globalObject, decoder);
    case MessageName::WebPage_URLSchemeTaskDidReceiveResponse:
        return jsValueForDecodedMessage<MessageName::WebPage_URLSchemeTaskDidReceiveResponse>(globalObject, decoder);
    case MessageName::WebPage_URLSchemeTaskDidReceiveData:
        return jsValueForDecodedMessage<MessageName::WebPage_URLSchemeTaskDidReceiveData>(globalObject, decoder);
    case MessageName::WebPage_URLSchemeTaskDidComplete:
        return jsValueForDecodedMessage<MessageName::WebPage_URLSchemeTaskDidComplete>(globalObject, decoder);
    case MessageName::WebPage_SetIsSuspended:
        return jsValueForDecodedMessage<MessageName::WebPage_SetIsSuspended>(globalObject, decoder);
#if ENABLE(ATTACHMENT_ELEMENT)
    case MessageName::WebPage_InsertAttachment:
        return jsValueForDecodedMessage<MessageName::WebPage_InsertAttachment>(globalObject, decoder);
    case MessageName::WebPage_UpdateAttachmentAttributes:
        return jsValueForDecodedMessage<MessageName::WebPage_UpdateAttachmentAttributes>(globalObject, decoder);
    case MessageName::WebPage_UpdateAttachmentThumbnail:
        return jsValueForDecodedMessage<MessageName::WebPage_UpdateAttachmentThumbnail>(globalObject, decoder);
    case MessageName::WebPage_UpdateAttachmentIcon:
        return jsValueForDecodedMessage<MessageName::WebPage_UpdateAttachmentIcon>(globalObject, decoder);
#endif
#if ENABLE(APPLICATION_MANIFEST)
    case MessageName::WebPage_GetApplicationManifest:
        return jsValueForDecodedMessage<MessageName::WebPage_GetApplicationManifest>(globalObject, decoder);
#endif
    case MessageName::WebPage_GetTextFragmentMatch:
        return jsValueForDecodedMessage<MessageName::WebPage_GetTextFragmentMatch>(globalObject, decoder);
    case MessageName::WebPage_SetDefersLoading:
        return jsValueForDecodedMessage<MessageName::WebPage_SetDefersLoading>(globalObject, decoder);
#if ENABLE(UI_PROCESS_PDF_HUD)
    case MessageName::WebPage_ZoomPDFIn:
        return jsValueForDecodedMessage<MessageName::WebPage_ZoomPDFIn>(globalObject, decoder);
    case MessageName::WebPage_ZoomPDFOut:
        return jsValueForDecodedMessage<MessageName::WebPage_ZoomPDFOut>(globalObject, decoder);
    case MessageName::WebPage_SavePDF:
        return jsValueForDecodedMessage<MessageName::WebPage_SavePDF>(globalObject, decoder);
    case MessageName::WebPage_OpenPDFWithPreview:
        return jsValueForDecodedMessage<MessageName::WebPage_OpenPDFWithPreview>(globalObject, decoder);
#endif
    case MessageName::WebPage_UpdateCurrentModifierState:
        return jsValueForDecodedMessage<MessageName::WebPage_UpdateCurrentModifierState>(globalObject, decoder);
    case MessageName::WebPage_SimulateDeviceOrientationChange:
        return jsValueForDecodedMessage<MessageName::WebPage_SimulateDeviceOrientationChange>(globalObject, decoder);
#if ENABLE(SPEECH_SYNTHESIS)
    case MessageName::WebPage_SpeakingErrorOccurred:
        return jsValueForDecodedMessage<MessageName::WebPage_SpeakingErrorOccurred>(globalObject, decoder);
    case MessageName::WebPage_BoundaryEventOccurred:
        return jsValueForDecodedMessage<MessageName::WebPage_BoundaryEventOccurred>(globalObject, decoder);
    case MessageName::WebPage_VoicesDidChange:
        return jsValueForDecodedMessage<MessageName::WebPage_VoicesDidChange>(globalObject, decoder);
#endif
    case MessageName::WebPage_SetCanShowPlaceholder:
        return jsValueForDecodedMessage<MessageName::WebPage_SetCanShowPlaceholder>(globalObject, decoder);
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::WebPage_WasLoadedWithDataTransferFromPrevalentResource:
        return jsValueForDecodedMessage<MessageName::WebPage_WasLoadedWithDataTransferFromPrevalentResource>(globalObject, decoder);
    case MessageName::WebPage_ClearLoadedSubresourceDomains:
        return jsValueForDecodedMessage<MessageName::WebPage_ClearLoadedSubresourceDomains>(globalObject, decoder);
    case MessageName::WebPage_GetLoadedSubresourceDomains:
        return jsValueForDecodedMessage<MessageName::WebPage_GetLoadedSubresourceDomains>(globalObject, decoder);
#endif
#if USE(SYSTEM_PREVIEW)
    case MessageName::WebPage_SystemPreviewActionTriggered:
        return jsValueForDecodedMessage<MessageName::WebPage_SystemPreviewActionTriggered>(globalObject, decoder);
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    case MessageName::WebPage_SendMessageToWebProcessExtension:
        return jsValueForDecodedMessage<MessageName::WebPage_SendMessageToWebProcessExtension>(globalObject, decoder);
    case MessageName::WebPage_SendMessageToWebProcessExtensionWithReply:
        return jsValueForDecodedMessage<MessageName::WebPage_SendMessageToWebProcessExtensionWithReply>(globalObject, decoder);
#endif
    case MessageName::WebPage_StartTextManipulations:
        return jsValueForDecodedMessage<MessageName::WebPage_StartTextManipulations>(globalObject, decoder);
    case MessageName::WebPage_CompleteTextManipulation:
        return jsValueForDecodedMessage<MessageName::WebPage_CompleteTextManipulation>(globalObject, decoder);
    case MessageName::WebPage_SetOverriddenMediaType:
        return jsValueForDecodedMessage<MessageName::WebPage_SetOverriddenMediaType>(globalObject, decoder);
    case MessageName::WebPage_GetProcessDisplayName:
        return jsValueForDecodedMessage<MessageName::WebPage_GetProcessDisplayName>(globalObject, decoder);
    case MessageName::WebPage_UpdateCORSDisablingPatterns:
        return jsValueForDecodedMessage<MessageName::WebPage_UpdateCORSDisablingPatterns>(globalObject, decoder);
    case MessageName::WebPage_SetIsTakingSnapshotsForApplicationSuspension:
        return jsValueForDecodedMessage<MessageName::WebPage_SetIsTakingSnapshotsForApplicationSuspension>(globalObject, decoder);
    case MessageName::WebPage_SetNeedsDOMWindowResizeEvent:
        return jsValueForDecodedMessage<MessageName::WebPage_SetNeedsDOMWindowResizeEvent>(globalObject, decoder);
    case MessageName::WebPage_SetHasResourceLoadClient:
        return jsValueForDecodedMessage<MessageName::WebPage_SetHasResourceLoadClient>(globalObject, decoder);
    case MessageName::WebPage_SetCanUseCredentialStorage:
        return jsValueForDecodedMessage<MessageName::WebPage_SetCanUseCredentialStorage>(globalObject, decoder);
#if ENABLE(APP_HIGHLIGHTS)
    case MessageName::WebPage_CreateAppHighlightInSelectedRange:
        return jsValueForDecodedMessage<MessageName::WebPage_CreateAppHighlightInSelectedRange>(globalObject, decoder);
    case MessageName::WebPage_RestoreAppHighlightsAndScrollToIndex:
        return jsValueForDecodedMessage<MessageName::WebPage_RestoreAppHighlightsAndScrollToIndex>(globalObject, decoder);
    case MessageName::WebPage_SetAppHighlightsVisibility:
        return jsValueForDecodedMessage<MessageName::WebPage_SetAppHighlightsVisibility>(globalObject, decoder);
#endif
    case MessageName::WebPage_HandleWheelEvent:
        return jsValueForDecodedMessage<MessageName::WebPage_HandleWheelEvent>(globalObject, decoder);
    case MessageName::WebPage_DispatchWheelEventWithoutScrolling:
        return jsValueForDecodedMessage<MessageName::WebPage_DispatchWheelEventWithoutScrolling>(globalObject, decoder);
    case MessageName::WebPage_LastNavigationWasAppInitiated:
        return jsValueForDecodedMessage<MessageName::WebPage_LastNavigationWasAppInitiated>(globalObject, decoder);
#if ENABLE(MEDIA_SESSION_COORDINATOR)
    case MessageName::WebPage_CreateMediaSessionCoordinator:
        return jsValueForDecodedMessage<MessageName::WebPage_CreateMediaSessionCoordinator>(globalObject, decoder);
#endif
#if ENABLE(IMAGE_ANALYSIS)
    case MessageName::WebPage_UpdateWithTextRecognitionResult:
        return jsValueForDecodedMessage<MessageName::WebPage_UpdateWithTextRecognitionResult>(globalObject, decoder);
    case MessageName::WebPage_StartVisualTranslation:
        return jsValueForDecodedMessage<MessageName::WebPage_StartVisualTranslation>(globalObject, decoder);
#endif
    case MessageName::WebPage_ScrollToRect:
        return jsValueForDecodedMessage<MessageName::WebPage_ScrollToRect>(globalObject, decoder);
    case MessageName::WebPage_NavigateServiceWorkerClient:
        return jsValueForDecodedMessage<MessageName::WebPage_NavigateServiceWorkerClient>(globalObject, decoder);
#if ENABLE(ARKIT_INLINE_PREVIEW_IOS)
    case MessageName::WebPage_ModelInlinePreviewDidLoad:
        return jsValueForDecodedMessage<MessageName::WebPage_ModelInlinePreviewDidLoad>(globalObject, decoder);
    case MessageName::WebPage_ModelInlinePreviewDidFailToLoad:
        return jsValueForDecodedMessage<MessageName::WebPage_ModelInlinePreviewDidFailToLoad>(globalObject, decoder);
#endif
#if HAVE(UIKIT_RESIZABLE_WINDOWS)
    case MessageName::WebPage_SetIsWindowResizingEnabled:
        return jsValueForDecodedMessage<MessageName::WebPage_SetIsWindowResizingEnabled>(globalObject, decoder);
#endif
#if ENABLE(INTERACTION_REGIONS_IN_EVENT_REGION)
    case MessageName::WebPage_SetInteractionRegionsEnabled:
        return jsValueForDecodedMessage<MessageName::WebPage_SetInteractionRegionsEnabled>(globalObject, decoder);
#endif
    case MessageName::WebPage_GenerateTestReport:
        return jsValueForDecodedMessage<MessageName::WebPage_GenerateTestReport>(globalObject, decoder);
#if ENABLE(NETWORK_CONNECTION_INTEGRITY)
    case MessageName::WebPage_SetLookalikeCharacterStrings:
        return jsValueForDecodedMessage<MessageName::WebPage_SetLookalikeCharacterStrings>(globalObject, decoder);
    case MessageName::WebPage_SetAllowedLookalikeCharacterStrings:
        return jsValueForDecodedMessage<MessageName::WebPage_SetAllowedLookalikeCharacterStrings>(globalObject, decoder);
#endif
#if ENABLE(ACCESSIBILITY_ANIMATION_CONTROL)
    case MessageName::WebPage_PauseAllAnimations:
        return jsValueForDecodedMessage<MessageName::WebPage_PauseAllAnimations>(globalObject, decoder);
    case MessageName::WebPage_PlayAllAnimations:
        return jsValueForDecodedMessage<MessageName::WebPage_PlayAllAnimations>(globalObject, decoder);
#endif
    case MessageName::StorageAreaMap_DispatchStorageEvent:
        return jsValueForDecodedMessage<MessageName::StorageAreaMap_DispatchStorageEvent>(globalObject, decoder);
    case MessageName::StorageAreaMap_ClearCache:
        return jsValueForDecodedMessage<MessageName::StorageAreaMap_ClearCache>(globalObject, decoder);
#if ENABLE(WEBXR)
    case MessageName::PlatformXRSystemProxy_SessionDidEnd:
        return jsValueForDecodedMessage<MessageName::PlatformXRSystemProxy_SessionDidEnd>(globalObject, decoder);
    case MessageName::PlatformXRSystemProxy_SessionDidUpdateVisibilityState:
        return jsValueForDecodedMessage<MessageName::PlatformXRSystemProxy_SessionDidUpdateVisibilityState>(globalObject, decoder);
#endif
#if PLATFORM(MAC)
    case MessageName::ViewGestureController_DidCollectGeometryForSmartMagnificationGesture:
        return jsValueForDecodedMessage<MessageName::ViewGestureController_DidCollectGeometryForSmartMagnificationGesture>(globalObject, decoder);
#endif
#if !PLATFORM(IOS_FAMILY)
    case MessageName::ViewGestureController_DidCollectGeometryForMagnificationGesture:
        return jsValueForDecodedMessage<MessageName::ViewGestureController_DidCollectGeometryForMagnificationGesture>(globalObject, decoder);
    case MessageName::ViewGestureController_DidHitRenderTreeSizeThreshold:
        return jsValueForDecodedMessage<MessageName::ViewGestureController_DidHitRenderTreeSizeThreshold>(globalObject, decoder);
#endif
    case MessageName::GtkSettingsManagerProxy_SettingsDidChange:
        return jsValueForDecodedMessage<MessageName::GtkSettingsManagerProxy_SettingsDidChange>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::ViewGestureGeometryCollector_CollectGeometryForSmartMagnificationGesture:
        return jsValueForDecodedMessage<MessageName::ViewGestureGeometryCollector_CollectGeometryForSmartMagnificationGesture>(globalObject, decoder);
#endif
#if !PLATFORM(IOS_FAMILY)
    case MessageName::ViewGestureGeometryCollector_CollectGeometryForMagnificationGesture:
        return jsValueForDecodedMessage<MessageName::ViewGestureGeometryCollector_CollectGeometryForMagnificationGesture>(globalObject, decoder);
    case MessageName::ViewGestureGeometryCollector_SetRenderTreeSizeNotificationThreshold:
        return jsValueForDecodedMessage<MessageName::ViewGestureGeometryCollector_SetRenderTreeSizeNotificationThreshold>(globalObject, decoder);
#endif
#if ENABLE(MEDIA_STREAM)
    case MessageName::UserMediaCaptureManager_ValidateUserMediaRequestConstraints:
        return jsValueForDecodedMessage<MessageName::UserMediaCaptureManager_ValidateUserMediaRequestConstraints>(globalObject, decoder);
    case MessageName::UserMediaCaptureManager_GetMediaStreamDevices:
        return jsValueForDecodedMessage<MessageName::UserMediaCaptureManager_GetMediaStreamDevices>(globalObject, decoder);
#endif
    default:
        break;
    }
    return std::nullopt;
}

std::optional<JSC::JSValue> jsValueForReplyArguments(JSC::JSGlobalObject* globalObject, MessageName name, Decoder& decoder)
{
    switch (name) {
#if ENABLE(GPU_PROCESS)
#if USE(AUDIO_SESSION)
    case MessageName::GPUConnectionToWebProcess_EnsureAudioSession:
        return jsValueForDecodedMessageReply<MessageName::GPUConnectionToWebProcess_EnsureAudioSession>(globalObject, decoder);
#endif
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::GPUProcess_CreateGPUConnectionToWebProcess:
        return jsValueForDecodedMessageReply<MessageName::GPUProcess_CreateGPUConnectionToWebProcess>(globalObject, decoder);
    case MessageName::GPUProcess_PrepareToSuspend:
        return jsValueForDecodedMessageReply<MessageName::GPUProcess_PrepareToSuspend>(globalObject, decoder);
#if ENABLE(MEDIA_STREAM)
    case MessageName::GPUProcess_UpdateCaptureAccess:
        return jsValueForDecodedMessageReply<MessageName::GPUProcess_UpdateCaptureAccess>(globalObject, decoder);
#endif
#if HAVE(SC_CONTENT_SHARING_SESSION)
    case MessageName::GPUProcess_ShowWindowPicker:
        return jsValueForDecodedMessageReply<MessageName::GPUProcess_ShowWindowPicker>(globalObject, decoder);
    case MessageName::GPUProcess_ShowScreenPicker:
        return jsValueForDecodedMessageReply<MessageName::GPUProcess_ShowScreenPicker>(globalObject, decoder);
#endif
#if ENABLE(VIDEO)
    case MessageName::GPUProcess_RequestBitmapImageForCurrentTime:
        return jsValueForDecodedMessageReply<MessageName::GPUProcess_RequestBitmapImageForCurrentTime>(globalObject, decoder);
#endif
    case MessageName::GPUProcess_WebProcessConnectionCountForTesting:
        return jsValueForDecodedMessageReply<MessageName::GPUProcess_WebProcessConnectionCountForTesting>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(WEBGL)
#if PLATFORM(COCOA)
    case MessageName::RemoteGraphicsContextGL_PrepareForDisplay:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_PrepareForDisplay>(globalObject, decoder);
#endif
#if USE(GRAPHICS_LAYER_WC)
    case MessageName::RemoteGraphicsContextGL_PrepareForDisplay:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_PrepareForDisplay>(globalObject, decoder);
#endif
#if USE(GBM)
    case MessageName::RemoteGraphicsContextGL_PrepareForDisplay:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_PrepareForDisplay>(globalObject, decoder);
#endif
#if !PLATFORM(COCOA) && !USE(GRAPHICS_LAYER_WC) && !USE(GBM)
    case MessageName::RemoteGraphicsContextGL_PrepareForDisplay:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_PrepareForDisplay>(globalObject, decoder);
#endif
    case MessageName::RemoteGraphicsContextGL_GetError:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetError>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_PaintRenderingResultsToCanvas:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_PaintRenderingResultsToCanvas>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_PaintCompositedResultsToCanvas:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_PaintCompositedResultsToCanvas>(globalObject, decoder);
#if ENABLE(VIDEO) && PLATFORM(COCOA)
    case MessageName::RemoteGraphicsContextGL_CopyTextureFromVideoFrame:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_CopyTextureFromVideoFrame>(globalObject, decoder);
#endif
#if ENABLE(MEDIA_STREAM)
    case MessageName::RemoteGraphicsContextGL_PaintCompositedResultsToVideoFrame:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_PaintCompositedResultsToVideoFrame>(globalObject, decoder);
#endif
    case MessageName::RemoteGraphicsContextGL_ReadnPixels0:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_ReadnPixels0>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_ReadnPixels2:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_ReadnPixels2>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_MoveErrorsToSyntheticErrorList:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_MoveErrorsToSyntheticErrorList>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CheckFramebufferStatus:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_CheckFramebufferStatus>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateBuffer:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_CreateBuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateFramebuffer:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_CreateFramebuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateProgram:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_CreateProgram>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateRenderbuffer:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_CreateRenderbuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateShader:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_CreateShader>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateTexture:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_CreateTexture>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetActiveAttrib:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetActiveAttrib>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetActiveUniform:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetActiveUniform>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetAttribLocation:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetAttribLocation>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetBufferParameteri:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetBufferParameteri>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetString:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetString>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetFloatv:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetFloatv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetIntegerv:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetIntegerv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetIntegeri_v:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetIntegeri_v>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetInteger64:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetInteger64>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetInteger64i:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetInteger64i>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetProgrami:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetProgrami>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetBooleanv:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetBooleanv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetFramebufferAttachmentParameteri:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetFramebufferAttachmentParameteri>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetProgramInfoLog:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetProgramInfoLog>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetRenderbufferParameteri:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetRenderbufferParameteri>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetShaderi:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetShaderi>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetShaderInfoLog:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetShaderInfoLog>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetShaderPrecisionFormat:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetShaderPrecisionFormat>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetShaderSource:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetShaderSource>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetTexParameterf:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetTexParameterf>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetTexParameteri:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetTexParameteri>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetUniformfv:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetUniformfv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetUniformiv:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetUniformiv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetUniformuiv:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetUniformuiv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetUniformLocation:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetUniformLocation>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetVertexAttribOffset:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetVertexAttribOffset>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsBuffer:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_IsBuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsEnabled:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_IsEnabled>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsFramebuffer:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_IsFramebuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsProgram:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_IsProgram>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsRenderbuffer:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_IsRenderbuffer>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsShader:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_IsShader>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsTexture:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_IsTexture>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateVertexArray:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_CreateVertexArray>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsVertexArray:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_IsVertexArray>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetBufferSubData:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetBufferSubData>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetFragDataLocation:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetFragDataLocation>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateQuery:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_CreateQuery>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsQuery:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_IsQuery>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetQuery:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetQuery>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetQueryObjectui:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetQueryObjectui>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateSampler:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_CreateSampler>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsSampler:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_IsSampler>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetSamplerParameterf:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetSamplerParameterf>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetSamplerParameteri:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetSamplerParameteri>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_FenceSync:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_FenceSync>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsSync:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_IsSync>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_ClientWaitSync:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_ClientWaitSync>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetSynci:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetSynci>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateTransformFeedback:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_CreateTransformFeedback>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsTransformFeedback:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_IsTransformFeedback>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetTransformFeedbackVarying:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetTransformFeedbackVarying>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetUniformIndices:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetUniformIndices>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetActiveUniforms:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetActiveUniforms>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetUniformBlockIndex:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetUniformBlockIndex>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetActiveUniformBlockName:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetActiveUniformBlockName>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetActiveUniformBlockiv:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetActiveUniformBlockiv>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetTranslatedShaderSourceANGLE:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetTranslatedShaderSourceANGLE>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_CreateQueryEXT:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_CreateQueryEXT>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_IsQueryEXT:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_IsQueryEXT>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetQueryiEXT:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetQueryiEXT>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetQueryObjectiEXT:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetQueryObjectiEXT>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetQueryObjectui64EXT:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetQueryObjectui64EXT>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetInteger64EXT:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetInteger64EXT>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_GetInternalformativ:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_GetInternalformativ>(globalObject, decoder);
    case MessageName::RemoteGraphicsContextGL_PaintRenderingResultsToPixelBuffer:
        return jsValueForDecodedMessageReply<MessageName::RemoteGraphicsContextGL_PaintRenderingResultsToPixelBuffer>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteRenderingBackend_GetPixelBufferForImageBuffer:
        return jsValueForDecodedMessageReply<MessageName::RemoteRenderingBackend_GetPixelBufferForImageBuffer>(globalObject, decoder);
    case MessageName::RemoteRenderingBackend_GetPixelBufferForImageBufferWithNewMemory:
        return jsValueForDecodedMessageReply<MessageName::RemoteRenderingBackend_GetPixelBufferForImageBufferWithNewMemory>(globalObject, decoder);
    case MessageName::RemoteRenderingBackend_GetShareableBitmapForImageBuffer:
        return jsValueForDecodedMessageReply<MessageName::RemoteRenderingBackend_GetShareableBitmapForImageBuffer>(globalObject, decoder);
    case MessageName::RemoteRenderingBackend_GetFilteredImageForImageBuffer:
        return jsValueForDecodedMessageReply<MessageName::RemoteRenderingBackend_GetFilteredImageForImageBuffer>(globalObject, decoder);
    case MessageName::RemoteRenderingBackend_PrepareBuffersForDisplay:
        return jsValueForDecodedMessageReply<MessageName::RemoteRenderingBackend_PrepareBuffersForDisplay>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteAdapter_RequestDevice:
        return jsValueForDecodedMessageReply<MessageName::RemoteAdapter_RequestDevice>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteBuffer_GetMappedRange:
        return jsValueForDecodedMessageReply<MessageName::RemoteBuffer_GetMappedRange>(globalObject, decoder);
    case MessageName::RemoteBuffer_MapAsync:
        return jsValueForDecodedMessageReply<MessageName::RemoteBuffer_MapAsync>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#if PLATFORM(COCOA)
    case MessageName::RemoteCompositorIntegration_RecreateRenderBuffers:
        return jsValueForDecodedMessageReply<MessageName::RemoteCompositorIntegration_RecreateRenderBuffers>(globalObject, decoder);
#endif
    case MessageName::RemoteCompositorIntegration_PrepareForDisplay:
        return jsValueForDecodedMessageReply<MessageName::RemoteCompositorIntegration_PrepareForDisplay>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteDevice_CreateComputePipelineAsync:
        return jsValueForDecodedMessageReply<MessageName::RemoteDevice_CreateComputePipelineAsync>(globalObject, decoder);
    case MessageName::RemoteDevice_CreateRenderPipelineAsync:
        return jsValueForDecodedMessageReply<MessageName::RemoteDevice_CreateRenderPipelineAsync>(globalObject, decoder);
    case MessageName::RemoteDevice_PopErrorScope:
        return jsValueForDecodedMessageReply<MessageName::RemoteDevice_PopErrorScope>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteGPU_RequestAdapter:
        return jsValueForDecodedMessageReply<MessageName::RemoteGPU_RequestAdapter>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteQueue_OnSubmittedWorkDone:
        return jsValueForDecodedMessageReply<MessageName::RemoteQueue_OnSubmittedWorkDone>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteShaderModule_CompilationInfo:
        return jsValueForDecodedMessageReply<MessageName::RemoteShaderModule_CompilationInfo>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(WEB_AUDIO)
    case MessageName::RemoteAudioDestinationManager_StartAudioDestination:
        return jsValueForDecodedMessageReply<MessageName::RemoteAudioDestinationManager_StartAudioDestination>(globalObject, decoder);
    case MessageName::RemoteAudioDestinationManager_StopAudioDestination:
        return jsValueForDecodedMessageReply<MessageName::RemoteAudioDestinationManager_StopAudioDestination>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && USE(AUDIO_SESSION)
    case MessageName::RemoteAudioSessionProxy_TryToSetActive:
        return jsValueForDecodedMessageReply<MessageName::RemoteAudioSessionProxy_TryToSetActive>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteCDMFactoryProxy_CreateCDM:
        return jsValueForDecodedMessageReply<MessageName::RemoteCDMFactoryProxy_CreateCDM>(globalObject, decoder);
    case MessageName::RemoteCDMFactoryProxy_SupportsKeySystem:
        return jsValueForDecodedMessageReply<MessageName::RemoteCDMFactoryProxy_SupportsKeySystem>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteCDMInstanceProxy_CreateSession:
        return jsValueForDecodedMessageReply<MessageName::RemoteCDMInstanceProxy_CreateSession>(globalObject, decoder);
    case MessageName::RemoteCDMInstanceProxy_InitializeWithConfiguration:
        return jsValueForDecodedMessageReply<MessageName::RemoteCDMInstanceProxy_InitializeWithConfiguration>(globalObject, decoder);
    case MessageName::RemoteCDMInstanceProxy_SetServerCertificate:
        return jsValueForDecodedMessageReply<MessageName::RemoteCDMInstanceProxy_SetServerCertificate>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteCDMInstanceSessionProxy_RequestLicense:
        return jsValueForDecodedMessageReply<MessageName::RemoteCDMInstanceSessionProxy_RequestLicense>(globalObject, decoder);
    case MessageName::RemoteCDMInstanceSessionProxy_UpdateLicense:
        return jsValueForDecodedMessageReply<MessageName::RemoteCDMInstanceSessionProxy_UpdateLicense>(globalObject, decoder);
    case MessageName::RemoteCDMInstanceSessionProxy_LoadSession:
        return jsValueForDecodedMessageReply<MessageName::RemoteCDMInstanceSessionProxy_LoadSession>(globalObject, decoder);
    case MessageName::RemoteCDMInstanceSessionProxy_CloseSession:
        return jsValueForDecodedMessageReply<MessageName::RemoteCDMInstanceSessionProxy_CloseSession>(globalObject, decoder);
    case MessageName::RemoteCDMInstanceSessionProxy_RemoveSessionData:
        return jsValueForDecodedMessageReply<MessageName::RemoteCDMInstanceSessionProxy_RemoveSessionData>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteCDMProxy_GetSupportedConfiguration:
        return jsValueForDecodedMessageReply<MessageName::RemoteCDMProxy_GetSupportedConfiguration>(globalObject, decoder);
    case MessageName::RemoteCDMProxy_CreateInstance:
        return jsValueForDecodedMessageReply<MessageName::RemoteCDMProxy_CreateInstance>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
    case MessageName::RemoteLegacyCDMFactoryProxy_CreateCDM:
        return jsValueForDecodedMessageReply<MessageName::RemoteLegacyCDMFactoryProxy_CreateCDM>(globalObject, decoder);
    case MessageName::RemoteLegacyCDMFactoryProxy_SupportsKeySystem:
        return jsValueForDecodedMessageReply<MessageName::RemoteLegacyCDMFactoryProxy_SupportsKeySystem>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
    case MessageName::RemoteLegacyCDMProxy_SupportsMIMEType:
        return jsValueForDecodedMessageReply<MessageName::RemoteLegacyCDMProxy_SupportsMIMEType>(globalObject, decoder);
    case MessageName::RemoteLegacyCDMProxy_CreateSession:
        return jsValueForDecodedMessageReply<MessageName::RemoteLegacyCDMProxy_CreateSession>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
    case MessageName::RemoteLegacyCDMSessionProxy_GenerateKeyRequest:
        return jsValueForDecodedMessageReply<MessageName::RemoteLegacyCDMSessionProxy_GenerateKeyRequest>(globalObject, decoder);
    case MessageName::RemoteLegacyCDMSessionProxy_Update:
        return jsValueForDecodedMessageReply<MessageName::RemoteLegacyCDMSessionProxy_Update>(globalObject, decoder);
    case MessageName::RemoteLegacyCDMSessionProxy_CachedKeyForKeyID:
        return jsValueForDecodedMessageReply<MessageName::RemoteLegacyCDMSessionProxy_CachedKeyForKeyID>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteMediaEngineConfigurationFactoryProxy_CreateDecodingConfiguration:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaEngineConfigurationFactoryProxy_CreateDecodingConfiguration>(globalObject, decoder);
    case MessageName::RemoteMediaEngineConfigurationFactoryProxy_CreateEncodingConfiguration:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaEngineConfigurationFactoryProxy_CreateEncodingConfiguration>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
    case MessageName::RemoteMediaPlayerManagerProxy_GetSupportedTypes:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaPlayerManagerProxy_GetSupportedTypes>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerManagerProxy_SupportsTypeAndCodecs:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaPlayerManagerProxy_SupportsTypeAndCodecs>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerManagerProxy_OriginsInMediaCache:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaPlayerManagerProxy_OriginsInMediaCache>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerManagerProxy_SupportsKeySystem:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaPlayerManagerProxy_SupportsKeySystem>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
    case MessageName::RemoteMediaPlayerProxy_Load:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaPlayerProxy_Load>(globalObject, decoder);
#if ENABLE(MEDIA_SOURCE)
    case MessageName::RemoteMediaPlayerProxy_LoadMediaSource:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaPlayerProxy_LoadMediaSource>(globalObject, decoder);
#endif
    case MessageName::RemoteMediaPlayerProxy_DidLoadingProgress:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaPlayerProxy_DidLoadingProgress>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_PerformTaskAtMediaTime:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaPlayerProxy_PerformTaskAtMediaTime>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_IsCrossOrigin:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaPlayerProxy_IsCrossOrigin>(globalObject, decoder);
#if PLATFORM(IOS_FAMILY)
    case MessageName::RemoteMediaPlayerProxy_ErrorLog:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaPlayerProxy_ErrorLog>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_AccessLog:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaPlayerProxy_AccessLog>(globalObject, decoder);
#endif
#if PLATFORM(COCOA)
    case MessageName::RemoteMediaPlayerProxy_NativeImageForCurrentTime:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaPlayerProxy_NativeImageForCurrentTime>(globalObject, decoder);
    case MessageName::RemoteMediaPlayerProxy_ColorSpace:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaPlayerProxy_ColorSpace>(globalObject, decoder);
#endif
    case MessageName::RemoteMediaPlayerProxy_VideoFrameForCurrentTimeIfChanged:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaPlayerProxy_VideoFrameForCurrentTimeIfChanged>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
    case MessageName::RemoteMediaResourceManager_ResponseReceived:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaResourceManager_ResponseReceived>(globalObject, decoder);
    case MessageName::RemoteMediaResourceManager_RedirectReceived:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaResourceManager_RedirectReceived>(globalObject, decoder);
    case MessageName::RemoteMediaResourceManager_DataReceived:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaResourceManager_DataReceived>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
    case MessageName::RemoteMediaSourceProxy_AddSourceBuffer:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaSourceProxy_AddSourceBuffer>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
    case MessageName::RemoteSourceBufferProxy_CanSwitchToType:
        return jsValueForDecodedMessageReply<MessageName::RemoteSourceBufferProxy_CanSwitchToType>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_Append:
        return jsValueForDecodedMessageReply<MessageName::RemoteSourceBufferProxy_Append>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_UpdateBufferedFromTrackBuffers:
        return jsValueForDecodedMessageReply<MessageName::RemoteSourceBufferProxy_UpdateBufferedFromTrackBuffers>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_RemoveCodedFrames:
        return jsValueForDecodedMessageReply<MessageName::RemoteSourceBufferProxy_RemoveCodedFrames>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_EvictCodedFrames:
        return jsValueForDecodedMessageReply<MessageName::RemoteSourceBufferProxy_EvictCodedFrames>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_BufferedSamplesForTrackId:
        return jsValueForDecodedMessageReply<MessageName::RemoteSourceBufferProxy_BufferedSamplesForTrackId>(globalObject, decoder);
    case MessageName::RemoteSourceBufferProxy_EnqueuedSamplesForTrackID:
        return jsValueForDecodedMessageReply<MessageName::RemoteSourceBufferProxy_EnqueuedSamplesForTrackID>(globalObject, decoder);
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
#if PLATFORM(COCOA)
    case MessageName::RemoteVideoFrameObjectHeap_PixelBuffer:
        return jsValueForDecodedMessageReply<MessageName::RemoteVideoFrameObjectHeap_PixelBuffer>(globalObject, decoder);
    case MessageName::RemoteVideoFrameObjectHeap_ConvertFrameBuffer:
        return jsValueForDecodedMessageReply<MessageName::RemoteVideoFrameObjectHeap_ConvertFrameBuffer>(globalObject, decoder);
#endif
#endif
#if USE(LIBWEBRTC) && PLATFORM(COCOA) && ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM) && PLATFORM(COCOA)
    case MessageName::RemoteAudioMediaStreamTrackRendererInternalUnitManager_CreateUnit:
        return jsValueForDecodedMessageReply<MessageName::RemoteAudioMediaStreamTrackRendererInternalUnitManager_CreateUnit>(globalObject, decoder);
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
    case MessageName::RemoteMediaRecorder_FetchData:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaRecorder_FetchData>(globalObject, decoder);
    case MessageName::RemoteMediaRecorder_StopRecording:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaRecorder_StopRecording>(globalObject, decoder);
    case MessageName::RemoteMediaRecorder_Pause:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaRecorder_Pause>(globalObject, decoder);
    case MessageName::RemoteMediaRecorder_Resume:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaRecorder_Resume>(globalObject, decoder);
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
    case MessageName::RemoteMediaRecorderManager_CreateRecorder:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaRecorderManager_CreateRecorder>(globalObject, decoder);
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
    case MessageName::RemoteSampleBufferDisplayLayerManager_CreateLayer:
        return jsValueForDecodedMessageReply<MessageName::RemoteSampleBufferDisplayLayerManager_CreateLayer>(globalObject, decoder);
#endif
    case MessageName::WebCookieManager_GetHostnamesWithCookies:
        return jsValueForDecodedMessageReply<MessageName::WebCookieManager_GetHostnamesWithCookies>(globalObject, decoder);
    case MessageName::WebCookieManager_DeleteCookiesForHostnames:
        return jsValueForDecodedMessageReply<MessageName::WebCookieManager_DeleteCookiesForHostnames>(globalObject, decoder);
    case MessageName::WebCookieManager_DeleteAllCookies:
        return jsValueForDecodedMessageReply<MessageName::WebCookieManager_DeleteAllCookies>(globalObject, decoder);
    case MessageName::WebCookieManager_SetCookie:
        return jsValueForDecodedMessageReply<MessageName::WebCookieManager_SetCookie>(globalObject, decoder);
    case MessageName::WebCookieManager_SetCookies:
        return jsValueForDecodedMessageReply<MessageName::WebCookieManager_SetCookies>(globalObject, decoder);
    case MessageName::WebCookieManager_GetAllCookies:
        return jsValueForDecodedMessageReply<MessageName::WebCookieManager_GetAllCookies>(globalObject, decoder);
    case MessageName::WebCookieManager_GetCookies:
        return jsValueForDecodedMessageReply<MessageName::WebCookieManager_GetCookies>(globalObject, decoder);
    case MessageName::WebCookieManager_DeleteCookie:
        return jsValueForDecodedMessageReply<MessageName::WebCookieManager_DeleteCookie>(globalObject, decoder);
    case MessageName::WebCookieManager_DeleteAllCookiesModifiedSince:
        return jsValueForDecodedMessageReply<MessageName::WebCookieManager_DeleteAllCookiesModifiedSince>(globalObject, decoder);
    case MessageName::WebCookieManager_SetHTTPCookieAcceptPolicy:
        return jsValueForDecodedMessageReply<MessageName::WebCookieManager_SetHTTPCookieAcceptPolicy>(globalObject, decoder);
    case MessageName::WebCookieManager_GetHTTPCookieAcceptPolicy:
        return jsValueForDecodedMessageReply<MessageName::WebCookieManager_GetHTTPCookieAcceptPolicy>(globalObject, decoder);
    case MessageName::NetworkBroadcastChannelRegistry_PostMessage:
        return jsValueForDecodedMessageReply<MessageName::NetworkBroadcastChannelRegistry_PostMessage>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_PerformSynchronousLoad:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_PerformSynchronousLoad>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_TestProcessIncomingSyncMessagesWhenWaitingForSyncReply:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_TestProcessIncomingSyncMessagesWhenWaitingForSyncReply>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_SendH2Ping:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_SendH2Ping>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_IsResourceLoadFinished:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_IsResourceLoadFinished>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_CookiesForDOM:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_CookiesForDOM>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_CookieRequestHeaderFieldValue:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_CookieRequestHeaderFieldValue>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_GetRawCookies:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_GetRawCookies>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_DeleteCookie:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_DeleteCookie>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_DomCookiesForHost:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_DomCookiesForHost>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_BlobSize:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_BlobSize>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_WriteBlobsToTemporaryFilesForIndexedDB:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_WriteBlobsToTemporaryFilesForIndexedDB>(globalObject, decoder);
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::NetworkConnectionToWebProcess_ResourceLoadStatisticsUpdated:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_ResourceLoadStatisticsUpdated>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_HasStorageAccess:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_HasStorageAccess>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_RequestStorageAccess:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_RequestStorageAccess>(globalObject, decoder);
#endif
    case MessageName::NetworkConnectionToWebProcess_GetNetworkLoadInformationResponse:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_GetNetworkLoadInformationResponse>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_GetNetworkLoadIntermediateInformation:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_GetNetworkLoadIntermediateInformation>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_TakeNetworkLoadInformationMetrics:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_TakeNetworkLoadInformationMetrics>(globalObject, decoder);
#if ENABLE(SERVICE_WORKER)
    case MessageName::NetworkConnectionToWebProcess_EstablishSWContextConnection:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_EstablishSWContextConnection>(globalObject, decoder);
#endif
    case MessageName::NetworkConnectionToWebProcess_EstablishSharedWorkerContextConnection:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_EstablishSharedWorkerContextConnection>(globalObject, decoder);
    case MessageName::NetworkConnectionToWebProcess_CreateRTCProvider:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_CreateRTCProvider>(globalObject, decoder);
#if ENABLE(WEB_RTC)
    case MessageName::NetworkConnectionToWebProcess_ConnectToRTCDataChannelRemoteSource:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_ConnectToRTCDataChannelRemoteSource>(globalObject, decoder);
#endif
    case MessageName::NetworkConnectionToWebProcess_TakeAllMessagesForPort:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_TakeAllMessagesForPort>(globalObject, decoder);
#if PLATFORM(MAC)
    case MessageName::NetworkConnectionToWebProcess_GetProcessDisplayName:
        return jsValueForDecodedMessageReply<MessageName::NetworkConnectionToWebProcess_GetProcessDisplayName>(globalObject, decoder);
#endif
#if ENABLE(CONTENT_EXTENSIONS)
#endif
    case MessageName::NetworkProcess_InitializeNetworkProcess:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_InitializeNetworkProcess>(globalObject, decoder);
    case MessageName::NetworkProcess_CreateNetworkConnectionToWebProcess:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_CreateNetworkConnectionToWebProcess>(globalObject, decoder);
    case MessageName::NetworkProcess_AddAllowedFirstPartyForCookies:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_AddAllowedFirstPartyForCookies>(globalObject, decoder);
    case MessageName::NetworkProcess_FetchWebsiteData:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_FetchWebsiteData>(globalObject, decoder);
    case MessageName::NetworkProcess_DeleteWebsiteData:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_DeleteWebsiteData>(globalObject, decoder);
    case MessageName::NetworkProcess_DeleteWebsiteDataForOrigins:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_DeleteWebsiteDataForOrigins>(globalObject, decoder);
    case MessageName::NetworkProcess_RenameOriginInWebsiteData:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_RenameOriginInWebsiteData>(globalObject, decoder);
    case MessageName::NetworkProcess_WebsiteDataOriginDirectoryForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_WebsiteDataOriginDirectoryForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_CancelDownload:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_CancelDownload>(globalObject, decoder);
    case MessageName::NetworkProcess_DataTaskWithRequest:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_DataTaskWithRequest>(globalObject, decoder);
    case MessageName::NetworkProcess_SetAllowsAnySSLCertificateForWebSocket:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetAllowsAnySSLCertificateForWebSocket>(globalObject, decoder);
    case MessageName::NetworkProcess_FlushCookies:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_FlushCookies>(globalObject, decoder);
    case MessageName::NetworkProcess_SetCacheModelSynchronouslyForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetCacheModelSynchronouslyForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_ProcessWillSuspendImminentlyForTestingSync:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_ProcessWillSuspendImminentlyForTestingSync>(globalObject, decoder);
    case MessageName::NetworkProcess_PrepareToSuspend:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_PrepareToSuspend>(globalObject, decoder);
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::NetworkProcess_ClearPrevalentResource:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_ClearPrevalentResource>(globalObject, decoder);
    case MessageName::NetworkProcess_ClearUserInteraction:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_ClearUserInteraction>(globalObject, decoder);
    case MessageName::NetworkProcess_DumpResourceLoadStatistics:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_DumpResourceLoadStatistics>(globalObject, decoder);
    case MessageName::NetworkProcess_UpdatePrevalentDomainsToBlockCookiesFor:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_UpdatePrevalentDomainsToBlockCookiesFor>(globalObject, decoder);
    case MessageName::NetworkProcess_IsGrandfathered:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_IsGrandfathered>(globalObject, decoder);
    case MessageName::NetworkProcess_IsPrevalentResource:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_IsPrevalentResource>(globalObject, decoder);
    case MessageName::NetworkProcess_IsVeryPrevalentResource:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_IsVeryPrevalentResource>(globalObject, decoder);
    case MessageName::NetworkProcess_SetLastSeen:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetLastSeen>(globalObject, decoder);
    case MessageName::NetworkProcess_MergeStatisticForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_MergeStatisticForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_InsertExpiredStatisticForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_InsertExpiredStatisticForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetPrevalentResource:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetPrevalentResource>(globalObject, decoder);
    case MessageName::NetworkProcess_SetPrevalentResourceForDebugMode:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetPrevalentResourceForDebugMode>(globalObject, decoder);
    case MessageName::NetworkProcess_IsResourceLoadStatisticsEphemeral:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_IsResourceLoadStatisticsEphemeral>(globalObject, decoder);
    case MessageName::NetworkProcess_HadUserInteraction:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_HadUserInteraction>(globalObject, decoder);
    case MessageName::NetworkProcess_IsRelationshipOnlyInDatabaseOnce:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_IsRelationshipOnlyInDatabaseOnce>(globalObject, decoder);
    case MessageName::NetworkProcess_HasLocalStorage:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_HasLocalStorage>(globalObject, decoder);
    case MessageName::NetworkProcess_GetAllStorageAccessEntries:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_GetAllStorageAccessEntries>(globalObject, decoder);
    case MessageName::NetworkProcess_IsRegisteredAsRedirectingTo:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_IsRegisteredAsRedirectingTo>(globalObject, decoder);
    case MessageName::NetworkProcess_IsRegisteredAsSubFrameUnder:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_IsRegisteredAsSubFrameUnder>(globalObject, decoder);
    case MessageName::NetworkProcess_IsRegisteredAsSubresourceUnder:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_IsRegisteredAsSubresourceUnder>(globalObject, decoder);
    case MessageName::NetworkProcess_DomainIDExistsInDatabase:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_DomainIDExistsInDatabase>(globalObject, decoder);
    case MessageName::NetworkProcess_LogUserInteraction:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_LogUserInteraction>(globalObject, decoder);
    case MessageName::NetworkProcess_ResetParametersToDefaultValues:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_ResetParametersToDefaultValues>(globalObject, decoder);
    case MessageName::NetworkProcess_ScheduleClearInMemoryAndPersistent:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_ScheduleClearInMemoryAndPersistent>(globalObject, decoder);
    case MessageName::NetworkProcess_ScheduleCookieBlockingUpdate:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_ScheduleCookieBlockingUpdate>(globalObject, decoder);
    case MessageName::NetworkProcess_ScheduleStatisticsAndDataRecordsProcessing:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_ScheduleStatisticsAndDataRecordsProcessing>(globalObject, decoder);
    case MessageName::NetworkProcess_StatisticsDatabaseHasAllTables:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_StatisticsDatabaseHasAllTables>(globalObject, decoder);
    case MessageName::NetworkProcess_SetCacheMaxAgeCapForPrevalentResources:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetCacheMaxAgeCapForPrevalentResources>(globalObject, decoder);
    case MessageName::NetworkProcess_SetGrandfathered:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetGrandfathered>(globalObject, decoder);
    case MessageName::NetworkProcess_GetResourceLoadStatisticsDataSummary:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_GetResourceLoadStatisticsDataSummary>(globalObject, decoder);
    case MessageName::NetworkProcess_SetGrandfatheringTime:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetGrandfatheringTime>(globalObject, decoder);
    case MessageName::NetworkProcess_SetMaxStatisticsEntries:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetMaxStatisticsEntries>(globalObject, decoder);
    case MessageName::NetworkProcess_SetMinimumTimeBetweenDataRecordsRemoval:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetMinimumTimeBetweenDataRecordsRemoval>(globalObject, decoder);
    case MessageName::NetworkProcess_SetPruneEntriesDownTo:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetPruneEntriesDownTo>(globalObject, decoder);
    case MessageName::NetworkProcess_SetShouldClassifyResourcesBeforeDataRecordsRemoval:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetShouldClassifyResourcesBeforeDataRecordsRemoval>(globalObject, decoder);
    case MessageName::NetworkProcess_SetNotifyPagesWhenDataRecordsWereScanned:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetNotifyPagesWhenDataRecordsWereScanned>(globalObject, decoder);
    case MessageName::NetworkProcess_SetResourceLoadStatisticsTimeAdvanceForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetResourceLoadStatisticsTimeAdvanceForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetIsRunningResourceLoadStatisticsTest:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetIsRunningResourceLoadStatisticsTest>(globalObject, decoder);
    case MessageName::NetworkProcess_SetResourceLoadStatisticsDebugMode:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetResourceLoadStatisticsDebugMode>(globalObject, decoder);
    case MessageName::NetworkProcess_SetVeryPrevalentResource:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetVeryPrevalentResource>(globalObject, decoder);
    case MessageName::NetworkProcess_SetSubframeUnderTopFrameDomain:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetSubframeUnderTopFrameDomain>(globalObject, decoder);
    case MessageName::NetworkProcess_SetSubresourceUnderTopFrameDomain:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetSubresourceUnderTopFrameDomain>(globalObject, decoder);
    case MessageName::NetworkProcess_SetSubresourceUniqueRedirectTo:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetSubresourceUniqueRedirectTo>(globalObject, decoder);
    case MessageName::NetworkProcess_SetSubresourceUniqueRedirectFrom:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetSubresourceUniqueRedirectFrom>(globalObject, decoder);
    case MessageName::NetworkProcess_SetTimeToLiveUserInteraction:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetTimeToLiveUserInteraction>(globalObject, decoder);
    case MessageName::NetworkProcess_SetTopFrameUniqueRedirectTo:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetTopFrameUniqueRedirectTo>(globalObject, decoder);
    case MessageName::NetworkProcess_SetTopFrameUniqueRedirectFrom:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetTopFrameUniqueRedirectFrom>(globalObject, decoder);
    case MessageName::NetworkProcess_ResetCacheMaxAgeCapForPrevalentResources:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_ResetCacheMaxAgeCapForPrevalentResources>(globalObject, decoder);
    case MessageName::NetworkProcess_SetCrossSiteLoadWithLinkDecorationForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetCrossSiteLoadWithLinkDecorationForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_ResetCrossSiteLoadsWithLinkDecorationForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_ResetCrossSiteLoadsWithLinkDecorationForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_DeleteCookiesForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_DeleteCookiesForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_HasIsolatedSession:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_HasIsolatedSession>(globalObject, decoder);
    case MessageName::NetworkProcess_CloseITPDatabase:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_CloseITPDatabase>(globalObject, decoder);
#endif
#if (ENABLE(TRACKING_PREVENTION) && ENABLE(APP_BOUND_DOMAINS))
    case MessageName::NetworkProcess_SetAppBoundDomainsForResourceLoadStatistics:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetAppBoundDomainsForResourceLoadStatistics>(globalObject, decoder);
#endif
#if (ENABLE(TRACKING_PREVENTION) && ENABLE(MANAGED_DOMAINS))
    case MessageName::NetworkProcess_SetManagedDomainsForResourceLoadStatistics:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetManagedDomainsForResourceLoadStatistics>(globalObject, decoder);
#endif
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::NetworkProcess_SetShouldDowngradeReferrerForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetShouldDowngradeReferrerForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetThirdPartyCookieBlockingMode:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetThirdPartyCookieBlockingMode>(globalObject, decoder);
    case MessageName::NetworkProcess_SetShouldEnbleSameSiteStrictEnforcementForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetShouldEnbleSameSiteStrictEnforcementForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetFirstPartyWebsiteDataRemovalModeForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetFirstPartyWebsiteDataRemovalModeForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetToSameSiteStrictCookiesForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetToSameSiteStrictCookiesForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetFirstPartyHostCNAMEDomainForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetFirstPartyHostCNAMEDomainForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetThirdPartyCNAMEDomainForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetThirdPartyCNAMEDomainForTesting>(globalObject, decoder);
#endif
    case MessageName::NetworkProcess_SyncLocalStorage:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SyncLocalStorage>(globalObject, decoder);
    case MessageName::NetworkProcess_DumpPrivateClickMeasurement:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_DumpPrivateClickMeasurement>(globalObject, decoder);
    case MessageName::NetworkProcess_ClearPrivateClickMeasurement:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_ClearPrivateClickMeasurement>(globalObject, decoder);
    case MessageName::NetworkProcess_SetPrivateClickMeasurementOverrideTimerForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetPrivateClickMeasurementOverrideTimerForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_MarkAttributedPrivateClickMeasurementsAsExpiredForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_MarkAttributedPrivateClickMeasurementsAsExpiredForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetPrivateClickMeasurementEphemeralMeasurementForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetPrivateClickMeasurementEphemeralMeasurementForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SimulatePrivateClickMeasurementSessionRestart:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SimulatePrivateClickMeasurementSessionRestart>(globalObject, decoder);
    case MessageName::NetworkProcess_SetPrivateClickMeasurementTokenPublicKeyURLForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetPrivateClickMeasurementTokenPublicKeyURLForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetPrivateClickMeasurementTokenSignatureURLForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetPrivateClickMeasurementTokenSignatureURLForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetPrivateClickMeasurementAttributionReportURLsForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetPrivateClickMeasurementAttributionReportURLsForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_MarkPrivateClickMeasurementsAsExpiredForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_MarkPrivateClickMeasurementsAsExpiredForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetPCMFraudPreventionValuesForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetPCMFraudPreventionValuesForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_SetPrivateClickMeasurementAppBundleIDForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetPrivateClickMeasurementAppBundleIDForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_ClosePCMDatabase:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_ClosePCMDatabase>(globalObject, decoder);
    case MessageName::NetworkProcess_SetServiceWorkerFetchTimeoutForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetServiceWorkerFetchTimeoutForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_ResetServiceWorkerFetchTimeoutForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_ResetServiceWorkerFetchTimeoutForTesting>(globalObject, decoder);
    case MessageName::NetworkProcess_ResetQuota:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_ResetQuota>(globalObject, decoder);
#if PLATFORM(IOS_FAMILY)
    case MessageName::NetworkProcess_SetBackupExclusionPeriodForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetBackupExclusionPeriodForTesting>(globalObject, decoder);
#endif
    case MessageName::NetworkProcess_ResetStoragePersistedState:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_ResetStoragePersistedState>(globalObject, decoder);
#if ENABLE(APP_BOUND_DOMAINS)
    case MessageName::NetworkProcess_HasAppBoundSession:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_HasAppBoundSession>(globalObject, decoder);
    case MessageName::NetworkProcess_ClearAppBoundSession:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_ClearAppBoundSession>(globalObject, decoder);
#endif
#if PLATFORM(IOS_FAMILY) && !PLATFORM(MACCATALYST)
    case MessageName::NetworkProcess_ClearServiceWorkerEntitlementOverride:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_ClearServiceWorkerEntitlementOverride>(globalObject, decoder);
#endif
    case MessageName::NetworkProcess_UpdateBundleIdentifier:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_UpdateBundleIdentifier>(globalObject, decoder);
    case MessageName::NetworkProcess_ClearBundleIdentifier:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_ClearBundleIdentifier>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::NetworkProcess_AppPrivacyReportTestingData:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_AppPrivacyReportTestingData>(globalObject, decoder);
    case MessageName::NetworkProcess_ClearAppPrivacyReportTestingData:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_ClearAppPrivacyReportTestingData>(globalObject, decoder);
#endif
    case MessageName::NetworkProcess_CountNonDefaultSessionSets:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_CountNonDefaultSessionSets>(globalObject, decoder);
#if ENABLE(SERVICE_WORKER)
    case MessageName::NetworkProcess_GetPendingPushMessages:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_GetPendingPushMessages>(globalObject, decoder);
    case MessageName::NetworkProcess_ProcessPushMessage:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_ProcessPushMessage>(globalObject, decoder);
    case MessageName::NetworkProcess_ProcessNotificationEvent:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_ProcessNotificationEvent>(globalObject, decoder);
#endif
    case MessageName::NetworkProcess_SetPushAndNotificationsEnabledForOrigin:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_SetPushAndNotificationsEnabledForOrigin>(globalObject, decoder);
    case MessageName::NetworkProcess_DeletePushAndNotificationRegistration:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_DeletePushAndNotificationRegistration>(globalObject, decoder);
    case MessageName::NetworkProcess_GetOriginsWithPushAndNotificationPermissions:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_GetOriginsWithPushAndNotificationPermissions>(globalObject, decoder);
    case MessageName::NetworkProcess_HasPushSubscriptionForTesting:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcess_HasPushSubscriptionForTesting>(globalObject, decoder);
    case MessageName::NetworkSocketChannel_SendString:
        return jsValueForDecodedMessageReply<MessageName::NetworkSocketChannel_SendString>(globalObject, decoder);
    case MessageName::NetworkSocketChannel_SendData:
        return jsValueForDecodedMessageReply<MessageName::NetworkSocketChannel_SendData>(globalObject, decoder);
#if ENABLE(SERVICE_WORKER)
#endif
#if ENABLE(SERVICE_WORKER)
#endif
#if ENABLE(SERVICE_WORKER)
    case MessageName::WebSWServerConnection_ScheduleUnregisterJobInServer:
        return jsValueForDecodedMessageReply<MessageName::WebSWServerConnection_ScheduleUnregisterJobInServer>(globalObject, decoder);
    case MessageName::WebSWServerConnection_MatchRegistration:
        return jsValueForDecodedMessageReply<MessageName::WebSWServerConnection_MatchRegistration>(globalObject, decoder);
    case MessageName::WebSWServerConnection_WhenRegistrationReady:
        return jsValueForDecodedMessageReply<MessageName::WebSWServerConnection_WhenRegistrationReady>(globalObject, decoder);
    case MessageName::WebSWServerConnection_GetRegistrations:
        return jsValueForDecodedMessageReply<MessageName::WebSWServerConnection_GetRegistrations>(globalObject, decoder);
    case MessageName::WebSWServerConnection_TerminateWorkerFromClient:
        return jsValueForDecodedMessageReply<MessageName::WebSWServerConnection_TerminateWorkerFromClient>(globalObject, decoder);
    case MessageName::WebSWServerConnection_WhenServiceWorkerIsTerminatedForTesting:
        return jsValueForDecodedMessageReply<MessageName::WebSWServerConnection_WhenServiceWorkerIsTerminatedForTesting>(globalObject, decoder);
    case MessageName::WebSWServerConnection_StoreRegistrationsOnDisk:
        return jsValueForDecodedMessageReply<MessageName::WebSWServerConnection_StoreRegistrationsOnDisk>(globalObject, decoder);
    case MessageName::WebSWServerConnection_SubscribeToPushService:
        return jsValueForDecodedMessageReply<MessageName::WebSWServerConnection_SubscribeToPushService>(globalObject, decoder);
    case MessageName::WebSWServerConnection_UnsubscribeFromPushService:
        return jsValueForDecodedMessageReply<MessageName::WebSWServerConnection_UnsubscribeFromPushService>(globalObject, decoder);
    case MessageName::WebSWServerConnection_GetPushSubscription:
        return jsValueForDecodedMessageReply<MessageName::WebSWServerConnection_GetPushSubscription>(globalObject, decoder);
    case MessageName::WebSWServerConnection_GetPushPermissionState:
        return jsValueForDecodedMessageReply<MessageName::WebSWServerConnection_GetPushPermissionState>(globalObject, decoder);
    case MessageName::WebSWServerConnection_EnableNavigationPreload:
        return jsValueForDecodedMessageReply<MessageName::WebSWServerConnection_EnableNavigationPreload>(globalObject, decoder);
    case MessageName::WebSWServerConnection_DisableNavigationPreload:
        return jsValueForDecodedMessageReply<MessageName::WebSWServerConnection_DisableNavigationPreload>(globalObject, decoder);
    case MessageName::WebSWServerConnection_SetNavigationPreloadHeaderValue:
        return jsValueForDecodedMessageReply<MessageName::WebSWServerConnection_SetNavigationPreloadHeaderValue>(globalObject, decoder);
    case MessageName::WebSWServerConnection_GetNavigationPreloadState:
        return jsValueForDecodedMessageReply<MessageName::WebSWServerConnection_GetNavigationPreloadState>(globalObject, decoder);
#endif
#if ENABLE(SERVICE_WORKER)
    case MessageName::WebSWServerToContextConnection_FindClientByVisibleIdentifier:
        return jsValueForDecodedMessageReply<MessageName::WebSWServerToContextConnection_FindClientByVisibleIdentifier>(globalObject, decoder);
    case MessageName::WebSWServerToContextConnection_Claim:
        return jsValueForDecodedMessageReply<MessageName::WebSWServerToContextConnection_Claim>(globalObject, decoder);
    case MessageName::WebSWServerToContextConnection_Focus:
        return jsValueForDecodedMessageReply<MessageName::WebSWServerToContextConnection_Focus>(globalObject, decoder);
    case MessageName::WebSWServerToContextConnection_Navigate:
        return jsValueForDecodedMessageReply<MessageName::WebSWServerToContextConnection_Navigate>(globalObject, decoder);
    case MessageName::WebSWServerToContextConnection_OpenWindow:
        return jsValueForDecodedMessageReply<MessageName::WebSWServerToContextConnection_OpenWindow>(globalObject, decoder);
#endif
    case MessageName::CacheStorageEngineConnection_Open:
        return jsValueForDecodedMessageReply<MessageName::CacheStorageEngineConnection_Open>(globalObject, decoder);
    case MessageName::CacheStorageEngineConnection_Remove:
        return jsValueForDecodedMessageReply<MessageName::CacheStorageEngineConnection_Remove>(globalObject, decoder);
    case MessageName::CacheStorageEngineConnection_Caches:
        return jsValueForDecodedMessageReply<MessageName::CacheStorageEngineConnection_Caches>(globalObject, decoder);
    case MessageName::CacheStorageEngineConnection_RetrieveRecords:
        return jsValueForDecodedMessageReply<MessageName::CacheStorageEngineConnection_RetrieveRecords>(globalObject, decoder);
    case MessageName::CacheStorageEngineConnection_DeleteMatchingRecords:
        return jsValueForDecodedMessageReply<MessageName::CacheStorageEngineConnection_DeleteMatchingRecords>(globalObject, decoder);
    case MessageName::CacheStorageEngineConnection_PutRecords:
        return jsValueForDecodedMessageReply<MessageName::CacheStorageEngineConnection_PutRecords>(globalObject, decoder);
    case MessageName::CacheStorageEngineConnection_ClearMemoryRepresentation:
        return jsValueForDecodedMessageReply<MessageName::CacheStorageEngineConnection_ClearMemoryRepresentation>(globalObject, decoder);
    case MessageName::CacheStorageEngineConnection_EngineRepresentation:
        return jsValueForDecodedMessageReply<MessageName::CacheStorageEngineConnection_EngineRepresentation>(globalObject, decoder);
    case MessageName::NetworkStorageManager_Persisted:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_Persisted>(globalObject, decoder);
    case MessageName::NetworkStorageManager_Persist:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_Persist>(globalObject, decoder);
    case MessageName::NetworkStorageManager_Estimate:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_Estimate>(globalObject, decoder);
    case MessageName::NetworkStorageManager_FileSystemGetDirectory:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_FileSystemGetDirectory>(globalObject, decoder);
    case MessageName::NetworkStorageManager_IsSameEntry:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_IsSameEntry>(globalObject, decoder);
    case MessageName::NetworkStorageManager_GetFileHandle:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_GetFileHandle>(globalObject, decoder);
    case MessageName::NetworkStorageManager_GetDirectoryHandle:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_GetDirectoryHandle>(globalObject, decoder);
    case MessageName::NetworkStorageManager_RemoveEntry:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_RemoveEntry>(globalObject, decoder);
    case MessageName::NetworkStorageManager_Resolve:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_Resolve>(globalObject, decoder);
    case MessageName::NetworkStorageManager_Move:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_Move>(globalObject, decoder);
    case MessageName::NetworkStorageManager_GetFile:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_GetFile>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CreateSyncAccessHandle:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_CreateSyncAccessHandle>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CloseSyncAccessHandle:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_CloseSyncAccessHandle>(globalObject, decoder);
    case MessageName::NetworkStorageManager_RequestNewCapacityForSyncAccessHandle:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_RequestNewCapacityForSyncAccessHandle>(globalObject, decoder);
    case MessageName::NetworkStorageManager_GetHandleNames:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_GetHandleNames>(globalObject, decoder);
    case MessageName::NetworkStorageManager_GetHandle:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_GetHandle>(globalObject, decoder);
    case MessageName::NetworkStorageManager_ConnectToStorageArea:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_ConnectToStorageArea>(globalObject, decoder);
    case MessageName::NetworkStorageManager_ConnectToStorageAreaSync:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_ConnectToStorageAreaSync>(globalObject, decoder);
    case MessageName::NetworkStorageManager_SetItem:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_SetItem>(globalObject, decoder);
    case MessageName::NetworkStorageManager_RemoveItem:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_RemoveItem>(globalObject, decoder);
    case MessageName::NetworkStorageManager_Clear:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_Clear>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CacheStorageOpenCache:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_CacheStorageOpenCache>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CacheStorageRemoveCache:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_CacheStorageRemoveCache>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CacheStorageAllCaches:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_CacheStorageAllCaches>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CacheStorageRetrieveRecords:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_CacheStorageRetrieveRecords>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CacheStorageRemoveRecords:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_CacheStorageRemoveRecords>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CacheStoragePutRecords:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_CacheStoragePutRecords>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CacheStorageClearMemoryRepresentation:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_CacheStorageClearMemoryRepresentation>(globalObject, decoder);
    case MessageName::NetworkStorageManager_CacheStorageRepresentation:
        return jsValueForDecodedMessageReply<MessageName::NetworkStorageManager_CacheStorageRepresentation>(globalObject, decoder);
#if ENABLE(WEB_RTC)
#endif
#if USE(LIBWEBRTC)
#endif
#if USE(LIBWEBRTC)
#endif
#if ENABLE(WEB_RTC)
#endif
    case MessageName::AuxiliaryProcess_MainThreadPing:
        return jsValueForDecodedMessageReply<MessageName::AuxiliaryProcess_MainThreadPing>(globalObject, decoder);
#if ENABLE(IPC_TESTING_API)
    case MessageName::IPCConnectionTester_SyncMessage:
        return jsValueForDecodedMessageReply<MessageName::IPCConnectionTester_SyncMessage>(globalObject, decoder);
#endif
#if ENABLE(IPC_TESTING_API)
    case MessageName::IPCStreamTester_SyncMessageReturningSharedMemory1:
        return jsValueForDecodedMessageReply<MessageName::IPCStreamTester_SyncMessageReturningSharedMemory1>(globalObject, decoder);
    case MessageName::IPCStreamTester_SyncCrashOnZero:
        return jsValueForDecodedMessageReply<MessageName::IPCStreamTester_SyncCrashOnZero>(globalObject, decoder);
    case MessageName::IPCStreamTester_CheckAutoreleasePool:
        return jsValueForDecodedMessageReply<MessageName::IPCStreamTester_CheckAutoreleasePool>(globalObject, decoder);
    case MessageName::IPCStreamTester_AsyncMessage:
        return jsValueForDecodedMessageReply<MessageName::IPCStreamTester_AsyncMessage>(globalObject, decoder);
#endif
#if ENABLE(IPC_TESTING_API)
#endif
#if ENABLE(IPC_TESTING_API)
    case MessageName::IPCTester_StopMessageTesting:
        return jsValueForDecodedMessageReply<MessageName::IPCTester_StopMessageTesting>(globalObject, decoder);
    case MessageName::IPCTester_ReleaseStreamTester:
        return jsValueForDecodedMessageReply<MessageName::IPCTester_ReleaseStreamTester>(globalObject, decoder);
    case MessageName::IPCTester_ReleaseConnectionTester:
        return jsValueForDecodedMessageReply<MessageName::IPCTester_ReleaseConnectionTester>(globalObject, decoder);
#endif
#if ENABLE(IPC_TESTING_API)
    case MessageName::IPCTesterReceiver_AsyncMessage:
        return jsValueForDecodedMessageReply<MessageName::IPCTesterReceiver_AsyncMessage>(globalObject, decoder);
#endif
    case MessageName::NotificationManagerMessageHandler_RequestSystemNotificationPermission:
        return jsValueForDecodedMessageReply<MessageName::NotificationManagerMessageHandler_RequestSystemNotificationPermission>(globalObject, decoder);
    case MessageName::NotificationManagerMessageHandler_ShowNotification:
        return jsValueForDecodedMessageReply<MessageName::NotificationManagerMessageHandler_ShowNotification>(globalObject, decoder);
#if ENABLE(MEDIA_SESSION_COORDINATOR)
    case MessageName::RemoteMediaSessionCoordinatorProxy_Join:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaSessionCoordinatorProxy_Join>(globalObject, decoder);
    case MessageName::RemoteMediaSessionCoordinatorProxy_CoordinateSeekTo:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaSessionCoordinatorProxy_CoordinateSeekTo>(globalObject, decoder);
    case MessageName::RemoteMediaSessionCoordinatorProxy_CoordinatePlay:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaSessionCoordinatorProxy_CoordinatePlay>(globalObject, decoder);
    case MessageName::RemoteMediaSessionCoordinatorProxy_CoordinatePause:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaSessionCoordinatorProxy_CoordinatePause>(globalObject, decoder);
    case MessageName::RemoteMediaSessionCoordinatorProxy_CoordinateSetTrack:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaSessionCoordinatorProxy_CoordinateSetTrack>(globalObject, decoder);
#endif
#if ENABLE(MEDIA_STREAM)
#endif
#if ENABLE(FULLSCREEN_API)
    case MessageName::WebFullScreenManagerProxy_SupportsFullScreen:
        return jsValueForDecodedMessageReply<MessageName::WebFullScreenManagerProxy_SupportsFullScreen>(globalObject, decoder);
#endif
    case MessageName::WebLockRegistryProxy_AbortLockRequest:
        return jsValueForDecodedMessageReply<MessageName::WebLockRegistryProxy_AbortLockRequest>(globalObject, decoder);
    case MessageName::WebLockRegistryProxy_Snapshot:
        return jsValueForDecodedMessageReply<MessageName::WebLockRegistryProxy_Snapshot>(globalObject, decoder);
    case MessageName::WebPageProxy_CreateNewPage:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_CreateNewPage>(globalObject, decoder);
    case MessageName::WebPageProxy_RunJavaScriptAlert:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_RunJavaScriptAlert>(globalObject, decoder);
    case MessageName::WebPageProxy_RunJavaScriptConfirm:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_RunJavaScriptConfirm>(globalObject, decoder);
    case MessageName::WebPageProxy_RunJavaScriptPrompt:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_RunJavaScriptPrompt>(globalObject, decoder);
    case MessageName::WebPageProxy_FocusFromServiceWorker:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_FocusFromServiceWorker>(globalObject, decoder);
    case MessageName::WebPageProxy_GetToolbarsAreVisible:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_GetToolbarsAreVisible>(globalObject, decoder);
    case MessageName::WebPageProxy_GetMenuBarIsVisible:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_GetMenuBarIsVisible>(globalObject, decoder);
    case MessageName::WebPageProxy_GetStatusBarIsVisible:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_GetStatusBarIsVisible>(globalObject, decoder);
    case MessageName::WebPageProxy_GetWindowFrame:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_GetWindowFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_ScreenToRootView:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ScreenToRootView>(globalObject, decoder);
    case MessageName::WebPageProxy_RootViewToScreen:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_RootViewToScreen>(globalObject, decoder);
    case MessageName::WebPageProxy_AccessibilityScreenToRootView:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_AccessibilityScreenToRootView>(globalObject, decoder);
    case MessageName::WebPageProxy_RootViewToAccessibilityScreen:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_RootViewToAccessibilityScreen>(globalObject, decoder);
    case MessageName::WebPageProxy_RunBeforeUnloadConfirmPanel:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_RunBeforeUnloadConfirmPanel>(globalObject, decoder);
    case MessageName::WebPageProxy_ShowShareSheet:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ShowShareSheet>(globalObject, decoder);
    case MessageName::WebPageProxy_ShowContactPicker:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ShowContactPicker>(globalObject, decoder);
    case MessageName::WebPageProxy_PrintFrame:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_PrintFrame>(globalObject, decoder);
    case MessageName::WebPageProxy_DecidePolicyForNavigationActionSync:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_DecidePolicyForNavigationActionSync>(globalObject, decoder);
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebPageProxy_InterpretKeyEvent:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_InterpretKeyEvent>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_BackForwardGoToItem:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_BackForwardGoToItem>(globalObject, decoder);
    case MessageName::WebPageProxy_BackForwardItemAtIndex:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_BackForwardItemAtIndex>(globalObject, decoder);
    case MessageName::WebPageProxy_BackForwardListContainsItem:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_BackForwardListContainsItem>(globalObject, decoder);
    case MessageName::WebPageProxy_BackForwardListCounts:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_BackForwardListCounts>(globalObject, decoder);
    case MessageName::WebPageProxy_CanUndoRedo:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_CanUndoRedo>(globalObject, decoder);
    case MessageName::WebPageProxy_ExecuteUndoRedo:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ExecuteUndoRedo>(globalObject, decoder);
#if ENABLE(IMAGE_ANALYSIS)
    case MessageName::WebPageProxy_RequestTextRecognition:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_RequestTextRecognition>(globalObject, decoder);
#endif
#if ENABLE(MEDIA_CONTROLS_CONTEXT_MENUS) && USE(UICONTEXTMENU)
    case MessageName::WebPageProxy_ShowMediaControlsContextMenu:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ShowMediaControlsContextMenu>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_RequestDOMPasteAccess:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_RequestDOMPasteAccess>(globalObject, decoder);
    case MessageName::WebPageProxy_ExceededDatabaseQuota:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ExceededDatabaseQuota>(globalObject, decoder);
    case MessageName::WebPageProxy_ReachedApplicationCacheOriginQuota:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ReachedApplicationCacheOriginQuota>(globalObject, decoder);
#if ENABLE(MEDIA_STREAM)
    case MessageName::WebPageProxy_EnumerateMediaDevicesForFrame:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_EnumerateMediaDevicesForFrame>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_RequestNotificationPermission:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_RequestNotificationPermission>(globalObject, decoder);
#if USE(UNIFIED_TEXT_CHECKING)
    case MessageName::WebPageProxy_CheckTextOfParagraph:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_CheckTextOfParagraph>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_CheckSpellingOfString:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_CheckSpellingOfString>(globalObject, decoder);
    case MessageName::WebPageProxy_CheckGrammarOfString:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_CheckGrammarOfString>(globalObject, decoder);
    case MessageName::WebPageProxy_SpellingUIIsShowing:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_SpellingUIIsShowing>(globalObject, decoder);
    case MessageName::WebPageProxy_GetGuessesForWord:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_GetGuessesForWord>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::WebPageProxy_ExecuteSavedCommandBySelector:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ExecuteSavedCommandBySelector>(globalObject, decoder);
    case MessageName::WebPageProxy_GetIsSpeaking:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_GetIsSpeaking>(globalObject, decoder);
#endif
#if USE(APPKIT)
    case MessageName::WebPageProxy_SubstitutionsPanelIsShowing:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_SubstitutionsPanelIsShowing>(globalObject, decoder);
#endif
#if PLATFORM(MAC)
    case MessageName::WebPageProxy_DismissCorrectionPanelSoon:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_DismissCorrectionPanelSoon>(globalObject, decoder);
#endif
#if USE(DICTATION_ALTERNATIVES)
    case MessageName::WebPageProxy_DictationAlternatives:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_DictationAlternatives>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_LoadRecentSearches:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_LoadRecentSearches>(globalObject, decoder);
#if ENABLE(PDFKIT_PLUGIN)
    case MessageName::WebPageProxy_ShowPDFContextMenu:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ShowPDFContextMenu>(globalObject, decoder);
#endif
#if ENABLE(WEB_CRYPTO)
    case MessageName::WebPageProxy_WrapCryptoKey:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_WrapCryptoKey>(globalObject, decoder);
    case MessageName::WebPageProxy_UnwrapCryptoKey:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_UnwrapCryptoKey>(globalObject, decoder);
#endif
#if USE(QUICK_LOOK)
    case MessageName::WebPageProxy_RequestPasswordForQuickLookDocumentInMainFrame:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_RequestPasswordForQuickLookDocumentInMainFrame>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_HandleSynchronousMessage:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_HandleSynchronousMessage>(globalObject, decoder);
    case MessageName::WebPageProxy_LoadSynchronousURLSchemeTask:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_LoadSynchronousURLSchemeTask>(globalObject, decoder);
#if ENABLE(DEVICE_ORIENTATION)
    case MessageName::WebPageProxy_ShouldAllowDeviceOrientationAndMotionAccess:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ShouldAllowDeviceOrientationAndMotionAccess>(globalObject, decoder);
#endif
#if ENABLE(ATTACHMENT_ELEMENT)
    case MessageName::WebPageProxy_SerializedAttachmentDataForIdentifiers:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_SerializedAttachmentDataForIdentifiers>(globalObject, decoder);
#endif
#if ENABLE(SPEECH_SYNTHESIS)
    case MessageName::WebPageProxy_SpeechSynthesisVoiceList:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_SpeechSynthesisVoiceList>(globalObject, decoder);
    case MessageName::WebPageProxy_SpeechSynthesisSpeak:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_SpeechSynthesisSpeak>(globalObject, decoder);
    case MessageName::WebPageProxy_SpeechSynthesisSetFinishedCallback:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_SpeechSynthesisSetFinishedCallback>(globalObject, decoder);
    case MessageName::WebPageProxy_SpeechSynthesisPause:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_SpeechSynthesisPause>(globalObject, decoder);
    case MessageName::WebPageProxy_SpeechSynthesisResume:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_SpeechSynthesisResume>(globalObject, decoder);
#endif
#if PLATFORM(GTK)
    case MessageName::WebPageProxy_ShowEmojiPicker:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ShowEmojiPicker>(globalObject, decoder);
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    case MessageName::WebPageProxy_SendMessageToWebViewWithReply:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_SendMessageToWebViewWithReply>(globalObject, decoder);
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW_MAC)
    case MessageName::WebPageProxy_ModelElementCreateRemotePreview:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ModelElementCreateRemotePreview>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementLoadRemotePreview:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ModelElementLoadRemotePreview>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementSizeDidChange:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ModelElementSizeDidChange>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelInlinePreviewUUIDs:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ModelInlinePreviewUUIDs>(globalObject, decoder);
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW)
    case MessageName::WebPageProxy_ModelElementGetCamera:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ModelElementGetCamera>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementSetCamera:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ModelElementSetCamera>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementIsPlayingAnimation:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ModelElementIsPlayingAnimation>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementSetAnimationIsPlaying:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ModelElementSetAnimationIsPlaying>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementIsLoopingAnimation:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ModelElementIsLoopingAnimation>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementSetIsLoopingAnimation:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ModelElementSetIsLoopingAnimation>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementAnimationDuration:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ModelElementAnimationDuration>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementAnimationCurrentTime:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ModelElementAnimationCurrentTime>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementSetAnimationCurrentTime:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ModelElementSetAnimationCurrentTime>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementHasAudio:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ModelElementHasAudio>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementIsMuted:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ModelElementIsMuted>(globalObject, decoder);
    case MessageName::WebPageProxy_ModelElementSetIsMuted:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ModelElementSetIsMuted>(globalObject, decoder);
#endif
#if ENABLE(APPLE_PAY_AMS_UI)
    case MessageName::WebPageProxy_StartApplePayAMSUISession:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_StartApplePayAMSUISession>(globalObject, decoder);
#endif
    case MessageName::WebPageProxy_RequestCookieConsent:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_RequestCookieConsent>(globalObject, decoder);
    case MessageName::WebPageProxy_ClassifyModalContainerControls:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_ClassifyModalContainerControls>(globalObject, decoder);
    case MessageName::WebPageProxy_DecidePolicyForModalContainer:
        return jsValueForDecodedMessageReply<MessageName::WebPageProxy_DecidePolicyForModalContainer>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_WriteCustomData:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_WriteCustomData>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_TypesSafeForDOMToReadAndWrite:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_TypesSafeForDOMToReadAndWrite>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_AllPasteboardItemInfo:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_AllPasteboardItemInfo>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_InformationForItemAtIndex:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_InformationForItemAtIndex>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_GetPasteboardItemsCount:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_GetPasteboardItemsCount>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_ReadStringFromPasteboard:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_ReadStringFromPasteboard>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_ReadURLFromPasteboard:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_ReadURLFromPasteboard>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_ReadBufferFromPasteboard:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_ReadBufferFromPasteboard>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_ContainsStringSafeForDOMToReadForType:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_ContainsStringSafeForDOMToReadForType>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::WebPasteboardProxy_GetNumberOfFiles:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_GetNumberOfFiles>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_GetPasteboardTypes:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_GetPasteboardTypes>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_GetPasteboardPathnamesForType:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_GetPasteboardPathnamesForType>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_GetPasteboardStringForType:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_GetPasteboardStringForType>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_GetPasteboardStringsForType:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_GetPasteboardStringsForType>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_GetPasteboardBufferForType:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_GetPasteboardBufferForType>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_GetPasteboardChangeCount:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_GetPasteboardChangeCount>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_GetPasteboardColor:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_GetPasteboardColor>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_GetPasteboardURL:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_GetPasteboardURL>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_AddPasteboardTypes:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_AddPasteboardTypes>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_SetPasteboardTypes:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_SetPasteboardTypes>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_SetPasteboardURL:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_SetPasteboardURL>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_SetPasteboardColor:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_SetPasteboardColor>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_SetPasteboardStringForType:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_SetPasteboardStringForType>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_SetPasteboardBufferForType:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_SetPasteboardBufferForType>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_ContainsURLStringSuitableForLoading:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_ContainsURLStringSuitableForLoading>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_URLStringSuitableForLoading:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_URLStringSuitableForLoading>(globalObject, decoder);
#endif
#if (PLATFORM(COCOA) && ENABLE(IPC_TESTING_API))
    case MessageName::WebPasteboardProxy_TestIPCSharedMemory:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_TestIPCSharedMemory>(globalObject, decoder);
#endif
#if PLATFORM(GTK)
    case MessageName::WebPasteboardProxy_GetTypes:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_GetTypes>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_ReadText:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_ReadText>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_ReadFilePaths:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_ReadFilePaths>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_ReadBuffer:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_ReadBuffer>(globalObject, decoder);
    case MessageName::WebPasteboardProxy_GetPasteboardChangeCount:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_GetPasteboardChangeCount>(globalObject, decoder);
#endif
#if USE(LIBWPE)
    case MessageName::WebPasteboardProxy_GetPasteboardTypes:
        return jsValueForDecodedMessageReply<MessageName::WebPasteboardProxy_GetPasteboardTypes>(globalObject, decoder);
#endif
    case MessageName::WebPermissionControllerProxy_Query:
        return jsValueForDecodedMessageReply<MessageName::WebPermissionControllerProxy_Query>(globalObject, decoder);
    case MessageName::WebProcessPool_HandleSynchronousMessage:
        return jsValueForDecodedMessageReply<MessageName::WebProcessPool_HandleSynchronousMessage>(globalObject, decoder);
#if ENABLE(GAMEPAD)
    case MessageName::WebProcessPool_StoppedUsingGamepads:
        return jsValueForDecodedMessageReply<MessageName::WebProcessPool_StoppedUsingGamepads>(globalObject, decoder);
    case MessageName::WebProcessPool_PlayGamepadEffect:
        return jsValueForDecodedMessageReply<MessageName::WebProcessPool_PlayGamepadEffect>(globalObject, decoder);
    case MessageName::WebProcessPool_StopGamepadEffects:
        return jsValueForDecodedMessageReply<MessageName::WebProcessPool_StopGamepadEffects>(globalObject, decoder);
#endif
    case MessageName::WebProcessProxy_ShouldTerminate:
        return jsValueForDecodedMessageReply<MessageName::WebProcessProxy_ShouldTerminate>(globalObject, decoder);
    case MessageName::WebProcessProxy_GetNetworkProcessConnection:
        return jsValueForDecodedMessageReply<MessageName::WebProcessProxy_GetNetworkProcessConnection>(globalObject, decoder);
#if PLATFORM(GTK) || PLATFORM(WPE)
    case MessageName::WebProcessProxy_SendMessageToWebContextWithReply:
        return jsValueForDecodedMessageReply<MessageName::WebProcessProxy_SendMessageToWebContextWithReply>(globalObject, decoder);
#endif
#if PLATFORM(MAC)
    case MessageName::WebProcessProxy_IsAXAuthenticated:
        return jsValueForDecodedMessageReply<MessageName::WebProcessProxy_IsAXAuthenticated>(globalObject, decoder);
#endif
    case MessageName::WebProcessProxy_GetNotifications:
        return jsValueForDecodedMessageReply<MessageName::WebProcessProxy_GetNotifications>(globalObject, decoder);
    case MessageName::WebScreenOrientationManagerProxy_CurrentOrientation:
        return jsValueForDecodedMessageReply<MessageName::WebScreenOrientationManagerProxy_CurrentOrientation>(globalObject, decoder);
    case MessageName::WebScreenOrientationManagerProxy_Lock:
        return jsValueForDecodedMessageReply<MessageName::WebScreenOrientationManagerProxy_Lock>(globalObject, decoder);
    case MessageName::DownloadProxy_DecideDestinationWithSuggestedFilename:
        return jsValueForDecodedMessageReply<MessageName::DownloadProxy_DecideDestinationWithSuggestedFilename>(globalObject, decoder);
#if ENABLE(GPU_PROCESS)
#if ENABLE(MEDIA_STREAM) && PLATFORM(IOS_FAMILY)
    case MessageName::GPUProcessProxy_StatusBarWasTapped:
        return jsValueForDecodedMessageReply<MessageName::GPUProcessProxy_StatusBarWasTapped>(globalObject, decoder);
#endif
#endif
    case MessageName::RemoteWebInspectorUIProxy_Load:
        return jsValueForDecodedMessageReply<MessageName::RemoteWebInspectorUIProxy_Load>(globalObject, decoder);
    case MessageName::RemoteWebInspectorUIProxy_PickColorFromScreen:
        return jsValueForDecodedMessageReply<MessageName::RemoteWebInspectorUIProxy_PickColorFromScreen>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_Load:
        return jsValueForDecodedMessageReply<MessageName::WebInspectorUIProxy_Load>(globalObject, decoder);
    case MessageName::WebInspectorUIProxy_PickColorFromScreen:
        return jsValueForDecodedMessageReply<MessageName::WebInspectorUIProxy_PickColorFromScreen>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_TestProcessIncomingSyncMessagesWhenWaitingForSyncReply:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcessProxy_TestProcessIncomingSyncMessagesWhenWaitingForSyncReply>(globalObject, decoder);
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::NetworkProcessProxy_RequestStorageAccessConfirm:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcessProxy_RequestStorageAccessConfirm>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_DeleteWebsiteDataInUIProcessForRegistrableDomains:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcessProxy_DeleteWebsiteDataInUIProcessForRegistrableDomains>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_SetDomainsWithCrossPageStorageAccess:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcessProxy_SetDomainsWithCrossPageStorageAccess>(globalObject, decoder);
#endif
    case MessageName::NetworkProcessProxy_EstablishRemoteWorkerContextConnectionToNetworkProcess:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcessProxy_EstablishRemoteWorkerContextConnectionToNetworkProcess>(globalObject, decoder);
#if ENABLE(APP_BOUND_DOMAINS)
    case MessageName::NetworkProcessProxy_GetAppBoundDomains:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcessProxy_GetAppBoundDomains>(globalObject, decoder);
#endif
    case MessageName::NetworkProcessProxy_RequestStorageSpace:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcessProxy_RequestStorageSpace>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_TriggerBrowsingContextGroupSwitchForNavigation:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcessProxy_TriggerBrowsingContextGroupSwitchForNavigation>(globalObject, decoder);
#if ENABLE(APPLE_PAY_REMOTE_UI_USES_SCENE)
    case MessageName::NetworkProcessProxy_GetWindowSceneIdentifierForPaymentPresentation:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcessProxy_GetWindowSceneIdentifierForPaymentPresentation>(globalObject, decoder);
#endif
    case MessageName::NetworkProcessProxy_DataTaskReceivedChallenge:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcessProxy_DataTaskReceivedChallenge>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_DataTaskWillPerformHTTPRedirection:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcessProxy_DataTaskWillPerformHTTPRedirection>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_DataTaskDidReceiveResponse:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcessProxy_DataTaskDidReceiveResponse>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_OpenWindowFromServiceWorker:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcessProxy_OpenWindowFromServiceWorker>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_NavigateServiceWorkerClient:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcessProxy_NavigateServiceWorkerClient>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_DeleteWebsiteDataInWebProcessesForOrigin:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcessProxy_DeleteWebsiteDataInWebProcessesForOrigin>(globalObject, decoder);
    case MessageName::NetworkProcessProxy_ReloadExecutionContextsForOrigin:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcessProxy_ReloadExecutionContextsForOrigin>(globalObject, decoder);
    case MessageName::WebUserContentControllerProxy_DidPostMessage:
        return jsValueForDecodedMessageReply<MessageName::WebUserContentControllerProxy_DidPostMessage>(globalObject, decoder);
#if ENABLE(WEBXR)
    case MessageName::PlatformXRSystem_EnumerateImmersiveXRDevices:
        return jsValueForDecodedMessageReply<MessageName::PlatformXRSystem_EnumerateImmersiveXRDevices>(globalObject, decoder);
    case MessageName::PlatformXRSystem_RequestPermissionOnSessionFeatures:
        return jsValueForDecodedMessageReply<MessageName::PlatformXRSystem_RequestPermissionOnSessionFeatures>(globalObject, decoder);
    case MessageName::PlatformXRSystem_RequestFrame:
        return jsValueForDecodedMessageReply<MessageName::PlatformXRSystem_RequestFrame>(globalObject, decoder);
#endif
    case MessageName::WebProcess_FetchWebsiteData:
        return jsValueForDecodedMessageReply<MessageName::WebProcess_FetchWebsiteData>(globalObject, decoder);
    case MessageName::WebProcess_DeleteWebsiteData:
        return jsValueForDecodedMessageReply<MessageName::WebProcess_DeleteWebsiteData>(globalObject, decoder);
    case MessageName::WebProcess_DeleteWebsiteDataForOrigins:
        return jsValueForDecodedMessageReply<MessageName::WebProcess_DeleteWebsiteDataForOrigins>(globalObject, decoder);
    case MessageName::WebProcess_DeleteWebsiteDataForOrigin:
        return jsValueForDecodedMessageReply<MessageName::WebProcess_DeleteWebsiteDataForOrigin>(globalObject, decoder);
    case MessageName::WebProcess_ReloadExecutionContextsForOrigin:
        return jsValueForDecodedMessageReply<MessageName::WebProcess_ReloadExecutionContextsForOrigin>(globalObject, decoder);
    case MessageName::WebProcess_DeleteAllCookies:
        return jsValueForDecodedMessageReply<MessageName::WebProcess_DeleteAllCookies>(globalObject, decoder);
    case MessageName::WebProcess_PrepareToSuspend:
        return jsValueForDecodedMessageReply<MessageName::WebProcess_PrepareToSuspend>(globalObject, decoder);
    case MessageName::WebProcess_EstablishRemoteWorkerContextConnectionToNetworkProcess:
        return jsValueForDecodedMessageReply<MessageName::WebProcess_EstablishRemoteWorkerContextConnectionToNetworkProcess>(globalObject, decoder);
    case MessageName::WebProcess_GetActivePagesOriginsForTesting:
        return jsValueForDecodedMessageReply<MessageName::WebProcess_GetActivePagesOriginsForTesting>(globalObject, decoder);
    case MessageName::WebProcess_IsJITEnabled:
        return jsValueForDecodedMessageReply<MessageName::WebProcess_IsJITEnabled>(globalObject, decoder);
    case MessageName::WebProcess_ClearCachedPage:
        return jsValueForDecodedMessageReply<MessageName::WebProcess_ClearCachedPage>(globalObject, decoder);
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::WebProcess_SeedResourceLoadStatisticsForTesting:
        return jsValueForDecodedMessageReply<MessageName::WebProcess_SeedResourceLoadStatisticsForTesting>(globalObject, decoder);
    case MessageName::WebProcess_SetThirdPartyCookieBlockingMode:
        return jsValueForDecodedMessageReply<MessageName::WebProcess_SetThirdPartyCookieBlockingMode>(globalObject, decoder);
    case MessageName::WebProcess_SetDomainsWithCrossPageStorageAccess:
        return jsValueForDecodedMessageReply<MessageName::WebProcess_SetDomainsWithCrossPageStorageAccess>(globalObject, decoder);
    case MessageName::WebProcess_SendResourceLoadStatisticsDataImmediately:
        return jsValueForDecodedMessageReply<MessageName::WebProcess_SendResourceLoadStatisticsDataImmediately>(globalObject, decoder);
#endif
    case MessageName::WebAutomationSessionProxy_ResolveChildFrameWithOrdinal:
        return jsValueForDecodedMessageReply<MessageName::WebAutomationSessionProxy_ResolveChildFrameWithOrdinal>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_ResolveChildFrameWithNodeHandle:
        return jsValueForDecodedMessageReply<MessageName::WebAutomationSessionProxy_ResolveChildFrameWithNodeHandle>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_ResolveChildFrameWithName:
        return jsValueForDecodedMessageReply<MessageName::WebAutomationSessionProxy_ResolveChildFrameWithName>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_ResolveParentFrame:
        return jsValueForDecodedMessageReply<MessageName::WebAutomationSessionProxy_ResolveParentFrame>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_ComputeElementLayout:
        return jsValueForDecodedMessageReply<MessageName::WebAutomationSessionProxy_ComputeElementLayout>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_GetComputedRole:
        return jsValueForDecodedMessageReply<MessageName::WebAutomationSessionProxy_GetComputedRole>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_GetComputedLabel:
        return jsValueForDecodedMessageReply<MessageName::WebAutomationSessionProxy_GetComputedLabel>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_SelectOptionElement:
        return jsValueForDecodedMessageReply<MessageName::WebAutomationSessionProxy_SelectOptionElement>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_SetFilesForInputFileUpload:
        return jsValueForDecodedMessageReply<MessageName::WebAutomationSessionProxy_SetFilesForInputFileUpload>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_SnapshotRectForScreenshot:
        return jsValueForDecodedMessageReply<MessageName::WebAutomationSessionProxy_SnapshotRectForScreenshot>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_GetCookiesForFrame:
        return jsValueForDecodedMessageReply<MessageName::WebAutomationSessionProxy_GetCookiesForFrame>(globalObject, decoder);
    case MessageName::WebAutomationSessionProxy_DeleteCookie:
        return jsValueForDecodedMessageReply<MessageName::WebAutomationSessionProxy_DeleteCookie>(globalObject, decoder);
#if ENABLE(FULLSCREEN_API)
#endif
#if ENABLE(GPU_PROCESS)
#if ENABLE(ROUTING_ARBITRATION)
    case MessageName::GPUProcessConnection_BeginRoutingArbitrationWithCategory:
        return jsValueForDecodedMessageReply<MessageName::GPUProcessConnection_BeginRoutingArbitrationWithCategory>(globalObject, decoder);
#endif
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(WEBGL)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
    case MessageName::MediaPlayerPrivateRemote_SendH2Ping:
        return jsValueForDecodedMessageReply<MessageName::MediaPlayerPrivateRemote_SendH2Ping>(globalObject, decoder);
#if PLATFORM(IOS_FAMILY)
    case MessageName::MediaPlayerPrivateRemote_GetRawCookies:
        return jsValueForDecodedMessageReply<MessageName::MediaPlayerPrivateRemote_GetRawCookies>(globalObject, decoder);
#endif
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
#endif
#if ENABLE(GPU_PROCESS) && USE(AUDIO_SESSION)
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateDidReceiveInitializationSegment:
        return jsValueForDecodedMessageReply<MessageName::SourceBufferPrivateRemote_SourceBufferPrivateDidReceiveInitializationSegment>(globalObject, decoder);
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateDurationChanged:
        return jsValueForDecodedMessageReply<MessageName::SourceBufferPrivateRemote_SourceBufferPrivateDurationChanged>(globalObject, decoder);
#endif
#if USE(LIBWEBRTC) && PLATFORM(COCOA) && ENABLE(GPU_PROCESS)
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
#endif
#if ENABLE(MEDIA_SESSION_COORDINATOR)
    case MessageName::RemoteMediaSessionCoordinator_SeekSessionToTime:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaSessionCoordinator_SeekSessionToTime>(globalObject, decoder);
    case MessageName::RemoteMediaSessionCoordinator_PlaySession:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaSessionCoordinator_PlaySession>(globalObject, decoder);
    case MessageName::RemoteMediaSessionCoordinator_PauseSession:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaSessionCoordinator_PauseSession>(globalObject, decoder);
    case MessageName::RemoteMediaSessionCoordinator_SetSessionTrack:
        return jsValueForDecodedMessageReply<MessageName::RemoteMediaSessionCoordinator_SetSessionTrack>(globalObject, decoder);
#endif
#if USE(LIBWEBRTC)
#endif
#if ENABLE(WEB_RTC)
#endif
#if ENABLE(WEB_RTC)
#endif
#if USE(LIBWEBRTC)
#endif
#if USE(LIBWEBRTC)
#endif
#if ENABLE(WEB_RTC)
    case MessageName::NetworkProcessConnection_ConnectToRTCDataChannelRemoteSource:
        return jsValueForDecodedMessageReply<MessageName::NetworkProcessConnection_ConnectToRTCDataChannelRemoteSource>(globalObject, decoder);
#endif
#if ENABLE(MEDIA_STREAM)
#endif
    case MessageName::WebSharedWorkerContextManagerConnection_PostConnectEvent:
        return jsValueForDecodedMessageReply<MessageName::WebSharedWorkerContextManagerConnection_PostConnectEvent>(globalObject, decoder);
    case MessageName::WebSharedWorkerObjectConnection_FetchScriptInClient:
        return jsValueForDecodedMessageReply<MessageName::WebSharedWorkerObjectConnection_FetchScriptInClient>(globalObject, decoder);
#if ENABLE(SERVICE_WORKER)
    case MessageName::WebSWClientConnection_SetServiceWorkerClientIsControlled:
        return jsValueForDecodedMessageReply<MessageName::WebSWClientConnection_SetServiceWorkerClientIsControlled>(globalObject, decoder);
    case MessageName::WebSWClientConnection_FocusServiceWorkerClient:
        return jsValueForDecodedMessageReply<MessageName::WebSWClientConnection_FocusServiceWorkerClient>(globalObject, decoder);
#endif
#if ENABLE(SERVICE_WORKER)
    case MessageName::WebSWContextManagerConnection_FirePushEvent:
        return jsValueForDecodedMessageReply<MessageName::WebSWContextManagerConnection_FirePushEvent>(globalObject, decoder);
    case MessageName::WebSWContextManagerConnection_FireNotificationEvent:
        return jsValueForDecodedMessageReply<MessageName::WebSWContextManagerConnection_FireNotificationEvent>(globalObject, decoder);
#endif
    case MessageName::WebBroadcastChannelRegistry_PostMessageToRemote:
        return jsValueForDecodedMessageReply<MessageName::WebBroadcastChannelRegistry_PostMessageToRemote>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::DrawingArea_UpdateGeometry:
        return jsValueForDecodedMessageReply<MessageName::DrawingArea_UpdateGeometry>(globalObject, decoder);
    case MessageName::DrawingArea_DispatchAfterEnsuringDrawing:
        return jsValueForDecodedMessageReply<MessageName::DrawingArea_DispatchAfterEnsuringDrawing>(globalObject, decoder);
#endif
#if ENABLE(IOS_TOUCH_EVENTS)
    case MessageName::EventDispatcher_TouchEvent:
        return jsValueForDecodedMessageReply<MessageName::EventDispatcher_TouchEvent>(globalObject, decoder);
#endif
    case MessageName::WebFrame_GetFrameInfo:
        return jsValueForDecodedMessageReply<MessageName::WebFrame_GetFrameInfo>(globalObject, decoder);
    case MessageName::WebPage_SetInitialFocus:
        return jsValueForDecodedMessageReply<MessageName::WebPage_SetInitialFocus>(globalObject, decoder);
    case MessageName::WebPage_SetActivityState:
        return jsValueForDecodedMessageReply<MessageName::WebPage_SetActivityState>(globalObject, decoder);
    case MessageName::WebPage_TestProcessIncomingSyncMessagesWhenWaitingForSyncReply:
        return jsValueForDecodedMessageReply<MessageName::WebPage_TestProcessIncomingSyncMessagesWhenWaitingForSyncReply>(globalObject, decoder);
    case MessageName::WebPage_ExecuteEditCommandWithCallback:
        return jsValueForDecodedMessageReply<MessageName::WebPage_ExecuteEditCommandWithCallback>(globalObject, decoder);
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebPage_SelectWithGesture:
        return jsValueForDecodedMessageReply<MessageName::WebPage_SelectWithGesture>(globalObject, decoder);
    case MessageName::WebPage_UpdateSelectionWithTouches:
        return jsValueForDecodedMessageReply<MessageName::WebPage_UpdateSelectionWithTouches>(globalObject, decoder);
    case MessageName::WebPage_SelectWithTwoTouches:
        return jsValueForDecodedMessageReply<MessageName::WebPage_SelectWithTwoTouches>(globalObject, decoder);
    case MessageName::WebPage_ExtendSelection:
        return jsValueForDecodedMessageReply<MessageName::WebPage_ExtendSelection>(globalObject, decoder);
    case MessageName::WebPage_ExtendSelectionForReplacement:
        return jsValueForDecodedMessageReply<MessageName::WebPage_ExtendSelectionForReplacement>(globalObject, decoder);
    case MessageName::WebPage_MoveSelectionByOffset:
        return jsValueForDecodedMessageReply<MessageName::WebPage_MoveSelectionByOffset>(globalObject, decoder);
    case MessageName::WebPage_SelectTextWithGranularityAtPoint:
        return jsValueForDecodedMessageReply<MessageName::WebPage_SelectTextWithGranularityAtPoint>(globalObject, decoder);
    case MessageName::WebPage_SelectPositionAtBoundaryWithDirection:
        return jsValueForDecodedMessageReply<MessageName::WebPage_SelectPositionAtBoundaryWithDirection>(globalObject, decoder);
    case MessageName::WebPage_MoveSelectionAtBoundaryWithDirection:
        return jsValueForDecodedMessageReply<MessageName::WebPage_MoveSelectionAtBoundaryWithDirection>(globalObject, decoder);
    case MessageName::WebPage_SelectPositionAtPoint:
        return jsValueForDecodedMessageReply<MessageName::WebPage_SelectPositionAtPoint>(globalObject, decoder);
    case MessageName::WebPage_BeginSelectionInDirection:
        return jsValueForDecodedMessageReply<MessageName::WebPage_BeginSelectionInDirection>(globalObject, decoder);
    case MessageName::WebPage_UpdateSelectionWithExtentPoint:
        return jsValueForDecodedMessageReply<MessageName::WebPage_UpdateSelectionWithExtentPoint>(globalObject, decoder);
    case MessageName::WebPage_UpdateSelectionWithExtentPointAndBoundary:
        return jsValueForDecodedMessageReply<MessageName::WebPage_UpdateSelectionWithExtentPointAndBoundary>(globalObject, decoder);
    case MessageName::WebPage_RequestDictationContext:
        return jsValueForDecodedMessageReply<MessageName::WebPage_RequestDictationContext>(globalObject, decoder);
#endif
#if (PLATFORM(IOS_FAMILY) && ENABLE(REVEAL))
    case MessageName::WebPage_RequestRVItemInCurrentSelectedRange:
        return jsValueForDecodedMessageReply<MessageName::WebPage_RequestRVItemInCurrentSelectedRange>(globalObject, decoder);
    case MessageName::WebPage_PrepareSelectionForContextMenuWithLocationInView:
        return jsValueForDecodedMessageReply<MessageName::WebPage_PrepareSelectionForContextMenuWithLocationInView>(globalObject, decoder);
#endif
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebPage_RequestAutocorrectionData:
        return jsValueForDecodedMessageReply<MessageName::WebPage_RequestAutocorrectionData>(globalObject, decoder);
    case MessageName::WebPage_ApplyAutocorrection:
        return jsValueForDecodedMessageReply<MessageName::WebPage_ApplyAutocorrection>(globalObject, decoder);
    case MessageName::WebPage_SyncApplyAutocorrection:
        return jsValueForDecodedMessageReply<MessageName::WebPage_SyncApplyAutocorrection>(globalObject, decoder);
    case MessageName::WebPage_RequestEvasionRectsAboveSelection:
        return jsValueForDecodedMessageReply<MessageName::WebPage_RequestEvasionRectsAboveSelection>(globalObject, decoder);
    case MessageName::WebPage_PerformActionOnElement:
        return jsValueForDecodedMessageReply<MessageName::WebPage_PerformActionOnElement>(globalObject, decoder);
    case MessageName::WebPage_FocusNextFocusedElement:
        return jsValueForDecodedMessageReply<MessageName::WebPage_FocusNextFocusedElement>(globalObject, decoder);
    case MessageName::WebPage_GetSelectionContext:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetSelectionContext>(globalObject, decoder);
    case MessageName::WebPage_GetRectsForGranularityWithSelectionOffset:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetRectsForGranularityWithSelectionOffset>(globalObject, decoder);
    case MessageName::WebPage_GetRectsAtSelectionOffsetWithText:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetRectsAtSelectionOffsetWithText>(globalObject, decoder);
    case MessageName::WebPage_RequestFocusedElementInformation:
        return jsValueForDecodedMessageReply<MessageName::WebPage_RequestFocusedElementInformation>(globalObject, decoder);
    case MessageName::WebPage_UpdateSelectionWithDelta:
        return jsValueForDecodedMessageReply<MessageName::WebPage_UpdateSelectionWithDelta>(globalObject, decoder);
    case MessageName::WebPage_RequestDocumentEditingContext:
        return jsValueForDecodedMessageReply<MessageName::WebPage_RequestDocumentEditingContext>(globalObject, decoder);
    case MessageName::WebPage_InsertTextPlaceholder:
        return jsValueForDecodedMessageReply<MessageName::WebPage_InsertTextPlaceholder>(globalObject, decoder);
    case MessageName::WebPage_RemoveTextPlaceholder:
        return jsValueForDecodedMessageReply<MessageName::WebPage_RemoveTextPlaceholder>(globalObject, decoder);
    case MessageName::WebPage_TextInputContextsInRect:
        return jsValueForDecodedMessageReply<MessageName::WebPage_TextInputContextsInRect>(globalObject, decoder);
    case MessageName::WebPage_FocusTextInputContextAndPlaceCaret:
        return jsValueForDecodedMessageReply<MessageName::WebPage_FocusTextInputContextAndPlaceCaret>(globalObject, decoder);
    case MessageName::WebPage_ClearServiceWorkerEntitlementOverride:
        return jsValueForDecodedMessageReply<MessageName::WebPage_ClearServiceWorkerEntitlementOverride>(globalObject, decoder);
#endif
    case MessageName::WebPage_RequestImageBitmap:
        return jsValueForDecodedMessageReply<MessageName::WebPage_RequestImageBitmap>(globalObject, decoder);
    case MessageName::WebPage_GetPDFFirstPageSize:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetPDFFirstPageSize>(globalObject, decoder);
    case MessageName::WebPage_GetContentsAsString:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetContentsAsString>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::WebPage_GetContentsAsAttributedString:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetContentsAsAttributedString>(globalObject, decoder);
#endif
#if ENABLE(MHTML)
    case MessageName::WebPage_GetContentsAsMHTMLData:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetContentsAsMHTMLData>(globalObject, decoder);
#endif
    case MessageName::WebPage_GetMainResourceDataOfFrame:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetMainResourceDataOfFrame>(globalObject, decoder);
    case MessageName::WebPage_GetResourceDataFromFrame:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetResourceDataFromFrame>(globalObject, decoder);
    case MessageName::WebPage_GetRenderTreeExternalRepresentation:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetRenderTreeExternalRepresentation>(globalObject, decoder);
    case MessageName::WebPage_GetSelectionOrContentsAsString:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetSelectionOrContentsAsString>(globalObject, decoder);
    case MessageName::WebPage_GetSelectionAsWebArchiveData:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetSelectionAsWebArchiveData>(globalObject, decoder);
    case MessageName::WebPage_GetSourceForFrame:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetSourceForFrame>(globalObject, decoder);
    case MessageName::WebPage_GetWebArchiveOfFrame:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetWebArchiveOfFrame>(globalObject, decoder);
    case MessageName::WebPage_RunJavaScriptInFrameInScriptWorld:
        return jsValueForDecodedMessageReply<MessageName::WebPage_RunJavaScriptInFrameInScriptWorld>(globalObject, decoder);
    case MessageName::WebPage_GetAccessibilityTreeData:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetAccessibilityTreeData>(globalObject, decoder);
    case MessageName::WebPage_ForceRepaint:
        return jsValueForDecodedMessageReply<MessageName::WebPage_ForceRepaint>(globalObject, decoder);
#if ENABLE(DATA_DETECTION)
    case MessageName::WebPage_DetectDataInAllFrames:
        return jsValueForDecodedMessageReply<MessageName::WebPage_DetectDataInAllFrames>(globalObject, decoder);
    case MessageName::WebPage_RemoveDataDetectedLinks:
        return jsValueForDecodedMessageReply<MessageName::WebPage_RemoveDataDetectedLinks>(globalObject, decoder);
#endif
    case MessageName::WebPage_Suspend:
        return jsValueForDecodedMessageReply<MessageName::WebPage_Suspend>(globalObject, decoder);
    case MessageName::WebPage_Resume:
        return jsValueForDecodedMessageReply<MessageName::WebPage_Resume>(globalObject, decoder);
    case MessageName::WebPage_TryClose:
        return jsValueForDecodedMessageReply<MessageName::WebPage_TryClose>(globalObject, decoder);
    case MessageName::WebPage_ValidateCommand:
        return jsValueForDecodedMessageReply<MessageName::WebPage_ValidateCommand>(globalObject, decoder);
    case MessageName::WebPage_RequestFontAttributesAtSelectionStart:
        return jsValueForDecodedMessageReply<MessageName::WebPage_RequestFontAttributesAtSelectionStart>(globalObject, decoder);
    case MessageName::WebPage_FindString:
        return jsValueForDecodedMessageReply<MessageName::WebPage_FindString>(globalObject, decoder);
    case MessageName::WebPage_ReplaceMatches:
        return jsValueForDecodedMessageReply<MessageName::WebPage_ReplaceMatches>(globalObject, decoder);
    case MessageName::WebPage_FindRectsForStringMatches:
        return jsValueForDecodedMessageReply<MessageName::WebPage_FindRectsForStringMatches>(globalObject, decoder);
    case MessageName::WebPage_FindTextRangesForStringMatches:
        return jsValueForDecodedMessageReply<MessageName::WebPage_FindTextRangesForStringMatches>(globalObject, decoder);
    case MessageName::WebPage_RequestRectForFoundTextRange:
        return jsValueForDecodedMessageReply<MessageName::WebPage_RequestRectForFoundTextRange>(globalObject, decoder);
    case MessageName::WebPage_AddLayerForFindOverlay:
        return jsValueForDecodedMessageReply<MessageName::WebPage_AddLayerForFindOverlay>(globalObject, decoder);
    case MessageName::WebPage_RemoveLayerForFindOverlay:
        return jsValueForDecodedMessageReply<MessageName::WebPage_RemoveLayerForFindOverlay>(globalObject, decoder);
#if PLATFORM(IOS_FAMILY) && ENABLE(DRAG_SUPPORT)
    case MessageName::WebPage_InsertDroppedImagePlaceholders:
        return jsValueForDecodedMessageReply<MessageName::WebPage_InsertDroppedImagePlaceholders>(globalObject, decoder);
#endif
#if ENABLE(MEDIA_STREAM)
    case MessageName::WebPage_UserMediaAccessWasGranted:
        return jsValueForDecodedMessageReply<MessageName::WebPage_UserMediaAccessWasGranted>(globalObject, decoder);
#endif
#if ENABLE(ENCRYPTED_MEDIA)
    case MessageName::WebPage_MediaKeySystemWasGranted:
        return jsValueForDecodedMessageReply<MessageName::WebPage_MediaKeySystemWasGranted>(globalObject, decoder);
#endif
    case MessageName::WebPage_RequestMediaPlaybackState:
        return jsValueForDecodedMessageReply<MessageName::WebPage_RequestMediaPlaybackState>(globalObject, decoder);
    case MessageName::WebPage_PauseAllMediaPlayback:
        return jsValueForDecodedMessageReply<MessageName::WebPage_PauseAllMediaPlayback>(globalObject, decoder);
    case MessageName::WebPage_SuspendAllMediaPlayback:
        return jsValueForDecodedMessageReply<MessageName::WebPage_SuspendAllMediaPlayback>(globalObject, decoder);
    case MessageName::WebPage_ResumeAllMediaPlayback:
        return jsValueForDecodedMessageReply<MessageName::WebPage_ResumeAllMediaPlayback>(globalObject, decoder);
    case MessageName::WebPage_IsLayerTreeFrozen:
        return jsValueForDecodedMessageReply<MessageName::WebPage_IsLayerTreeFrozen>(globalObject, decoder);
    case MessageName::WebPage_EndPrinting:
        return jsValueForDecodedMessageReply<MessageName::WebPage_EndPrinting>(globalObject, decoder);
    case MessageName::WebPage_EndPrintingDuringDOMPrintOperation:
        return jsValueForDecodedMessageReply<MessageName::WebPage_EndPrintingDuringDOMPrintOperation>(globalObject, decoder);
    case MessageName::WebPage_ComputePagesForPrinting:
        return jsValueForDecodedMessageReply<MessageName::WebPage_ComputePagesForPrinting>(globalObject, decoder);
    case MessageName::WebPage_ComputePagesForPrintingDuringDOMPrintOperation:
        return jsValueForDecodedMessageReply<MessageName::WebPage_ComputePagesForPrintingDuringDOMPrintOperation>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::WebPage_DrawRectToImage:
        return jsValueForDecodedMessageReply<MessageName::WebPage_DrawRectToImage>(globalObject, decoder);
    case MessageName::WebPage_DrawRectToImageDuringDOMPrintOperation:
        return jsValueForDecodedMessageReply<MessageName::WebPage_DrawRectToImageDuringDOMPrintOperation>(globalObject, decoder);
    case MessageName::WebPage_DrawPagesToPDF:
        return jsValueForDecodedMessageReply<MessageName::WebPage_DrawPagesToPDF>(globalObject, decoder);
    case MessageName::WebPage_DrawPagesToPDFDuringDOMPrintOperation:
        return jsValueForDecodedMessageReply<MessageName::WebPage_DrawPagesToPDFDuringDOMPrintOperation>(globalObject, decoder);
#endif
#if (PLATFORM(COCOA) && PLATFORM(IOS_FAMILY))
    case MessageName::WebPage_ComputePagesForPrintingiOS:
        return jsValueForDecodedMessageReply<MessageName::WebPage_ComputePagesForPrintingiOS>(globalObject, decoder);
    case MessageName::WebPage_DrawToPDFiOS:
        return jsValueForDecodedMessageReply<MessageName::WebPage_DrawToPDFiOS>(globalObject, decoder);
#endif
#if PLATFORM(COCOA)
    case MessageName::WebPage_DrawToPDF:
        return jsValueForDecodedMessageReply<MessageName::WebPage_DrawToPDF>(globalObject, decoder);
#endif
#if PLATFORM(GTK)
    case MessageName::WebPage_DrawPagesForPrinting:
        return jsValueForDecodedMessageReply<MessageName::WebPage_DrawPagesForPrinting>(globalObject, decoder);
    case MessageName::WebPage_DrawPagesForPrintingDuringDOMPrintOperation:
        return jsValueForDecodedMessageReply<MessageName::WebPage_DrawPagesForPrintingDuringDOMPrintOperation>(globalObject, decoder);
#endif
    case MessageName::WebPage_SetMuted:
        return jsValueForDecodedMessageReply<MessageName::WebPage_SetMuted>(globalObject, decoder);
    case MessageName::WebPage_StopMediaCapture:
        return jsValueForDecodedMessageReply<MessageName::WebPage_StopMediaCapture>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::WebPage_GetStringSelectionForPasteboard:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetStringSelectionForPasteboard>(globalObject, decoder);
    case MessageName::WebPage_GetDataSelectionForPasteboard:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetDataSelectionForPasteboard>(globalObject, decoder);
    case MessageName::WebPage_ReadSelectionFromPasteboard:
        return jsValueForDecodedMessageReply<MessageName::WebPage_ReadSelectionFromPasteboard>(globalObject, decoder);
    case MessageName::WebPage_ShouldDelayWindowOrderingEvent:
        return jsValueForDecodedMessageReply<MessageName::WebPage_ShouldDelayWindowOrderingEvent>(globalObject, decoder);
    case MessageName::WebPage_AddDictationAlternative:
        return jsValueForDecodedMessageReply<MessageName::WebPage_AddDictationAlternative>(globalObject, decoder);
    case MessageName::WebPage_DictationAlternativesAtSelection:
        return jsValueForDecodedMessageReply<MessageName::WebPage_DictationAlternativesAtSelection>(globalObject, decoder);
    case MessageName::WebPage_HasMarkedText:
        return jsValueForDecodedMessageReply<MessageName::WebPage_HasMarkedText>(globalObject, decoder);
    case MessageName::WebPage_GetMarkedRangeAsync:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetMarkedRangeAsync>(globalObject, decoder);
    case MessageName::WebPage_GetSelectedRangeAsync:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetSelectedRangeAsync>(globalObject, decoder);
    case MessageName::WebPage_CharacterIndexForPointAsync:
        return jsValueForDecodedMessageReply<MessageName::WebPage_CharacterIndexForPointAsync>(globalObject, decoder);
    case MessageName::WebPage_FirstRectForCharacterRangeAsync:
        return jsValueForDecodedMessageReply<MessageName::WebPage_FirstRectForCharacterRangeAsync>(globalObject, decoder);
#endif
#if PLATFORM(MAC)
    case MessageName::WebPage_AttributedSubstringForCharacterRangeAsync:
        return jsValueForDecodedMessageReply<MessageName::WebPage_AttributedSubstringForCharacterRangeAsync>(globalObject, decoder);
#endif
#if ENABLE(IMAGE_ANALYSIS_ENHANCEMENTS)
    case MessageName::WebPage_ShouldAllowRemoveBackground:
        return jsValueForDecodedMessageReply<MessageName::WebPage_ShouldAllowRemoveBackground>(globalObject, decoder);
#endif
    case MessageName::WebPage_GetBytecodeProfile:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetBytecodeProfile>(globalObject, decoder);
    case MessageName::WebPage_GetSamplingProfilerOutput:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetSamplingProfilerOutput>(globalObject, decoder);
    case MessageName::WebPage_TakeSnapshot:
        return jsValueForDecodedMessageReply<MessageName::WebPage_TakeSnapshot>(globalObject, decoder);
#if PLATFORM(COCOA)
    case MessageName::WebPage_RequestActiveNowPlayingSessionInfo:
        return jsValueForDecodedMessageReply<MessageName::WebPage_RequestActiveNowPlayingSessionInfo>(globalObject, decoder);
#endif
    case MessageName::WebPage_DidGetLoadDecisionForIcon:
        return jsValueForDecodedMessageReply<MessageName::WebPage_DidGetLoadDecisionForIcon>(globalObject, decoder);
    case MessageName::WebPage_URLSchemeTaskWillPerformRedirection:
        return jsValueForDecodedMessageReply<MessageName::WebPage_URLSchemeTaskWillPerformRedirection>(globalObject, decoder);
#if ENABLE(ATTACHMENT_ELEMENT)
    case MessageName::WebPage_InsertAttachment:
        return jsValueForDecodedMessageReply<MessageName::WebPage_InsertAttachment>(globalObject, decoder);
    case MessageName::WebPage_UpdateAttachmentAttributes:
        return jsValueForDecodedMessageReply<MessageName::WebPage_UpdateAttachmentAttributes>(globalObject, decoder);
#endif
#if ENABLE(APPLICATION_MANIFEST)
    case MessageName::WebPage_GetApplicationManifest:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetApplicationManifest>(globalObject, decoder);
#endif
    case MessageName::WebPage_GetTextFragmentMatch:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetTextFragmentMatch>(globalObject, decoder);
#if ENABLE(UI_PROCESS_PDF_HUD)
    case MessageName::WebPage_SavePDF:
        return jsValueForDecodedMessageReply<MessageName::WebPage_SavePDF>(globalObject, decoder);
    case MessageName::WebPage_OpenPDFWithPreview:
        return jsValueForDecodedMessageReply<MessageName::WebPage_OpenPDFWithPreview>(globalObject, decoder);
#endif
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::WebPage_GetLoadedSubresourceDomains:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetLoadedSubresourceDomains>(globalObject, decoder);
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    case MessageName::WebPage_SendMessageToWebProcessExtensionWithReply:
        return jsValueForDecodedMessageReply<MessageName::WebPage_SendMessageToWebProcessExtensionWithReply>(globalObject, decoder);
#endif
    case MessageName::WebPage_StartTextManipulations:
        return jsValueForDecodedMessageReply<MessageName::WebPage_StartTextManipulations>(globalObject, decoder);
    case MessageName::WebPage_CompleteTextManipulation:
        return jsValueForDecodedMessageReply<MessageName::WebPage_CompleteTextManipulation>(globalObject, decoder);
    case MessageName::WebPage_GetProcessDisplayName:
        return jsValueForDecodedMessageReply<MessageName::WebPage_GetProcessDisplayName>(globalObject, decoder);
    case MessageName::WebPage_DispatchWheelEventWithoutScrolling:
        return jsValueForDecodedMessageReply<MessageName::WebPage_DispatchWheelEventWithoutScrolling>(globalObject, decoder);
    case MessageName::WebPage_LastNavigationWasAppInitiated:
        return jsValueForDecodedMessageReply<MessageName::WebPage_LastNavigationWasAppInitiated>(globalObject, decoder);
#if ENABLE(MEDIA_SESSION_COORDINATOR)
    case MessageName::WebPage_CreateMediaSessionCoordinator:
        return jsValueForDecodedMessageReply<MessageName::WebPage_CreateMediaSessionCoordinator>(globalObject, decoder);
#endif
#if ENABLE(IMAGE_ANALYSIS)
    case MessageName::WebPage_UpdateWithTextRecognitionResult:
        return jsValueForDecodedMessageReply<MessageName::WebPage_UpdateWithTextRecognitionResult>(globalObject, decoder);
#endif
    case MessageName::WebPage_NavigateServiceWorkerClient:
        return jsValueForDecodedMessageReply<MessageName::WebPage_NavigateServiceWorkerClient>(globalObject, decoder);
#if ENABLE(ACCESSIBILITY_ANIMATION_CONTROL)
    case MessageName::WebPage_PauseAllAnimations:
        return jsValueForDecodedMessageReply<MessageName::WebPage_PauseAllAnimations>(globalObject, decoder);
    case MessageName::WebPage_PlayAllAnimations:
        return jsValueForDecodedMessageReply<MessageName::WebPage_PlayAllAnimations>(globalObject, decoder);
#endif
#if ENABLE(WEBXR)
#endif
#if ENABLE(MEDIA_STREAM)
    case MessageName::UserMediaCaptureManager_ValidateUserMediaRequestConstraints:
        return jsValueForDecodedMessageReply<MessageName::UserMediaCaptureManager_ValidateUserMediaRequestConstraints>(globalObject, decoder);
    case MessageName::UserMediaCaptureManager_GetMediaStreamDevices:
        return jsValueForDecodedMessageReply<MessageName::UserMediaCaptureManager_GetMediaStreamDevices>(globalObject, decoder);
#endif
    default:
        break;
    }
    return std::nullopt;
}

Vector<ASCIILiteral> serializedIdentifiers()
{
    static_assert(sizeof(uint64_t) == sizeof(WebCore::BroadcastChannelIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::DOMCacheIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::DictationContext));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::DisplayList::ItemBufferIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::ElementIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::FetchIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::FileSystemHandleIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::FileSystemSyncAccessHandleIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::ImageDecoderIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::LayerHostingContextIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::LibWebRTCSocketIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::MediaKeySystemRequestIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::MediaPlayerIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::MediaSessionIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::OpaqueOriginIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::PageIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::PlaybackTargetClientContextIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::PushSubscriptionIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::PortIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::ProcessIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::RealtimeMediaSourceIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::RenderingResourceIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::ResourceLoaderIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::SWServerConnectionIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::ServiceWorkerIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::ServiceWorkerJobIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::ServiceWorkerRegistrationIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::SharedWorkerIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::SleepDisablerIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::SpeechRecognitionConnectionClientIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::TextCheckingRequestIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::TextManipulationTokenIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::UserMediaRequestIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebCore::WebSocketIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::AudioMediaStreamTrackRendererInternalUnitIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::AuthenticationChallengeIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::ContentWorldIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::DataTaskIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::FormSubmitListenerIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::GeolocationIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::GraphicsContextGLIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::IPCConnectionTesterIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::IPCStreamTesterIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::LegacyCustomProtocolID));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::LibWebRTCResolverIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::MDNSRegisterIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::MarkSurfacesAsVolatileRequestIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::MediaRecorderIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::NetworkResourceLoadIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::PDFPluginIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::PageGroupIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::PlaybackSessionContextIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::QuotaIncreaseRequestIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::RemoteAudioDestinationIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::RemoteAudioHardwareListenerIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::RemoteAudioSessionIdentifier));
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    static_assert(sizeof(uint64_t) == sizeof(WebKit::RemoteCDMIdentifier));
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    static_assert(sizeof(uint64_t) == sizeof(WebKit::RemoteCDMInstanceIdentifier));
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    static_assert(sizeof(uint64_t) == sizeof(WebKit::RemoteCDMInstanceSessionIdentifier));
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
    static_assert(sizeof(uint64_t) == sizeof(WebKit::RemoteLegacyCDMIdentifier));
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
    static_assert(sizeof(uint64_t) == sizeof(WebKit::RemoteLegacyCDMSessionIdentifier));
#endif
    static_assert(sizeof(uint64_t) == sizeof(WebKit::RemoteMediaResourceIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::RemoteSerializedImageBufferIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::RemoteVideoFrameIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::RemoteRemoteCommandListenerIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::RenderingBackendIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::RenderingUpdateID));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::SampleBufferDisplayLayerIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::StorageAreaIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::StorageAreaImplIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::StorageAreaMapIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::StorageNamespaceIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::TapIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::TrackPrivateRemoteIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::TransactionID));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::UserContentControllerIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::VideoDecoderIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::VideoEncoderIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::WebExtensionContextIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::WebExtensionControllerIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::WebGPUIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::WebPageProxyIdentifier));
    static_assert(sizeof(uint64_t) == sizeof(WebKit::WebURLSchemeHandlerIdentifier));
    return {
        "WebCore::BroadcastChannelIdentifier"_s,
        "WebCore::DOMCacheIdentifier"_s,
        "WebCore::DictationContext"_s,
        "WebCore::DisplayList::ItemBufferIdentifier"_s,
        "WebCore::ElementIdentifier"_s,
        "WebCore::FetchIdentifier"_s,
        "WebCore::FileSystemHandleIdentifier"_s,
        "WebCore::FileSystemSyncAccessHandleIdentifier"_s,
        "WebCore::ImageDecoderIdentifier"_s,
        "WebCore::LayerHostingContextIdentifier"_s,
        "WebCore::LibWebRTCSocketIdentifier"_s,
        "WebCore::MediaKeySystemRequestIdentifier"_s,
        "WebCore::MediaPlayerIdentifier"_s,
        "WebCore::MediaSessionIdentifier"_s,
        "WebCore::OpaqueOriginIdentifier"_s,
        "WebCore::PageIdentifier"_s,
        "WebCore::PlaybackTargetClientContextIdentifier"_s,
        "WebCore::PushSubscriptionIdentifier"_s,
        "WebCore::PortIdentifier"_s,
        "WebCore::ProcessIdentifier"_s,
        "WebCore::RealtimeMediaSourceIdentifier"_s,
        "WebCore::RenderingResourceIdentifier"_s,
        "WebCore::ResourceLoaderIdentifier"_s,
        "WebCore::SWServerConnectionIdentifier"_s,
        "WebCore::ServiceWorkerIdentifier"_s,
        "WebCore::ServiceWorkerJobIdentifier"_s,
        "WebCore::ServiceWorkerRegistrationIdentifier"_s,
        "WebCore::SharedWorkerIdentifier"_s,
        "WebCore::SleepDisablerIdentifier"_s,
        "WebCore::SpeechRecognitionConnectionClientIdentifier"_s,
        "WebCore::TextCheckingRequestIdentifier"_s,
        "WebCore::TextManipulationTokenIdentifier"_s,
        "WebCore::UserMediaRequestIdentifier"_s,
        "WebCore::WebSocketIdentifier"_s,
        "WebKit::AudioMediaStreamTrackRendererInternalUnitIdentifier"_s,
        "WebKit::AuthenticationChallengeIdentifier"_s,
        "WebKit::ContentWorldIdentifier"_s,
        "WebKit::DataTaskIdentifier"_s,
        "WebKit::FormSubmitListenerIdentifier"_s,
        "WebKit::GeolocationIdentifier"_s,
        "WebKit::GraphicsContextGLIdentifier"_s,
        "WebKit::IPCConnectionTesterIdentifier"_s,
        "WebKit::IPCStreamTesterIdentifier"_s,
        "WebKit::LegacyCustomProtocolID"_s,
        "WebKit::LibWebRTCResolverIdentifier"_s,
        "WebKit::MDNSRegisterIdentifier"_s,
        "WebKit::MarkSurfacesAsVolatileRequestIdentifier"_s,
        "WebKit::MediaRecorderIdentifier"_s,
        "WebKit::NetworkResourceLoadIdentifier"_s,
        "WebKit::PDFPluginIdentifier"_s,
        "WebKit::PageGroupIdentifier"_s,
        "WebKit::PlaybackSessionContextIdentifier"_s,
        "WebKit::QuotaIncreaseRequestIdentifier"_s,
        "WebKit::RemoteAudioDestinationIdentifier"_s,
        "WebKit::RemoteAudioHardwareListenerIdentifier"_s,
        "WebKit::RemoteAudioSessionIdentifier"_s,
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
        "WebKit::RemoteCDMIdentifier"_s,
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
        "WebKit::RemoteCDMInstanceIdentifier"_s,
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
        "WebKit::RemoteCDMInstanceSessionIdentifier"_s,
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
        "WebKit::RemoteLegacyCDMIdentifier"_s,
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
        "WebKit::RemoteLegacyCDMSessionIdentifier"_s,
#endif
        "WebKit::RemoteMediaResourceIdentifier"_s,
        "WebKit::RemoteSerializedImageBufferIdentifier"_s,
        "WebKit::RemoteVideoFrameIdentifier"_s,
        "WebKit::RemoteRemoteCommandListenerIdentifier"_s,
        "WebKit::RenderingBackendIdentifier"_s,
        "WebKit::RenderingUpdateID"_s,
        "WebKit::SampleBufferDisplayLayerIdentifier"_s,
        "WebKit::StorageAreaIdentifier"_s,
        "WebKit::StorageAreaImplIdentifier"_s,
        "WebKit::StorageAreaMapIdentifier"_s,
        "WebKit::StorageNamespaceIdentifier"_s,
        "WebKit::TapIdentifier"_s,
        "WebKit::TrackPrivateRemoteIdentifier"_s,
        "WebKit::TransactionID"_s,
        "WebKit::UserContentControllerIdentifier"_s,
        "WebKit::VideoDecoderIdentifier"_s,
        "WebKit::VideoEncoderIdentifier"_s,
        "WebKit::WebExtensionContextIdentifier"_s,
        "WebKit::WebExtensionControllerIdentifier"_s,
        "WebKit::WebGPUIdentifier"_s,
        "WebKit::WebPageProxyIdentifier"_s,
        "WebKit::WebURLSchemeHandlerIdentifier"_s,
    };
}

#endif // ENABLE(IPC_TESTING_API)

std::optional<Vector<ArgumentDescription>> messageArgumentDescriptions(MessageName name)
{
    switch (name) {
#if ENABLE(GPU_PROCESS)
    case MessageName::GPUConnectionToWebProcess_CreateRenderingBackend:
        return Vector<ArgumentDescription> {
            { "creationParameters", "WebKit::RemoteRenderingBackendCreationParameters", nullptr, false },
            { "connectionHandle", "IPC::StreamServerConnection::Handle", nullptr, false },
        };
    case MessageName::GPUConnectionToWebProcess_ReleaseRenderingBackend:
        return Vector<ArgumentDescription> {
            { "renderingBackendIdentifier", "WebKit::RenderingBackendIdentifier", nullptr, false },
        };
    case MessageName::GPUConnectionToWebProcess_ReleaseSerializedImageBuffer:
        return Vector<ArgumentDescription> {
            { "reference", "WebKit::RemoteSerializedImageBufferWriteReference", nullptr, false },
        };
#if ENABLE(WEBGL)
    case MessageName::GPUConnectionToWebProcess_CreateGraphicsContextGL:
        return Vector<ArgumentDescription> {
            { "attributes", "WebCore::GraphicsContextGLAttributes", nullptr, false },
            { "graphicsContextGLIdentifier", "WebKit::GraphicsContextGLIdentifier", nullptr, false },
            { "renderingBackendIdentifier", "WebKit::RenderingBackendIdentifier", nullptr, false },
            { "serverConnection", "IPC::StreamServerConnection::Handle", nullptr, false },
        };
    case MessageName::GPUConnectionToWebProcess_ReleaseGraphicsContextGL:
        return Vector<ArgumentDescription> {
            { "graphicsContextGLIdentifier", "WebKit::GraphicsContextGLIdentifier", nullptr, false },
        };
#endif
    case MessageName::GPUConnectionToWebProcess_CreateRemoteGPU:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
            { "renderingBackendIdentifier", "WebKit::RenderingBackendIdentifier", nullptr, false },
            { "serverConnection", "IPC::StreamServerConnection::Handle", nullptr, false },
        };
    case MessageName::GPUConnectionToWebProcess_ClearNowPlayingInfo:
        return Vector<ArgumentDescription> { };
    case MessageName::GPUConnectionToWebProcess_SetNowPlayingInfo:
        return Vector<ArgumentDescription> {
            { "nowPlayingInfo", "WebCore::NowPlayingInfo", nullptr, false },
        };
#if USE(AUDIO_SESSION)
    case MessageName::GPUConnectionToWebProcess_EnsureAudioSession:
        return Vector<ArgumentDescription> { };
#endif
#if PLATFORM(IOS_FAMILY)
    case MessageName::GPUConnectionToWebProcess_EnsureMediaSessionHelper:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(VP9)
    case MessageName::GPUConnectionToWebProcess_EnableVP9Decoders:
        return Vector<ArgumentDescription> {
            { "shouldEnableVP8Decoder", "bool", nullptr, false },
            { "shouldEnableVP9Decoder", "bool", nullptr, false },
            { "shouldEnableVP9SWDecoder", "bool", nullptr, false },
        };
#endif
#if HAVE(VISIBILITY_PROPAGATION_VIEW)
    case MessageName::GPUConnectionToWebProcess_CreateVisibilityPropagationContextForPage:
        return Vector<ArgumentDescription> {
            { "pageProxyID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "canShowWhileLocked", "bool", nullptr, false },
        };
    case MessageName::GPUConnectionToWebProcess_DestroyVisibilityPropagationContextForPage:
        return Vector<ArgumentDescription> {
            { "pageProxyID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
        };
#endif
    case MessageName::GPUConnectionToWebProcess_SetMediaOverridesForTesting:
        return Vector<ArgumentDescription> {
            { "configuration", "WebKit::MediaOverridesForTesting", nullptr, false },
        };
    case MessageName::GPUConnectionToWebProcess_CreateAudioHardwareListener:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteAudioHardwareListenerIdentifier", nullptr, false },
        };
    case MessageName::GPUConnectionToWebProcess_ReleaseAudioHardwareListener:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteAudioHardwareListenerIdentifier", nullptr, false },
        };
    case MessageName::GPUConnectionToWebProcess_CreateRemoteCommandListener:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteRemoteCommandListenerIdentifier", nullptr, false },
        };
    case MessageName::GPUConnectionToWebProcess_ReleaseRemoteCommandListener:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteRemoteCommandListenerIdentifier", nullptr, false },
        };
    case MessageName::GPUConnectionToWebProcess_SetUserPreferredLanguages:
        return Vector<ArgumentDescription> {
            { "languages", "Vector<String>", nullptr, false },
        };
    case MessageName::GPUConnectionToWebProcess_ConfigureLoggingChannel:
        return Vector<ArgumentDescription> {
            { "channelName", "String", nullptr, false },
            { "state", "uint8_t", "WTFLogChannelState", false },
            { "level", "uint8_t", "WTFLogLevel", false },
        };
#if USE(GRAPHICS_LAYER_WC)
    case MessageName::GPUConnectionToWebProcess_CreateWCLayerTreeHost:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::WCLayerTreeHostIdentifier", nullptr, false },
            { "nativeWindow", "uint64_t", nullptr, false },
            { "usesOffscreenRendering", "bool", nullptr, false },
        };
    case MessageName::GPUConnectionToWebProcess_ReleaseWCLayerTreeHost:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::WCLayerTreeHostIdentifier", nullptr, false },
        };
#endif
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::GPUProcess_InitializeGPUProcess:
        return Vector<ArgumentDescription> {
            { "processCreationParameters", "WebKit::GPUProcessCreationParameters", nullptr, false },
        };
    case MessageName::GPUProcess_CreateGPUConnectionToWebProcess:
        return Vector<ArgumentDescription> {
            { "processIdentifier", "WebCore::ProcessIdentifier", nullptr, false },
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "connectionHandle", "IPC::Connection::Handle", nullptr, false },
            { "parameters", "WebKit::GPUProcessConnectionParameters", nullptr, false },
        };
    case MessageName::GPUProcess_UpdateGPUProcessPreferences:
        return Vector<ArgumentDescription> {
            { "preferences", "WebKit::GPUProcessPreferences", nullptr, false },
        };
    case MessageName::GPUProcess_PrepareToSuspend:
        return Vector<ArgumentDescription> {
            { "isSuspensionImminent", "bool", nullptr, false },
            { "estimatedSuspendTime", "MonotonicTime", nullptr, false },
        };
    case MessageName::GPUProcess_ProcessDidResume:
        return Vector<ArgumentDescription> { };
    case MessageName::GPUProcess_AddSession:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "parameters", "WebKit::GPUProcessSessionParameters", nullptr, false },
        };
    case MessageName::GPUProcess_RemoveSession:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
#if ENABLE(MEDIA_STREAM)
    case MessageName::GPUProcess_SetMockCaptureDevicesEnabled:
        return Vector<ArgumentDescription> {
            { "isEnabled", "bool", nullptr, false },
        };
    case MessageName::GPUProcess_SetOrientationForMediaCapture:
        return Vector<ArgumentDescription> {
            { "orientation", "uint64_t", nullptr, false },
        };
    case MessageName::GPUProcess_UpdateSandboxAccess:
        return Vector<ArgumentDescription> {
            { "extensions", "Vector<WebKit::SandboxExtension::Handle>", nullptr, false },
        };
    case MessageName::GPUProcess_UpdateCaptureAccess:
        return Vector<ArgumentDescription> {
            { "allowAudioCapture", "bool", nullptr, false },
            { "allowVideoCapture", "bool", nullptr, false },
            { "allowDisplayCapture", "bool", nullptr, false },
            { "processID", "WebCore::ProcessIdentifier", nullptr, false },
        };
    case MessageName::GPUProcess_UpdateCaptureOrigin:
        return Vector<ArgumentDescription> {
            { "originData", "WebCore::SecurityOriginData", nullptr, false },
            { "processID", "WebCore::ProcessIdentifier", nullptr, false },
        };
    case MessageName::GPUProcess_AddMockMediaDevice:
        return Vector<ArgumentDescription> {
            { "device", "WebCore::MockMediaDevice", nullptr, false },
        };
    case MessageName::GPUProcess_ClearMockMediaDevices:
        return Vector<ArgumentDescription> { };
    case MessageName::GPUProcess_RemoveMockMediaDevice:
        return Vector<ArgumentDescription> {
            { "persistentId", "String", nullptr, false },
        };
    case MessageName::GPUProcess_SetMockMediaDeviceIsEphemeral:
        return Vector<ArgumentDescription> {
            { "persistentId", "String", nullptr, false },
            { "isEphemeral", "bool", nullptr, false },
        };
    case MessageName::GPUProcess_ResetMockMediaDevices:
        return Vector<ArgumentDescription> { };
    case MessageName::GPUProcess_SetMockCaptureDevicesInterrupted:
        return Vector<ArgumentDescription> {
            { "isCameraInterrupted", "bool", nullptr, false },
            { "isMicrophoneInterrupted", "bool", nullptr, false },
        };
    case MessageName::GPUProcess_TriggerMockMicrophoneConfigurationChange:
        return Vector<ArgumentDescription> { };
#endif
#if PLATFORM(MAC)
    case MessageName::GPUProcess_DisplayConfigurationChanged:
        return Vector<ArgumentDescription> {
            { "displayID", "CGDirectDisplayID", nullptr, false },
            { "flags", "CGDisplayChangeSummaryFlags", nullptr, false },
        };
    case MessageName::GPUProcess_SetScreenProperties:
        return Vector<ArgumentDescription> {
            { "screenProperties", "WebCore::ScreenProperties", nullptr, false },
        };
#endif
#if HAVE(SC_CONTENT_SHARING_SESSION)
    case MessageName::GPUProcess_ShowWindowPicker:
        return Vector<ArgumentDescription> { };
    case MessageName::GPUProcess_ShowScreenPicker:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(VIDEO)
    case MessageName::GPUProcess_RequestBitmapImageForCurrentTime:
        return Vector<ArgumentDescription> {
            { "processIdentifier", "WebCore::ProcessIdentifier", nullptr, false },
            { "playerIdentifier", "WebCore::MediaPlayerIdentifier", nullptr, false },
        };
#endif
#if ENABLE(CFPREFS_DIRECT_MODE)
    case MessageName::GPUProcess_NotifyPreferencesChanged:
        return Vector<ArgumentDescription> {
            { "domain", "String", nullptr, false },
            { "key", "String", nullptr, false },
            { "encodedValue", "String", nullptr, true },
        };
#endif
#if PLATFORM(MAC)
    case MessageName::GPUProcess_OpenDirectoryCacheInvalidated:
        return Vector<ArgumentDescription> {
            { "handle", "WebKit::SandboxExtension::Handle", nullptr, false },
        };
#endif
#if HAVE(AUDIO_COMPONENT_SERVER_REGISTRATIONS)
    case MessageName::GPUProcess_ConsumeAudioComponentRegistrations:
        return Vector<ArgumentDescription> {
            { "registrationData", "IPC::SharedBufferReference", nullptr, false },
        };
#endif
#if HAVE(POWERLOG_TASK_MODE_QUERY)
    case MessageName::GPUProcess_EnablePowerLogging:
        return Vector<ArgumentDescription> {
            { "handle", "WebKit::SandboxExtension::Handle", nullptr, false },
        };
#endif
    case MessageName::GPUProcess_WebProcessConnectionCountForTesting:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteDisplayListRecorder_Save:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteDisplayListRecorder_Restore:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteDisplayListRecorder_Translate:
        return Vector<ArgumentDescription> {
            { "x", "float", nullptr, false },
            { "y", "float", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_Rotate:
        return Vector<ArgumentDescription> {
            { "angle", "float", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_Scale:
        return Vector<ArgumentDescription> {
            { "scale", "WebCore::FloatSize", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_SetCTM:
        return Vector<ArgumentDescription> {
            { "ctm", "WebCore::AffineTransform", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_ConcatenateCTM:
        return Vector<ArgumentDescription> {
            { "ctm", "WebCore::AffineTransform", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_SetInlineFillColor:
        return Vector<ArgumentDescription> {
            { "item", "WebCore::DisplayList::SetInlineFillColor", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_SetInlineStrokeColor:
        return Vector<ArgumentDescription> {
            { "item", "WebCore::DisplayList::SetInlineStrokeColor", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_SetStrokeThickness:
        return Vector<ArgumentDescription> {
            { "thickness", "float", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_SetState:
        return Vector<ArgumentDescription> {
            { "item", "WebCore::DisplayList::SetState", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_SetLineCap:
        return Vector<ArgumentDescription> {
            { "lineCap", "uint8_t", "WebCore::LineCap", false },
        };
    case MessageName::RemoteDisplayListRecorder_SetLineDash:
        return Vector<ArgumentDescription> {
            { "item", "WebCore::DisplayList::SetLineDash", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_SetLineJoin:
        return Vector<ArgumentDescription> {
            { "lineJoin", "uint8_t", "WebCore::LineJoin", false },
        };
    case MessageName::RemoteDisplayListRecorder_SetMiterLimit:
        return Vector<ArgumentDescription> {
            { "limit", "float", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_ClearShadow:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteDisplayListRecorder_Clip:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::FloatRect", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_ClipOut:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::FloatRect", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_ClipToImageBuffer:
        return Vector<ArgumentDescription> {
            { "renderingResourceIdentifier", "WebCore::RenderingResourceIdentifier", nullptr, false },
            { "destinationRect", "WebCore::FloatRect", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_ClipOutToPath:
        return Vector<ArgumentDescription> {
            { "path", "WebCore::Path", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_ClipPath:
        return Vector<ArgumentDescription> {
            { "path", "WebCore::Path", nullptr, false },
            { "windRule", "bool", "WebCore::WindRule", false },
        };
    case MessageName::RemoteDisplayListRecorder_DrawGlyphs:
        return Vector<ArgumentDescription> {
            { "item", "WebCore::DisplayList::DrawGlyphs", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_DrawDecomposedGlyphs:
        return Vector<ArgumentDescription> {
            { "fontIdentifier", "WebCore::RenderingResourceIdentifier", nullptr, false },
            { "decomposedGlyphsIdentifier", "WebCore::RenderingResourceIdentifier", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_DrawFilteredImageBuffer:
        return Vector<ArgumentDescription> {
            { "sourceImageIdentifier", "WebCore::RenderingResourceIdentifier", nullptr, true },
            { "sourceImageRect", "WebCore::FloatRect", nullptr, false },
            { "filter", "Ref<WebCore::Filter>", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_DrawImageBuffer:
        return Vector<ArgumentDescription> {
            { "imageBufferIdentifier", "WebCore::RenderingResourceIdentifier", nullptr, false },
            { "destinationRect", "WebCore::FloatRect", nullptr, false },
            { "srcRect", "WebCore::FloatRect", nullptr, false },
            { "options", "WebCore::ImagePaintingOptions", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_DrawNativeImage:
        return Vector<ArgumentDescription> {
            { "imageIdentifier", "WebCore::RenderingResourceIdentifier", nullptr, false },
            { "imageSize", "WebCore::FloatSize", nullptr, false },
            { "destRect", "WebCore::FloatRect", nullptr, false },
            { "srcRect", "WebCore::FloatRect", nullptr, false },
            { "options", "WebCore::ImagePaintingOptions", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_DrawSystemImage:
        return Vector<ArgumentDescription> {
            { "systemImage", "Ref<WebCore::SystemImage>", nullptr, false },
            { "destinationRect", "WebCore::FloatRect", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_DrawPattern:
        return Vector<ArgumentDescription> {
            { "imageIdentifier", "WebCore::RenderingResourceIdentifier", nullptr, false },
            { "destRect", "WebCore::FloatRect", nullptr, false },
            { "tileRect", "WebCore::FloatRect", nullptr, false },
            { "transform", "WebCore::AffineTransform", nullptr, false },
            { "phase", "WebCore::FloatPoint", nullptr, false },
            { "spacing", "WebCore::FloatSize", nullptr, false },
            { "options", "WebCore::ImagePaintingOptions", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_BeginTransparencyLayer:
        return Vector<ArgumentDescription> {
            { "opacity", "float", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_EndTransparencyLayer:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteDisplayListRecorder_DrawRect:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::FloatRect", nullptr, false },
            { "borderThickness", "float", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_DrawLine:
        return Vector<ArgumentDescription> {
            { "point1", "WebCore::FloatPoint", nullptr, false },
            { "point2", "WebCore::FloatPoint", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_DrawLinesForText:
        return Vector<ArgumentDescription> {
            { "item", "WebCore::DisplayList::DrawLinesForText", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_DrawDotsForDocumentMarker:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::FloatRect", nullptr, false },
            { "style", "WebCore::DocumentMarkerLineStyle", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_DrawEllipse:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::FloatRect", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_DrawPath:
        return Vector<ArgumentDescription> {
            { "path", "WebCore::Path", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_DrawFocusRingPath:
        return Vector<ArgumentDescription> {
            { "path", "WebCore::Path", nullptr, false },
            { "outlineWidth", "float", nullptr, false },
            { "color", "WebCore::Color", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_DrawFocusRingRects:
        return Vector<ArgumentDescription> {
            { "rects", "Vector<WebCore::FloatRect>", nullptr, false },
            { "outlineOffset", "float", nullptr, false },
            { "outlineWidth", "float", nullptr, false },
            { "color", "WebCore::Color", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_FillRect:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::FloatRect", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_FillRectWithColor:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::FloatRect", nullptr, false },
            { "color", "WebCore::Color", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_FillRectWithGradient:
        return Vector<ArgumentDescription> {
            { "item", "WebCore::DisplayList::FillRectWithGradient", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_FillCompositedRect:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::FloatRect", nullptr, false },
            { "color", "WebCore::Color", nullptr, false },
            { "op", "uint8_t", "WebCore::CompositeOperator", false },
            { "blendMode", "uint8_t", "WebCore::BlendMode", false },
        };
    case MessageName::RemoteDisplayListRecorder_FillRoundedRect:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::FloatRoundedRect", nullptr, false },
            { "color", "WebCore::Color", nullptr, false },
            { "blendMode", "uint8_t", "WebCore::BlendMode", false },
        };
    case MessageName::RemoteDisplayListRecorder_FillRectWithRoundedHole:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::FloatRect", nullptr, false },
            { "roundedHoleRect", "WebCore::FloatRoundedRect", nullptr, false },
            { "color", "WebCore::Color", nullptr, false },
        };
#if ENABLE(INLINE_PATH_DATA)
    case MessageName::RemoteDisplayListRecorder_FillLine:
        return Vector<ArgumentDescription> {
            { "data", "WebCore::LineData", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_FillArc:
        return Vector<ArgumentDescription> {
            { "data", "WebCore::ArcData", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_FillQuadCurve:
        return Vector<ArgumentDescription> {
            { "data", "WebCore::QuadCurveData", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_FillBezierCurve:
        return Vector<ArgumentDescription> {
            { "data", "WebCore::BezierCurveData", nullptr, false },
        };
#endif
    case MessageName::RemoteDisplayListRecorder_FillPath:
        return Vector<ArgumentDescription> {
            { "path", "WebCore::Path", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_FillEllipse:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::FloatRect", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_ConvertToLuminanceMask:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteDisplayListRecorder_TransformToColorSpace:
        return Vector<ArgumentDescription> {
            { "colorSpace", "WebCore::DestinationColorSpace", nullptr, false },
        };
#if ENABLE(VIDEO)
    case MessageName::RemoteDisplayListRecorder_PaintFrameForMedia:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::MediaPlayerIdentifier", nullptr, false },
            { "destination", "WebCore::FloatRect", nullptr, false },
        };
#endif
    case MessageName::RemoteDisplayListRecorder_StrokeRect:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::FloatRect", nullptr, false },
            { "lineWidth", "float", nullptr, false },
        };
#if ENABLE(INLINE_PATH_DATA)
    case MessageName::RemoteDisplayListRecorder_StrokeLine:
        return Vector<ArgumentDescription> {
            { "data", "WebCore::LineData", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_StrokeLineWithColorAndThickness:
        return Vector<ArgumentDescription> {
            { "color", "WebCore::DisplayList::SetInlineStrokeColor", nullptr, false },
            { "thickness", "float", nullptr, false },
            { "data", "WebCore::LineData", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_StrokeArc:
        return Vector<ArgumentDescription> {
            { "data", "WebCore::ArcData", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_StrokeQuadCurve:
        return Vector<ArgumentDescription> {
            { "data", "WebCore::QuadCurveData", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_StrokeBezierCurve:
        return Vector<ArgumentDescription> {
            { "data", "WebCore::BezierCurveData", nullptr, false },
        };
#endif
    case MessageName::RemoteDisplayListRecorder_StrokePath:
        return Vector<ArgumentDescription> {
            { "path", "WebCore::Path", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_StrokeEllipse:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::FloatRect", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_ClearRect:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::FloatRect", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_DrawControlPart:
        return Vector<ArgumentDescription> {
            { "part", "Ref<WebCore::ControlPart>", nullptr, false },
            { "borderRect", "WebCore::FloatRoundedRect", nullptr, false },
            { "deviceScaleFactor", "float", nullptr, false },
            { "style", "WebCore::ControlStyle", nullptr, false },
        };
#if USE(CG)
    case MessageName::RemoteDisplayListRecorder_ApplyStrokePattern:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteDisplayListRecorder_ApplyFillPattern:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::RemoteDisplayListRecorder_ApplyDeviceScaleFactor:
        return Vector<ArgumentDescription> {
            { "scaleFactor", "float", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_FlushContext:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::DisplayListRecorderFlushIdentifier", nullptr, false },
        };
#if PLATFORM(COCOA) && ENABLE(VIDEO)
    case MessageName::RemoteDisplayListRecorder_PaintVideoFrame:
        return Vector<ArgumentDescription> {
            { "frame", "WebKit::SharedVideoFrame", nullptr, false },
            { "rect", "WebCore::FloatRect", nullptr, false },
            { "shouldDiscardAlpha", "bool", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_SetSharedVideoFrameSemaphore:
        return Vector<ArgumentDescription> {
            { "semaphore", "IPC::Semaphore", nullptr, false },
        };
    case MessageName::RemoteDisplayListRecorder_SetSharedVideoFrameMemory:
        return Vector<ArgumentDescription> {
            { "storageHandle", "WebKit::SharedMemory::Handle", nullptr, false },
        };
#endif
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(WEBGL)
    case MessageName::RemoteGraphicsContextGL_Reshape:
        return Vector<ArgumentDescription> {
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
        };
#if PLATFORM(COCOA)
    case MessageName::RemoteGraphicsContextGL_PrepareForDisplay:
        return Vector<ArgumentDescription> { };
#endif
#if USE(GRAPHICS_LAYER_WC)
    case MessageName::RemoteGraphicsContextGL_PrepareForDisplay:
        return Vector<ArgumentDescription> { };
#endif
#if USE(GBM)
    case MessageName::RemoteGraphicsContextGL_PrepareForDisplay:
        return Vector<ArgumentDescription> { };
#endif
#if !PLATFORM(COCOA) && !USE(GRAPHICS_LAYER_WC) && !USE(GBM)
    case MessageName::RemoteGraphicsContextGL_PrepareForDisplay:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::RemoteGraphicsContextGL_EnsureExtensionEnabled:
        return Vector<ArgumentDescription> {
            { "extension", "String", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_MarkContextChanged:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteGraphicsContextGL_SynthesizeGLError:
        return Vector<ArgumentDescription> {
            { "error", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetError:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteGraphicsContextGL_PaintRenderingResultsToCanvas:
        return Vector<ArgumentDescription> {
            { "imageBuffer", "WebCore::RenderingResourceIdentifier", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_PaintCompositedResultsToCanvas:
        return Vector<ArgumentDescription> {
            { "imageBuffer", "WebCore::RenderingResourceIdentifier", nullptr, false },
        };
#if ENABLE(VIDEO) && PLATFORM(COCOA)
    case MessageName::RemoteGraphicsContextGL_CopyTextureFromVideoFrame:
        return Vector<ArgumentDescription> {
            { "frame", "WebKit::SharedVideoFrame", nullptr, false },
            { "texture", "uint32_t", nullptr, false },
            { "target", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
            { "internalFormat", "uint32_t", nullptr, false },
            { "format", "uint32_t", nullptr, false },
            { "type", "uint32_t", nullptr, false },
            { "premultiplyAlpha", "bool", nullptr, false },
            { "flipY", "bool", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_SetSharedVideoFrameSemaphore:
        return Vector<ArgumentDescription> {
            { "semaphore", "IPC::Semaphore", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_SetSharedVideoFrameMemory:
        return Vector<ArgumentDescription> {
            { "storageHandle", "WebKit::SharedMemory::Handle", nullptr, false },
        };
#endif
#if ENABLE(MEDIA_STREAM)
    case MessageName::RemoteGraphicsContextGL_PaintCompositedResultsToVideoFrame:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::RemoteGraphicsContextGL_SimulateEventForTesting:
        return Vector<ArgumentDescription> {
            { "event", "WebCore::GraphicsContextGL::SimulatedEventForTesting", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_ReadnPixels0:
        return Vector<ArgumentDescription> {
            { "x", "int32_t", nullptr, false },
            { "y", "int32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
            { "format", "uint32_t", nullptr, false },
            { "type", "uint32_t", nullptr, false },
            { "data", "IPC::ArrayReference<uint8_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_ReadnPixels1:
        return Vector<ArgumentDescription> {
            { "x", "int32_t", nullptr, false },
            { "y", "int32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
            { "format", "uint32_t", nullptr, false },
            { "type", "uint32_t", nullptr, false },
            { "offset", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_ReadnPixels2:
        return Vector<ArgumentDescription> {
            { "x", "int32_t", nullptr, false },
            { "y", "int32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
            { "format", "uint32_t", nullptr, false },
            { "type", "uint32_t", nullptr, false },
            { "handle", "WebKit::SharedMemory::Handle", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_MultiDrawArraysANGLE:
        return Vector<ArgumentDescription> {
            { "mode", "uint32_t", nullptr, false },
            { "firstsAndCounts", "IPC::ArrayReferenceTuple<int32_t, int32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_MultiDrawArraysInstancedANGLE:
        return Vector<ArgumentDescription> {
            { "mode", "uint32_t", nullptr, false },
            { "firstsCountsAandInstanceCounts", "IPC::ArrayReferenceTuple<int32_t, int32_t, int32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_MultiDrawElementsANGLE:
        return Vector<ArgumentDescription> {
            { "mode", "uint32_t", nullptr, false },
            { "countsAndOffsets", "IPC::ArrayReferenceTuple<int32_t, int32_t>", nullptr, false },
            { "type", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_MultiDrawElementsInstancedANGLE:
        return Vector<ArgumentDescription> {
            { "mode", "uint32_t", nullptr, false },
            { "countsOffsetsAndInstanceCounts", "IPC::ArrayReferenceTuple<int32_t, int32_t, int32_t>", nullptr, false },
            { "type", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_MultiDrawArraysInstancedBaseInstanceANGLE:
        return Vector<ArgumentDescription> {
            { "mode", "uint32_t", nullptr, false },
            { "firstsCountsInstanceCountsAndBaseInstances", "IPC::ArrayReferenceTuple<int32_t, int32_t, int32_t, uint32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_MultiDrawElementsInstancedBaseVertexBaseInstanceANGLE:
        return Vector<ArgumentDescription> {
            { "mode", "uint32_t", nullptr, false },
            { "countsOffsetsInstanceCountsBaseVerticesAndBaseInstances", "IPC::ArrayReferenceTuple<int32_t, int32_t, int32_t, int32_t, uint32_t>", nullptr, false },
            { "type", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_MoveErrorsToSyntheticErrorList:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteGraphicsContextGL_ActiveTexture:
        return Vector<ArgumentDescription> {
            { "texture", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_AttachShader:
        return Vector<ArgumentDescription> {
            { "program", "uint32_t", nullptr, false },
            { "shader", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BindAttribLocation:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
            { "index", "uint32_t", nullptr, false },
            { "name", "String", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BindBuffer:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "arg1", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BindFramebuffer:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "arg1", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BindRenderbuffer:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "arg1", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BindTexture:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "arg1", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BlendColor:
        return Vector<ArgumentDescription> {
            { "red", "float", nullptr, false },
            { "green", "float", nullptr, false },
            { "blue", "float", nullptr, false },
            { "alpha", "float", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BlendEquation:
        return Vector<ArgumentDescription> {
            { "mode", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BlendEquationSeparate:
        return Vector<ArgumentDescription> {
            { "modeRGB", "uint32_t", nullptr, false },
            { "modeAlpha", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BlendFunc:
        return Vector<ArgumentDescription> {
            { "sfactor", "uint32_t", nullptr, false },
            { "dfactor", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BlendFuncSeparate:
        return Vector<ArgumentDescription> {
            { "srcRGB", "uint32_t", nullptr, false },
            { "dstRGB", "uint32_t", nullptr, false },
            { "srcAlpha", "uint32_t", nullptr, false },
            { "dstAlpha", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CheckFramebufferStatus:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Clear:
        return Vector<ArgumentDescription> {
            { "mask", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_ClearColor:
        return Vector<ArgumentDescription> {
            { "red", "float", nullptr, false },
            { "green", "float", nullptr, false },
            { "blue", "float", nullptr, false },
            { "alpha", "float", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_ClearDepth:
        return Vector<ArgumentDescription> {
            { "depth", "float", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_ClearStencil:
        return Vector<ArgumentDescription> {
            { "s", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_ColorMask:
        return Vector<ArgumentDescription> {
            { "red", "bool", nullptr, false },
            { "green", "bool", nullptr, false },
            { "blue", "bool", nullptr, false },
            { "alpha", "bool", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CompileShader:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CopyTexImage2D:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
            { "internalformat", "uint32_t", nullptr, false },
            { "x", "int32_t", nullptr, false },
            { "y", "int32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
            { "border", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CopyTexSubImage2D:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
            { "xoffset", "int32_t", nullptr, false },
            { "yoffset", "int32_t", nullptr, false },
            { "x", "int32_t", nullptr, false },
            { "y", "int32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CreateBuffer:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteGraphicsContextGL_CreateFramebuffer:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteGraphicsContextGL_CreateProgram:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteGraphicsContextGL_CreateRenderbuffer:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteGraphicsContextGL_CreateShader:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CreateTexture:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteGraphicsContextGL_CullFace:
        return Vector<ArgumentDescription> {
            { "mode", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DeleteBuffer:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DeleteFramebuffer:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DeleteProgram:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DeleteRenderbuffer:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DeleteShader:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DeleteTexture:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DepthFunc:
        return Vector<ArgumentDescription> {
            { "func", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DepthMask:
        return Vector<ArgumentDescription> {
            { "flag", "bool", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DepthRange:
        return Vector<ArgumentDescription> {
            { "zNear", "float", nullptr, false },
            { "zFar", "float", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DetachShader:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
            { "arg1", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Disable:
        return Vector<ArgumentDescription> {
            { "cap", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DisableVertexAttribArray:
        return Vector<ArgumentDescription> {
            { "index", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DrawArrays:
        return Vector<ArgumentDescription> {
            { "mode", "uint32_t", nullptr, false },
            { "first", "int32_t", nullptr, false },
            { "count", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DrawElements:
        return Vector<ArgumentDescription> {
            { "mode", "uint32_t", nullptr, false },
            { "count", "int32_t", nullptr, false },
            { "type", "uint32_t", nullptr, false },
            { "offset", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Enable:
        return Vector<ArgumentDescription> {
            { "cap", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_EnableVertexAttribArray:
        return Vector<ArgumentDescription> {
            { "index", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Finish:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteGraphicsContextGL_Flush:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteGraphicsContextGL_FramebufferRenderbuffer:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "attachment", "uint32_t", nullptr, false },
            { "renderbuffertarget", "uint32_t", nullptr, false },
            { "arg3", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_FramebufferTexture2D:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "attachment", "uint32_t", nullptr, false },
            { "textarget", "uint32_t", nullptr, false },
            { "arg3", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_FrontFace:
        return Vector<ArgumentDescription> {
            { "mode", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GenerateMipmap:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetActiveAttrib:
        return Vector<ArgumentDescription> {
            { "program", "uint32_t", nullptr, false },
            { "index", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetActiveUniform:
        return Vector<ArgumentDescription> {
            { "program", "uint32_t", nullptr, false },
            { "index", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetAttribLocation:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
            { "name", "String", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetBufferParameteri:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetString:
        return Vector<ArgumentDescription> {
            { "name", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetFloatv:
        return Vector<ArgumentDescription> {
            { "pname", "uint32_t", nullptr, false },
            { "valueSize", "size_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetIntegerv:
        return Vector<ArgumentDescription> {
            { "pname", "uint32_t", nullptr, false },
            { "valueSize", "size_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetIntegeri_v:
        return Vector<ArgumentDescription> {
            { "pname", "uint32_t", nullptr, false },
            { "index", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetInteger64:
        return Vector<ArgumentDescription> {
            { "pname", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetInteger64i:
        return Vector<ArgumentDescription> {
            { "pname", "uint32_t", nullptr, false },
            { "index", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetProgrami:
        return Vector<ArgumentDescription> {
            { "program", "uint32_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetBooleanv:
        return Vector<ArgumentDescription> {
            { "pname", "uint32_t", nullptr, false },
            { "valueSize", "size_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetFramebufferAttachmentParameteri:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "attachment", "uint32_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetProgramInfoLog:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetRenderbufferParameteri:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetShaderi:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetShaderInfoLog:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetShaderPrecisionFormat:
        return Vector<ArgumentDescription> {
            { "shaderType", "uint32_t", nullptr, false },
            { "precisionType", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetShaderSource:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetTexParameterf:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetTexParameteri:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetUniformfv:
        return Vector<ArgumentDescription> {
            { "program", "uint32_t", nullptr, false },
            { "location", "int32_t", nullptr, false },
            { "valueSize", "size_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetUniformiv:
        return Vector<ArgumentDescription> {
            { "program", "uint32_t", nullptr, false },
            { "location", "int32_t", nullptr, false },
            { "valueSize", "size_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetUniformuiv:
        return Vector<ArgumentDescription> {
            { "program", "uint32_t", nullptr, false },
            { "location", "int32_t", nullptr, false },
            { "valueSize", "size_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetUniformLocation:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
            { "name", "String", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetVertexAttribOffset:
        return Vector<ArgumentDescription> {
            { "index", "uint32_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Hint:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "mode", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsBuffer:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsEnabled:
        return Vector<ArgumentDescription> {
            { "cap", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsFramebuffer:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsProgram:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsRenderbuffer:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsShader:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsTexture:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_LineWidth:
        return Vector<ArgumentDescription> {
            { "arg0", "float", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_LinkProgram:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_PixelStorei:
        return Vector<ArgumentDescription> {
            { "pname", "uint32_t", nullptr, false },
            { "param", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_PolygonOffset:
        return Vector<ArgumentDescription> {
            { "factor", "float", nullptr, false },
            { "units", "float", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_RenderbufferStorage:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "internalformat", "uint32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_SampleCoverage:
        return Vector<ArgumentDescription> {
            { "value", "float", nullptr, false },
            { "invert", "bool", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Scissor:
        return Vector<ArgumentDescription> {
            { "x", "int32_t", nullptr, false },
            { "y", "int32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_ShaderSource:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
            { "arg1", "String", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_StencilFunc:
        return Vector<ArgumentDescription> {
            { "func", "uint32_t", nullptr, false },
            { "ref", "int32_t", nullptr, false },
            { "mask", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_StencilFuncSeparate:
        return Vector<ArgumentDescription> {
            { "face", "uint32_t", nullptr, false },
            { "func", "uint32_t", nullptr, false },
            { "ref", "int32_t", nullptr, false },
            { "mask", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_StencilMask:
        return Vector<ArgumentDescription> {
            { "mask", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_StencilMaskSeparate:
        return Vector<ArgumentDescription> {
            { "face", "uint32_t", nullptr, false },
            { "mask", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_StencilOp:
        return Vector<ArgumentDescription> {
            { "fail", "uint32_t", nullptr, false },
            { "zfail", "uint32_t", nullptr, false },
            { "zpass", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_StencilOpSeparate:
        return Vector<ArgumentDescription> {
            { "face", "uint32_t", nullptr, false },
            { "fail", "uint32_t", nullptr, false },
            { "zfail", "uint32_t", nullptr, false },
            { "zpass", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_TexParameterf:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
            { "param", "float", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_TexParameteri:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
            { "param", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform1f:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "x", "float", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform1fv:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "v", "IPC::ArrayReference<float>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform1i:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "x", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform1iv:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "v", "IPC::ArrayReference<int32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform2f:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "x", "float", nullptr, false },
            { "y", "float", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform2fv:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "v", "IPC::ArrayReference<float>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform2i:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "x", "int32_t", nullptr, false },
            { "y", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform2iv:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "v", "IPC::ArrayReference<int32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform3f:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "x", "float", nullptr, false },
            { "y", "float", nullptr, false },
            { "z", "float", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform3fv:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "v", "IPC::ArrayReference<float>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform3i:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "x", "int32_t", nullptr, false },
            { "y", "int32_t", nullptr, false },
            { "z", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform3iv:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "v", "IPC::ArrayReference<int32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform4f:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "x", "float", nullptr, false },
            { "y", "float", nullptr, false },
            { "z", "float", nullptr, false },
            { "w", "float", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform4fv:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "v", "IPC::ArrayReference<float>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform4i:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "x", "int32_t", nullptr, false },
            { "y", "int32_t", nullptr, false },
            { "z", "int32_t", nullptr, false },
            { "w", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform4iv:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "v", "IPC::ArrayReference<int32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_UniformMatrix2fv:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "transpose", "bool", nullptr, false },
            { "value", "IPC::ArrayReference<float>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_UniformMatrix3fv:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "transpose", "bool", nullptr, false },
            { "value", "IPC::ArrayReference<float>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_UniformMatrix4fv:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "transpose", "bool", nullptr, false },
            { "value", "IPC::ArrayReference<float>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_UseProgram:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_ValidateProgram:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_VertexAttrib1f:
        return Vector<ArgumentDescription> {
            { "index", "uint32_t", nullptr, false },
            { "x", "float", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_VertexAttrib1fv:
        return Vector<ArgumentDescription> {
            { "index", "uint32_t", nullptr, false },
            { "values", "IPC::ArrayReference<float, 1>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_VertexAttrib2f:
        return Vector<ArgumentDescription> {
            { "index", "uint32_t", nullptr, false },
            { "x", "float", nullptr, false },
            { "y", "float", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_VertexAttrib2fv:
        return Vector<ArgumentDescription> {
            { "index", "uint32_t", nullptr, false },
            { "values", "IPC::ArrayReference<float, 2>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_VertexAttrib3f:
        return Vector<ArgumentDescription> {
            { "index", "uint32_t", nullptr, false },
            { "x", "float", nullptr, false },
            { "y", "float", nullptr, false },
            { "z", "float", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_VertexAttrib3fv:
        return Vector<ArgumentDescription> {
            { "index", "uint32_t", nullptr, false },
            { "values", "IPC::ArrayReference<float, 3>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_VertexAttrib4f:
        return Vector<ArgumentDescription> {
            { "index", "uint32_t", nullptr, false },
            { "x", "float", nullptr, false },
            { "y", "float", nullptr, false },
            { "z", "float", nullptr, false },
            { "w", "float", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_VertexAttrib4fv:
        return Vector<ArgumentDescription> {
            { "index", "uint32_t", nullptr, false },
            { "values", "IPC::ArrayReference<float, 4>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_VertexAttribPointer:
        return Vector<ArgumentDescription> {
            { "index", "uint32_t", nullptr, false },
            { "size", "int32_t", nullptr, false },
            { "type", "uint32_t", nullptr, false },
            { "normalized", "bool", nullptr, false },
            { "stride", "int32_t", nullptr, false },
            { "offset", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Viewport:
        return Vector<ArgumentDescription> {
            { "x", "int32_t", nullptr, false },
            { "y", "int32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BufferData0:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "arg1", "uint64_t", nullptr, false },
            { "usage", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BufferData1:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "data", "IPC::ArrayReference<uint8_t>", nullptr, false },
            { "usage", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BufferSubData:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "offset", "uint64_t", nullptr, false },
            { "data", "IPC::ArrayReference<uint8_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_TexImage2D0:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
            { "internalformat", "uint32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
            { "border", "int32_t", nullptr, false },
            { "format", "uint32_t", nullptr, false },
            { "type", "uint32_t", nullptr, false },
            { "pixels", "IPC::ArrayReference<uint8_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_TexImage2D1:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
            { "internalformat", "uint32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
            { "border", "int32_t", nullptr, false },
            { "format", "uint32_t", nullptr, false },
            { "type", "uint32_t", nullptr, false },
            { "offset", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_TexSubImage2D0:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
            { "xoffset", "int32_t", nullptr, false },
            { "yoffset", "int32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
            { "format", "uint32_t", nullptr, false },
            { "type", "uint32_t", nullptr, false },
            { "pixels", "IPC::ArrayReference<uint8_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_TexSubImage2D1:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
            { "xoffset", "int32_t", nullptr, false },
            { "yoffset", "int32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
            { "format", "uint32_t", nullptr, false },
            { "type", "uint32_t", nullptr, false },
            { "offset", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CompressedTexImage2D0:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
            { "internalformat", "uint32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
            { "border", "int32_t", nullptr, false },
            { "imageSize", "int32_t", nullptr, false },
            { "data", "IPC::ArrayReference<uint8_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CompressedTexImage2D1:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
            { "internalformat", "uint32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
            { "border", "int32_t", nullptr, false },
            { "imageSize", "int32_t", nullptr, false },
            { "offset", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CompressedTexSubImage2D0:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
            { "xoffset", "int32_t", nullptr, false },
            { "yoffset", "int32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
            { "format", "uint32_t", nullptr, false },
            { "imageSize", "int32_t", nullptr, false },
            { "data", "IPC::ArrayReference<uint8_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CompressedTexSubImage2D1:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
            { "xoffset", "int32_t", nullptr, false },
            { "yoffset", "int32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
            { "format", "uint32_t", nullptr, false },
            { "imageSize", "int32_t", nullptr, false },
            { "offset", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DrawArraysInstanced:
        return Vector<ArgumentDescription> {
            { "mode", "uint32_t", nullptr, false },
            { "first", "int32_t", nullptr, false },
            { "count", "int32_t", nullptr, false },
            { "primcount", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DrawElementsInstanced:
        return Vector<ArgumentDescription> {
            { "mode", "uint32_t", nullptr, false },
            { "count", "int32_t", nullptr, false },
            { "type", "uint32_t", nullptr, false },
            { "offset", "uint64_t", nullptr, false },
            { "primcount", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_VertexAttribDivisor:
        return Vector<ArgumentDescription> {
            { "index", "uint32_t", nullptr, false },
            { "divisor", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CreateVertexArray:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteGraphicsContextGL_DeleteVertexArray:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsVertexArray:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BindVertexArray:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CopyBufferSubData:
        return Vector<ArgumentDescription> {
            { "readTarget", "uint32_t", nullptr, false },
            { "writeTarget", "uint32_t", nullptr, false },
            { "readOffset", "uint64_t", nullptr, false },
            { "writeOffset", "uint64_t", nullptr, false },
            { "arg4", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetBufferSubData:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "offset", "uint64_t", nullptr, false },
            { "dataSize", "size_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BlitFramebuffer:
        return Vector<ArgumentDescription> {
            { "srcX0", "int32_t", nullptr, false },
            { "srcY0", "int32_t", nullptr, false },
            { "srcX1", "int32_t", nullptr, false },
            { "srcY1", "int32_t", nullptr, false },
            { "dstX0", "int32_t", nullptr, false },
            { "dstY0", "int32_t", nullptr, false },
            { "dstX1", "int32_t", nullptr, false },
            { "dstY1", "int32_t", nullptr, false },
            { "mask", "uint32_t", nullptr, false },
            { "filter", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_FramebufferTextureLayer:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "attachment", "uint32_t", nullptr, false },
            { "texture", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
            { "layer", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_InvalidateFramebuffer:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "attachments", "IPC::ArrayReference<uint32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_InvalidateSubFramebuffer:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "attachments", "IPC::ArrayReference<uint32_t>", nullptr, false },
            { "x", "int32_t", nullptr, false },
            { "y", "int32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_ReadBuffer:
        return Vector<ArgumentDescription> {
            { "src", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_RenderbufferStorageMultisample:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "samples", "int32_t", nullptr, false },
            { "internalformat", "uint32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_TexStorage2D:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "levels", "int32_t", nullptr, false },
            { "internalformat", "uint32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_TexStorage3D:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "levels", "int32_t", nullptr, false },
            { "internalformat", "uint32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
            { "depth", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_TexImage3D0:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
            { "internalformat", "int32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
            { "depth", "int32_t", nullptr, false },
            { "border", "int32_t", nullptr, false },
            { "format", "uint32_t", nullptr, false },
            { "type", "uint32_t", nullptr, false },
            { "pixels", "IPC::ArrayReference<uint8_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_TexImage3D1:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
            { "internalformat", "int32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
            { "depth", "int32_t", nullptr, false },
            { "border", "int32_t", nullptr, false },
            { "format", "uint32_t", nullptr, false },
            { "type", "uint32_t", nullptr, false },
            { "offset", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_TexSubImage3D0:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
            { "xoffset", "int32_t", nullptr, false },
            { "yoffset", "int32_t", nullptr, false },
            { "zoffset", "int32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
            { "depth", "int32_t", nullptr, false },
            { "format", "uint32_t", nullptr, false },
            { "type", "uint32_t", nullptr, false },
            { "pixels", "IPC::ArrayReference<uint8_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_TexSubImage3D1:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
            { "xoffset", "int32_t", nullptr, false },
            { "yoffset", "int32_t", nullptr, false },
            { "zoffset", "int32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
            { "depth", "int32_t", nullptr, false },
            { "format", "uint32_t", nullptr, false },
            { "type", "uint32_t", nullptr, false },
            { "offset", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CopyTexSubImage3D:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
            { "xoffset", "int32_t", nullptr, false },
            { "yoffset", "int32_t", nullptr, false },
            { "zoffset", "int32_t", nullptr, false },
            { "x", "int32_t", nullptr, false },
            { "y", "int32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CompressedTexImage3D0:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
            { "internalformat", "uint32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
            { "depth", "int32_t", nullptr, false },
            { "border", "int32_t", nullptr, false },
            { "imageSize", "int32_t", nullptr, false },
            { "data", "IPC::ArrayReference<uint8_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CompressedTexImage3D1:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
            { "internalformat", "uint32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
            { "depth", "int32_t", nullptr, false },
            { "border", "int32_t", nullptr, false },
            { "imageSize", "int32_t", nullptr, false },
            { "offset", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CompressedTexSubImage3D0:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
            { "xoffset", "int32_t", nullptr, false },
            { "yoffset", "int32_t", nullptr, false },
            { "zoffset", "int32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
            { "depth", "int32_t", nullptr, false },
            { "format", "uint32_t", nullptr, false },
            { "imageSize", "int32_t", nullptr, false },
            { "data", "IPC::ArrayReference<uint8_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CompressedTexSubImage3D1:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "level", "int32_t", nullptr, false },
            { "xoffset", "int32_t", nullptr, false },
            { "yoffset", "int32_t", nullptr, false },
            { "zoffset", "int32_t", nullptr, false },
            { "width", "int32_t", nullptr, false },
            { "height", "int32_t", nullptr, false },
            { "depth", "int32_t", nullptr, false },
            { "format", "uint32_t", nullptr, false },
            { "imageSize", "int32_t", nullptr, false },
            { "offset", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetFragDataLocation:
        return Vector<ArgumentDescription> {
            { "program", "uint32_t", nullptr, false },
            { "name", "String", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform1ui:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "v0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform2ui:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "v0", "uint32_t", nullptr, false },
            { "v1", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform3ui:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "v0", "uint32_t", nullptr, false },
            { "v1", "uint32_t", nullptr, false },
            { "v2", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform4ui:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "v0", "uint32_t", nullptr, false },
            { "v1", "uint32_t", nullptr, false },
            { "v2", "uint32_t", nullptr, false },
            { "v3", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform1uiv:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "data", "IPC::ArrayReference<uint32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform2uiv:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "data", "IPC::ArrayReference<uint32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform3uiv:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "data", "IPC::ArrayReference<uint32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_Uniform4uiv:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "data", "IPC::ArrayReference<uint32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_UniformMatrix2x3fv:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "transpose", "bool", nullptr, false },
            { "data", "IPC::ArrayReference<float>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_UniformMatrix3x2fv:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "transpose", "bool", nullptr, false },
            { "data", "IPC::ArrayReference<float>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_UniformMatrix2x4fv:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "transpose", "bool", nullptr, false },
            { "data", "IPC::ArrayReference<float>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_UniformMatrix4x2fv:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "transpose", "bool", nullptr, false },
            { "data", "IPC::ArrayReference<float>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_UniformMatrix3x4fv:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "transpose", "bool", nullptr, false },
            { "data", "IPC::ArrayReference<float>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_UniformMatrix4x3fv:
        return Vector<ArgumentDescription> {
            { "location", "int32_t", nullptr, false },
            { "transpose", "bool", nullptr, false },
            { "data", "IPC::ArrayReference<float>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_VertexAttribI4i:
        return Vector<ArgumentDescription> {
            { "index", "uint32_t", nullptr, false },
            { "x", "int32_t", nullptr, false },
            { "y", "int32_t", nullptr, false },
            { "z", "int32_t", nullptr, false },
            { "w", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_VertexAttribI4iv:
        return Vector<ArgumentDescription> {
            { "index", "uint32_t", nullptr, false },
            { "values", "IPC::ArrayReference<int32_t, 4>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_VertexAttribI4ui:
        return Vector<ArgumentDescription> {
            { "index", "uint32_t", nullptr, false },
            { "x", "uint32_t", nullptr, false },
            { "y", "uint32_t", nullptr, false },
            { "z", "uint32_t", nullptr, false },
            { "w", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_VertexAttribI4uiv:
        return Vector<ArgumentDescription> {
            { "index", "uint32_t", nullptr, false },
            { "values", "IPC::ArrayReference<uint32_t, 4>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_VertexAttribIPointer:
        return Vector<ArgumentDescription> {
            { "index", "uint32_t", nullptr, false },
            { "size", "int32_t", nullptr, false },
            { "type", "uint32_t", nullptr, false },
            { "stride", "int32_t", nullptr, false },
            { "offset", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DrawRangeElements:
        return Vector<ArgumentDescription> {
            { "mode", "uint32_t", nullptr, false },
            { "start", "uint32_t", nullptr, false },
            { "end", "uint32_t", nullptr, false },
            { "count", "int32_t", nullptr, false },
            { "type", "uint32_t", nullptr, false },
            { "offset", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DrawBuffers:
        return Vector<ArgumentDescription> {
            { "bufs", "IPC::ArrayReference<uint32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_ClearBufferiv:
        return Vector<ArgumentDescription> {
            { "buffer", "uint32_t", nullptr, false },
            { "drawbuffer", "int32_t", nullptr, false },
            { "values", "IPC::ArrayReference<int32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_ClearBufferuiv:
        return Vector<ArgumentDescription> {
            { "buffer", "uint32_t", nullptr, false },
            { "drawbuffer", "int32_t", nullptr, false },
            { "values", "IPC::ArrayReference<uint32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_ClearBufferfv:
        return Vector<ArgumentDescription> {
            { "buffer", "uint32_t", nullptr, false },
            { "drawbuffer", "int32_t", nullptr, false },
            { "values", "IPC::ArrayReference<float>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_ClearBufferfi:
        return Vector<ArgumentDescription> {
            { "buffer", "uint32_t", nullptr, false },
            { "drawbuffer", "int32_t", nullptr, false },
            { "depth", "float", nullptr, false },
            { "stencil", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CreateQuery:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteGraphicsContextGL_DeleteQuery:
        return Vector<ArgumentDescription> {
            { "query", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsQuery:
        return Vector<ArgumentDescription> {
            { "query", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BeginQuery:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "query", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_EndQuery:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetQuery:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetQueryObjectui:
        return Vector<ArgumentDescription> {
            { "query", "uint32_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CreateSampler:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteGraphicsContextGL_DeleteSampler:
        return Vector<ArgumentDescription> {
            { "sampler", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsSampler:
        return Vector<ArgumentDescription> {
            { "sampler", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BindSampler:
        return Vector<ArgumentDescription> {
            { "unit", "uint32_t", nullptr, false },
            { "sampler", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_SamplerParameteri:
        return Vector<ArgumentDescription> {
            { "sampler", "uint32_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
            { "param", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_SamplerParameterf:
        return Vector<ArgumentDescription> {
            { "sampler", "uint32_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
            { "param", "float", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetSamplerParameterf:
        return Vector<ArgumentDescription> {
            { "sampler", "uint32_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetSamplerParameteri:
        return Vector<ArgumentDescription> {
            { "sampler", "uint32_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_FenceSync:
        return Vector<ArgumentDescription> {
            { "condition", "uint32_t", nullptr, false },
            { "flags", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsSync:
        return Vector<ArgumentDescription> {
            { "arg0", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DeleteSync:
        return Vector<ArgumentDescription> {
            { "arg0", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_ClientWaitSync:
        return Vector<ArgumentDescription> {
            { "arg0", "uint64_t", nullptr, false },
            { "flags", "uint32_t", nullptr, false },
            { "timeout", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_WaitSync:
        return Vector<ArgumentDescription> {
            { "arg0", "uint64_t", nullptr, false },
            { "flags", "uint32_t", nullptr, false },
            { "timeout", "int64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetSynci:
        return Vector<ArgumentDescription> {
            { "arg0", "uint64_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CreateTransformFeedback:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteGraphicsContextGL_DeleteTransformFeedback:
        return Vector<ArgumentDescription> {
            { "id", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsTransformFeedback:
        return Vector<ArgumentDescription> {
            { "id", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BindTransformFeedback:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "id", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BeginTransformFeedback:
        return Vector<ArgumentDescription> {
            { "primitiveMode", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_EndTransformFeedback:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteGraphicsContextGL_TransformFeedbackVaryings:
        return Vector<ArgumentDescription> {
            { "program", "uint32_t", nullptr, false },
            { "varyings", "Vector<String>", nullptr, false },
            { "bufferMode", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetTransformFeedbackVarying:
        return Vector<ArgumentDescription> {
            { "program", "uint32_t", nullptr, false },
            { "index", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_PauseTransformFeedback:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteGraphicsContextGL_ResumeTransformFeedback:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteGraphicsContextGL_BindBufferBase:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "index", "uint32_t", nullptr, false },
            { "buffer", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BindBufferRange:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "index", "uint32_t", nullptr, false },
            { "buffer", "uint32_t", nullptr, false },
            { "offset", "uint64_t", nullptr, false },
            { "arg4", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetUniformIndices:
        return Vector<ArgumentDescription> {
            { "program", "uint32_t", nullptr, false },
            { "uniformNames", "Vector<String>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetActiveUniforms:
        return Vector<ArgumentDescription> {
            { "program", "uint32_t", nullptr, false },
            { "uniformIndices", "Vector<uint32_t>", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetUniformBlockIndex:
        return Vector<ArgumentDescription> {
            { "program", "uint32_t", nullptr, false },
            { "uniformBlockName", "String", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetActiveUniformBlockName:
        return Vector<ArgumentDescription> {
            { "program", "uint32_t", nullptr, false },
            { "uniformBlockIndex", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_UniformBlockBinding:
        return Vector<ArgumentDescription> {
            { "program", "uint32_t", nullptr, false },
            { "uniformBlockIndex", "uint32_t", nullptr, false },
            { "uniformBlockBinding", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetActiveUniformBlockiv:
        return Vector<ArgumentDescription> {
            { "program", "uint32_t", nullptr, false },
            { "uniformBlockIndex", "uint32_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
            { "paramsSize", "size_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetTranslatedShaderSourceANGLE:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DrawBuffersEXT:
        return Vector<ArgumentDescription> {
            { "bufs", "IPC::ArrayReference<uint32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CreateQueryEXT:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteGraphicsContextGL_DeleteQueryEXT:
        return Vector<ArgumentDescription> {
            { "query", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsQueryEXT:
        return Vector<ArgumentDescription> {
            { "query", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BeginQueryEXT:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "query", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_EndQueryEXT:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_QueryCounterEXT:
        return Vector<ArgumentDescription> {
            { "query", "uint32_t", nullptr, false },
            { "target", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetQueryiEXT:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetQueryObjectiEXT:
        return Vector<ArgumentDescription> {
            { "query", "uint32_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetQueryObjectui64EXT:
        return Vector<ArgumentDescription> {
            { "query", "uint32_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetInteger64EXT:
        return Vector<ArgumentDescription> {
            { "pname", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_EnableiOES:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "index", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DisableiOES:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "index", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BlendEquationiOES:
        return Vector<ArgumentDescription> {
            { "buf", "uint32_t", nullptr, false },
            { "mode", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BlendEquationSeparateiOES:
        return Vector<ArgumentDescription> {
            { "buf", "uint32_t", nullptr, false },
            { "modeRGB", "uint32_t", nullptr, false },
            { "modeAlpha", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BlendFunciOES:
        return Vector<ArgumentDescription> {
            { "buf", "uint32_t", nullptr, false },
            { "src", "uint32_t", nullptr, false },
            { "dst", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_BlendFuncSeparateiOES:
        return Vector<ArgumentDescription> {
            { "buf", "uint32_t", nullptr, false },
            { "srcRGB", "uint32_t", nullptr, false },
            { "dstRGB", "uint32_t", nullptr, false },
            { "srcAlpha", "uint32_t", nullptr, false },
            { "dstAlpha", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_ColorMaskiOES:
        return Vector<ArgumentDescription> {
            { "buf", "uint32_t", nullptr, false },
            { "red", "bool", nullptr, false },
            { "green", "bool", nullptr, false },
            { "blue", "bool", nullptr, false },
            { "alpha", "bool", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DrawArraysInstancedBaseInstanceANGLE:
        return Vector<ArgumentDescription> {
            { "mode", "uint32_t", nullptr, false },
            { "first", "int32_t", nullptr, false },
            { "count", "int32_t", nullptr, false },
            { "instanceCount", "int32_t", nullptr, false },
            { "baseInstance", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_DrawElementsInstancedBaseVertexBaseInstanceANGLE:
        return Vector<ArgumentDescription> {
            { "mode", "uint32_t", nullptr, false },
            { "count", "int32_t", nullptr, false },
            { "type", "uint32_t", nullptr, false },
            { "offset", "uint64_t", nullptr, false },
            { "instanceCount", "int32_t", nullptr, false },
            { "baseVertex", "int32_t", nullptr, false },
            { "baseInstance", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_ProvokingVertexANGLE:
        return Vector<ArgumentDescription> {
            { "provokeMode", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetInternalformativ:
        return Vector<ArgumentDescription> {
            { "target", "uint32_t", nullptr, false },
            { "internalformat", "uint32_t", nullptr, false },
            { "pname", "uint32_t", nullptr, false },
            { "paramsSize", "size_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_SetDrawingBufferColorSpace:
        return Vector<ArgumentDescription> {
            { "arg0", "WebCore::DestinationColorSpace", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_PaintRenderingResultsToPixelBuffer:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteRenderingBackend_CreateImageBuffer:
        return Vector<ArgumentDescription> {
            { "logicalSize", "WebCore::FloatSize", nullptr, false },
            { "renderingMode", "WebCore::RenderingMode", nullptr, false },
            { "renderingPurpose", "WebCore::RenderingPurpose", nullptr, false },
            { "resolutionScale", "float", nullptr, false },
            { "colorSpace", "WebCore::DestinationColorSpace", nullptr, false },
            { "pixelFormat", "uint8_t", "WebCore::PixelFormat", false },
            { "renderingResourceIdentifier", "WebCore::RenderingResourceIdentifier", nullptr, false },
        };
    case MessageName::RemoteRenderingBackend_GetPixelBufferForImageBuffer:
        return Vector<ArgumentDescription> {
            { "imageBuffer", "WebCore::RenderingResourceIdentifier", nullptr, false },
            { "outputFormat", "WebCore::PixelBufferFormat", nullptr, false },
            { "srcRect", "WebCore::IntRect", nullptr, false },
        };
    case MessageName::RemoteRenderingBackend_GetPixelBufferForImageBufferWithNewMemory:
        return Vector<ArgumentDescription> {
            { "imageBuffer", "WebCore::RenderingResourceIdentifier", nullptr, false },
            { "handle", "WebKit::SharedMemory::Handle", nullptr, false },
            { "outputFormat", "WebCore::PixelBufferFormat", nullptr, false },
            { "srcRect", "WebCore::IntRect", nullptr, false },
        };
    case MessageName::RemoteRenderingBackend_DestroyGetPixelBufferSharedMemory:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteRenderingBackend_PutPixelBufferForImageBuffer:
        return Vector<ArgumentDescription> {
            { "imageBuffer", "WebCore::RenderingResourceIdentifier", nullptr, false },
            { "pixelBuffer", "Ref<WebCore::PixelBuffer>", nullptr, false },
            { "srcRect", "WebCore::IntRect", nullptr, false },
            { "destPoint", "WebCore::IntPoint", nullptr, false },
            { "destFormat", "uint8_t", "WebCore::AlphaPremultiplication", false },
        };
    case MessageName::RemoteRenderingBackend_GetShareableBitmapForImageBuffer:
        return Vector<ArgumentDescription> {
            { "imageBuffer", "WebCore::RenderingResourceIdentifier", nullptr, false },
            { "preserveResolution", "bool", "WebCore::PreserveResolution", false },
        };
    case MessageName::RemoteRenderingBackend_GetFilteredImageForImageBuffer:
        return Vector<ArgumentDescription> {
            { "imageBuffer", "WebCore::RenderingResourceIdentifier", nullptr, false },
            { "filter", "Ref<WebCore::Filter>", nullptr, false },
        };
    case MessageName::RemoteRenderingBackend_CacheNativeImage:
        return Vector<ArgumentDescription> {
            { "handle", "WebKit::ShareableBitmapHandle", nullptr, false },
            { "renderingResourceIdentifier", "WebCore::RenderingResourceIdentifier", nullptr, false },
        };
    case MessageName::RemoteRenderingBackend_CacheFont:
        return Vector<ArgumentDescription> {
            { "font", "Ref<WebCore::Font>", nullptr, false },
        };
    case MessageName::RemoteRenderingBackend_CacheDecomposedGlyphs:
        return Vector<ArgumentDescription> {
            { "decomposedGlyphs", "Ref<WebCore::DecomposedGlyphs>", nullptr, false },
        };
    case MessageName::RemoteRenderingBackend_ReleaseAllResources:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteRenderingBackend_ReleaseResource:
        return Vector<ArgumentDescription> {
            { "renderingResourceIdentifier", "WebCore::RenderingResourceIdentifier", nullptr, false },
        };
    case MessageName::RemoteRenderingBackend_PrepareBuffersForDisplay:
        return Vector<ArgumentDescription> {
            { "swapBuffersInput", "Vector<WebKit::PrepareBackingStoreBuffersInputData>", nullptr, false },
        };
    case MessageName::RemoteRenderingBackend_MarkSurfacesVolatile:
        return Vector<ArgumentDescription> {
            { "requestIdentifier", "WebKit::MarkSurfacesAsVolatileRequestIdentifier", nullptr, false },
            { "renderingResourceIdentifiers", "Vector<WebCore::RenderingResourceIdentifier>", nullptr, false },
        };
    case MessageName::RemoteRenderingBackend_FinalizeRenderingUpdate:
        return Vector<ArgumentDescription> {
            { "renderingUpdateID", "WebKit::RenderingUpdateID", nullptr, false },
        };
    case MessageName::RemoteRenderingBackend_MoveToSerializedBuffer:
        return Vector<ArgumentDescription> {
            { "sourceImageBuffer", "WebCore::RenderingResourceIdentifier", nullptr, false },
            { "destination", "WebKit::RemoteSerializedImageBufferWriteReference", nullptr, false },
        };
    case MessageName::RemoteRenderingBackend_MoveToImageBuffer:
        return Vector<ArgumentDescription> {
            { "source", "WebKit::RemoteSerializedImageBufferWriteReference", nullptr, false },
            { "destinationImageBuffer", "WebCore::RenderingResourceIdentifier", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteAdapter_RequestDevice:
        return Vector<ArgumentDescription> {
            { "deviceDescriptor", "WebKit::WebGPU::DeviceDescriptor", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
            { "queueIdentifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteBindGroup_SetLabel:
        return Vector<ArgumentDescription> {
            { "label", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteBindGroupLayout_SetLabel:
        return Vector<ArgumentDescription> {
            { "label", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteBuffer_GetMappedRange:
        return Vector<ArgumentDescription> {
            { "offset", "PAL::WebGPU::Size64", nullptr, false },
            { "size", "PAL::WebGPU::Size64", nullptr, true },
        };
    case MessageName::RemoteBuffer_MapAsync:
        return Vector<ArgumentDescription> {
            { "mapModeFlags", "PAL::WebGPU::MapModeFlags", nullptr, false },
            { "offset", "PAL::WebGPU::Size64", nullptr, false },
            { "size", "PAL::WebGPU::Size64", nullptr, true },
        };
    case MessageName::RemoteBuffer_Unmap:
        return Vector<ArgumentDescription> {
            { "data", "Vector<uint8_t>", nullptr, false },
        };
    case MessageName::RemoteBuffer_Destroy:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteBuffer_SetLabel:
        return Vector<ArgumentDescription> {
            { "label", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteCommandBuffer_SetLabel:
        return Vector<ArgumentDescription> {
            { "label", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteCommandEncoder_BeginRenderPass:
        return Vector<ArgumentDescription> {
            { "descriptor", "WebKit::WebGPU::RenderPassDescriptor", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteCommandEncoder_BeginComputePass:
        return Vector<ArgumentDescription> {
            { "descriptor", "WebKit::WebGPU::ComputePassDescriptor", nullptr, true },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteCommandEncoder_CopyBufferToBuffer:
        return Vector<ArgumentDescription> {
            { "source", "WebKit::WebGPUIdentifier", nullptr, false },
            { "sourceOffset", "PAL::WebGPU::Size64", nullptr, false },
            { "destination", "WebKit::WebGPUIdentifier", nullptr, false },
            { "destinationOffset", "PAL::WebGPU::Size64", nullptr, false },
            { "size", "PAL::WebGPU::Size64", nullptr, false },
        };
    case MessageName::RemoteCommandEncoder_CopyBufferToTexture:
        return Vector<ArgumentDescription> {
            { "source", "WebKit::WebGPU::ImageCopyBuffer", nullptr, false },
            { "destination", "WebKit::WebGPU::ImageCopyTexture", nullptr, false },
            { "copySize", "WebKit::WebGPU::Extent3D", nullptr, false },
        };
    case MessageName::RemoteCommandEncoder_CopyTextureToBuffer:
        return Vector<ArgumentDescription> {
            { "source", "WebKit::WebGPU::ImageCopyTexture", nullptr, false },
            { "destination", "WebKit::WebGPU::ImageCopyBuffer", nullptr, false },
            { "copySize", "WebKit::WebGPU::Extent3D", nullptr, false },
        };
    case MessageName::RemoteCommandEncoder_CopyTextureToTexture:
        return Vector<ArgumentDescription> {
            { "source", "WebKit::WebGPU::ImageCopyTexture", nullptr, false },
            { "destination", "WebKit::WebGPU::ImageCopyTexture", nullptr, false },
            { "copySize", "WebKit::WebGPU::Extent3D", nullptr, false },
        };
    case MessageName::RemoteCommandEncoder_ClearBuffer:
        return Vector<ArgumentDescription> {
            { "buffer", "WebKit::WebGPUIdentifier", nullptr, false },
            { "offset", "PAL::WebGPU::Size64", nullptr, false },
            { "size", "PAL::WebGPU::Size64", nullptr, true },
        };
    case MessageName::RemoteCommandEncoder_PushDebugGroup:
        return Vector<ArgumentDescription> {
            { "groupLabel", "String", nullptr, false },
        };
    case MessageName::RemoteCommandEncoder_PopDebugGroup:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteCommandEncoder_InsertDebugMarker:
        return Vector<ArgumentDescription> {
            { "markerLabel", "String", nullptr, false },
        };
    case MessageName::RemoteCommandEncoder_WriteTimestamp:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
            { "queryIndex", "PAL::WebGPU::Size32", nullptr, false },
        };
    case MessageName::RemoteCommandEncoder_ResolveQuerySet:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
            { "firstQuery", "PAL::WebGPU::Size32", nullptr, false },
            { "queryCount", "PAL::WebGPU::Size32", nullptr, false },
            { "destination", "WebKit::WebGPUIdentifier", nullptr, false },
            { "destinationOffset", "PAL::WebGPU::Size64", nullptr, false },
        };
    case MessageName::RemoteCommandEncoder_Finish:
        return Vector<ArgumentDescription> {
            { "descriptor", "WebKit::WebGPU::CommandBufferDescriptor", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteCommandEncoder_SetLabel:
        return Vector<ArgumentDescription> {
            { "label", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
#if PLATFORM(COCOA)
    case MessageName::RemoteCompositorIntegration_RecreateRenderBuffers:
        return Vector<ArgumentDescription> {
            { "width", "int", nullptr, false },
            { "height", "int", nullptr, false },
        };
#endif
    case MessageName::RemoteCompositorIntegration_PrepareForDisplay:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteComputePassEncoder_SetPipeline:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteComputePassEncoder_Dispatch:
        return Vector<ArgumentDescription> {
            { "workgroupCountX", "PAL::WebGPU::Size32", nullptr, false },
            { "workgroupCountY", "PAL::WebGPU::Size32", nullptr, false },
            { "workgroupCountZ", "PAL::WebGPU::Size32", nullptr, false },
        };
    case MessageName::RemoteComputePassEncoder_DispatchIndirect:
        return Vector<ArgumentDescription> {
            { "indirectBuffer", "WebKit::WebGPUIdentifier", nullptr, false },
            { "indirectOffset", "PAL::WebGPU::Size64", nullptr, false },
        };
    case MessageName::RemoteComputePassEncoder_End:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteComputePassEncoder_SetBindGroup:
        return Vector<ArgumentDescription> {
            { "index", "PAL::WebGPU::Index32", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
            { "dynamicOffsets", "Vector<PAL::WebGPU::BufferDynamicOffset>", nullptr, true },
        };
    case MessageName::RemoteComputePassEncoder_PushDebugGroup:
        return Vector<ArgumentDescription> {
            { "groupLabel", "String", nullptr, false },
        };
    case MessageName::RemoteComputePassEncoder_PopDebugGroup:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteComputePassEncoder_InsertDebugMarker:
        return Vector<ArgumentDescription> {
            { "markerLabel", "String", nullptr, false },
        };
    case MessageName::RemoteComputePassEncoder_SetLabel:
        return Vector<ArgumentDescription> {
            { "label", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteComputePipeline_GetBindGroupLayout:
        return Vector<ArgumentDescription> {
            { "index", "uint32_t", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteComputePipeline_SetLabel:
        return Vector<ArgumentDescription> {
            { "label", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteDevice_Destroy:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteDevice_CreateBuffer:
        return Vector<ArgumentDescription> {
            { "descriptor", "WebKit::WebGPU::BufferDescriptor", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteDevice_CreateTexture:
        return Vector<ArgumentDescription> {
            { "descriptor", "WebKit::WebGPU::TextureDescriptor", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteDevice_CreateSampler:
        return Vector<ArgumentDescription> {
            { "descriptor", "WebKit::WebGPU::SamplerDescriptor", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteDevice_ImportExternalTexture:
        return Vector<ArgumentDescription> {
            { "descriptor", "WebKit::WebGPU::ExternalTextureDescriptor", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteDevice_CreateBindGroupLayout:
        return Vector<ArgumentDescription> {
            { "descriptor", "WebKit::WebGPU::BindGroupLayoutDescriptor", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteDevice_CreatePipelineLayout:
        return Vector<ArgumentDescription> {
            { "descriptor", "WebKit::WebGPU::PipelineLayoutDescriptor", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteDevice_CreateBindGroup:
        return Vector<ArgumentDescription> {
            { "descriptor", "WebKit::WebGPU::BindGroupDescriptor", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteDevice_CreateShaderModule:
        return Vector<ArgumentDescription> {
            { "descriptor", "WebKit::WebGPU::ShaderModuleDescriptor", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteDevice_CreateComputePipeline:
        return Vector<ArgumentDescription> {
            { "descriptor", "WebKit::WebGPU::ComputePipelineDescriptor", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteDevice_CreateRenderPipeline:
        return Vector<ArgumentDescription> {
            { "descriptor", "WebKit::WebGPU::RenderPipelineDescriptor", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteDevice_CreateComputePipelineAsync:
        return Vector<ArgumentDescription> {
            { "descriptor", "WebKit::WebGPU::ComputePipelineDescriptor", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteDevice_CreateRenderPipelineAsync:
        return Vector<ArgumentDescription> {
            { "descriptor", "WebKit::WebGPU::RenderPipelineDescriptor", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteDevice_CreateCommandEncoder:
        return Vector<ArgumentDescription> {
            { "descriptor", "WebKit::WebGPU::CommandEncoderDescriptor", nullptr, true },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteDevice_CreateRenderBundleEncoder:
        return Vector<ArgumentDescription> {
            { "descriptor", "WebKit::WebGPU::RenderBundleEncoderDescriptor", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteDevice_CreateQuerySet:
        return Vector<ArgumentDescription> {
            { "descriptor", "WebKit::WebGPU::QuerySetDescriptor", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteDevice_PushErrorScope:
        return Vector<ArgumentDescription> {
            { "errorFilter", "PAL::WebGPU::ErrorFilter", nullptr, false },
        };
    case MessageName::RemoteDevice_PopErrorScope:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteDevice_SetLabel:
        return Vector<ArgumentDescription> {
            { "label", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteExternalTexture_SetLabel:
        return Vector<ArgumentDescription> {
            { "label", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteGPU_RequestAdapter:
        return Vector<ArgumentDescription> {
            { "options", "WebKit::WebGPU::RequestAdapterOptions", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteGPU_CreatePresentationContext:
        return Vector<ArgumentDescription> {
            { "descriptor", "WebKit::WebGPU::PresentationContextDescriptor", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteGPU_CreateCompositorIntegration:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemotePipelineLayout_SetLabel:
        return Vector<ArgumentDescription> {
            { "label", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemotePresentationContext_Configure:
        return Vector<ArgumentDescription> {
            { "configuration", "WebKit::WebGPU::CanvasConfiguration", nullptr, false },
        };
    case MessageName::RemotePresentationContext_Unconfigure:
        return Vector<ArgumentDescription> { };
    case MessageName::RemotePresentationContext_GetCurrentTexture:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteQuerySet_Destroy:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteQuerySet_SetLabel:
        return Vector<ArgumentDescription> {
            { "label", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteQueue_Submit:
        return Vector<ArgumentDescription> {
            { "commandBuffers", "Vector<WebKit::WebGPUIdentifier>", nullptr, false },
        };
    case MessageName::RemoteQueue_OnSubmittedWorkDone:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteQueue_WriteBuffer:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
            { "bufferOffset", "PAL::WebGPU::Size64", nullptr, false },
            { "data", "Vector<uint8_t>", nullptr, false },
        };
    case MessageName::RemoteQueue_WriteTexture:
        return Vector<ArgumentDescription> {
            { "destination", "WebKit::WebGPU::ImageCopyTexture", nullptr, false },
            { "data", "Vector<uint8_t>", nullptr, false },
            { "imageDataLayout", "WebKit::WebGPU::ImageDataLayout", nullptr, false },
            { "size", "WebKit::WebGPU::Extent3D", nullptr, false },
        };
    case MessageName::RemoteQueue_CopyExternalImageToTexture:
        return Vector<ArgumentDescription> {
            { "source", "WebKit::WebGPU::ImageCopyExternalImage", nullptr, false },
            { "destination", "WebKit::WebGPU::ImageCopyTextureTagged", nullptr, false },
            { "copySize", "WebKit::WebGPU::Extent3D", nullptr, false },
        };
    case MessageName::RemoteQueue_SetLabel:
        return Vector<ArgumentDescription> {
            { "label", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteRenderBundle_SetLabel:
        return Vector<ArgumentDescription> {
            { "label", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteRenderBundleEncoder_SetPipeline:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteRenderBundleEncoder_SetIndexBuffer:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
            { "indexFormat", "PAL::WebGPU::IndexFormat", nullptr, false },
            { "offset", "PAL::WebGPU::Size64", nullptr, true },
            { "size", "PAL::WebGPU::Size64", nullptr, true },
        };
    case MessageName::RemoteRenderBundleEncoder_SetVertexBuffer:
        return Vector<ArgumentDescription> {
            { "slot", "PAL::WebGPU::Index32", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
            { "offset", "PAL::WebGPU::Size64", nullptr, true },
            { "size", "PAL::WebGPU::Size64", nullptr, true },
        };
    case MessageName::RemoteRenderBundleEncoder_Draw:
        return Vector<ArgumentDescription> {
            { "vertexCount", "PAL::WebGPU::Size32", nullptr, false },
            { "instanceCount", "PAL::WebGPU::Size32", nullptr, true },
            { "firstVertex", "PAL::WebGPU::Size32", nullptr, true },
            { "firstInstance", "PAL::WebGPU::Size32", nullptr, true },
        };
    case MessageName::RemoteRenderBundleEncoder_DrawIndexed:
        return Vector<ArgumentDescription> {
            { "indexCount", "PAL::WebGPU::Size32", nullptr, false },
            { "instanceCount", "PAL::WebGPU::Size32", nullptr, true },
            { "firstIndex", "PAL::WebGPU::Size32", nullptr, true },
            { "baseVertex", "PAL::WebGPU::SignedOffset32", nullptr, true },
            { "firstInstance", "PAL::WebGPU::Size32", nullptr, true },
        };
    case MessageName::RemoteRenderBundleEncoder_DrawIndirect:
        return Vector<ArgumentDescription> {
            { "indirectBuffer", "WebKit::WebGPUIdentifier", nullptr, false },
            { "indirectOffset", "PAL::WebGPU::Size64", nullptr, false },
        };
    case MessageName::RemoteRenderBundleEncoder_DrawIndexedIndirect:
        return Vector<ArgumentDescription> {
            { "indirectBuffer", "WebKit::WebGPUIdentifier", nullptr, false },
            { "indirectOffset", "PAL::WebGPU::Size64", nullptr, false },
        };
    case MessageName::RemoteRenderBundleEncoder_SetBindGroup:
        return Vector<ArgumentDescription> {
            { "index", "PAL::WebGPU::Index32", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
            { "dynamicOffsets", "Vector<PAL::WebGPU::BufferDynamicOffset>", nullptr, true },
        };
    case MessageName::RemoteRenderBundleEncoder_PushDebugGroup:
        return Vector<ArgumentDescription> {
            { "groupLabel", "String", nullptr, false },
        };
    case MessageName::RemoteRenderBundleEncoder_PopDebugGroup:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteRenderBundleEncoder_InsertDebugMarker:
        return Vector<ArgumentDescription> {
            { "markerLabel", "String", nullptr, false },
        };
    case MessageName::RemoteRenderBundleEncoder_Finish:
        return Vector<ArgumentDescription> {
            { "descriptor", "WebKit::WebGPU::RenderBundleDescriptor", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteRenderBundleEncoder_SetLabel:
        return Vector<ArgumentDescription> {
            { "label", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteRenderPassEncoder_SetPipeline:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteRenderPassEncoder_SetIndexBuffer:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
            { "indexFormat", "PAL::WebGPU::IndexFormat", nullptr, false },
            { "offset", "PAL::WebGPU::Size64", nullptr, true },
            { "size", "PAL::WebGPU::Size64", nullptr, true },
        };
    case MessageName::RemoteRenderPassEncoder_SetVertexBuffer:
        return Vector<ArgumentDescription> {
            { "slot", "PAL::WebGPU::Index32", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
            { "offset", "PAL::WebGPU::Size64", nullptr, true },
            { "size", "PAL::WebGPU::Size64", nullptr, true },
        };
    case MessageName::RemoteRenderPassEncoder_Draw:
        return Vector<ArgumentDescription> {
            { "vertexCount", "PAL::WebGPU::Size32", nullptr, false },
            { "instanceCount", "PAL::WebGPU::Size32", nullptr, true },
            { "firstVertex", "PAL::WebGPU::Size32", nullptr, true },
            { "firstInstance", "PAL::WebGPU::Size32", nullptr, true },
        };
    case MessageName::RemoteRenderPassEncoder_DrawIndexed:
        return Vector<ArgumentDescription> {
            { "indexCount", "PAL::WebGPU::Size32", nullptr, false },
            { "instanceCount", "PAL::WebGPU::Size32", nullptr, true },
            { "firstIndex", "PAL::WebGPU::Size32", nullptr, true },
            { "baseVertex", "PAL::WebGPU::SignedOffset32", nullptr, true },
            { "firstInstance", "PAL::WebGPU::Size32", nullptr, true },
        };
    case MessageName::RemoteRenderPassEncoder_DrawIndirect:
        return Vector<ArgumentDescription> {
            { "indirectBuffer", "WebKit::WebGPUIdentifier", nullptr, false },
            { "indirectOffset", "PAL::WebGPU::Size64", nullptr, false },
        };
    case MessageName::RemoteRenderPassEncoder_DrawIndexedIndirect:
        return Vector<ArgumentDescription> {
            { "indirectBuffer", "WebKit::WebGPUIdentifier", nullptr, false },
            { "indirectOffset", "PAL::WebGPU::Size64", nullptr, false },
        };
    case MessageName::RemoteRenderPassEncoder_SetBindGroup:
        return Vector<ArgumentDescription> {
            { "index", "PAL::WebGPU::Index32", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
            { "dynamicOffsets", "Vector<PAL::WebGPU::BufferDynamicOffset>", nullptr, true },
        };
    case MessageName::RemoteRenderPassEncoder_PushDebugGroup:
        return Vector<ArgumentDescription> {
            { "groupLabel", "String", nullptr, false },
        };
    case MessageName::RemoteRenderPassEncoder_PopDebugGroup:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteRenderPassEncoder_InsertDebugMarker:
        return Vector<ArgumentDescription> {
            { "markerLabel", "String", nullptr, false },
        };
    case MessageName::RemoteRenderPassEncoder_SetViewport:
        return Vector<ArgumentDescription> {
            { "x", "float", nullptr, false },
            { "y", "float", nullptr, false },
            { "width", "float", nullptr, false },
            { "height", "float", nullptr, false },
            { "minDepth", "float", nullptr, false },
            { "maxDepth", "float", nullptr, false },
        };
    case MessageName::RemoteRenderPassEncoder_SetScissorRect:
        return Vector<ArgumentDescription> {
            { "x", "PAL::WebGPU::IntegerCoordinate", nullptr, false },
            { "y", "PAL::WebGPU::IntegerCoordinate", nullptr, false },
            { "width", "PAL::WebGPU::IntegerCoordinate", nullptr, false },
            { "height", "PAL::WebGPU::IntegerCoordinate", nullptr, false },
        };
    case MessageName::RemoteRenderPassEncoder_SetBlendConstant:
        return Vector<ArgumentDescription> {
            { "color", "WebKit::WebGPU::Color", nullptr, false },
        };
    case MessageName::RemoteRenderPassEncoder_SetStencilReference:
        return Vector<ArgumentDescription> {
            { "stencilValue", "PAL::WebGPU::StencilValue", nullptr, false },
        };
    case MessageName::RemoteRenderPassEncoder_BeginOcclusionQuery:
        return Vector<ArgumentDescription> {
            { "queryIndex", "PAL::WebGPU::Size32", nullptr, false },
        };
    case MessageName::RemoteRenderPassEncoder_EndOcclusionQuery:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteRenderPassEncoder_ExecuteBundles:
        return Vector<ArgumentDescription> {
            { "bundles", "Vector<WebKit::WebGPUIdentifier>", nullptr, false },
        };
    case MessageName::RemoteRenderPassEncoder_End:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteRenderPassEncoder_SetLabel:
        return Vector<ArgumentDescription> {
            { "label", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteRenderPipeline_GetBindGroupLayout:
        return Vector<ArgumentDescription> {
            { "index", "uint32_t", nullptr, false },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteRenderPipeline_SetLabel:
        return Vector<ArgumentDescription> {
            { "label", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteSampler_SetLabel:
        return Vector<ArgumentDescription> {
            { "label", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteShaderModule_CompilationInfo:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteShaderModule_SetLabel:
        return Vector<ArgumentDescription> {
            { "label", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteTexture_CreateView:
        return Vector<ArgumentDescription> {
            { "descriptor", "WebKit::WebGPU::TextureViewDescriptor", nullptr, true },
            { "identifier", "WebKit::WebGPUIdentifier", nullptr, false },
        };
    case MessageName::RemoteTexture_Destroy:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteTexture_SetLabel:
        return Vector<ArgumentDescription> {
            { "label", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteTextureView_SetLabel:
        return Vector<ArgumentDescription> {
            { "label", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(WEB_AUDIO)
    case MessageName::RemoteAudioDestinationManager_CreateAudioDestination:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteAudioDestinationIdentifier", nullptr, false },
            { "inputDeviceId", "String", nullptr, false },
            { "numberOfInputChannels", "uint32_t", nullptr, false },
            { "numberOfOutputChannels", "uint32_t", nullptr, false },
            { "sampleRate", "float", nullptr, false },
            { "hardwareSampleRate", "float", nullptr, false },
            { "renderSemaphore", "IPC::Semaphore", nullptr, false },
        };
    case MessageName::RemoteAudioDestinationManager_DeleteAudioDestination:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteAudioDestinationIdentifier", nullptr, false },
        };
    case MessageName::RemoteAudioDestinationManager_StartAudioDestination:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteAudioDestinationIdentifier", nullptr, false },
        };
    case MessageName::RemoteAudioDestinationManager_StopAudioDestination:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteAudioDestinationIdentifier", nullptr, false },
        };
#if PLATFORM(COCOA)
    case MessageName::RemoteAudioDestinationManager_AudioSamplesStorageChanged:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteAudioDestinationIdentifier", nullptr, false },
            { "storageHandle", "WebKit::ConsumerSharedCARingBuffer::Handle", nullptr, false },
        };
#endif
#endif
#if ENABLE(GPU_PROCESS) && USE(AUDIO_SESSION)
    case MessageName::RemoteAudioSessionProxy_SetCategory:
        return Vector<ArgumentDescription> {
            { "type", "uint8_t", "WebCore::AudioSession::CategoryType", false },
            { "policy", "uint8_t", "WebCore::RouteSharingPolicy", false },
        };
    case MessageName::RemoteAudioSessionProxy_SetPreferredBufferSize:
        return Vector<ArgumentDescription> {
            { "preferredBufferSize", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteAudioSessionProxy_TryToSetActive:
        return Vector<ArgumentDescription> {
            { "active", "bool", nullptr, false },
        };
    case MessageName::RemoteAudioSessionProxy_SetIsPlayingToBluetoothOverride:
        return Vector<ArgumentDescription> {
            { "value", "bool", nullptr, true },
        };
    case MessageName::RemoteAudioSessionProxy_TriggerBeginInterruptionForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteAudioSessionProxy_TriggerEndInterruptionForTesting:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteCDMFactoryProxy_CreateCDM:
        return Vector<ArgumentDescription> {
            { "keySystem", "String", nullptr, false },
        };
    case MessageName::RemoteCDMFactoryProxy_SupportsKeySystem:
        return Vector<ArgumentDescription> {
            { "keySystem", "String", nullptr, false },
        };
    case MessageName::RemoteCDMFactoryProxy_RemoveInstance:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteCDMInstanceIdentifier", nullptr, false },
        };
    case MessageName::RemoteCDMFactoryProxy_RemoveSession:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteCDMInstanceSessionIdentifier", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteCDMInstanceProxy_CreateSession:
        return Vector<ArgumentDescription> {
            { "logIdentifier", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteCDMInstanceProxy_InitializeWithConfiguration:
        return Vector<ArgumentDescription> {
            { "configuration", "WebCore::CDMKeySystemConfiguration", nullptr, false },
            { "distinctiveIdentifiersAllowed", "bool", "WebCore::CDMInstance::AllowDistinctiveIdentifiers", false },
            { "persistentStateAllowed", "bool", "WebCore::CDMInstance::AllowPersistentState", false },
        };
    case MessageName::RemoteCDMInstanceProxy_SetServerCertificate:
        return Vector<ArgumentDescription> {
            { "certificate", "Ref<WebCore::SharedBuffer>", nullptr, false },
        };
    case MessageName::RemoteCDMInstanceProxy_SetStorageDirectory:
        return Vector<ArgumentDescription> {
            { "directory", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteCDMInstanceSessionProxy_SetLogIdentifier:
        return Vector<ArgumentDescription> {
            { "logIdentifier", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteCDMInstanceSessionProxy_RequestLicense:
        return Vector<ArgumentDescription> {
            { "type", "WebCore::CDMInstanceSession::LicenseType", nullptr, false },
            { "initDataType", "AtomString", nullptr, false },
            { "initData", "RefPtr<WebCore::SharedBuffer>", nullptr, false },
        };
    case MessageName::RemoteCDMInstanceSessionProxy_UpdateLicense:
        return Vector<ArgumentDescription> {
            { "sessionId", "String", nullptr, false },
            { "type", "WebCore::CDMInstanceSession::LicenseType", nullptr, false },
            { "response", "RefPtr<WebCore::SharedBuffer>", nullptr, false },
        };
    case MessageName::RemoteCDMInstanceSessionProxy_LoadSession:
        return Vector<ArgumentDescription> {
            { "type", "WebCore::CDMInstanceSession::LicenseType", nullptr, false },
            { "sessionId", "String", nullptr, false },
            { "origin", "String", nullptr, false },
        };
    case MessageName::RemoteCDMInstanceSessionProxy_CloseSession:
        return Vector<ArgumentDescription> {
            { "sessionId", "String", nullptr, false },
        };
    case MessageName::RemoteCDMInstanceSessionProxy_RemoveSessionData:
        return Vector<ArgumentDescription> {
            { "sessionId", "String", nullptr, false },
            { "type", "WebCore::CDMInstanceSession::LicenseType", nullptr, false },
        };
    case MessageName::RemoteCDMInstanceSessionProxy_StoreRecordOfKeyUsage:
        return Vector<ArgumentDescription> {
            { "sessionId", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteCDMProxy_GetSupportedConfiguration:
        return Vector<ArgumentDescription> {
            { "candidateConfiguration", "WebCore::CDMKeySystemConfiguration", nullptr, false },
            { "access", "bool", "WebCore::CDMPrivate::LocalStorageAccess", false },
        };
    case MessageName::RemoteCDMProxy_CreateInstance:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteCDMProxy_LoadAndInitialize:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteCDMProxy_SetLogIdentifier:
        return Vector<ArgumentDescription> {
            { "logIdentifier", "uint64_t", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
    case MessageName::RemoteLegacyCDMFactoryProxy_CreateCDM:
        return Vector<ArgumentDescription> {
            { "keySystem", "String", nullptr, false },
            { "playerId", "WebCore::MediaPlayerIdentifier", nullptr, true },
        };
    case MessageName::RemoteLegacyCDMFactoryProxy_SupportsKeySystem:
        return Vector<ArgumentDescription> {
            { "keySystem", "String", nullptr, false },
            { "mimeType", "String", nullptr, true },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
    case MessageName::RemoteLegacyCDMProxy_SupportsMIMEType:
        return Vector<ArgumentDescription> {
            { "mimeType", "String", nullptr, false },
        };
    case MessageName::RemoteLegacyCDMProxy_CreateSession:
        return Vector<ArgumentDescription> {
            { "mediaKeysStorageDirectory", "String", nullptr, false },
            { "logIdentifier", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteLegacyCDMProxy_SetPlayerId:
        return Vector<ArgumentDescription> {
            { "playerId", "WebCore::MediaPlayerIdentifier", nullptr, true },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
    case MessageName::RemoteLegacyCDMSessionProxy_GenerateKeyRequest:
        return Vector<ArgumentDescription> {
            { "mimeType", "String", nullptr, false },
            { "initData", "RefPtr<WebCore::SharedBuffer>", nullptr, false },
        };
    case MessageName::RemoteLegacyCDMSessionProxy_ReleaseKeys:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteLegacyCDMSessionProxy_Update:
        return Vector<ArgumentDescription> {
            { "update", "RefPtr<WebCore::SharedBuffer>", nullptr, false },
        };
    case MessageName::RemoteLegacyCDMSessionProxy_CachedKeyForKeyID:
        return Vector<ArgumentDescription> {
            { "keyId", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteMediaEngineConfigurationFactoryProxy_CreateDecodingConfiguration:
        return Vector<ArgumentDescription> {
            { "configuration", "WebCore::MediaDecodingConfiguration", nullptr, false },
        };
    case MessageName::RemoteMediaEngineConfigurationFactoryProxy_CreateEncodingConfiguration:
        return Vector<ArgumentDescription> {
            { "configuration", "WebCore::MediaEncodingConfiguration", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
    case MessageName::RemoteMediaPlayerManagerProxy_CreateMediaPlayer:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::MediaPlayerIdentifier", nullptr, false },
            { "remoteEngineIdentifier", "uint8_t", "WebCore::MediaPlayerEnums::MediaEngineIdentifier", false },
            { "proxyConfiguration", "WebKit::RemoteMediaPlayerProxyConfiguration", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerManagerProxy_DeleteMediaPlayer:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::MediaPlayerIdentifier", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerManagerProxy_GetSupportedTypes:
        return Vector<ArgumentDescription> {
            { "remoteEngineIdentifier", "uint8_t", "WebCore::MediaPlayerEnums::MediaEngineIdentifier", false },
        };
    case MessageName::RemoteMediaPlayerManagerProxy_SupportsTypeAndCodecs:
        return Vector<ArgumentDescription> {
            { "remoteEngineIdentifier", "uint8_t", "WebCore::MediaPlayerEnums::MediaEngineIdentifier", false },
            { "type", "WebCore::MediaEngineSupportParameters", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerManagerProxy_OriginsInMediaCache:
        return Vector<ArgumentDescription> {
            { "remoteEngineIdentifier", "uint8_t", "WebCore::MediaPlayerEnums::MediaEngineIdentifier", false },
            { "path", "String", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerManagerProxy_ClearMediaCache:
        return Vector<ArgumentDescription> {
            { "remoteEngineIdentifier", "uint8_t", "WebCore::MediaPlayerEnums::MediaEngineIdentifier", false },
            { "path", "String", nullptr, false },
            { "modifiedSince", "WallTime", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerManagerProxy_ClearMediaCacheForOrigins:
        return Vector<ArgumentDescription> {
            { "remoteEngineIdentifier", "uint8_t", "WebCore::MediaPlayerEnums::MediaEngineIdentifier", false },
            { "path", "String", nullptr, false },
            { "origins", "HashSet<WebCore::SecurityOriginData>", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerManagerProxy_SupportsKeySystem:
        return Vector<ArgumentDescription> {
            { "remoteEngineIdentifier", "uint8_t", "WebCore::MediaPlayerEnums::MediaEngineIdentifier", false },
            { "keySystem", "String", nullptr, false },
            { "mimeType", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
    case MessageName::RemoteMediaPlayerProxy_PrepareForPlayback:
        return Vector<ArgumentDescription> {
            { "privateMode", "bool", nullptr, false },
            { "preload", "uint8_t", "WebCore::MediaPlayerEnums::Preload", false },
            { "preservesPitch", "bool", nullptr, false },
            { "prepareForRendering", "bool", nullptr, false },
            { "presentationSize", "WebCore::IntSize", nullptr, false },
            { "videoContentScale", "float", nullptr, false },
            { "mode", "uint8_t", "WebCore::DynamicRangeMode", false },
        };
    case MessageName::RemoteMediaPlayerProxy_Load:
        return Vector<ArgumentDescription> {
            { "url", "URL", nullptr, false },
            { "sandboxExtension", "WebKit::SandboxExtension::Handle", nullptr, true },
            { "contentType", "WebCore::ContentType", nullptr, false },
            { "keySystem", "String", nullptr, false },
            { "requiresRemotePlayback", "bool", nullptr, false },
        };
#if ENABLE(MEDIA_SOURCE)
    case MessageName::RemoteMediaPlayerProxy_LoadMediaSource:
        return Vector<ArgumentDescription> {
            { "url", "URL", nullptr, false },
            { "contentType", "WebCore::ContentType", nullptr, false },
            { "webMParserEnabled", "bool", nullptr, false },
            { "mediaSourceIdentifier", "WebKit::RemoteMediaSourceIdentifier", nullptr, false },
        };
#endif
    case MessageName::RemoteMediaPlayerProxy_CancelLoad:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaPlayerProxy_PrepareToPlay:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaPlayerProxy_Play:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaPlayerProxy_Pause:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaPlayerProxy_SetVolume:
        return Vector<ArgumentDescription> {
            { "volume", "double", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_SetMuted:
        return Vector<ArgumentDescription> {
            { "muted", "bool", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_Seek:
        return Vector<ArgumentDescription> {
            { "time", "MediaTime", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_SeekWithTolerance:
        return Vector<ArgumentDescription> {
            { "time", "MediaTime", nullptr, false },
            { "negativeTolerance", "MediaTime", nullptr, false },
            { "positiveTolerance", "MediaTime", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_SetPreload:
        return Vector<ArgumentDescription> {
            { "preload", "uint8_t", "WebCore::MediaPlayerEnums::Preload", false },
        };
    case MessageName::RemoteMediaPlayerProxy_SetPrivateBrowsingMode:
        return Vector<ArgumentDescription> {
            { "privateMode", "bool", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_SetPreservesPitch:
        return Vector<ArgumentDescription> {
            { "preservesPitch", "bool", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_SetPitchCorrectionAlgorithm:
        return Vector<ArgumentDescription> {
            { "algorithm", "WebCore::MediaPlayer::PitchCorrectionAlgorithm", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_PrepareForRendering:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaPlayerProxy_SetPageIsVisible:
        return Vector<ArgumentDescription> {
            { "visible", "bool", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_SetShouldMaintainAspectRatio:
        return Vector<ArgumentDescription> {
            { "maintainAspectRatio", "bool", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_AcceleratedRenderingStateChanged:
        return Vector<ArgumentDescription> {
            { "canBeAccelerated", "bool", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_SetShouldDisableSleep:
        return Vector<ArgumentDescription> {
            { "disable", "bool", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_SetRate:
        return Vector<ArgumentDescription> {
            { "rate", "double", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_DidLoadingProgress:
        return Vector<ArgumentDescription> { };
#if ENABLE(VIDEO_PRESENTATION_MODE)
    case MessageName::RemoteMediaPlayerProxy_UpdateVideoFullscreenInlineImage:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaPlayerProxy_SetVideoFullscreenMode:
        return Vector<ArgumentDescription> {
            { "mode", "WebCore::MediaPlayer::VideoFullscreenMode", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_SetVideoFullscreenGravity:
        return Vector<ArgumentDescription> {
            { "gravity", "uint8_t", "WebCore::MediaPlayerEnums::VideoGravity", false },
        };
    case MessageName::RemoteMediaPlayerProxy_VideoFullscreenStandbyChanged:
        return Vector<ArgumentDescription> {
            { "standby", "bool", nullptr, false },
        };
#endif
    case MessageName::RemoteMediaPlayerProxy_SetBufferingPolicy:
        return Vector<ArgumentDescription> {
            { "policy", "WebCore::MediaPlayer::BufferingPolicy", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_SetPresentationSize:
        return Vector<ArgumentDescription> {
            { "size", "WebCore::IntSize", nullptr, false },
        };
#if PLATFORM(COCOA)
    case MessageName::RemoteMediaPlayerProxy_SetVideoInlineSizeFenced:
        return Vector<ArgumentDescription> {
            { "size", "WebCore::FloatSize", nullptr, false },
            { "machSendRight", "MachSendRight", nullptr, false },
        };
#endif
#if ENABLE(WIRELESS_PLAYBACK_TARGET)
    case MessageName::RemoteMediaPlayerProxy_SetWirelessVideoPlaybackDisabled:
        return Vector<ArgumentDescription> {
            { "disabled", "bool", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_SetWirelessPlaybackTarget:
        return Vector<ArgumentDescription> {
            { "target", "WebCore::MediaPlaybackTargetContext", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_SetShouldPlayToPlaybackTarget:
        return Vector<ArgumentDescription> {
            { "shouldPlay", "bool", nullptr, false },
        };
#endif
#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
    case MessageName::RemoteMediaPlayerProxy_SetLegacyCDMSession:
        return Vector<ArgumentDescription> {
            { "instanceId", "WebKit::RemoteLegacyCDMSessionIdentifier", nullptr, true },
        };
    case MessageName::RemoteMediaPlayerProxy_KeyAdded:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteMediaPlayerProxy_CdmInstanceAttached:
        return Vector<ArgumentDescription> {
            { "instanceId", "WebKit::RemoteCDMInstanceIdentifier", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_CdmInstanceDetached:
        return Vector<ArgumentDescription> {
            { "instanceId", "WebKit::RemoteCDMInstanceIdentifier", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_AttemptToDecryptWithInstance:
        return Vector<ArgumentDescription> {
            { "instanceId", "WebKit::RemoteCDMInstanceIdentifier", nullptr, false },
        };
#endif
#if ENABLE(LEGACY_ENCRYPTED_MEDIA) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteMediaPlayerProxy_SetShouldContinueAfterKeyNeeded:
        return Vector<ArgumentDescription> {
            { "should", "bool", nullptr, false },
        };
#endif
    case MessageName::RemoteMediaPlayerProxy_BeginSimulatedHDCPError:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaPlayerProxy_EndSimulatedHDCPError:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaPlayerProxy_NotifyActiveSourceBuffersChanged:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaPlayerProxy_ApplicationWillResignActive:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaPlayerProxy_ApplicationDidBecomeActive:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaPlayerProxy_NotifyTrackModeChanged:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaPlayerProxy_TracksChanged:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaPlayerProxy_AudioTrackSetEnabled:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
            { "enabled", "bool", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_VideoTrackSetSelected:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
            { "selected", "bool", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_TextTrackSetMode:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
            { "mode", "uint8_t", "WebCore::InbandTextTrackPrivate::Mode", false },
        };
    case MessageName::RemoteMediaPlayerProxy_PerformTaskAtMediaTime:
        return Vector<ArgumentDescription> {
            { "mediaTime", "MediaTime", nullptr, false },
            { "messageTime", "MonotonicTime", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_IsCrossOrigin:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::SecurityOriginData", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_SetVideoPlaybackMetricsUpdateInterval:
        return Vector<ArgumentDescription> {
            { "interval", "double", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_SetPreferredDynamicRangeMode:
        return Vector<ArgumentDescription> {
            { "mode", "uint8_t", "WebCore::DynamicRangeMode", false },
        };
#if PLATFORM(IOS_FAMILY)
    case MessageName::RemoteMediaPlayerProxy_ErrorLog:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaPlayerProxy_AccessLog:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(WEB_AUDIO)
    case MessageName::RemoteMediaPlayerProxy_CreateAudioSourceProvider:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaPlayerProxy_SetShouldEnableAudioSourceProvider:
        return Vector<ArgumentDescription> {
            { "shouldEnable", "bool", nullptr, false },
        };
#endif
#if PLATFORM(COCOA)
    case MessageName::RemoteMediaPlayerProxy_NativeImageForCurrentTime:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaPlayerProxy_ColorSpace:
        return Vector<ArgumentDescription> { };
#endif
#if (PLATFORM(COCOA) && !HAVE(AVSAMPLEBUFFERDISPLAYLAYER_COPYDISPLAYEDPIXELBUFFER))
    case MessageName::RemoteMediaPlayerProxy_WillBeAskedToPaintGL:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::RemoteMediaPlayerProxy_VideoFrameForCurrentTimeIfChanged:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaPlayerProxy_PlayAtHostTime:
        return Vector<ArgumentDescription> {
            { "time", "MonotonicTime", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_PauseAtHostTime:
        return Vector<ArgumentDescription> {
            { "time", "MonotonicTime", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_StartVideoFrameMetadataGathering:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaPlayerProxy_StopVideoFrameMetadataGathering:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaPlayerProxy_PlayerContentBoxRectChanged:
        return Vector<ArgumentDescription> {
            { "contentRect", "WebCore::LayoutRect", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_SetShouldDisableHDR:
        return Vector<ArgumentDescription> {
            { "shouldDisable", "bool", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
    case MessageName::RemoteMediaResourceManager_ResponseReceived:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteMediaResourceIdentifier", nullptr, false },
            { "response", "WebCore::ResourceResponse", nullptr, false },
            { "didPassAccessControlCheck", "bool", nullptr, false },
        };
    case MessageName::RemoteMediaResourceManager_RedirectReceived:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteMediaResourceIdentifier", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "response", "WebCore::ResourceResponse", nullptr, false },
        };
    case MessageName::RemoteMediaResourceManager_DataSent:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteMediaResourceIdentifier", nullptr, false },
            { "bytesSent", "uint64_t", nullptr, false },
            { "totalBytesToBeSent", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteMediaResourceManager_DataReceived:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteMediaResourceIdentifier", nullptr, false },
            { "data", "IPC::SharedBufferReference", nullptr, false },
        };
    case MessageName::RemoteMediaResourceManager_AccessControlCheckFailed:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteMediaResourceIdentifier", nullptr, false },
            { "error", "WebCore::ResourceError", nullptr, false },
        };
    case MessageName::RemoteMediaResourceManager_LoadFailed:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteMediaResourceIdentifier", nullptr, false },
            { "error", "WebCore::ResourceError", nullptr, false },
        };
    case MessageName::RemoteMediaResourceManager_LoadFinished:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteMediaResourceIdentifier", nullptr, false },
            { "metrics", "WebCore::NetworkLoadMetrics", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
    case MessageName::RemoteMediaSourceProxy_AddSourceBuffer:
        return Vector<ArgumentDescription> {
            { "contentType", "WebCore::ContentType", nullptr, false },
        };
    case MessageName::RemoteMediaSourceProxy_DurationChanged:
        return Vector<ArgumentDescription> {
            { "duration", "MediaTime", nullptr, false },
        };
    case MessageName::RemoteMediaSourceProxy_BufferedChanged:
        return Vector<ArgumentDescription> {
            { "buffered", "WebCore::PlatformTimeRanges", nullptr, false },
        };
    case MessageName::RemoteMediaSourceProxy_SetReadyState:
        return Vector<ArgumentDescription> {
            { "readyState", "WebCore::MediaPlayerEnums::ReadyState", nullptr, false },
        };
    case MessageName::RemoteMediaSourceProxy_SetIsSeeking:
        return Vector<ArgumentDescription> {
            { "isSeeking", "bool", nullptr, false },
        };
    case MessageName::RemoteMediaSourceProxy_WaitForSeekCompleted:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaSourceProxy_SeekCompleted:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaSourceProxy_SetTimeFudgeFactor:
        return Vector<ArgumentDescription> {
            { "fudgeFactor", "MediaTime", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteRemoteCommandListenerProxy_UpdateSupportedCommands:
        return Vector<ArgumentDescription> {
            { "commands", "Vector<WebCore::PlatformMediaSession::RemoteControlCommandType>", nullptr, false },
            { "supportsSeeking", "bool", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
    case MessageName::RemoteSourceBufferProxy_SetActive:
        return Vector<ArgumentDescription> {
            { "active", "bool", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_CanSwitchToType:
        return Vector<ArgumentDescription> {
            { "contentType", "WebCore::ContentType", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_SetMode:
        return Vector<ArgumentDescription> {
            { "appendMode", "WebCore::SourceBufferAppendMode", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_Append:
        return Vector<ArgumentDescription> {
            { "data", "IPC::SharedBufferReference", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_Abort:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteSourceBufferProxy_ResetParserState:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteSourceBufferProxy_RemovedFromMediaSource:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteSourceBufferProxy_SetMediaSourceEnded:
        return Vector<ArgumentDescription> {
            { "isEnded", "bool", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_SetReadyState:
        return Vector<ArgumentDescription> {
            { "state", "WebCore::MediaPlayer::ReadyState", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_StartChangingType:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteSourceBufferProxy_UpdateBufferedFromTrackBuffers:
        return Vector<ArgumentDescription> {
            { "sourceIsEnded", "bool", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_AddTrackBuffer:
        return Vector<ArgumentDescription> {
            { "remoteIdentifier", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_ResetTrackBuffers:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteSourceBufferProxy_ClearTrackBuffers:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteSourceBufferProxy_SetAllTrackBuffersNeedRandomAccess:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteSourceBufferProxy_RemoveCodedFrames:
        return Vector<ArgumentDescription> {
            { "start", "MediaTime", nullptr, false },
            { "end", "MediaTime", nullptr, false },
            { "currentTime", "MediaTime", nullptr, false },
            { "isEnded", "bool", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_EvictCodedFrames:
        return Vector<ArgumentDescription> {
            { "newDataSize", "uint64_t", nullptr, false },
            { "maximumBufferSize", "uint64_t", nullptr, false },
            { "currentTime", "MediaTime", nullptr, false },
            { "isEnded", "bool", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_ReenqueueMediaIfNeeded:
        return Vector<ArgumentDescription> {
            { "currentMediaTime", "MediaTime", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_SetGroupStartTimestamp:
        return Vector<ArgumentDescription> {
            { "timestamp", "MediaTime", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_SetGroupStartTimestampToEndTimestamp:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteSourceBufferProxy_SetShouldGenerateTimestamps:
        return Vector<ArgumentDescription> {
            { "shouldGenerateTimestamps", "bool", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_ResetTimestampOffsetInTrackBuffers:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteSourceBufferProxy_SetTimestampOffset:
        return Vector<ArgumentDescription> {
            { "timestampOffset", "MediaTime", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_SetAppendWindowStart:
        return Vector<ArgumentDescription> {
            { "appendWindowStart", "MediaTime", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_SetAppendWindowEnd:
        return Vector<ArgumentDescription> {
            { "appendWindowEnd", "MediaTime", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_SeekToTime:
        return Vector<ArgumentDescription> {
            { "mediaTime", "MediaTime", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_UpdateTrackIds:
        return Vector<ArgumentDescription> {
            { "identifierPairs", "Vector<std::pair<WebKit::TrackPrivateRemoteIdentifier, WebKit::TrackPrivateRemoteIdentifier>>", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_BufferedSamplesForTrackId:
        return Vector<ArgumentDescription> {
            { "remoteIdentifier", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_EnqueuedSamplesForTrackID:
        return Vector<ArgumentDescription> {
            { "remoteIdentifier", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
#if PLATFORM(COCOA)
    case MessageName::RemoteVideoFrameObjectHeap_GetVideoFrameBuffer:
        return Vector<ArgumentDescription> {
            { "read", "WebKit::RemoteVideoFrameReadReference", nullptr, false },
            { "canSendIOSurface", "bool", nullptr, false },
        };
    case MessageName::RemoteVideoFrameObjectHeap_PixelBuffer:
        return Vector<ArgumentDescription> {
            { "read", "WebKit::RemoteVideoFrameReadReference", nullptr, false },
        };
    case MessageName::RemoteVideoFrameObjectHeap_SetSharedVideoFrameSemaphore:
        return Vector<ArgumentDescription> {
            { "semaphore", "IPC::Semaphore", nullptr, false },
        };
    case MessageName::RemoteVideoFrameObjectHeap_SetSharedVideoFrameMemory:
        return Vector<ArgumentDescription> {
            { "storageHandle", "WebKit::SharedMemory::Handle", nullptr, false },
        };
    case MessageName::RemoteVideoFrameObjectHeap_ConvertFrameBuffer:
        return Vector<ArgumentDescription> {
            { "frame", "WebKit::SharedVideoFrame", nullptr, false },
        };
#endif
    case MessageName::RemoteVideoFrameObjectHeap_ReleaseVideoFrame:
        return Vector<ArgumentDescription> {
            { "write", "WebKit::RemoteVideoFrameWriteReference", nullptr, false },
        };
#endif
#if USE(LIBWEBRTC) && PLATFORM(COCOA) && ENABLE(GPU_PROCESS)
    case MessageName::LibWebRTCCodecsProxy_CreateDecoder:
        return Vector<ArgumentDescription> {
            { "id", "WebKit::VideoDecoderIdentifier", nullptr, false },
            { "codecType", "uint8_t", "WebKit::VideoCodecType", false },
            { "useRemoteFrames", "bool", nullptr, false },
            { "enableAdditionalLogging", "bool", nullptr, false },
        };
    case MessageName::LibWebRTCCodecsProxy_ReleaseDecoder:
        return Vector<ArgumentDescription> {
            { "id", "WebKit::VideoDecoderIdentifier", nullptr, false },
        };
    case MessageName::LibWebRTCCodecsProxy_FlushDecoder:
        return Vector<ArgumentDescription> {
            { "id", "WebKit::VideoDecoderIdentifier", nullptr, false },
        };
    case MessageName::LibWebRTCCodecsProxy_SetDecoderFormatDescription:
        return Vector<ArgumentDescription> {
            { "id", "WebKit::VideoDecoderIdentifier", nullptr, false },
            { "description", "IPC::DataReference", nullptr, false },
            { "width", "uint16_t", nullptr, false },
            { "height", "uint16_t", nullptr, false },
        };
    case MessageName::LibWebRTCCodecsProxy_DecodeFrame:
        return Vector<ArgumentDescription> {
            { "id", "WebKit::VideoDecoderIdentifier", nullptr, false },
            { "timeStamp", "int64_t", nullptr, false },
            { "data", "IPC::DataReference", nullptr, false },
        };
    case MessageName::LibWebRTCCodecsProxy_SetFrameSize:
        return Vector<ArgumentDescription> {
            { "id", "WebKit::VideoDecoderIdentifier", nullptr, false },
            { "width", "uint16_t", nullptr, false },
            { "height", "uint16_t", nullptr, false },
        };
    case MessageName::LibWebRTCCodecsProxy_CreateEncoder:
        return Vector<ArgumentDescription> {
            { "id", "WebKit::VideoEncoderIdentifier", nullptr, false },
            { "codecType", "uint8_t", "WebKit::VideoCodecType", false },
            { "parameters", "Vector<std::pair<String, String>>", nullptr, false },
            { "useLowLatency", "bool", nullptr, false },
            { "useAnnexB", "bool", nullptr, false },
        };
    case MessageName::LibWebRTCCodecsProxy_ReleaseEncoder:
        return Vector<ArgumentDescription> {
            { "id", "WebKit::VideoEncoderIdentifier", nullptr, false },
        };
    case MessageName::LibWebRTCCodecsProxy_InitializeEncoder:
        return Vector<ArgumentDescription> {
            { "id", "WebKit::VideoEncoderIdentifier", nullptr, false },
            { "width", "uint16_t", nullptr, false },
            { "height", "uint16_t", nullptr, false },
            { "startBitrate", "unsigned", nullptr, false },
            { "maxBitrate", "unsigned", nullptr, false },
            { "minBitrate", "unsigned", nullptr, false },
            { "maxFramerate", "uint32_t", nullptr, false },
        };
    case MessageName::LibWebRTCCodecsProxy_EncodeFrame:
        return Vector<ArgumentDescription> {
            { "id", "WebKit::VideoEncoderIdentifier", nullptr, false },
            { "buffer", "WebKit::SharedVideoFrame", nullptr, false },
            { "timeStamp", "uint32_t", nullptr, false },
            { "shouldEncodeAsKeyFrame", "bool", nullptr, false },
        };
    case MessageName::LibWebRTCCodecsProxy_FlushEncoder:
        return Vector<ArgumentDescription> {
            { "id", "WebKit::VideoEncoderIdentifier", nullptr, false },
        };
    case MessageName::LibWebRTCCodecsProxy_SetEncodeRates:
        return Vector<ArgumentDescription> {
            { "id", "WebKit::VideoEncoderIdentifier", nullptr, false },
            { "bitRate", "uint32_t", nullptr, false },
            { "frameRate", "uint32_t", nullptr, false },
        };
    case MessageName::LibWebRTCCodecsProxy_SetSharedVideoFrameSemaphore:
        return Vector<ArgumentDescription> {
            { "id", "WebKit::VideoEncoderIdentifier", nullptr, false },
            { "semaphore", "IPC::Semaphore", nullptr, false },
        };
    case MessageName::LibWebRTCCodecsProxy_SetSharedVideoFrameMemory:
        return Vector<ArgumentDescription> {
            { "id", "WebKit::VideoEncoderIdentifier", nullptr, false },
            { "storageHandle", "WebKit::SharedMemory::Handle", nullptr, false },
        };
    case MessageName::LibWebRTCCodecsProxy_SetRTCLoggingLevel:
        return Vector<ArgumentDescription> {
            { "level", "WTFLogLevel", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM) && PLATFORM(COCOA)
    case MessageName::RemoteAudioMediaStreamTrackRendererInternalUnitManager_CreateUnit:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::AudioMediaStreamTrackRendererInternalUnitIdentifier", nullptr, false },
        };
    case MessageName::RemoteAudioMediaStreamTrackRendererInternalUnitManager_DeleteUnit:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::AudioMediaStreamTrackRendererInternalUnitIdentifier", nullptr, false },
        };
    case MessageName::RemoteAudioMediaStreamTrackRendererInternalUnitManager_StartUnit:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::AudioMediaStreamTrackRendererInternalUnitIdentifier", nullptr, false },
            { "storageHandle", "WebKit::ConsumerSharedCARingBuffer::Handle", nullptr, false },
            { "renderSemaphore", "IPC::Semaphore", nullptr, false },
        };
    case MessageName::RemoteAudioMediaStreamTrackRendererInternalUnitManager_StopUnit:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::AudioMediaStreamTrackRendererInternalUnitIdentifier", nullptr, false },
        };
    case MessageName::RemoteAudioMediaStreamTrackRendererInternalUnitManager_SetAudioOutputDevice:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::AudioMediaStreamTrackRendererInternalUnitIdentifier", nullptr, false },
            { "deviceId", "String", nullptr, false },
        };
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
    case MessageName::RemoteMediaRecorder_AudioSamplesStorageChanged:
        return Vector<ArgumentDescription> {
            { "storageHandle", "WebKit::ConsumerSharedCARingBuffer::Handle", nullptr, false },
            { "description", "WebCore::CAAudioStreamDescription", nullptr, false },
        };
    case MessageName::RemoteMediaRecorder_AudioSamplesAvailable:
        return Vector<ArgumentDescription> {
            { "time", "MediaTime", nullptr, false },
            { "numberOfFrames", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteMediaRecorder_VideoFrameAvailable:
        return Vector<ArgumentDescription> {
            { "frame", "WebKit::SharedVideoFrame", nullptr, false },
        };
    case MessageName::RemoteMediaRecorder_FetchData:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaRecorder_StopRecording:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaRecorder_Pause:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaRecorder_Resume:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaRecorder_SetSharedVideoFrameSemaphore:
        return Vector<ArgumentDescription> {
            { "semaphore", "IPC::Semaphore", nullptr, false },
        };
    case MessageName::RemoteMediaRecorder_SetSharedVideoFrameMemory:
        return Vector<ArgumentDescription> {
            { "storageHandle", "WebKit::SharedMemory::Handle", nullptr, false },
        };
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
    case MessageName::RemoteMediaRecorderManager_CreateRecorder:
        return Vector<ArgumentDescription> {
            { "id", "WebKit::MediaRecorderIdentifier", nullptr, false },
            { "hasAudio", "bool", nullptr, false },
            { "hasVideo", "bool", nullptr, false },
            { "options", "WebCore::MediaRecorderPrivateOptions", nullptr, false },
        };
    case MessageName::RemoteMediaRecorderManager_ReleaseRecorder:
        return Vector<ArgumentDescription> {
            { "id", "WebKit::MediaRecorderIdentifier", nullptr, false },
        };
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
#if !RELEASE_LOG_DISABLED
    case MessageName::RemoteSampleBufferDisplayLayer_SetLogIdentifier:
        return Vector<ArgumentDescription> {
            { "logIdentifier", "String", nullptr, false },
        };
#endif
    case MessageName::RemoteSampleBufferDisplayLayer_UpdateDisplayMode:
        return Vector<ArgumentDescription> {
            { "hideDisplayLayer", "bool", nullptr, false },
            { "hideRootLayer", "bool", nullptr, false },
        };
    case MessageName::RemoteSampleBufferDisplayLayer_UpdateAffineTransform:
        return Vector<ArgumentDescription> {
            { "transform", "CGAffineTransform", nullptr, false },
        };
    case MessageName::RemoteSampleBufferDisplayLayer_UpdateBoundsAndPosition:
        return Vector<ArgumentDescription> {
            { "bounds", "CGRect", nullptr, false },
            { "rotation", "WebCore::VideoFrame::Rotation", nullptr, false },
        };
    case MessageName::RemoteSampleBufferDisplayLayer_Flush:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteSampleBufferDisplayLayer_FlushAndRemoveImage:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteSampleBufferDisplayLayer_EnqueueVideoFrame:
        return Vector<ArgumentDescription> {
            { "frame", "WebKit::SharedVideoFrame", nullptr, false },
        };
    case MessageName::RemoteSampleBufferDisplayLayer_ClearVideoFrames:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteSampleBufferDisplayLayer_Play:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteSampleBufferDisplayLayer_Pause:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteSampleBufferDisplayLayer_SetSharedVideoFrameSemaphore:
        return Vector<ArgumentDescription> {
            { "semaphore", "IPC::Semaphore", nullptr, false },
        };
    case MessageName::RemoteSampleBufferDisplayLayer_SetSharedVideoFrameMemory:
        return Vector<ArgumentDescription> {
            { "storageHandle", "WebKit::SharedMemory::Handle", nullptr, false },
        };
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
    case MessageName::RemoteSampleBufferDisplayLayerManager_CreateLayer:
        return Vector<ArgumentDescription> {
            { "id", "WebKit::SampleBufferDisplayLayerIdentifier", nullptr, false },
            { "hideRootLayer", "bool", nullptr, false },
            { "size", "WebCore::IntSize", nullptr, false },
        };
    case MessageName::RemoteSampleBufferDisplayLayerManager_ReleaseLayer:
        return Vector<ArgumentDescription> {
            { "id", "WebKit::SampleBufferDisplayLayerIdentifier", nullptr, false },
        };
#endif
    case MessageName::WebCookieManager_GetHostnamesWithCookies:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::WebCookieManager_DeleteCookiesForHostnames:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "hostnames", "Vector<String>", nullptr, false },
        };
    case MessageName::WebCookieManager_DeleteAllCookies:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::WebCookieManager_SetCookie:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "cookie", "Vector<WebCore::Cookie>", nullptr, false },
        };
    case MessageName::WebCookieManager_SetCookies:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "cookies", "Vector<WebCore::Cookie>", nullptr, false },
            { "url", "URL", nullptr, false },
            { "mainDocumentURL", "URL", nullptr, false },
        };
    case MessageName::WebCookieManager_GetAllCookies:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::WebCookieManager_GetCookies:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "url", "URL", nullptr, false },
        };
    case MessageName::WebCookieManager_DeleteCookie:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "cookie", "WebCore::Cookie", nullptr, false },
        };
    case MessageName::WebCookieManager_DeleteAllCookiesModifiedSince:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "time", "WallTime", nullptr, false },
        };
    case MessageName::WebCookieManager_SetHTTPCookieAcceptPolicy:
        return Vector<ArgumentDescription> {
            { "policy", "uint8_t", "WebCore::HTTPCookieAcceptPolicy", false },
        };
    case MessageName::WebCookieManager_GetHTTPCookieAcceptPolicy:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::WebCookieManager_StartObservingCookieChanges:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::WebCookieManager_StopObservingCookieChanges:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
#if USE(SOUP)
    case MessageName::WebCookieManager_SetCookiePersistentStorage:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "storagePath", "String", nullptr, false },
            { "storageType", "bool", "WebKit::SoupCookiePersistentStorageType", false },
        };
#endif
    case MessageName::NetworkBroadcastChannelRegistry_RegisterChannel:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::ClientOrigin", nullptr, false },
            { "name", "String", nullptr, false },
        };
    case MessageName::NetworkBroadcastChannelRegistry_UnregisterChannel:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::ClientOrigin", nullptr, false },
            { "name", "String", nullptr, false },
        };
    case MessageName::NetworkBroadcastChannelRegistry_PostMessage:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::ClientOrigin", nullptr, false },
            { "name", "String", nullptr, false },
            { "message", "WebCore::MessageWithMessagePorts", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_ScheduleResourceLoad:
        return Vector<ArgumentDescription> {
            { "resourceLoadParameters", "WebKit::NetworkResourceLoadParameters", nullptr, false },
            { "existingLoaderToResume", "WebKit::NetworkResourceLoadIdentifier", nullptr, true },
        };
    case MessageName::NetworkConnectionToWebProcess_PerformSynchronousLoad:
        return Vector<ArgumentDescription> {
            { "resourceLoadParameters", "WebKit::NetworkResourceLoadParameters", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_TestProcessIncomingSyncMessagesWhenWaitingForSyncReply:
        return Vector<ArgumentDescription> {
            { "pageID", "WebKit::WebPageProxyIdentifier", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_LoadPing:
        return Vector<ArgumentDescription> {
            { "resourceLoadParameters", "WebKit::NetworkResourceLoadParameters", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_RemoveLoadIdentifier:
        return Vector<ArgumentDescription> {
            { "resourceLoadIdentifier", "WebCore::ResourceLoaderIdentifier", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_PageLoadCompleted:
        return Vector<ArgumentDescription> {
            { "webPageID", "WebCore::PageIdentifier", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_BrowsingContextRemoved:
        return Vector<ArgumentDescription> {
            { "webPageProxyID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "webPageID", "WebCore::PageIdentifier", nullptr, false },
            { "webFrameID", "WebCore::FrameIdentifier", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_PrefetchDNS:
        return Vector<ArgumentDescription> {
            { "hostname", "String", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_SendH2Ping:
        return Vector<ArgumentDescription> {
            { "parameters", "WebKit::NetworkResourceLoadParameters", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_PreconnectTo:
        return Vector<ArgumentDescription> {
            { "preconnectionIdentifier", "WebCore::ResourceLoaderIdentifier", nullptr, true },
            { "loadParameters", "WebKit::NetworkResourceLoadParameters", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_IsResourceLoadFinished:
        return Vector<ArgumentDescription> {
            { "resourceLoadIdentifier", "WebCore::ResourceLoaderIdentifier", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_StartDownload:
        return Vector<ArgumentDescription> {
            { "downloadID", "WebKit::DownloadID", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "isNavigatingToAppBoundDomain", "WebKit::NavigatingToAppBoundDomain", nullptr, true },
            { "suggestedName", "String", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_ConvertMainResourceLoadToDownload:
        return Vector<ArgumentDescription> {
            { "mainResourceLoadIdentifier", "WebCore::ResourceLoaderIdentifier", nullptr, true },
            { "downloadID", "WebKit::DownloadID", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "response", "WebCore::ResourceResponse", nullptr, false },
            { "isNavigatingToAppBoundDomain", "WebKit::NavigatingToAppBoundDomain", nullptr, true },
        };
    case MessageName::NetworkConnectionToWebProcess_CookiesForDOM:
        return Vector<ArgumentDescription> {
            { "firstParty", "URL", nullptr, false },
            { "sameSiteInfo", "WebCore::SameSiteInfo", nullptr, false },
            { "url", "URL", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "includeSecureCookies", "bool", "WebCore::IncludeSecureCookies", false },
            { "applyTrackingPrevention", "bool", "WebCore::ApplyTrackingPrevention", false },
            { "shouldRelaxThirdPartyCookieBlocking", "bool", "WebCore::ShouldRelaxThirdPartyCookieBlocking", false },
        };
    case MessageName::NetworkConnectionToWebProcess_SetCookiesFromDOM:
        return Vector<ArgumentDescription> {
            { "firstParty", "URL", nullptr, false },
            { "sameSiteInfo", "WebCore::SameSiteInfo", nullptr, false },
            { "url", "URL", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "applyTrackingPrevention", "bool", "WebCore::ApplyTrackingPrevention", false },
            { "cookieString", "String", nullptr, false },
            { "shouldRelaxThirdPartyCookieBlocking", "bool", "WebCore::ShouldRelaxThirdPartyCookieBlocking", false },
        };
    case MessageName::NetworkConnectionToWebProcess_CookieRequestHeaderFieldValue:
        return Vector<ArgumentDescription> {
            { "firstParty", "URL", nullptr, false },
            { "sameSiteInfo", "WebCore::SameSiteInfo", nullptr, false },
            { "url", "URL", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
            { "includeSecureCookies", "bool", "WebCore::IncludeSecureCookies", false },
            { "applyTrackingPrevention", "bool", "WebCore::ApplyTrackingPrevention", false },
            { "shouldRelaxThirdPartyCookieBlocking", "bool", "WebCore::ShouldRelaxThirdPartyCookieBlocking", false },
        };
    case MessageName::NetworkConnectionToWebProcess_GetRawCookies:
        return Vector<ArgumentDescription> {
            { "firstParty", "URL", nullptr, false },
            { "sameSiteInfo", "WebCore::SameSiteInfo", nullptr, false },
            { "url", "URL", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
            { "applyTrackingPrevention", "bool", "WebCore::ApplyTrackingPrevention", false },
            { "shouldRelaxThirdPartyCookieBlocking", "bool", "WebCore::ShouldRelaxThirdPartyCookieBlocking", false },
        };
    case MessageName::NetworkConnectionToWebProcess_SetRawCookie:
        return Vector<ArgumentDescription> {
            { "cookie", "WebCore::Cookie", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_DeleteCookie:
        return Vector<ArgumentDescription> {
            { "url", "URL", nullptr, false },
            { "cookieName", "String", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_DomCookiesForHost:
        return Vector<ArgumentDescription> {
            { "host", "URL", nullptr, false },
            { "subscribeToCookieChangeNotifications", "bool", nullptr, false },
        };
#if HAVE(COOKIE_CHANGE_LISTENER_API)
    case MessageName::NetworkConnectionToWebProcess_UnsubscribeFromCookieChangeNotifications:
        return Vector<ArgumentDescription> {
            { "hosts", "HashSet<String>", nullptr, false },
        };
#endif
    case MessageName::NetworkConnectionToWebProcess_RegisterFileBlobURL:
        return Vector<ArgumentDescription> {
            { "url", "URL", nullptr, false },
            { "path", "String", nullptr, false },
            { "replacementPath", "String", nullptr, false },
            { "extensionHandle", "WebKit::SandboxExtension::Handle", nullptr, false },
            { "contentType", "String", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_RegisterBlobURL:
        return Vector<ArgumentDescription> {
            { "url", "URL", nullptr, false },
            { "blobParts", "Vector<WebCore::BlobPart>", nullptr, false },
            { "contentType", "String", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_RegisterBlobURLFromURL:
        return Vector<ArgumentDescription> {
            { "url", "URL", nullptr, false },
            { "srcURL", "URL", nullptr, false },
            { "policyContainer", "WebCore::PolicyContainer", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_RegisterBlobURLOptionallyFileBacked:
        return Vector<ArgumentDescription> {
            { "url", "URL", nullptr, false },
            { "srcURL", "URL", nullptr, false },
            { "fileBackedPath", "String", nullptr, false },
            { "contentType", "String", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_RegisterBlobURLForSlice:
        return Vector<ArgumentDescription> {
            { "url", "URL", nullptr, false },
            { "srcURL", "URL", nullptr, false },
            { "start", "int64_t", nullptr, false },
            { "end", "int64_t", nullptr, false },
            { "contentType", "String", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_UnregisterBlobURL:
        return Vector<ArgumentDescription> {
            { "url", "URL", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_BlobSize:
        return Vector<ArgumentDescription> {
            { "url", "URL", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_WriteBlobsToTemporaryFilesForIndexedDB:
        return Vector<ArgumentDescription> {
            { "blobURLs", "Vector<String>", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_RegisterBlobURLHandle:
        return Vector<ArgumentDescription> {
            { "url", "URL", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_UnregisterBlobURLHandle:
        return Vector<ArgumentDescription> {
            { "url", "URL", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_SetCaptureExtraNetworkLoadMetricsEnabled:
        return Vector<ArgumentDescription> {
            { "enabled", "bool", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_CreateSocketStream:
        return Vector<ArgumentDescription> {
            { "url", "URL", nullptr, false },
            { "cachePartition", "String", nullptr, false },
            { "identifier", "WebCore::WebSocketIdentifier", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_CreateSocketChannel:
        return Vector<ArgumentDescription> {
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "protocol", "String", nullptr, false },
            { "identifier", "WebCore::WebSocketIdentifier", nullptr, false },
            { "webPageProxyID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
            { "clientOrigin", "WebCore::ClientOrigin", nullptr, false },
            { "hadMainFrameMainResourcePrivateRelayed", "bool", nullptr, false },
            { "allowPrivacyProxy", "bool", nullptr, false },
            { "networkConnectionIntegrityPolicy", "OptionSet<WebCore::NetworkConnectionIntegrity>", nullptr, false },
            { "shouldRelaxThirdPartyCookieBlocking", "bool", "WebCore::ShouldRelaxThirdPartyCookieBlocking", false },
            { "storedCredentialsPolicy", "uint8_t", "WebCore::StoredCredentialsPolicy", false },
        };
    case MessageName::NetworkConnectionToWebProcess_ClearPageSpecificData:
        return Vector<ArgumentDescription> {
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
        };
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::NetworkConnectionToWebProcess_RemoveStorageAccessForFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_LogUserInteraction:
        return Vector<ArgumentDescription> {
            { "domain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_ResourceLoadStatisticsUpdated:
        return Vector<ArgumentDescription> {
            { "statistics", "Vector<WebCore::ResourceLoadStatistics>", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_HasStorageAccess:
        return Vector<ArgumentDescription> {
            { "subFrameDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "topFrameDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_RequestStorageAccess:
        return Vector<ArgumentDescription> {
            { "subFrameDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "topFrameDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "webPageID", "WebCore::PageIdentifier", nullptr, false },
            { "webPageProxyID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "scope", "bool", "WebCore::StorageAccessScope", false },
        };
    case MessageName::NetworkConnectionToWebProcess_RequestStorageAccessUnderOpener:
        return Vector<ArgumentDescription> {
            { "domainInNeedOfStorageAccess", "WebCore::RegistrableDomain", nullptr, false },
            { "openerPageID", "WebCore::PageIdentifier", nullptr, false },
            { "openerDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
#endif
    case MessageName::NetworkConnectionToWebProcess_AddOriginAccessAllowListEntry:
        return Vector<ArgumentDescription> {
            { "sourceOrigin", "String", nullptr, false },
            { "destinationProtocol", "String", nullptr, false },
            { "destinationHost", "String", nullptr, false },
            { "allowDestinationSubdomains", "bool", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_RemoveOriginAccessAllowListEntry:
        return Vector<ArgumentDescription> {
            { "sourceOrigin", "String", nullptr, false },
            { "destinationProtocol", "String", nullptr, false },
            { "destinationHost", "String", nullptr, false },
            { "allowDestinationSubdomains", "bool", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_ResetOriginAccessAllowLists:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkConnectionToWebProcess_GetNetworkLoadInformationResponse:
        return Vector<ArgumentDescription> {
            { "resourceLoadIdentifier", "WebCore::ResourceLoaderIdentifier", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_GetNetworkLoadIntermediateInformation:
        return Vector<ArgumentDescription> {
            { "resourceLoadIdentifier", "WebCore::ResourceLoaderIdentifier", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_TakeNetworkLoadInformationMetrics:
        return Vector<ArgumentDescription> {
            { "resourceLoadIdentifier", "WebCore::ResourceLoaderIdentifier", nullptr, false },
        };
#if ENABLE(SERVICE_WORKER)
    case MessageName::NetworkConnectionToWebProcess_EstablishSWContextConnection:
        return Vector<ArgumentDescription> {
            { "webPageProxyID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "domain", "WebCore::RegistrableDomain", nullptr, false },
            { "serviceWorkerPageIdentifier", "WebCore::ScriptExecutionContextIdentifier", nullptr, true },
        };
    case MessageName::NetworkConnectionToWebProcess_CloseSWContextConnection:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::NetworkConnectionToWebProcess_EstablishSharedWorkerContextConnection:
        return Vector<ArgumentDescription> {
            { "webPageProxyID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "domain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_CloseSharedWorkerContextConnection:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkConnectionToWebProcess_CreateRTCProvider:
        return Vector<ArgumentDescription> { };
#if ENABLE(WEB_RTC)
    case MessageName::NetworkConnectionToWebProcess_ConnectToRTCDataChannelRemoteSource:
        return Vector<ArgumentDescription> {
            { "source", "WebCore::RTCDataChannelIdentifier", nullptr, false },
            { "handler", "WebCore::RTCDataChannelIdentifier", nullptr, false },
        };
#endif
    case MessageName::NetworkConnectionToWebProcess_UpdateQuotaBasedOnSpaceUsageForTesting:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::ClientOrigin", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_CreateNewMessagePortChannel:
        return Vector<ArgumentDescription> {
            { "port1", "WebCore::MessagePortIdentifier", nullptr, false },
            { "port2", "WebCore::MessagePortIdentifier", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_EntangleLocalPortInThisProcessToRemote:
        return Vector<ArgumentDescription> {
            { "local", "WebCore::MessagePortIdentifier", nullptr, false },
            { "remote", "WebCore::MessagePortIdentifier", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_MessagePortDisentangled:
        return Vector<ArgumentDescription> {
            { "local", "WebCore::MessagePortIdentifier", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_MessagePortClosed:
        return Vector<ArgumentDescription> {
            { "local", "WebCore::MessagePortIdentifier", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_TakeAllMessagesForPort:
        return Vector<ArgumentDescription> {
            { "port", "WebCore::MessagePortIdentifier", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_PostMessageToRemote:
        return Vector<ArgumentDescription> {
            { "message", "WebCore::MessageWithMessagePorts", nullptr, false },
            { "remote", "WebCore::MessagePortIdentifier", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_DidDeliverMessagePortMessages:
        return Vector<ArgumentDescription> {
            { "messageBatchIdentifier", "uint64_t", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_RegisterURLSchemesAsCORSEnabled:
        return Vector<ArgumentDescription> {
            { "schemes", "Vector<String>", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_SetCORSDisablingPatterns:
        return Vector<ArgumentDescription> {
            { "pageIdentifier", "WebCore::PageIdentifier", nullptr, false },
            { "patterns", "Vector<String>", nullptr, false },
        };
#if PLATFORM(MAC)
    case MessageName::NetworkConnectionToWebProcess_GetProcessDisplayName:
        return Vector<ArgumentDescription> {
            { "auditToken", "audit_token_t", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_UpdateActivePages:
        return Vector<ArgumentDescription> {
            { "name", "String", nullptr, false },
            { "activePagesOrigins", "Vector<String>", nullptr, false },
            { "auditToken", "audit_token_t", nullptr, false },
        };
#endif
    case MessageName::NetworkConnectionToWebProcess_SetResourceLoadSchedulingMode:
        return Vector<ArgumentDescription> {
            { "webPageID", "WebCore::PageIdentifier", nullptr, false },
            { "mode", "uint8_t", "WebCore::LoadSchedulingMode", false },
        };
    case MessageName::NetworkConnectionToWebProcess_PrioritizeResourceLoads:
        return Vector<ArgumentDescription> {
            { "loadIdentifiers", "Vector<WebCore::ResourceLoaderIdentifier>", nullptr, false },
        };
#if ENABLE(CONTENT_FILTERING_IN_NETWORKING_PROCESS)
    case MessageName::NetworkConnectionToWebProcess_InstallMockContentFilter:
        return Vector<ArgumentDescription> {
            { "settings", "WebCore::MockContentFilterSettings", nullptr, false },
        };
#endif
#if ENABLE(CONTENT_EXTENSIONS)
    case MessageName::NetworkContentRuleListManager_Remove:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::UserContentControllerIdentifier", nullptr, false },
        };
    case MessageName::NetworkContentRuleListManager_AddContentRuleLists:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::UserContentControllerIdentifier", nullptr, false },
            { "contentFilters", "Vector<std::pair<WebKit::WebCompiledContentRuleListData, URL>>", nullptr, false },
        };
    case MessageName::NetworkContentRuleListManager_RemoveContentRuleList:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::UserContentControllerIdentifier", nullptr, false },
            { "name", "String", nullptr, false },
        };
    case MessageName::NetworkContentRuleListManager_RemoveAllContentRuleLists:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::UserContentControllerIdentifier", nullptr, false },
        };
#endif
    case MessageName::NetworkProcess_InitializeNetworkProcess:
        return Vector<ArgumentDescription> {
            { "processCreationParameters", "WebKit::NetworkProcessCreationParameters", nullptr, false },
        };
    case MessageName::NetworkProcess_CreateNetworkConnectionToWebProcess:
        return Vector<ArgumentDescription> {
            { "processIdentifier", "WebCore::ProcessIdentifier", nullptr, false },
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_AddAllowedFirstPartyForCookies:
        return Vector<ArgumentDescription> {
            { "processIdentifier", "WebCore::ProcessIdentifier", nullptr, false },
            { "firstPartyForCookies", "WebCore::RegistrableDomain", nullptr, false },
            { "loadedWebArchive", "bool", "WebKit::LoadedWebArchive", false },
        };
#if USE(SOUP)
    case MessageName::NetworkProcess_SetIgnoreTLSErrors:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "ignoreTLSErrors", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_UserPreferredLanguagesChanged:
        return Vector<ArgumentDescription> {
            { "languages", "Vector<String>", nullptr, false },
        };
    case MessageName::NetworkProcess_SetNetworkProxySettings:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "settings", "WebCore::SoupNetworkProxySettings", nullptr, false },
        };
    case MessageName::NetworkProcess_PrefetchDNS:
        return Vector<ArgumentDescription> {
            { "hostname", "String", nullptr, false },
        };
    case MessageName::NetworkProcess_SetPersistentCredentialStorageEnabled:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "enabled", "bool", nullptr, false },
        };
#endif
#if USE(CURL)
    case MessageName::NetworkProcess_SetNetworkProxySettings:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "settings", "WebCore::CurlProxySettings", nullptr, false },
        };
#endif
    case MessageName::NetworkProcess_AddWebsiteDataStore:
        return Vector<ArgumentDescription> {
            { "websiteDataStoreParameters", "WebKit::WebsiteDataStoreParameters", nullptr, false },
        };
    case MessageName::NetworkProcess_DestroySession:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_FetchWebsiteData:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "websiteDataTypes", "OptionSet<WebKit::WebsiteDataType>", nullptr, false },
            { "fetchOptions", "OptionSet<WebKit::WebsiteDataFetchOption>", nullptr, false },
        };
    case MessageName::NetworkProcess_DeleteWebsiteData:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "websiteDataTypes", "OptionSet<WebKit::WebsiteDataType>", nullptr, false },
            { "modifiedSince", "WallTime", nullptr, false },
        };
    case MessageName::NetworkProcess_DeleteWebsiteDataForOrigins:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "websiteDataTypes", "OptionSet<WebKit::WebsiteDataType>", nullptr, false },
            { "origins", "Vector<WebCore::SecurityOriginData>", nullptr, false },
            { "cookieHostNames", "Vector<String>", nullptr, false },
            { "HSTSCacheHostNames", "Vector<String>", nullptr, false },
            { "registrableDomains", "Vector<WebCore::RegistrableDomain>", nullptr, false },
        };
    case MessageName::NetworkProcess_RenameOriginInWebsiteData:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "oldOrigin", "WebCore::SecurityOriginData", nullptr, false },
            { "newOrigin", "WebCore::SecurityOriginData", nullptr, false },
            { "websiteDataTypes", "OptionSet<WebKit::WebsiteDataType>", nullptr, false },
        };
    case MessageName::NetworkProcess_WebsiteDataOriginDirectoryForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "origin", "WebCore::ClientOrigin", nullptr, false },
            { "websiteDataType", "uint32_t", "WebKit::WebsiteDataType", false },
        };
    case MessageName::NetworkProcess_DownloadRequest:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "downloadID", "WebKit::DownloadID", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "isNavigatingToAppBoundDomain", "WebKit::NavigatingToAppBoundDomain", nullptr, true },
            { "suggestedFilename", "String", nullptr, false },
        };
    case MessageName::NetworkProcess_ResumeDownload:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "downloadID", "WebKit::DownloadID", nullptr, false },
            { "resumeData", "IPC::DataReference", nullptr, false },
            { "path", "String", nullptr, false },
            { "sandboxExtensionHandle", "WebKit::SandboxExtension::Handle", nullptr, false },
            { "callDownloadDidStart", "bool", "WebKit::CallDownloadDidStart", false },
        };
    case MessageName::NetworkProcess_CancelDownload:
        return Vector<ArgumentDescription> {
            { "downloadID", "WebKit::DownloadID", nullptr, false },
        };
#if PLATFORM(COCOA)
    case MessageName::NetworkProcess_PublishDownloadProgress:
        return Vector<ArgumentDescription> {
            { "downloadID", "WebKit::DownloadID", nullptr, false },
            { "url", "URL", nullptr, false },
            { "sandboxExtensionHandle", "WebKit::SandboxExtension::Handle", nullptr, false },
        };
#endif
    case MessageName::NetworkProcess_DataTaskWithRequest:
        return Vector<ArgumentDescription> {
            { "pageID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "requestBody", "IPC::FormDataReference", nullptr, false },
        };
    case MessageName::NetworkProcess_CancelDataTask:
        return Vector<ArgumentDescription> {
            { "taskIdentifier", "WebKit::DataTaskIdentifier", nullptr, false },
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_ApplicationDidEnterBackground:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_ApplicationWillEnterForeground:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_ContinueWillSendRequest:
        return Vector<ArgumentDescription> {
            { "downloadID", "WebKit::DownloadID", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
        };
    case MessageName::NetworkProcess_SetAllowsAnySSLCertificateForWebSocket:
        return Vector<ArgumentDescription> {
            { "enabled", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_FlushCookies:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_AllowSpecificHTTPSCertificateForHost:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "certificate", "WebCore::CertificateInfo", nullptr, false },
            { "host", "String", nullptr, false },
        };
    case MessageName::NetworkProcess_AllowTLSCertificateChainForLocalPCMTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "certificate", "WebCore::CertificateInfo", nullptr, false },
        };
    case MessageName::NetworkProcess_SetCacheModel:
        return Vector<ArgumentDescription> {
            { "cacheModel", "uint8_t", "WebKit::CacheModel", false },
        };
    case MessageName::NetworkProcess_SetCacheModelSynchronouslyForTesting:
        return Vector<ArgumentDescription> {
            { "cacheModel", "uint8_t", "WebKit::CacheModel", false },
        };
    case MessageName::NetworkProcess_ProcessWillSuspendImminentlyForTestingSync:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_PrepareToSuspend:
        return Vector<ArgumentDescription> {
            { "isSuspensionImminent", "bool", nullptr, false },
            { "estimatedSuspendTime", "MonotonicTime", nullptr, false },
        };
    case MessageName::NetworkProcess_ProcessDidResume:
        return Vector<ArgumentDescription> {
            { "forForegroundActivity", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_NotifyMediaStreamingActivity:
        return Vector<ArgumentDescription> {
            { "activity", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_PreconnectTo:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "webPageProxyID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "webPageID", "WebCore::PageIdentifier", nullptr, false },
            { "url", "URL", nullptr, false },
            { "userAgent", "String", nullptr, false },
            { "storedCredentialsPolicy", "uint8_t", "WebCore::StoredCredentialsPolicy", false },
            { "isNavigatingToAppBoundDomain", "WebKit::NavigatingToAppBoundDomain", nullptr, true },
            { "lastNavigationWasAppInitiated", "bool", "WebKit::LastNavigationWasAppInitiated", false },
        };
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::NetworkProcess_ClearPrevalentResource:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "resourceDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_ClearUserInteraction:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "resourceDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_DumpResourceLoadStatistics:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_SetTrackingPreventionEnabled:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "enabled", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_SetResourceLoadStatisticsLogTestingEvent:
        return Vector<ArgumentDescription> {
            { "enabled", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_UpdatePrevalentDomainsToBlockCookiesFor:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "domainsToBlock", "Vector<WebCore::RegistrableDomain>", nullptr, false },
        };
    case MessageName::NetworkProcess_IsGrandfathered:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "targetDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_IsPrevalentResource:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "targetDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_IsVeryPrevalentResource:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "targetDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_SetLastSeen:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "resourceDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "seconds", "Seconds", nullptr, false },
        };
    case MessageName::NetworkProcess_MergeStatisticForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "resourceDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "topFrameDomain1", "WebCore::RegistrableDomain", nullptr, false },
            { "topFrameDomain2", "WebCore::RegistrableDomain", nullptr, false },
            { "lastSeen", "Seconds", nullptr, false },
            { "hadUserInteraction", "bool", nullptr, false },
            { "mostRecentUserInteraction", "Seconds", nullptr, false },
            { "isGrandfathered", "bool", nullptr, false },
            { "isPrevalent", "bool", nullptr, false },
            { "isVeryPrevalent", "bool", nullptr, false },
            { "dataRecordsRemoved", "uint64_t", nullptr, false },
        };
    case MessageName::NetworkProcess_InsertExpiredStatisticForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "resourceDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "numberOfOperatingDaysPassed", "uint64_t", nullptr, false },
            { "hadUserInteraction", "bool", nullptr, false },
            { "isScheduledForAllButCookieDataRemoval", "bool", nullptr, false },
            { "isPrevalent", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_SetPrevalentResource:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "resourceDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_SetPrevalentResourceForDebugMode:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "resourceDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_IsResourceLoadStatisticsEphemeral:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_HadUserInteraction:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "resourceDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_IsRelationshipOnlyInDatabaseOnce:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "subDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "topDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_HasLocalStorage:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "resourceDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_GetAllStorageAccessEntries:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_IsRegisteredAsRedirectingTo:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "redirectedFromDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "redirectedToDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_IsRegisteredAsSubFrameUnder:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "subFrameDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "topFrameDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_IsRegisteredAsSubresourceUnder:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "subresourceDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "topFrameDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_DomainIDExistsInDatabase:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "domainID", "int", nullptr, false },
        };
    case MessageName::NetworkProcess_LogFrameNavigation:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "targetDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "topFrameDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "sourceDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "isRedirect", "bool", nullptr, false },
            { "isMainFrame", "bool", nullptr, false },
            { "delayAfterMainFrameDocumentLoad", "Seconds", nullptr, false },
            { "wasPotentiallyInitiatedByUser", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_LogUserInteraction:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "topFrameDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_ResetParametersToDefaultValues:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_ScheduleClearInMemoryAndPersistent:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "modifiedSince", "WallTime", nullptr, true },
            { "shouldGrandfather", "bool", "WebKit::ShouldGrandfatherStatistics", false },
        };
    case MessageName::NetworkProcess_ScheduleCookieBlockingUpdate:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_ScheduleStatisticsAndDataRecordsProcessing:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_StatisticsDatabaseHasAllTables:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_SetCacheMaxAgeCapForPrevalentResources:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "seconds", "Seconds", nullptr, false },
        };
    case MessageName::NetworkProcess_SetGrandfathered:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "resourceDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "isGrandfathered", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_GetResourceLoadStatisticsDataSummary:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_SetGrandfatheringTime:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "seconds", "Seconds", nullptr, false },
        };
    case MessageName::NetworkProcess_SetMaxStatisticsEntries:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "maximumEntryCount", "uint64_t", nullptr, false },
        };
    case MessageName::NetworkProcess_SetMinimumTimeBetweenDataRecordsRemoval:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "seconds", "Seconds", nullptr, false },
        };
    case MessageName::NetworkProcess_SetPruneEntriesDownTo:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "maximumEntryCount", "uint64_t", nullptr, false },
        };
    case MessageName::NetworkProcess_SetShouldClassifyResourcesBeforeDataRecordsRemoval:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "value", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_SetNotifyPagesWhenDataRecordsWereScanned:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "value", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_SetResourceLoadStatisticsTimeAdvanceForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "time", "Seconds", nullptr, false },
        };
    case MessageName::NetworkProcess_SetIsRunningResourceLoadStatisticsTest:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "value", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_SetResourceLoadStatisticsDebugMode:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "debugMode", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_SetVeryPrevalentResource:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "resourceDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_SetSubframeUnderTopFrameDomain:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "subFrameDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "topFrameDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_SetSubresourceUnderTopFrameDomain:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "subresourceDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "topFrameDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_SetSubresourceUniqueRedirectTo:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "subresourceDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "redirectedToDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_SetSubresourceUniqueRedirectFrom:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "subresourceDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "redirectedFromDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_SetTimeToLiveUserInteraction:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "seconds", "Seconds", nullptr, false },
        };
    case MessageName::NetworkProcess_SetTopFrameUniqueRedirectTo:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "topFrameDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "redirectedToDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_SetTopFrameUniqueRedirectFrom:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "topFrameDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "redirectedFromDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_ResetCacheMaxAgeCapForPrevalentResources:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_DidCommitCrossSiteLoadWithDataTransfer:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "fromDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "toDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "navigationDataTransfer", "OptionSet<WebCore::CrossSiteNavigationDataTransfer::Flag>", nullptr, false },
            { "webPageProxyID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "webPageID", "WebCore::PageIdentifier", nullptr, false },
        };
    case MessageName::NetworkProcess_SetCrossSiteLoadWithLinkDecorationForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "fromDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "toDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_ResetCrossSiteLoadsWithLinkDecorationForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_DeleteCookiesForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "domain", "WebCore::RegistrableDomain", nullptr, false },
            { "includeHttpOnlyCookies", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_HasIsolatedSession:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "domain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_CloseITPDatabase:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
#endif
#if (ENABLE(TRACKING_PREVENTION) && ENABLE(APP_BOUND_DOMAINS))
    case MessageName::NetworkProcess_SetAppBoundDomainsForResourceLoadStatistics:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "appBoundDomains", "HashSet<WebCore::RegistrableDomain>", nullptr, false },
        };
#endif
#if (ENABLE(TRACKING_PREVENTION) && ENABLE(MANAGED_DOMAINS))
    case MessageName::NetworkProcess_SetManagedDomainsForResourceLoadStatistics:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "appBoundDomains", "HashSet<WebCore::RegistrableDomain>", nullptr, false },
        };
#endif
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::NetworkProcess_SetShouldDowngradeReferrerForTesting:
        return Vector<ArgumentDescription> {
            { "enabled", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_SetThirdPartyCookieBlockingMode:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "blockingMode", "uint8_t", "WebCore::ThirdPartyCookieBlockingMode", false },
        };
    case MessageName::NetworkProcess_SetShouldEnbleSameSiteStrictEnforcementForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "enabled", "bool", "WebCore::SameSiteStrictEnforcementEnabled", false },
        };
    case MessageName::NetworkProcess_SetFirstPartyWebsiteDataRemovalModeForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "mode", "uint8_t", "WebCore::FirstPartyWebsiteDataRemovalMode", false },
        };
    case MessageName::NetworkProcess_SetToSameSiteStrictCookiesForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "domain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_SetFirstPartyHostCNAMEDomainForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "firstPartyHost", "String", nullptr, false },
            { "cnameDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcess_SetThirdPartyCNAMEDomainForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "domain", "WebCore::RegistrableDomain", nullptr, false },
        };
#endif
    case MessageName::NetworkProcess_SetPrivateClickMeasurementDebugMode:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "debugMode", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_SetSessionIsControlledByAutomation:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "controlled", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_RegisterURLSchemeAsSecure:
        return Vector<ArgumentDescription> {
            { "scheme", "String", nullptr, false },
        };
    case MessageName::NetworkProcess_RegisterURLSchemeAsBypassingContentSecurityPolicy:
        return Vector<ArgumentDescription> {
            { "scheme", "String", nullptr, false },
        };
    case MessageName::NetworkProcess_RegisterURLSchemeAsLocal:
        return Vector<ArgumentDescription> {
            { "scheme", "String", nullptr, false },
        };
    case MessageName::NetworkProcess_RegisterURLSchemeAsNoAccess:
        return Vector<ArgumentDescription> {
            { "scheme", "String", nullptr, false },
        };
    case MessageName::NetworkProcess_SyncLocalStorage:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_StorePrivateClickMeasurement:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "privateClickMeasurement", "WebCore::PrivateClickMeasurement", nullptr, false },
        };
    case MessageName::NetworkProcess_DumpPrivateClickMeasurement:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_ClearPrivateClickMeasurement:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_SetPrivateClickMeasurementOverrideTimerForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "value", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_MarkAttributedPrivateClickMeasurementsAsExpiredForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_SetPrivateClickMeasurementEphemeralMeasurementForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "value", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_SimulatePrivateClickMeasurementSessionRestart:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_SetPrivateClickMeasurementTokenPublicKeyURLForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "url", "URL", nullptr, false },
        };
    case MessageName::NetworkProcess_SetPrivateClickMeasurementTokenSignatureURLForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "url", "URL", nullptr, false },
        };
    case MessageName::NetworkProcess_SetPrivateClickMeasurementAttributionReportURLsForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "sourceURL", "URL", nullptr, false },
            { "destinationURL", "URL", nullptr, false },
        };
    case MessageName::NetworkProcess_MarkPrivateClickMeasurementsAsExpiredForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_SetPCMFraudPreventionValuesForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "secretToken", "String", nullptr, false },
            { "unlinkableToken", "String", nullptr, false },
            { "signature", "String", nullptr, false },
            { "keyID", "String", nullptr, false },
        };
    case MessageName::NetworkProcess_SetPrivateClickMeasurementAppBundleIDForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "appBundleIDForTesting", "String", nullptr, false },
        };
    case MessageName::NetworkProcess_ClosePCMDatabase:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_SetServiceWorkerFetchTimeoutForTesting:
        return Vector<ArgumentDescription> {
            { "seconds", "Seconds", nullptr, false },
        };
    case MessageName::NetworkProcess_ResetServiceWorkerFetchTimeoutForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_ResetQuota:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
#if PLATFORM(IOS_FAMILY)
    case MessageName::NetworkProcess_SetBackupExclusionPeriodForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "period", "Seconds", nullptr, false },
        };
#endif
    case MessageName::NetworkProcess_ResetStoragePersistedState:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_CloneSessionStorageForWebPage:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "sourcePage", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "destinationPage", "WebKit::WebPageProxyIdentifier", nullptr, false },
        };
    case MessageName::NetworkProcess_DidIncreaseQuota:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "origin", "WebCore::ClientOrigin", nullptr, false },
            { "identifier", "WebKit::QuotaIncreaseRequestIdentifier", nullptr, false },
            { "newQuota", "uint64_t", nullptr, true },
        };
#if ENABLE(APP_BOUND_DOMAINS)
    case MessageName::NetworkProcess_HasAppBoundSession:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_ClearAppBoundSession:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
#endif
#if PLATFORM(IOS_FAMILY) && !PLATFORM(MACCATALYST)
    case MessageName::NetworkProcess_DisableServiceWorkerEntitlement:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_ClearServiceWorkerEntitlementOverride:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::NetworkProcess_UpdateBundleIdentifier:
        return Vector<ArgumentDescription> {
            { "bundleIdentifier", "String", nullptr, false },
        };
    case MessageName::NetworkProcess_ClearBundleIdentifier:
        return Vector<ArgumentDescription> { };
#if PLATFORM(COCOA)
    case MessageName::NetworkProcess_AppPrivacyReportTestingData:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_ClearAppPrivacyReportTestingData:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
#endif
    case MessageName::NetworkProcess_AddWebPageNetworkParameters:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "pageID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "parameters", "WebKit::WebPageNetworkParameters", nullptr, false },
        };
    case MessageName::NetworkProcess_RemoveWebPageNetworkParameters:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "pageID", "WebKit::WebPageProxyIdentifier", nullptr, false },
        };
    case MessageName::NetworkProcess_CountNonDefaultSessionSets:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
#if ENABLE(CFPREFS_DIRECT_MODE)
    case MessageName::NetworkProcess_NotifyPreferencesChanged:
        return Vector<ArgumentDescription> {
            { "domain", "String", nullptr, false },
            { "key", "String", nullptr, false },
            { "encodedValue", "String", nullptr, true },
        };
#endif
    case MessageName::NetworkProcess_TerminateRemoteWorkerContextConnectionWhenPossible:
        return Vector<ArgumentDescription> {
            { "workerType", "bool", "WebKit::RemoteWorkerType", false },
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "registrableDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "processIdentifier", "WebCore::ProcessIdentifier", nullptr, false },
        };
#if ENABLE(SERVICE_WORKER)
    case MessageName::NetworkProcess_GetPendingPushMessages:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_ProcessPushMessage:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "pushMessage", "WebKit::WebPushMessage", nullptr, false },
            { "pushPermissionState", "uint8_t", "WebCore::PushPermissionState", false },
        };
    case MessageName::NetworkProcess_ProcessNotificationEvent:
        return Vector<ArgumentDescription> {
            { "data", "WebCore::NotificationData", nullptr, false },
            { "eventType", "bool", "WebCore::NotificationEventType", false },
        };
#endif
    case MessageName::NetworkProcess_SetPushAndNotificationsEnabledForOrigin:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "origin", "WebCore::SecurityOriginData", nullptr, false },
            { "enabled", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_DeletePushAndNotificationRegistration:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "origin", "WebCore::SecurityOriginData", nullptr, false },
        };
    case MessageName::NetworkProcess_GetOriginsWithPushAndNotificationPermissions:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcess_HasPushSubscriptionForTesting:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "scopeURL", "URL", nullptr, false },
        };
#if ENABLE(INSPECTOR_NETWORK_THROTTLING)
    case MessageName::NetworkProcess_SetEmulatedConditions:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "bytesPerSecondLimit", "int64_t", nullptr, true },
        };
#endif
    case MessageName::NetworkResourceLoader_ContinueWillSendRequest:
        return Vector<ArgumentDescription> {
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "isAllowedToAskUserForCredentials", "bool", nullptr, false },
        };
    case MessageName::NetworkResourceLoader_ContinueDidReceiveResponse:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkSocketChannel_SendString:
        return Vector<ArgumentDescription> {
            { "message", "IPC::DataReference", nullptr, false },
        };
    case MessageName::NetworkSocketChannel_SendData:
        return Vector<ArgumentDescription> {
            { "data", "IPC::DataReference", nullptr, false },
        };
    case MessageName::NetworkSocketChannel_Close:
        return Vector<ArgumentDescription> {
            { "code", "int32_t", nullptr, false },
            { "reason", "String", nullptr, false },
        };
    case MessageName::NetworkSocketStream_SendData:
        return Vector<ArgumentDescription> {
            { "data", "IPC::DataReference", nullptr, false },
            { "identifier", "uint64_t", nullptr, false },
        };
    case MessageName::NetworkSocketStream_SendHandshake:
        return Vector<ArgumentDescription> {
            { "data", "IPC::DataReference", nullptr, false },
            { "headerFieldProxy", "WebCore::CookieRequestHeaderFieldProxy", nullptr, true },
            { "identifier", "uint64_t", nullptr, false },
        };
    case MessageName::NetworkSocketStream_Close:
        return Vector<ArgumentDescription> { };
#if ENABLE(SERVICE_WORKER)
    case MessageName::ServiceWorkerDownloadTask_DidFail:
        return Vector<ArgumentDescription> {
            { "error", "WebCore::ResourceError", nullptr, false },
        };
    case MessageName::ServiceWorkerDownloadTask_DidReceiveData:
        return Vector<ArgumentDescription> {
            { "data", "IPC::SharedBufferReference", nullptr, false },
            { "encodedDataLength", "uint64_t", nullptr, false },
        };
    case MessageName::ServiceWorkerDownloadTask_DidReceiveFormData:
        return Vector<ArgumentDescription> {
            { "data", "IPC::FormDataReference", nullptr, false },
        };
    case MessageName::ServiceWorkerDownloadTask_DidFinish:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(SERVICE_WORKER)
    case MessageName::ServiceWorkerFetchTask_DidNotHandle:
        return Vector<ArgumentDescription> { };
    case MessageName::ServiceWorkerFetchTask_DidFail:
        return Vector<ArgumentDescription> {
            { "error", "WebCore::ResourceError", nullptr, false },
        };
    case MessageName::ServiceWorkerFetchTask_DidReceiveRedirectResponse:
        return Vector<ArgumentDescription> {
            { "response", "WebCore::ResourceResponse", nullptr, false },
        };
    case MessageName::ServiceWorkerFetchTask_DidReceiveResponse:
        return Vector<ArgumentDescription> {
            { "response", "WebCore::ResourceResponse", nullptr, false },
            { "needsContinueDidReceiveResponseMessage", "bool", nullptr, false },
        };
    case MessageName::ServiceWorkerFetchTask_DidReceiveData:
        return Vector<ArgumentDescription> {
            { "data", "IPC::SharedBufferReference", nullptr, false },
            { "encodedDataLength", "uint64_t", nullptr, false },
        };
    case MessageName::ServiceWorkerFetchTask_DidReceiveFormData:
        return Vector<ArgumentDescription> {
            { "data", "IPC::FormDataReference", nullptr, false },
        };
    case MessageName::ServiceWorkerFetchTask_DidFinish:
        return Vector<ArgumentDescription> {
            { "metrics", "WebCore::NetworkLoadMetrics", nullptr, false },
        };
    case MessageName::ServiceWorkerFetchTask_UsePreload:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(SERVICE_WORKER)
    case MessageName::WebSWServerConnection_ScheduleJobInServer:
        return Vector<ArgumentDescription> {
            { "jobData", "WebCore::ServiceWorkerJobData", nullptr, false },
        };
    case MessageName::WebSWServerConnection_ScheduleUnregisterJobInServer:
        return Vector<ArgumentDescription> {
            { "jobIdentifier", "WebCore::ServiceWorkerJobIdentifier", nullptr, false },
            { "identifier", "WebCore::ServiceWorkerRegistrationIdentifier", nullptr, false },
            { "documentIdentifier", "WebCore::ServiceWorkerOrClientIdentifier", nullptr, false },
        };
    case MessageName::WebSWServerConnection_FinishFetchingScriptInServer:
        return Vector<ArgumentDescription> {
            { "jobDataIdentifier", "WebCore::ServiceWorkerJobDataIdentifier", nullptr, false },
            { "registrationKey", "WebCore::ServiceWorkerRegistrationKey", nullptr, false },
            { "result", "WebCore::WorkerFetchResult", nullptr, false },
        };
    case MessageName::WebSWServerConnection_AddServiceWorkerRegistrationInServer:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerRegistrationIdentifier", nullptr, false },
        };
    case MessageName::WebSWServerConnection_RemoveServiceWorkerRegistrationInServer:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerRegistrationIdentifier", nullptr, false },
        };
    case MessageName::WebSWServerConnection_PostMessageToServiceWorker:
        return Vector<ArgumentDescription> {
            { "destination", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "message", "WebCore::MessageWithMessagePorts", nullptr, false },
            { "source", "WebCore::ServiceWorkerOrClientIdentifier", nullptr, false },
        };
    case MessageName::WebSWServerConnection_DidResolveRegistrationPromise:
        return Vector<ArgumentDescription> {
            { "key", "WebCore::ServiceWorkerRegistrationKey", nullptr, false },
        };
    case MessageName::WebSWServerConnection_MatchRegistration:
        return Vector<ArgumentDescription> {
            { "topOrigin", "WebCore::SecurityOriginData", nullptr, false },
            { "clientURL", "URL", nullptr, false },
        };
    case MessageName::WebSWServerConnection_WhenRegistrationReady:
        return Vector<ArgumentDescription> {
            { "topOrigin", "WebCore::SecurityOriginData", nullptr, false },
            { "clientURL", "URL", nullptr, false },
        };
    case MessageName::WebSWServerConnection_GetRegistrations:
        return Vector<ArgumentDescription> {
            { "topOrigin", "WebCore::SecurityOriginData", nullptr, false },
            { "clientURL", "URL", nullptr, false },
        };
    case MessageName::WebSWServerConnection_RegisterServiceWorkerClient:
        return Vector<ArgumentDescription> {
            { "clientOrigin", "WebCore::ClientOrigin", nullptr, false },
            { "data", "WebCore::ServiceWorkerClientData", nullptr, false },
            { "controllingServiceWorkerRegistrationIdentifier", "WebCore::ServiceWorkerRegistrationIdentifier", nullptr, true },
            { "userAgent", "String", nullptr, false },
        };
    case MessageName::WebSWServerConnection_UnregisterServiceWorkerClient:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ScriptExecutionContextIdentifier", nullptr, false },
        };
    case MessageName::WebSWServerConnection_TerminateWorkerFromClient:
        return Vector<ArgumentDescription> {
            { "workerIdentifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
        };
    case MessageName::WebSWServerConnection_WhenServiceWorkerIsTerminatedForTesting:
        return Vector<ArgumentDescription> {
            { "workerIdentifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
        };
    case MessageName::WebSWServerConnection_SetThrottleState:
        return Vector<ArgumentDescription> {
            { "isThrottleable", "bool", nullptr, false },
        };
    case MessageName::WebSWServerConnection_StoreRegistrationsOnDisk:
        return Vector<ArgumentDescription> { };
    case MessageName::WebSWServerConnection_SubscribeToPushService:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerRegistrationIdentifier", nullptr, false },
            { "applicationServerKey", "Vector<uint8_t>", nullptr, false },
        };
    case MessageName::WebSWServerConnection_UnsubscribeFromPushService:
        return Vector<ArgumentDescription> {
            { "serviceWorkerRegistrationIdentifier", "WebCore::ServiceWorkerRegistrationIdentifier", nullptr, false },
            { "pushSubscriptionIdentifier", "WebCore::PushSubscriptionIdentifier", nullptr, false },
        };
    case MessageName::WebSWServerConnection_GetPushSubscription:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerRegistrationIdentifier", nullptr, false },
        };
    case MessageName::WebSWServerConnection_GetPushPermissionState:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerRegistrationIdentifier", nullptr, false },
        };
    case MessageName::WebSWServerConnection_EnableNavigationPreload:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerRegistrationIdentifier", nullptr, false },
        };
    case MessageName::WebSWServerConnection_DisableNavigationPreload:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerRegistrationIdentifier", nullptr, false },
        };
    case MessageName::WebSWServerConnection_SetNavigationPreloadHeaderValue:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerRegistrationIdentifier", nullptr, false },
            { "value", "String", nullptr, false },
        };
    case MessageName::WebSWServerConnection_GetNavigationPreloadState:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerRegistrationIdentifier", nullptr, false },
        };
#endif
#if ENABLE(SERVICE_WORKER)
    case MessageName::WebSWServerToContextConnection_ScriptContextFailedToStart:
        return Vector<ArgumentDescription> {
            { "jobDataIdentifier", "WebCore::ServiceWorkerJobDataIdentifier", nullptr, true },
            { "serviceWorkerIdentifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "message", "String", nullptr, false },
        };
    case MessageName::WebSWServerToContextConnection_ScriptContextStarted:
        return Vector<ArgumentDescription> {
            { "jobDataIdentifier", "WebCore::ServiceWorkerJobDataIdentifier", nullptr, true },
            { "serviceWorkerIdentifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "doesHandleFetch", "bool", nullptr, false },
        };
    case MessageName::WebSWServerToContextConnection_DidFinishInstall:
        return Vector<ArgumentDescription> {
            { "jobDataIdentifier", "WebCore::ServiceWorkerJobDataIdentifier", nullptr, true },
            { "serviceWorkerIdentifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "wasSuccessful", "bool", nullptr, false },
        };
    case MessageName::WebSWServerToContextConnection_DidFinishActivation:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
        };
    case MessageName::WebSWServerToContextConnection_SetServiceWorkerHasPendingEvents:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "hasPendingEvents", "bool", nullptr, false },
        };
    case MessageName::WebSWServerToContextConnection_SkipWaiting:
        return Vector<ArgumentDescription> {
            { "matchAllRequestIdentifier", "uint64_t", nullptr, false },
            { "identifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
        };
    case MessageName::WebSWServerToContextConnection_WorkerTerminated:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
        };
    case MessageName::WebSWServerToContextConnection_FindClientByVisibleIdentifier:
        return Vector<ArgumentDescription> {
            { "serviceWorkerIdentifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "clientIdentifier", "String", nullptr, false },
        };
    case MessageName::WebSWServerToContextConnection_MatchAll:
        return Vector<ArgumentDescription> {
            { "matchAllRequestIdentifier", "uint64_t", nullptr, false },
            { "serviceWorkerIdentifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "options", "WebCore::ServiceWorkerClientQueryOptions", nullptr, false },
        };
    case MessageName::WebSWServerToContextConnection_Claim:
        return Vector<ArgumentDescription> {
            { "serviceWorkerIdentifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
        };
    case MessageName::WebSWServerToContextConnection_Focus:
        return Vector<ArgumentDescription> {
            { "serviceWorkerClientIdentifier", "WebCore::ScriptExecutionContextIdentifier", nullptr, false },
        };
    case MessageName::WebSWServerToContextConnection_Navigate:
        return Vector<ArgumentDescription> {
            { "clientIdentifier", "WebCore::ScriptExecutionContextIdentifier", nullptr, false },
            { "serviceWorkerIdentifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "url", "URL", nullptr, false },
        };
    case MessageName::WebSWServerToContextConnection_SetScriptResource:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "scriptURL", "URL", nullptr, false },
            { "script", "WebCore::ServiceWorkerContextData::ImportedScript", nullptr, false },
        };
    case MessageName::WebSWServerToContextConnection_PostMessageToServiceWorkerClient:
        return Vector<ArgumentDescription> {
            { "destination", "WebCore::ScriptExecutionContextIdentifier", nullptr, false },
            { "message", "WebCore::MessageWithMessagePorts", nullptr, false },
            { "source", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "sourceOrigin", "String", nullptr, false },
        };
    case MessageName::WebSWServerToContextConnection_DidFailHeartBeatCheck:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
        };
    case MessageName::WebSWServerToContextConnection_SetAsInspected:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "isInspected", "bool", nullptr, false },
        };
    case MessageName::WebSWServerToContextConnection_OpenWindow:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "url", "URL", nullptr, false },
        };
#endif
    case MessageName::WebSharedWorkerServerConnection_RequestSharedWorker:
        return Vector<ArgumentDescription> {
            { "sharedWorkerKey", "WebCore::SharedWorkerKey", nullptr, false },
            { "sharedWorkerObjectIdentifier", "WebCore::SharedWorkerObjectIdentifier", nullptr, false },
            { "port", "WebCore::TransferredMessagePort", nullptr, false },
            { "workerOptions", "WebCore::WorkerOptions", nullptr, false },
        };
    case MessageName::WebSharedWorkerServerConnection_SharedWorkerObjectIsGoingAway:
        return Vector<ArgumentDescription> {
            { "sharedWorkerKey", "WebCore::SharedWorkerKey", nullptr, false },
            { "sharedWorkerObjectIdentifier", "WebCore::SharedWorkerObjectIdentifier", nullptr, false },
        };
    case MessageName::WebSharedWorkerServerConnection_SuspendForBackForwardCache:
        return Vector<ArgumentDescription> {
            { "sharedWorkerKey", "WebCore::SharedWorkerKey", nullptr, false },
            { "sharedWorkerObjectIdentifier", "WebCore::SharedWorkerObjectIdentifier", nullptr, false },
        };
    case MessageName::WebSharedWorkerServerConnection_ResumeForBackForwardCache:
        return Vector<ArgumentDescription> {
            { "sharedWorkerKey", "WebCore::SharedWorkerKey", nullptr, false },
            { "sharedWorkerObjectIdentifier", "WebCore::SharedWorkerObjectIdentifier", nullptr, false },
        };
    case MessageName::WebSharedWorkerServerToContextConnection_PostExceptionToWorkerObject:
        return Vector<ArgumentDescription> {
            { "sharedWorkerIdentifier", "WebCore::SharedWorkerIdentifier", nullptr, false },
            { "errorMessage", "String", nullptr, false },
            { "lineNumber", "int", nullptr, false },
            { "columnNumber", "int", nullptr, false },
            { "sourceURL", "String", nullptr, false },
        };
    case MessageName::WebSharedWorkerServerToContextConnection_SharedWorkerTerminated:
        return Vector<ArgumentDescription> {
            { "sharedWorkerIdentifier", "WebCore::SharedWorkerIdentifier", nullptr, false },
        };
    case MessageName::CacheStorageEngineConnection_Reference:
        return Vector<ArgumentDescription> {
            { "cacheIdentifier", "WebCore::DOMCacheIdentifier", nullptr, false },
        };
    case MessageName::CacheStorageEngineConnection_Dereference:
        return Vector<ArgumentDescription> {
            { "cacheIdentifier", "WebCore::DOMCacheIdentifier", nullptr, false },
        };
    case MessageName::CacheStorageEngineConnection_Open:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::ClientOrigin", nullptr, false },
            { "cacheName", "String", nullptr, false },
        };
    case MessageName::CacheStorageEngineConnection_Remove:
        return Vector<ArgumentDescription> {
            { "cacheIdentifier", "WebCore::DOMCacheIdentifier", nullptr, false },
        };
    case MessageName::CacheStorageEngineConnection_Caches:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::ClientOrigin", nullptr, false },
            { "updateCounter", "uint64_t", nullptr, false },
        };
    case MessageName::CacheStorageEngineConnection_RetrieveRecords:
        return Vector<ArgumentDescription> {
            { "cacheIdentifier", "WebCore::DOMCacheIdentifier", nullptr, false },
            { "options", "WebCore::RetrieveRecordsOptions", nullptr, false },
        };
    case MessageName::CacheStorageEngineConnection_DeleteMatchingRecords:
        return Vector<ArgumentDescription> {
            { "cacheIdentifier", "WebCore::DOMCacheIdentifier", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "options", "WebCore::CacheQueryOptions", nullptr, false },
        };
    case MessageName::CacheStorageEngineConnection_PutRecords:
        return Vector<ArgumentDescription> {
            { "cacheIdentifier", "WebCore::DOMCacheIdentifier", nullptr, false },
            { "record", "Vector<WebCore::DOMCacheEngine::Record>", nullptr, false },
        };
    case MessageName::CacheStorageEngineConnection_ClearMemoryRepresentation:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::ClientOrigin", nullptr, false },
        };
    case MessageName::CacheStorageEngineConnection_EngineRepresentation:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkStorageManager_Persisted:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::ClientOrigin", nullptr, false },
        };
    case MessageName::NetworkStorageManager_Persist:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::ClientOrigin", nullptr, false },
        };
    case MessageName::NetworkStorageManager_Estimate:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::ClientOrigin", nullptr, false },
        };
    case MessageName::NetworkStorageManager_FileSystemGetDirectory:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::ClientOrigin", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CloseHandle:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::FileSystemHandleIdentifier", nullptr, false },
        };
    case MessageName::NetworkStorageManager_IsSameEntry:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::FileSystemHandleIdentifier", nullptr, false },
            { "targetIdentifier", "WebCore::FileSystemHandleIdentifier", nullptr, false },
        };
    case MessageName::NetworkStorageManager_GetFileHandle:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::FileSystemHandleIdentifier", nullptr, false },
            { "name", "String", nullptr, false },
            { "createIfNecessary", "bool", nullptr, false },
        };
    case MessageName::NetworkStorageManager_GetDirectoryHandle:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::FileSystemHandleIdentifier", nullptr, false },
            { "name", "String", nullptr, false },
            { "createIfNecessary", "bool", nullptr, false },
        };
    case MessageName::NetworkStorageManager_RemoveEntry:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::FileSystemHandleIdentifier", nullptr, false },
            { "name", "String", nullptr, false },
            { "deleteRecursively", "bool", nullptr, false },
        };
    case MessageName::NetworkStorageManager_Resolve:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::FileSystemHandleIdentifier", nullptr, false },
            { "targetIdentifier", "WebCore::FileSystemHandleIdentifier", nullptr, false },
        };
    case MessageName::NetworkStorageManager_Move:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::FileSystemHandleIdentifier", nullptr, false },
            { "destinationIdentifier", "WebCore::FileSystemHandleIdentifier", nullptr, false },
            { "newName", "String", nullptr, false },
        };
    case MessageName::NetworkStorageManager_GetFile:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::FileSystemHandleIdentifier", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CreateSyncAccessHandle:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::FileSystemHandleIdentifier", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CloseSyncAccessHandle:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::FileSystemHandleIdentifier", nullptr, false },
            { "accessHandleIdentifier", "WebCore::FileSystemSyncAccessHandleIdentifier", nullptr, false },
        };
    case MessageName::NetworkStorageManager_RequestNewCapacityForSyncAccessHandle:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::FileSystemHandleIdentifier", nullptr, false },
            { "accessHandleIdentifier", "WebCore::FileSystemSyncAccessHandleIdentifier", nullptr, false },
            { "newCapacity", "uint64_t", nullptr, false },
        };
    case MessageName::NetworkStorageManager_GetHandleNames:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::FileSystemHandleIdentifier", nullptr, false },
        };
    case MessageName::NetworkStorageManager_GetHandle:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::FileSystemHandleIdentifier", nullptr, false },
            { "name", "String", nullptr, false },
        };
    case MessageName::NetworkStorageManager_ConnectToStorageArea:
        return Vector<ArgumentDescription> {
            { "type", "WebCore::StorageType", nullptr, false },
            { "sourceIdentifier", "WebKit::StorageAreaMapIdentifier", nullptr, false },
            { "namespaceIdentifier", "WebKit::StorageNamespaceIdentifier", nullptr, true },
            { "origin", "WebCore::ClientOrigin", nullptr, false },
        };
    case MessageName::NetworkStorageManager_ConnectToStorageAreaSync:
        return Vector<ArgumentDescription> {
            { "type", "WebCore::StorageType", nullptr, false },
            { "sourceIdentifier", "WebKit::StorageAreaMapIdentifier", nullptr, false },
            { "namespaceIdentifier", "WebKit::StorageNamespaceIdentifier", nullptr, true },
            { "origin", "WebCore::ClientOrigin", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CancelConnectToStorageArea:
        return Vector<ArgumentDescription> {
            { "type", "WebCore::StorageType", nullptr, false },
            { "namespaceIdentifier", "WebKit::StorageNamespaceIdentifier", nullptr, true },
            { "origin", "WebCore::ClientOrigin", nullptr, false },
        };
    case MessageName::NetworkStorageManager_DisconnectFromStorageArea:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::StorageAreaIdentifier", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CloneSessionStorageNamespace:
        return Vector<ArgumentDescription> {
            { "fromStorageNamespaceID", "WebKit::StorageNamespaceIdentifier", nullptr, false },
            { "toStorageNamespaceID", "WebKit::StorageNamespaceIdentifier", nullptr, false },
        };
    case MessageName::NetworkStorageManager_SetItem:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::StorageAreaIdentifier", nullptr, false },
            { "implIdentifier", "WebKit::StorageAreaImplIdentifier", nullptr, false },
            { "key", "String", nullptr, false },
            { "value", "String", nullptr, false },
            { "urlString", "String", nullptr, false },
        };
    case MessageName::NetworkStorageManager_RemoveItem:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::StorageAreaIdentifier", nullptr, false },
            { "implIdentifier", "WebKit::StorageAreaImplIdentifier", nullptr, false },
            { "key", "String", nullptr, false },
            { "urlString", "String", nullptr, false },
        };
    case MessageName::NetworkStorageManager_Clear:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::StorageAreaIdentifier", nullptr, false },
            { "implIdentifier", "WebKit::StorageAreaImplIdentifier", nullptr, false },
            { "urlString", "String", nullptr, false },
        };
    case MessageName::NetworkStorageManager_OpenDatabase:
        return Vector<ArgumentDescription> {
            { "requestData", "WebCore::IDBRequestData", nullptr, false },
        };
    case MessageName::NetworkStorageManager_OpenDBRequestCancelled:
        return Vector<ArgumentDescription> {
            { "requestData", "WebCore::IDBRequestData", nullptr, false },
        };
    case MessageName::NetworkStorageManager_DeleteDatabase:
        return Vector<ArgumentDescription> {
            { "requestData", "WebCore::IDBRequestData", nullptr, false },
        };
    case MessageName::NetworkStorageManager_EstablishTransaction:
        return Vector<ArgumentDescription> {
            { "databaseConnectionIdentifier", "uint64_t", nullptr, false },
            { "info", "WebCore::IDBTransactionInfo", nullptr, false },
        };
    case MessageName::NetworkStorageManager_DatabaseConnectionPendingClose:
        return Vector<ArgumentDescription> {
            { "databaseConnectionIdentifier", "uint64_t", nullptr, false },
        };
    case MessageName::NetworkStorageManager_DatabaseConnectionClosed:
        return Vector<ArgumentDescription> {
            { "databaseConnectionIdentifier", "uint64_t", nullptr, false },
        };
    case MessageName::NetworkStorageManager_AbortOpenAndUpgradeNeeded:
        return Vector<ArgumentDescription> {
            { "databaseConnectionIdentifier", "uint64_t", nullptr, false },
            { "transactionIdentifier", "WebCore::IDBResourceIdentifier", nullptr, true },
        };
    case MessageName::NetworkStorageManager_DidFireVersionChangeEvent:
        return Vector<ArgumentDescription> {
            { "databaseConnectionIdentifier", "uint64_t", nullptr, false },
            { "requestIdentifier", "WebCore::IDBResourceIdentifier", nullptr, false },
            { "connectionClosedOnBehalfOfServer", "WebCore::IndexedDB::ConnectionClosedOnBehalfOfServer", nullptr, false },
        };
    case MessageName::NetworkStorageManager_DidFinishHandlingVersionChangeTransaction:
        return Vector<ArgumentDescription> {
            { "databaseConnectionIdentifier", "uint64_t", nullptr, false },
            { "transactionIdentifier", "WebCore::IDBResourceIdentifier", nullptr, false },
        };
    case MessageName::NetworkStorageManager_AbortTransaction:
        return Vector<ArgumentDescription> {
            { "transactionIdentifier", "WebCore::IDBResourceIdentifier", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CommitTransaction:
        return Vector<ArgumentDescription> {
            { "transactionIdentifier", "WebCore::IDBResourceIdentifier", nullptr, false },
            { "pendingRequestCount", "uint64_t", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CreateObjectStore:
        return Vector<ArgumentDescription> {
            { "requestData", "WebCore::IDBRequestData", nullptr, false },
            { "info", "WebCore::IDBObjectStoreInfo", nullptr, false },
        };
    case MessageName::NetworkStorageManager_DeleteObjectStore:
        return Vector<ArgumentDescription> {
            { "requestData", "WebCore::IDBRequestData", nullptr, false },
            { "objectStoreName", "String", nullptr, false },
        };
    case MessageName::NetworkStorageManager_RenameObjectStore:
        return Vector<ArgumentDescription> {
            { "requestData", "WebCore::IDBRequestData", nullptr, false },
            { "objectStoreIdentifier", "uint64_t", nullptr, false },
            { "newName", "String", nullptr, false },
        };
    case MessageName::NetworkStorageManager_ClearObjectStore:
        return Vector<ArgumentDescription> {
            { "requestData", "WebCore::IDBRequestData", nullptr, false },
            { "objectStoreIdentifier", "uint64_t", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CreateIndex:
        return Vector<ArgumentDescription> {
            { "requestData", "WebCore::IDBRequestData", nullptr, false },
            { "info", "WebCore::IDBIndexInfo", nullptr, false },
        };
    case MessageName::NetworkStorageManager_DeleteIndex:
        return Vector<ArgumentDescription> {
            { "requestData", "WebCore::IDBRequestData", nullptr, false },
            { "objectStoreIdentifier", "uint64_t", nullptr, false },
            { "indexName", "String", nullptr, false },
        };
    case MessageName::NetworkStorageManager_RenameIndex:
        return Vector<ArgumentDescription> {
            { "requestData", "WebCore::IDBRequestData", nullptr, false },
            { "objectStoreIdentifier", "uint64_t", nullptr, false },
            { "indexIdentifier", "uint64_t", nullptr, false },
            { "newName", "String", nullptr, false },
        };
    case MessageName::NetworkStorageManager_PutOrAdd:
        return Vector<ArgumentDescription> {
            { "requestData", "WebCore::IDBRequestData", nullptr, false },
            { "key", "WebCore::IDBKeyData", nullptr, false },
            { "value", "WebCore::IDBValue", nullptr, false },
            { "overwriteMode", "WebCore::IndexedDB::ObjectStoreOverwriteMode", nullptr, false },
        };
    case MessageName::NetworkStorageManager_GetRecord:
        return Vector<ArgumentDescription> {
            { "requestData", "WebCore::IDBRequestData", nullptr, false },
            { "getRecordData", "WebCore::IDBGetRecordData", nullptr, false },
        };
    case MessageName::NetworkStorageManager_GetAllRecords:
        return Vector<ArgumentDescription> {
            { "requestData", "WebCore::IDBRequestData", nullptr, false },
            { "getAllRecordsData", "WebCore::IDBGetAllRecordsData", nullptr, false },
        };
    case MessageName::NetworkStorageManager_GetCount:
        return Vector<ArgumentDescription> {
            { "requestData", "WebCore::IDBRequestData", nullptr, false },
            { "range", "WebCore::IDBKeyRangeData", nullptr, false },
        };
    case MessageName::NetworkStorageManager_DeleteRecord:
        return Vector<ArgumentDescription> {
            { "requestData", "WebCore::IDBRequestData", nullptr, false },
            { "range", "WebCore::IDBKeyRangeData", nullptr, false },
        };
    case MessageName::NetworkStorageManager_OpenCursor:
        return Vector<ArgumentDescription> {
            { "requestData", "WebCore::IDBRequestData", nullptr, false },
            { "info", "WebCore::IDBCursorInfo", nullptr, false },
        };
    case MessageName::NetworkStorageManager_IterateCursor:
        return Vector<ArgumentDescription> {
            { "requestData", "WebCore::IDBRequestData", nullptr, false },
            { "data", "WebCore::IDBIterateCursorData", nullptr, false },
        };
    case MessageName::NetworkStorageManager_GetAllDatabaseNamesAndVersions:
        return Vector<ArgumentDescription> {
            { "requestIdentifier", "WebCore::IDBResourceIdentifier", nullptr, false },
            { "origin", "WebCore::ClientOrigin", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CacheStorageOpenCache:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::ClientOrigin", nullptr, false },
            { "cacheName", "String", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CacheStorageRemoveCache:
        return Vector<ArgumentDescription> {
            { "cacheIdentifier", "WebCore::DOMCacheIdentifier", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CacheStorageAllCaches:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::ClientOrigin", nullptr, false },
            { "updateCounter", "uint64_t", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CacheStorageReference:
        return Vector<ArgumentDescription> {
            { "cacheIdentifier", "WebCore::DOMCacheIdentifier", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CacheStorageDereference:
        return Vector<ArgumentDescription> {
            { "cacheIdentifier", "WebCore::DOMCacheIdentifier", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CacheStorageRetrieveRecords:
        return Vector<ArgumentDescription> {
            { "cacheIdentifier", "WebCore::DOMCacheIdentifier", nullptr, false },
            { "options", "WebCore::RetrieveRecordsOptions", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CacheStorageRemoveRecords:
        return Vector<ArgumentDescription> {
            { "cacheIdentifier", "WebCore::DOMCacheIdentifier", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "options", "WebCore::CacheQueryOptions", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CacheStoragePutRecords:
        return Vector<ArgumentDescription> {
            { "cacheIdentifier", "WebCore::DOMCacheIdentifier", nullptr, false },
            { "record", "Vector<WebCore::DOMCacheEngine::Record>", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CacheStorageClearMemoryRepresentation:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::ClientOrigin", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CacheStorageRepresentation:
        return Vector<ArgumentDescription> { };
#if ENABLE(WEB_RTC)
    case MessageName::NetworkMDNSRegister_UnregisterMDNSNames:
        return Vector<ArgumentDescription> {
            { "documentIdentifier", "WebCore::ScriptExecutionContextIdentifier", nullptr, false },
        };
    case MessageName::NetworkMDNSRegister_RegisterMDNSName:
        return Vector<ArgumentDescription> {
            { "requestIdentifier", "WebKit::MDNSRegisterIdentifier", nullptr, false },
            { "documentIdentifier", "WebCore::ScriptExecutionContextIdentifier", nullptr, false },
            { "ipAddress", "String", nullptr, false },
        };
#endif
#if USE(LIBWEBRTC)
    case MessageName::NetworkRTCMonitor_StartUpdatingIfNeeded:
        return Vector<ArgumentDescription> {
            { "enumerateAllInterfaces", "bool", nullptr, false },
        };
    case MessageName::NetworkRTCMonitor_StopUpdating:
        return Vector<ArgumentDescription> { };
#endif
#if USE(LIBWEBRTC)
    case MessageName::NetworkRTCProvider_CreateUDPSocket:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::LibWebRTCSocketIdentifier", nullptr, false },
            { "localAddress", "WebKit::RTCNetwork::SocketAddress", nullptr, false },
            { "minPort", "uint16_t", nullptr, false },
            { "maxPort", "uint16_t", nullptr, false },
            { "pageIdentifier", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "isFirstParty", "bool", nullptr, false },
            { "isRelayDisabled", "bool", nullptr, false },
            { "domain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkRTCProvider_CreateClientTCPSocket:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::LibWebRTCSocketIdentifier", nullptr, false },
            { "localAddress", "WebKit::RTCNetwork::SocketAddress", nullptr, false },
            { "remoteAddress", "WebKit::RTCNetwork::SocketAddress", nullptr, false },
            { "userAgent", "String", nullptr, false },
            { "options", "int", nullptr, false },
            { "pageIdentifier", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "isFirstParty", "bool", nullptr, false },
            { "isRelayDisabled", "bool", nullptr, false },
            { "domain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkRTCProvider_WrapNewTCPConnection:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::LibWebRTCSocketIdentifier", nullptr, false },
            { "newConnectionSocketIdentifier", "WebCore::LibWebRTCSocketIdentifier", nullptr, false },
        };
    case MessageName::NetworkRTCProvider_SetPlatformTCPSocketsEnabled:
        return Vector<ArgumentDescription> {
            { "enabled", "bool", nullptr, false },
        };
    case MessageName::NetworkRTCProvider_SetPlatformUDPSocketsEnabled:
        return Vector<ArgumentDescription> {
            { "enabled", "bool", nullptr, false },
        };
    case MessageName::NetworkRTCProvider_CreateResolver:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::LibWebRTCResolverIdentifier", nullptr, false },
            { "address", "String", nullptr, false },
        };
    case MessageName::NetworkRTCProvider_StopResolver:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::LibWebRTCResolverIdentifier", nullptr, false },
        };
    case MessageName::NetworkRTCProvider_SendToSocket:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::LibWebRTCSocketIdentifier", nullptr, false },
            { "data", "IPC::DataReference", nullptr, false },
            { "address", "WebKit::RTCNetwork::SocketAddress", nullptr, false },
            { "options", "WebKit::RTCPacketOptions", nullptr, false },
        };
    case MessageName::NetworkRTCProvider_CloseSocket:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::LibWebRTCSocketIdentifier", nullptr, false },
        };
    case MessageName::NetworkRTCProvider_SetSocketOption:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::LibWebRTCSocketIdentifier", nullptr, false },
            { "option", "int", nullptr, false },
            { "value", "int", nullptr, false },
        };
#endif
#if ENABLE(WEB_RTC)
    case MessageName::RTCDataChannelRemoteManagerProxy_SendData:
        return Vector<ArgumentDescription> {
            { "source", "WebCore::RTCDataChannelIdentifier", nullptr, false },
            { "isRaw", "bool", nullptr, false },
            { "text", "IPC::DataReference", nullptr, false },
        };
    case MessageName::RTCDataChannelRemoteManagerProxy_Close:
        return Vector<ArgumentDescription> {
            { "source", "WebCore::RTCDataChannelIdentifier", nullptr, false },
        };
    case MessageName::RTCDataChannelRemoteManagerProxy_ChangeReadyState:
        return Vector<ArgumentDescription> {
            { "handler", "WebCore::RTCDataChannelIdentifier", nullptr, false },
            { "state", "uint8_t", "WebCore::RTCDataChannelState", false },
        };
    case MessageName::RTCDataChannelRemoteManagerProxy_ReceiveData:
        return Vector<ArgumentDescription> {
            { "handler", "WebCore::RTCDataChannelIdentifier", nullptr, false },
            { "isRaw", "bool", nullptr, false },
            { "data", "IPC::DataReference", nullptr, false },
        };
    case MessageName::RTCDataChannelRemoteManagerProxy_DetectError:
        return Vector<ArgumentDescription> {
            { "handler", "WebCore::RTCDataChannelIdentifier", nullptr, false },
            { "type", "uint8_t", "WebCore::RTCErrorDetailType", false },
            { "message", "String", nullptr, false },
        };
    case MessageName::RTCDataChannelRemoteManagerProxy_BufferedAmountIsDecreasing:
        return Vector<ArgumentDescription> {
            { "handler", "WebCore::RTCDataChannelIdentifier", nullptr, false },
            { "amount", "size_t", nullptr, false },
        };
#endif
    case MessageName::AuxiliaryProcess_ShutDown:
        return Vector<ArgumentDescription> { };
    case MessageName::AuxiliaryProcess_SetProcessSuppressionEnabled:
        return Vector<ArgumentDescription> {
            { "flag", "bool", nullptr, false },
        };
    case MessageName::AuxiliaryProcess_MainThreadPing:
        return Vector<ArgumentDescription> { };
#if OS(LINUX)
    case MessageName::AuxiliaryProcess_DidReceiveMemoryPressureEvent:
        return Vector<ArgumentDescription> {
            { "isCritical", "bool", nullptr, false },
        };
#endif
#if ENABLE(IPC_TESTING_API)
    case MessageName::IPCConnectionTester_AsyncMessage:
        return Vector<ArgumentDescription> {
            { "value", "uint32_t", nullptr, false },
        };
    case MessageName::IPCConnectionTester_SyncMessage:
        return Vector<ArgumentDescription> {
            { "value", "uint32_t", nullptr, false },
        };
#endif
#if ENABLE(IPC_TESTING_API)
    case MessageName::IPCStreamTester_SyncMessageReturningSharedMemory1:
        return Vector<ArgumentDescription> {
            { "byteCount", "uint32_t", nullptr, false },
        };
    case MessageName::IPCStreamTester_SyncCrashOnZero:
        return Vector<ArgumentDescription> {
            { "value", "int32_t", nullptr, false },
        };
    case MessageName::IPCStreamTester_CheckAutoreleasePool:
        return Vector<ArgumentDescription> { };
    case MessageName::IPCStreamTester_AsyncMessage:
        return Vector<ArgumentDescription> {
            { "value", "bool", nullptr, false },
        };
#endif
#if ENABLE(IPC_TESTING_API)
    case MessageName::IPCStreamTesterProxy_WasCreated:
        return Vector<ArgumentDescription> {
            { "streamWakeUpSemaphore", "IPC::Semaphore", nullptr, false },
            { "streamClientWaitSemaphore", "IPC::Semaphore", nullptr, false },
        };
#endif
#if ENABLE(IPC_TESTING_API)
    case MessageName::IPCTester_StartMessageTesting:
        return Vector<ArgumentDescription> {
            { "driverName", "String", nullptr, false },
        };
    case MessageName::IPCTester_StopMessageTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::IPCTester_CreateStreamTester:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::IPCStreamTesterIdentifier", nullptr, false },
            { "serverConnection", "IPC::StreamServerConnection::Handle", nullptr, false },
        };
    case MessageName::IPCTester_ReleaseStreamTester:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::IPCStreamTesterIdentifier", nullptr, false },
        };
    case MessageName::IPCTester_CreateConnectionTester:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::IPCConnectionTesterIdentifier", nullptr, false },
            { "connection", "IPC::Connection::Handle", nullptr, false },
        };
    case MessageName::IPCTester_CreateConnectionTesterAndSendAsyncMessages:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::IPCConnectionTesterIdentifier", nullptr, false },
            { "connection", "IPC::Connection::Handle", nullptr, false },
            { "messageCount", "uint32_t", nullptr, false },
        };
    case MessageName::IPCTester_ReleaseConnectionTester:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::IPCConnectionTesterIdentifier", nullptr, false },
        };
    case MessageName::IPCTester_SendSameSemaphoreBack:
        return Vector<ArgumentDescription> {
            { "semaphore", "IPC::Semaphore", nullptr, false },
        };
    case MessageName::IPCTester_SendSemaphoreBackAndSignalProtocol:
        return Vector<ArgumentDescription> {
            { "semaphore", "IPC::Semaphore", nullptr, false },
        };
    case MessageName::IPCTester_SendAsyncMessageToReceiver:
        return Vector<ArgumentDescription> {
            { "arg0", "uint32_t", nullptr, false },
        };
#endif
#if ENABLE(IPC_TESTING_API)
    case MessageName::IPCTesterReceiver_AsyncMessage:
        return Vector<ArgumentDescription> {
            { "arg1", "uint32_t", nullptr, false },
        };
#endif
    case MessageName::WebConnection_HandleMessage:
        return Vector<ArgumentDescription> {
            { "messageName", "String", nullptr, false },
            { "messageBody", "WebKit::UserData", nullptr, false },
        };
    case MessageName::AuthenticationManager_CompleteAuthenticationChallenge:
        return Vector<ArgumentDescription> {
            { "challengeID", "WebKit::AuthenticationChallengeIdentifier", nullptr, false },
            { "disposition", "uint8_t", "WebKit::AuthenticationChallengeDisposition", false },
            { "credential", "WebCore::Credential", nullptr, false },
        };
    case MessageName::NotificationManagerMessageHandler_RequestSystemNotificationPermission:
        return Vector<ArgumentDescription> {
            { "originIdentifier", "String", nullptr, false },
        };
    case MessageName::NotificationManagerMessageHandler_ShowNotification:
        return Vector<ArgumentDescription> {
            { "notificationData", "WebCore::NotificationData", nullptr, false },
            { "notificationResources", "RefPtr<WebCore::NotificationResources>", nullptr, false },
        };
    case MessageName::NotificationManagerMessageHandler_CancelNotification:
        return Vector<ArgumentDescription> {
            { "notificationID", "UUID", nullptr, false },
        };
    case MessageName::NotificationManagerMessageHandler_ClearNotifications:
        return Vector<ArgumentDescription> {
            { "notificationIDs", "Vector<UUID>", nullptr, false },
        };
    case MessageName::NotificationManagerMessageHandler_DidDestroyNotification:
        return Vector<ArgumentDescription> {
            { "notificationID", "UUID", nullptr, false },
        };
    case MessageName::DrawingAreaProxy_EnterAcceleratedCompositingMode:
        return Vector<ArgumentDescription> {
            { "backingStoreStateID", "uint64_t", nullptr, false },
            { "context", "WebKit::LayerTreeContext", nullptr, false },
        };
    case MessageName::DrawingAreaProxy_UpdateAcceleratedCompositingMode:
        return Vector<ArgumentDescription> {
            { "backingStoreStateID", "uint64_t", nullptr, false },
            { "context", "WebKit::LayerTreeContext", nullptr, false },
        };
    case MessageName::DrawingAreaProxy_DidFirstLayerFlush:
        return Vector<ArgumentDescription> {
            { "backingStoreStateID", "uint64_t", nullptr, false },
            { "context", "WebKit::LayerTreeContext", nullptr, false },
        };
    case MessageName::DrawingAreaProxy_DispatchPresentationCallbacksAfterFlushingLayers:
        return Vector<ArgumentDescription> {
            { "callbackIDs", "Vector<IPC::AsyncReplyID>", nullptr, false },
        };
#if USE(COORDINATED_GRAPHICS) || USE(TEXTURE_MAPPER)
    case MessageName::DrawingAreaProxy_Update:
        return Vector<ArgumentDescription> {
            { "stateID", "uint64_t", nullptr, false },
            { "updateInfo", "WebKit::UpdateInfo", nullptr, false },
        };
    case MessageName::DrawingAreaProxy_DidUpdateBackingStoreState:
        return Vector<ArgumentDescription> {
            { "backingStoreStateID", "uint64_t", nullptr, false },
            { "updateInfo", "WebKit::UpdateInfo", nullptr, false },
            { "context", "WebKit::LayerTreeContext", nullptr, false },
        };
    case MessageName::DrawingAreaProxy_ExitAcceleratedCompositingMode:
        return Vector<ArgumentDescription> {
            { "backingStoreStateID", "uint64_t", nullptr, false },
            { "updateInfo", "WebKit::UpdateInfo", nullptr, false },
        };
#endif
#if ENABLE(MEDIA_SESSION_COORDINATOR)
    case MessageName::RemoteMediaSessionCoordinatorProxy_Join:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaSessionCoordinatorProxy_Leave:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaSessionCoordinatorProxy_CoordinateSeekTo:
        return Vector<ArgumentDescription> {
            { "time", "double", nullptr, false },
        };
    case MessageName::RemoteMediaSessionCoordinatorProxy_CoordinatePlay:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaSessionCoordinatorProxy_CoordinatePause:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaSessionCoordinatorProxy_CoordinateSetTrack:
        return Vector<ArgumentDescription> {
            { "trackIdentifier", "String", nullptr, false },
        };
    case MessageName::RemoteMediaSessionCoordinatorProxy_PositionStateChanged:
        return Vector<ArgumentDescription> {
            { "state", "WebCore::MediaPositionState", nullptr, true },
        };
    case MessageName::RemoteMediaSessionCoordinatorProxy_ReadyStateChanged:
        return Vector<ArgumentDescription> {
            { "state", "uint8_t", "WebCore::MediaSessionReadyState", false },
        };
    case MessageName::RemoteMediaSessionCoordinatorProxy_PlaybackStateChanged:
        return Vector<ArgumentDescription> {
            { "state", "uint8_t", "WebCore::MediaSessionPlaybackState", false },
        };
    case MessageName::RemoteMediaSessionCoordinatorProxy_TrackIdentifierChanged:
        return Vector<ArgumentDescription> {
            { "trackIdentifier", "String", nullptr, false },
        };
#endif
#if ENABLE(MEDIA_STREAM)
    case MessageName::SpeechRecognitionRemoteRealtimeMediaSourceManager_RemoteAudioSamplesAvailable:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::RealtimeMediaSourceIdentifier", nullptr, false },
            { "time", "MediaTime", nullptr, false },
            { "numberOfFrames", "uint64_t", nullptr, false },
        };
    case MessageName::SpeechRecognitionRemoteRealtimeMediaSourceManager_RemoteCaptureFailed:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::RealtimeMediaSourceIdentifier", nullptr, false },
        };
    case MessageName::SpeechRecognitionRemoteRealtimeMediaSourceManager_RemoteSourceStopped:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::RealtimeMediaSourceIdentifier", nullptr, false },
        };
#if PLATFORM(COCOA)
    case MessageName::SpeechRecognitionRemoteRealtimeMediaSourceManager_SetStorage:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::RealtimeMediaSourceIdentifier", nullptr, false },
            { "storageHandle", "WebKit::ConsumerSharedCARingBuffer::Handle", nullptr, false },
            { "description", "WebCore::CAAudioStreamDescription", nullptr, false },
        };
#endif
#endif
    case MessageName::SpeechRecognitionServer_Start:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::SpeechRecognitionConnectionClientIdentifier", nullptr, false },
            { "lang", "String", nullptr, false },
            { "continuous", "bool", nullptr, false },
            { "interimResults", "bool", nullptr, false },
            { "maxAlternatives", "uint64_t", nullptr, false },
            { "origin", "WebCore::ClientOrigin", nullptr, false },
            { "frameIdentifier", "WebCore::FrameIdentifier", nullptr, false },
        };
    case MessageName::SpeechRecognitionServer_Stop:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::SpeechRecognitionConnectionClientIdentifier", nullptr, false },
        };
    case MessageName::SpeechRecognitionServer_Abort:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::SpeechRecognitionConnectionClientIdentifier", nullptr, false },
        };
    case MessageName::SpeechRecognitionServer_Invalidate:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::SpeechRecognitionConnectionClientIdentifier", nullptr, false },
        };
    case MessageName::VisitedLinkStore_AddVisitedLinkHashFromPage:
        return Vector<ArgumentDescription> {
            { "pageProxyID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "linkHash", "WebCore::SharedStringHash", nullptr, false },
        };
    case MessageName::WebFrameProxy_DidCreateSubframe:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
        };
#if ENABLE(FULLSCREEN_API)
    case MessageName::WebFullScreenManagerProxy_SupportsFullScreen:
        return Vector<ArgumentDescription> {
            { "withKeyboard", "bool", nullptr, false },
        };
    case MessageName::WebFullScreenManagerProxy_EnterFullScreen:
        return Vector<ArgumentDescription> {
            { "blocksReturnToFullscreenFromPictureInPicture", "bool", nullptr, false },
            { "isVideoElementWithControls", "bool", nullptr, false },
            { "videoDimensions", "WebCore::FloatSize", nullptr, false },
        };
    case MessageName::WebFullScreenManagerProxy_ExitFullScreen:
        return Vector<ArgumentDescription> { };
    case MessageName::WebFullScreenManagerProxy_BeganEnterFullScreen:
        return Vector<ArgumentDescription> {
            { "initialRect", "WebCore::IntRect", nullptr, false },
            { "finalRect", "WebCore::IntRect", nullptr, false },
        };
    case MessageName::WebFullScreenManagerProxy_BeganExitFullScreen:
        return Vector<ArgumentDescription> {
            { "initialRect", "WebCore::IntRect", nullptr, false },
            { "finalRect", "WebCore::IntRect", nullptr, false },
        };
    case MessageName::WebFullScreenManagerProxy_Close:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebGeolocationManagerProxy_StartUpdating:
        return Vector<ArgumentDescription> {
            { "registrableDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "pageProxyID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "authorizationToken", "String", nullptr, false },
            { "enableHighAccuracy", "bool", nullptr, false },
        };
    case MessageName::WebGeolocationManagerProxy_StopUpdating:
        return Vector<ArgumentDescription> {
            { "registrableDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::WebGeolocationManagerProxy_SetEnableHighAccuracy:
        return Vector<ArgumentDescription> {
            { "registrableDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "enabled", "bool", nullptr, false },
        };
    case MessageName::WebLockRegistryProxy_RequestLock:
        return Vector<ArgumentDescription> {
            { "clientOrigin", "WebCore::ClientOrigin", nullptr, false },
            { "identifier", "WebCore::WebLockIdentifier", nullptr, false },
            { "clientID", "WebCore::ScriptExecutionContextIdentifier", nullptr, false },
            { "name", "String", nullptr, false },
            { "mode", "bool", "WebCore::WebLockMode", false },
            { "steal", "bool", nullptr, false },
            { "ifAvailable", "bool", nullptr, false },
        };
    case MessageName::WebLockRegistryProxy_ReleaseLock:
        return Vector<ArgumentDescription> {
            { "clientOrigin", "WebCore::ClientOrigin", nullptr, false },
            { "identifier", "WebCore::WebLockIdentifier", nullptr, false },
            { "clientID", "WebCore::ScriptExecutionContextIdentifier", nullptr, false },
            { "name", "String", nullptr, false },
        };
    case MessageName::WebLockRegistryProxy_AbortLockRequest:
        return Vector<ArgumentDescription> {
            { "clientOrigin", "WebCore::ClientOrigin", nullptr, false },
            { "identifier", "WebCore::WebLockIdentifier", nullptr, false },
            { "clientID", "WebCore::ScriptExecutionContextIdentifier", nullptr, false },
            { "name", "String", nullptr, false },
        };
    case MessageName::WebLockRegistryProxy_Snapshot:
        return Vector<ArgumentDescription> {
            { "clientOrigin", "WebCore::ClientOrigin", nullptr, false },
        };
    case MessageName::WebLockRegistryProxy_ClientIsGoingAway:
        return Vector<ArgumentDescription> {
            { "clientOrigin", "WebCore::ClientOrigin", nullptr, false },
            { "clientId", "WebCore::ScriptExecutionContextIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_CreateNewPage:
        return Vector<ArgumentDescription> {
            { "originatingFrameInfoData", "WebKit::FrameInfoData", nullptr, false },
            { "originatingPageID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "windowFeatures", "WebCore::WindowFeatures", nullptr, false },
            { "navigationActionData", "WebKit::NavigationActionData", nullptr, false },
        };
    case MessageName::WebPageProxy_ShowPage:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_ClosePage:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_RunJavaScriptAlert:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "frameInfo", "WebKit::FrameInfoData", nullptr, false },
            { "message", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_RunJavaScriptConfirm:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "frameInfo", "WebKit::FrameInfoData", nullptr, false },
            { "message", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_RunJavaScriptPrompt:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "frameInfo", "WebKit::FrameInfoData", nullptr, false },
            { "message", "String", nullptr, false },
            { "defaultValue", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_MouseDidMoveOverElement:
        return Vector<ArgumentDescription> {
            { "hitTestResultData", "WebKit::WebHitTestResultData", nullptr, false },
            { "modifiers", "uint32_t", nullptr, false },
            { "userData", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebPageProxy_DidChangeViewportProperties:
        return Vector<ArgumentDescription> {
            { "attributes", "WebCore::ViewportAttributes", nullptr, false },
        };
    case MessageName::WebPageProxy_DidReceiveEvent:
        return Vector<ArgumentDescription> {
            { "eventType", "int8_t", "WebKit::WebEventType", false },
            { "handled", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_SetCursor:
        return Vector<ArgumentDescription> {
            { "cursor", "WebCore::Cursor", nullptr, false },
        };
    case MessageName::WebPageProxy_SetCursorHiddenUntilMouseMoves:
        return Vector<ArgumentDescription> {
            { "hiddenUntilMouseMoves", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_SetStatusText:
        return Vector<ArgumentDescription> {
            { "statusText", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_SetFocus:
        return Vector<ArgumentDescription> {
            { "focused", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_TakeFocus:
        return Vector<ArgumentDescription> {
            { "direction", "uint8_t", "WebCore::FocusDirection", false },
        };
    case MessageName::WebPageProxy_FocusFromServiceWorker:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_FocusedFrameChanged:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
        };
    case MessageName::WebPageProxy_SetRenderTreeSize:
        return Vector<ArgumentDescription> {
            { "treeSize", "uint64_t", nullptr, false },
        };
    case MessageName::WebPageProxy_SetToolbarsAreVisible:
        return Vector<ArgumentDescription> {
            { "toolbarsAreVisible", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_GetToolbarsAreVisible:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_SetMenuBarIsVisible:
        return Vector<ArgumentDescription> {
            { "menuBarIsVisible", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_GetMenuBarIsVisible:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_SetStatusBarIsVisible:
        return Vector<ArgumentDescription> {
            { "statusBarIsVisible", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_GetStatusBarIsVisible:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_SetIsResizable:
        return Vector<ArgumentDescription> {
            { "isResizable", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_SetWindowFrame:
        return Vector<ArgumentDescription> {
            { "windowFrame", "WebCore::FloatRect", nullptr, false },
        };
    case MessageName::WebPageProxy_GetWindowFrame:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_ScreenToRootView:
        return Vector<ArgumentDescription> {
            { "screenPoint", "WebCore::IntPoint", nullptr, false },
        };
    case MessageName::WebPageProxy_RootViewToScreen:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::IntRect", nullptr, false },
        };
    case MessageName::WebPageProxy_AccessibilityScreenToRootView:
        return Vector<ArgumentDescription> {
            { "screenPoint", "WebCore::IntPoint", nullptr, false },
        };
    case MessageName::WebPageProxy_RootViewToAccessibilityScreen:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::IntRect", nullptr, false },
        };
#if PLATFORM(COCOA) || PLATFORM(GTK)
    case MessageName::WebPageProxy_ShowValidationMessage:
        return Vector<ArgumentDescription> {
            { "anchorRect", "WebCore::IntRect", nullptr, false },
            { "message", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_HideValidationMessage:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebPageProxy_RunBeforeUnloadConfirmPanel:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "frameInfo", "WebKit::FrameInfoData", nullptr, false },
            { "message", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_PageDidScroll:
        return Vector<ArgumentDescription> {
            { "scrollPosition", "WebCore::IntPoint", nullptr, false },
        };
    case MessageName::WebPageProxy_SetHasActiveAnimatedScrolls:
        return Vector<ArgumentDescription> {
            { "hasActiveAnimatedScrolls", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_RunOpenPanel:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "frameInfo", "WebKit::FrameInfoData", nullptr, false },
            { "parameters", "WebCore::FileChooserSettings", nullptr, false },
        };
    case MessageName::WebPageProxy_ShowShareSheet:
        return Vector<ArgumentDescription> {
            { "shareData", "WebCore::ShareDataWithParsedURL", nullptr, false },
        };
    case MessageName::WebPageProxy_ShowContactPicker:
        return Vector<ArgumentDescription> {
            { "requestData", "WebCore::ContactsRequestData", nullptr, false },
        };
    case MessageName::WebPageProxy_PrintFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "title", "String", nullptr, false },
            { "pdfFirstPageSize", "WebCore::FloatSize", nullptr, false },
        };
    case MessageName::WebPageProxy_RunModal:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_NotifyScrollerThumbIsVisibleInRect:
        return Vector<ArgumentDescription> {
            { "scrollerThumb", "WebCore::IntRect", nullptr, false },
        };
    case MessageName::WebPageProxy_RecommendedScrollbarStyleDidChange:
        return Vector<ArgumentDescription> {
            { "newStyle", "int32_t", nullptr, false },
        };
    case MessageName::WebPageProxy_DidChangeScrollbarsForMainFrame:
        return Vector<ArgumentDescription> {
            { "hasHorizontalScrollbar", "bool", nullptr, false },
            { "hasVerticalScrollbar", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_DidChangeScrollOffsetPinningForMainFrame:
        return Vector<ArgumentDescription> {
            { "pinnedState", "WebCore::RectEdges<bool>", nullptr, false },
        };
    case MessageName::WebPageProxy_DidChangePageCount:
        return Vector<ArgumentDescription> {
            { "pageCount", "unsigned", nullptr, false },
        };
    case MessageName::WebPageProxy_ThemeColorChanged:
        return Vector<ArgumentDescription> {
            { "themeColor", "WebCore::Color", nullptr, false },
        };
    case MessageName::WebPageProxy_PageExtendedBackgroundColorDidChange:
        return Vector<ArgumentDescription> {
            { "backgroundColor", "WebCore::Color", nullptr, false },
        };
    case MessageName::WebPageProxy_SampledPageTopColorChanged:
        return Vector<ArgumentDescription> {
            { "sampledPageTopColor", "WebCore::Color", nullptr, false },
        };
    case MessageName::WebPageProxy_SetCanShortCircuitHorizontalWheelEvents:
        return Vector<ArgumentDescription> {
            { "canShortCircuitHorizontalWheelEvents", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_DidChangeContentSize:
        return Vector<ArgumentDescription> {
            { "newSize", "WebCore::IntSize", nullptr, false },
        };
    case MessageName::WebPageProxy_DidChangeIntrinsicContentSize:
        return Vector<ArgumentDescription> {
            { "newIntrinsicContentSize", "WebCore::IntSize", nullptr, false },
        };
#if ENABLE(INPUT_TYPE_COLOR)
    case MessageName::WebPageProxy_ShowColorPicker:
        return Vector<ArgumentDescription> {
            { "initialColor", "WebCore::Color", nullptr, false },
            { "elementRect", "WebCore::IntRect", nullptr, false },
            { "suggestions", "Vector<WebCore::Color>", nullptr, false },
        };
    case MessageName::WebPageProxy_SetColorPickerColor:
        return Vector<ArgumentDescription> {
            { "color", "WebCore::Color", nullptr, false },
        };
    case MessageName::WebPageProxy_EndColorPicker:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(DATALIST_ELEMENT)
    case MessageName::WebPageProxy_ShowDataListSuggestions:
        return Vector<ArgumentDescription> {
            { "suggestionInformation", "WebCore::DataListSuggestionInformation", nullptr, false },
        };
    case MessageName::WebPageProxy_HandleKeydownInDataList:
        return Vector<ArgumentDescription> {
            { "key", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_EndDataListSuggestions:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(DATE_AND_TIME_INPUT_TYPES)
    case MessageName::WebPageProxy_ShowDateTimePicker:
        return Vector<ArgumentDescription> {
            { "params", "WebCore::DateTimeChooserParameters", nullptr, false },
        };
    case MessageName::WebPageProxy_EndDateTimePicker:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebPageProxy_DecidePolicyForResponse:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "frameInfo", "WebKit::FrameInfoData", nullptr, false },
            { "policyCheckIdentifier", "WebCore::PolicyCheckIdentifier", nullptr, false },
            { "navigationID", "uint64_t", nullptr, false },
            { "response", "WebCore::ResourceResponse", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "canShowMIMEType", "bool", nullptr, false },
            { "downloadAttribute", "String", nullptr, false },
            { "listenerID", "uint64_t", nullptr, false },
        };
    case MessageName::WebPageProxy_DecidePolicyForNavigationActionAsync:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "frameInfo", "WebKit::FrameInfoData", nullptr, false },
            { "policyCheckIdentifier", "WebCore::PolicyCheckIdentifier", nullptr, false },
            { "navigationID", "uint64_t", nullptr, false },
            { "navigationActionData", "WebKit::NavigationActionData", nullptr, false },
            { "originatingFrameInfoData", "WebKit::FrameInfoData", nullptr, false },
            { "originatingPageID", "WebKit::WebPageProxyIdentifier", nullptr, true },
            { "originalRequest", "WebCore::ResourceRequest", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "requestBody", "IPC::FormDataReference", nullptr, false },
            { "redirectResponse", "WebCore::ResourceResponse", nullptr, false },
            { "listenerID", "uint64_t", nullptr, false },
        };
    case MessageName::WebPageProxy_DecidePolicyForNavigationActionSync:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "isMainFrame", "bool", nullptr, false },
            { "frameInfo", "WebKit::FrameInfoData", nullptr, false },
            { "policyCheckIdentifier", "WebCore::PolicyCheckIdentifier", nullptr, false },
            { "navigationID", "uint64_t", nullptr, false },
            { "navigationActionData", "WebKit::NavigationActionData", nullptr, false },
            { "originatingFrameInfoData", "WebKit::FrameInfoData", nullptr, false },
            { "originatingPageID", "WebKit::WebPageProxyIdentifier", nullptr, true },
            { "originalRequest", "WebCore::ResourceRequest", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "requestBody", "IPC::FormDataReference", nullptr, false },
            { "redirectResponse", "WebCore::ResourceResponse", nullptr, false },
        };
    case MessageName::WebPageProxy_DecidePolicyForNewWindowAction:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "frameInfo", "WebKit::FrameInfoData", nullptr, false },
            { "policyCheckIdentifier", "WebCore::PolicyCheckIdentifier", nullptr, false },
            { "navigationActionData", "WebKit::NavigationActionData", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "frameName", "String", nullptr, false },
            { "listenerID", "uint64_t", nullptr, false },
        };
    case MessageName::WebPageProxy_DidChangeProgress:
        return Vector<ArgumentDescription> {
            { "value", "double", nullptr, false },
        };
    case MessageName::WebPageProxy_DidFinishProgress:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_DidStartProgress:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_SetNetworkRequestsInProgress:
        return Vector<ArgumentDescription> {
            { "networkRequestsInProgress", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_DidCreateMainFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_DidStartProvisionalLoadForFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "frameInfo", "WebKit::FrameInfoData", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "navigationID", "uint64_t", nullptr, false },
            { "url", "URL", nullptr, false },
            { "unreachableURL", "URL", nullptr, false },
            { "userData", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebPageProxy_DidReceiveServerRedirectForProvisionalLoadForFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "navigationID", "uint64_t", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "userData", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebPageProxy_WillPerformClientRedirectForFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "url", "String", nullptr, false },
            { "delay", "double", nullptr, false },
            { "lockBackForwardList", "bool", "WebCore::LockBackForwardList", false },
        };
    case MessageName::WebPageProxy_DidCancelClientRedirectForFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_DidChangeProvisionalURLForFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "navigationID", "uint64_t", nullptr, false },
            { "url", "URL", nullptr, false },
        };
    case MessageName::WebPageProxy_DidFailProvisionalLoadForFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "frameInfo", "WebKit::FrameInfoData", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "navigationID", "uint64_t", nullptr, false },
            { "provisionalURL", "String", nullptr, false },
            { "error", "WebCore::ResourceError", nullptr, false },
            { "willContinueLoading", "bool", "WebCore::WillContinueLoading", false },
            { "userData", "WebKit::UserData", nullptr, false },
            { "willInternallyHandleFailure", "bool", "WebCore::WillInternallyHandleFailure", false },
        };
    case MessageName::WebPageProxy_DidCommitLoadForFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "frameInfo", "WebKit::FrameInfoData", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "navigationID", "uint64_t", nullptr, false },
            { "mimeType", "String", nullptr, false },
            { "hasCustomContentProvider", "bool", nullptr, false },
            { "loadType", "uint8_t", "WebCore::FrameLoadType", false },
            { "certificateInfo", "WebCore::CertificateInfo", nullptr, false },
            { "usedLegacyTLS", "bool", nullptr, false },
            { "wasPrivateRelayed", "bool", nullptr, false },
            { "containsPluginDocument", "bool", nullptr, false },
            { "hasInsecureContent", "bool", "WebCore::HasInsecureContent", false },
            { "mouseEventPolicy", "uint8_t", "WebCore::MouseEventPolicy", false },
            { "userData", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebPageProxy_DidFailLoadForFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "frameInfo", "WebKit::FrameInfoData", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "navigationID", "uint64_t", nullptr, false },
            { "error", "WebCore::ResourceError", nullptr, false },
            { "userData", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebPageProxy_DidFinishDocumentLoadForFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "navigationID", "uint64_t", nullptr, false },
            { "userData", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebPageProxy_DidFinishLoadForFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "frameInfo", "WebKit::FrameInfoData", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "navigationID", "uint64_t", nullptr, false },
            { "userData", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebPageProxy_DidFirstLayoutForFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "userData", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebPageProxy_DidFirstVisuallyNonEmptyLayoutForFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "userData", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebPageProxy_DidReachLayoutMilestone:
        return Vector<ArgumentDescription> {
            { "layoutMilestones", "OptionSet<WebCore::LayoutMilestone>", nullptr, false },
        };
    case MessageName::WebPageProxy_DidReceiveTitleForFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "title", "String", nullptr, false },
            { "userData", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebPageProxy_DidDisplayInsecureContentForFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "userData", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebPageProxy_DidRunInsecureContentForFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "userData", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebPageProxy_DidSameDocumentNavigationForFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "navigationID", "uint64_t", nullptr, false },
            { "type", "uint8_t", "WebKit::SameDocumentNavigationType", false },
            { "url", "URL", nullptr, false },
            { "userData", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebPageProxy_DidSameDocumentNavigationForFrameViaJSHistoryAPI:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "type", "uint8_t", "WebKit::SameDocumentNavigationType", false },
            { "url", "URL", nullptr, false },
            { "navigationActionData", "WebKit::NavigationActionData", nullptr, false },
            { "userData", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebPageProxy_DidChangeMainDocument:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_DidExplicitOpenForFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "url", "URL", nullptr, false },
            { "mimeType", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_DidDestroyNavigation:
        return Vector<ArgumentDescription> {
            { "navigationID", "uint64_t", nullptr, false },
        };
    case MessageName::WebPageProxy_UpdateRemoteFrameSize:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "size", "WebCore::IntSize", nullptr, false },
        };
    case MessageName::WebPageProxy_MainFramePluginHandlesPageScaleGestureDidChange:
        return Vector<ArgumentDescription> {
            { "mainFramePluginHandlesPageScaleGesture", "bool", nullptr, false },
        };
#if PLATFORM(GTK) || PLATFORM(WPE)
    case MessageName::WebPageProxy_DidInitiateLoadForResource:
        return Vector<ArgumentDescription> {
            { "resourceID", "WebCore::ResourceLoaderIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
        };
    case MessageName::WebPageProxy_DidSendRequestForResource:
        return Vector<ArgumentDescription> {
            { "resourceID", "WebCore::ResourceLoaderIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "redirectResponse", "WebCore::ResourceResponse", nullptr, false },
        };
    case MessageName::WebPageProxy_DidReceiveResponseForResource:
        return Vector<ArgumentDescription> {
            { "resourceID", "WebCore::ResourceLoaderIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "response", "WebCore::ResourceResponse", nullptr, false },
        };
    case MessageName::WebPageProxy_DidFinishLoadForResource:
        return Vector<ArgumentDescription> {
            { "resourceID", "WebCore::ResourceLoaderIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "error", "WebCore::ResourceError", nullptr, false },
        };
#endif
    case MessageName::WebPageProxy_DidNavigateWithNavigationData:
        return Vector<ArgumentDescription> {
            { "store", "WebKit::WebNavigationDataStore", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_DidPerformClientRedirect:
        return Vector<ArgumentDescription> {
            { "sourceURLString", "String", nullptr, false },
            { "destinationURLString", "String", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_DidPerformServerRedirect:
        return Vector<ArgumentDescription> {
            { "sourceURLString", "String", nullptr, false },
            { "destinationURLString", "String", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_DidUpdateHistoryTitle:
        return Vector<ArgumentDescription> {
            { "title", "String", nullptr, false },
            { "url", "String", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_DidFinishLoadingDataForCustomContentProvider:
        return Vector<ArgumentDescription> {
            { "suggestedFilename", "String", nullptr, false },
            { "data", "IPC::DataReference", nullptr, false },
        };
    case MessageName::WebPageProxy_WillSubmitForm:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "sourceFrameID", "WebCore::FrameIdentifier", nullptr, false },
            { "textFieldValues", "Vector<std::pair<String, String>>", nullptr, false },
            { "listenerID", "WebKit::FormSubmitListenerIdentifier", nullptr, false },
            { "userData", "WebKit::UserData", nullptr, false },
        };
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebPageProxy_InterpretKeyEvent:
        return Vector<ArgumentDescription> {
            { "state", "WebKit::EditorState", nullptr, false },
            { "isCharEvent", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_DidReceivePositionInformation:
        return Vector<ArgumentDescription> {
            { "information", "WebKit::InteractionInformationAtPosition", nullptr, false },
        };
    case MessageName::WebPageProxy_SaveImageToLibrary:
        return Vector<ArgumentDescription> {
            { "handle", "WebKit::SharedMemory::Handle", nullptr, false },
            { "authorizationToken", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_ShowPlaybackTargetPicker:
        return Vector<ArgumentDescription> {
            { "hasVideo", "bool", nullptr, false },
            { "elementRect", "WebCore::IntRect", nullptr, false },
            { "policy", "uint8_t", "WebCore::RouteSharingPolicy", false },
            { "routingContextUID", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_CommitPotentialTapFailed:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_DidNotHandleTapAsClick:
        return Vector<ArgumentDescription> {
            { "point", "WebCore::IntPoint", nullptr, false },
        };
    case MessageName::WebPageProxy_DidCompleteSyntheticClick:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_DisableDoubleTapGesturesDuringTapIfNecessary:
        return Vector<ArgumentDescription> {
            { "requestID", "WebKit::TapIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_HandleSmartMagnificationInformationForPotentialTap:
        return Vector<ArgumentDescription> {
            { "requestID", "WebKit::TapIdentifier", nullptr, false },
            { "renderRect", "WebCore::FloatRect", nullptr, false },
            { "fitEntireRect", "bool", nullptr, false },
            { "viewportMinimumScale", "double", nullptr, false },
            { "viewportMaximumScale", "double", nullptr, false },
            { "nodeIsRootLevel", "bool", nullptr, false },
        };
#endif
#if ENABLE(DATA_DETECTION)
    case MessageName::WebPageProxy_SetDataDetectionResult:
        return Vector<ArgumentDescription> {
            { "dataDetectionResult", "WebKit::DataDetectionResult", nullptr, false },
        };
    case MessageName::WebPageProxy_HandleClickForDataDetectionResult:
        return Vector<ArgumentDescription> {
            { "info", "WebCore::DataDetectorElementInfo", nullptr, false },
            { "location", "WebCore::IntPoint", nullptr, false },
        };
#endif
    case MessageName::WebPageProxy_PageScaleFactorDidChange:
        return Vector<ArgumentDescription> {
            { "scaleFactor", "double", nullptr, false },
        };
    case MessageName::WebPageProxy_PluginScaleFactorDidChange:
        return Vector<ArgumentDescription> {
            { "zoomFactor", "double", nullptr, false },
        };
    case MessageName::WebPageProxy_PluginZoomFactorDidChange:
        return Vector<ArgumentDescription> {
            { "zoomFactor", "double", nullptr, false },
        };
#if PLATFORM(GTK) || PLATFORM(WPE)
    case MessageName::WebPageProxy_BindAccessibilityTree:
        return Vector<ArgumentDescription> {
            { "plugID", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_SetInputMethodState:
        return Vector<ArgumentDescription> {
            { "state", "WebKit::InputMethodState", nullptr, true },
        };
#endif
    case MessageName::WebPageProxy_BackForwardAddItem:
        return Vector<ArgumentDescription> {
            { "itemState", "WebKit::BackForwardListItemState", nullptr, false },
        };
    case MessageName::WebPageProxy_BackForwardGoToItem:
        return Vector<ArgumentDescription> {
            { "itemID", "WebCore::BackForwardItemIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_BackForwardItemAtIndex:
        return Vector<ArgumentDescription> {
            { "itemIndex", "int32_t", nullptr, false },
        };
    case MessageName::WebPageProxy_BackForwardListContainsItem:
        return Vector<ArgumentDescription> {
            { "itemID", "WebCore::BackForwardItemIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_BackForwardListCounts:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_BackForwardClear:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_WillGoToBackForwardListItem:
        return Vector<ArgumentDescription> {
            { "itemID", "WebCore::BackForwardItemIdentifier", nullptr, false },
            { "inBackForwardCache", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_RegisterEditCommandForUndo:
        return Vector<ArgumentDescription> {
            { "commandID", "uint64_t", nullptr, false },
            { "label", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_ClearAllEditCommands:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_RegisterInsertionUndoGrouping:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_CanUndoRedo:
        return Vector<ArgumentDescription> {
            { "undoOrRedo", "bool", "WebKit::UndoOrRedo", false },
        };
    case MessageName::WebPageProxy_ExecuteUndoRedo:
        return Vector<ArgumentDescription> {
            { "undoOrRedo", "bool", "WebKit::UndoOrRedo", false },
        };
    case MessageName::WebPageProxy_LogDiagnosticMessageFromWebProcess:
        return Vector<ArgumentDescription> {
            { "message", "String", nullptr, false },
            { "description", "String", nullptr, false },
            { "shouldSample", "bool", "WebCore::ShouldSample", false },
        };
    case MessageName::WebPageProxy_LogDiagnosticMessageWithResultFromWebProcess:
        return Vector<ArgumentDescription> {
            { "message", "String", nullptr, false },
            { "description", "String", nullptr, false },
            { "result", "uint32_t", nullptr, false },
            { "shouldSample", "bool", "WebCore::ShouldSample", false },
        };
    case MessageName::WebPageProxy_LogDiagnosticMessageWithValueFromWebProcess:
        return Vector<ArgumentDescription> {
            { "message", "String", nullptr, false },
            { "description", "String", nullptr, false },
            { "value", "double", nullptr, false },
            { "significantFigures", "unsigned", nullptr, false },
            { "shouldSample", "bool", "WebCore::ShouldSample", false },
        };
    case MessageName::WebPageProxy_LogDiagnosticMessageWithEnhancedPrivacyFromWebProcess:
        return Vector<ArgumentDescription> {
            { "message", "String", nullptr, false },
            { "description", "String", nullptr, false },
            { "shouldSample", "bool", "WebCore::ShouldSample", false },
        };
    case MessageName::WebPageProxy_LogDiagnosticMessageWithValueDictionaryFromWebProcess:
        return Vector<ArgumentDescription> {
            { "message", "String", nullptr, false },
            { "description", "String", nullptr, false },
            { "value", "WebCore::DiagnosticLoggingClient::ValueDictionary", nullptr, false },
            { "shouldSample", "bool", "WebCore::ShouldSample", false },
        };
    case MessageName::WebPageProxy_LogDiagnosticMessageWithDomainFromWebProcess:
        return Vector<ArgumentDescription> {
            { "message", "String", nullptr, false },
            { "domain", "WebCore::DiagnosticLoggingDomain", nullptr, false },
        };
    case MessageName::WebPageProxy_LogScrollingEvent:
        return Vector<ArgumentDescription> {
            { "eventType", "uint32_t", nullptr, false },
            { "timestamp", "MonotonicTime", nullptr, false },
            { "data", "uint64_t", nullptr, false },
        };
    case MessageName::WebPageProxy_EditorStateChanged:
        return Vector<ArgumentDescription> {
            { "editorState", "WebKit::EditorState", nullptr, false },
        };
    case MessageName::WebPageProxy_CompositionWasCanceled:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_SetHasHadSelectionChangesFromUserInteraction:
        return Vector<ArgumentDescription> {
            { "hasHadUserSelectionChanges", "bool", nullptr, false },
        };
#if ENABLE(IMAGE_ANALYSIS)
    case MessageName::WebPageProxy_RequestTextRecognition:
        return Vector<ArgumentDescription> {
            { "imageURL", "URL", nullptr, false },
            { "imageData", "WebKit::ShareableBitmapHandle", nullptr, false },
            { "source", "String", nullptr, false },
            { "target", "String", nullptr, false },
        };
#endif
#if HAVE(TRANSLATION_UI_SERVICES) && ENABLE(CONTEXT_MENUS)
    case MessageName::WebPageProxy_HandleContextMenuTranslation:
        return Vector<ArgumentDescription> {
            { "info", "WebCore::TranslationContextMenuInfo", nullptr, false },
        };
#endif
#if ENABLE(MEDIA_CONTROLS_CONTEXT_MENUS) && USE(UICONTEXTMENU)
    case MessageName::WebPageProxy_ShowMediaControlsContextMenu:
        return Vector<ArgumentDescription> {
            { "targetFrame", "WebCore::FloatRect", nullptr, false },
            { "items", "Vector<WebCore::MediaControlsContextMenuItem>", nullptr, false },
        };
#endif
#if HAVE(TOUCH_BAR)
    case MessageName::WebPageProxy_SetIsTouchBarUpdateSupressedForHiddenContentEditable:
        return Vector<ArgumentDescription> {
            { "isTouchBarUpdateSupressed", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_SetIsNeverRichlyEditableForTouchBar:
        return Vector<ArgumentDescription> {
            { "isNeverRichlyEditable", "bool", nullptr, false },
        };
#endif
    case MessageName::WebPageProxy_RequestDOMPasteAccess:
        return Vector<ArgumentDescription> {
            { "pasteAccessCategory", "uint8_t", "WebCore::DOMPasteAccessCategory", false },
            { "elementRect", "WebCore::IntRect", nullptr, false },
            { "originIdentifier", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_DidCountStringMatches:
        return Vector<ArgumentDescription> {
            { "string", "String", nullptr, false },
            { "matchCount", "uint32_t", nullptr, false },
        };
    case MessageName::WebPageProxy_SetTextIndicator:
        return Vector<ArgumentDescription> {
            { "indicator", "WebCore::TextIndicatorData", nullptr, false },
            { "lifetime", "uint64_t", nullptr, false },
        };
    case MessageName::WebPageProxy_ClearTextIndicator:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_DidFindString:
        return Vector<ArgumentDescription> {
            { "string", "String", nullptr, false },
            { "matchRect", "Vector<WebCore::IntRect>", nullptr, false },
            { "matchCount", "uint32_t", nullptr, false },
            { "matchIndex", "int32_t", nullptr, false },
            { "didWrapAround", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_DidFailToFindString:
        return Vector<ArgumentDescription> {
            { "string", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_DidFindStringMatches:
        return Vector<ArgumentDescription> {
            { "string", "String", nullptr, false },
            { "matches", "Vector<Vector<WebCore::IntRect>>", nullptr, false },
            { "firstIndexAfterSelection", "int32_t", nullptr, false },
        };
    case MessageName::WebPageProxy_DidGetImageForFindMatch:
        return Vector<ArgumentDescription> {
            { "parameters", "WebCore::ImageBufferBackendParameters", nullptr, false },
            { "contentImageHandle", "WebKit::ShareableBitmapHandle", nullptr, false },
            { "matchIndex", "uint32_t", nullptr, false },
        };
    case MessageName::WebPageProxy_ShowPopupMenu:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::IntRect", nullptr, false },
            { "textDirection", "uint64_t", nullptr, false },
            { "items", "Vector<WebKit::WebPopupItem>", nullptr, false },
            { "selectedIndex", "int32_t", nullptr, false },
            { "data", "WebKit::PlatformPopupMenuData", nullptr, false },
        };
    case MessageName::WebPageProxy_HidePopupMenu:
        return Vector<ArgumentDescription> { };
#if ENABLE(CONTEXT_MENUS)
    case MessageName::WebPageProxy_ShowContextMenu:
        return Vector<ArgumentDescription> {
            { "contextMenuContextData", "WebKit::ContextMenuContextData", nullptr, false },
            { "userData", "WebKit::UserData", nullptr, false },
        };
#endif
#if ENABLE(SERVICE_WORKER)
    case MessageName::WebPageProxy_DidFinishServiceWorkerPageRegistration:
        return Vector<ArgumentDescription> {
            { "success", "bool", nullptr, false },
        };
#endif
    case MessageName::WebPageProxy_ExceededDatabaseQuota:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "originIdentifier", "String", nullptr, false },
            { "databaseName", "String", nullptr, false },
            { "databaseDisplayName", "String", nullptr, false },
            { "currentQuota", "uint64_t", nullptr, false },
            { "currentOriginUsage", "uint64_t", nullptr, false },
            { "currentDatabaseUsage", "uint64_t", nullptr, false },
            { "expectedUsage", "uint64_t", nullptr, false },
        };
    case MessageName::WebPageProxy_ReachedApplicationCacheOriginQuota:
        return Vector<ArgumentDescription> {
            { "originIdentifier", "String", nullptr, false },
            { "currentQuota", "uint64_t", nullptr, false },
            { "totalBytesNeeded", "uint64_t", nullptr, false },
        };
    case MessageName::WebPageProxy_RequestGeolocationPermissionForFrame:
        return Vector<ArgumentDescription> {
            { "geolocationID", "WebKit::GeolocationIdentifier", nullptr, false },
            { "frameInfo", "WebKit::FrameInfoData", nullptr, false },
        };
    case MessageName::WebPageProxy_RevokeGeolocationAuthorizationToken:
        return Vector<ArgumentDescription> {
            { "authorizationToken", "String", nullptr, false },
        };
#if ENABLE(MEDIA_STREAM)
    case MessageName::WebPageProxy_RequestUserMediaPermissionForFrame:
        return Vector<ArgumentDescription> {
            { "userMediaID", "WebCore::UserMediaRequestIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "userMediaDocumentOriginIdentifier", "WebCore::SecurityOriginData", nullptr, false },
            { "topLevelDocumentOriginIdentifier", "WebCore::SecurityOriginData", nullptr, false },
            { "request", "WebCore::MediaStreamRequest", nullptr, false },
        };
    case MessageName::WebPageProxy_EnumerateMediaDevicesForFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "userMediaDocumentOriginIdentifier", "WebCore::SecurityOriginData", nullptr, false },
            { "topLevelDocumentOriginIdentifier", "WebCore::SecurityOriginData", nullptr, false },
        };
    case MessageName::WebPageProxy_BeginMonitoringCaptureDevices:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(ENCRYPTED_MEDIA)
    case MessageName::WebPageProxy_RequestMediaKeySystemPermissionForFrame:
        return Vector<ArgumentDescription> {
            { "mediaKeySystemID", "WebCore::MediaKeySystemRequestIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "topLevelDocumentOriginIdentifier", "WebCore::SecurityOriginData", nullptr, false },
            { "keySystem", "String", nullptr, false },
        };
#endif
    case MessageName::WebPageProxy_RequestNotificationPermission:
        return Vector<ArgumentDescription> {
            { "originIdentifier", "String", nullptr, false },
        };
#if USE(UNIFIED_TEXT_CHECKING)
    case MessageName::WebPageProxy_CheckTextOfParagraph:
        return Vector<ArgumentDescription> {
            { "text", "String", nullptr, false },
            { "checkingTypes", "OptionSet<WebCore::TextCheckingType>", nullptr, false },
            { "insertionPoint", "int32_t", nullptr, false },
        };
#endif
    case MessageName::WebPageProxy_CheckSpellingOfString:
        return Vector<ArgumentDescription> {
            { "text", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_CheckGrammarOfString:
        return Vector<ArgumentDescription> {
            { "text", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_SpellingUIIsShowing:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_UpdateSpellingUIWithMisspelledWord:
        return Vector<ArgumentDescription> {
            { "misspelledWord", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_UpdateSpellingUIWithGrammarString:
        return Vector<ArgumentDescription> {
            { "badGrammarPhrase", "String", nullptr, false },
            { "grammarDetail", "WebCore::GrammarDetail", nullptr, false },
        };
    case MessageName::WebPageProxy_GetGuessesForWord:
        return Vector<ArgumentDescription> {
            { "word", "String", nullptr, false },
            { "context", "String", nullptr, false },
            { "insertionPoint", "int32_t", nullptr, false },
        };
    case MessageName::WebPageProxy_LearnWord:
        return Vector<ArgumentDescription> {
            { "word", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_IgnoreWord:
        return Vector<ArgumentDescription> {
            { "word", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_RequestCheckingOfString:
        return Vector<ArgumentDescription> {
            { "requestID", "WebKit::TextCheckerRequestID", nullptr, false },
            { "request", "WebCore::TextCheckingRequestData", nullptr, false },
            { "insertionPoint", "int32_t", nullptr, false },
        };
#if ENABLE(DRAG_SUPPORT)
    case MessageName::WebPageProxy_DidPerformDragControllerAction:
        return Vector<ArgumentDescription> {
            { "dragOperation", "WebCore::DragOperation", nullptr, true },
            { "dragHandlingMethod", "uint8_t", "WebCore::DragHandlingMethod", false },
            { "mouseIsOverFileInput", "bool", nullptr, false },
            { "numberOfItemsToBeAccepted", "unsigned", nullptr, false },
            { "insertionRect", "WebCore::IntRect", nullptr, false },
            { "editableElementRect", "WebCore::IntRect", nullptr, false },
        };
    case MessageName::WebPageProxy_DidEndDragging:
        return Vector<ArgumentDescription> { };
#endif
#if PLATFORM(COCOA) && ENABLE(DRAG_SUPPORT)
    case MessageName::WebPageProxy_StartDrag:
        return Vector<ArgumentDescription> {
            { "dragItem", "WebCore::DragItem", nullptr, false },
            { "dragImage", "WebKit::ShareableBitmapHandle", nullptr, false },
        };
    case MessageName::WebPageProxy_SetPromisedDataForImage:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "imageHandle", "WebKit::SharedMemory::Handle", nullptr, false },
            { "filename", "String", nullptr, false },
            { "extension", "String", nullptr, false },
            { "title", "String", nullptr, false },
            { "url", "String", nullptr, false },
            { "visibleURL", "String", nullptr, false },
            { "archiveHandle", "WebKit::SharedMemory::Handle", nullptr, false },
            { "originIdentifier", "String", nullptr, false },
        };
#endif
#if PLATFORM(GTK) && ENABLE(DRAG_SUPPORT)
    case MessageName::WebPageProxy_StartDrag:
        return Vector<ArgumentDescription> {
            { "selectionData", "WebCore::SelectionData", nullptr, false },
            { "dragOperationMask", "OptionSet<WebCore::DragOperation>", nullptr, false },
            { "dragImage", "WebKit::ShareableBitmapHandle", nullptr, false },
            { "dragImageHotspot", "WebCore::IntPoint", nullptr, false },
        };
#endif
#if ENABLE(DRAG_SUPPORT)
    case MessageName::WebPageProxy_DidPerformDragOperation:
        return Vector<ArgumentDescription> {
            { "handled", "bool", nullptr, false },
        };
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(DRAG_SUPPORT)
    case MessageName::WebPageProxy_DidHandleDragStartRequest:
        return Vector<ArgumentDescription> {
            { "started", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_DidHandleAdditionalDragItemsRequest:
        return Vector<ArgumentDescription> {
            { "added", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_WillReceiveEditDragSnapshot:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_DidReceiveEditDragSnapshot:
        return Vector<ArgumentDescription> {
            { "textIndicator", "WebCore::TextIndicatorData", nullptr, true },
        };
#endif
#if PLATFORM(COCOA)
    case MessageName::WebPageProxy_DidPerformDictionaryLookup:
        return Vector<ArgumentDescription> {
            { "dictionaryPopupInfo", "WebCore::DictionaryPopupInfo", nullptr, false },
        };
    case MessageName::WebPageProxy_ExecuteSavedCommandBySelector:
        return Vector<ArgumentDescription> {
            { "selector", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_RegisterWebProcessAccessibilityToken:
        return Vector<ArgumentDescription> {
            { "data", "IPC::DataReference", nullptr, false },
        };
    case MessageName::WebPageProxy_GetIsSpeaking:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_Speak:
        return Vector<ArgumentDescription> {
            { "string", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_StopSpeaking:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_MakeFirstResponder:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_AssistiveTechnologyMakeFirstResponder:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_SearchWithSpotlight:
        return Vector<ArgumentDescription> {
            { "string", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_SearchTheWeb:
        return Vector<ArgumentDescription> {
            { "string", "String", nullptr, false },
        };
#endif
#if HAVE(TOUCH_BAR)
    case MessageName::WebPageProxy_TouchBarMenuDataChanged:
        return Vector<ArgumentDescription> {
            { "touchBarMenuData", "WebKit::TouchBarMenuData", nullptr, false },
        };
    case MessageName::WebPageProxy_TouchBarMenuItemDataAdded:
        return Vector<ArgumentDescription> {
            { "touchBarMenuItemData", "WebKit::TouchBarMenuItemData", nullptr, false },
        };
    case MessageName::WebPageProxy_TouchBarMenuItemDataRemoved:
        return Vector<ArgumentDescription> {
            { "touchBarMenuItemData", "WebKit::TouchBarMenuItemData", nullptr, false },
        };
#endif
#if USE(APPKIT)
    case MessageName::WebPageProxy_SubstitutionsPanelIsShowing:
        return Vector<ArgumentDescription> { };
#endif
#if USE(AUTOMATIC_TEXT_REPLACEMENT)
    case MessageName::WebPageProxy_toggleSmartInsertDelete:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_toggleAutomaticQuoteSubstitution:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_toggleAutomaticLinkDetection:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_toggleAutomaticDashSubstitution:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_toggleAutomaticTextReplacement:
        return Vector<ArgumentDescription> { };
#endif
#if PLATFORM(MAC)
    case MessageName::WebPageProxy_ShowCorrectionPanel:
        return Vector<ArgumentDescription> {
            { "panelType", "uint8_t", "WebCore::AlternativeTextType", false },
            { "boundingBoxOfReplacedString", "WebCore::FloatRect", nullptr, false },
            { "replacedString", "String", nullptr, false },
            { "replacementString", "String", nullptr, false },
            { "alternativeReplacementStrings", "Vector<String>", nullptr, false },
        };
    case MessageName::WebPageProxy_DismissCorrectionPanel:
        return Vector<ArgumentDescription> {
            { "reason", "int32_t", nullptr, false },
        };
    case MessageName::WebPageProxy_DismissCorrectionPanelSoon:
        return Vector<ArgumentDescription> {
            { "reason", "int32_t", nullptr, false },
        };
    case MessageName::WebPageProxy_RecordAutocorrectionResponse:
        return Vector<ArgumentDescription> {
            { "response", "int32_t", nullptr, false },
            { "replacedString", "String", nullptr, false },
            { "replacementString", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_SetEditableElementIsFocused:
        return Vector<ArgumentDescription> {
            { "editableElementIsFocused", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_SetCaretDecorationVisibility:
        return Vector<ArgumentDescription> {
            { "visibility", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_HandleAcceptsFirstMouse:
        return Vector<ArgumentDescription> {
            { "acceptsFirstMouse", "bool", nullptr, false },
        };
#endif
    case MessageName::WebPageProxy_DidUpdateRenderingAfterCommittingLoad:
        return Vector<ArgumentDescription> { };
#if USE(DICTATION_ALTERNATIVES)
    case MessageName::WebPageProxy_ShowDictationAlternativeUI:
        return Vector<ArgumentDescription> {
            { "boundingBoxOfDictatedText", "WebCore::FloatRect", nullptr, false },
            { "dictationContext", "WebCore::DictationContext", nullptr, false },
        };
    case MessageName::WebPageProxy_RemoveDictationAlternatives:
        return Vector<ArgumentDescription> {
            { "dictationContext", "WebCore::DictationContext", nullptr, false },
        };
    case MessageName::WebPageProxy_DictationAlternatives:
        return Vector<ArgumentDescription> {
            { "dictationContext", "WebCore::DictationContext", nullptr, false },
        };
#endif
#if ENABLE(VIDEO)
    case MessageName::WebPageProxy_BeginTextRecognitionForVideoInElementFullScreen:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::MediaPlayerIdentifier", nullptr, false },
            { "videoBounds", "WebCore::FloatRect", nullptr, false },
        };
    case MessageName::WebPageProxy_CancelTextRecognitionForVideoInElementFullScreen:
        return Vector<ArgumentDescription> { };
#endif
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebPageProxy_CouldNotRestorePageState:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_RestorePageState:
        return Vector<ArgumentDescription> {
            { "scrollPosition", "WebCore::FloatPoint", nullptr, true },
            { "scrollOrigin", "WebCore::FloatPoint", nullptr, false },
            { "obscuredInsetsOnSave", "WebCore::RectEdges<float>", nullptr, false },
            { "scale", "double", nullptr, false },
        };
    case MessageName::WebPageProxy_RestorePageCenterAndScale:
        return Vector<ArgumentDescription> {
            { "unobscuredCenter", "WebCore::FloatPoint", nullptr, true },
            { "scale", "double", nullptr, false },
        };
    case MessageName::WebPageProxy_DidGetTapHighlightGeometries:
        return Vector<ArgumentDescription> {
            { "requestID", "WebKit::TapIdentifier", nullptr, false },
            { "color", "WebCore::Color", nullptr, false },
            { "geometries", "Vector<WebCore::FloatQuad>", nullptr, false },
            { "topLeftRadius", "WebCore::IntSize", nullptr, false },
            { "topRightRadius", "WebCore::IntSize", nullptr, false },
            { "bottomLeftRadius", "WebCore::IntSize", nullptr, false },
            { "bottomRightRadius", "WebCore::IntSize", nullptr, false },
            { "nodeHasBuiltInClickHandling", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_ElementDidFocus:
        return Vector<ArgumentDescription> {
            { "information", "WebKit::FocusedElementInformation", nullptr, false },
            { "userIsInteracting", "bool", nullptr, false },
            { "blurPreviousNode", "bool", nullptr, false },
            { "activityStateChanges", "OptionSet<WebCore::ActivityState::Flag>", nullptr, false },
            { "userData", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebPageProxy_ElementDidBlur:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_UpdateInputContextAfterBlurringAndRefocusingElement:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_FocusedElementDidChangeInputMode:
        return Vector<ArgumentDescription> {
            { "mode", "uint8_t", "WebCore::InputMode", false },
        };
    case MessageName::WebPageProxy_ScrollingNodeScrollWillStartScroll:
        return Vector<ArgumentDescription> {
            { "nodeID", "uint64_t", nullptr, false },
        };
    case MessageName::WebPageProxy_ScrollingNodeScrollDidEndScroll:
        return Vector<ArgumentDescription> {
            { "nodeID", "uint64_t", nullptr, false },
        };
    case MessageName::WebPageProxy_ShowInspectorHighlight:
        return Vector<ArgumentDescription> {
            { "highlight", "WebCore::InspectorOverlay::Highlight", nullptr, false },
        };
    case MessageName::WebPageProxy_HideInspectorHighlight:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_ShowInspectorIndication:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_HideInspectorIndication:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_EnableInspectorNodeSearch:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_DisableInspectorNodeSearch:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_UpdateStringForFind:
        return Vector<ArgumentDescription> {
            { "findString", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_HandleAutocorrectionContext:
        return Vector<ArgumentDescription> {
            { "context", "WebKit::WebAutocorrectionContext", nullptr, false },
        };
    case MessageName::WebPageProxy_ShowDataDetectorsUIForPositionInformation:
        return Vector<ArgumentDescription> {
            { "information", "WebKit::InteractionInformationAtPosition", nullptr, false },
        };
#endif
    case MessageName::WebPageProxy_DidChangeInspectorFrontendCount:
        return Vector<ArgumentDescription> {
            { "count", "uint32_t", nullptr, false },
        };
    case MessageName::WebPageProxy_CreateInspectorTarget:
        return Vector<ArgumentDescription> {
            { "targetId", "String", nullptr, false },
            { "type", "uint8_t", "Inspector::InspectorTargetType", false },
        };
    case MessageName::WebPageProxy_DestroyInspectorTarget:
        return Vector<ArgumentDescription> {
            { "targetId", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_SendMessageToInspectorFrontend:
        return Vector<ArgumentDescription> {
            { "targetId", "String", nullptr, false },
            { "message", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_SaveRecentSearches:
        return Vector<ArgumentDescription> {
            { "name", "String", nullptr, false },
            { "searchItems", "Vector<WebCore::RecentSearch>", nullptr, false },
        };
    case MessageName::WebPageProxy_LoadRecentSearches:
        return Vector<ArgumentDescription> {
            { "name", "String", nullptr, false },
        };
#if ENABLE(PDFKIT_PLUGIN) && !ENABLE(UI_PROCESS_PDF_HUD)
    case MessageName::WebPageProxy_SavePDFToFileInDownloadsFolder:
        return Vector<ArgumentDescription> {
            { "suggestedFilename", "String", nullptr, false },
            { "originatingURL", "URL", nullptr, false },
            { "data", "IPC::DataReference", nullptr, false },
        };
    case MessageName::WebPageProxy_SavePDFToTemporaryFolderAndOpenWithNativeApplication:
        return Vector<ArgumentDescription> {
            { "suggestedFilename", "String", nullptr, false },
            { "sourceFrameInfo", "WebKit::FrameInfoData", nullptr, false },
            { "data", "IPC::DataReference", nullptr, false },
            { "pdfUUID", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_OpenPDFFromTemporaryFolderWithNativeApplication:
        return Vector<ArgumentDescription> {
            { "sourceFrameInfo", "WebKit::FrameInfoData", nullptr, false },
            { "pdfUUID", "String", nullptr, false },
        };
#endif
#if ENABLE(PDFKIT_PLUGIN)
    case MessageName::WebPageProxy_ShowPDFContextMenu:
        return Vector<ArgumentDescription> {
            { "contextMenu", "WebKit::PDFContextMenu", nullptr, false },
            { "identifier", "WebKit::PDFPluginIdentifier", nullptr, false },
        };
#endif
    case MessageName::WebPageProxy_DidUpdateActivityState:
        return Vector<ArgumentDescription> { };
#if ENABLE(WEB_CRYPTO)
    case MessageName::WebPageProxy_WrapCryptoKey:
        return Vector<ArgumentDescription> {
            { "key", "Vector<uint8_t>", nullptr, false },
        };
    case MessageName::WebPageProxy_UnwrapCryptoKey:
        return Vector<ArgumentDescription> {
            { "wrappedKey", "Vector<uint8_t>", nullptr, false },
        };
#endif
#if (ENABLE(TELEPHONE_NUMBER_DETECTION) && PLATFORM(MAC))
    case MessageName::WebPageProxy_ShowTelephoneNumberMenu:
        return Vector<ArgumentDescription> {
            { "telephoneNumber", "String", nullptr, false },
            { "point", "WebCore::IntPoint", nullptr, false },
            { "rect", "WebCore::IntRect", nullptr, false },
        };
#endif
#if USE(QUICK_LOOK)
    case MessageName::WebPageProxy_DidStartLoadForQuickLookDocumentInMainFrame:
        return Vector<ArgumentDescription> {
            { "fileName", "String", nullptr, false },
            { "uti", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_DidFinishLoadForQuickLookDocumentInMainFrame:
        return Vector<ArgumentDescription> {
            { "resource", "WebKit::ShareableResource::Handle", nullptr, false },
        };
    case MessageName::WebPageProxy_RequestPasswordForQuickLookDocumentInMainFrame:
        return Vector<ArgumentDescription> {
            { "fileName", "String", nullptr, false },
        };
#endif
#if ENABLE(CONTENT_FILTERING)
    case MessageName::WebPageProxy_ContentFilterDidBlockLoadForFrame:
        return Vector<ArgumentDescription> {
            { "unblockHandler", "WebCore::ContentFilterUnblockHandler", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
        };
#endif
#if ENABLE(ACCESSIBILITY_ANIMATION_CONTROL)
    case MessageName::WebPageProxy_IsAnyAnimationAllowedToPlayDidChange:
        return Vector<ArgumentDescription> {
            { "anyAnimationCanPlay", "bool", nullptr, false },
        };
#endif
    case MessageName::WebPageProxy_IsPlayingMediaDidChange:
        return Vector<ArgumentDescription> {
            { "state", "OptionSet<WebCore::MediaProducerMediaState>", nullptr, false },
        };
    case MessageName::WebPageProxy_HandleAutoplayEvent:
        return Vector<ArgumentDescription> {
            { "event", "uint8_t", "WebCore::AutoplayEvent", false },
            { "flags", "OptionSet<WebCore::AutoplayEventFlags>", nullptr, false },
        };
#if PLATFORM(MAC)
    case MessageName::WebPageProxy_DidPerformImmediateActionHitTest:
        return Vector<ArgumentDescription> {
            { "result", "WebKit::WebHitTestResultData", nullptr, false },
            { "contentPreventsDefault", "bool", nullptr, false },
            { "userData", "WebKit::UserData", nullptr, false },
        };
#endif
    case MessageName::WebPageProxy_HandleMessage:
        return Vector<ArgumentDescription> {
            { "messageName", "String", nullptr, false },
            { "messageBody", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebPageProxy_HandleSynchronousMessage:
        return Vector<ArgumentDescription> {
            { "messageName", "String", nullptr, false },
            { "messageBody", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebPageProxy_HandleAutoFillButtonClick:
        return Vector<ArgumentDescription> {
            { "userData", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebPageProxy_DidResignInputElementStrongPasswordAppearance:
        return Vector<ArgumentDescription> {
            { "userData", "WebKit::UserData", nullptr, false },
        };
#if ENABLE(CONTENT_EXTENSIONS)
    case MessageName::WebPageProxy_ContentRuleListNotification:
        return Vector<ArgumentDescription> {
            { "url", "URL", nullptr, false },
            { "results", "WebCore::ContentRuleListResults", nullptr, false },
        };
#endif
#if ENABLE(WIRELESS_PLAYBACK_TARGET) && !PLATFORM(IOS_FAMILY)
    case MessageName::WebPageProxy_AddPlaybackTargetPickerClient:
        return Vector<ArgumentDescription> {
            { "contextId", "WebCore::PlaybackTargetClientContextIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_RemovePlaybackTargetPickerClient:
        return Vector<ArgumentDescription> {
            { "contextId", "WebCore::PlaybackTargetClientContextIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_ShowPlaybackTargetPicker:
        return Vector<ArgumentDescription> {
            { "contextId", "WebCore::PlaybackTargetClientContextIdentifier", nullptr, false },
            { "pickerLocation", "WebCore::FloatRect", nullptr, false },
            { "hasVideo", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_PlaybackTargetPickerClientStateDidChange:
        return Vector<ArgumentDescription> {
            { "contextId", "WebCore::PlaybackTargetClientContextIdentifier", nullptr, false },
            { "mediaState", "OptionSet<WebCore::MediaProducerMediaState>", nullptr, false },
        };
    case MessageName::WebPageProxy_SetMockMediaPlaybackTargetPickerEnabled:
        return Vector<ArgumentDescription> {
            { "enabled", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_SetMockMediaPlaybackTargetPickerState:
        return Vector<ArgumentDescription> {
            { "name", "String", nullptr, false },
            { "pickerState", "uint8_t", "WebCore::MediaPlaybackTargetContext::MockState", false },
        };
    case MessageName::WebPageProxy_MockMediaPlaybackTargetPickerDismissPopup:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(VIDEO_PRESENTATION_MODE)
    case MessageName::WebPageProxy_SetMockVideoPresentationModeEnabled:
        return Vector<ArgumentDescription> {
            { "enabled", "bool", nullptr, false },
        };
#endif
#if ENABLE(POINTER_LOCK)
    case MessageName::WebPageProxy_RequestPointerLock:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_RequestPointerUnlock:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebPageProxy_DidFailToSuspendAfterProcessSwap:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_DidSuspendAfterProcessSwap:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_ImageOrMediaDocumentSizeChanged:
        return Vector<ArgumentDescription> {
            { "newSize", "WebCore::IntSize", nullptr, false },
        };
    case MessageName::WebPageProxy_UseFixedLayoutDidChange:
        return Vector<ArgumentDescription> {
            { "useFixedLayout", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_FixedLayoutSizeDidChange:
        return Vector<ArgumentDescription> {
            { "fixedLayoutSize", "WebCore::IntSize", nullptr, false },
        };
    case MessageName::WebPageProxy_DidRestoreScrollPosition:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_RequestScrollToRect:
        return Vector<ArgumentDescription> {
            { "targetRect", "WebCore::FloatRect", nullptr, false },
            { "origin", "WebCore::FloatPoint", nullptr, false },
        };
    case MessageName::WebPageProxy_GetLoadDecisionForIcon:
        return Vector<ArgumentDescription> {
            { "icon", "WebCore::LinkIcon", nullptr, false },
            { "callbackID", "WebKit::CallbackID", nullptr, false },
        };
#if PLATFORM(MAC)
    case MessageName::WebPageProxy_DidHandleAcceptedCandidate:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebPageProxy_SetIsUsingHighPerformanceWebGL:
        return Vector<ArgumentDescription> {
            { "isUsingHighPerformanceWebGL", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_StartURLSchemeTask:
        return Vector<ArgumentDescription> {
            { "parameters", "WebKit::URLSchemeTaskParameters", nullptr, false },
        };
    case MessageName::WebPageProxy_StopURLSchemeTask:
        return Vector<ArgumentDescription> {
            { "handlerIdentifier", "WebKit::WebURLSchemeHandlerIdentifier", nullptr, false },
            { "taskIdentifier", "WebCore::ResourceLoaderIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_LoadSynchronousURLSchemeTask:
        return Vector<ArgumentDescription> {
            { "parameters", "WebKit::URLSchemeTaskParameters", nullptr, false },
        };
#if ENABLE(DEVICE_ORIENTATION)
    case MessageName::WebPageProxy_ShouldAllowDeviceOrientationAndMotionAccess:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "frameInfo", "WebKit::FrameInfoData", nullptr, false },
            { "mayPrompt", "bool", nullptr, false },
        };
#endif
#if ENABLE(ATTACHMENT_ELEMENT)
    case MessageName::WebPageProxy_RegisterAttachmentIdentifierFromData:
        return Vector<ArgumentDescription> {
            { "identifier", "String", nullptr, false },
            { "contentType", "String", nullptr, false },
            { "preferredFileName", "String", nullptr, false },
            { "data", "IPC::SharedBufferReference", nullptr, false },
        };
    case MessageName::WebPageProxy_RegisterAttachmentIdentifierFromFilePath:
        return Vector<ArgumentDescription> {
            { "identifier", "String", nullptr, false },
            { "contentType", "String", nullptr, false },
            { "filePath", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_RegisterAttachmentIdentifier:
        return Vector<ArgumentDescription> {
            { "identifier", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_RegisterAttachmentsFromSerializedData:
        return Vector<ArgumentDescription> {
            { "data", "Vector<WebCore::SerializedAttachmentData>", nullptr, false },
        };
    case MessageName::WebPageProxy_CloneAttachmentData:
        return Vector<ArgumentDescription> {
            { "fromIdentifier", "String", nullptr, false },
            { "toIdentifier", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_DidInsertAttachmentWithIdentifier:
        return Vector<ArgumentDescription> {
            { "identifier", "String", nullptr, false },
            { "source", "String", nullptr, false },
            { "hasEnclosingImage", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_DidRemoveAttachmentWithIdentifier:
        return Vector<ArgumentDescription> {
            { "identifier", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_SerializedAttachmentDataForIdentifiers:
        return Vector<ArgumentDescription> {
            { "identifiers", "Vector<String>", nullptr, false },
        };
#endif
#if (ENABLE(ATTACHMENT_ELEMENT) && PLATFORM(IOS_FAMILY))
    case MessageName::WebPageProxy_WritePromisedAttachmentToPasteboard:
        return Vector<ArgumentDescription> {
            { "info", "WebCore::PromisedAttachmentInfo", nullptr, false },
            { "authorizationToken", "String", nullptr, false },
        };
#endif
#if ENABLE(ATTACHMENT_ELEMENT)
    case MessageName::WebPageProxy_RequestAttachmentIcon:
        return Vector<ArgumentDescription> {
            { "identifier", "String", nullptr, false },
            { "contentType", "String", nullptr, false },
            { "path", "String", nullptr, false },
            { "title", "String", nullptr, false },
            { "size", "WebCore::FloatSize", nullptr, false },
        };
#endif
#if ENABLE(APP_HIGHLIGHTS)
    case MessageName::WebPageProxy_StoreAppHighlight:
        return Vector<ArgumentDescription> {
            { "info", "WebCore::AppHighlight", nullptr, false },
        };
#endif
#if ENABLE(SPEECH_SYNTHESIS)
    case MessageName::WebPageProxy_SpeechSynthesisVoiceList:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_SpeechSynthesisSpeak:
        return Vector<ArgumentDescription> {
            { "text", "String", nullptr, false },
            { "lang", "String", nullptr, false },
            { "volume", "float", nullptr, false },
            { "rate", "float", nullptr, false },
            { "pitch", "float", nullptr, false },
            { "startTime", "MonotonicTime", nullptr, false },
            { "voiceURI", "String", nullptr, false },
            { "voiceName", "String", nullptr, false },
            { "voiceLang", "String", nullptr, false },
            { "localService", "bool", nullptr, false },
            { "defaultVoice", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_SpeechSynthesisSetFinishedCallback:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_SpeechSynthesisCancel:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_SpeechSynthesisPause:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_SpeechSynthesisResume:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_SpeechSynthesisResetState:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(UI_PROCESS_PDF_HUD)
    case MessageName::WebPageProxy_CreatePDFHUD:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::PDFPluginIdentifier", nullptr, false },
            { "boundingBox", "WebCore::IntRect", nullptr, false },
        };
    case MessageName::WebPageProxy_UpdatePDFHUDLocation:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::PDFPluginIdentifier", nullptr, false },
            { "boundingBox", "WebCore::IntRect", nullptr, false },
        };
    case MessageName::WebPageProxy_RemovePDFHUD:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::PDFPluginIdentifier", nullptr, false },
        };
#endif
    case MessageName::WebPageProxy_ConfigureLoggingChannel:
        return Vector<ArgumentDescription> {
            { "channelName", "String", nullptr, false },
            { "state", "uint8_t", "WTFLogChannelState", false },
            { "level", "uint8_t", "WTFLogLevel", false },
        };
#if PLATFORM(GTK)
    case MessageName::WebPageProxy_ShowEmojiPicker:
        return Vector<ArgumentDescription> {
            { "caretRect", "WebCore::IntRect", nullptr, false },
        };
#endif
#if HAVE(VISIBILITY_PROPAGATION_VIEW)
    case MessageName::WebPageProxy_DidCreateContextInWebProcessForVisibilityPropagation:
        return Vector<ArgumentDescription> {
            { "contextID", "WebKit::LayerHostingContextID", nullptr, false },
        };
#endif
#if ENABLE(WEB_AUTHN)
    case MessageName::WebPageProxy_SetMockWebAuthenticationConfiguration:
        return Vector<ArgumentDescription> {
            { "configuration", "WebCore::MockWebAuthenticationConfiguration", nullptr, false },
        };
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    case MessageName::WebPageProxy_SendMessageToWebView:
        return Vector<ArgumentDescription> {
            { "userMessage", "WebKit::UserMessage", nullptr, false },
        };
    case MessageName::WebPageProxy_SendMessageToWebViewWithReply:
        return Vector<ArgumentDescription> {
            { "userMessage", "WebKit::UserMessage", nullptr, false },
        };
#endif
    case MessageName::WebPageProxy_DidFindTextManipulationItems:
        return Vector<ArgumentDescription> {
            { "items", "Vector<WebCore::TextManipulationItem>", nullptr, false },
        };
#if ENABLE(MEDIA_USAGE)
    case MessageName::WebPageProxy_AddMediaUsageManagerSession:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::MediaSessionIdentifier", nullptr, false },
            { "bundleIdentifier", "String", nullptr, false },
            { "pageURL", "URL", nullptr, false },
        };
    case MessageName::WebPageProxy_UpdateMediaUsageManagerSessionState:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::MediaSessionIdentifier", nullptr, false },
            { "info", "WebCore::MediaUsageInfo", nullptr, false },
        };
    case MessageName::WebPageProxy_RemoveMediaUsageManagerSession:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::MediaSessionIdentifier", nullptr, false },
        };
#endif
    case MessageName::WebPageProxy_SetHasExecutedAppBoundBehaviorBeforeNavigation:
        return Vector<ArgumentDescription> { };
#if PLATFORM(MAC)
    case MessageName::WebPageProxy_ChangeUniversalAccessZoomFocus:
        return Vector<ArgumentDescription> {
            { "viewRect", "WebCore::IntRect", nullptr, false },
            { "caretRect", "WebCore::IntRect", nullptr, false },
        };
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW_IOS)
    case MessageName::WebPageProxy_TakeModelElementFullscreen:
        return Vector<ArgumentDescription> {
            { "modelIdentifier", "WebKit::ModelIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementSetInteractionEnabled:
        return Vector<ArgumentDescription> {
            { "modelIdentifier", "WebKit::ModelIdentifier", nullptr, false },
            { "isInteractionEnabled", "bool", nullptr, false },
        };
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW_MAC)
    case MessageName::WebPageProxy_ModelElementCreateRemotePreview:
        return Vector<ArgumentDescription> {
            { "uuid", "String", nullptr, false },
            { "size", "WebCore::FloatSize", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementLoadRemotePreview:
        return Vector<ArgumentDescription> {
            { "uuid", "String", nullptr, false },
            { "url", "URL", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementDestroyRemotePreview:
        return Vector<ArgumentDescription> {
            { "uuid", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementSizeDidChange:
        return Vector<ArgumentDescription> {
            { "uuid", "String", nullptr, false },
            { "size", "WebCore::FloatSize", nullptr, false },
        };
    case MessageName::WebPageProxy_HandleMouseDownForModelElement:
        return Vector<ArgumentDescription> {
            { "uuid", "String", nullptr, false },
            { "flippedLocationInElement", "WebCore::LayoutPoint", nullptr, false },
            { "timestamp", "MonotonicTime", nullptr, false },
        };
    case MessageName::WebPageProxy_HandleMouseMoveForModelElement:
        return Vector<ArgumentDescription> {
            { "uuid", "String", nullptr, false },
            { "flippedLocationInElement", "WebCore::LayoutPoint", nullptr, false },
            { "timestamp", "MonotonicTime", nullptr, false },
        };
    case MessageName::WebPageProxy_HandleMouseUpForModelElement:
        return Vector<ArgumentDescription> {
            { "uuid", "String", nullptr, false },
            { "flippedLocationInElement", "WebCore::LayoutPoint", nullptr, false },
            { "timestamp", "MonotonicTime", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelInlinePreviewUUIDs:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW)
    case MessageName::WebPageProxy_ModelElementGetCamera:
        return Vector<ArgumentDescription> {
            { "modelIdentifier", "WebKit::ModelIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementSetCamera:
        return Vector<ArgumentDescription> {
            { "modelIdentifier", "WebKit::ModelIdentifier", nullptr, false },
            { "camera", "WebCore::HTMLModelElementCamera", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementIsPlayingAnimation:
        return Vector<ArgumentDescription> {
            { "modelIdentifier", "WebKit::ModelIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementSetAnimationIsPlaying:
        return Vector<ArgumentDescription> {
            { "modelIdentifier", "WebKit::ModelIdentifier", nullptr, false },
            { "isPlaying", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementIsLoopingAnimation:
        return Vector<ArgumentDescription> {
            { "modelIdentifier", "WebKit::ModelIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementSetIsLoopingAnimation:
        return Vector<ArgumentDescription> {
            { "modelIdentifier", "WebKit::ModelIdentifier", nullptr, false },
            { "isLooping", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementAnimationDuration:
        return Vector<ArgumentDescription> {
            { "modelIdentifier", "WebKit::ModelIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementAnimationCurrentTime:
        return Vector<ArgumentDescription> {
            { "modelIdentifier", "WebKit::ModelIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementSetAnimationCurrentTime:
        return Vector<ArgumentDescription> {
            { "modelIdentifier", "WebKit::ModelIdentifier", nullptr, false },
            { "currentTime", "Seconds", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementHasAudio:
        return Vector<ArgumentDescription> {
            { "modelIdentifier", "WebKit::ModelIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementIsMuted:
        return Vector<ArgumentDescription> {
            { "modelIdentifier", "WebKit::ModelIdentifier", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementSetIsMuted:
        return Vector<ArgumentDescription> {
            { "modelIdentifier", "WebKit::ModelIdentifier", nullptr, false },
            { "isMuted", "bool", nullptr, false },
        };
#endif
#if ENABLE(APPLE_PAY_AMS_UI)
    case MessageName::WebPageProxy_StartApplePayAMSUISession:
        return Vector<ArgumentDescription> {
            { "originatingURL", "URL", nullptr, false },
            { "request", "WebCore::ApplePayAMSUIRequest", nullptr, false },
        };
    case MessageName::WebPageProxy_AbortApplePayAMSUISession:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebPageProxy_RequestCookieConsent:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_ClassifyModalContainerControls:
        return Vector<ArgumentDescription> {
            { "texts", "Vector<String>", nullptr, false },
        };
    case MessageName::WebPageProxy_DecidePolicyForModalContainer:
        return Vector<ArgumentDescription> {
            { "types", "OptionSet<WebCore::ModalContainerControlType>", nullptr, false },
        };
    case MessageName::WebPageProxy_DidChangeLookalikeCharacters:
        return Vector<ArgumentDescription> {
            { "originalURL", "URL", nullptr, false },
            { "adjustedURL", "URL", nullptr, false },
        };
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebPasteboardProxy_WriteURLToPasteboard:
        return Vector<ArgumentDescription> {
            { "url", "WebCore::PasteboardURL", nullptr, false },
            { "pasteboardName", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_WriteWebContentToPasteboard:
        return Vector<ArgumentDescription> {
            { "content", "WebCore::PasteboardWebContent", nullptr, false },
            { "pasteboardName", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_WriteImageToPasteboard:
        return Vector<ArgumentDescription> {
            { "pasteboardImage", "WebCore::PasteboardImage", nullptr, false },
            { "pasteboardName", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_WriteStringToPasteboard:
        return Vector<ArgumentDescription> {
            { "pasteboardType", "String", nullptr, false },
            { "text", "String", nullptr, false },
            { "pasteboardName", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_UpdateSupportedTypeIdentifiers:
        return Vector<ArgumentDescription> {
            { "identifiers", "Vector<String>", nullptr, false },
            { "pasteboardName", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
#endif
    case MessageName::WebPasteboardProxy_WriteCustomData:
        return Vector<ArgumentDescription> {
            { "data", "Vector<WebCore::PasteboardCustomData>", nullptr, false },
            { "pasteboardName", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_TypesSafeForDOMToReadAndWrite:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "origin", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_AllPasteboardItemInfo:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "changeCount", "int64_t", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_InformationForItemAtIndex:
        return Vector<ArgumentDescription> {
            { "index", "size_t", nullptr, false },
            { "pasteboardName", "String", nullptr, false },
            { "changeCount", "int64_t", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_GetPasteboardItemsCount:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_ReadStringFromPasteboard:
        return Vector<ArgumentDescription> {
            { "index", "size_t", nullptr, false },
            { "pasteboardType", "String", nullptr, false },
            { "pasteboardName", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_ReadURLFromPasteboard:
        return Vector<ArgumentDescription> {
            { "index", "size_t", nullptr, false },
            { "pasteboardName", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_ReadBufferFromPasteboard:
        return Vector<ArgumentDescription> {
            { "index", "size_t", nullptr, true },
            { "pasteboardType", "String", nullptr, false },
            { "pasteboardName", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_ContainsStringSafeForDOMToReadForType:
        return Vector<ArgumentDescription> {
            { "type", "String", nullptr, false },
            { "pasteboardName", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
#if PLATFORM(COCOA)
    case MessageName::WebPasteboardProxy_GetNumberOfFiles:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_GetPasteboardTypes:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_GetPasteboardPathnamesForType:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "pasteboardType", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_GetPasteboardStringForType:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "pasteboardType", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_GetPasteboardStringsForType:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "pasteboardType", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_GetPasteboardBufferForType:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "pasteboardType", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_GetPasteboardChangeCount:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_GetPasteboardColor:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_GetPasteboardURL:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_AddPasteboardTypes:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "pasteboardTypes", "Vector<String>", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_SetPasteboardTypes:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "pasteboardTypes", "Vector<String>", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_SetPasteboardURL:
        return Vector<ArgumentDescription> {
            { "pasteboardURL", "WebCore::PasteboardURL", nullptr, false },
            { "pasteboardName", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_SetPasteboardColor:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "color", "WebCore::Color", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_SetPasteboardStringForType:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "pasteboardType", "String", nullptr, false },
            { "string", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_SetPasteboardBufferForType:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "pasteboardType", "String", nullptr, false },
            { "buffer", "RefPtr<WebCore::SharedBuffer>", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_ContainsURLStringSuitableForLoading:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_URLStringSuitableForLoading:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
#endif
#if (PLATFORM(COCOA) && ENABLE(IPC_TESTING_API))
    case MessageName::WebPasteboardProxy_TestIPCSharedMemory:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "pasteboardType", "String", nullptr, false },
            { "handle", "WebKit::SharedMemory::Handle", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, true },
        };
#endif
#if PLATFORM(GTK)
    case MessageName::WebPasteboardProxy_GetTypes:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_ReadText:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_ReadFilePaths:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_ReadBuffer:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "pasteboardType", "String", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_WriteToClipboard:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
            { "pasteboardContent", "WebCore::SelectionData", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_ClearClipboard:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_GetPasteboardChangeCount:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
        };
#endif
#if USE(LIBWPE)
    case MessageName::WebPasteboardProxy_GetPasteboardTypes:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPasteboardProxy_WriteWebContentToPasteboard:
        return Vector<ArgumentDescription> {
            { "content", "WebCore::PasteboardWebContent", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_WriteStringToPasteboard:
        return Vector<ArgumentDescription> {
            { "pasteboardType", "String", nullptr, false },
            { "text", "String", nullptr, false },
        };
#endif
    case MessageName::WebPermissionControllerProxy_Query:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::ClientOrigin", nullptr, false },
            { "descriptor", "WebCore::PermissionDescriptor", nullptr, false },
            { "identifier", "WebKit::WebPageProxyIdentifier", nullptr, true },
            { "source", "uint8_t", "WebCore::PermissionQuerySource", false },
        };
    case MessageName::WebProcessPool_HandleMessage:
        return Vector<ArgumentDescription> {
            { "messageName", "String", nullptr, false },
            { "messageBody", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebProcessPool_HandleSynchronousMessage:
        return Vector<ArgumentDescription> {
            { "messageName", "String", nullptr, false },
            { "messageBody", "WebKit::UserData", nullptr, false },
        };
#if ENABLE(GAMEPAD)
    case MessageName::WebProcessPool_StartedUsingGamepads:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcessPool_StoppedUsingGamepads:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcessPool_PlayGamepadEffect:
        return Vector<ArgumentDescription> {
            { "gamepadIndex", "unsigned", nullptr, false },
            { "gamepadID", "String", nullptr, false },
            { "type", "uint8_t", "WebCore::GamepadHapticEffectType", false },
            { "parameters", "WebCore::GamepadEffectParameters", nullptr, false },
        };
    case MessageName::WebProcessPool_StopGamepadEffects:
        return Vector<ArgumentDescription> {
            { "gamepadIndex", "unsigned", nullptr, false },
            { "gamepadID", "String", nullptr, false },
        };
#endif
    case MessageName::WebProcessPool_ReportWebContentCPUTime:
        return Vector<ArgumentDescription> {
            { "cpuTime", "Seconds", nullptr, false },
            { "activityState", "uint64_t", nullptr, false },
        };
    case MessageName::WebProcessProxy_UpdateBackForwardItem:
        return Vector<ArgumentDescription> {
            { "backForwardListItemState", "WebKit::BackForwardListItemState", nullptr, false },
        };
    case MessageName::WebProcessProxy_DidDestroyFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "pageID", "WebKit::WebPageProxyIdentifier", nullptr, false },
        };
    case MessageName::WebProcessProxy_DidDestroyUserGestureToken:
        return Vector<ArgumentDescription> {
            { "userGestureTokenID", "uint64_t", nullptr, false },
        };
    case MessageName::WebProcessProxy_ShouldTerminate:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcessProxy_EnableSuddenTermination:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcessProxy_DisableSuddenTermination:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcessProxy_GetNetworkProcessConnection:
        return Vector<ArgumentDescription> { };
#if ENABLE(GPU_PROCESS)
    case MessageName::WebProcessProxy_CreateGPUProcessConnection:
        return Vector<ArgumentDescription> {
            { "connectionHandle", "IPC::Connection::Handle", nullptr, false },
            { "parameters", "WebKit::GPUProcessConnectionParameters", nullptr, false },
        };
#endif
    case MessageName::WebProcessProxy_SetIsHoldingLockedFiles:
        return Vector<ArgumentDescription> {
            { "isHoldingLockedFiles", "bool", nullptr, false },
        };
    case MessageName::WebProcessProxy_DidExceedActiveMemoryLimit:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcessProxy_DidExceedInactiveMemoryLimit:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcessProxy_DidExceedCPULimit:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcessProxy_StopResponsivenessTimer:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcessProxy_DidReceiveBackgroundResponsivenessPing:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcessProxy_MemoryPressureStatusChanged:
        return Vector<ArgumentDescription> {
            { "isUnderMemoryPressure", "bool", nullptr, false },
        };
    case MessageName::WebProcessProxy_DidCollectPrewarmInformation:
        return Vector<ArgumentDescription> {
            { "domain", "WebCore::RegistrableDomain", nullptr, false },
            { "prewarmInformation", "WebCore::PrewarmInformation", nullptr, false },
        };
#if PLATFORM(COCOA)
    case MessageName::WebProcessProxy_CacheMediaMIMETypes:
        return Vector<ArgumentDescription> {
            { "types", "Vector<String>", nullptr, false },
        };
#endif
#if PLATFORM(MAC)
    case MessageName::WebProcessProxy_RequestHighPerformanceGPU:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcessProxy_ReleaseHighPerformanceGPU:
        return Vector<ArgumentDescription> { };
#endif
#if HAVE(CVDISPLAYLINK)
    case MessageName::WebProcessProxy_StartDisplayLink:
        return Vector<ArgumentDescription> {
            { "observerID", "WebKit::DisplayLinkObserverID", nullptr, false },
            { "displayID", "uint32_t", nullptr, false },
            { "preferredFramesPerSecond", "unsigned", nullptr, false },
        };
    case MessageName::WebProcessProxy_StopDisplayLink:
        return Vector<ArgumentDescription> {
            { "observerID", "WebKit::DisplayLinkObserverID", nullptr, false },
            { "displayID", "uint32_t", nullptr, false },
        };
    case MessageName::WebProcessProxy_SetDisplayLinkPreferredFramesPerSecond:
        return Vector<ArgumentDescription> {
            { "observerID", "WebKit::DisplayLinkObserverID", nullptr, false },
            { "displayID", "uint32_t", nullptr, false },
            { "preferredFramesPerSecond", "unsigned", nullptr, false },
        };
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    case MessageName::WebProcessProxy_SendMessageToWebContext:
        return Vector<ArgumentDescription> {
            { "userMessage", "WebKit::UserMessage", nullptr, false },
        };
    case MessageName::WebProcessProxy_SendMessageToWebContextWithReply:
        return Vector<ArgumentDescription> {
            { "userMessage", "WebKit::UserMessage", nullptr, false },
        };
#endif
    case MessageName::WebProcessProxy_DidCreateSleepDisabler:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::SleepDisablerIdentifier", nullptr, false },
            { "reason", "String", nullptr, false },
            { "display", "bool", nullptr, false },
        };
    case MessageName::WebProcessProxy_DidDestroySleepDisabler:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::SleepDisablerIdentifier", nullptr, false },
        };
    case MessageName::WebProcessProxy_CreateSpeechRecognitionServer:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::PageIdentifier", nullptr, false },
        };
    case MessageName::WebProcessProxy_DestroySpeechRecognitionServer:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::PageIdentifier", nullptr, false },
        };
    case MessageName::WebProcessProxy_SystemBeep:
        return Vector<ArgumentDescription> { };
#if PLATFORM(MAC)
    case MessageName::WebProcessProxy_IsAXAuthenticated:
        return Vector<ArgumentDescription> {
            { "auditToken", "audit_token_t", nullptr, false },
        };
#endif
#if HAVE(MEDIA_ACCESSIBILITY_FRAMEWORK)
    case MessageName::WebProcessProxy_SetCaptionDisplayMode:
        return Vector<ArgumentDescription> {
            { "mode", "WebCore::CaptionUserPreferences::CaptionDisplayMode", nullptr, false },
        };
    case MessageName::WebProcessProxy_SetCaptionLanguage:
        return Vector<ArgumentDescription> {
            { "language", "String", nullptr, false },
        };
#endif
    case MessageName::WebProcessProxy_GetNotifications:
        return Vector<ArgumentDescription> {
            { "registrationURL", "URL", nullptr, false },
            { "tag", "String", nullptr, false },
        };
    case MessageName::WebProcessProxy_SetAppBadge:
        return Vector<ArgumentDescription> {
            { "pageIdentifier", "WebKit::WebPageProxyIdentifier", nullptr, true },
            { "origin", "WebCore::SecurityOriginData", nullptr, false },
            { "badge", "uint64_t", nullptr, true },
        };
    case MessageName::WebProcessProxy_SetClientBadge:
        return Vector<ArgumentDescription> {
            { "pageIdentifier", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "origin", "WebCore::SecurityOriginData", nullptr, false },
            { "badge", "uint64_t", nullptr, true },
        };
    case MessageName::WebScreenOrientationManagerProxy_CurrentOrientation:
        return Vector<ArgumentDescription> { };
    case MessageName::WebScreenOrientationManagerProxy_Lock:
        return Vector<ArgumentDescription> {
            { "orientation", "uint8_t", "WebCore::ScreenOrientationLockType", false },
        };
    case MessageName::WebScreenOrientationManagerProxy_Unlock:
        return Vector<ArgumentDescription> { };
    case MessageName::WebScreenOrientationManagerProxy_SetShouldSendChangeNotification:
        return Vector<ArgumentDescription> {
            { "shouldSend", "bool", nullptr, false },
        };
    case MessageName::WebAutomationSession_DidEvaluateJavaScriptFunction:
        return Vector<ArgumentDescription> {
            { "callbackID", "uint64_t", nullptr, false },
            { "result", "String", nullptr, false },
            { "errorType", "String", nullptr, false },
        };
    case MessageName::WebAutomationSession_DidTakeScreenshot:
        return Vector<ArgumentDescription> {
            { "callbackID", "uint64_t", nullptr, false },
            { "imageDataHandle", "WebKit::ShareableBitmapHandle", nullptr, false },
            { "errorType", "String", nullptr, false },
        };
    case MessageName::DownloadProxy_DidStart:
        return Vector<ArgumentDescription> {
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "suggestedFilename", "String", nullptr, false },
        };
    case MessageName::DownloadProxy_DidReceiveAuthenticationChallenge:
        return Vector<ArgumentDescription> {
            { "challenge", "WebCore::AuthenticationChallenge", nullptr, false },
            { "challengeID", "WebKit::AuthenticationChallengeIdentifier", nullptr, false },
        };
    case MessageName::DownloadProxy_WillSendRequest:
        return Vector<ArgumentDescription> {
            { "redirectRequest", "WebCore::ResourceRequest", nullptr, false },
            { "redirectResponse", "WebCore::ResourceResponse", nullptr, false },
        };
    case MessageName::DownloadProxy_DecideDestinationWithSuggestedFilename:
        return Vector<ArgumentDescription> {
            { "response", "WebCore::ResourceResponse", nullptr, false },
            { "suggestedFilename", "String", nullptr, false },
        };
    case MessageName::DownloadProxy_DidReceiveData:
        return Vector<ArgumentDescription> {
            { "bytesWritten", "uint64_t", nullptr, false },
            { "totalBytesWritten", "uint64_t", nullptr, false },
            { "totalBytesExpectedToWrite", "uint64_t", nullptr, false },
        };
    case MessageName::DownloadProxy_DidCreateDestination:
        return Vector<ArgumentDescription> {
            { "path", "String", nullptr, false },
        };
    case MessageName::DownloadProxy_DidFinish:
        return Vector<ArgumentDescription> { };
    case MessageName::DownloadProxy_DidFail:
        return Vector<ArgumentDescription> {
            { "error", "WebCore::ResourceError", nullptr, false },
            { "resumeData", "IPC::DataReference", nullptr, false },
        };
#if ENABLE(GPU_PROCESS)
#if HAVE(VISIBILITY_PROPAGATION_VIEW)
    case MessageName::GPUProcessProxy_DidCreateContextForVisibilityPropagation:
        return Vector<ArgumentDescription> {
            { "pageProxyID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "contextID", "WebKit::LayerHostingContextID", nullptr, false },
        };
#endif
#if ENABLE(VP9)
    case MessageName::GPUProcessProxy_SetHasVP9HardwareDecoder:
        return Vector<ArgumentDescription> {
            { "hasVP9HardwareDecoder", "bool", nullptr, false },
        };
    case MessageName::GPUProcessProxy_SetHasVP9ExtensionSupport:
        return Vector<ArgumentDescription> {
            { "hasVP9ExtensionSupport", "bool", nullptr, false },
        };
#endif
#if ENABLE(MEDIA_STREAM) && PLATFORM(IOS_FAMILY)
    case MessageName::GPUProcessProxy_StatusBarWasTapped:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::GPUProcessProxy_ProcessIsReadyToExit:
        return Vector<ArgumentDescription> { };
    case MessageName::GPUProcessProxy_TerminateWebProcess:
        return Vector<ArgumentDescription> {
            { "webProcessIdentifier", "WebCore::ProcessIdentifier", nullptr, false },
        };
#endif
    case MessageName::RemoteWebInspectorUIProxy_FrontendLoaded:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteWebInspectorUIProxy_FrontendDidClose:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteWebInspectorUIProxy_Reopen:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteWebInspectorUIProxy_ResetState:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteWebInspectorUIProxy_BringToFront:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteWebInspectorUIProxy_Save:
        return Vector<ArgumentDescription> {
            { "saveDatas", "Vector<WebCore::InspectorFrontendClient::SaveData>", nullptr, false },
            { "forceSaveAs", "bool", nullptr, false },
        };
    case MessageName::RemoteWebInspectorUIProxy_Load:
        return Vector<ArgumentDescription> {
            { "path", "String", nullptr, false },
        };
    case MessageName::RemoteWebInspectorUIProxy_PickColorFromScreen:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteWebInspectorUIProxy_SetForcedAppearance:
        return Vector<ArgumentDescription> {
            { "appearance", "WebCore::InspectorFrontendClient::Appearance", nullptr, false },
        };
    case MessageName::RemoteWebInspectorUIProxy_SetSheetRect:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::FloatRect", nullptr, false },
        };
    case MessageName::RemoteWebInspectorUIProxy_StartWindowDrag:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteWebInspectorUIProxy_OpenURLExternally:
        return Vector<ArgumentDescription> {
            { "url", "String", nullptr, false },
        };
    case MessageName::RemoteWebInspectorUIProxy_RevealFileExternally:
        return Vector<ArgumentDescription> {
            { "path", "String", nullptr, false },
        };
    case MessageName::RemoteWebInspectorUIProxy_ShowCertificate:
        return Vector<ArgumentDescription> {
            { "certificateInfo", "WebCore::CertificateInfo", nullptr, false },
        };
    case MessageName::RemoteWebInspectorUIProxy_SendMessageToBackend:
        return Vector<ArgumentDescription> {
            { "message", "String", nullptr, false },
        };
    case MessageName::RemoteWebInspectorUIProxy_SetInspectorPageDeveloperExtrasEnabled:
        return Vector<ArgumentDescription> {
            { "enabled", "bool", nullptr, false },
        };
    case MessageName::WebInspectorUIProxy_OpenLocalInspectorFrontend:
        return Vector<ArgumentDescription> {
            { "canAttach", "bool", nullptr, false },
            { "underTest", "bool", nullptr, false },
        };
    case MessageName::WebInspectorUIProxy_SetFrontendConnection:
        return Vector<ArgumentDescription> {
            { "connectionHandle", "IPC::Connection::Handle", nullptr, false },
        };
    case MessageName::WebInspectorUIProxy_SendMessageToBackend:
        return Vector<ArgumentDescription> {
            { "message", "String", nullptr, false },
        };
    case MessageName::WebInspectorUIProxy_FrontendLoaded:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUIProxy_DidClose:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUIProxy_BringToFront:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUIProxy_Reopen:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUIProxy_ResetState:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUIProxy_SetForcedAppearance:
        return Vector<ArgumentDescription> {
            { "appearance", "WebCore::InspectorFrontendClient::Appearance", nullptr, false },
        };
    case MessageName::WebInspectorUIProxy_OpenURLExternally:
        return Vector<ArgumentDescription> {
            { "url", "String", nullptr, false },
        };
    case MessageName::WebInspectorUIProxy_RevealFileExternally:
        return Vector<ArgumentDescription> {
            { "path", "String", nullptr, false },
        };
    case MessageName::WebInspectorUIProxy_InspectedURLChanged:
        return Vector<ArgumentDescription> {
            { "urlString", "String", nullptr, false },
        };
    case MessageName::WebInspectorUIProxy_ShowCertificate:
        return Vector<ArgumentDescription> {
            { "certificateInfo", "WebCore::CertificateInfo", nullptr, false },
        };
    case MessageName::WebInspectorUIProxy_ElementSelectionChanged:
        return Vector<ArgumentDescription> {
            { "active", "bool", nullptr, false },
        };
    case MessageName::WebInspectorUIProxy_TimelineRecordingChanged:
        return Vector<ArgumentDescription> {
            { "active", "bool", nullptr, false },
        };
    case MessageName::WebInspectorUIProxy_SetDeveloperPreferenceOverride:
        return Vector<ArgumentDescription> {
            { "developerPreference", "WebCore::InspectorClient::DeveloperPreference", nullptr, false },
            { "overrideValue", "bool", nullptr, true },
        };
#if ENABLE(INSPECTOR_NETWORK_THROTTLING)
    case MessageName::WebInspectorUIProxy_SetEmulatedConditions:
        return Vector<ArgumentDescription> {
            { "bytesPerSecondLimit", "int64_t", nullptr, true },
        };
#endif
    case MessageName::WebInspectorUIProxy_Save:
        return Vector<ArgumentDescription> {
            { "saveDatas", "Vector<WebCore::InspectorFrontendClient::SaveData>", nullptr, false },
            { "forceSaveAs", "bool", nullptr, false },
        };
    case MessageName::WebInspectorUIProxy_Load:
        return Vector<ArgumentDescription> {
            { "path", "String", nullptr, false },
        };
    case MessageName::WebInspectorUIProxy_PickColorFromScreen:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUIProxy_AttachBottom:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUIProxy_AttachRight:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUIProxy_AttachLeft:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUIProxy_Detach:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUIProxy_AttachAvailabilityChanged:
        return Vector<ArgumentDescription> {
            { "available", "bool", nullptr, false },
        };
    case MessageName::WebInspectorUIProxy_SetAttachedWindowHeight:
        return Vector<ArgumentDescription> {
            { "height", "unsigned", nullptr, false },
        };
    case MessageName::WebInspectorUIProxy_SetAttachedWindowWidth:
        return Vector<ArgumentDescription> {
            { "width", "unsigned", nullptr, false },
        };
    case MessageName::WebInspectorUIProxy_SetSheetRect:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::FloatRect", nullptr, false },
        };
    case MessageName::WebInspectorUIProxy_StartWindowDrag:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUIProxy_SetInspectorPageDeveloperExtrasEnabled:
        return Vector<ArgumentDescription> {
            { "enabled", "bool", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_DidReceiveAuthenticationChallenge:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "pageID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "topOrigin", "WebCore::SecurityOriginData", nullptr, true },
            { "challenge", "WebCore::AuthenticationChallenge", nullptr, false },
            { "negotiatedLegacyTLS", "bool", nullptr, false },
            { "challengeID", "WebKit::AuthenticationChallengeIdentifier", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_NegotiatedLegacyTLS:
        return Vector<ArgumentDescription> {
            { "pageID", "WebKit::WebPageProxyIdentifier", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_DidNegotiateModernTLS:
        return Vector<ArgumentDescription> {
            { "pageID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "url", "URL", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_DidFailLoadDueToNetworkConnectionIntegrity:
        return Vector<ArgumentDescription> {
            { "pageID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "url", "URL", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_TestProcessIncomingSyncMessagesWhenWaitingForSyncReply:
        return Vector<ArgumentDescription> {
            { "pageID", "WebKit::WebPageProxyIdentifier", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_TerminateUnresponsiveServiceWorkerProcesses:
        return Vector<ArgumentDescription> {
            { "processIdentifier", "WebCore::ProcessIdentifier", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_LogDiagnosticMessage:
        return Vector<ArgumentDescription> {
            { "pageID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "message", "String", nullptr, false },
            { "description", "String", nullptr, false },
            { "shouldSample", "bool", "WebCore::ShouldSample", false },
        };
    case MessageName::NetworkProcessProxy_LogDiagnosticMessageWithResult:
        return Vector<ArgumentDescription> {
            { "pageID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "message", "String", nullptr, false },
            { "description", "String", nullptr, false },
            { "result", "uint32_t", nullptr, false },
            { "shouldSample", "bool", "WebCore::ShouldSample", false },
        };
    case MessageName::NetworkProcessProxy_LogDiagnosticMessageWithValue:
        return Vector<ArgumentDescription> {
            { "pageID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "message", "String", nullptr, false },
            { "description", "String", nullptr, false },
            { "value", "double", nullptr, false },
            { "significantFigures", "unsigned", nullptr, false },
            { "shouldSample", "bool", "WebCore::ShouldSample", false },
        };
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::NetworkProcessProxy_LogTestingEvent:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "event", "String", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_NotifyResourceLoadStatisticsProcessed:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcessProxy_NotifyWebsiteDataDeletionForRegistrableDomainsFinished:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcessProxy_NotifyWebsiteDataScanForRegistrableDomainsFinished:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcessProxy_RequestStorageAccessConfirm:
        return Vector<ArgumentDescription> {
            { "pageID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "subFrameDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "topFrameDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_DeleteWebsiteDataInUIProcessForRegistrableDomains:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "dataTypes", "OptionSet<WebKit::WebsiteDataType>", nullptr, false },
            { "fetchOptions", "OptionSet<WebKit::WebsiteDataFetchOption>", nullptr, false },
            { "domains", "Vector<WebCore::RegistrableDomain>", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_DidCommitCrossSiteLoadWithDataTransferFromPrevalentResource:
        return Vector<ArgumentDescription> {
            { "pageID", "WebKit::WebPageProxyIdentifier", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_SetDomainsWithUserInteraction:
        return Vector<ArgumentDescription> {
            { "domains", "HashSet<WebCore::RegistrableDomain>", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_SetDomainsWithCrossPageStorageAccess:
        return Vector<ArgumentDescription> {
            { "domains", "HashMap<WebCore::RegistrableDomain, WebCore::RegistrableDomain>", nullptr, false },
        };
#endif
#if ENABLE(CONTENT_EXTENSIONS)
    case MessageName::NetworkProcessProxy_ContentExtensionRules:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::UserContentControllerIdentifier", nullptr, false },
        };
#endif
    case MessageName::NetworkProcessProxy_TerminateWebProcess:
        return Vector<ArgumentDescription> {
            { "webProcessIdentifier", "WebCore::ProcessIdentifier", nullptr, false },
        };
#if ENABLE(SERVICE_WORKER)
    case MessageName::NetworkProcessProxy_StartServiceWorkerBackgroundProcessing:
        return Vector<ArgumentDescription> {
            { "serviceWorkerProcessIdentifier", "WebCore::ProcessIdentifier", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_EndServiceWorkerBackgroundProcessing:
        return Vector<ArgumentDescription> {
            { "serviceWorkerProcessIdentifier", "WebCore::ProcessIdentifier", nullptr, false },
        };
#endif
    case MessageName::NetworkProcessProxy_EstablishRemoteWorkerContextConnectionToNetworkProcess:
        return Vector<ArgumentDescription> {
            { "workerType", "bool", "WebKit::RemoteWorkerType", false },
            { "registrableDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "requestingProcessIdentifier", "WebCore::ProcessIdentifier", nullptr, true },
            { "serviceWorkerPageIdentifier", "WebCore::ScriptExecutionContextIdentifier", nullptr, true },
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_RemoteWorkerContextConnectionNoLongerNeeded:
        return Vector<ArgumentDescription> {
            { "workerType", "bool", "WebKit::RemoteWorkerType", false },
            { "identifier", "WebCore::ProcessIdentifier", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_RegisterRemoteWorkerClientProcess:
        return Vector<ArgumentDescription> {
            { "workerType", "bool", "WebKit::RemoteWorkerType", false },
            { "clientProcessIdentifier", "WebCore::ProcessIdentifier", nullptr, false },
            { "remoteWorkerProcessIdentifier", "WebCore::ProcessIdentifier", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_UnregisterRemoteWorkerClientProcess:
        return Vector<ArgumentDescription> {
            { "workerType", "bool", "WebKit::RemoteWorkerType", false },
            { "clientProcessIdentifier", "WebCore::ProcessIdentifier", nullptr, false },
            { "remoteWorkerProcessIdentifier", "WebCore::ProcessIdentifier", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_SetWebProcessHasUploads:
        return Vector<ArgumentDescription> {
            { "processID", "WebCore::ProcessIdentifier", nullptr, false },
            { "hasUpload", "bool", nullptr, false },
        };
#if ENABLE(APP_BOUND_DOMAINS)
    case MessageName::NetworkProcessProxy_GetAppBoundDomains:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
#endif
    case MessageName::NetworkProcessProxy_RequestStorageSpace:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "origin", "WebCore::ClientOrigin", nullptr, false },
            { "quota", "uint64_t", nullptr, false },
            { "currentSize", "uint64_t", nullptr, false },
            { "spaceRequired", "uint64_t", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_IncreaseQuota:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "origin", "WebCore::ClientOrigin", nullptr, false },
            { "identifier", "WebKit::QuotaIncreaseRequestIdentifier", nullptr, false },
            { "quota", "uint64_t", nullptr, false },
            { "currentSize", "uint64_t", nullptr, false },
            { "spaceRequested", "uint64_t", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_ResourceLoadDidSendRequest:
        return Vector<ArgumentDescription> {
            { "pageIdentifier", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "resourceLoadInfo", "WebKit::ResourceLoadInfo", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "httpBody", "IPC::FormDataReference", nullptr, true },
        };
    case MessageName::NetworkProcessProxy_ResourceLoadDidPerformHTTPRedirection:
        return Vector<ArgumentDescription> {
            { "pageIdentifier", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "resourceLoadInfo", "WebKit::ResourceLoadInfo", nullptr, false },
            { "resourceResponse", "WebCore::ResourceResponse", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_ResourceLoadDidReceiveChallenge:
        return Vector<ArgumentDescription> {
            { "pageIdentifier", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "resourceLoadInfo", "WebKit::ResourceLoadInfo", nullptr, false },
            { "challenge", "WebCore::AuthenticationChallenge", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_ResourceLoadDidReceiveResponse:
        return Vector<ArgumentDescription> {
            { "pageIdentifier", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "resourceLoadInfo", "WebKit::ResourceLoadInfo", nullptr, false },
            { "response", "WebCore::ResourceResponse", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_ResourceLoadDidCompleteWithError:
        return Vector<ArgumentDescription> {
            { "pageIdentifier", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "resourceLoadInfo", "WebKit::ResourceLoadInfo", nullptr, false },
            { "response", "WebCore::ResourceResponse", nullptr, false },
            { "error", "WebCore::ResourceError", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_TriggerBrowsingContextGroupSwitchForNavigation:
        return Vector<ArgumentDescription> {
            { "pageIdentifier", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "navigationID", "uint64_t", nullptr, false },
            { "browsingContextGroupSwitchDecision", "uint8_t", "WebCore::BrowsingContextGroupSwitchDecision", false },
            { "responseDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "existingNetworkResourceLoadIdentifierToResume", "WebKit::NetworkResourceLoadIdentifier", nullptr, false },
        };
#if USE(SOUP)
    case MessageName::NetworkProcessProxy_DidExceedMemoryLimit:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(APPLE_PAY_REMOTE_UI_USES_SCENE)
    case MessageName::NetworkProcessProxy_GetWindowSceneIdentifierForPaymentPresentation:
        return Vector<ArgumentDescription> {
            { "webPageProxyIdentifier", "WebKit::WebPageProxyIdentifier", nullptr, false },
        };
#endif
    case MessageName::NetworkProcessProxy_DataTaskReceivedChallenge:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::DataTaskIdentifier", nullptr, false },
            { "challenge", "WebCore::AuthenticationChallenge", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_DataTaskWillPerformHTTPRedirection:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::DataTaskIdentifier", nullptr, false },
            { "response", "WebCore::ResourceResponse", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_DataTaskDidReceiveResponse:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::DataTaskIdentifier", nullptr, false },
            { "response", "WebCore::ResourceResponse", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_DataTaskDidReceiveData:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::DataTaskIdentifier", nullptr, false },
            { "data", "IPC::DataReference", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_DataTaskDidCompleteWithError:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::DataTaskIdentifier", nullptr, false },
            { "error", "WebCore::ResourceError", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_OpenWindowFromServiceWorker:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "urlString", "String", nullptr, false },
            { "serviceWorkerOrigin", "WebCore::SecurityOriginData", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_NavigateServiceWorkerClient:
        return Vector<ArgumentDescription> {
            { "frameIdentifier", "WebCore::FrameIdentifier", nullptr, false },
            { "documentIdentifier", "WebCore::ScriptExecutionContextIdentifier", nullptr, false },
            { "url", "URL", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_CookiesDidChange:
        return Vector<ArgumentDescription> {
            { "sessionID", "PAL::SessionID", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_DeleteWebsiteDataInWebProcessesForOrigin:
        return Vector<ArgumentDescription> {
            { "websiteDataTypes", "OptionSet<WebKit::WebsiteDataType>", nullptr, false },
            { "origin", "WebCore::ClientOrigin", nullptr, false },
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "webPageProxyID", "WebKit::WebPageProxyIdentifier", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_ReloadExecutionContextsForOrigin:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::ClientOrigin", nullptr, false },
            { "sessionID", "PAL::SessionID", nullptr, false },
            { "triggeringFrame", "WebCore::FrameIdentifier", nullptr, true },
        };
#if ENABLE(NETWORK_ISSUE_REPORTING)
    case MessageName::NetworkProcessProxy_ReportNetworkIssue:
        return Vector<ArgumentDescription> {
            { "webPageProxyIdentifier", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "requestURL", "URL", nullptr, false },
        };
#endif
    case MessageName::WebUserContentControllerProxy_DidPostMessage:
        return Vector<ArgumentDescription> {
            { "pageID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "frameInfoData", "WebKit::FrameInfoData", nullptr, false },
            { "messageHandlerID", "uint64_t", nullptr, false },
            { "message", "IPC::DataReference", nullptr, false },
        };
#if ENABLE(WEBXR)
    case MessageName::PlatformXRSystem_EnumerateImmersiveXRDevices:
        return Vector<ArgumentDescription> { };
    case MessageName::PlatformXRSystem_RequestPermissionOnSessionFeatures:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::SecurityOriginData", nullptr, false },
            { "mode", "PlatformXR::SessionMode", nullptr, false },
            { "granted", "Vector<PlatformXR::SessionFeature>", nullptr, false },
            { "consentRequired", "Vector<PlatformXR::SessionFeature>", nullptr, false },
            { "consentOptional", "Vector<PlatformXR::SessionFeature>", nullptr, false },
        };
    case MessageName::PlatformXRSystem_InitializeTrackingAndRendering:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::SecurityOriginData", nullptr, false },
            { "mode", "PlatformXR::SessionMode", nullptr, false },
            { "requestedFeatures", "Vector<PlatformXR::SessionFeature>", nullptr, false },
        };
    case MessageName::PlatformXRSystem_ShutDownTrackingAndRendering:
        return Vector<ArgumentDescription> { };
    case MessageName::PlatformXRSystem_RequestFrame:
        return Vector<ArgumentDescription> { };
    case MessageName::PlatformXRSystem_SubmitFrame:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebProcess_InitializeWebProcess:
        return Vector<ArgumentDescription> {
            { "processCreationParameters", "WebKit::WebProcessCreationParameters", nullptr, false },
        };
    case MessageName::WebProcess_SetWebsiteDataStoreParameters:
        return Vector<ArgumentDescription> {
            { "parameters", "WebKit::WebProcessDataStoreParameters", nullptr, false },
        };
    case MessageName::WebProcess_CreateWebPage:
        return Vector<ArgumentDescription> {
            { "newPageID", "WebCore::PageIdentifier", nullptr, false },
            { "pageCreationParameters", "WebKit::WebPageCreationParameters", nullptr, false },
        };
    case MessageName::WebProcess_PrewarmWithDomainInformation:
        return Vector<ArgumentDescription> {
            { "prewarmInformation", "WebCore::PrewarmInformation", nullptr, false },
        };
    case MessageName::WebProcess_SetCacheModel:
        return Vector<ArgumentDescription> {
            { "cacheModel", "uint8_t", "WebKit::CacheModel", false },
        };
    case MessageName::WebProcess_RegisterURLSchemeAsEmptyDocument:
        return Vector<ArgumentDescription> {
            { "scheme", "String", nullptr, false },
        };
    case MessageName::WebProcess_RegisterURLSchemeAsSecure:
        return Vector<ArgumentDescription> {
            { "scheme", "String", nullptr, false },
        };
    case MessageName::WebProcess_RegisterURLSchemeAsBypassingContentSecurityPolicy:
        return Vector<ArgumentDescription> {
            { "scheme", "String", nullptr, false },
        };
    case MessageName::WebProcess_SetDomainRelaxationForbiddenForURLScheme:
        return Vector<ArgumentDescription> {
            { "scheme", "String", nullptr, false },
        };
    case MessageName::WebProcess_RegisterURLSchemeAsLocal:
        return Vector<ArgumentDescription> {
            { "scheme", "String", nullptr, false },
        };
    case MessageName::WebProcess_RegisterURLSchemeAsNoAccess:
        return Vector<ArgumentDescription> {
            { "scheme", "String", nullptr, false },
        };
    case MessageName::WebProcess_RegisterURLSchemeAsDisplayIsolated:
        return Vector<ArgumentDescription> {
            { "scheme", "String", nullptr, false },
        };
    case MessageName::WebProcess_RegisterURLSchemeAsCORSEnabled:
        return Vector<ArgumentDescription> {
            { "scheme", "String", nullptr, false },
        };
    case MessageName::WebProcess_RegisterURLSchemeAsCachePartitioned:
        return Vector<ArgumentDescription> {
            { "scheme", "String", nullptr, false },
        };
    case MessageName::WebProcess_RegisterURLSchemeAsCanDisplayOnlyIfCanRequest:
        return Vector<ArgumentDescription> {
            { "scheme", "String", nullptr, false },
        };
    case MessageName::WebProcess_SetDefaultRequestTimeoutInterval:
        return Vector<ArgumentDescription> {
            { "timeoutInterval", "double", nullptr, false },
        };
    case MessageName::WebProcess_SetAlwaysUsesComplexTextCodePath:
        return Vector<ArgumentDescription> {
            { "alwaysUseComplexText", "bool", nullptr, false },
        };
    case MessageName::WebProcess_SetDisableFontSubpixelAntialiasingForTesting:
        return Vector<ArgumentDescription> {
            { "disable", "bool", nullptr, false },
        };
    case MessageName::WebProcess_SetTrackingPreventionEnabled:
        return Vector<ArgumentDescription> {
            { "enabled", "bool", nullptr, false },
        };
    case MessageName::WebProcess_ClearResourceLoadStatistics:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_UserPreferredLanguagesChanged:
        return Vector<ArgumentDescription> {
            { "languages", "Vector<String>", nullptr, false },
        };
    case MessageName::WebProcess_FullKeyboardAccessModeChanged:
        return Vector<ArgumentDescription> {
            { "fullKeyboardAccessEnabled", "bool", nullptr, false },
        };
#if HAVE(MOUSE_DEVICE_OBSERVATION)
    case MessageName::WebProcess_SetHasMouseDevice:
        return Vector<ArgumentDescription> {
            { "hasMouseDevice", "bool", nullptr, false },
        };
#endif
#if HAVE(STYLUS_DEVICE_OBSERVATION)
    case MessageName::WebProcess_SetHasStylusDevice:
        return Vector<ArgumentDescription> {
            { "hasStylusDevice", "bool", nullptr, false },
        };
#endif
    case MessageName::WebProcess_StartMemorySampler:
        return Vector<ArgumentDescription> {
            { "sampleLogFileHandle", "WebKit::SandboxExtension::Handle", nullptr, false },
            { "sampleLogFilePath", "String", nullptr, false },
            { "interval", "double", nullptr, false },
        };
    case MessageName::WebProcess_StopMemorySampler:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_SetTextCheckerState:
        return Vector<ArgumentDescription> {
            { "textCheckerState", "WebKit::TextCheckerState", nullptr, false },
        };
    case MessageName::WebProcess_SetEnhancedAccessibility:
        return Vector<ArgumentDescription> {
            { "flag", "bool", nullptr, false },
        };
    case MessageName::WebProcess_GarbageCollectJavaScriptObjects:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_SetJavaScriptGarbageCollectorTimerEnabled:
        return Vector<ArgumentDescription> {
            { "enable", "bool", nullptr, false },
        };
    case MessageName::WebProcess_SetInjectedBundleParameter:
        return Vector<ArgumentDescription> {
            { "parameter", "String", nullptr, false },
            { "value", "IPC::DataReference", nullptr, false },
        };
    case MessageName::WebProcess_SetInjectedBundleParameters:
        return Vector<ArgumentDescription> {
            { "parameters", "IPC::DataReference", nullptr, false },
        };
    case MessageName::WebProcess_HandleInjectedBundleMessage:
        return Vector<ArgumentDescription> {
            { "messageName", "String", nullptr, false },
            { "messageBody", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebProcess_FetchWebsiteData:
        return Vector<ArgumentDescription> {
            { "websiteDataTypes", "OptionSet<WebKit::WebsiteDataType>", nullptr, false },
        };
    case MessageName::WebProcess_DeleteWebsiteData:
        return Vector<ArgumentDescription> {
            { "websiteDataTypes", "OptionSet<WebKit::WebsiteDataType>", nullptr, false },
            { "modifiedSince", "WallTime", nullptr, false },
        };
    case MessageName::WebProcess_DeleteWebsiteDataForOrigins:
        return Vector<ArgumentDescription> {
            { "websiteDataTypes", "OptionSet<WebKit::WebsiteDataType>", nullptr, false },
            { "origins", "Vector<WebCore::SecurityOriginData>", nullptr, false },
        };
    case MessageName::WebProcess_DeleteWebsiteDataForOrigin:
        return Vector<ArgumentDescription> {
            { "websiteDataTypes", "OptionSet<WebKit::WebsiteDataType>", nullptr, false },
            { "origin", "WebCore::ClientOrigin", nullptr, false },
        };
    case MessageName::WebProcess_ReloadExecutionContextsForOrigin:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::ClientOrigin", nullptr, false },
            { "triggeringFrame", "WebCore::FrameIdentifier", nullptr, true },
        };
    case MessageName::WebProcess_DeleteAllCookies:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_SetHiddenPageDOMTimerThrottlingIncreaseLimit:
        return Vector<ArgumentDescription> {
            { "milliseconds", "int", nullptr, false },
        };
    case MessageName::WebProcess_SetMemoryCacheDisabled:
        return Vector<ArgumentDescription> {
            { "disabled", "bool", nullptr, false },
        };
#if ENABLE(SERVICE_CONTROLS)
    case MessageName::WebProcess_SetEnabledServices:
        return Vector<ArgumentDescription> {
            { "hasImageServices", "bool", nullptr, false },
            { "hasSelectionServices", "bool", nullptr, false },
            { "hasRichContentServices", "bool", nullptr, false },
        };
#endif
    case MessageName::WebProcess_EnsureAutomationSessionProxy:
        return Vector<ArgumentDescription> {
            { "sessionIdentifier", "String", nullptr, false },
        };
    case MessageName::WebProcess_DestroyAutomationSessionProxy:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_PrepareToSuspend:
        return Vector<ArgumentDescription> {
            { "isSuspensionImminent", "bool", nullptr, false },
            { "estimatedSuspendTime", "MonotonicTime", nullptr, false },
        };
    case MessageName::WebProcess_ProcessDidResume:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_BackgroundResponsivenessPing:
        return Vector<ArgumentDescription> { };
#if ENABLE(GAMEPAD)
    case MessageName::WebProcess_SetInitialGamepads:
        return Vector<ArgumentDescription> {
            { "gamepadDatas", "Vector<WebKit::GamepadData>", nullptr, false },
        };
    case MessageName::WebProcess_GamepadConnected:
        return Vector<ArgumentDescription> {
            { "gamepadData", "WebKit::GamepadData", nullptr, false },
            { "eventVisibility", "bool", "WebCore::EventMakesGamepadsVisible", false },
        };
    case MessageName::WebProcess_GamepadDisconnected:
        return Vector<ArgumentDescription> {
            { "index", "unsigned", nullptr, false },
        };
#endif
    case MessageName::WebProcess_EstablishRemoteWorkerContextConnectionToNetworkProcess:
        return Vector<ArgumentDescription> {
            { "workerType", "bool", "WebKit::RemoteWorkerType", false },
            { "pageGroupID", "WebKit::PageGroupIdentifier", nullptr, false },
            { "webPageProxyID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "store", "WebKit::WebPreferencesStore", nullptr, false },
            { "domain", "WebCore::RegistrableDomain", nullptr, false },
            { "serviceWorkerPageIdentifier", "WebCore::ScriptExecutionContextIdentifier", nullptr, true },
            { "initializationData", "WebKit::RemoteWorkerInitializationData", nullptr, false },
        };
    case MessageName::WebProcess_SetHasSuspendedPageProxy:
        return Vector<ArgumentDescription> {
            { "hasSuspendedPageProxy", "bool", nullptr, false },
        };
    case MessageName::WebProcess_SetIsInProcessCache:
        return Vector<ArgumentDescription> {
            { "isInProcessCache", "bool", nullptr, false },
        };
    case MessageName::WebProcess_MarkIsNoLongerPrewarmed:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_GetActivePagesOriginsForTesting:
        return Vector<ArgumentDescription> { };
#if PLATFORM(COCOA)
    case MessageName::WebProcess_SetScreenProperties:
        return Vector<ArgumentDescription> {
            { "screenProperties", "WebCore::ScreenProperties", nullptr, false },
        };
#endif
#if PLATFORM(MAC)
    case MessageName::WebProcess_ScrollerStylePreferenceChanged:
        return Vector<ArgumentDescription> {
            { "useOvelayScrollbars", "bool", nullptr, false },
        };
    case MessageName::WebProcess_DisplayConfigurationChanged:
        return Vector<ArgumentDescription> {
            { "displayID", "CGDirectDisplayID", nullptr, false },
            { "flags", "CGDisplayChangeSummaryFlags", nullptr, false },
        };
#endif
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebProcess_UserInterfaceIdiomDidChange:
        return Vector<ArgumentDescription> {
            { "isPhoneOrWatch", "bool", nullptr, false },
        };
#endif
#if PLATFORM(IOS_FAMILY) && !PLATFORM(MACCATALYST)
    case MessageName::WebProcess_BacklightLevelDidChange:
        return Vector<ArgumentDescription> {
            { "backlightLevel", "float", nullptr, false },
        };
#endif
#if PLATFORM(MAC) || PLATFORM(MACCATALYST)
    case MessageName::WebProcess_ColorPreferencesDidChange:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebProcess_IsJITEnabled:
        return Vector<ArgumentDescription> { };
#if PLATFORM(COCOA)
    case MessageName::WebProcess_SetMediaMIMETypes:
        return Vector<ArgumentDescription> {
            { "types", "Vector<String>", nullptr, false },
        };
#endif
#if (PLATFORM(COCOA) && ENABLE(REMOTE_INSPECTOR))
    case MessageName::WebProcess_EnableRemoteWebInspector:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(MEDIA_STREAM)
    case MessageName::WebProcess_AddMockMediaDevice:
        return Vector<ArgumentDescription> {
            { "device", "WebCore::MockMediaDevice", nullptr, false },
        };
    case MessageName::WebProcess_ClearMockMediaDevices:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_RemoveMockMediaDevice:
        return Vector<ArgumentDescription> {
            { "persistentId", "String", nullptr, false },
        };
    case MessageName::WebProcess_SetMockMediaDeviceIsEphemeral:
        return Vector<ArgumentDescription> {
            { "persistentId", "String", nullptr, false },
            { "isEphemeral", "bool", nullptr, false },
        };
    case MessageName::WebProcess_ResetMockMediaDevices:
        return Vector<ArgumentDescription> { };
#endif
#if (ENABLE(MEDIA_STREAM) && ENABLE(SANDBOX_EXTENSIONS))
    case MessageName::WebProcess_GrantUserMediaDeviceSandboxExtensions:
        return Vector<ArgumentDescription> {
            { "sandboxExtensions", "WebKit::MediaDeviceSandboxExtensions", nullptr, false },
        };
    case MessageName::WebProcess_RevokeUserMediaDeviceSandboxExtensions:
        return Vector<ArgumentDescription> {
            { "sandboxExtensionIDs", "Vector<String>", nullptr, false },
        };
#endif
    case MessageName::WebProcess_ClearCurrentModifierStateForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_SetBackForwardCacheCapacity:
        return Vector<ArgumentDescription> {
            { "capacity", "unsigned", nullptr, false },
        };
    case MessageName::WebProcess_ClearCachedPage:
        return Vector<ArgumentDescription> {
            { "backForwardItemID", "WebCore::BackForwardItemIdentifier", nullptr, false },
        };
#if PLATFORM(GTK) || PLATFORM(WPE)
    case MessageName::WebProcess_SendMessageToWebProcessExtension:
        return Vector<ArgumentDescription> {
            { "userMessage", "WebKit::UserMessage", nullptr, false },
        };
#endif
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::WebProcess_SeedResourceLoadStatisticsForTesting:
        return Vector<ArgumentDescription> {
            { "firstPartyDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "thirdPartyDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "shouldScheduleNotification", "bool", nullptr, false },
        };
    case MessageName::WebProcess_SetThirdPartyCookieBlockingMode:
        return Vector<ArgumentDescription> {
            { "blockingMode", "uint8_t", "WebCore::ThirdPartyCookieBlockingMode", false },
        };
    case MessageName::WebProcess_SetDomainsWithUserInteraction:
        return Vector<ArgumentDescription> {
            { "domains", "HashSet<WebCore::RegistrableDomain>", nullptr, false },
        };
    case MessageName::WebProcess_SetDomainsWithCrossPageStorageAccess:
        return Vector<ArgumentDescription> {
            { "domains", "HashMap<WebCore::RegistrableDomain, WebCore::RegistrableDomain>", nullptr, false },
        };
    case MessageName::WebProcess_SendResourceLoadStatisticsDataImmediately:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebProcess_GrantAccessToAssetServices:
        return Vector<ArgumentDescription> {
            { "assetServicesHandles", "Vector<WebKit::SandboxExtension::Handle>", nullptr, false },
        };
    case MessageName::WebProcess_RevokeAccessToAssetServices:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_SwitchFromStaticFontRegistryToUserFontRegistry:
        return Vector<ArgumentDescription> {
            { "fontMachExtensionHandles", "Vector<WebKit::SandboxExtension::Handle>", nullptr, false },
        };
#if PLATFORM(COCOA)
    case MessageName::WebProcess_DisableURLSchemeCheckInDataDetectors:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_UnblockServicesRequiredByAccessibility:
        return Vector<ArgumentDescription> {
            { "handleArray", "Vector<WebKit::SandboxExtension::Handle>", nullptr, false },
        };
#endif
#if (PLATFORM(COCOA) && ENABLE(CFPREFS_DIRECT_MODE))
    case MessageName::WebProcess_NotifyPreferencesChanged:
        return Vector<ArgumentDescription> {
            { "domain", "String", nullptr, false },
            { "key", "String", nullptr, false },
            { "encodedValue", "String", nullptr, true },
        };
#endif
#if PLATFORM(COCOA)
    case MessageName::WebProcess_PowerSourceDidChange:
        return Vector<ArgumentDescription> {
            { "hasAC", "bool", nullptr, false },
        };
#endif
#if PLATFORM(GTK) && !USE(GTK4)
    case MessageName::WebProcess_SetUseSystemAppearanceForScrollbars:
        return Vector<ArgumentDescription> {
            { "useSystemAppearanceForScrollbars", "bool", nullptr, false },
        };
#endif
#if PLATFORM(COCOA)
    case MessageName::WebProcess_DidWriteToPasteboardAsynchronously:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
        };
#endif
#if HAVE(CVDISPLAYLINK)
    case MessageName::WebProcess_DisplayWasRefreshed:
        return Vector<ArgumentDescription> {
            { "displayID", "uint32_t", nullptr, false },
            { "update", "WebCore::DisplayUpdate", nullptr, false },
        };
#endif
#if PLATFORM(MAC)
    case MessageName::WebProcess_SystemWillPowerOn:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_SystemWillSleep:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_SystemDidWake:
        return Vector<ArgumentDescription> { };
#endif
#if HAVE(AUDIO_COMPONENT_SERVER_REGISTRATIONS)
    case MessageName::WebProcess_ConsumeAudioComponentRegistrations:
        return Vector<ArgumentDescription> {
            { "registrationData", "IPC::SharedBufferReference", nullptr, false },
        };
#endif
#if PLATFORM(COCOA)
    case MessageName::WebProcess_AccessibilityPreferencesDidChange:
        return Vector<ArgumentDescription> {
            { "accessibilityPreferences", "WebKit::AccessibilityPreferences", nullptr, false },
        };
#endif
#if HAVE(MEDIA_ACCESSIBILITY_FRAMEWORK)
    case MessageName::WebProcess_SetMediaAccessibilityPreferences:
        return Vector<ArgumentDescription> {
            { "displayMode", "WebCore::CaptionUserPreferences::CaptionDisplayMode", nullptr, false },
            { "languages", "Vector<String>", nullptr, false },
        };
#endif
#if PLATFORM(MAC)
    case MessageName::WebProcess_OpenDirectoryCacheInvalidated:
        return Vector<ArgumentDescription> {
            { "handle", "WebKit::SandboxExtension::Handle", nullptr, false },
            { "machBootstrapHandle", "WebKit::SandboxExtension::Handle", nullptr, false },
        };
#endif
    case MessageName::WebAutomationSessionProxy_EvaluateJavaScriptFunction:
        return Vector<ArgumentDescription> {
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
            { "function", "String", nullptr, false },
            { "arguments", "Vector<String>", nullptr, false },
            { "expectsImplicitCallbackArgument", "bool", nullptr, false },
            { "callbackTimeout", "double", nullptr, true },
            { "callbackID", "uint64_t", nullptr, false },
        };
    case MessageName::WebAutomationSessionProxy_ResolveChildFrameWithOrdinal:
        return Vector<ArgumentDescription> {
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
            { "ordinal", "uint32_t", nullptr, false },
        };
    case MessageName::WebAutomationSessionProxy_ResolveChildFrameWithNodeHandle:
        return Vector<ArgumentDescription> {
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
            { "nodeHandle", "String", nullptr, false },
        };
    case MessageName::WebAutomationSessionProxy_ResolveChildFrameWithName:
        return Vector<ArgumentDescription> {
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
            { "name", "String", nullptr, false },
        };
    case MessageName::WebAutomationSessionProxy_ResolveParentFrame:
        return Vector<ArgumentDescription> {
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
        };
    case MessageName::WebAutomationSessionProxy_ComputeElementLayout:
        return Vector<ArgumentDescription> {
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
            { "nodeHandle", "String", nullptr, false },
            { "scrollIntoViewIfNeeded", "bool", nullptr, false },
            { "coordinateSystem", "uint8_t", "WebKit::CoordinateSystem", false },
        };
    case MessageName::WebAutomationSessionProxy_GetComputedRole:
        return Vector<ArgumentDescription> {
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
            { "nodeHandle", "String", nullptr, false },
        };
    case MessageName::WebAutomationSessionProxy_GetComputedLabel:
        return Vector<ArgumentDescription> {
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
            { "nodeHandle", "String", nullptr, false },
        };
    case MessageName::WebAutomationSessionProxy_SelectOptionElement:
        return Vector<ArgumentDescription> {
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
            { "nodeHandle", "String", nullptr, false },
        };
    case MessageName::WebAutomationSessionProxy_SetFilesForInputFileUpload:
        return Vector<ArgumentDescription> {
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
            { "nodeHandle", "String", nullptr, false },
            { "filenames", "Vector<String>", nullptr, false },
        };
    case MessageName::WebAutomationSessionProxy_TakeScreenshot:
        return Vector<ArgumentDescription> {
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
            { "nodeHandle", "String", nullptr, false },
            { "scrollIntoViewIfNeeded", "bool", nullptr, false },
            { "clipToViewport", "bool", nullptr, false },
            { "callbackID", "uint64_t", nullptr, false },
        };
    case MessageName::WebAutomationSessionProxy_SnapshotRectForScreenshot:
        return Vector<ArgumentDescription> {
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
            { "nodeHandle", "String", nullptr, false },
            { "scrollIntoViewIfNeeded", "bool", nullptr, false },
            { "clipToViewport", "bool", nullptr, false },
        };
    case MessageName::WebAutomationSessionProxy_GetCookiesForFrame:
        return Vector<ArgumentDescription> {
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
        };
    case MessageName::WebAutomationSessionProxy_DeleteCookie:
        return Vector<ArgumentDescription> {
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
            { "cookieName", "String", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_DidDeleteDatabase:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::IDBResultData", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_DidOpenDatabase:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::IDBResultData", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_DidAbortTransaction:
        return Vector<ArgumentDescription> {
            { "transactionIdentifier", "WebCore::IDBResourceIdentifier", nullptr, false },
            { "error", "WebCore::IDBError", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_DidCommitTransaction:
        return Vector<ArgumentDescription> {
            { "transactionIdentifier", "WebCore::IDBResourceIdentifier", nullptr, false },
            { "error", "WebCore::IDBError", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_DidCreateObjectStore:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::IDBResultData", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_DidDeleteObjectStore:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::IDBResultData", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_DidRenameObjectStore:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::IDBResultData", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_DidClearObjectStore:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::IDBResultData", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_DidCreateIndex:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::IDBResultData", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_DidDeleteIndex:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::IDBResultData", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_DidRenameIndex:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::IDBResultData", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_DidPutOrAdd:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::IDBResultData", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_DidGetRecord:
        return Vector<ArgumentDescription> {
            { "result", "WebKit::WebIDBResult", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_DidGetAllRecords:
        return Vector<ArgumentDescription> {
            { "result", "WebKit::WebIDBResult", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_DidGetCount:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::IDBResultData", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_DidDeleteRecord:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::IDBResultData", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_DidOpenCursor:
        return Vector<ArgumentDescription> {
            { "result", "WebKit::WebIDBResult", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_DidIterateCursor:
        return Vector<ArgumentDescription> {
            { "result", "WebKit::WebIDBResult", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_FireVersionChangeEvent:
        return Vector<ArgumentDescription> {
            { "databaseConnectionIdentifier", "uint64_t", nullptr, false },
            { "requestIdentifier", "WebCore::IDBResourceIdentifier", nullptr, false },
            { "requestedVersion", "uint64_t", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_DidStartTransaction:
        return Vector<ArgumentDescription> {
            { "transactionIdentifier", "WebCore::IDBResourceIdentifier", nullptr, false },
            { "error", "WebCore::IDBError", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_DidCloseFromServer:
        return Vector<ArgumentDescription> {
            { "databaseConnectionIdentifier", "uint64_t", nullptr, false },
            { "error", "WebCore::IDBError", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_NotifyOpenDBRequestBlocked:
        return Vector<ArgumentDescription> {
            { "requestIdentifier", "WebCore::IDBResourceIdentifier", nullptr, false },
            { "oldVersion", "uint64_t", nullptr, false },
            { "newVersion", "uint64_t", nullptr, false },
        };
    case MessageName::WebIDBConnectionToServer_DidGetAllDatabaseNamesAndVersions:
        return Vector<ArgumentDescription> {
            { "requestIdentifier", "WebCore::IDBResourceIdentifier", nullptr, false },
            { "databases", "Vector<WebCore::IDBDatabaseNameAndVersion>", nullptr, false },
        };
#if ENABLE(FULLSCREEN_API)
    case MessageName::WebFullScreenManager_RequestRestoreFullScreen:
        return Vector<ArgumentDescription> { };
    case MessageName::WebFullScreenManager_RequestExitFullScreen:
        return Vector<ArgumentDescription> { };
    case MessageName::WebFullScreenManager_WillEnterFullScreen:
        return Vector<ArgumentDescription> { };
    case MessageName::WebFullScreenManager_DidEnterFullScreen:
        return Vector<ArgumentDescription> { };
    case MessageName::WebFullScreenManager_WillExitFullScreen:
        return Vector<ArgumentDescription> { };
    case MessageName::WebFullScreenManager_DidExitFullScreen:
        return Vector<ArgumentDescription> { };
    case MessageName::WebFullScreenManager_SetAnimatingFullScreen:
        return Vector<ArgumentDescription> {
            { "animating", "bool", nullptr, false },
        };
    case MessageName::WebFullScreenManager_SaveScrollPosition:
        return Vector<ArgumentDescription> { };
    case MessageName::WebFullScreenManager_RestoreScrollPosition:
        return Vector<ArgumentDescription> { };
    case MessageName::WebFullScreenManager_SetFullscreenInsets:
        return Vector<ArgumentDescription> {
            { "insets", "WebCore::RectEdges<float>", nullptr, false },
        };
    case MessageName::WebFullScreenManager_SetFullscreenAutoHideDuration:
        return Vector<ArgumentDescription> {
            { "duration", "Seconds", nullptr, false },
        };
    case MessageName::WebFullScreenManager_SetFullscreenControlsHidden:
        return Vector<ArgumentDescription> {
            { "hidden", "bool", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::GPUProcessConnection_DidInitialize:
        return Vector<ArgumentDescription> {
            { "info", "WebKit::GPUProcessConnectionInfo", nullptr, true },
        };
    case MessageName::GPUProcessConnection_DidReceiveRemoteCommand:
        return Vector<ArgumentDescription> {
            { "type", "uint8_t", "WebCore::PlatformMediaSession::RemoteControlCommandType", false },
            { "argument", "WebCore::PlatformMediaSession::RemoteCommandArgument", nullptr, false },
        };
#if ENABLE(ROUTING_ARBITRATION)
    case MessageName::GPUProcessConnection_BeginRoutingArbitrationWithCategory:
        return Vector<ArgumentDescription> {
            { "category", "uint8_t", "WebCore::AudioSession::CategoryType", false },
        };
    case MessageName::GPUProcessConnection_EndRoutingArbitration:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(MEDIA_STREAM) && PLATFORM(COCOA)
    case MessageName::GPUProcessConnection_ResetAudioMediaStreamTrackRendererInternalUnit:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::AudioMediaStreamTrackRendererInternalUnitIdentifier", nullptr, false },
        };
#endif
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(WEBGL)
    case MessageName::RemoteGraphicsContextGLProxy_WasCreated:
        return Vector<ArgumentDescription> {
            { "didSucceed", "bool", nullptr, false },
            { "streamWakeUpSemaphore", "IPC::Semaphore", nullptr, false },
            { "streamClientWaitSemaphore", "IPC::Semaphore", nullptr, false },
            { "availableExtensions", "String", nullptr, false },
            { "requestableExtensions", "String", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGLProxy_WasLost:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteGraphicsContextGLProxy_WasChanged:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteRenderingBackendProxy_DidCreateImageBufferBackend:
        return Vector<ArgumentDescription> {
            { "handle", "WebKit::ImageBufferBackendHandle", nullptr, false },
            { "renderingResourceIdentifier", "WebCore::RenderingResourceIdentifier", nullptr, false },
        };
    case MessageName::RemoteRenderingBackendProxy_DidFlush:
        return Vector<ArgumentDescription> {
            { "flushIdentifier", "WebKit::DisplayListRecorderFlushIdentifier", nullptr, false },
        };
    case MessageName::RemoteRenderingBackendProxy_DidFinalizeRenderingUpdate:
        return Vector<ArgumentDescription> {
            { "didRenderingUpdateID", "WebKit::RenderingUpdateID", nullptr, false },
        };
    case MessageName::RemoteRenderingBackendProxy_DidInitialize:
        return Vector<ArgumentDescription> {
            { "wakeUpSemaphore", "IPC::Semaphore", nullptr, false },
            { "clientWaitSemaphore", "IPC::Semaphore", nullptr, false },
        };
    case MessageName::RemoteRenderingBackendProxy_DidMarkLayersAsVolatile:
        return Vector<ArgumentDescription> {
            { "requestIdentifier", "WebKit::MarkSurfacesAsVolatileRequestIdentifier", nullptr, false },
            { "markedVolatileBufferIdentifiers", "Vector<WebCore::RenderingResourceIdentifier>", nullptr, false },
            { "didMarkAllLayersAsVolatile", "bool", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteGPUProxy_WasCreated:
        return Vector<ArgumentDescription> {
            { "didSucceed", "bool", nullptr, false },
            { "streamWakeUpSemaphore", "IPC::Semaphore", nullptr, false },
            { "streamClientWaitSemaphore", "IPC::Semaphore", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteAudioHardwareListener_AudioHardwareDidBecomeActive:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteAudioHardwareListener_AudioHardwareDidBecomeInactive:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteAudioHardwareListener_AudioOutputDeviceChanged:
        return Vector<ArgumentDescription> {
            { "bufferSizeMinimum", "size_t", nullptr, false },
            { "bufferSizeMaximum", "size_t", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
    case MessageName::MediaPlayerPrivateRemote_NetworkStateChanged:
        return Vector<ArgumentDescription> {
            { "state", "WebKit::RemoteMediaPlayerState", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_ReadyStateChanged:
        return Vector<ArgumentDescription> {
            { "state", "WebKit::RemoteMediaPlayerState", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_FirstVideoFrameAvailable:
        return Vector<ArgumentDescription> { };
    case MessageName::MediaPlayerPrivateRemote_VolumeChanged:
        return Vector<ArgumentDescription> {
            { "volume", "double", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_MuteChanged:
        return Vector<ArgumentDescription> {
            { "mute", "bool", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_TimeChanged:
        return Vector<ArgumentDescription> {
            { "state", "WebKit::RemoteMediaPlayerState", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_DurationChanged:
        return Vector<ArgumentDescription> {
            { "state", "WebKit::RemoteMediaPlayerState", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_RateChanged:
        return Vector<ArgumentDescription> {
            { "rate", "double", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_PlaybackStateChanged:
        return Vector<ArgumentDescription> {
            { "paused", "bool", nullptr, false },
            { "mediaTime", "MediaTime", nullptr, false },
            { "wallTime", "MonotonicTime", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_EngineFailedToLoad:
        return Vector<ArgumentDescription> {
            { "platformErrorCode", "int64_t", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_UpdateCachedState:
        return Vector<ArgumentDescription> {
            { "state", "WebKit::RemoteMediaPlayerState", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_CharacteristicChanged:
        return Vector<ArgumentDescription> {
            { "state", "WebKit::RemoteMediaPlayerState", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_SizeChanged:
        return Vector<ArgumentDescription> {
            { "naturalSize", "WebCore::FloatSize", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_RenderingModeChanged:
        return Vector<ArgumentDescription> { };
    case MessageName::MediaPlayerPrivateRemote_CurrentTimeChanged:
        return Vector<ArgumentDescription> {
            { "mediaTime", "MediaTime", nullptr, false },
            { "wallTime", "MonotonicTime", nullptr, false },
            { "timeIsProgressing", "bool", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_AddRemoteAudioTrack:
        return Vector<ArgumentDescription> {
            { "trackID", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
            { "configuration", "WebKit::AudioTrackPrivateRemoteConfiguration", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_RemoveRemoteAudioTrack:
        return Vector<ArgumentDescription> {
            { "trackID", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_RemoteAudioTrackConfigurationChanged:
        return Vector<ArgumentDescription> {
            { "trackID", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
            { "configuration", "WebKit::AudioTrackPrivateRemoteConfiguration", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_AddRemoteVideoTrack:
        return Vector<ArgumentDescription> {
            { "trackID", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
            { "configuration", "WebKit::VideoTrackPrivateRemoteConfiguration", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_RemoveRemoteVideoTrack:
        return Vector<ArgumentDescription> {
            { "trackID", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_RemoteVideoTrackConfigurationChanged:
        return Vector<ArgumentDescription> {
            { "trackID", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
            { "configuration", "WebKit::VideoTrackPrivateRemoteConfiguration", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_AddRemoteTextTrack:
        return Vector<ArgumentDescription> {
            { "trackID", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
            { "configuration", "WebKit::TextTrackPrivateRemoteConfiguration", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_RemoveRemoteTextTrack:
        return Vector<ArgumentDescription> {
            { "trackID", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_RemoteTextTrackConfigurationChanged:
        return Vector<ArgumentDescription> {
            { "trackID", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
            { "configuration", "WebKit::TextTrackPrivateRemoteConfiguration", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_ParseWebVTTFileHeader:
        return Vector<ArgumentDescription> {
            { "trackID", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
            { "header", "String", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_ParseWebVTTCueData:
        return Vector<ArgumentDescription> {
            { "trackID", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
            { "data", "IPC::DataReference", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_ParseWebVTTCueDataStruct:
        return Vector<ArgumentDescription> {
            { "trackID", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
            { "cueData", "WebCore::ISOWebVTTCue", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_AddDataCue:
        return Vector<ArgumentDescription> {
            { "trackID", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
            { "start", "MediaTime", nullptr, false },
            { "end", "MediaTime", nullptr, false },
            { "data", "IPC::DataReference", nullptr, false },
        };
#if ENABLE(DATACUE_VALUE)
    case MessageName::MediaPlayerPrivateRemote_AddDataCueWithType:
        return Vector<ArgumentDescription> {
            { "trackID", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
            { "start", "MediaTime", nullptr, false },
            { "end", "MediaTime", nullptr, false },
            { "cueData", "WebCore::SerializedPlatformDataCueValue", nullptr, false },
            { "type", "String", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_UpdateDataCue:
        return Vector<ArgumentDescription> {
            { "trackID", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
            { "start", "MediaTime", nullptr, false },
            { "end", "MediaTime", nullptr, false },
            { "cueData", "WebCore::SerializedPlatformDataCueValue", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_RemoveDataCue:
        return Vector<ArgumentDescription> {
            { "trackID", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
            { "start", "MediaTime", nullptr, false },
            { "end", "MediaTime", nullptr, false },
            { "cueData", "WebCore::SerializedPlatformDataCueValue", nullptr, false },
        };
#endif
    case MessageName::MediaPlayerPrivateRemote_AddGenericCue:
        return Vector<ArgumentDescription> {
            { "trackID", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
            { "cue", "WebCore::GenericCueData", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_UpdateGenericCue:
        return Vector<ArgumentDescription> {
            { "trackID", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
            { "cue", "WebCore::GenericCueData", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_RemoveGenericCue:
        return Vector<ArgumentDescription> {
            { "trackID", "WebKit::TrackPrivateRemoteIdentifier", nullptr, false },
            { "cue", "WebCore::GenericCueData", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_RequestResource:
        return Vector<ArgumentDescription> {
            { "remoteMediaResourceIdentifier", "WebKit::RemoteMediaResourceIdentifier", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "options", "uint8_t", "WebCore::PlatformMediaResourceLoader::LoadOptions", false },
        };
    case MessageName::MediaPlayerPrivateRemote_RemoveResource:
        return Vector<ArgumentDescription> {
            { "remoteMediaResourceIdentifier", "WebKit::RemoteMediaResourceIdentifier", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_SendH2Ping:
        return Vector<ArgumentDescription> {
            { "url", "URL", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_ResourceNotSupported:
        return Vector<ArgumentDescription> { };
    case MessageName::MediaPlayerPrivateRemote_ActiveSourceBuffersChanged:
        return Vector<ArgumentDescription> { };
#if ENABLE(ENCRYPTED_MEDIA)
    case MessageName::MediaPlayerPrivateRemote_WaitingForKeyChanged:
        return Vector<ArgumentDescription> {
            { "waitingForKey", "bool", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_InitializationDataEncountered:
        return Vector<ArgumentDescription> {
            { "initDataType", "String", nullptr, false },
            { "initData", "IPC::DataReference", nullptr, false },
        };
#endif
#if ENABLE(LEGACY_ENCRYPTED_MEDIA)
    case MessageName::MediaPlayerPrivateRemote_MediaPlayerKeyNeeded:
        return Vector<ArgumentDescription> {
            { "message", "IPC::DataReference", nullptr, false },
        };
#endif
#if ENABLE(WIRELESS_PLAYBACK_TARGET)
    case MessageName::MediaPlayerPrivateRemote_CurrentPlaybackTargetIsWirelessChanged:
        return Vector<ArgumentDescription> {
            { "isCurrentPlaybackTargetWireless", "bool", nullptr, false },
        };
#endif
#if PLATFORM(IOS_FAMILY)
    case MessageName::MediaPlayerPrivateRemote_GetRawCookies:
        return Vector<ArgumentDescription> {
            { "url", "URL", nullptr, false },
        };
#endif
#if PLATFORM(COCOA)
    case MessageName::MediaPlayerPrivateRemote_PushVideoFrameMetadata:
        return Vector<ArgumentDescription> {
            { "metadata", "WebCore::VideoFrameMetadata", nullptr, false },
            { "frameProperties", "WebKit::RemoteVideoFrameProxy::Properties", nullptr, false },
        };
    case MessageName::MediaPlayerPrivateRemote_LayerHostingContextIdChanged:
        return Vector<ArgumentDescription> {
            { "inlineLayerHostingContextId", "WebKit::LayerHostingContextID", nullptr, true },
            { "presentationSize", "WebCore::IntSize", nullptr, false },
        };
#endif
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
    case MessageName::MediaSourcePrivateRemote_SeekToTime:
        return Vector<ArgumentDescription> {
            { "time", "MediaTime", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && USE(AUDIO_SESSION)
    case MessageName::RemoteAudioSession_ConfigurationChanged:
        return Vector<ArgumentDescription> {
            { "configuration", "WebKit::RemoteAudioSessionConfiguration", nullptr, false },
        };
    case MessageName::RemoteAudioSession_BeginInterruption:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteAudioSession_EndInterruption:
        return Vector<ArgumentDescription> {
            { "flags", "WebCore::AudioSession::MayResume", nullptr, false },
        };
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS)
    case MessageName::RemoteAudioSourceProviderManager_AudioStorageChanged:
        return Vector<ArgumentDescription> {
            { "id", "WebCore::MediaPlayerIdentifier", nullptr, false },
            { "storageHandle", "WebKit::ConsumerSharedCARingBuffer::Handle", nullptr, false },
            { "description", "WebCore::CAAudioStreamDescription", nullptr, false },
        };
    case MessageName::RemoteAudioSourceProviderManager_AudioSamplesAvailable:
        return Vector<ArgumentDescription> {
            { "id", "WebCore::MediaPlayerIdentifier", nullptr, false },
            { "startFrame", "uint64_t", nullptr, false },
            { "numberOfFrames", "uint64_t", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteCDMInstance_UnrequestedInitializationDataReceived:
        return Vector<ArgumentDescription> {
            { "type", "String", nullptr, false },
            { "initData", "Ref<WebCore::SharedBuffer>", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteCDMInstanceSession_UpdateKeyStatuses:
        return Vector<ArgumentDescription> {
            { "keyStatuses", "WebCore::CDMInstanceSession::KeyStatusVector", nullptr, false },
        };
    case MessageName::RemoteCDMInstanceSession_SendMessage:
        return Vector<ArgumentDescription> {
            { "type", "uint8_t", "WebCore::CDMMessageType", false },
            { "message", "RefPtr<WebCore::SharedBuffer>", nullptr, false },
        };
    case MessageName::RemoteCDMInstanceSession_SessionIdChanged:
        return Vector<ArgumentDescription> {
            { "sessionId", "String", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
    case MessageName::RemoteLegacyCDMSession_SendMessage:
        return Vector<ArgumentDescription> {
            { "message", "RefPtr<WebCore::SharedBuffer>", nullptr, false },
            { "destinationURL", "String", nullptr, false },
        };
    case MessageName::RemoteLegacyCDMSession_SendError:
        return Vector<ArgumentDescription> {
            { "errorCode", "uint8_t", "WebCore::LegacyCDMSessionClient::MediaKeyErrorCode", false },
            { "systemCode", "uint32_t", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteRemoteCommandListener_DidReceiveRemoteControlCommand:
        return Vector<ArgumentDescription> {
            { "command", "uint8_t", "WebCore::PlatformMediaSession::RemoteControlCommandType", false },
            { "argument", "WebCore::PlatformMediaSession::RemoteCommandArgument", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateDidReceiveInitializationSegment:
        return Vector<ArgumentDescription> {
            { "segmentConfiguration", "WebKit::InitializationSegmentInfo", nullptr, false },
        };
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateStreamEndedWithDecodeError:
        return Vector<ArgumentDescription> { };
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateAppendError:
        return Vector<ArgumentDescription> {
            { "decodeError", "bool", nullptr, false },
        };
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateAppendComplete:
        return Vector<ArgumentDescription> {
            { "appendResult", "WebCore::SourceBufferPrivateClient::AppendResult", nullptr, false },
            { "buffered", "WebCore::PlatformTimeRanges", nullptr, false },
            { "totalTrackBufferSizeInBytes", "uint64_t", nullptr, false },
            { "timeStampOffset", "MediaTime", nullptr, false },
        };
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateHighestPresentationTimestampChanged:
        return Vector<ArgumentDescription> {
            { "timestamp", "MediaTime", nullptr, false },
        };
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateDurationChanged:
        return Vector<ArgumentDescription> {
            { "duration", "MediaTime", nullptr, false },
        };
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateDidParseSample:
        return Vector<ArgumentDescription> {
            { "sampleDuration", "double", nullptr, false },
        };
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateDidDropSample:
        return Vector<ArgumentDescription> { };
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateBufferedDirtyChanged:
        return Vector<ArgumentDescription> {
            { "dirty", "bool", nullptr, false },
        };
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateDidReceiveRenderingError:
        return Vector<ArgumentDescription> {
            { "errorCode", "int64_t", nullptr, false },
        };
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateReportExtraMemoryCost:
        return Vector<ArgumentDescription> {
            { "extraMemory", "uint64_t", nullptr, false },
        };
#endif
#if USE(LIBWEBRTC) && PLATFORM(COCOA) && ENABLE(GPU_PROCESS)
    case MessageName::LibWebRTCCodecs_FailedDecoding:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::VideoDecoderIdentifier", nullptr, false },
        };
    case MessageName::LibWebRTCCodecs_FlushDecoderCompleted:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::VideoDecoderIdentifier", nullptr, false },
        };
    case MessageName::LibWebRTCCodecs_CompletedDecoding:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::VideoDecoderIdentifier", nullptr, false },
            { "timeStamp", "int64_t", nullptr, false },
            { "timeStampNs", "int64_t", nullptr, false },
            { "frame", "WebKit::RemoteVideoFrameProxy::Properties", nullptr, false },
        };
    case MessageName::LibWebRTCCodecs_CompletedDecodingCV:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::VideoDecoderIdentifier", nullptr, false },
            { "timeStamp", "int64_t", nullptr, false },
            { "timeStampNs", "int64_t", nullptr, false },
            { "pixelBuffer", "RetainPtr<CVPixelBufferRef>", nullptr, false },
        };
    case MessageName::LibWebRTCCodecs_CompletedEncoding:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::VideoEncoderIdentifier", nullptr, false },
            { "data", "IPC::DataReference", nullptr, false },
            { "info", "webrtc::WebKitEncodedFrameInfo", nullptr, false },
        };
    case MessageName::LibWebRTCCodecs_FlushEncoderCompleted:
        return Vector<ArgumentDescription> {
            { "id", "WebKit::VideoEncoderIdentifier", nullptr, false },
        };
    case MessageName::LibWebRTCCodecs_SetEncodingDescription:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::VideoEncoderIdentifier", nullptr, false },
            { "data", "IPC::DataReference", nullptr, false },
        };
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
    case MessageName::RemoteVideoFrameObjectHeapProxyProcessor_SetSharedVideoFrameSemaphore:
        return Vector<ArgumentDescription> {
            { "semaphore", "IPC::Semaphore", nullptr, false },
        };
    case MessageName::RemoteVideoFrameObjectHeapProxyProcessor_SetSharedVideoFrameMemory:
        return Vector<ArgumentDescription> {
            { "storageHandle", "WebKit::SharedMemory::Handle", nullptr, false },
        };
    case MessageName::RemoteVideoFrameObjectHeapProxyProcessor_NewVideoFrameBuffer:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteVideoFrameIdentifier", nullptr, false },
            { "frame", "WebKit::SharedVideoFrame::Buffer", nullptr, true },
        };
    case MessageName::RemoteVideoFrameObjectHeapProxyProcessor_NewConvertedVideoFrameBuffer:
        return Vector<ArgumentDescription> {
            { "frame", "WebKit::SharedVideoFrame::Buffer", nullptr, true },
        };
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
    case MessageName::SampleBufferDisplayLayer_SetDidFail:
        return Vector<ArgumentDescription> {
            { "value", "bool", nullptr, false },
        };
#endif
    case MessageName::WebGeolocationManager_DidChangePosition:
        return Vector<ArgumentDescription> {
            { "registrableDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "position", "WebCore::GeolocationPositionData", nullptr, false },
        };
    case MessageName::WebGeolocationManager_DidFailToDeterminePosition:
        return Vector<ArgumentDescription> {
            { "registrableDomain", "WebCore::RegistrableDomain", nullptr, false },
            { "errorMessage", "String", nullptr, false },
        };
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebGeolocationManager_ResetPermissions:
        return Vector<ArgumentDescription> {
            { "registrableDomain", "WebCore::RegistrableDomain", nullptr, false },
        };
#endif
    case MessageName::RemoteWebInspectorUI_Initialize:
        return Vector<ArgumentDescription> {
            { "debuggableInfo", "WebKit::DebuggableInfoData", nullptr, false },
            { "backendCommandsURL", "String", nullptr, false },
        };
    case MessageName::RemoteWebInspectorUI_UpdateFindString:
        return Vector<ArgumentDescription> {
            { "findString", "String", nullptr, false },
        };
#if ENABLE(INSPECTOR_TELEMETRY)
    case MessageName::RemoteWebInspectorUI_SetDiagnosticLoggingAvailable:
        return Vector<ArgumentDescription> {
            { "available", "bool", nullptr, false },
        };
#endif
    case MessageName::RemoteWebInspectorUI_ShowConsole:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteWebInspectorUI_ShowResources:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteWebInspectorUI_SendMessageToFrontend:
        return Vector<ArgumentDescription> {
            { "message", "String", nullptr, false },
        };
    case MessageName::WebInspector_Show:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspector_Close:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspector_SetAttached:
        return Vector<ArgumentDescription> {
            { "attached", "bool", nullptr, false },
        };
    case MessageName::WebInspector_ShowConsole:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspector_ShowResources:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspector_ShowMainResourceForFrame:
        return Vector<ArgumentDescription> {
            { "frameIdentifier", "WebCore::FrameIdentifier", nullptr, false },
        };
    case MessageName::WebInspector_StartPageProfiling:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspector_StopPageProfiling:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspector_StartElementSelection:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspector_StopElementSelection:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspector_SetFrontendConnection:
        return Vector<ArgumentDescription> {
            { "connectionHandle", "IPC::Connection::Handle", nullptr, false },
        };
    case MessageName::WebInspectorInterruptDispatcher_NotifyNeedDebuggerBreak:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUI_EstablishConnection:
        return Vector<ArgumentDescription> {
            { "inspectedPageIdentifier", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "debuggableInfo", "WebKit::DebuggableInfoData", nullptr, false },
            { "underTest", "bool", nullptr, false },
            { "inspectionLevel", "unsigned", nullptr, false },
        };
    case MessageName::WebInspectorUI_UpdateConnection:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUI_AttachedBottom:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUI_AttachedRight:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUI_AttachedLeft:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUI_Detached:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUI_SetDockingUnavailable:
        return Vector<ArgumentDescription> {
            { "unavailable", "bool", nullptr, false },
        };
    case MessageName::WebInspectorUI_SetIsVisible:
        return Vector<ArgumentDescription> {
            { "visible", "bool", nullptr, false },
        };
    case MessageName::WebInspectorUI_UpdateFindString:
        return Vector<ArgumentDescription> {
            { "findString", "String", nullptr, false },
        };
#if ENABLE(INSPECTOR_TELEMETRY)
    case MessageName::WebInspectorUI_SetDiagnosticLoggingAvailable:
        return Vector<ArgumentDescription> {
            { "available", "bool", nullptr, false },
        };
#endif
    case MessageName::WebInspectorUI_ShowConsole:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUI_ShowResources:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUI_ShowMainResourceForFrame:
        return Vector<ArgumentDescription> {
            { "frameIdentifier", "String", nullptr, false },
        };
    case MessageName::WebInspectorUI_StartPageProfiling:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUI_StopPageProfiling:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUI_StartElementSelection:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUI_StopElementSelection:
        return Vector<ArgumentDescription> { };
    case MessageName::WebInspectorUI_SendMessageToFrontend:
        return Vector<ArgumentDescription> {
            { "message", "String", nullptr, false },
        };
    case MessageName::WebInspectorUI_EvaluateInFrontendForTesting:
        return Vector<ArgumentDescription> {
            { "expression", "String", nullptr, false },
        };
#if ENABLE(MEDIA_SESSION_COORDINATOR)
    case MessageName::RemoteMediaSessionCoordinator_SeekSessionToTime:
        return Vector<ArgumentDescription> {
            { "time", "double", nullptr, false },
        };
    case MessageName::RemoteMediaSessionCoordinator_PlaySession:
        return Vector<ArgumentDescription> {
            { "atTime", "double", nullptr, true },
            { "hostTime", "MonotonicTime", nullptr, true },
        };
    case MessageName::RemoteMediaSessionCoordinator_PauseSession:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaSessionCoordinator_SetSessionTrack:
        return Vector<ArgumentDescription> {
            { "trackIdentifier", "String", nullptr, false },
        };
    case MessageName::RemoteMediaSessionCoordinator_CoordinatorStateChanged:
        return Vector<ArgumentDescription> {
            { "state", "uint8_t", "WebCore::MediaSessionCoordinatorState", false },
        };
#endif
#if USE(LIBWEBRTC)
    case MessageName::LibWebRTCNetwork_SignalReadPacket:
        return Vector<ArgumentDescription> {
            { "socketIdentifier", "WebCore::LibWebRTCSocketIdentifier", nullptr, false },
            { "data", "IPC::DataReference", nullptr, false },
            { "address", "WebKit::RTCNetwork::IPAddress", nullptr, false },
            { "port", "uint16_t", nullptr, false },
            { "timestamp", "int64_t", nullptr, false },
        };
    case MessageName::LibWebRTCNetwork_SignalSentPacket:
        return Vector<ArgumentDescription> {
            { "socketIdentifier", "WebCore::LibWebRTCSocketIdentifier", nullptr, false },
            { "packetSize", "int", nullptr, false },
            { "timestamp", "int64_t", nullptr, false },
        };
    case MessageName::LibWebRTCNetwork_SignalAddressReady:
        return Vector<ArgumentDescription> {
            { "socketIdentifier", "WebCore::LibWebRTCSocketIdentifier", nullptr, false },
            { "address", "WebKit::RTCNetwork::SocketAddress", nullptr, false },
        };
    case MessageName::LibWebRTCNetwork_SignalConnect:
        return Vector<ArgumentDescription> {
            { "socketIdentifier", "WebCore::LibWebRTCSocketIdentifier", nullptr, false },
        };
    case MessageName::LibWebRTCNetwork_SignalClose:
        return Vector<ArgumentDescription> {
            { "socketIdentifier", "WebCore::LibWebRTCSocketIdentifier", nullptr, false },
            { "error", "int", nullptr, false },
        };
#endif
#if ENABLE(WEB_RTC)
    case MessageName::RTCDataChannelRemoteManager_SendData:
        return Vector<ArgumentDescription> {
            { "source", "WebCore::RTCDataChannelIdentifier", nullptr, false },
            { "isRaw", "bool", nullptr, false },
            { "text", "IPC::DataReference", nullptr, false },
        };
    case MessageName::RTCDataChannelRemoteManager_Close:
        return Vector<ArgumentDescription> {
            { "source", "WebCore::RTCDataChannelIdentifier", nullptr, false },
        };
    case MessageName::RTCDataChannelRemoteManager_ChangeReadyState:
        return Vector<ArgumentDescription> {
            { "handler", "WebCore::RTCDataChannelIdentifier", nullptr, false },
            { "state", "uint8_t", "WebCore::RTCDataChannelState", false },
        };
    case MessageName::RTCDataChannelRemoteManager_ReceiveData:
        return Vector<ArgumentDescription> {
            { "handler", "WebCore::RTCDataChannelIdentifier", nullptr, false },
            { "isRaw", "bool", nullptr, false },
            { "data", "IPC::DataReference", nullptr, false },
        };
    case MessageName::RTCDataChannelRemoteManager_DetectError:
        return Vector<ArgumentDescription> {
            { "handler", "WebCore::RTCDataChannelIdentifier", nullptr, false },
            { "type", "uint8_t", "WebCore::RTCErrorDetailType", false },
            { "message", "String", nullptr, false },
        };
    case MessageName::RTCDataChannelRemoteManager_BufferedAmountIsDecreasing:
        return Vector<ArgumentDescription> {
            { "handler", "WebCore::RTCDataChannelIdentifier", nullptr, false },
            { "amount", "size_t", nullptr, false },
        };
#endif
#if ENABLE(WEB_RTC)
    case MessageName::WebMDNSRegister_FinishedRegisteringMDNSName:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::MDNSRegisterIdentifier", nullptr, false },
            { "mdnsName", "String", nullptr, false },
            { "error", "WebCore::MDNSRegisterError", nullptr, true },
        };
#endif
#if USE(LIBWEBRTC)
    case MessageName::WebRTCMonitor_NetworksChanged:
        return Vector<ArgumentDescription> {
            { "networks", "Vector<WebKit::RTCNetwork>", nullptr, false },
            { "defaultIPV4Address", "WebKit::RTCNetwork::IPAddress", nullptr, false },
            { "defaultIPV6Address", "WebKit::RTCNetwork::IPAddress", nullptr, false },
        };
#endif
#if USE(LIBWEBRTC)
    case MessageName::WebRTCResolver_SetResolvedAddress:
        return Vector<ArgumentDescription> {
            { "addresses", "Vector<WebKit::RTCNetwork::IPAddress>", nullptr, false },
        };
    case MessageName::WebRTCResolver_ResolvedAddressError:
        return Vector<ArgumentDescription> {
            { "error", "int", nullptr, false },
        };
#endif
#if ENABLE(SHAREABLE_RESOURCE)
    case MessageName::NetworkProcessConnection_DidCacheResource:
        return Vector<ArgumentDescription> {
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "resource", "WebKit::ShareableResource::Handle", nullptr, false },
        };
#endif
    case MessageName::NetworkProcessConnection_DidFinishPingLoad:
        return Vector<ArgumentDescription> {
            { "pingLoadIdentifier", "WebCore::ResourceLoaderIdentifier", nullptr, false },
            { "error", "WebCore::ResourceError", nullptr, false },
            { "response", "WebCore::ResourceResponse", nullptr, false },
        };
    case MessageName::NetworkProcessConnection_DidFinishPreconnection:
        return Vector<ArgumentDescription> {
            { "preconnectionIdentifier", "WebCore::ResourceLoaderIdentifier", nullptr, false },
            { "error", "WebCore::ResourceError", nullptr, false },
        };
    case MessageName::NetworkProcessConnection_SetOnLineState:
        return Vector<ArgumentDescription> {
            { "isOnLine", "bool", nullptr, false },
        };
    case MessageName::NetworkProcessConnection_CookieAcceptPolicyChanged:
        return Vector<ArgumentDescription> {
            { "policy", "uint8_t", "WebCore::HTTPCookieAcceptPolicy", false },
        };
#if HAVE(COOKIE_CHANGE_LISTENER_API)
    case MessageName::NetworkProcessConnection_CookiesAdded:
        return Vector<ArgumentDescription> {
            { "host", "String", nullptr, false },
            { "cookies", "Vector<WebCore::Cookie>", nullptr, false },
        };
    case MessageName::NetworkProcessConnection_CookiesDeleted:
        return Vector<ArgumentDescription> {
            { "host", "String", nullptr, false },
            { "cookies", "Vector<WebCore::Cookie>", nullptr, false },
        };
    case MessageName::NetworkProcessConnection_AllCookiesDeleted:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(WEB_RTC)
    case MessageName::NetworkProcessConnection_ConnectToRTCDataChannelRemoteSource:
        return Vector<ArgumentDescription> {
            { "source", "WebCore::RTCDataChannelIdentifier", nullptr, false },
            { "handler", "WebCore::RTCDataChannelIdentifier", nullptr, false },
        };
#endif
    case MessageName::NetworkProcessConnection_MessagesAvailableForPort:
        return Vector<ArgumentDescription> {
            { "port", "WebCore::MessagePortIdentifier", nullptr, false },
        };
    case MessageName::NetworkProcessConnection_BroadcastConsoleMessage:
        return Vector<ArgumentDescription> {
            { "messageSource", "uint8_t", "JSC::MessageSource", false },
            { "messageLevel", "uint8_t", "JSC::MessageLevel", false },
            { "message", "String", nullptr, false },
        };
    case MessageName::WebResourceLoader_WillSendRequest:
        return Vector<ArgumentDescription> {
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "requestBody", "IPC::FormDataReference", nullptr, false },
            { "redirectResponse", "WebCore::ResourceResponse", nullptr, false },
        };
    case MessageName::WebResourceLoader_SetWorkerStart:
        return Vector<ArgumentDescription> {
            { "value", "MonotonicTime", nullptr, false },
        };
    case MessageName::WebResourceLoader_DidSendData:
        return Vector<ArgumentDescription> {
            { "bytesSent", "uint64_t", nullptr, false },
            { "totalBytesToBeSent", "uint64_t", nullptr, false },
        };
    case MessageName::WebResourceLoader_DidReceiveResponse:
        return Vector<ArgumentDescription> {
            { "response", "WebCore::ResourceResponse", nullptr, false },
            { "privateRelayed", "bool", "WebKit::PrivateRelayed", false },
            { "needsContinueDidReceiveResponseMessage", "bool", nullptr, false },
            { "optionalNetworkLoadMetrics", "WebCore::NetworkLoadMetrics", nullptr, true },
        };
    case MessageName::WebResourceLoader_DidReceiveData:
        return Vector<ArgumentDescription> {
            { "data", "IPC::SharedBufferReference", nullptr, false },
            { "encodedDataLength", "uint64_t", nullptr, false },
        };
    case MessageName::WebResourceLoader_DidFinishResourceLoad:
        return Vector<ArgumentDescription> {
            { "networkLoadMetrics", "WebCore::NetworkLoadMetrics", nullptr, false },
        };
    case MessageName::WebResourceLoader_DidFailResourceLoad:
        return Vector<ArgumentDescription> {
            { "error", "WebCore::ResourceError", nullptr, false },
        };
    case MessageName::WebResourceLoader_DidFailServiceWorkerLoad:
        return Vector<ArgumentDescription> {
            { "error", "WebCore::ResourceError", nullptr, false },
        };
    case MessageName::WebResourceLoader_ServiceWorkerDidNotHandle:
        return Vector<ArgumentDescription> { };
    case MessageName::WebResourceLoader_DidBlockAuthenticationChallenge:
        return Vector<ArgumentDescription> { };
    case MessageName::WebResourceLoader_StopLoadingAfterXFrameOptionsOrContentSecurityPolicyDenied:
        return Vector<ArgumentDescription> {
            { "response", "WebCore::ResourceResponse", nullptr, false },
        };
#if ENABLE(SHAREABLE_RESOURCE)
    case MessageName::WebResourceLoader_DidReceiveResource:
        return Vector<ArgumentDescription> {
            { "resource", "WebKit::ShareableResource::Handle", nullptr, false },
        };
#endif
#if ENABLE(CONTENT_FILTERING_IN_NETWORKING_PROCESS)
    case MessageName::WebResourceLoader_ContentFilterDidBlockLoad:
        return Vector<ArgumentDescription> {
            { "unblockHandler", "WebCore::ContentFilterUnblockHandler", nullptr, false },
            { "unblockRequestDeniedScript", "String", nullptr, false },
            { "error", "WebCore::ResourceError", nullptr, false },
            { "blockedPageURL", "URL", nullptr, false },
            { "substituteData", "WebCore::SubstituteData", nullptr, false },
        };
#endif
    case MessageName::WebSocketChannel_DidConnect:
        return Vector<ArgumentDescription> {
            { "subprotocol", "String", nullptr, false },
            { "extensions", "String", nullptr, false },
        };
    case MessageName::WebSocketChannel_DidClose:
        return Vector<ArgumentDescription> {
            { "code", "unsigned short", nullptr, false },
            { "reason", "String", nullptr, false },
        };
    case MessageName::WebSocketChannel_DidReceiveText:
        return Vector<ArgumentDescription> {
            { "message", "String", nullptr, false },
        };
    case MessageName::WebSocketChannel_DidReceiveBinaryData:
        return Vector<ArgumentDescription> {
            { "data", "IPC::DataReference", nullptr, false },
        };
    case MessageName::WebSocketChannel_DidReceiveMessageError:
        return Vector<ArgumentDescription> {
            { "errorMessage", "String", nullptr, false },
        };
    case MessageName::WebSocketChannel_DidSendHandshakeRequest:
        return Vector<ArgumentDescription> {
            { "request", "WebCore::ResourceRequest", nullptr, false },
        };
    case MessageName::WebSocketChannel_DidReceiveHandshakeResponse:
        return Vector<ArgumentDescription> {
            { "response", "WebCore::ResourceResponse", nullptr, false },
        };
    case MessageName::WebSocketStream_DidOpenSocketStream:
        return Vector<ArgumentDescription> { };
    case MessageName::WebSocketStream_DidCloseSocketStream:
        return Vector<ArgumentDescription> { };
    case MessageName::WebSocketStream_DidReceiveSocketStreamData:
        return Vector<ArgumentDescription> {
            { "data", "IPC::DataReference", nullptr, false },
        };
    case MessageName::WebSocketStream_DidFailToReceiveSocketStreamData:
        return Vector<ArgumentDescription> { };
    case MessageName::WebSocketStream_DidUpdateBufferedAmount:
        return Vector<ArgumentDescription> {
            { "bufferedAmount", "uint64_t", nullptr, false },
        };
    case MessageName::WebSocketStream_DidFailSocketStream:
        return Vector<ArgumentDescription> {
            { "error", "WebCore::SocketStreamError", nullptr, false },
        };
    case MessageName::WebSocketStream_DidSendData:
        return Vector<ArgumentDescription> {
            { "identifier", "uint64_t", nullptr, false },
            { "success", "bool", nullptr, false },
        };
    case MessageName::WebSocketStream_DidSendHandshake:
        return Vector<ArgumentDescription> {
            { "identifier", "uint64_t", nullptr, false },
            { "success", "bool", nullptr, false },
            { "didAccessSecureCookies", "bool", nullptr, false },
        };
    case MessageName::WebNotificationManager_DidShowNotification:
        return Vector<ArgumentDescription> {
            { "notificationID", "UUID", nullptr, false },
        };
    case MessageName::WebNotificationManager_DidClickNotification:
        return Vector<ArgumentDescription> {
            { "notificationID", "UUID", nullptr, false },
        };
    case MessageName::WebNotificationManager_DidCloseNotifications:
        return Vector<ArgumentDescription> {
            { "notificationIDs", "Vector<UUID>", nullptr, false },
        };
    case MessageName::WebNotificationManager_DidUpdateNotificationDecision:
        return Vector<ArgumentDescription> {
            { "originString", "String", nullptr, false },
            { "allowed", "bool", nullptr, false },
        };
    case MessageName::WebNotificationManager_DidRemoveNotificationDecisions:
        return Vector<ArgumentDescription> {
            { "originStrings", "Vector<String>", nullptr, false },
        };
#if ENABLE(MEDIA_STREAM)
    case MessageName::SpeechRecognitionRealtimeMediaSourceManager_CreateSource:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::RealtimeMediaSourceIdentifier", nullptr, false },
            { "device", "WebCore::CaptureDevice", nullptr, false },
            { "pageIdentifier", "WebCore::PageIdentifier", nullptr, false },
        };
    case MessageName::SpeechRecognitionRealtimeMediaSourceManager_DeleteSource:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::RealtimeMediaSourceIdentifier", nullptr, false },
        };
    case MessageName::SpeechRecognitionRealtimeMediaSourceManager_Start:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::RealtimeMediaSourceIdentifier", nullptr, false },
        };
    case MessageName::SpeechRecognitionRealtimeMediaSourceManager_Stop:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::RealtimeMediaSourceIdentifier", nullptr, false },
        };
#if ENABLE(SANDBOX_EXTENSIONS)
    case MessageName::SpeechRecognitionRealtimeMediaSourceManager_GrantSandboxExtensions:
        return Vector<ArgumentDescription> {
            { "bootstrapHandle", "WebKit::SandboxExtension::Handle", nullptr, false },
            { "sandboxExtensionHandleForTCCD", "WebKit::SandboxExtension::Handle", nullptr, false },
            { "sandboxExtensionHandleForMicrophone", "WebKit::SandboxExtension::Handle", nullptr, false },
        };
    case MessageName::SpeechRecognitionRealtimeMediaSourceManager_RevokeSandboxExtensions:
        return Vector<ArgumentDescription> { };
#endif
#endif
    case MessageName::WebSharedWorkerContextManagerConnection_LaunchSharedWorker:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::ClientOrigin", nullptr, false },
            { "sharedWorkerIdentifier", "WebCore::SharedWorkerIdentifier", nullptr, false },
            { "workerOptions", "WebCore::WorkerOptions", nullptr, false },
            { "workerFetchResult", "WebCore::WorkerFetchResult", nullptr, false },
            { "workerInitializationData", "WebCore::WorkerInitializationData", nullptr, false },
        };
    case MessageName::WebSharedWorkerContextManagerConnection_PostConnectEvent:
        return Vector<ArgumentDescription> {
            { "sharedWorkerIdentifier", "WebCore::SharedWorkerIdentifier", nullptr, false },
            { "port", "WebCore::TransferredMessagePort", nullptr, false },
            { "sourceOrigin", "String", nullptr, false },
        };
    case MessageName::WebSharedWorkerContextManagerConnection_TerminateSharedWorker:
        return Vector<ArgumentDescription> {
            { "sharedWorkerIdentifier", "WebCore::SharedWorkerIdentifier", nullptr, false },
        };
    case MessageName::WebSharedWorkerContextManagerConnection_UpdatePreferencesStore:
        return Vector<ArgumentDescription> {
            { "store", "WebKit::WebPreferencesStore", nullptr, false },
        };
    case MessageName::WebSharedWorkerContextManagerConnection_SetUserAgent:
        return Vector<ArgumentDescription> {
            { "userAgent", "String", nullptr, false },
        };
    case MessageName::WebSharedWorkerContextManagerConnection_Close:
        return Vector<ArgumentDescription> { };
    case MessageName::WebSharedWorkerContextManagerConnection_SuspendSharedWorker:
        return Vector<ArgumentDescription> {
            { "sharedWorkerIdentifier", "WebCore::SharedWorkerIdentifier", nullptr, false },
        };
    case MessageName::WebSharedWorkerContextManagerConnection_ResumeSharedWorker:
        return Vector<ArgumentDescription> {
            { "sharedWorkerIdentifier", "WebCore::SharedWorkerIdentifier", nullptr, false },
        };
    case MessageName::WebSharedWorkerObjectConnection_FetchScriptInClient:
        return Vector<ArgumentDescription> {
            { "url", "URL", nullptr, false },
            { "sharedWorkerObjectIdentifier", "WebCore::SharedWorkerObjectIdentifier", nullptr, false },
            { "workerOptions", "WebCore::WorkerOptions", nullptr, false },
        };
    case MessageName::WebSharedWorkerObjectConnection_NotifyWorkerObjectOfLoadCompletion:
        return Vector<ArgumentDescription> {
            { "sharedWorkerObjectIdentifier", "WebCore::SharedWorkerObjectIdentifier", nullptr, false },
            { "error", "WebCore::ResourceError", nullptr, false },
        };
    case MessageName::WebSharedWorkerObjectConnection_PostExceptionToWorkerObject:
        return Vector<ArgumentDescription> {
            { "sharedWorkerObjectIdentifier", "WebCore::SharedWorkerObjectIdentifier", nullptr, false },
            { "errorMessage", "String", nullptr, false },
            { "lineNumber", "int", nullptr, false },
            { "columnNumber", "int", nullptr, false },
            { "sourceURL", "String", nullptr, false },
        };
#if ENABLE(SERVICE_WORKER)
    case MessageName::WebSWClientConnection_JobRejectedInServer:
        return Vector<ArgumentDescription> {
            { "jobDataIdentifier", "WebCore::ServiceWorkerJobIdentifier", nullptr, false },
            { "exception", "WebCore::ExceptionData", nullptr, false },
        };
    case MessageName::WebSWClientConnection_RegistrationJobResolvedInServer:
        return Vector<ArgumentDescription> {
            { "jobDataIdentifier", "WebCore::ServiceWorkerJobIdentifier", nullptr, false },
            { "registration", "WebCore::ServiceWorkerRegistrationData", nullptr, false },
            { "shouldNotifyWhenResolved", "bool", "WebCore::ShouldNotifyWhenResolved", false },
        };
    case MessageName::WebSWClientConnection_StartScriptFetchForServer:
        return Vector<ArgumentDescription> {
            { "jobDataIdentifier", "WebCore::ServiceWorkerJobIdentifier", nullptr, false },
            { "registrationKey", "WebCore::ServiceWorkerRegistrationKey", nullptr, false },
            { "cachePolicy", "uint8_t", "WebCore::FetchOptions::Cache", false },
        };
    case MessageName::WebSWClientConnection_UpdateRegistrationState:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerRegistrationIdentifier", nullptr, false },
            { "state", "uint8_t", "WebCore::ServiceWorkerRegistrationState", false },
            { "serviceWorkerIdentifier", "WebCore::ServiceWorkerData", nullptr, true },
        };
    case MessageName::WebSWClientConnection_UpdateWorkerState:
        return Vector<ArgumentDescription> {
            { "serviceWorkerIdentifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "state", "uint8_t", "WebCore::ServiceWorkerState", false },
        };
    case MessageName::WebSWClientConnection_FireUpdateFoundEvent:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerRegistrationIdentifier", nullptr, false },
        };
    case MessageName::WebSWClientConnection_SetRegistrationLastUpdateTime:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerRegistrationIdentifier", nullptr, false },
            { "lastUpdateTime", "WallTime", nullptr, false },
        };
    case MessageName::WebSWClientConnection_SetRegistrationUpdateViaCache:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerRegistrationIdentifier", nullptr, false },
            { "updateViaCache", "uint8_t", "WebCore::ServiceWorkerUpdateViaCache", false },
        };
    case MessageName::WebSWClientConnection_NotifyClientsOfControllerChange:
        return Vector<ArgumentDescription> {
            { "contextIdentifiers", "HashSet<WebCore::ScriptExecutionContextIdentifier>", nullptr, false },
            { "newController", "WebCore::ServiceWorkerData", nullptr, false },
        };
    case MessageName::WebSWClientConnection_SetSWOriginTableIsImported:
        return Vector<ArgumentDescription> { };
    case MessageName::WebSWClientConnection_SetSWOriginTableSharedMemory:
        return Vector<ArgumentDescription> {
            { "handle", "WebKit::SharedMemory::Handle", nullptr, false },
        };
    case MessageName::WebSWClientConnection_PostMessageToServiceWorkerClient:
        return Vector<ArgumentDescription> {
            { "destinationContextIdentifier", "WebCore::ScriptExecutionContextIdentifier", nullptr, false },
            { "message", "WebCore::MessageWithMessagePorts", nullptr, false },
            { "source", "WebCore::ServiceWorkerData", nullptr, false },
            { "sourceOrigin", "String", nullptr, false },
        };
    case MessageName::WebSWClientConnection_SetServiceWorkerClientIsControlled:
        return Vector<ArgumentDescription> {
            { "workerIdentifier", "WebCore::ScriptExecutionContextIdentifier", nullptr, false },
            { "data", "WebCore::ServiceWorkerRegistrationData", nullptr, false },
        };
    case MessageName::WebSWClientConnection_FocusServiceWorkerClient:
        return Vector<ArgumentDescription> {
            { "clientIdentifier", "WebCore::ScriptExecutionContextIdentifier", nullptr, false },
        };
#endif
#if ENABLE(SERVICE_WORKER)
    case MessageName::WebSWContextManagerConnection_InstallServiceWorker:
        return Vector<ArgumentDescription> {
            { "contextData", "WebCore::ServiceWorkerContextData", nullptr, false },
            { "workerData", "WebCore::ServiceWorkerData", nullptr, false },
            { "userAgent", "String", nullptr, false },
            { "workerThreadMode", "bool", "WebCore::WorkerThreadMode", false },
        };
    case MessageName::WebSWContextManagerConnection_UpdateAppInitiatedValue:
        return Vector<ArgumentDescription> {
            { "serviceWorkerIdentifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "lastNavigationWasAppInitiated", "bool", "WebCore::LastNavigationWasAppInitiated", false },
        };
    case MessageName::WebSWContextManagerConnection_StartFetch:
        return Vector<ArgumentDescription> {
            { "serverConnectionIdentifier", "WebCore::SWServerConnectionIdentifier", nullptr, false },
            { "serviceWorkerIdentifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "fetchIdentifier", "WebCore::FetchIdentifier", nullptr, false },
            { "request", "WebCore::ResourceRequest", nullptr, false },
            { "options", "WebCore::FetchOptions", nullptr, false },
            { "requestBody", "IPC::FormDataReference", nullptr, false },
            { "referrer", "String", nullptr, false },
            { "isServiceWorkerNavigationPreloadEnabled", "bool", nullptr, false },
            { "clientIdentifier", "String", nullptr, false },
            { "resutlingClientIdentifier", "String", nullptr, false },
        };
    case MessageName::WebSWContextManagerConnection_CancelFetch:
        return Vector<ArgumentDescription> {
            { "serverConnectionIdentifier", "WebCore::SWServerConnectionIdentifier", nullptr, false },
            { "serviceWorkerIdentifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "fetchIdentifier", "WebCore::FetchIdentifier", nullptr, false },
        };
    case MessageName::WebSWContextManagerConnection_ContinueDidReceiveFetchResponse:
        return Vector<ArgumentDescription> {
            { "serverConnectionIdentifier", "WebCore::SWServerConnectionIdentifier", nullptr, false },
            { "serviceWorkerIdentifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "fetchIdentifier", "WebCore::FetchIdentifier", nullptr, false },
        };
    case MessageName::WebSWContextManagerConnection_PostMessageToServiceWorker:
        return Vector<ArgumentDescription> {
            { "destinationIdentifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "message", "WebCore::MessageWithMessagePorts", nullptr, false },
            { "sourceData", "WebCore::ServiceWorkerOrClientData", nullptr, false },
        };
    case MessageName::WebSWContextManagerConnection_FireInstallEvent:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
        };
    case MessageName::WebSWContextManagerConnection_FireActivateEvent:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
        };
    case MessageName::WebSWContextManagerConnection_FirePushEvent:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "data", "IPC::DataReference", nullptr, true },
        };
    case MessageName::WebSWContextManagerConnection_FireNotificationEvent:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "data", "WebCore::NotificationData", nullptr, false },
            { "type", "bool", "WebCore::NotificationEventType", false },
        };
    case MessageName::WebSWContextManagerConnection_TerminateWorker:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
        };
#if ENABLE(SHAREABLE_RESOURCE) && PLATFORM(COCOA)
    case MessageName::WebSWContextManagerConnection_DidSaveScriptsToDisk:
        return Vector<ArgumentDescription> {
            { "serviceWorkerIdentifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "script", "WebCore::ScriptBuffer", nullptr, false },
            { "importedScripts", "HashMap<URL, WebCore::ScriptBuffer>", nullptr, false },
        };
#endif
    case MessageName::WebSWContextManagerConnection_MatchAllCompleted:
        return Vector<ArgumentDescription> {
            { "matchAllRequestIdentifier", "uint64_t", nullptr, false },
            { "clientsData", "Vector<WebCore::ServiceWorkerClientData>", nullptr, false },
        };
    case MessageName::WebSWContextManagerConnection_SkipWaitingCompleted:
        return Vector<ArgumentDescription> {
            { "matchAllRequestIdentifier", "uint64_t", nullptr, false },
        };
    case MessageName::WebSWContextManagerConnection_SetUserAgent:
        return Vector<ArgumentDescription> {
            { "userAgent", "String", nullptr, false },
        };
    case MessageName::WebSWContextManagerConnection_UpdatePreferencesStore:
        return Vector<ArgumentDescription> {
            { "store", "WebKit::WebPreferencesStore", nullptr, false },
        };
    case MessageName::WebSWContextManagerConnection_Close:
        return Vector<ArgumentDescription> { };
    case MessageName::WebSWContextManagerConnection_SetThrottleState:
        return Vector<ArgumentDescription> {
            { "isThrottleable", "bool", nullptr, false },
        };
    case MessageName::WebSWContextManagerConnection_ConvertFetchToDownload:
        return Vector<ArgumentDescription> {
            { "serverConnectionIdentifier", "WebCore::SWServerConnectionIdentifier", nullptr, false },
            { "serviceWorkerIdentifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "identifier", "WebCore::FetchIdentifier", nullptr, false },
        };
    case MessageName::WebSWContextManagerConnection_NavigationPreloadIsReady:
        return Vector<ArgumentDescription> {
            { "serverConnectionIdentifier", "WebCore::SWServerConnectionIdentifier", nullptr, false },
            { "serviceWorkerIdentifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "identifier", "WebCore::FetchIdentifier", nullptr, false },
            { "response", "WebCore::ResourceResponse", nullptr, false },
        };
    case MessageName::WebSWContextManagerConnection_NavigationPreloadFailed:
        return Vector<ArgumentDescription> {
            { "serverConnectionIdentifier", "WebCore::SWServerConnectionIdentifier", nullptr, false },
            { "serviceWorkerIdentifier", "WebCore::ServiceWorkerIdentifier", nullptr, false },
            { "identifier", "WebCore::FetchIdentifier", nullptr, false },
            { "error", "WebCore::ResourceError", nullptr, false },
        };
#endif
    case MessageName::WebUserContentController_AddContentWorlds:
        return Vector<ArgumentDescription> {
            { "worlds", "Vector<std::pair<WebKit::ContentWorldIdentifier, String>>", nullptr, false },
        };
    case MessageName::WebUserContentController_RemoveContentWorlds:
        return Vector<ArgumentDescription> {
            { "worldIdentifiers", "Vector<WebKit::ContentWorldIdentifier>", nullptr, false },
        };
    case MessageName::WebUserContentController_AddUserScripts:
        return Vector<ArgumentDescription> {
            { "userScripts", "Vector<WebKit::WebUserScriptData>", nullptr, false },
            { "immediately", "bool", "WebKit::InjectUserScriptImmediately", false },
        };
    case MessageName::WebUserContentController_RemoveUserScript:
        return Vector<ArgumentDescription> {
            { "worldIdentifier", "WebKit::ContentWorldIdentifier", nullptr, false },
            { "identifier", "uint64_t", nullptr, false },
        };
    case MessageName::WebUserContentController_RemoveAllUserScripts:
        return Vector<ArgumentDescription> {
            { "worldIdentifiers", "Vector<WebKit::ContentWorldIdentifier>", nullptr, false },
        };
    case MessageName::WebUserContentController_AddUserStyleSheets:
        return Vector<ArgumentDescription> {
            { "userStyleSheets", "Vector<WebKit::WebUserStyleSheetData>", nullptr, false },
        };
    case MessageName::WebUserContentController_RemoveUserStyleSheet:
        return Vector<ArgumentDescription> {
            { "worldIdentifier", "WebKit::ContentWorldIdentifier", nullptr, false },
            { "identifier", "uint64_t", nullptr, false },
        };
    case MessageName::WebUserContentController_RemoveAllUserStyleSheets:
        return Vector<ArgumentDescription> {
            { "worldIdentifiers", "Vector<WebKit::ContentWorldIdentifier>", nullptr, false },
        };
    case MessageName::WebUserContentController_AddUserScriptMessageHandlers:
        return Vector<ArgumentDescription> {
            { "scriptMessageHandlers", "Vector<WebKit::WebScriptMessageHandlerData>", nullptr, false },
        };
    case MessageName::WebUserContentController_RemoveUserScriptMessageHandler:
        return Vector<ArgumentDescription> {
            { "worldIdentifier", "WebKit::ContentWorldIdentifier", nullptr, false },
            { "identifier", "uint64_t", nullptr, false },
        };
    case MessageName::WebUserContentController_RemoveAllUserScriptMessageHandlersForWorlds:
        return Vector<ArgumentDescription> {
            { "worldIdentifiers", "Vector<WebKit::ContentWorldIdentifier>", nullptr, false },
        };
    case MessageName::WebUserContentController_RemoveAllUserScriptMessageHandlers:
        return Vector<ArgumentDescription> { };
#if ENABLE(CONTENT_EXTENSIONS)
    case MessageName::WebUserContentController_AddContentRuleLists:
        return Vector<ArgumentDescription> {
            { "contentFilters", "Vector<std::pair<WebKit::WebCompiledContentRuleListData, URL>>", nullptr, false },
        };
    case MessageName::WebUserContentController_RemoveContentRuleList:
        return Vector<ArgumentDescription> {
            { "name", "String", nullptr, false },
        };
    case MessageName::WebUserContentController_RemoveAllContentRuleLists:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::RemoteWebLockRegistry_DidCompleteLockRequest:
        return Vector<ArgumentDescription> {
            { "lockIdentifier", "WebCore::WebLockIdentifier", nullptr, false },
            { "clientID", "WebCore::ScriptExecutionContextIdentifier", nullptr, false },
            { "success", "bool", nullptr, false },
        };
    case MessageName::RemoteWebLockRegistry_DidStealLock:
        return Vector<ArgumentDescription> {
            { "lockIdentifier", "WebCore::WebLockIdentifier", nullptr, false },
            { "clientID", "WebCore::ScriptExecutionContextIdentifier", nullptr, false },
        };
    case MessageName::WebBroadcastChannelRegistry_PostMessageToRemote:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::ClientOrigin", nullptr, false },
            { "name", "String", nullptr, false },
            { "message", "WebCore::MessageWithMessagePorts", nullptr, false },
        };
    case MessageName::WebFileSystemStorageConnection_InvalidateAccessHandle:
        return Vector<ArgumentDescription> {
            { "identifier", "WebCore::FileSystemSyncAccessHandleIdentifier", nullptr, false },
        };
    case MessageName::WebPermissionController_permissionChanged:
        return Vector<ArgumentDescription> {
            { "permissionName", "uint8_t", "WebCore::PermissionName", false },
            { "topOrigin", "WebCore::SecurityOriginData", nullptr, false },
        };
    case MessageName::WebScreenOrientationManager_OrientationDidChange:
        return Vector<ArgumentDescription> {
            { "orientation", "uint8_t", "WebCore::ScreenOrientationType", false },
        };
    case MessageName::WebSpeechRecognitionConnection_DidReceiveUpdate:
        return Vector<ArgumentDescription> {
            { "update", "WebCore::SpeechRecognitionUpdate", nullptr, false },
        };
#if USE(COORDINATED_GRAPHICS) || USE(TEXTURE_MAPPER)
    case MessageName::DrawingArea_UpdateBackingStoreState:
        return Vector<ArgumentDescription> {
            { "backingStoreStateID", "uint64_t", nullptr, false },
            { "respondImmediately", "bool", nullptr, false },
            { "deviceScaleFactor", "float", nullptr, false },
            { "size", "WebCore::IntSize", nullptr, false },
            { "scrollOffset", "WebCore::IntSize", nullptr, false },
        };
    case MessageName::DrawingArea_TargetRefreshRateDidChange:
        return Vector<ArgumentDescription> {
            { "rate", "unsigned", nullptr, false },
        };
#endif
    case MessageName::DrawingArea_DisplayDidRefresh:
        return Vector<ArgumentDescription> { };
#if PLATFORM(COCOA)
    case MessageName::DrawingArea_UpdateGeometry:
        return Vector<ArgumentDescription> {
            { "viewSize", "WebCore::IntSize", nullptr, false },
            { "flushSynchronously", "bool", nullptr, false },
            { "fencePort", "MachSendRight", nullptr, false },
        };
    case MessageName::DrawingArea_SetDeviceScaleFactor:
        return Vector<ArgumentDescription> {
            { "deviceScaleFactor", "float", nullptr, false },
        };
    case MessageName::DrawingArea_SetColorSpace:
        return Vector<ArgumentDescription> {
            { "colorSpace", "WebCore::DestinationColorSpace", nullptr, true },
        };
    case MessageName::DrawingArea_SetViewExposedRect:
        return Vector<ArgumentDescription> {
            { "viewExposedRect", "WebCore::FloatRect", nullptr, true },
        };
    case MessageName::DrawingArea_AcceleratedAnimationDidStart:
        return Vector<ArgumentDescription> {
            { "layerID", "WebCore::GraphicsLayer::PlatformLayerID", nullptr, false },
            { "key", "String", nullptr, false },
            { "startTime", "MonotonicTime", nullptr, false },
        };
    case MessageName::DrawingArea_AcceleratedAnimationDidEnd:
        return Vector<ArgumentDescription> {
            { "layerID", "WebCore::GraphicsLayer::PlatformLayerID", nullptr, false },
            { "key", "String", nullptr, false },
        };
    case MessageName::DrawingArea_DispatchAfterEnsuringDrawing:
        return Vector<ArgumentDescription> { };
#endif
#if USE(GRAPHICS_LAYER_WC)
    case MessageName::DrawingArea_UpdateGeometry:
        return Vector<ArgumentDescription> {
            { "backingStoreStateID", "uint64_t", nullptr, false },
            { "viewSize", "WebCore::IntSize", nullptr, false },
        };
#endif
#if PLATFORM(COCOA) || PLATFORM(GTK)
    case MessageName::DrawingArea_AdjustTransientZoom:
        return Vector<ArgumentDescription> {
            { "scale", "double", nullptr, false },
            { "origin", "WebCore::FloatPoint", nullptr, false },
        };
    case MessageName::DrawingArea_CommitTransientZoom:
        return Vector<ArgumentDescription> {
            { "scale", "double", nullptr, false },
            { "origin", "WebCore::FloatPoint", nullptr, false },
        };
#endif
    case MessageName::EventDispatcher_WheelEvent:
        return Vector<ArgumentDescription> {
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "event", "WebKit::WebWheelEvent", nullptr, false },
            { "rubberBandableEdges", "WebCore::RectEdges<bool>", nullptr, false },
        };
#if ENABLE(IOS_TOUCH_EVENTS)
    case MessageName::EventDispatcher_TouchEvent:
        return Vector<ArgumentDescription> {
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "event", "WebKit::WebTouchEvent", nullptr, false },
        };
    case MessageName::EventDispatcher_TouchEventWithoutCallback:
        return Vector<ArgumentDescription> {
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "event", "WebKit::WebTouchEvent", nullptr, false },
        };
#endif
#if ENABLE(MAC_GESTURE_EVENTS)
    case MessageName::EventDispatcher_GestureEvent:
        return Vector<ArgumentDescription> {
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "event", "WebKit::WebGestureEvent", nullptr, false },
        };
#endif
#if HAVE(CVDISPLAYLINK)
    case MessageName::EventDispatcher_DisplayWasRefreshed:
        return Vector<ArgumentDescription> {
            { "displayID", "uint32_t", nullptr, false },
            { "update", "WebCore::DisplayUpdate", nullptr, false },
            { "sendToMainThread", "bool", nullptr, false },
        };
#endif
#if ENABLE(MOMENTUM_EVENT_DISPATCHER)
    case MessageName::EventDispatcher_SetScrollingAccelerationCurve:
        return Vector<ArgumentDescription> {
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "curve", "WebKit::ScrollingAccelerationCurve", nullptr, true },
        };
#endif
    case MessageName::EventDispatcher_PageScreenDidChange:
        return Vector<ArgumentDescription> {
            { "pageID", "WebCore::PageIdentifier", nullptr, false },
            { "displayID", "uint32_t", nullptr, false },
            { "nominalFramesPerSecond", "unsigned", nullptr, true },
        };
    case MessageName::VisitedLinkTableController_SetVisitedLinkTable:
        return Vector<ArgumentDescription> {
            { "handle", "WebKit::SharedMemory::Handle", nullptr, false },
        };
    case MessageName::VisitedLinkTableController_VisitedLinkStateChanged:
        return Vector<ArgumentDescription> {
            { "linkHashes", "Vector<WebCore::SharedStringHash>", nullptr, false },
        };
    case MessageName::VisitedLinkTableController_AllVisitedLinkStateChanged:
        return Vector<ArgumentDescription> { };
    case MessageName::VisitedLinkTableController_RemoveAllVisitedLinks:
        return Vector<ArgumentDescription> { };
    case MessageName::WebFrame_GetFrameInfo:
        return Vector<ArgumentDescription> { };
    case MessageName::WebFrame_ContinueWillSubmitForm:
        return Vector<ArgumentDescription> {
            { "listenerID", "WebKit::FormSubmitListenerIdentifier", nullptr, false },
        };
    case MessageName::WebFrame_DidCommitLoadInAnotherProcess:
        return Vector<ArgumentDescription> {
            { "layerHostingContextIdentifier", "WebCore::LayerHostingContextIdentifier", nullptr, false },
        };
    case MessageName::WebFrame_DidFinishLoadInAnotherProcess:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_SetInitialFocus:
        return Vector<ArgumentDescription> {
            { "forward", "bool", nullptr, false },
            { "isKeyboardEventValid", "bool", nullptr, false },
            { "event", "WebKit::WebKeyboardEvent", nullptr, false },
        };
    case MessageName::WebPage_SetActivityState:
        return Vector<ArgumentDescription> {
            { "activityState", "OptionSet<WebCore::ActivityState::Flag>", nullptr, false },
            { "activityStateChangeID", "WebKit::ActivityStateChangeID", nullptr, false },
        };
    case MessageName::WebPage_SetLayerHostingMode:
        return Vector<ArgumentDescription> {
            { "layerHostingMode", "uint8_t", "WebKit::LayerHostingMode", false },
        };
    case MessageName::WebPage_SetBackgroundColor:
        return Vector<ArgumentDescription> {
            { "color", "WebCore::Color", nullptr, true },
        };
    case MessageName::WebPage_AddConsoleMessage:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "messageSource", "uint8_t", "JSC::MessageSource", false },
            { "messageLevel", "uint8_t", "JSC::MessageLevel", false },
            { "message", "String", nullptr, false },
            { "requestID", "WebCore::ResourceLoaderIdentifier", nullptr, true },
        };
    case MessageName::WebPage_EnqueueSecurityPolicyViolationEvent:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "eventInit", "WebCore::SecurityPolicyViolationEventInit", nullptr, false },
        };
    case MessageName::WebPage_SendReportToEndpoints:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "baseURL", "URL", nullptr, false },
            { "endpointURIs", "Vector<String>", nullptr, false },
            { "endpointTokens", "Vector<String>", nullptr, false },
            { "reportData", "IPC::FormDataReference", nullptr, false },
            { "reportType", "uint8_t", "WebCore::ViolationReportType", false },
        };
    case MessageName::WebPage_NotifyReportObservers:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "report", "Ref<WebCore::Report>", nullptr, false },
        };
    case MessageName::WebPage_TestProcessIncomingSyncMessagesWhenWaitingForSyncReply:
        return Vector<ArgumentDescription> { };
#if PLATFORM(COCOA)
    case MessageName::WebPage_SetTopContentInsetFenced:
        return Vector<ArgumentDescription> {
            { "contentInset", "float", nullptr, false },
            { "machSendRight", "MachSendRight", nullptr, false },
        };
#endif
    case MessageName::WebPage_SetTopContentInset:
        return Vector<ArgumentDescription> {
            { "contentInset", "float", nullptr, false },
        };
    case MessageName::WebPage_SetUnderlayColor:
        return Vector<ArgumentDescription> {
            { "color", "WebCore::Color", nullptr, false },
        };
    case MessageName::WebPage_SetUnderPageBackgroundColorOverride:
        return Vector<ArgumentDescription> {
            { "underPageBackgroundColorOverride", "WebCore::Color", nullptr, false },
        };
    case MessageName::WebPage_ViewWillStartLiveResize:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_ViewWillEndLiveResize:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_ExecuteEditCommandWithCallback:
        return Vector<ArgumentDescription> {
            { "name", "String", nullptr, false },
            { "argument", "String", nullptr, false },
        };
    case MessageName::WebPage_KeyEvent:
        return Vector<ArgumentDescription> {
            { "event", "WebKit::WebKeyboardEvent", nullptr, false },
        };
    case MessageName::WebPage_MouseEvent:
        return Vector<ArgumentDescription> {
            { "event", "WebKit::WebMouseEvent", nullptr, false },
            { "sandboxExtensions", "Vector<WebKit::SandboxExtension::Handle>", nullptr, true },
        };
#if ENABLE(NOTIFICATIONS)
    case MessageName::WebPage_ClearNotificationPermissionState:
        return Vector<ArgumentDescription> { };
#endif
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebPage_SetViewportConfigurationViewLayoutSize:
        return Vector<ArgumentDescription> {
            { "size", "WebCore::FloatSize", nullptr, false },
            { "scaleFactor", "double", nullptr, false },
            { "minimumEffectiveDeviceWidth", "double", nullptr, false },
        };
    case MessageName::WebPage_SetDeviceOrientation:
        return Vector<ArgumentDescription> {
            { "deviceOrientation", "int32_t", nullptr, false },
        };
    case MessageName::WebPage_SetOverrideViewportArguments:
        return Vector<ArgumentDescription> {
            { "arguments", "WebCore::ViewportArguments", nullptr, true },
        };
    case MessageName::WebPage_DynamicViewportSizeUpdate:
        return Vector<ArgumentDescription> {
            { "viewLayoutSize", "WebCore::FloatSize", nullptr, false },
            { "minimumUnobscuredSize", "WebCore::FloatSize", nullptr, false },
            { "maximumUnobscuredSize", "WebCore::FloatSize", nullptr, false },
            { "targetExposedContentRect", "WebCore::FloatRect", nullptr, false },
            { "targetUnobscuredRect", "WebCore::FloatRect", nullptr, false },
            { "targetUnobscuredRectInScrollViewCoordinates", "WebCore::FloatRect", nullptr, false },
            { "targetUnobscuredSafeAreaInsets", "WebCore::RectEdges<float>", nullptr, false },
            { "scale", "double", nullptr, false },
            { "deviceOrientation", "int32_t", nullptr, false },
            { "minimumEffectiveDeviceWidth", "double", nullptr, false },
            { "dynamicViewportSizeUpdateID", "uint64_t", nullptr, false },
        };
    case MessageName::WebPage_SetScreenIsBeingCaptured:
        return Vector<ArgumentDescription> {
            { "captured", "bool", nullptr, false },
        };
    case MessageName::WebPage_AttemptSyntheticClick:
        return Vector<ArgumentDescription> {
            { "point", "WebCore::IntPoint", nullptr, false },
            { "modifiers", "OptionSet<WebKit::WebEventModifier>", nullptr, false },
            { "lastLayerTreeTransactionId", "WebKit::TransactionID", nullptr, false },
        };
    case MessageName::WebPage_PotentialTapAtPosition:
        return Vector<ArgumentDescription> {
            { "requestID", "WebKit::TapIdentifier", nullptr, false },
            { "point", "WebCore::FloatPoint", nullptr, false },
            { "shouldRequestMagnificationInformation", "bool", nullptr, false },
        };
    case MessageName::WebPage_CommitPotentialTap:
        return Vector<ArgumentDescription> {
            { "modifiers", "OptionSet<WebKit::WebEventModifier>", nullptr, false },
            { "lastLayerTreeTransactionId", "WebKit::TransactionID", nullptr, false },
            { "pointerId", "WebCore::PointerID", nullptr, false },
        };
    case MessageName::WebPage_CancelPotentialTap:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_TapHighlightAtPosition:
        return Vector<ArgumentDescription> {
            { "requestID", "WebKit::TapIdentifier", nullptr, false },
            { "point", "WebCore::FloatPoint", nullptr, false },
        };
    case MessageName::WebPage_DidRecognizeLongPress:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_HandleDoubleTapForDoubleClickAtPoint:
        return Vector<ArgumentDescription> {
            { "point", "WebCore::IntPoint", nullptr, false },
            { "modifiers", "OptionSet<WebKit::WebEventModifier>", nullptr, false },
            { "lastLayerTreeTransactionId", "WebKit::TransactionID", nullptr, false },
        };
    case MessageName::WebPage_InspectorNodeSearchMovedToPosition:
        return Vector<ArgumentDescription> {
            { "point", "WebCore::FloatPoint", nullptr, false },
        };
    case MessageName::WebPage_InspectorNodeSearchEndedAtPosition:
        return Vector<ArgumentDescription> {
            { "point", "WebCore::FloatPoint", nullptr, false },
        };
    case MessageName::WebPage_BlurFocusedElement:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_SelectWithGesture:
        return Vector<ArgumentDescription> {
            { "point", "WebCore::IntPoint", nullptr, false },
            { "gestureType", "uint8_t", "WebKit::GestureType", false },
            { "gestureState", "uint8_t", "WebKit::GestureRecognizerState", false },
            { "isInteractingWithFocusedElement", "bool", nullptr, false },
        };
    case MessageName::WebPage_UpdateSelectionWithTouches:
        return Vector<ArgumentDescription> {
            { "point", "WebCore::IntPoint", nullptr, false },
            { "touches", "uint8_t", "WebKit::SelectionTouch", false },
            { "baseIsStart", "bool", nullptr, false },
        };
    case MessageName::WebPage_SelectWithTwoTouches:
        return Vector<ArgumentDescription> {
            { "from", "WebCore::IntPoint", nullptr, false },
            { "to", "WebCore::IntPoint", nullptr, false },
            { "gestureType", "uint8_t", "WebKit::GestureType", false },
            { "gestureState", "uint8_t", "WebKit::GestureRecognizerState", false },
        };
    case MessageName::WebPage_ExtendSelection:
        return Vector<ArgumentDescription> {
            { "granularity", "uint8_t", "WebCore::TextGranularity", false },
        };
    case MessageName::WebPage_SelectWordBackward:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_ExtendSelectionForReplacement:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_MoveSelectionByOffset:
        return Vector<ArgumentDescription> {
            { "offset", "int32_t", nullptr, false },
        };
    case MessageName::WebPage_SelectTextWithGranularityAtPoint:
        return Vector<ArgumentDescription> {
            { "point", "WebCore::IntPoint", nullptr, false },
            { "granularity", "uint8_t", "WebCore::TextGranularity", false },
            { "isInteractingWithFocusedElement", "bool", nullptr, false },
        };
    case MessageName::WebPage_SelectPositionAtBoundaryWithDirection:
        return Vector<ArgumentDescription> {
            { "point", "WebCore::IntPoint", nullptr, false },
            { "granularity", "uint8_t", "WebCore::TextGranularity", false },
            { "direction", "uint8_t", "WebCore::SelectionDirection", false },
            { "isInteractingWithFocusedElement", "bool", nullptr, false },
        };
    case MessageName::WebPage_MoveSelectionAtBoundaryWithDirection:
        return Vector<ArgumentDescription> {
            { "granularity", "uint8_t", "WebCore::TextGranularity", false },
            { "direction", "uint8_t", "WebCore::SelectionDirection", false },
        };
    case MessageName::WebPage_SelectPositionAtPoint:
        return Vector<ArgumentDescription> {
            { "point", "WebCore::IntPoint", nullptr, false },
            { "isInteractingWithFocusedElement", "bool", nullptr, false },
        };
    case MessageName::WebPage_BeginSelectionInDirection:
        return Vector<ArgumentDescription> {
            { "direction", "uint8_t", "WebCore::SelectionDirection", false },
        };
    case MessageName::WebPage_UpdateSelectionWithExtentPoint:
        return Vector<ArgumentDescription> {
            { "point", "WebCore::IntPoint", nullptr, false },
            { "isInteractingWithFocusedElement", "bool", nullptr, false },
            { "respectSelectionAnchor", "bool", "WebKit::RespectSelectionAnchor", false },
        };
    case MessageName::WebPage_UpdateSelectionWithExtentPointAndBoundary:
        return Vector<ArgumentDescription> {
            { "point", "WebCore::IntPoint", nullptr, false },
            { "granularity", "uint8_t", "WebCore::TextGranularity", false },
            { "isInteractingWithFocusedElement", "bool", nullptr, false },
        };
    case MessageName::WebPage_RequestDictationContext:
        return Vector<ArgumentDescription> { };
#endif
#if (PLATFORM(IOS_FAMILY) && ENABLE(REVEAL))
    case MessageName::WebPage_RequestRVItemInCurrentSelectedRange:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_PrepareSelectionForContextMenuWithLocationInView:
        return Vector<ArgumentDescription> {
            { "point", "WebCore::IntPoint", nullptr, false },
        };
#endif
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebPage_WillInsertFinalDictationResult:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_DidInsertFinalDictationResult:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_ReplaceDictatedText:
        return Vector<ArgumentDescription> {
            { "oldText", "String", nullptr, false },
            { "newText", "String", nullptr, false },
        };
    case MessageName::WebPage_ReplaceSelectedText:
        return Vector<ArgumentDescription> {
            { "oldText", "String", nullptr, false },
            { "newText", "String", nullptr, false },
        };
    case MessageName::WebPage_RequestAutocorrectionData:
        return Vector<ArgumentDescription> {
            { "textForAutocorrection", "String", nullptr, false },
        };
    case MessageName::WebPage_ApplyAutocorrection:
        return Vector<ArgumentDescription> {
            { "correction", "String", nullptr, false },
            { "originalText", "String", nullptr, false },
        };
    case MessageName::WebPage_SyncApplyAutocorrection:
        return Vector<ArgumentDescription> {
            { "correction", "String", nullptr, false },
            { "originalText", "String", nullptr, false },
        };
    case MessageName::WebPage_HandleAutocorrectionContextRequest:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_RequestEvasionRectsAboveSelection:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_RequestPositionInformation:
        return Vector<ArgumentDescription> {
            { "request", "WebKit::InteractionInformationRequest", nullptr, false },
        };
    case MessageName::WebPage_StartInteractionWithElementContextOrPosition:
        return Vector<ArgumentDescription> {
            { "elementContext", "WebCore::ElementContext", nullptr, true },
            { "point", "WebCore::IntPoint", nullptr, false },
        };
    case MessageName::WebPage_StopInteraction:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_PerformActionOnElement:
        return Vector<ArgumentDescription> {
            { "action", "uint32_t", nullptr, false },
            { "authenticationToken", "String", nullptr, false },
        };
    case MessageName::WebPage_FocusNextFocusedElement:
        return Vector<ArgumentDescription> {
            { "isForward", "bool", nullptr, false },
        };
    case MessageName::WebPage_AutofillLoginCredentials:
        return Vector<ArgumentDescription> {
            { "username", "String", nullptr, false },
            { "password", "String", nullptr, false },
        };
    case MessageName::WebPage_SetFocusedElementValue:
        return Vector<ArgumentDescription> {
            { "context", "WebCore::ElementContext", nullptr, false },
            { "value", "String", nullptr, false },
        };
    case MessageName::WebPage_SetFocusedElementSelectedIndex:
        return Vector<ArgumentDescription> {
            { "context", "WebCore::ElementContext", nullptr, false },
            { "index", "uint32_t", nullptr, false },
            { "allowMultipleSelection", "bool", nullptr, false },
        };
    case MessageName::WebPage_ApplicationWillResignActive:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_ApplicationDidEnterBackground:
        return Vector<ArgumentDescription> {
            { "isSuspendedUnderLock", "bool", nullptr, false },
        };
    case MessageName::WebPage_ApplicationDidFinishSnapshottingAfterEnteringBackground:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_ApplicationWillEnterForeground:
        return Vector<ArgumentDescription> {
            { "isSuspendedUnderLock", "bool", nullptr, false },
        };
    case MessageName::WebPage_ApplicationDidBecomeActive:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_ApplicationDidEnterBackgroundForMedia:
        return Vector<ArgumentDescription> {
            { "isSuspendedUnderLock", "bool", nullptr, false },
        };
    case MessageName::WebPage_ApplicationWillEnterForegroundForMedia:
        return Vector<ArgumentDescription> {
            { "isSuspendedUnderLock", "bool", nullptr, false },
        };
    case MessageName::WebPage_ContentSizeCategoryDidChange:
        return Vector<ArgumentDescription> {
            { "contentSizeCategory", "String", nullptr, false },
        };
    case MessageName::WebPage_GetSelectionContext:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_SetAllowsMediaDocumentInlinePlayback:
        return Vector<ArgumentDescription> {
            { "allows", "bool", nullptr, false },
        };
    case MessageName::WebPage_HandleTwoFingerTapAtPoint:
        return Vector<ArgumentDescription> {
            { "point", "WebCore::IntPoint", nullptr, false },
            { "modifiers", "OptionSet<WebKit::WebEventModifier>", nullptr, false },
            { "requestID", "WebKit::TapIdentifier", nullptr, false },
        };
    case MessageName::WebPage_SetForceAlwaysUserScalable:
        return Vector<ArgumentDescription> {
            { "userScalable", "bool", nullptr, false },
        };
    case MessageName::WebPage_GetRectsForGranularityWithSelectionOffset:
        return Vector<ArgumentDescription> {
            { "granularity", "uint8_t", "WebCore::TextGranularity", false },
            { "offset", "int32_t", nullptr, false },
        };
    case MessageName::WebPage_GetRectsAtSelectionOffsetWithText:
        return Vector<ArgumentDescription> {
            { "offset", "int32_t", nullptr, false },
            { "text", "String", nullptr, false },
        };
    case MessageName::WebPage_StoreSelectionForAccessibility:
        return Vector<ArgumentDescription> {
            { "shouldStore", "bool", nullptr, false },
        };
    case MessageName::WebPage_StartAutoscrollAtPosition:
        return Vector<ArgumentDescription> {
            { "positionInWindow", "WebCore::FloatPoint", nullptr, false },
        };
    case MessageName::WebPage_CancelAutoscroll:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_RequestFocusedElementInformation:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_HardwareKeyboardAvailabilityChanged:
        return Vector<ArgumentDescription> {
            { "keyboardIsAttached", "bool", nullptr, false },
        };
    case MessageName::WebPage_SetIsShowingInputViewForFocusedElement:
        return Vector<ArgumentDescription> {
            { "showingInputView", "bool", nullptr, false },
        };
    case MessageName::WebPage_UpdateSelectionWithDelta:
        return Vector<ArgumentDescription> {
            { "locationDelta", "int64_t", nullptr, false },
            { "lengthDelta", "int64_t", nullptr, false },
        };
    case MessageName::WebPage_RequestDocumentEditingContext:
        return Vector<ArgumentDescription> {
            { "request", "WebKit::DocumentEditingContextRequest", nullptr, false },
        };
    case MessageName::WebPage_GenerateSyntheticEditingCommand:
        return Vector<ArgumentDescription> {
            { "command", "uint8_t", "WebKit::SyntheticEditingCommandType", false },
        };
    case MessageName::WebPage_SetShouldRevealCurrentSelectionAfterInsertion:
        return Vector<ArgumentDescription> {
            { "shouldRevealCurrentSelectionAfterInsertion", "bool", nullptr, false },
        };
    case MessageName::WebPage_InsertTextPlaceholder:
        return Vector<ArgumentDescription> {
            { "size", "WebCore::IntSize", nullptr, false },
        };
    case MessageName::WebPage_RemoveTextPlaceholder:
        return Vector<ArgumentDescription> {
            { "placeholder", "WebCore::ElementContext", nullptr, false },
        };
    case MessageName::WebPage_TextInputContextsInRect:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::FloatRect", nullptr, false },
        };
    case MessageName::WebPage_FocusTextInputContextAndPlaceCaret:
        return Vector<ArgumentDescription> {
            { "context", "WebCore::ElementContext", nullptr, false },
            { "point", "WebCore::IntPoint", nullptr, false },
        };
    case MessageName::WebPage_ClearServiceWorkerEntitlementOverride:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebPage_RequestImageBitmap:
        return Vector<ArgumentDescription> {
            { "elementContext", "WebCore::ElementContext", nullptr, false },
        };
    case MessageName::WebPage_SetControlledByAutomation:
        return Vector<ArgumentDescription> {
            { "controlled", "bool", nullptr, false },
        };
    case MessageName::WebPage_ConnectInspector:
        return Vector<ArgumentDescription> {
            { "targetId", "String", nullptr, false },
            { "connectionType", "Inspector::FrontendChannel::ConnectionType", nullptr, false },
        };
    case MessageName::WebPage_DisconnectInspector:
        return Vector<ArgumentDescription> {
            { "targetId", "String", nullptr, false },
        };
    case MessageName::WebPage_SendMessageToTargetBackend:
        return Vector<ArgumentDescription> {
            { "targetId", "String", nullptr, false },
            { "message", "String", nullptr, false },
        };
#if ENABLE(REMOTE_INSPECTOR)
    case MessageName::WebPage_SetIndicating:
        return Vector<ArgumentDescription> {
            { "indicating", "bool", nullptr, false },
        };
#endif
#if ENABLE(IOS_TOUCH_EVENTS)
    case MessageName::WebPage_ResetPotentialTapSecurityOrigin:
        return Vector<ArgumentDescription> { };
#endif
#if !ENABLE(IOS_TOUCH_EVENTS) && ENABLE(TOUCH_EVENTS)
    case MessageName::WebPage_TouchEvent:
        return Vector<ArgumentDescription> {
            { "event", "WebKit::WebTouchEvent", nullptr, false },
        };
#endif
    case MessageName::WebPage_CancelPointer:
        return Vector<ArgumentDescription> {
            { "pointerId", "WebCore::PointerID", nullptr, false },
            { "documentPoint", "WebCore::IntPoint", nullptr, false },
        };
    case MessageName::WebPage_TouchWithIdentifierWasRemoved:
        return Vector<ArgumentDescription> {
            { "pointerId", "WebCore::PointerID", nullptr, false },
        };
#if ENABLE(INPUT_TYPE_COLOR)
    case MessageName::WebPage_DidEndColorPicker:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_DidChooseColor:
        return Vector<ArgumentDescription> {
            { "color", "WebCore::Color", nullptr, false },
        };
#endif
#if ENABLE(DATALIST_ELEMENT)
    case MessageName::WebPage_DidSelectDataListOption:
        return Vector<ArgumentDescription> {
            { "selectedOption", "String", nullptr, false },
        };
    case MessageName::WebPage_DidCloseSuggestions:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(DATE_AND_TIME_INPUT_TYPES)
    case MessageName::WebPage_DidChooseDate:
        return Vector<ArgumentDescription> {
            { "date", "String", nullptr, false },
        };
    case MessageName::WebPage_DidEndDateTimePicker:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebPage_ScrollBy:
        return Vector<ArgumentDescription> {
            { "scrollDirection", "uint32_t", nullptr, false },
            { "scrollGranularity", "uint8_t", "WebCore::ScrollGranularity", false },
        };
    case MessageName::WebPage_CenterSelectionInVisibleArea:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_GoToBackForwardItem:
        return Vector<ArgumentDescription> {
            { "navigationID", "uint64_t", nullptr, false },
            { "backForwardItemID", "WebCore::BackForwardItemIdentifier", nullptr, false },
            { "backForwardType", "uint8_t", "WebCore::FrameLoadType", false },
            { "shouldTreatAsContinuingLoad", "uint8_t", "WebCore::ShouldTreatAsContinuingLoad", false },
            { "websitePolicies", "WebKit::WebsitePoliciesData", nullptr, true },
            { "lastNavigationWasAppInitiated", "bool", nullptr, false },
            { "existingNetworkResourceLoadIdentifierToResume", "WebKit::NetworkResourceLoadIdentifier", nullptr, true },
            { "topPrivatelyControlledDomain", "String", nullptr, true },
        };
    case MessageName::WebPage_TryRestoreScrollPosition:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_LoadURLInFrame:
        return Vector<ArgumentDescription> {
            { "url", "URL", nullptr, false },
            { "referrer", "String", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
        };
    case MessageName::WebPage_LoadDataInFrame:
        return Vector<ArgumentDescription> {
            { "data", "IPC::DataReference", nullptr, false },
            { "MIMEType", "String", nullptr, false },
            { "encodingName", "String", nullptr, false },
            { "baseURL", "URL", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
        };
    case MessageName::WebPage_LoadRequest:
        return Vector<ArgumentDescription> {
            { "loadParameters", "WebKit::LoadParameters", nullptr, false },
        };
    case MessageName::WebPage_LoadRequestWaitingForProcessLaunch:
        return Vector<ArgumentDescription> {
            { "loadParameters", "WebKit::LoadParameters", nullptr, false },
            { "resourceDirectoryURL", "URL", nullptr, false },
            { "pageID", "WebKit::WebPageProxyIdentifier", nullptr, false },
            { "checkAssumedReadAccessToResourceURL", "bool", nullptr, false },
        };
    case MessageName::WebPage_LoadData:
        return Vector<ArgumentDescription> {
            { "loadParameters", "WebKit::LoadParameters", nullptr, false },
        };
    case MessageName::WebPage_LoadSimulatedRequestAndResponse:
        return Vector<ArgumentDescription> {
            { "loadParameters", "WebKit::LoadParameters", nullptr, false },
            { "simulatedResponse", "WebCore::ResourceResponse", nullptr, false },
        };
    case MessageName::WebPage_LoadAlternateHTML:
        return Vector<ArgumentDescription> {
            { "loadParameters", "WebKit::LoadParameters", nullptr, false },
        };
    case MessageName::WebPage_NavigateToPDFLinkWithSimulatedClick:
        return Vector<ArgumentDescription> {
            { "url", "String", nullptr, false },
            { "documentPoint", "WebCore::IntPoint", nullptr, false },
            { "screenPoint", "WebCore::IntPoint", nullptr, false },
        };
    case MessageName::WebPage_GetPDFFirstPageSize:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
        };
    case MessageName::WebPage_Reload:
        return Vector<ArgumentDescription> {
            { "navigationID", "uint64_t", nullptr, false },
            { "reloadOptions", "OptionSet<WebCore::ReloadOption>", nullptr, false },
            { "sandboxExtensionHandle", "WebKit::SandboxExtension::Handle", nullptr, false },
        };
    case MessageName::WebPage_StopLoading:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_StopLoadingDueToProcessSwap:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_RestoreSession:
        return Vector<ArgumentDescription> {
            { "itemStates", "Vector<WebKit::BackForwardListItemState>", nullptr, false },
        };
    case MessageName::WebPage_UpdateBackForwardListForReattach:
        return Vector<ArgumentDescription> {
            { "itemStates", "Vector<WebKit::BackForwardListItemState>", nullptr, false },
        };
    case MessageName::WebPage_SetCurrentHistoryItemForReattach:
        return Vector<ArgumentDescription> {
            { "itemState", "WebKit::BackForwardListItemState", nullptr, false },
        };
    case MessageName::WebPage_DidRemoveBackForwardItem:
        return Vector<ArgumentDescription> {
            { "backForwardItemID", "WebCore::BackForwardItemIdentifier", nullptr, false },
        };
    case MessageName::WebPage_UpdateWebsitePolicies:
        return Vector<ArgumentDescription> {
            { "websitePolicies", "WebKit::WebsitePoliciesData", nullptr, false },
        };
    case MessageName::WebPage_NotifyUserScripts:
        return Vector<ArgumentDescription> { };
#if ENABLE(CONTENT_FILTERING_IN_NETWORKING_PROCESS)
    case MessageName::WebPage_DidReceivePolicyDecision:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "listenerID", "uint64_t", nullptr, false },
            { "policyDecision", "WebKit::PolicyDecision", nullptr, false },
        };
#endif
#if !ENABLE(CONTENT_FILTERING_IN_NETWORKING_PROCESS)
    case MessageName::WebPage_DidReceivePolicyDecision:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "listenerID", "uint64_t", nullptr, false },
            { "policyDecision", "WebKit::PolicyDecision", nullptr, false },
            { "networkExtensionsSandboxExtensions", "Vector<WebKit::SandboxExtension::Handle>", nullptr, false },
        };
#endif
    case MessageName::WebPage_ClearSelection:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_RestoreSelectionInFocusedEditableElement:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_GetContentsAsString:
        return Vector<ArgumentDescription> {
            { "inChildFrames", "bool", "WebKit::ContentAsStringIncludesChildFrames", false },
        };
#if PLATFORM(COCOA)
    case MessageName::WebPage_GetContentsAsAttributedString:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(MHTML)
    case MessageName::WebPage_GetContentsAsMHTMLData:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebPage_GetMainResourceDataOfFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
        };
    case MessageName::WebPage_GetResourceDataFromFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "resourceURL", "String", nullptr, false },
        };
    case MessageName::WebPage_GetRenderTreeExternalRepresentation:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_GetSelectionOrContentsAsString:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_GetSelectionAsWebArchiveData:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_GetSourceForFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
        };
    case MessageName::WebPage_GetWebArchiveOfFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
        };
    case MessageName::WebPage_RunJavaScriptInFrameInScriptWorld:
        return Vector<ArgumentDescription> {
            { "parameters", "WebCore::RunJavaScriptParameters", nullptr, false },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
            { "world", "std::pair<WebKit::ContentWorldIdentifier, String>", nullptr, false },
        };
    case MessageName::WebPage_GetAccessibilityTreeData:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_ForceRepaint:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_SelectAll:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_ScheduleFullEditorStateUpdate:
        return Vector<ArgumentDescription> { };
#if PLATFORM(COCOA)
    case MessageName::WebPage_PerformDictionaryLookupOfCurrentSelection:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_PerformDictionaryLookupAtLocation:
        return Vector<ArgumentDescription> {
            { "point", "WebCore::FloatPoint", nullptr, false },
        };
#endif
#if ENABLE(DATA_DETECTION)
    case MessageName::WebPage_DetectDataInAllFrames:
        return Vector<ArgumentDescription> {
            { "types", "OptionSet<WebCore::DataDetectorType>", nullptr, false },
        };
    case MessageName::WebPage_RemoveDataDetectedLinks:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebPage_ChangeFont:
        return Vector<ArgumentDescription> {
            { "changes", "WebCore::FontChanges", nullptr, false },
        };
    case MessageName::WebPage_ChangeFontAttributes:
        return Vector<ArgumentDescription> {
            { "changes", "WebCore::FontAttributeChanges", nullptr, false },
        };
    case MessageName::WebPage_PreferencesDidChange:
        return Vector<ArgumentDescription> {
            { "store", "WebKit::WebPreferencesStore", nullptr, false },
        };
    case MessageName::WebPage_PreferencesDidChangeDuringDOMPrintOperation:
        return Vector<ArgumentDescription> {
            { "store", "WebKit::WebPreferencesStore", nullptr, false },
        };
    case MessageName::WebPage_SetUserAgent:
        return Vector<ArgumentDescription> {
            { "userAgent", "String", nullptr, false },
        };
    case MessageName::WebPage_SetCustomTextEncodingName:
        return Vector<ArgumentDescription> {
            { "encodingName", "String", nullptr, false },
        };
    case MessageName::WebPage_SuspendActiveDOMObjectsAndAnimations:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_ResumeActiveDOMObjectsAndAnimations:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_Suspend:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_Resume:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_Close:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_TryClose:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_SetEditable:
        return Vector<ArgumentDescription> {
            { "editable", "bool", nullptr, false },
        };
    case MessageName::WebPage_ValidateCommand:
        return Vector<ArgumentDescription> {
            { "name", "String", nullptr, false },
        };
    case MessageName::WebPage_ExecuteEditCommand:
        return Vector<ArgumentDescription> {
            { "name", "String", nullptr, false },
            { "argument", "String", nullptr, false },
        };
    case MessageName::WebPage_IncreaseListLevel:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_DecreaseListLevel:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_ChangeListType:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_SetBaseWritingDirection:
        return Vector<ArgumentDescription> {
            { "direction", "uint8_t", "WebCore::WritingDirection", false },
        };
    case MessageName::WebPage_SetNeedsFontAttributes:
        return Vector<ArgumentDescription> {
            { "needsFontAttributes", "bool", nullptr, false },
        };
    case MessageName::WebPage_RequestFontAttributesAtSelectionStart:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_DidRemoveEditCommand:
        return Vector<ArgumentDescription> {
            { "commandID", "uint64_t", nullptr, false },
        };
    case MessageName::WebPage_ReapplyEditCommand:
        return Vector<ArgumentDescription> {
            { "commandID", "uint64_t", nullptr, false },
        };
    case MessageName::WebPage_UnapplyEditCommand:
        return Vector<ArgumentDescription> {
            { "commandID", "uint64_t", nullptr, false },
        };
    case MessageName::WebPage_SetPageAndTextZoomFactors:
        return Vector<ArgumentDescription> {
            { "pageZoomFactor", "double", nullptr, false },
            { "textZoomFactor", "double", nullptr, false },
        };
    case MessageName::WebPage_SetPageZoomFactor:
        return Vector<ArgumentDescription> {
            { "zoomFactor", "double", nullptr, false },
        };
    case MessageName::WebPage_SetTextZoomFactor:
        return Vector<ArgumentDescription> {
            { "zoomFactor", "double", nullptr, false },
        };
    case MessageName::WebPage_WindowScreenDidChange:
        return Vector<ArgumentDescription> {
            { "displayID", "uint32_t", nullptr, false },
            { "nominalFramesPerSecond", "unsigned", nullptr, true },
        };
    case MessageName::WebPage_AccessibilitySettingsDidChange:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_ScalePage:
        return Vector<ArgumentDescription> {
            { "scale", "double", nullptr, false },
            { "origin", "WebCore::IntPoint", nullptr, false },
        };
    case MessageName::WebPage_ScalePageInViewCoordinates:
        return Vector<ArgumentDescription> {
            { "scale", "double", nullptr, false },
            { "centerInViewCoordinates", "WebCore::IntPoint", nullptr, false },
        };
    case MessageName::WebPage_ScaleView:
        return Vector<ArgumentDescription> {
            { "scale", "double", nullptr, false },
        };
    case MessageName::WebPage_SetUseFixedLayout:
        return Vector<ArgumentDescription> {
            { "fixed", "bool", nullptr, false },
        };
    case MessageName::WebPage_SetFixedLayoutSize:
        return Vector<ArgumentDescription> {
            { "size", "WebCore::IntSize", nullptr, false },
        };
    case MessageName::WebPage_SetDefaultUnobscuredSize:
        return Vector<ArgumentDescription> {
            { "size", "WebCore::FloatSize", nullptr, false },
        };
    case MessageName::WebPage_SetMinimumUnobscuredSize:
        return Vector<ArgumentDescription> {
            { "size", "WebCore::FloatSize", nullptr, false },
        };
    case MessageName::WebPage_SetMaximumUnobscuredSize:
        return Vector<ArgumentDescription> {
            { "size", "WebCore::FloatSize", nullptr, false },
        };
    case MessageName::WebPage_ListenForLayoutMilestones:
        return Vector<ArgumentDescription> {
            { "layoutMilestones", "OptionSet<WebCore::LayoutMilestone>", nullptr, false },
        };
    case MessageName::WebPage_SetSuppressScrollbarAnimations:
        return Vector<ArgumentDescription> {
            { "suppressAnimations", "bool", nullptr, false },
        };
    case MessageName::WebPage_SetEnableVerticalRubberBanding:
        return Vector<ArgumentDescription> {
            { "enableVerticalRubberBanding", "bool", nullptr, false },
        };
    case MessageName::WebPage_SetEnableHorizontalRubberBanding:
        return Vector<ArgumentDescription> {
            { "enableHorizontalRubberBanding", "bool", nullptr, false },
        };
    case MessageName::WebPage_SetBackgroundExtendsBeyondPage:
        return Vector<ArgumentDescription> {
            { "backgroundExtendsBeyondPage", "bool", nullptr, false },
        };
    case MessageName::WebPage_SetPaginationMode:
        return Vector<ArgumentDescription> {
            { "mode", "uint32_t", nullptr, false },
        };
    case MessageName::WebPage_SetPaginationBehavesLikeColumns:
        return Vector<ArgumentDescription> {
            { "behavesLikeColumns", "bool", nullptr, false },
        };
    case MessageName::WebPage_SetPageLength:
        return Vector<ArgumentDescription> {
            { "pageLength", "double", nullptr, false },
        };
    case MessageName::WebPage_SetGapBetweenPages:
        return Vector<ArgumentDescription> {
            { "gap", "double", nullptr, false },
        };
    case MessageName::WebPage_PostInjectedBundleMessage:
        return Vector<ArgumentDescription> {
            { "messageName", "String", nullptr, false },
            { "messageBody", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebPage_FindString:
        return Vector<ArgumentDescription> {
            { "string", "String", nullptr, false },
            { "findOptions", "OptionSet<WebKit::FindOptions>", nullptr, false },
            { "maxMatchCount", "unsigned", nullptr, false },
        };
    case MessageName::WebPage_FindStringMatches:
        return Vector<ArgumentDescription> {
            { "string", "String", nullptr, false },
            { "findOptions", "OptionSet<WebKit::FindOptions>", nullptr, false },
            { "maxMatchCount", "unsigned", nullptr, false },
        };
    case MessageName::WebPage_GetImageForFindMatch:
        return Vector<ArgumentDescription> {
            { "matchIndex", "uint32_t", nullptr, false },
        };
    case MessageName::WebPage_SelectFindMatch:
        return Vector<ArgumentDescription> {
            { "matchIndex", "uint32_t", nullptr, false },
        };
    case MessageName::WebPage_IndicateFindMatch:
        return Vector<ArgumentDescription> {
            { "matchIndex", "uint32_t", nullptr, false },
        };
    case MessageName::WebPage_HideFindUI:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_CountStringMatches:
        return Vector<ArgumentDescription> {
            { "string", "String", nullptr, false },
            { "findOptions", "OptionSet<WebKit::FindOptions>", nullptr, false },
            { "maxMatchCount", "unsigned", nullptr, false },
        };
    case MessageName::WebPage_ReplaceMatches:
        return Vector<ArgumentDescription> {
            { "matchIndices", "Vector<uint32_t>", nullptr, false },
            { "replacementText", "String", nullptr, false },
            { "selectionOnly", "bool", nullptr, false },
        };
    case MessageName::WebPage_FindRectsForStringMatches:
        return Vector<ArgumentDescription> {
            { "string", "String", nullptr, false },
            { "findOptions", "OptionSet<WebKit::FindOptions>", nullptr, false },
            { "maxMatchCount", "unsigned", nullptr, false },
        };
    case MessageName::WebPage_HideFindIndicator:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_FindTextRangesForStringMatches:
        return Vector<ArgumentDescription> {
            { "string", "String", nullptr, false },
            { "findOptions", "OptionSet<WebKit::FindOptions>", nullptr, false },
            { "maxMatchCount", "unsigned", nullptr, false },
        };
    case MessageName::WebPage_ReplaceFoundTextRangeWithString:
        return Vector<ArgumentDescription> {
            { "range", "WebKit::WebFoundTextRange", nullptr, false },
            { "string", "String", nullptr, false },
        };
    case MessageName::WebPage_DecorateTextRangeWithStyle:
        return Vector<ArgumentDescription> {
            { "range", "WebKit::WebFoundTextRange", nullptr, false },
            { "style", "uint8_t", "WebKit::FindDecorationStyle", false },
        };
    case MessageName::WebPage_ScrollTextRangeToVisible:
        return Vector<ArgumentDescription> {
            { "range", "WebKit::WebFoundTextRange", nullptr, false },
        };
    case MessageName::WebPage_ClearAllDecoratedFoundText:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_DidBeginTextSearchOperation:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_DidEndTextSearchOperation:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_RequestRectForFoundTextRange:
        return Vector<ArgumentDescription> {
            { "range", "WebKit::WebFoundTextRange", nullptr, false },
        };
    case MessageName::WebPage_AddLayerForFindOverlay:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_RemoveLayerForFindOverlay:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_AddMIMETypeWithCustomContentProvider:
        return Vector<ArgumentDescription> {
            { "mimeType", "String", nullptr, false },
        };
#if PLATFORM(GTK) && ENABLE(DRAG_SUPPORT)
    case MessageName::WebPage_PerformDragControllerAction:
        return Vector<ArgumentDescription> {
            { "action", "uint8_t", "WebKit::DragControllerAction", false },
            { "clientPosition", "WebCore::IntPoint", nullptr, false },
            { "globalPosition", "WebCore::IntPoint", nullptr, false },
            { "draggingSourceOperationMask", "OptionSet<WebCore::DragOperation>", nullptr, false },
            { "selection", "WebCore::SelectionData", nullptr, false },
            { "flags", "OptionSet<WebCore::DragApplicationFlags>", nullptr, false },
        };
#endif
#if !PLATFORM(GTK) && ENABLE(DRAG_SUPPORT)
    case MessageName::WebPage_PerformDragControllerAction:
        return Vector<ArgumentDescription> {
            { "action", "uint8_t", "WebKit::DragControllerAction", false },
            { "dragData", "WebCore::DragData", nullptr, false },
            { "sandboxExtensionHandle", "WebKit::SandboxExtension::Handle", nullptr, false },
            { "sandboxExtensionsForUpload", "Vector<WebKit::SandboxExtension::Handle>", nullptr, false },
        };
#endif
#if ENABLE(DRAG_SUPPORT)
    case MessageName::WebPage_DidStartDrag:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_DragEnded:
        return Vector<ArgumentDescription> {
            { "clientPosition", "WebCore::IntPoint", nullptr, false },
            { "globalPosition", "WebCore::IntPoint", nullptr, false },
            { "dragOperationMask", "OptionSet<WebCore::DragOperation>", nullptr, false },
        };
    case MessageName::WebPage_DragCancelled:
        return Vector<ArgumentDescription> { };
#endif
#if PLATFORM(IOS_FAMILY) && ENABLE(DRAG_SUPPORT)
    case MessageName::WebPage_RequestDragStart:
        return Vector<ArgumentDescription> {
            { "clientPosition", "WebCore::IntPoint", nullptr, false },
            { "globalPosition", "WebCore::IntPoint", nullptr, false },
            { "allowedActionsMask", "OptionSet<WebCore::DragSourceAction>", nullptr, false },
        };
    case MessageName::WebPage_RequestAdditionalItemsForDragSession:
        return Vector<ArgumentDescription> {
            { "clientPosition", "WebCore::IntPoint", nullptr, false },
            { "globalPosition", "WebCore::IntPoint", nullptr, false },
            { "allowedActionsMask", "OptionSet<WebCore::DragSourceAction>", nullptr, false },
        };
    case MessageName::WebPage_InsertDroppedImagePlaceholders:
        return Vector<ArgumentDescription> {
            { "imageSize", "Vector<WebCore::IntSize>", nullptr, false },
        };
    case MessageName::WebPage_DidConcludeDrop:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebPage_DidChangeSelectedIndexForActivePopupMenu:
        return Vector<ArgumentDescription> {
            { "newIndex", "int32_t", nullptr, false },
        };
    case MessageName::WebPage_SetTextForActivePopupMenu:
        return Vector<ArgumentDescription> {
            { "index", "int32_t", nullptr, false },
        };
#if PLATFORM(GTK)
    case MessageName::WebPage_FailedToShowPopupMenu:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(CONTEXT_MENUS)
    case MessageName::WebPage_DidShowContextMenu:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_DidDismissContextMenu:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_DidSelectItemFromActiveContextMenu:
        return Vector<ArgumentDescription> {
            { "menuItem", "WebKit::WebContextMenuItemData", nullptr, false },
        };
    case MessageName::WebPage_ContextMenuForKeyEvent:
        return Vector<ArgumentDescription> { };
#endif
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebPage_DidChooseFilesForOpenPanelWithDisplayStringAndIcon:
        return Vector<ArgumentDescription> {
            { "fileURLs", "Vector<String>", nullptr, false },
            { "displayString", "String", nullptr, false },
            { "iconData", "IPC::DataReference", nullptr, false },
            { "machBootstrapHandle", "WebKit::SandboxExtension::Handle", nullptr, false },
            { "frontboardServicesSandboxExtension", "WebKit::SandboxExtension::Handle", nullptr, false },
            { "iconServicesSandboxExtension", "WebKit::SandboxExtension::Handle", nullptr, false },
        };
#endif
    case MessageName::WebPage_DidChooseFilesForOpenPanel:
        return Vector<ArgumentDescription> {
            { "fileURLs", "Vector<String>", nullptr, false },
            { "replacementURLs", "Vector<String>", nullptr, false },
        };
    case MessageName::WebPage_DidCancelForOpenPanel:
        return Vector<ArgumentDescription> { };
#if ENABLE(SANDBOX_EXTENSIONS)
    case MessageName::WebPage_ExtendSandboxForFilesFromOpenPanel:
        return Vector<ArgumentDescription> {
            { "sandboxExtensions", "Vector<WebKit::SandboxExtension::Handle>", nullptr, false },
        };
#endif
    case MessageName::WebPage_AdvanceToNextMisspelling:
        return Vector<ArgumentDescription> {
            { "startBeforeSelection", "bool", nullptr, false },
        };
    case MessageName::WebPage_ChangeSpellingToWord:
        return Vector<ArgumentDescription> {
            { "word", "String", nullptr, false },
        };
    case MessageName::WebPage_DidFinishCheckingText:
        return Vector<ArgumentDescription> {
            { "requestID", "WebKit::TextCheckerRequestID", nullptr, false },
            { "result", "Vector<WebCore::TextCheckingResult>", nullptr, false },
        };
    case MessageName::WebPage_DidCancelCheckingText:
        return Vector<ArgumentDescription> {
            { "requestID", "WebKit::TextCheckerRequestID", nullptr, false },
        };
#if USE(APPKIT)
    case MessageName::WebPage_UppercaseWord:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_LowercaseWord:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_CapitalizeWord:
        return Vector<ArgumentDescription> { };
#endif
#if PLATFORM(COCOA)
    case MessageName::WebPage_SetSmartInsertDeleteEnabled:
        return Vector<ArgumentDescription> {
            { "isSmartInsertDeleteEnabled", "bool", nullptr, false },
        };
#endif
#if ENABLE(GEOLOCATION)
    case MessageName::WebPage_DidReceiveGeolocationPermissionDecision:
        return Vector<ArgumentDescription> {
            { "geolocationID", "WebKit::GeolocationIdentifier", nullptr, false },
            { "authorizationToken", "String", nullptr, false },
        };
#endif
#if ENABLE(MEDIA_STREAM)
    case MessageName::WebPage_UserMediaAccessWasGranted:
        return Vector<ArgumentDescription> {
            { "userMediaID", "WebCore::UserMediaRequestIdentifier", nullptr, false },
            { "audioDevice", "WebCore::CaptureDevice", nullptr, false },
            { "videoDevice", "WebCore::CaptureDevice", nullptr, false },
            { "mediaDeviceIdentifierHashSalts", "WebCore::MediaDeviceHashSalts", nullptr, false },
            { "sandboxExtensionHandles", "Vector<WebKit::SandboxExtension::Handle>", nullptr, false },
        };
    case MessageName::WebPage_UserMediaAccessWasDenied:
        return Vector<ArgumentDescription> {
            { "userMediaID", "WebCore::UserMediaRequestIdentifier", nullptr, false },
            { "reason", "uint64_t", nullptr, false },
            { "invalidConstraint", "String", nullptr, false },
        };
    case MessageName::WebPage_CaptureDevicesChanged:
        return Vector<ArgumentDescription> { };
#endif
#if (ENABLE(MEDIA_STREAM) && USE(GSTREAMER))
    case MessageName::WebPage_SetOrientationForMediaCapture:
        return Vector<ArgumentDescription> {
            { "rotation", "uint64_t", nullptr, false },
        };
    case MessageName::WebPage_SetMockCaptureDevicesInterrupted:
        return Vector<ArgumentDescription> {
            { "isCameraInterrupted", "bool", nullptr, false },
            { "isMicrophoneInterrupted", "bool", nullptr, false },
        };
#endif
#if ENABLE(ENCRYPTED_MEDIA)
    case MessageName::WebPage_MediaKeySystemWasGranted:
        return Vector<ArgumentDescription> {
            { "mediaKeySystemID", "WebCore::MediaKeySystemRequestIdentifier", nullptr, false },
        };
    case MessageName::WebPage_MediaKeySystemWasDenied:
        return Vector<ArgumentDescription> {
            { "mediaKeySystemID", "WebCore::MediaKeySystemRequestIdentifier", nullptr, false },
            { "message", "String", nullptr, false },
        };
#endif
    case MessageName::WebPage_RequestMediaPlaybackState:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_PauseAllMediaPlayback:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_SuspendAllMediaPlayback:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_ResumeAllMediaPlayback:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_FreezeLayerTreeDueToSwipeAnimation:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_UnfreezeLayerTreeDueToSwipeAnimation:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_IsLayerTreeFrozen:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_BeginPrinting:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "printInfo", "WebKit::PrintInfo", nullptr, false },
        };
    case MessageName::WebPage_BeginPrintingDuringDOMPrintOperation:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "printInfo", "WebKit::PrintInfo", nullptr, false },
        };
    case MessageName::WebPage_EndPrinting:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_EndPrintingDuringDOMPrintOperation:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_ComputePagesForPrinting:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "printInfo", "WebKit::PrintInfo", nullptr, false },
        };
    case MessageName::WebPage_ComputePagesForPrintingDuringDOMPrintOperation:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "printInfo", "WebKit::PrintInfo", nullptr, false },
        };
#if PLATFORM(COCOA)
    case MessageName::WebPage_DrawRectToImage:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "printInfo", "WebKit::PrintInfo", nullptr, false },
            { "rect", "WebCore::IntRect", nullptr, false },
            { "imageSize", "WebCore::IntSize", nullptr, false },
        };
    case MessageName::WebPage_DrawRectToImageDuringDOMPrintOperation:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "printInfo", "WebKit::PrintInfo", nullptr, false },
            { "rect", "WebCore::IntRect", nullptr, false },
            { "imageSize", "WebCore::IntSize", nullptr, false },
        };
    case MessageName::WebPage_DrawPagesToPDF:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "printInfo", "WebKit::PrintInfo", nullptr, false },
            { "first", "uint32_t", nullptr, false },
            { "count", "uint32_t", nullptr, false },
        };
    case MessageName::WebPage_DrawPagesToPDFDuringDOMPrintOperation:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "printInfo", "WebKit::PrintInfo", nullptr, false },
            { "first", "uint32_t", nullptr, false },
            { "count", "uint32_t", nullptr, false },
        };
#endif
#if (PLATFORM(COCOA) && PLATFORM(IOS_FAMILY))
    case MessageName::WebPage_ComputePagesForPrintingiOS:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "printInfo", "WebKit::PrintInfo", nullptr, false },
        };
    case MessageName::WebPage_DrawToPDFiOS:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "printInfo", "WebKit::PrintInfo", nullptr, false },
            { "pageCount", "size_t", nullptr, false },
        };
#endif
#if PLATFORM(COCOA)
    case MessageName::WebPage_DrawToPDF:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "rect", "WebCore::FloatRect", nullptr, true },
            { "allowTransparentBackground", "bool", nullptr, false },
        };
#endif
#if PLATFORM(GTK)
    case MessageName::WebPage_DrawPagesForPrinting:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "printInfo", "WebKit::PrintInfo", nullptr, false },
        };
    case MessageName::WebPage_DrawPagesForPrintingDuringDOMPrintOperation:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
            { "printInfo", "WebKit::PrintInfo", nullptr, false },
        };
#endif
    case MessageName::WebPage_SetMediaVolume:
        return Vector<ArgumentDescription> {
            { "volume", "float", nullptr, false },
        };
    case MessageName::WebPage_SetMuted:
        return Vector<ArgumentDescription> {
            { "muted", "OptionSet<WebCore::MediaProducerMutedState>", nullptr, false },
        };
    case MessageName::WebPage_SetMayStartMediaWhenInWindow:
        return Vector<ArgumentDescription> {
            { "mayStartMedia", "bool", nullptr, false },
        };
    case MessageName::WebPage_StopMediaCapture:
        return Vector<ArgumentDescription> {
            { "kind", "uint8_t", "WebCore::MediaProducerMediaCaptureKind", false },
        };
    case MessageName::WebPage_SetCanRunBeforeUnloadConfirmPanel:
        return Vector<ArgumentDescription> {
            { "canRunBeforeUnloadConfirmPanel", "bool", nullptr, false },
        };
    case MessageName::WebPage_SetCanRunModal:
        return Vector<ArgumentDescription> {
            { "canRunModal", "bool", nullptr, false },
        };
#if PLATFORM(GTK) || PLATFORM(WPE)
    case MessageName::WebPage_CancelComposition:
        return Vector<ArgumentDescription> {
            { "text", "String", nullptr, false },
        };
    case MessageName::WebPage_DeleteSurrounding:
        return Vector<ArgumentDescription> {
            { "offset", "int64_t", nullptr, false },
            { "characterCount", "unsigned", nullptr, false },
        };
#endif
#if PLATFORM(GTK)
    case MessageName::WebPage_CollapseSelectionInFrame:
        return Vector<ArgumentDescription> {
            { "frameID", "WebCore::FrameIdentifier", nullptr, false },
        };
#endif
#if PLATFORM(COCOA)
    case MessageName::WebPage_WindowAndViewFramesChanged:
        return Vector<ArgumentDescription> {
            { "windowFrameInScreenCoordinates", "WebCore::FloatRect", nullptr, false },
            { "windowFrameInUnflippedScreenCoordinates", "WebCore::FloatRect", nullptr, false },
            { "viewFrameInWindowCoordinates", "WebCore::FloatRect", nullptr, false },
            { "accessibilityViewCoordinates", "WebCore::FloatPoint", nullptr, false },
        };
    case MessageName::WebPage_SetMainFrameIsScrollable:
        return Vector<ArgumentDescription> {
            { "isScrollable", "bool", nullptr, false },
        };
    case MessageName::WebPage_RegisterUIProcessAccessibilityTokens:
        return Vector<ArgumentDescription> {
            { "elemenToken", "IPC::DataReference", nullptr, false },
            { "windowToken", "IPC::DataReference", nullptr, false },
        };
    case MessageName::WebPage_GetStringSelectionForPasteboard:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_GetDataSelectionForPasteboard:
        return Vector<ArgumentDescription> {
            { "pasteboardType", "String", nullptr, false },
        };
    case MessageName::WebPage_ReadSelectionFromPasteboard:
        return Vector<ArgumentDescription> {
            { "pasteboardName", "String", nullptr, false },
        };
    case MessageName::WebPage_ReplaceSelectionWithPasteboardData:
        return Vector<ArgumentDescription> {
            { "types", "Vector<String>", nullptr, false },
            { "data", "IPC::DataReference", nullptr, false },
        };
    case MessageName::WebPage_ShouldDelayWindowOrderingEvent:
        return Vector<ArgumentDescription> {
            { "event", "WebKit::WebMouseEvent", nullptr, false },
        };
    case MessageName::WebPage_SetTextAsync:
        return Vector<ArgumentDescription> {
            { "text", "String", nullptr, false },
        };
    case MessageName::WebPage_InsertTextAsync:
        return Vector<ArgumentDescription> {
            { "text", "String", nullptr, false },
            { "replacementRange", "WebKit::EditingRange", nullptr, false },
            { "options", "WebKit::InsertTextOptions", nullptr, false },
        };
    case MessageName::WebPage_InsertDictatedTextAsync:
        return Vector<ArgumentDescription> {
            { "text", "String", nullptr, false },
            { "replacementRange", "WebKit::EditingRange", nullptr, false },
            { "dictationAlternatives", "Vector<WebCore::DictationAlternative>", nullptr, false },
            { "options", "WebKit::InsertTextOptions", nullptr, false },
        };
    case MessageName::WebPage_AddDictationAlternative:
        return Vector<ArgumentDescription> {
            { "text", "String", nullptr, false },
            { "context", "WebCore::DictationContext", nullptr, false },
        };
    case MessageName::WebPage_DictationAlternativesAtSelection:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_ClearDictationAlternatives:
        return Vector<ArgumentDescription> {
            { "contexts", "Vector<WebCore::DictationContext>", nullptr, false },
        };
    case MessageName::WebPage_HasMarkedText:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_GetMarkedRangeAsync:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_GetSelectedRangeAsync:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_CharacterIndexForPointAsync:
        return Vector<ArgumentDescription> {
            { "point", "WebCore::IntPoint", nullptr, false },
        };
    case MessageName::WebPage_FirstRectForCharacterRangeAsync:
        return Vector<ArgumentDescription> {
            { "range", "WebKit::EditingRange", nullptr, false },
        };
    case MessageName::WebPage_SetCompositionAsync:
        return Vector<ArgumentDescription> {
            { "text", "String", nullptr, false },
            { "underlines", "Vector<WebCore::CompositionUnderline>", nullptr, false },
            { "highlights", "Vector<WebCore::CompositionHighlight>", nullptr, false },
            { "selectionRange", "WebKit::EditingRange", nullptr, false },
            { "replacementRange", "WebKit::EditingRange", nullptr, false },
        };
    case MessageName::WebPage_ConfirmCompositionAsync:
        return Vector<ArgumentDescription> { };
#endif
#if PLATFORM(MAC)
    case MessageName::WebPage_AttributedSubstringForCharacterRangeAsync:
        return Vector<ArgumentDescription> {
            { "range", "WebKit::EditingRange", nullptr, false },
        };
    case MessageName::WebPage_RequestAcceptsFirstMouse:
        return Vector<ArgumentDescription> {
            { "eventNumber", "int", nullptr, false },
            { "event", "WebKit::WebMouseEvent", nullptr, false },
        };
#endif
#if ENABLE(IMAGE_ANALYSIS_ENHANCEMENTS)
    case MessageName::WebPage_ReplaceImageForRemoveBackground:
        return Vector<ArgumentDescription> {
            { "context", "WebCore::ElementContext", nullptr, false },
            { "types", "Vector<String>", nullptr, false },
            { "data", "IPC::DataReference", nullptr, false },
        };
    case MessageName::WebPage_ShouldAllowRemoveBackground:
        return Vector<ArgumentDescription> {
            { "context", "WebCore::ElementContext", nullptr, false },
        };
#endif
    case MessageName::WebPage_SetAlwaysShowsHorizontalScroller:
        return Vector<ArgumentDescription> {
            { "alwaysShowsHorizontalScroller", "bool", nullptr, false },
        };
    case MessageName::WebPage_SetAlwaysShowsVerticalScroller:
        return Vector<ArgumentDescription> {
            { "alwaysShowsVerticalScroller", "bool", nullptr, false },
        };
    case MessageName::WebPage_SetMinimumSizeForAutoLayout:
        return Vector<ArgumentDescription> {
            { "size", "WebCore::IntSize", nullptr, false },
        };
    case MessageName::WebPage_SetSizeToContentAutoSizeMaximumSize:
        return Vector<ArgumentDescription> {
            { "size", "WebCore::IntSize", nullptr, false },
        };
    case MessageName::WebPage_SetAutoSizingShouldExpandToViewHeight:
        return Vector<ArgumentDescription> {
            { "shouldExpand", "bool", nullptr, false },
        };
    case MessageName::WebPage_SetViewportSizeForCSSViewportUnits:
        return Vector<ArgumentDescription> {
            { "viewportSize", "WebCore::FloatSize", nullptr, true },
        };
#if PLATFORM(COCOA)
    case MessageName::WebPage_HandleAlternativeTextUIResult:
        return Vector<ArgumentDescription> {
            { "result", "String", nullptr, false },
        };
#endif
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebPage_WillStartUserTriggeredZooming:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebPage_SetScrollPinningBehavior:
        return Vector<ArgumentDescription> {
            { "pinning", "uint8_t", "WebCore::ScrollPinningBehavior", false },
        };
    case MessageName::WebPage_SetScrollbarOverlayStyle:
        return Vector<ArgumentDescription> {
            { "scrollbarStyle", "uint32_t", nullptr, true },
        };
    case MessageName::WebPage_GetBytecodeProfile:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_GetSamplingProfilerOutput:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_TakeSnapshot:
        return Vector<ArgumentDescription> {
            { "snapshotRect", "WebCore::IntRect", nullptr, false },
            { "bitmapSize", "WebCore::IntSize", nullptr, false },
            { "options", "uint32_t", nullptr, false },
        };
#if PLATFORM(MAC)
    case MessageName::WebPage_PerformImmediateActionHitTestAtLocation:
        return Vector<ArgumentDescription> {
            { "location", "WebCore::FloatPoint", nullptr, false },
        };
    case MessageName::WebPage_ImmediateActionDidUpdate:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_ImmediateActionDidCancel:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_ImmediateActionDidComplete:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_DataDetectorsDidPresentUI:
        return Vector<ArgumentDescription> {
            { "pageOverlay", "WebCore::PageOverlay::PageOverlayID", nullptr, false },
        };
    case MessageName::WebPage_DataDetectorsDidChangeUI:
        return Vector<ArgumentDescription> {
            { "pageOverlay", "WebCore::PageOverlay::PageOverlayID", nullptr, false },
        };
    case MessageName::WebPage_DataDetectorsDidHideUI:
        return Vector<ArgumentDescription> {
            { "pageOverlay", "WebCore::PageOverlay::PageOverlayID", nullptr, false },
        };
    case MessageName::WebPage_HandleAcceptedCandidate:
        return Vector<ArgumentDescription> {
            { "acceptedCandidate", "WebCore::TextCheckingResult", nullptr, false },
        };
    case MessageName::WebPage_SetUseSystemAppearance:
        return Vector<ArgumentDescription> {
            { "useSystemAppearance", "bool", nullptr, false },
        };
    case MessageName::WebPage_SemanticContextDidChange:
        return Vector<ArgumentDescription> {
            { "useFormSemanticContext", "bool", nullptr, false },
        };
    case MessageName::WebPage_SetHeaderBannerHeightForTesting:
        return Vector<ArgumentDescription> {
            { "height", "int", nullptr, false },
        };
    case MessageName::WebPage_SetFooterBannerHeightForTesting:
        return Vector<ArgumentDescription> {
            { "height", "int", nullptr, false },
        };
    case MessageName::WebPage_DidEndMagnificationGesture:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebPage_EffectiveAppearanceDidChange:
        return Vector<ArgumentDescription> {
            { "useDarkAppearance", "bool", nullptr, false },
            { "useElevatedUserInterfaceLevel", "bool", nullptr, false },
        };
#if HAVE(APP_ACCENT_COLORS)
    case MessageName::WebPage_SetAccentColor:
        return Vector<ArgumentDescription> {
            { "color", "WebCore::Color", nullptr, false },
        };
#endif
#if PLATFORM(COCOA)
    case MessageName::WebPage_RequestActiveNowPlayingSessionInfo:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebPage_SetShouldDispatchFakeMouseMoveEvents:
        return Vector<ArgumentDescription> {
            { "shouldDispatchFakeMouseMoveEvents", "bool", nullptr, false },
        };
#if ENABLE(WIRELESS_PLAYBACK_TARGET) && !PLATFORM(IOS_FAMILY)
    case MessageName::WebPage_PlaybackTargetSelected:
        return Vector<ArgumentDescription> {
            { "contextId", "WebCore::PlaybackTargetClientContextIdentifier", nullptr, false },
            { "target", "WebCore::MediaPlaybackTargetContext", nullptr, false },
        };
    case MessageName::WebPage_PlaybackTargetAvailabilityDidChange:
        return Vector<ArgumentDescription> {
            { "contextId", "WebCore::PlaybackTargetClientContextIdentifier", nullptr, false },
            { "available", "bool", nullptr, false },
        };
    case MessageName::WebPage_SetShouldPlayToPlaybackTarget:
        return Vector<ArgumentDescription> {
            { "contextId", "WebCore::PlaybackTargetClientContextIdentifier", nullptr, false },
            { "shouldPlay", "bool", nullptr, false },
        };
    case MessageName::WebPage_PlaybackTargetPickerWasDismissed:
        return Vector<ArgumentDescription> {
            { "contextId", "WebCore::PlaybackTargetClientContextIdentifier", nullptr, false },
        };
#endif
#if ENABLE(POINTER_LOCK)
    case MessageName::WebPage_DidAcquirePointerLock:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_DidNotAcquirePointerLock:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_DidLosePointerLock:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebPage_ClearWheelEventTestMonitor:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_SetShouldScaleViewToFitDocument:
        return Vector<ArgumentDescription> {
            { "shouldScaleViewToFitDocument", "bool", nullptr, false },
        };
    case MessageName::WebPage_SetUserInterfaceLayoutDirection:
        return Vector<ArgumentDescription> {
            { "direction", "uint32_t", nullptr, false },
        };
    case MessageName::WebPage_DidGetLoadDecisionForIcon:
        return Vector<ArgumentDescription> {
            { "decision", "bool", nullptr, false },
            { "loadIdentifier", "WebKit::CallbackID", nullptr, false },
        };
    case MessageName::WebPage_SetUseIconLoadingClient:
        return Vector<ArgumentDescription> {
            { "useIconLoadingClient", "bool", nullptr, false },
        };
#if ENABLE(GAMEPAD)
    case MessageName::WebPage_GamepadActivity:
        return Vector<ArgumentDescription> {
            { "gamepadDatas", "Vector<WebKit::GamepadData>", nullptr, false },
            { "eventVisibility", "bool", "WebCore::EventMakesGamepadsVisible", false },
        };
#endif
    case MessageName::WebPage_RegisterURLSchemeHandler:
        return Vector<ArgumentDescription> {
            { "handlerIdentifier", "WebKit::WebURLSchemeHandlerIdentifier", nullptr, false },
            { "scheme", "String", nullptr, false },
        };
    case MessageName::WebPage_URLSchemeTaskWillPerformRedirection:
        return Vector<ArgumentDescription> {
            { "handlerIdentifier", "WebKit::WebURLSchemeHandlerIdentifier", nullptr, false },
            { "taskIdentifier", "WebCore::ResourceLoaderIdentifier", nullptr, false },
            { "response", "WebCore::ResourceResponse", nullptr, false },
            { "proposedRequest", "WebCore::ResourceRequest", nullptr, false },
        };
    case MessageName::WebPage_URLSchemeTaskDidPerformRedirection:
        return Vector<ArgumentDescription> {
            { "handlerIdentifier", "WebKit::WebURLSchemeHandlerIdentifier", nullptr, false },
            { "taskIdentifier", "WebCore::ResourceLoaderIdentifier", nullptr, false },
            { "response", "WebCore::ResourceResponse", nullptr, false },
            { "proposedRequest", "WebCore::ResourceRequest", nullptr, false },
        };
    case MessageName::WebPage_URLSchemeTaskDidReceiveResponse:
        return Vector<ArgumentDescription> {
            { "handlerIdentifier", "WebKit::WebURLSchemeHandlerIdentifier", nullptr, false },
            { "taskIdentifier", "WebCore::ResourceLoaderIdentifier", nullptr, false },
            { "response", "WebCore::ResourceResponse", nullptr, false },
        };
    case MessageName::WebPage_URLSchemeTaskDidReceiveData:
        return Vector<ArgumentDescription> {
            { "handlerIdentifier", "WebKit::WebURLSchemeHandlerIdentifier", nullptr, false },
            { "taskIdentifier", "WebCore::ResourceLoaderIdentifier", nullptr, false },
            { "data", "Ref<WebCore::SharedBuffer>", nullptr, false },
        };
    case MessageName::WebPage_URLSchemeTaskDidComplete:
        return Vector<ArgumentDescription> {
            { "handlerIdentifier", "WebKit::WebURLSchemeHandlerIdentifier", nullptr, false },
            { "taskIdentifier", "WebCore::ResourceLoaderIdentifier", nullptr, false },
            { "error", "WebCore::ResourceError", nullptr, false },
        };
    case MessageName::WebPage_SetIsSuspended:
        return Vector<ArgumentDescription> {
            { "suspended", "bool", nullptr, false },
        };
#if ENABLE(ATTACHMENT_ELEMENT)
    case MessageName::WebPage_InsertAttachment:
        return Vector<ArgumentDescription> {
            { "identifier", "String", nullptr, false },
            { "fileSize", "uint64_t", nullptr, true },
            { "fileName", "String", nullptr, false },
            { "contentType", "String", nullptr, false },
        };
    case MessageName::WebPage_UpdateAttachmentAttributes:
        return Vector<ArgumentDescription> {
            { "identifier", "String", nullptr, false },
            { "fileSize", "uint64_t", nullptr, true },
            { "contentType", "String", nullptr, false },
            { "fileName", "String", nullptr, false },
            { "enclosingImageData", "IPC::SharedBufferReference", nullptr, false },
        };
    case MessageName::WebPage_UpdateAttachmentThumbnail:
        return Vector<ArgumentDescription> {
            { "identifier", "String", nullptr, false },
            { "qlThumbnailHandle", "WebKit::ShareableBitmapHandle", nullptr, false },
        };
    case MessageName::WebPage_UpdateAttachmentIcon:
        return Vector<ArgumentDescription> {
            { "identifier", "String", nullptr, false },
            { "icon", "WebKit::ShareableBitmapHandle", nullptr, false },
            { "size", "WebCore::FloatSize", nullptr, false },
        };
#endif
#if ENABLE(APPLICATION_MANIFEST)
    case MessageName::WebPage_GetApplicationManifest:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebPage_GetTextFragmentMatch:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_SetDefersLoading:
        return Vector<ArgumentDescription> {
            { "defersLoading", "bool", nullptr, false },
        };
#if ENABLE(UI_PROCESS_PDF_HUD)
    case MessageName::WebPage_ZoomPDFIn:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::PDFPluginIdentifier", nullptr, false },
        };
    case MessageName::WebPage_ZoomPDFOut:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::PDFPluginIdentifier", nullptr, false },
        };
    case MessageName::WebPage_SavePDF:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::PDFPluginIdentifier", nullptr, false },
        };
    case MessageName::WebPage_OpenPDFWithPreview:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::PDFPluginIdentifier", nullptr, false },
        };
#endif
    case MessageName::WebPage_UpdateCurrentModifierState:
        return Vector<ArgumentDescription> {
            { "modifiers", "OptionSet<WebCore::PlatformEvent::Modifier>", nullptr, false },
        };
    case MessageName::WebPage_SimulateDeviceOrientationChange:
        return Vector<ArgumentDescription> {
            { "alpha", "double", nullptr, false },
            { "beta", "double", nullptr, false },
            { "gamma", "double", nullptr, false },
        };
#if ENABLE(SPEECH_SYNTHESIS)
    case MessageName::WebPage_SpeakingErrorOccurred:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_BoundaryEventOccurred:
        return Vector<ArgumentDescription> {
            { "wordBoundary", "bool", nullptr, false },
            { "charIndex", "unsigned", nullptr, false },
            { "charLength", "unsigned", nullptr, false },
        };
    case MessageName::WebPage_VoicesDidChange:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebPage_SetCanShowPlaceholder:
        return Vector<ArgumentDescription> {
            { "context", "WebCore::ElementContext", nullptr, false },
            { "canShowPlaceholder", "bool", nullptr, false },
        };
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::WebPage_WasLoadedWithDataTransferFromPrevalentResource:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_ClearLoadedSubresourceDomains:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_GetLoadedSubresourceDomains:
        return Vector<ArgumentDescription> { };
#endif
#if USE(SYSTEM_PREVIEW)
    case MessageName::WebPage_SystemPreviewActionTriggered:
        return Vector<ArgumentDescription> {
            { "previewInfo", "WebCore::SystemPreviewInfo", nullptr, false },
            { "message", "String", nullptr, false },
        };
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    case MessageName::WebPage_SendMessageToWebProcessExtension:
        return Vector<ArgumentDescription> {
            { "userMessage", "WebKit::UserMessage", nullptr, false },
        };
    case MessageName::WebPage_SendMessageToWebProcessExtensionWithReply:
        return Vector<ArgumentDescription> {
            { "userMessage", "WebKit::UserMessage", nullptr, false },
        };
#endif
    case MessageName::WebPage_StartTextManipulations:
        return Vector<ArgumentDescription> {
            { "exclusionRules", "Vector<WebCore::TextManipulationController::ExclusionRule>", nullptr, false },
        };
    case MessageName::WebPage_CompleteTextManipulation:
        return Vector<ArgumentDescription> {
            { "items", "Vector<WebCore::TextManipulationItem>", nullptr, false },
        };
    case MessageName::WebPage_SetOverriddenMediaType:
        return Vector<ArgumentDescription> {
            { "mediaType", "String", nullptr, false },
        };
    case MessageName::WebPage_GetProcessDisplayName:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_UpdateCORSDisablingPatterns:
        return Vector<ArgumentDescription> {
            { "patterns", "Vector<String>", nullptr, false },
        };
    case MessageName::WebPage_SetIsTakingSnapshotsForApplicationSuspension:
        return Vector<ArgumentDescription> {
            { "isTakingSnapshotsForApplicationSuspension", "bool", nullptr, false },
        };
    case MessageName::WebPage_SetNeedsDOMWindowResizeEvent:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_SetHasResourceLoadClient:
        return Vector<ArgumentDescription> {
            { "has", "bool", nullptr, false },
        };
    case MessageName::WebPage_SetCanUseCredentialStorage:
        return Vector<ArgumentDescription> {
            { "canUse", "bool", nullptr, false },
        };
#if ENABLE(APP_HIGHLIGHTS)
    case MessageName::WebPage_CreateAppHighlightInSelectedRange:
        return Vector<ArgumentDescription> {
            { "createNewGroup", "bool", "WebCore::CreateNewGroupForHighlight", false },
            { "requestOrigin", "bool", "WebCore::HighlightRequestOriginatedInApp", false },
        };
    case MessageName::WebPage_RestoreAppHighlightsAndScrollToIndex:
        return Vector<ArgumentDescription> {
            { "memoryHandles", "Vector<WebKit::SharedMemory::Handle>", nullptr, false },
            { "index", "unsigned", nullptr, true },
        };
    case MessageName::WebPage_SetAppHighlightsVisibility:
        return Vector<ArgumentDescription> {
            { "highlightVisibility", "bool", "WebCore::HighlightVisibility", false },
        };
#endif
    case MessageName::WebPage_HandleWheelEvent:
        return Vector<ArgumentDescription> {
            { "event", "WebKit::WebWheelEvent", nullptr, false },
            { "processingSteps", "OptionSet<WebCore::WheelEventProcessingSteps>", nullptr, false },
        };
    case MessageName::WebPage_DispatchWheelEventWithoutScrolling:
        return Vector<ArgumentDescription> {
            { "event", "WebKit::WebWheelEvent", nullptr, false },
        };
    case MessageName::WebPage_LastNavigationWasAppInitiated:
        return Vector<ArgumentDescription> { };
#if ENABLE(MEDIA_SESSION_COORDINATOR)
    case MessageName::WebPage_CreateMediaSessionCoordinator:
        return Vector<ArgumentDescription> {
            { "identifier", "String", nullptr, false },
        };
#endif
#if ENABLE(IMAGE_ANALYSIS)
    case MessageName::WebPage_UpdateWithTextRecognitionResult:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::TextRecognitionResult", nullptr, false },
            { "element", "WebCore::ElementContext", nullptr, false },
            { "location", "WebCore::FloatPoint", nullptr, false },
        };
    case MessageName::WebPage_StartVisualTranslation:
        return Vector<ArgumentDescription> {
            { "sourceLanguageIdentifier", "String", nullptr, false },
            { "targetLanguageIdentifier", "String", nullptr, false },
        };
#endif
    case MessageName::WebPage_ScrollToRect:
        return Vector<ArgumentDescription> {
            { "targetRect", "WebCore::FloatRect", nullptr, false },
            { "origin", "WebCore::FloatPoint", nullptr, false },
        };
    case MessageName::WebPage_NavigateServiceWorkerClient:
        return Vector<ArgumentDescription> {
            { "documentIdentifier", "WebCore::ScriptExecutionContextIdentifier", nullptr, false },
            { "url", "URL", nullptr, false },
        };
#if ENABLE(ARKIT_INLINE_PREVIEW_IOS)
    case MessageName::WebPage_ModelInlinePreviewDidLoad:
        return Vector<ArgumentDescription> {
            { "layerID", "WebCore::GraphicsLayer::PlatformLayerID", nullptr, false },
        };
    case MessageName::WebPage_ModelInlinePreviewDidFailToLoad:
        return Vector<ArgumentDescription> {
            { "layerID", "WebCore::GraphicsLayer::PlatformLayerID", nullptr, false },
            { "error", "WebCore::ResourceError", nullptr, false },
        };
#endif
#if HAVE(UIKIT_RESIZABLE_WINDOWS)
    case MessageName::WebPage_SetIsWindowResizingEnabled:
        return Vector<ArgumentDescription> {
            { "hasResizableWindows", "bool", nullptr, false },
        };
#endif
#if ENABLE(INTERACTION_REGIONS_IN_EVENT_REGION)
    case MessageName::WebPage_SetInteractionRegionsEnabled:
        return Vector<ArgumentDescription> {
            { "enable", "bool", nullptr, false },
        };
#endif
    case MessageName::WebPage_GenerateTestReport:
        return Vector<ArgumentDescription> {
            { "message", "String", nullptr, false },
            { "group", "String", nullptr, false },
        };
#if ENABLE(NETWORK_CONNECTION_INTEGRITY)
    case MessageName::WebPage_SetLookalikeCharacterStrings:
        return Vector<ArgumentDescription> {
            { "strings", "Vector<String>", nullptr, false },
        };
    case MessageName::WebPage_SetAllowedLookalikeCharacterStrings:
        return Vector<ArgumentDescription> {
            { "allowedStrings", "Vector<WebCore::LookalikeCharactersSanitizationData>", nullptr, false },
        };
#endif
#if ENABLE(ACCESSIBILITY_ANIMATION_CONTROL)
    case MessageName::WebPage_PauseAllAnimations:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_PlayAllAnimations:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::StorageAreaMap_DispatchStorageEvent:
        return Vector<ArgumentDescription> {
            { "storageAreaImplID", "WebKit::StorageAreaImplIdentifier", nullptr, true },
            { "key", "String", nullptr, false },
            { "oldValue", "String", nullptr, false },
            { "newValue", "String", nullptr, false },
            { "urlString", "String", nullptr, false },
            { "messageIdentifier", "uint64_t", nullptr, false },
        };
    case MessageName::StorageAreaMap_ClearCache:
        return Vector<ArgumentDescription> {
            { "messageIdentifier", "uint64_t", nullptr, false },
        };
#if ENABLE(WEBXR)
    case MessageName::PlatformXRSystemProxy_SessionDidEnd:
        return Vector<ArgumentDescription> {
            { "deviceIdentifier", "WebKit::XRDeviceIdentifier", nullptr, false },
        };
    case MessageName::PlatformXRSystemProxy_SessionDidUpdateVisibilityState:
        return Vector<ArgumentDescription> {
            { "deviceIdentifier", "WebKit::XRDeviceIdentifier", nullptr, false },
            { "visibilityState", "PlatformXR::VisibilityState", nullptr, false },
        };
#endif
#if PLATFORM(MAC)
    case MessageName::ViewGestureController_DidCollectGeometryForSmartMagnificationGesture:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::FloatPoint", nullptr, false },
            { "renderRect", "WebCore::FloatRect", nullptr, false },
            { "visibleContentBounds", "WebCore::FloatRect", nullptr, false },
            { "fitEntireRect", "bool", nullptr, false },
            { "viewportMinimumScale", "double", nullptr, false },
            { "viewportMaximumScale", "double", nullptr, false },
        };
#endif
#if !PLATFORM(IOS_FAMILY)
    case MessageName::ViewGestureController_DidCollectGeometryForMagnificationGesture:
        return Vector<ArgumentDescription> {
            { "visibleContentBounds", "WebCore::FloatRect", nullptr, false },
            { "frameHandlesMagnificationGesture", "bool", nullptr, false },
        };
    case MessageName::ViewGestureController_DidHitRenderTreeSizeThreshold:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::GtkSettingsManagerProxy_SettingsDidChange:
        return Vector<ArgumentDescription> {
            { "settingsState", "WebKit::GtkSettingsState", nullptr, false },
        };
#if PLATFORM(COCOA)
    case MessageName::ViewGestureGeometryCollector_CollectGeometryForSmartMagnificationGesture:
        return Vector<ArgumentDescription> {
            { "origin", "WebCore::FloatPoint", nullptr, false },
        };
#endif
#if !PLATFORM(IOS_FAMILY)
    case MessageName::ViewGestureGeometryCollector_CollectGeometryForMagnificationGesture:
        return Vector<ArgumentDescription> { };
    case MessageName::ViewGestureGeometryCollector_SetRenderTreeSizeNotificationThreshold:
        return Vector<ArgumentDescription> {
            { "size", "uint64_t", nullptr, false },
        };
#endif
#if ENABLE(MEDIA_STREAM)
    case MessageName::UserMediaCaptureManager_ValidateUserMediaRequestConstraints:
        return Vector<ArgumentDescription> {
            { "request", "WebCore::MediaStreamRequest", nullptr, false },
            { "mediaDeviceIdentifierHashSalts", "WebCore::MediaDeviceHashSalts", nullptr, false },
        };
    case MessageName::UserMediaCaptureManager_GetMediaStreamDevices:
        return Vector<ArgumentDescription> { };
#endif
    default:
        break;
    }
    return std::nullopt;
}

std::optional<Vector<ArgumentDescription>> messageReplyArgumentDescriptions(MessageName name)
{
    switch (name) {
#if ENABLE(GPU_PROCESS)
#if USE(AUDIO_SESSION)
    case MessageName::GPUConnectionToWebProcess_EnsureAudioSession:
        return Vector<ArgumentDescription> {
            { "configuration", "WebKit::RemoteAudioSessionConfiguration", nullptr, false },
        };
#endif
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::GPUProcess_CreateGPUConnectionToWebProcess:
        return Vector<ArgumentDescription> { };
    case MessageName::GPUProcess_PrepareToSuspend:
        return Vector<ArgumentDescription> { };
#if ENABLE(MEDIA_STREAM)
    case MessageName::GPUProcess_UpdateCaptureAccess:
        return Vector<ArgumentDescription> { };
#endif
#if HAVE(SC_CONTENT_SHARING_SESSION)
    case MessageName::GPUProcess_ShowWindowPicker:
        return Vector<ArgumentDescription> {
            { "device", "WebCore::CaptureDevice", nullptr, true },
        };
    case MessageName::GPUProcess_ShowScreenPicker:
        return Vector<ArgumentDescription> {
            { "device", "WebCore::CaptureDevice", nullptr, true },
        };
#endif
#if ENABLE(VIDEO)
    case MessageName::GPUProcess_RequestBitmapImageForCurrentTime:
        return Vector<ArgumentDescription> {
            { "handle", "WebKit::ShareableBitmapHandle", nullptr, false },
        };
#endif
    case MessageName::GPUProcess_WebProcessConnectionCountForTesting:
        return Vector<ArgumentDescription> {
            { "count", "uint64_t", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(WEBGL)
#if PLATFORM(COCOA)
    case MessageName::RemoteGraphicsContextGL_PrepareForDisplay:
        return Vector<ArgumentDescription> {
            { "displayBuffer", "MachSendRight", nullptr, false },
        };
#endif
#if USE(GRAPHICS_LAYER_WC)
    case MessageName::RemoteGraphicsContextGL_PrepareForDisplay:
        return Vector<ArgumentDescription> {
            { "contentBuffer", "WebKit::WCContentBufferIdentifier", nullptr, true },
        };
#endif
#if USE(GBM)
    case MessageName::RemoteGraphicsContextGL_PrepareForDisplay:
        return Vector<ArgumentDescription> {
            { "dmabufObject", "WebCore::DMABufObject", nullptr, false },
        };
#endif
#if !PLATFORM(COCOA) && !USE(GRAPHICS_LAYER_WC) && !USE(GBM)
    case MessageName::RemoteGraphicsContextGL_PrepareForDisplay:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::RemoteGraphicsContextGL_GetError:
        return Vector<ArgumentDescription> {
            { "returnValue", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_PaintRenderingResultsToCanvas:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteGraphicsContextGL_PaintCompositedResultsToCanvas:
        return Vector<ArgumentDescription> { };
#if ENABLE(VIDEO) && PLATFORM(COCOA)
    case MessageName::RemoteGraphicsContextGL_CopyTextureFromVideoFrame:
        return Vector<ArgumentDescription> {
            { "success", "bool", nullptr, false },
        };
#endif
#if ENABLE(MEDIA_STREAM)
    case MessageName::RemoteGraphicsContextGL_PaintCompositedResultsToVideoFrame:
        return Vector<ArgumentDescription> {
            { "properties", "WebKit::RemoteVideoFrameProxy::Properties", nullptr, true },
        };
#endif
    case MessageName::RemoteGraphicsContextGL_ReadnPixels0:
        return Vector<ArgumentDescription> {
            { "data", "IPC::ArrayReference<uint8_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_ReadnPixels2:
        return Vector<ArgumentDescription> {
            { "success", "bool", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_MoveErrorsToSyntheticErrorList:
        return Vector<ArgumentDescription> {
            { "returnValue", "bool", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CheckFramebufferStatus:
        return Vector<ArgumentDescription> {
            { "returnValue", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CreateBuffer:
        return Vector<ArgumentDescription> {
            { "returnValue", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CreateFramebuffer:
        return Vector<ArgumentDescription> {
            { "returnValue", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CreateProgram:
        return Vector<ArgumentDescription> {
            { "returnValue", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CreateRenderbuffer:
        return Vector<ArgumentDescription> {
            { "returnValue", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CreateShader:
        return Vector<ArgumentDescription> {
            { "returnValue", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CreateTexture:
        return Vector<ArgumentDescription> {
            { "returnValue", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetActiveAttrib:
        return Vector<ArgumentDescription> {
            { "returnValue", "bool", nullptr, false },
            { "arg2", "WebCore::GraphicsContextGLActiveInfo", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetActiveUniform:
        return Vector<ArgumentDescription> {
            { "returnValue", "bool", nullptr, false },
            { "arg2", "WebCore::GraphicsContextGLActiveInfo", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetAttribLocation:
        return Vector<ArgumentDescription> {
            { "returnValue", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetBufferParameteri:
        return Vector<ArgumentDescription> {
            { "returnValue", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetString:
        return Vector<ArgumentDescription> {
            { "returnValue", "String", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetFloatv:
        return Vector<ArgumentDescription> {
            { "value", "IPC::ArrayReference<float>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetIntegerv:
        return Vector<ArgumentDescription> {
            { "value", "IPC::ArrayReference<int32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetIntegeri_v:
        return Vector<ArgumentDescription> {
            { "value", "IPC::ArrayReference<int32_t, 4>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetInteger64:
        return Vector<ArgumentDescription> {
            { "returnValue", "int64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetInteger64i:
        return Vector<ArgumentDescription> {
            { "returnValue", "int64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetProgrami:
        return Vector<ArgumentDescription> {
            { "returnValue", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetBooleanv:
        return Vector<ArgumentDescription> {
            { "value", "IPC::ArrayReference<bool>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetFramebufferAttachmentParameteri:
        return Vector<ArgumentDescription> {
            { "returnValue", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetProgramInfoLog:
        return Vector<ArgumentDescription> {
            { "returnValue", "String", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetRenderbufferParameteri:
        return Vector<ArgumentDescription> {
            { "returnValue", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetShaderi:
        return Vector<ArgumentDescription> {
            { "returnValue", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetShaderInfoLog:
        return Vector<ArgumentDescription> {
            { "returnValue", "String", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetShaderPrecisionFormat:
        return Vector<ArgumentDescription> {
            { "range", "IPC::ArrayReference<int32_t, 2>", nullptr, false },
            { "precision", "GCGLint", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetShaderSource:
        return Vector<ArgumentDescription> {
            { "returnValue", "String", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetTexParameterf:
        return Vector<ArgumentDescription> {
            { "returnValue", "float", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetTexParameteri:
        return Vector<ArgumentDescription> {
            { "returnValue", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetUniformfv:
        return Vector<ArgumentDescription> {
            { "value", "IPC::ArrayReference<float>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetUniformiv:
        return Vector<ArgumentDescription> {
            { "value", "IPC::ArrayReference<int32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetUniformuiv:
        return Vector<ArgumentDescription> {
            { "value", "IPC::ArrayReference<uint32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetUniformLocation:
        return Vector<ArgumentDescription> {
            { "returnValue", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetVertexAttribOffset:
        return Vector<ArgumentDescription> {
            { "returnValue", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsBuffer:
        return Vector<ArgumentDescription> {
            { "returnValue", "bool", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsEnabled:
        return Vector<ArgumentDescription> {
            { "returnValue", "bool", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsFramebuffer:
        return Vector<ArgumentDescription> {
            { "returnValue", "bool", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsProgram:
        return Vector<ArgumentDescription> {
            { "returnValue", "bool", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsRenderbuffer:
        return Vector<ArgumentDescription> {
            { "returnValue", "bool", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsShader:
        return Vector<ArgumentDescription> {
            { "returnValue", "bool", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsTexture:
        return Vector<ArgumentDescription> {
            { "returnValue", "bool", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CreateVertexArray:
        return Vector<ArgumentDescription> {
            { "returnValue", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsVertexArray:
        return Vector<ArgumentDescription> {
            { "returnValue", "bool", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetBufferSubData:
        return Vector<ArgumentDescription> {
            { "data", "IPC::ArrayReference<uint8_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetFragDataLocation:
        return Vector<ArgumentDescription> {
            { "returnValue", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CreateQuery:
        return Vector<ArgumentDescription> {
            { "returnValue", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsQuery:
        return Vector<ArgumentDescription> {
            { "returnValue", "bool", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetQuery:
        return Vector<ArgumentDescription> {
            { "returnValue", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetQueryObjectui:
        return Vector<ArgumentDescription> {
            { "returnValue", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CreateSampler:
        return Vector<ArgumentDescription> {
            { "returnValue", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsSampler:
        return Vector<ArgumentDescription> {
            { "returnValue", "bool", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetSamplerParameterf:
        return Vector<ArgumentDescription> {
            { "returnValue", "float", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetSamplerParameteri:
        return Vector<ArgumentDescription> {
            { "returnValue", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_FenceSync:
        return Vector<ArgumentDescription> {
            { "returnValue", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsSync:
        return Vector<ArgumentDescription> {
            { "returnValue", "bool", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_ClientWaitSync:
        return Vector<ArgumentDescription> {
            { "returnValue", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetSynci:
        return Vector<ArgumentDescription> {
            { "returnValue", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CreateTransformFeedback:
        return Vector<ArgumentDescription> {
            { "returnValue", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsTransformFeedback:
        return Vector<ArgumentDescription> {
            { "returnValue", "bool", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetTransformFeedbackVarying:
        return Vector<ArgumentDescription> {
            { "arg2", "WebCore::GraphicsContextGLActiveInfo", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetUniformIndices:
        return Vector<ArgumentDescription> {
            { "returnValue", "Vector<uint32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetActiveUniforms:
        return Vector<ArgumentDescription> {
            { "returnValue", "Vector<int32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetUniformBlockIndex:
        return Vector<ArgumentDescription> {
            { "returnValue", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetActiveUniformBlockName:
        return Vector<ArgumentDescription> {
            { "returnValue", "String", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetActiveUniformBlockiv:
        return Vector<ArgumentDescription> {
            { "params", "IPC::ArrayReference<int32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetTranslatedShaderSourceANGLE:
        return Vector<ArgumentDescription> {
            { "returnValue", "String", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_CreateQueryEXT:
        return Vector<ArgumentDescription> {
            { "returnValue", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_IsQueryEXT:
        return Vector<ArgumentDescription> {
            { "returnValue", "bool", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetQueryiEXT:
        return Vector<ArgumentDescription> {
            { "returnValue", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetQueryObjectiEXT:
        return Vector<ArgumentDescription> {
            { "returnValue", "int32_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetQueryObjectui64EXT:
        return Vector<ArgumentDescription> {
            { "returnValue", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetInteger64EXT:
        return Vector<ArgumentDescription> {
            { "returnValue", "int64_t", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_GetInternalformativ:
        return Vector<ArgumentDescription> {
            { "params", "IPC::ArrayReference<int32_t>", nullptr, false },
        };
    case MessageName::RemoteGraphicsContextGL_PaintRenderingResultsToPixelBuffer:
        return Vector<ArgumentDescription> {
            { "returnValue", "RefPtr<WebCore::PixelBuffer>", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteRenderingBackend_GetPixelBufferForImageBuffer:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteRenderingBackend_GetPixelBufferForImageBufferWithNewMemory:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteRenderingBackend_GetShareableBitmapForImageBuffer:
        return Vector<ArgumentDescription> {
            { "handle", "WebKit::ShareableBitmapHandle", nullptr, false },
        };
    case MessageName::RemoteRenderingBackend_GetFilteredImageForImageBuffer:
        return Vector<ArgumentDescription> {
            { "handle", "WebKit::ShareableBitmapHandle", nullptr, false },
        };
    case MessageName::RemoteRenderingBackend_PrepareBuffersForDisplay:
        return Vector<ArgumentDescription> {
            { "swapBuffersOutput", "Vector<WebKit::PrepareBackingStoreBuffersOutputData>", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteAdapter_RequestDevice:
        return Vector<ArgumentDescription> {
            { "supportedFeatures", "WebKit::WebGPU::SupportedFeatures", nullptr, false },
            { "supportedLimits", "WebKit::WebGPU::SupportedLimits", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteBuffer_GetMappedRange:
        return Vector<ArgumentDescription> {
            { "data", "Vector<uint8_t>", nullptr, true },
        };
    case MessageName::RemoteBuffer_MapAsync:
        return Vector<ArgumentDescription> {
            { "data", "Vector<uint8_t>", nullptr, true },
        };
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#if PLATFORM(COCOA)
    case MessageName::RemoteCompositorIntegration_RecreateRenderBuffers:
        return Vector<ArgumentDescription> {
            { "renderBuffers", "Vector<MachSendRight>", nullptr, false },
        };
#endif
    case MessageName::RemoteCompositorIntegration_PrepareForDisplay:
        return Vector<ArgumentDescription> {
            { "dummy", "bool", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteDevice_CreateComputePipelineAsync:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteDevice_CreateRenderPipelineAsync:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteDevice_PopErrorScope:
        return Vector<ArgumentDescription> {
            { "error", "WebKit::WebGPU::Error", nullptr, true },
        };
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteGPU_RequestAdapter:
        return Vector<ArgumentDescription> {
            { "response", "WebKit::RemoteGPU::RequestAdapterResponse", nullptr, true },
        };
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteQueue_OnSubmittedWorkDone:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteShaderModule_CompilationInfo:
        return Vector<ArgumentDescription> {
            { "messages", "Vector<WebKit::WebGPU::CompilationMessage>", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(WEB_AUDIO)
    case MessageName::RemoteAudioDestinationManager_StartAudioDestination:
        return Vector<ArgumentDescription> {
            { "isPlaying", "bool", nullptr, false },
        };
    case MessageName::RemoteAudioDestinationManager_StopAudioDestination:
        return Vector<ArgumentDescription> {
            { "isPlaying", "bool", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && USE(AUDIO_SESSION)
    case MessageName::RemoteAudioSessionProxy_TryToSetActive:
        return Vector<ArgumentDescription> {
            { "suceeded", "bool", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteCDMFactoryProxy_CreateCDM:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteCDMIdentifier", nullptr, false },
            { "configuration", "WebKit::RemoteCDMConfiguration", nullptr, false },
        };
    case MessageName::RemoteCDMFactoryProxy_SupportsKeySystem:
        return Vector<ArgumentDescription> {
            { "result", "bool", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteCDMInstanceProxy_CreateSession:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteCDMInstanceSessionIdentifier", nullptr, false },
        };
    case MessageName::RemoteCDMInstanceProxy_InitializeWithConfiguration:
        return Vector<ArgumentDescription> {
            { "success", "bool", "WebCore::CDMInstance::SuccessValue", false },
        };
    case MessageName::RemoteCDMInstanceProxy_SetServerCertificate:
        return Vector<ArgumentDescription> {
            { "success", "bool", "WebCore::CDMInstance::SuccessValue", false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteCDMInstanceSessionProxy_RequestLicense:
        return Vector<ArgumentDescription> {
            { "message", "RefPtr<WebCore::SharedBuffer>", nullptr, false },
            { "sessionId", "String", nullptr, false },
            { "needsIndividualization", "bool", nullptr, false },
            { "succeeded", "bool", nullptr, false },
        };
    case MessageName::RemoteCDMInstanceSessionProxy_UpdateLicense:
        return Vector<ArgumentDescription> {
            { "sessionWasClosed", "bool", nullptr, false },
            { "changedKeys", "WebCore::CDMInstanceSession::KeyStatusVector", nullptr, true },
            { "changedExpiration", "double", nullptr, true },
            { "message", "WebCore::CDMInstanceSession::Message", nullptr, true },
            { "succeeded", "bool", nullptr, false },
        };
    case MessageName::RemoteCDMInstanceSessionProxy_LoadSession:
        return Vector<ArgumentDescription> {
            { "changedKeys", "WebCore::CDMInstanceSession::KeyStatusVector", nullptr, true },
            { "changedExpiration", "double", nullptr, true },
            { "message", "WebCore::CDMInstanceSession::Message", nullptr, true },
            { "succeeded", "bool", nullptr, false },
            { "loadFailure", "uint8_t", "WebCore::CDMInstanceSession::SessionLoadFailure", false },
        };
    case MessageName::RemoteCDMInstanceSessionProxy_CloseSession:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteCDMInstanceSessionProxy_RemoveSessionData:
        return Vector<ArgumentDescription> {
            { "changedKeys", "WebCore::CDMInstanceSession::KeyStatusVector", nullptr, false },
            { "message", "RefPtr<WebCore::SharedBuffer>", nullptr, false },
            { "succeeded", "bool", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
    case MessageName::RemoteCDMProxy_GetSupportedConfiguration:
        return Vector<ArgumentDescription> {
            { "accumulatedConfiguration", "WebCore::CDMKeySystemConfiguration", nullptr, true },
        };
    case MessageName::RemoteCDMProxy_CreateInstance:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteCDMInstanceIdentifier", nullptr, false },
            { "configuration", "WebKit::RemoteCDMInstanceConfiguration", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
    case MessageName::RemoteLegacyCDMFactoryProxy_CreateCDM:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteLegacyCDMIdentifier", nullptr, false },
        };
    case MessageName::RemoteLegacyCDMFactoryProxy_SupportsKeySystem:
        return Vector<ArgumentDescription> {
            { "result", "bool", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
    case MessageName::RemoteLegacyCDMProxy_SupportsMIMEType:
        return Vector<ArgumentDescription> {
            { "supports", "bool", nullptr, false },
        };
    case MessageName::RemoteLegacyCDMProxy_CreateSession:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::RemoteLegacyCDMSessionIdentifier", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
    case MessageName::RemoteLegacyCDMSessionProxy_GenerateKeyRequest:
        return Vector<ArgumentDescription> {
            { "nextMessage", "RefPtr<WebCore::SharedBuffer>", nullptr, false },
            { "destinationURL", "String", nullptr, false },
            { "errorCode", "unsigned short", nullptr, false },
            { "systemCode", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteLegacyCDMSessionProxy_Update:
        return Vector<ArgumentDescription> {
            { "succeeded", "bool", nullptr, false },
            { "nextMessage", "RefPtr<WebCore::SharedBuffer>", nullptr, false },
            { "errorCode", "unsigned short", nullptr, false },
            { "systemCode", "uint32_t", nullptr, false },
        };
    case MessageName::RemoteLegacyCDMSessionProxy_CachedKeyForKeyID:
        return Vector<ArgumentDescription> {
            { "key", "RefPtr<WebCore::SharedBuffer>", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS)
    case MessageName::RemoteMediaEngineConfigurationFactoryProxy_CreateDecodingConfiguration:
        return Vector<ArgumentDescription> {
            { "decodingInfo", "WebCore::MediaCapabilitiesDecodingInfo", nullptr, false },
        };
    case MessageName::RemoteMediaEngineConfigurationFactoryProxy_CreateEncodingConfiguration:
        return Vector<ArgumentDescription> {
            { "decodingInfo", "WebCore::MediaCapabilitiesEncodingInfo", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
    case MessageName::RemoteMediaPlayerManagerProxy_GetSupportedTypes:
        return Vector<ArgumentDescription> {
            { "types", "Vector<String>", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerManagerProxy_SupportsTypeAndCodecs:
        return Vector<ArgumentDescription> {
            { "support", "uint8_t", "WebCore::MediaPlayerEnums::SupportsType", false },
        };
    case MessageName::RemoteMediaPlayerManagerProxy_OriginsInMediaCache:
        return Vector<ArgumentDescription> {
            { "origins", "HashSet<WebCore::SecurityOriginData>", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerManagerProxy_SupportsKeySystem:
        return Vector<ArgumentDescription> {
            { "supports", "bool", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
    case MessageName::RemoteMediaPlayerProxy_Load:
        return Vector<ArgumentDescription> {
            { "playerConfiguration", "WebKit::RemoteMediaPlayerConfiguration", nullptr, false },
        };
#if ENABLE(MEDIA_SOURCE)
    case MessageName::RemoteMediaPlayerProxy_LoadMediaSource:
        return Vector<ArgumentDescription> {
            { "playerConfiguration", "WebKit::RemoteMediaPlayerConfiguration", nullptr, false },
        };
#endif
    case MessageName::RemoteMediaPlayerProxy_DidLoadingProgress:
        return Vector<ArgumentDescription> {
            { "flag", "bool", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_PerformTaskAtMediaTime:
        return Vector<ArgumentDescription> {
            { "mediaTime", "MediaTime", nullptr, true },
            { "monotonicTime", "MonotonicTime", nullptr, true },
        };
    case MessageName::RemoteMediaPlayerProxy_IsCrossOrigin:
        return Vector<ArgumentDescription> {
            { "crossOrigin", "bool", nullptr, true },
        };
#if PLATFORM(IOS_FAMILY)
    case MessageName::RemoteMediaPlayerProxy_ErrorLog:
        return Vector<ArgumentDescription> {
            { "errorLog", "String", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_AccessLog:
        return Vector<ArgumentDescription> {
            { "accessLog", "String", nullptr, false },
        };
#endif
#if PLATFORM(COCOA)
    case MessageName::RemoteMediaPlayerProxy_NativeImageForCurrentTime:
        return Vector<ArgumentDescription> {
            { "sendRight", "MachSendRight", nullptr, true },
            { "colorSpace", "WebCore::DestinationColorSpace", nullptr, false },
        };
    case MessageName::RemoteMediaPlayerProxy_ColorSpace:
        return Vector<ArgumentDescription> {
            { "colorSpace", "WebCore::DestinationColorSpace", nullptr, false },
        };
#endif
    case MessageName::RemoteMediaPlayerProxy_VideoFrameForCurrentTimeIfChanged:
        return Vector<ArgumentDescription> {
            { "videoFrame", "WebKit::RemoteVideoFrameProxy::Properties", nullptr, true },
            { "changed", "bool", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
    case MessageName::RemoteMediaResourceManager_ResponseReceived:
        return Vector<ArgumentDescription> {
            { "shouldContinue", "bool", "WebCore::ShouldContinuePolicyCheck", false },
        };
    case MessageName::RemoteMediaResourceManager_RedirectReceived:
        return Vector<ArgumentDescription> {
            { "returnRequest", "WebCore::ResourceRequest", nullptr, false },
        };
    case MessageName::RemoteMediaResourceManager_DataReceived:
        return Vector<ArgumentDescription> {
            { "remoteData", "WebKit::SharedMemory::Handle", nullptr, true },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
    case MessageName::RemoteMediaSourceProxy_AddSourceBuffer:
        return Vector<ArgumentDescription> {
            { "status", "uint8_t", "WebCore::MediaSourcePrivate::AddStatus", false },
            { "remoteSourceBufferIdentifier", "WebKit::RemoteSourceBufferIdentifier", nullptr, true },
        };
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
    case MessageName::RemoteSourceBufferProxy_CanSwitchToType:
        return Vector<ArgumentDescription> {
            { "canSwitch", "bool", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_Append:
        return Vector<ArgumentDescription> {
            { "remoteData", "WebKit::SharedMemory::Handle", nullptr, true },
        };
    case MessageName::RemoteSourceBufferProxy_UpdateBufferedFromTrackBuffers:
        return Vector<ArgumentDescription> {
            { "buffered", "WebCore::PlatformTimeRanges", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_RemoveCodedFrames:
        return Vector<ArgumentDescription> {
            { "buffered", "WebCore::PlatformTimeRanges", nullptr, false },
            { "totalTrackBufferSizeInBytes", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_EvictCodedFrames:
        return Vector<ArgumentDescription> {
            { "totalTrackBufferSizeInBytes", "uint64_t", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_BufferedSamplesForTrackId:
        return Vector<ArgumentDescription> {
            { "samples", "Vector<String>", nullptr, false },
        };
    case MessageName::RemoteSourceBufferProxy_EnqueuedSamplesForTrackID:
        return Vector<ArgumentDescription> {
            { "samples", "Vector<String>", nullptr, false },
        };
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
#if PLATFORM(COCOA)
    case MessageName::RemoteVideoFrameObjectHeap_PixelBuffer:
        return Vector<ArgumentDescription> {
            { "result", "RetainPtr<CVPixelBufferRef>", nullptr, false },
        };
    case MessageName::RemoteVideoFrameObjectHeap_ConvertFrameBuffer:
        return Vector<ArgumentDescription> {
            { "space", "WebCore::DestinationColorSpace", nullptr, false },
        };
#endif
#endif
#if USE(LIBWEBRTC) && PLATFORM(COCOA) && ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM) && PLATFORM(COCOA)
    case MessageName::RemoteAudioMediaStreamTrackRendererInternalUnitManager_CreateUnit:
        return Vector<ArgumentDescription> {
            { "description", "WebCore::CAAudioStreamDescription", nullptr, true },
            { "frameChunkSize", "size_t", nullptr, false },
        };
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
    case MessageName::RemoteMediaRecorder_FetchData:
        return Vector<ArgumentDescription> {
            { "buffer", "IPC::DataReference", nullptr, false },
            { "timeCode", "double", nullptr, false },
        };
    case MessageName::RemoteMediaRecorder_StopRecording:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaRecorder_Pause:
        return Vector<ArgumentDescription> { };
    case MessageName::RemoteMediaRecorder_Resume:
        return Vector<ArgumentDescription> { };
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
    case MessageName::RemoteMediaRecorderManager_CreateRecorder:
        return Vector<ArgumentDescription> {
            { "creationError", "WebCore::ExceptionData", nullptr, true },
            { "mimeType", "String", nullptr, false },
            { "audioBitRate", "unsigned", nullptr, false },
            { "videoBitRate", "unsigned", nullptr, false },
        };
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
    case MessageName::RemoteSampleBufferDisplayLayerManager_CreateLayer:
        return Vector<ArgumentDescription> {
            { "contextID", "WebKit::LayerHostingContextID", nullptr, true },
        };
#endif
    case MessageName::WebCookieManager_GetHostnamesWithCookies:
        return Vector<ArgumentDescription> {
            { "hostnames", "Vector<String>", nullptr, false },
        };
    case MessageName::WebCookieManager_DeleteCookiesForHostnames:
        return Vector<ArgumentDescription> { };
    case MessageName::WebCookieManager_DeleteAllCookies:
        return Vector<ArgumentDescription> { };
    case MessageName::WebCookieManager_SetCookie:
        return Vector<ArgumentDescription> { };
    case MessageName::WebCookieManager_SetCookies:
        return Vector<ArgumentDescription> { };
    case MessageName::WebCookieManager_GetAllCookies:
        return Vector<ArgumentDescription> {
            { "cookies", "Vector<WebCore::Cookie>", nullptr, false },
        };
    case MessageName::WebCookieManager_GetCookies:
        return Vector<ArgumentDescription> {
            { "cookies", "Vector<WebCore::Cookie>", nullptr, false },
        };
    case MessageName::WebCookieManager_DeleteCookie:
        return Vector<ArgumentDescription> { };
    case MessageName::WebCookieManager_DeleteAllCookiesModifiedSince:
        return Vector<ArgumentDescription> { };
    case MessageName::WebCookieManager_SetHTTPCookieAcceptPolicy:
        return Vector<ArgumentDescription> { };
    case MessageName::WebCookieManager_GetHTTPCookieAcceptPolicy:
        return Vector<ArgumentDescription> {
            { "policy", "uint8_t", "WebCore::HTTPCookieAcceptPolicy", false },
        };
    case MessageName::NetworkBroadcastChannelRegistry_PostMessage:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkConnectionToWebProcess_PerformSynchronousLoad:
        return Vector<ArgumentDescription> {
            { "error", "WebCore::ResourceError", nullptr, false },
            { "response", "WebCore::ResourceResponse", nullptr, false },
            { "data", "Vector<uint8_t>", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_TestProcessIncomingSyncMessagesWhenWaitingForSyncReply:
        return Vector<ArgumentDescription> {
            { "handled", "bool", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_SendH2Ping:
        return Vector<ArgumentDescription> {
            { "result", "Expected<Seconds, WebCore::ResourceError>", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_IsResourceLoadFinished:
        return Vector<ArgumentDescription> {
            { "isFinished", "bool", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_CookiesForDOM:
        return Vector<ArgumentDescription> {
            { "cookieString", "String", nullptr, false },
            { "didAccessSecureCookies", "bool", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_CookieRequestHeaderFieldValue:
        return Vector<ArgumentDescription> {
            { "cookieString", "String", nullptr, false },
            { "didAccessSecureCookies", "bool", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_GetRawCookies:
        return Vector<ArgumentDescription> {
            { "cookies", "Vector<WebCore::Cookie>", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_DeleteCookie:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkConnectionToWebProcess_DomCookiesForHost:
        return Vector<ArgumentDescription> {
            { "cookies", "Vector<WebCore::Cookie>", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_BlobSize:
        return Vector<ArgumentDescription> {
            { "resultSize", "uint64_t", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_WriteBlobsToTemporaryFilesForIndexedDB:
        return Vector<ArgumentDescription> {
            { "fileNames", "Vector<String>", nullptr, false },
        };
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::NetworkConnectionToWebProcess_ResourceLoadStatisticsUpdated:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkConnectionToWebProcess_HasStorageAccess:
        return Vector<ArgumentDescription> {
            { "hasStorageAccess", "bool", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_RequestStorageAccess:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::RequestStorageAccessResult", nullptr, false },
        };
#endif
    case MessageName::NetworkConnectionToWebProcess_GetNetworkLoadInformationResponse:
        return Vector<ArgumentDescription> {
            { "response", "WebCore::ResourceResponse", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_GetNetworkLoadIntermediateInformation:
        return Vector<ArgumentDescription> {
            { "transactions", "Vector<WebCore::NetworkTransactionInformation>", nullptr, false },
        };
    case MessageName::NetworkConnectionToWebProcess_TakeNetworkLoadInformationMetrics:
        return Vector<ArgumentDescription> {
            { "networkMetrics", "WebCore::NetworkLoadMetrics", nullptr, false },
        };
#if ENABLE(SERVICE_WORKER)
    case MessageName::NetworkConnectionToWebProcess_EstablishSWContextConnection:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::NetworkConnectionToWebProcess_EstablishSharedWorkerContextConnection:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkConnectionToWebProcess_CreateRTCProvider:
        return Vector<ArgumentDescription> { };
#if ENABLE(WEB_RTC)
    case MessageName::NetworkConnectionToWebProcess_ConnectToRTCDataChannelRemoteSource:
        return Vector<ArgumentDescription> {
            { "result", "bool", nullptr, true },
        };
#endif
    case MessageName::NetworkConnectionToWebProcess_TakeAllMessagesForPort:
        return Vector<ArgumentDescription> {
            { "messages", "Vector<WebCore::MessageWithMessagePorts>", nullptr, false },
            { "messageBatchIdentifier", "uint64_t", nullptr, false },
        };
#if PLATFORM(MAC)
    case MessageName::NetworkConnectionToWebProcess_GetProcessDisplayName:
        return Vector<ArgumentDescription> {
            { "displayName", "String", nullptr, false },
        };
#endif
#if ENABLE(CONTENT_EXTENSIONS)
#endif
    case MessageName::NetworkProcess_InitializeNetworkProcess:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_CreateNetworkConnectionToWebProcess:
        return Vector<ArgumentDescription> {
            { "connectionHandle", "IPC::Connection::Handle", nullptr, true },
            { "cookieAcceptPolicy", "uint8_t", "WebCore::HTTPCookieAcceptPolicy", false },
        };
    case MessageName::NetworkProcess_AddAllowedFirstPartyForCookies:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_FetchWebsiteData:
        return Vector<ArgumentDescription> {
            { "websiteData", "WebKit::WebsiteData", nullptr, false },
        };
    case MessageName::NetworkProcess_DeleteWebsiteData:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_DeleteWebsiteDataForOrigins:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_RenameOriginInWebsiteData:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_WebsiteDataOriginDirectoryForTesting:
        return Vector<ArgumentDescription> {
            { "directory", "String", nullptr, false },
        };
    case MessageName::NetworkProcess_CancelDownload:
        return Vector<ArgumentDescription> {
            { "resumeData", "IPC::DataReference", nullptr, false },
        };
    case MessageName::NetworkProcess_DataTaskWithRequest:
        return Vector<ArgumentDescription> {
            { "taskIdentifier", "WebKit::DataTaskIdentifier", nullptr, false },
        };
    case MessageName::NetworkProcess_SetAllowsAnySSLCertificateForWebSocket:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_FlushCookies:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetCacheModelSynchronouslyForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_ProcessWillSuspendImminentlyForTestingSync:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_PrepareToSuspend:
        return Vector<ArgumentDescription> { };
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::NetworkProcess_ClearPrevalentResource:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_ClearUserInteraction:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_DumpResourceLoadStatistics:
        return Vector<ArgumentDescription> {
            { "dumpedStatistics", "String", nullptr, false },
        };
    case MessageName::NetworkProcess_UpdatePrevalentDomainsToBlockCookiesFor:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_IsGrandfathered:
        return Vector<ArgumentDescription> {
            { "isGrandfathered", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_IsPrevalentResource:
        return Vector<ArgumentDescription> {
            { "isPrevalent", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_IsVeryPrevalentResource:
        return Vector<ArgumentDescription> {
            { "isVeryPrevalent", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_SetLastSeen:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_MergeStatisticForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_InsertExpiredStatisticForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetPrevalentResource:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetPrevalentResourceForDebugMode:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_IsResourceLoadStatisticsEphemeral:
        return Vector<ArgumentDescription> {
            { "isEphemeral", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_HadUserInteraction:
        return Vector<ArgumentDescription> {
            { "hadUserInteraction", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_IsRelationshipOnlyInDatabaseOnce:
        return Vector<ArgumentDescription> {
            { "hadUserInteraction", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_HasLocalStorage:
        return Vector<ArgumentDescription> {
            { "hadUserInteraction", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_GetAllStorageAccessEntries:
        return Vector<ArgumentDescription> {
            { "domains", "Vector<String>", nullptr, false },
        };
    case MessageName::NetworkProcess_IsRegisteredAsRedirectingTo:
        return Vector<ArgumentDescription> {
            { "isRedirectingTo", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_IsRegisteredAsSubFrameUnder:
        return Vector<ArgumentDescription> {
            { "isSubframeUnder", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_IsRegisteredAsSubresourceUnder:
        return Vector<ArgumentDescription> {
            { "isSubresourceUnder", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_DomainIDExistsInDatabase:
        return Vector<ArgumentDescription> {
            { "exists", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_LogUserInteraction:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_ResetParametersToDefaultValues:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_ScheduleClearInMemoryAndPersistent:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_ScheduleCookieBlockingUpdate:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_ScheduleStatisticsAndDataRecordsProcessing:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_StatisticsDatabaseHasAllTables:
        return Vector<ArgumentDescription> {
            { "hasAllTables", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_SetCacheMaxAgeCapForPrevalentResources:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetGrandfathered:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_GetResourceLoadStatisticsDataSummary:
        return Vector<ArgumentDescription> {
            { "thirdPartyData", "Vector<WebKit::WebResourceLoadStatisticsStore::ThirdPartyData>", nullptr, false },
        };
    case MessageName::NetworkProcess_SetGrandfatheringTime:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetMaxStatisticsEntries:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetMinimumTimeBetweenDataRecordsRemoval:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetPruneEntriesDownTo:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetShouldClassifyResourcesBeforeDataRecordsRemoval:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetNotifyPagesWhenDataRecordsWereScanned:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetResourceLoadStatisticsTimeAdvanceForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetIsRunningResourceLoadStatisticsTest:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetResourceLoadStatisticsDebugMode:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetVeryPrevalentResource:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetSubframeUnderTopFrameDomain:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetSubresourceUnderTopFrameDomain:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetSubresourceUniqueRedirectTo:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetSubresourceUniqueRedirectFrom:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetTimeToLiveUserInteraction:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetTopFrameUniqueRedirectTo:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetTopFrameUniqueRedirectFrom:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_ResetCacheMaxAgeCapForPrevalentResources:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetCrossSiteLoadWithLinkDecorationForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_ResetCrossSiteLoadsWithLinkDecorationForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_DeleteCookiesForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_HasIsolatedSession:
        return Vector<ArgumentDescription> {
            { "hasIsolatedSession", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_CloseITPDatabase:
        return Vector<ArgumentDescription> { };
#endif
#if (ENABLE(TRACKING_PREVENTION) && ENABLE(APP_BOUND_DOMAINS))
    case MessageName::NetworkProcess_SetAppBoundDomainsForResourceLoadStatistics:
        return Vector<ArgumentDescription> { };
#endif
#if (ENABLE(TRACKING_PREVENTION) && ENABLE(MANAGED_DOMAINS))
    case MessageName::NetworkProcess_SetManagedDomainsForResourceLoadStatistics:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::NetworkProcess_SetShouldDowngradeReferrerForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetThirdPartyCookieBlockingMode:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetShouldEnbleSameSiteStrictEnforcementForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetFirstPartyWebsiteDataRemovalModeForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetToSameSiteStrictCookiesForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetFirstPartyHostCNAMEDomainForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetThirdPartyCNAMEDomainForTesting:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::NetworkProcess_SyncLocalStorage:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_DumpPrivateClickMeasurement:
        return Vector<ArgumentDescription> {
            { "privateClickMeasurementState", "String", nullptr, false },
        };
    case MessageName::NetworkProcess_ClearPrivateClickMeasurement:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetPrivateClickMeasurementOverrideTimerForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_MarkAttributedPrivateClickMeasurementsAsExpiredForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetPrivateClickMeasurementEphemeralMeasurementForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SimulatePrivateClickMeasurementSessionRestart:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetPrivateClickMeasurementTokenPublicKeyURLForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetPrivateClickMeasurementTokenSignatureURLForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetPrivateClickMeasurementAttributionReportURLsForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_MarkPrivateClickMeasurementsAsExpiredForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetPCMFraudPreventionValuesForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetPrivateClickMeasurementAppBundleIDForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_ClosePCMDatabase:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_SetServiceWorkerFetchTimeoutForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_ResetServiceWorkerFetchTimeoutForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_ResetQuota:
        return Vector<ArgumentDescription> { };
#if PLATFORM(IOS_FAMILY)
    case MessageName::NetworkProcess_SetBackupExclusionPeriodForTesting:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::NetworkProcess_ResetStoragePersistedState:
        return Vector<ArgumentDescription> { };
#if ENABLE(APP_BOUND_DOMAINS)
    case MessageName::NetworkProcess_HasAppBoundSession:
        return Vector<ArgumentDescription> {
            { "hasAppBoundSession", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_ClearAppBoundSession:
        return Vector<ArgumentDescription> { };
#endif
#if PLATFORM(IOS_FAMILY) && !PLATFORM(MACCATALYST)
    case MessageName::NetworkProcess_ClearServiceWorkerEntitlementOverride:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::NetworkProcess_UpdateBundleIdentifier:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_ClearBundleIdentifier:
        return Vector<ArgumentDescription> { };
#if PLATFORM(COCOA)
    case MessageName::NetworkProcess_AppPrivacyReportTestingData:
        return Vector<ArgumentDescription> {
            { "data", "WebKit::AppPrivacyReportTestingData", nullptr, false },
        };
    case MessageName::NetworkProcess_ClearAppPrivacyReportTestingData:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::NetworkProcess_CountNonDefaultSessionSets:
        return Vector<ArgumentDescription> {
            { "count", "size_t", nullptr, false },
        };
#if ENABLE(SERVICE_WORKER)
    case MessageName::NetworkProcess_GetPendingPushMessages:
        return Vector<ArgumentDescription> {
            { "messages", "Vector<WebKit::WebPushMessage>", nullptr, false },
        };
    case MessageName::NetworkProcess_ProcessPushMessage:
        return Vector<ArgumentDescription> {
            { "didSucceed", "bool", nullptr, false },
        };
    case MessageName::NetworkProcess_ProcessNotificationEvent:
        return Vector<ArgumentDescription> {
            { "didSucceed", "bool", nullptr, false },
        };
#endif
    case MessageName::NetworkProcess_SetPushAndNotificationsEnabledForOrigin:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcess_DeletePushAndNotificationRegistration:
        return Vector<ArgumentDescription> {
            { "errorMessage", "String", nullptr, false },
        };
    case MessageName::NetworkProcess_GetOriginsWithPushAndNotificationPermissions:
        return Vector<ArgumentDescription> {
            { "origins", "Vector<WebCore::SecurityOriginData>", nullptr, false },
        };
    case MessageName::NetworkProcess_HasPushSubscriptionForTesting:
        return Vector<ArgumentDescription> {
            { "hasSubscription", "bool", nullptr, false },
        };
    case MessageName::NetworkSocketChannel_SendString:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkSocketChannel_SendData:
        return Vector<ArgumentDescription> { };
#if ENABLE(SERVICE_WORKER)
#endif
#if ENABLE(SERVICE_WORKER)
#endif
#if ENABLE(SERVICE_WORKER)
    case MessageName::WebSWServerConnection_ScheduleUnregisterJobInServer:
        return Vector<ArgumentDescription> {
            { "result", "Expected<bool, WebCore::ExceptionData>", nullptr, false },
        };
    case MessageName::WebSWServerConnection_MatchRegistration:
        return Vector<ArgumentDescription> {
            { "registration", "WebCore::ServiceWorkerRegistrationData", nullptr, true },
        };
    case MessageName::WebSWServerConnection_WhenRegistrationReady:
        return Vector<ArgumentDescription> {
            { "registration", "WebCore::ServiceWorkerRegistrationData", nullptr, true },
        };
    case MessageName::WebSWServerConnection_GetRegistrations:
        return Vector<ArgumentDescription> {
            { "registrations", "Vector<WebCore::ServiceWorkerRegistrationData>", nullptr, false },
        };
    case MessageName::WebSWServerConnection_TerminateWorkerFromClient:
        return Vector<ArgumentDescription> { };
    case MessageName::WebSWServerConnection_WhenServiceWorkerIsTerminatedForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::WebSWServerConnection_StoreRegistrationsOnDisk:
        return Vector<ArgumentDescription> { };
    case MessageName::WebSWServerConnection_SubscribeToPushService:
        return Vector<ArgumentDescription> {
            { "result", "Expected<WebCore::PushSubscriptionData, WebCore::ExceptionData>", nullptr, false },
        };
    case MessageName::WebSWServerConnection_UnsubscribeFromPushService:
        return Vector<ArgumentDescription> {
            { "result", "Expected<bool, WebCore::ExceptionData>", nullptr, false },
        };
    case MessageName::WebSWServerConnection_GetPushSubscription:
        return Vector<ArgumentDescription> {
            { "result", "Expected<std::optional<WebCore::PushSubscriptionData>, WebCore::ExceptionData>", nullptr, false },
        };
    case MessageName::WebSWServerConnection_GetPushPermissionState:
        return Vector<ArgumentDescription> {
            { "result", "Expected<uint8_t, WebCore::ExceptionData>", nullptr, false },
        };
    case MessageName::WebSWServerConnection_EnableNavigationPreload:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::ExceptionData", nullptr, true },
        };
    case MessageName::WebSWServerConnection_DisableNavigationPreload:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::ExceptionData", nullptr, true },
        };
    case MessageName::WebSWServerConnection_SetNavigationPreloadHeaderValue:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::ExceptionData", nullptr, true },
        };
    case MessageName::WebSWServerConnection_GetNavigationPreloadState:
        return Vector<ArgumentDescription> {
            { "result", "Expected<WebCore::NavigationPreloadState, WebCore::ExceptionData>", nullptr, false },
        };
#endif
#if ENABLE(SERVICE_WORKER)
    case MessageName::WebSWServerToContextConnection_FindClientByVisibleIdentifier:
        return Vector<ArgumentDescription> {
            { "data", "WebCore::ServiceWorkerClientData", nullptr, true },
        };
    case MessageName::WebSWServerToContextConnection_Claim:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::ExceptionData", nullptr, true },
        };
    case MessageName::WebSWServerToContextConnection_Focus:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::ServiceWorkerClientData", nullptr, true },
        };
    case MessageName::WebSWServerToContextConnection_Navigate:
        return Vector<ArgumentDescription> {
            { "result", "Expected<std::optional<WebCore::ServiceWorkerClientData>, WebCore::ExceptionData>", nullptr, false },
        };
    case MessageName::WebSWServerToContextConnection_OpenWindow:
        return Vector<ArgumentDescription> {
            { "newClientData", "Expected<std::optional<WebCore::ServiceWorkerClientData>, WebCore::ExceptionData>", nullptr, false },
        };
#endif
    case MessageName::CacheStorageEngineConnection_Open:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::DOMCacheEngine::CacheIdentifierOrError", nullptr, false },
        };
    case MessageName::CacheStorageEngineConnection_Remove:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::DOMCacheEngine::RemoveCacheIdentifierOrError", nullptr, false },
        };
    case MessageName::CacheStorageEngineConnection_Caches:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::DOMCacheEngine::CacheInfosOrError", nullptr, false },
        };
    case MessageName::CacheStorageEngineConnection_RetrieveRecords:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::DOMCacheEngine::RecordsOrError", nullptr, false },
        };
    case MessageName::CacheStorageEngineConnection_DeleteMatchingRecords:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::DOMCacheEngine::RecordIdentifiersOrError", nullptr, false },
        };
    case MessageName::CacheStorageEngineConnection_PutRecords:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::DOMCacheEngine::RecordIdentifiersOrError", nullptr, false },
        };
    case MessageName::CacheStorageEngineConnection_ClearMemoryRepresentation:
        return Vector<ArgumentDescription> {
            { "error", "WebCore::DOMCacheEngine::Error", nullptr, true },
        };
    case MessageName::CacheStorageEngineConnection_EngineRepresentation:
        return Vector<ArgumentDescription> {
            { "representation", "String", nullptr, false },
        };
    case MessageName::NetworkStorageManager_Persisted:
        return Vector<ArgumentDescription> {
            { "persisted", "bool", nullptr, false },
        };
    case MessageName::NetworkStorageManager_Persist:
        return Vector<ArgumentDescription> {
            { "persisted", "bool", nullptr, false },
        };
    case MessageName::NetworkStorageManager_Estimate:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::StorageEstimate", nullptr, true },
        };
    case MessageName::NetworkStorageManager_FileSystemGetDirectory:
        return Vector<ArgumentDescription> {
            { "result", "Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>", nullptr, false },
        };
    case MessageName::NetworkStorageManager_IsSameEntry:
        return Vector<ArgumentDescription> {
            { "result", "bool", nullptr, false },
        };
    case MessageName::NetworkStorageManager_GetFileHandle:
        return Vector<ArgumentDescription> {
            { "result", "Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>", nullptr, false },
        };
    case MessageName::NetworkStorageManager_GetDirectoryHandle:
        return Vector<ArgumentDescription> {
            { "result", "Expected<WebCore::FileSystemHandleIdentifier, WebKit::FileSystemStorageError>", nullptr, false },
        };
    case MessageName::NetworkStorageManager_RemoveEntry:
        return Vector<ArgumentDescription> {
            { "result", "WebKit::FileSystemStorageError", nullptr, true },
        };
    case MessageName::NetworkStorageManager_Resolve:
        return Vector<ArgumentDescription> {
            { "result", "Expected<Vector<String>, WebKit::FileSystemStorageError>", nullptr, false },
        };
    case MessageName::NetworkStorageManager_Move:
        return Vector<ArgumentDescription> {
            { "result", "WebKit::FileSystemStorageError", nullptr, true },
        };
    case MessageName::NetworkStorageManager_GetFile:
        return Vector<ArgumentDescription> {
            { "result", "Expected<String, WebKit::FileSystemStorageError>", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CreateSyncAccessHandle:
        return Vector<ArgumentDescription> {
            { "result", "Expected<WebKit::FileSystemSyncAccessHandleInfo, WebKit::FileSystemStorageError>", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CloseSyncAccessHandle:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkStorageManager_RequestNewCapacityForSyncAccessHandle:
        return Vector<ArgumentDescription> {
            { "result", "uint64_t", nullptr, true },
        };
    case MessageName::NetworkStorageManager_GetHandleNames:
        return Vector<ArgumentDescription> {
            { "result", "Expected<Vector<String>, WebKit::FileSystemStorageError>", nullptr, false },
        };
    case MessageName::NetworkStorageManager_GetHandle:
        return Vector<ArgumentDescription> {
            { "result", "Expected<std::pair<WebCore::FileSystemHandleIdentifier, bool>, WebKit::FileSystemStorageError>", nullptr, false },
        };
    case MessageName::NetworkStorageManager_ConnectToStorageArea:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::StorageAreaIdentifier", nullptr, false },
            { "items", "HashMap<String, String>", nullptr, false },
            { "messageIdentifier", "uint64_t", nullptr, false },
        };
    case MessageName::NetworkStorageManager_ConnectToStorageAreaSync:
        return Vector<ArgumentDescription> {
            { "identifier", "WebKit::StorageAreaIdentifier", nullptr, false },
            { "items", "HashMap<String, String>", nullptr, false },
            { "messageIdentifier", "uint64_t", nullptr, false },
        };
    case MessageName::NetworkStorageManager_SetItem:
        return Vector<ArgumentDescription> {
            { "hasError", "bool", nullptr, false },
            { "allItems", "HashMap<String, String>", nullptr, false },
        };
    case MessageName::NetworkStorageManager_RemoveItem:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkStorageManager_Clear:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkStorageManager_CacheStorageOpenCache:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::DOMCacheEngine::CacheIdentifierOrError", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CacheStorageRemoveCache:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::DOMCacheEngine::RemoveCacheIdentifierOrError", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CacheStorageAllCaches:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::DOMCacheEngine::CacheInfosOrError", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CacheStorageRetrieveRecords:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::DOMCacheEngine::RecordsOrError", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CacheStorageRemoveRecords:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::DOMCacheEngine::RecordIdentifiersOrError", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CacheStoragePutRecords:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::DOMCacheEngine::RecordIdentifiersOrError", nullptr, false },
        };
    case MessageName::NetworkStorageManager_CacheStorageClearMemoryRepresentation:
        return Vector<ArgumentDescription> {
            { "error", "WebCore::DOMCacheEngine::Error", nullptr, true },
        };
    case MessageName::NetworkStorageManager_CacheStorageRepresentation:
        return Vector<ArgumentDescription> {
            { "representation", "String", nullptr, false },
        };
#if ENABLE(WEB_RTC)
#endif
#if USE(LIBWEBRTC)
#endif
#if USE(LIBWEBRTC)
#endif
#if ENABLE(WEB_RTC)
#endif
    case MessageName::AuxiliaryProcess_MainThreadPing:
        return Vector<ArgumentDescription> { };
#if ENABLE(IPC_TESTING_API)
    case MessageName::IPCConnectionTester_SyncMessage:
        return Vector<ArgumentDescription> {
            { "sameValue", "uint32_t", nullptr, false },
        };
#endif
#if ENABLE(IPC_TESTING_API)
    case MessageName::IPCStreamTester_SyncMessageReturningSharedMemory1:
        return Vector<ArgumentDescription> {
            { "handle", "WebKit::SharedMemory::Handle", nullptr, false },
        };
    case MessageName::IPCStreamTester_SyncCrashOnZero:
        return Vector<ArgumentDescription> {
            { "sameValue", "int32_t", nullptr, false },
        };
    case MessageName::IPCStreamTester_CheckAutoreleasePool:
        return Vector<ArgumentDescription> {
            { "previousUseCount", "int32_t", nullptr, false },
        };
    case MessageName::IPCStreamTester_AsyncMessage:
        return Vector<ArgumentDescription> {
            { "reply", "bool", nullptr, false },
        };
#endif
#if ENABLE(IPC_TESTING_API)
#endif
#if ENABLE(IPC_TESTING_API)
    case MessageName::IPCTester_StopMessageTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::IPCTester_ReleaseStreamTester:
        return Vector<ArgumentDescription> { };
    case MessageName::IPCTester_ReleaseConnectionTester:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(IPC_TESTING_API)
    case MessageName::IPCTesterReceiver_AsyncMessage:
        return Vector<ArgumentDescription> {
            { "reply", "uint32_t", nullptr, false },
        };
#endif
    case MessageName::NotificationManagerMessageHandler_RequestSystemNotificationPermission:
        return Vector<ArgumentDescription> {
            { "allowed", "bool", nullptr, false },
        };
    case MessageName::NotificationManagerMessageHandler_ShowNotification:
        return Vector<ArgumentDescription> { };
#if ENABLE(MEDIA_SESSION_COORDINATOR)
    case MessageName::RemoteMediaSessionCoordinatorProxy_Join:
        return Vector<ArgumentDescription> {
            { "error", "WebCore::ExceptionData", nullptr, true },
        };
    case MessageName::RemoteMediaSessionCoordinatorProxy_CoordinateSeekTo:
        return Vector<ArgumentDescription> {
            { "error", "WebCore::ExceptionData", nullptr, true },
        };
    case MessageName::RemoteMediaSessionCoordinatorProxy_CoordinatePlay:
        return Vector<ArgumentDescription> {
            { "error", "WebCore::ExceptionData", nullptr, true },
        };
    case MessageName::RemoteMediaSessionCoordinatorProxy_CoordinatePause:
        return Vector<ArgumentDescription> {
            { "error", "WebCore::ExceptionData", nullptr, true },
        };
    case MessageName::RemoteMediaSessionCoordinatorProxy_CoordinateSetTrack:
        return Vector<ArgumentDescription> {
            { "error", "WebCore::ExceptionData", nullptr, true },
        };
#endif
#if ENABLE(MEDIA_STREAM)
#endif
#if ENABLE(FULLSCREEN_API)
    case MessageName::WebFullScreenManagerProxy_SupportsFullScreen:
        return Vector<ArgumentDescription> {
            { "supportsFullScreen", "bool", nullptr, false },
        };
#endif
    case MessageName::WebLockRegistryProxy_AbortLockRequest:
        return Vector<ArgumentDescription> {
            { "wasAborted", "bool", nullptr, false },
        };
    case MessageName::WebLockRegistryProxy_Snapshot:
        return Vector<ArgumentDescription> {
            { "snapshot", "WebCore::WebLockManagerSnapshot", nullptr, false },
        };
    case MessageName::WebPageProxy_CreateNewPage:
        return Vector<ArgumentDescription> {
            { "newPageID", "WebCore::PageIdentifier", nullptr, true },
            { "newPageParameters", "WebKit::WebPageCreationParameters", nullptr, true },
        };
    case MessageName::WebPageProxy_RunJavaScriptAlert:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_RunJavaScriptConfirm:
        return Vector<ArgumentDescription> {
            { "result", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_RunJavaScriptPrompt:
        return Vector<ArgumentDescription> {
            { "result", "String", nullptr, false },
        };
    case MessageName::WebPageProxy_FocusFromServiceWorker:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_GetToolbarsAreVisible:
        return Vector<ArgumentDescription> {
            { "toolbarsAreVisible", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_GetMenuBarIsVisible:
        return Vector<ArgumentDescription> {
            { "menuBarIsVisible", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_GetStatusBarIsVisible:
        return Vector<ArgumentDescription> {
            { "statusBarIsVisible", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_GetWindowFrame:
        return Vector<ArgumentDescription> {
            { "windowFrame", "WebCore::FloatRect", nullptr, false },
        };
    case MessageName::WebPageProxy_ScreenToRootView:
        return Vector<ArgumentDescription> {
            { "windowPoint", "WebCore::IntPoint", nullptr, false },
        };
    case MessageName::WebPageProxy_RootViewToScreen:
        return Vector<ArgumentDescription> {
            { "screenFrame", "WebCore::IntRect", nullptr, false },
        };
    case MessageName::WebPageProxy_AccessibilityScreenToRootView:
        return Vector<ArgumentDescription> {
            { "windowPoint", "WebCore::IntPoint", nullptr, false },
        };
    case MessageName::WebPageProxy_RootViewToAccessibilityScreen:
        return Vector<ArgumentDescription> {
            { "screenFrame", "WebCore::IntRect", nullptr, false },
        };
    case MessageName::WebPageProxy_RunBeforeUnloadConfirmPanel:
        return Vector<ArgumentDescription> {
            { "shouldClose", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_ShowShareSheet:
        return Vector<ArgumentDescription> {
            { "granted", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_ShowContactPicker:
        return Vector<ArgumentDescription> {
            { "info", "Vector<WebCore::ContactInfo>", nullptr, true },
        };
    case MessageName::WebPageProxy_PrintFrame:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_DecidePolicyForNavigationActionSync:
        return Vector<ArgumentDescription> {
            { "PolicyDecision", "WebKit::PolicyDecision", nullptr, false },
        };
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebPageProxy_InterpretKeyEvent:
        return Vector<ArgumentDescription> {
            { "handled", "bool", nullptr, false },
        };
#endif
    case MessageName::WebPageProxy_BackForwardGoToItem:
        return Vector<ArgumentDescription> {
            { "counts", "WebKit::WebBackForwardListCounts", nullptr, false },
        };
    case MessageName::WebPageProxy_BackForwardItemAtIndex:
        return Vector<ArgumentDescription> {
            { "itemID", "WebCore::BackForwardItemIdentifier", nullptr, true },
        };
    case MessageName::WebPageProxy_BackForwardListContainsItem:
        return Vector<ArgumentDescription> {
            { "contains", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_BackForwardListCounts:
        return Vector<ArgumentDescription> {
            { "counts", "WebKit::WebBackForwardListCounts", nullptr, false },
        };
    case MessageName::WebPageProxy_CanUndoRedo:
        return Vector<ArgumentDescription> {
            { "result", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_ExecuteUndoRedo:
        return Vector<ArgumentDescription> { };
#if ENABLE(IMAGE_ANALYSIS)
    case MessageName::WebPageProxy_RequestTextRecognition:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::TextRecognitionResult", nullptr, false },
        };
#endif
#if ENABLE(MEDIA_CONTROLS_CONTEXT_MENUS) && USE(UICONTEXTMENU)
    case MessageName::WebPageProxy_ShowMediaControlsContextMenu:
        return Vector<ArgumentDescription> {
            { "selectedItemID", "WebCore::MediaControlsContextMenuItem::ID", nullptr, false },
        };
#endif
    case MessageName::WebPageProxy_RequestDOMPasteAccess:
        return Vector<ArgumentDescription> {
            { "response", "uint8_t", "WebCore::DOMPasteAccessResponse", false },
        };
    case MessageName::WebPageProxy_ExceededDatabaseQuota:
        return Vector<ArgumentDescription> {
            { "newQuota", "uint64_t", nullptr, false },
        };
    case MessageName::WebPageProxy_ReachedApplicationCacheOriginQuota:
        return Vector<ArgumentDescription> {
            { "newQuota", "uint64_t", nullptr, false },
        };
#if ENABLE(MEDIA_STREAM)
    case MessageName::WebPageProxy_EnumerateMediaDevicesForFrame:
        return Vector<ArgumentDescription> {
            { "devices", "Vector<WebCore::CaptureDevice>", nullptr, false },
            { "mediaDeviceIdentifierHashSalts", "WebCore::MediaDeviceHashSalts", nullptr, false },
        };
#endif
    case MessageName::WebPageProxy_RequestNotificationPermission:
        return Vector<ArgumentDescription> {
            { "allowed", "bool", nullptr, false },
        };
#if USE(UNIFIED_TEXT_CHECKING)
    case MessageName::WebPageProxy_CheckTextOfParagraph:
        return Vector<ArgumentDescription> {
            { "results", "Vector<WebCore::TextCheckingResult>", nullptr, false },
        };
#endif
    case MessageName::WebPageProxy_CheckSpellingOfString:
        return Vector<ArgumentDescription> {
            { "misspellingLocation", "int32_t", nullptr, false },
            { "misspellingLength", "int32_t", nullptr, false },
        };
    case MessageName::WebPageProxy_CheckGrammarOfString:
        return Vector<ArgumentDescription> {
            { "results", "Vector<WebCore::GrammarDetail>", nullptr, false },
            { "badGrammarLocation", "int32_t", nullptr, false },
            { "badGrammarLength", "int32_t", nullptr, false },
        };
    case MessageName::WebPageProxy_SpellingUIIsShowing:
        return Vector<ArgumentDescription> {
            { "isShowing", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_GetGuessesForWord:
        return Vector<ArgumentDescription> {
            { "guesses", "Vector<String>", nullptr, false },
        };
#if PLATFORM(COCOA)
    case MessageName::WebPageProxy_ExecuteSavedCommandBySelector:
        return Vector<ArgumentDescription> {
            { "handled", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_GetIsSpeaking:
        return Vector<ArgumentDescription> {
            { "isSpeaking", "bool", nullptr, false },
        };
#endif
#if USE(APPKIT)
    case MessageName::WebPageProxy_SubstitutionsPanelIsShowing:
        return Vector<ArgumentDescription> {
            { "isShowing", "bool", nullptr, false },
        };
#endif
#if PLATFORM(MAC)
    case MessageName::WebPageProxy_DismissCorrectionPanelSoon:
        return Vector<ArgumentDescription> {
            { "result", "String", nullptr, false },
        };
#endif
#if USE(DICTATION_ALTERNATIVES)
    case MessageName::WebPageProxy_DictationAlternatives:
        return Vector<ArgumentDescription> {
            { "alternatives", "Vector<String>", nullptr, false },
        };
#endif
    case MessageName::WebPageProxy_LoadRecentSearches:
        return Vector<ArgumentDescription> {
            { "result", "Vector<WebCore::RecentSearch>", nullptr, false },
        };
#if ENABLE(PDFKIT_PLUGIN)
    case MessageName::WebPageProxy_ShowPDFContextMenu:
        return Vector<ArgumentDescription> {
            { "selectedItem", "int32_t", nullptr, true },
        };
#endif
#if ENABLE(WEB_CRYPTO)
    case MessageName::WebPageProxy_WrapCryptoKey:
        return Vector<ArgumentDescription> {
            { "succeeded", "bool", nullptr, false },
            { "wrappedKey", "Vector<uint8_t>", nullptr, false },
        };
    case MessageName::WebPageProxy_UnwrapCryptoKey:
        return Vector<ArgumentDescription> {
            { "succeeded", "bool", nullptr, false },
            { "key", "Vector<uint8_t>", nullptr, false },
        };
#endif
#if USE(QUICK_LOOK)
    case MessageName::WebPageProxy_RequestPasswordForQuickLookDocumentInMainFrame:
        return Vector<ArgumentDescription> {
            { "password", "String", nullptr, false },
        };
#endif
    case MessageName::WebPageProxy_HandleSynchronousMessage:
        return Vector<ArgumentDescription> {
            { "returnData", "WebKit::UserData", nullptr, false },
        };
    case MessageName::WebPageProxy_LoadSynchronousURLSchemeTask:
        return Vector<ArgumentDescription> {
            { "response", "WebCore::ResourceResponse", nullptr, false },
            { "error", "WebCore::ResourceError", nullptr, false },
            { "data", "Vector<uint8_t>", nullptr, false },
        };
#if ENABLE(DEVICE_ORIENTATION)
    case MessageName::WebPageProxy_ShouldAllowDeviceOrientationAndMotionAccess:
        return Vector<ArgumentDescription> {
            { "permissionState", "uint8_t", "WebCore::DeviceOrientationOrMotionPermissionState", false },
        };
#endif
#if ENABLE(ATTACHMENT_ELEMENT)
    case MessageName::WebPageProxy_SerializedAttachmentDataForIdentifiers:
        return Vector<ArgumentDescription> {
            { "seralizedData", "Vector<WebCore::SerializedAttachmentData>", nullptr, false },
        };
#endif
#if ENABLE(SPEECH_SYNTHESIS)
    case MessageName::WebPageProxy_SpeechSynthesisVoiceList:
        return Vector<ArgumentDescription> {
            { "voiceList", "Vector<WebKit::WebSpeechSynthesisVoice>", nullptr, false },
        };
    case MessageName::WebPageProxy_SpeechSynthesisSpeak:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_SpeechSynthesisSetFinishedCallback:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_SpeechSynthesisPause:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPageProxy_SpeechSynthesisResume:
        return Vector<ArgumentDescription> { };
#endif
#if PLATFORM(GTK)
    case MessageName::WebPageProxy_ShowEmojiPicker:
        return Vector<ArgumentDescription> {
            { "result", "String", nullptr, false },
        };
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    case MessageName::WebPageProxy_SendMessageToWebViewWithReply:
        return Vector<ArgumentDescription> {
            { "replyMessage", "WebKit::UserMessage", nullptr, false },
        };
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW_MAC)
    case MessageName::WebPageProxy_ModelElementCreateRemotePreview:
        return Vector<ArgumentDescription> {
            { "result", "Expected<std::pair<String, uint32_t>, WebCore::ResourceError>", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementLoadRemotePreview:
        return Vector<ArgumentDescription> {
            { "error", "WebCore::ResourceError", nullptr, true },
        };
    case MessageName::WebPageProxy_ModelElementSizeDidChange:
        return Vector<ArgumentDescription> {
            { "result", "Expected<MachSendRight, WebCore::ResourceError>", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelInlinePreviewUUIDs:
        return Vector<ArgumentDescription> {
            { "uuids", "Vector<String>", nullptr, false },
        };
#endif
#if ENABLE(ARKIT_INLINE_PREVIEW)
    case MessageName::WebPageProxy_ModelElementGetCamera:
        return Vector<ArgumentDescription> {
            { "result", "Expected<WebCore::HTMLModelElementCamera, WebCore::ResourceError>", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementSetCamera:
        return Vector<ArgumentDescription> {
            { "success", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementIsPlayingAnimation:
        return Vector<ArgumentDescription> {
            { "result", "Expected<bool, WebCore::ResourceError>", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementSetAnimationIsPlaying:
        return Vector<ArgumentDescription> {
            { "success", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementIsLoopingAnimation:
        return Vector<ArgumentDescription> {
            { "result", "Expected<bool, WebCore::ResourceError>", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementSetIsLoopingAnimation:
        return Vector<ArgumentDescription> {
            { "success", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementAnimationDuration:
        return Vector<ArgumentDescription> {
            { "result", "Expected<Seconds, WebCore::ResourceError>", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementAnimationCurrentTime:
        return Vector<ArgumentDescription> {
            { "result", "Expected<Seconds, WebCore::ResourceError>", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementSetAnimationCurrentTime:
        return Vector<ArgumentDescription> {
            { "success", "bool", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementHasAudio:
        return Vector<ArgumentDescription> {
            { "result", "Expected<bool, WebCore::ResourceError>", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementIsMuted:
        return Vector<ArgumentDescription> {
            { "result", "Expected<bool, WebCore::ResourceError>", nullptr, false },
        };
    case MessageName::WebPageProxy_ModelElementSetIsMuted:
        return Vector<ArgumentDescription> {
            { "success", "bool", nullptr, false },
        };
#endif
#if ENABLE(APPLE_PAY_AMS_UI)
    case MessageName::WebPageProxy_StartApplePayAMSUISession:
        return Vector<ArgumentDescription> {
            { "result", "bool", nullptr, true },
        };
#endif
    case MessageName::WebPageProxy_RequestCookieConsent:
        return Vector<ArgumentDescription> {
            { "result", "uint8_t", "WebCore::CookieConsentDecisionResult", false },
        };
    case MessageName::WebPageProxy_ClassifyModalContainerControls:
        return Vector<ArgumentDescription> {
            { "types", "Vector<WebCore::ModalContainerControlType>", nullptr, false },
        };
    case MessageName::WebPageProxy_DecidePolicyForModalContainer:
        return Vector<ArgumentDescription> {
            { "decision", "uint8_t", "WebCore::ModalContainerDecision", false },
        };
    case MessageName::WebPasteboardProxy_WriteCustomData:
        return Vector<ArgumentDescription> {
            { "changeCount", "int64_t", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_TypesSafeForDOMToReadAndWrite:
        return Vector<ArgumentDescription> {
            { "types", "Vector<String>", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_AllPasteboardItemInfo:
        return Vector<ArgumentDescription> {
            { "allInfo", "Vector<WebCore::PasteboardItemInfo>", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_InformationForItemAtIndex:
        return Vector<ArgumentDescription> {
            { "info", "WebCore::PasteboardItemInfo", nullptr, true },
        };
    case MessageName::WebPasteboardProxy_GetPasteboardItemsCount:
        return Vector<ArgumentDescription> {
            { "itemsCount", "uint64_t", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_ReadStringFromPasteboard:
        return Vector<ArgumentDescription> {
            { "string", "String", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_ReadURLFromPasteboard:
        return Vector<ArgumentDescription> {
            { "url", "String", nullptr, false },
            { "title", "String", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_ReadBufferFromPasteboard:
        return Vector<ArgumentDescription> {
            { "buffer", "RefPtr<WebCore::SharedBuffer>", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_ContainsStringSafeForDOMToReadForType:
        return Vector<ArgumentDescription> {
            { "result", "bool", nullptr, false },
        };
#if PLATFORM(COCOA)
    case MessageName::WebPasteboardProxy_GetNumberOfFiles:
        return Vector<ArgumentDescription> {
            { "numberOfFiles", "uint64_t", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_GetPasteboardTypes:
        return Vector<ArgumentDescription> {
            { "types", "Vector<String>", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_GetPasteboardPathnamesForType:
        return Vector<ArgumentDescription> {
            { "pathnames", "Vector<String>", nullptr, false },
            { "sandboxExtensions", "Vector<WebKit::SandboxExtension::Handle>", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_GetPasteboardStringForType:
        return Vector<ArgumentDescription> {
            { "string", "String", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_GetPasteboardStringsForType:
        return Vector<ArgumentDescription> {
            { "strings", "Vector<String>", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_GetPasteboardBufferForType:
        return Vector<ArgumentDescription> {
            { "buffer", "RefPtr<WebCore::SharedBuffer>", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_GetPasteboardChangeCount:
        return Vector<ArgumentDescription> {
            { "changeCount", "int64_t", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_GetPasteboardColor:
        return Vector<ArgumentDescription> {
            { "color", "WebCore::Color", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_GetPasteboardURL:
        return Vector<ArgumentDescription> {
            { "urlString", "String", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_AddPasteboardTypes:
        return Vector<ArgumentDescription> {
            { "changeCount", "int64_t", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_SetPasteboardTypes:
        return Vector<ArgumentDescription> {
            { "changeCount", "int64_t", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_SetPasteboardURL:
        return Vector<ArgumentDescription> {
            { "changeCount", "int64_t", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_SetPasteboardColor:
        return Vector<ArgumentDescription> {
            { "changeCount", "int64_t", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_SetPasteboardStringForType:
        return Vector<ArgumentDescription> {
            { "changeCount", "int64_t", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_SetPasteboardBufferForType:
        return Vector<ArgumentDescription> {
            { "changeCount", "int64_t", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_ContainsURLStringSuitableForLoading:
        return Vector<ArgumentDescription> {
            { "result", "bool", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_URLStringSuitableForLoading:
        return Vector<ArgumentDescription> {
            { "url", "String", nullptr, false },
            { "title", "String", nullptr, false },
        };
#endif
#if (PLATFORM(COCOA) && ENABLE(IPC_TESTING_API))
    case MessageName::WebPasteboardProxy_TestIPCSharedMemory:
        return Vector<ArgumentDescription> {
            { "changeCount", "int64_t", nullptr, false },
            { "buffer", "String", nullptr, false },
        };
#endif
#if PLATFORM(GTK)
    case MessageName::WebPasteboardProxy_GetTypes:
        return Vector<ArgumentDescription> {
            { "types", "Vector<String>", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_ReadText:
        return Vector<ArgumentDescription> {
            { "text", "String", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_ReadFilePaths:
        return Vector<ArgumentDescription> {
            { "types", "Vector<String>", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_ReadBuffer:
        return Vector<ArgumentDescription> {
            { "buffer", "RefPtr<WebCore::SharedBuffer>", nullptr, false },
        };
    case MessageName::WebPasteboardProxy_GetPasteboardChangeCount:
        return Vector<ArgumentDescription> {
            { "changeCount", "int64_t", nullptr, false },
        };
#endif
#if USE(LIBWPE)
    case MessageName::WebPasteboardProxy_GetPasteboardTypes:
        return Vector<ArgumentDescription> {
            { "types", "Vector<String>", nullptr, false },
        };
#endif
    case MessageName::WebPermissionControllerProxy_Query:
        return Vector<ArgumentDescription> {
            { "state", "WebCore::PermissionState", nullptr, true },
        };
    case MessageName::WebProcessPool_HandleSynchronousMessage:
        return Vector<ArgumentDescription> {
            { "returnData", "WebKit::UserData", nullptr, false },
        };
#if ENABLE(GAMEPAD)
    case MessageName::WebProcessPool_StoppedUsingGamepads:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcessPool_PlayGamepadEffect:
        return Vector<ArgumentDescription> {
            { "success", "bool", nullptr, false },
        };
    case MessageName::WebProcessPool_StopGamepadEffects:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebProcessProxy_ShouldTerminate:
        return Vector<ArgumentDescription> {
            { "shouldTerminate", "bool", nullptr, false },
        };
    case MessageName::WebProcessProxy_GetNetworkProcessConnection:
        return Vector<ArgumentDescription> {
            { "connectionInfo", "WebKit::NetworkProcessConnectionInfo", nullptr, false },
        };
#if PLATFORM(GTK) || PLATFORM(WPE)
    case MessageName::WebProcessProxy_SendMessageToWebContextWithReply:
        return Vector<ArgumentDescription> {
            { "replyMessage", "WebKit::UserMessage", nullptr, false },
        };
#endif
#if PLATFORM(MAC)
    case MessageName::WebProcessProxy_IsAXAuthenticated:
        return Vector<ArgumentDescription> {
            { "authenticated", "bool", nullptr, false },
        };
#endif
    case MessageName::WebProcessProxy_GetNotifications:
        return Vector<ArgumentDescription> {
            { "result", "Vector<WebCore::NotificationData>", nullptr, false },
        };
    case MessageName::WebScreenOrientationManagerProxy_CurrentOrientation:
        return Vector<ArgumentDescription> {
            { "orientation", "uint8_t", "WebCore::ScreenOrientationType", false },
        };
    case MessageName::WebScreenOrientationManagerProxy_Lock:
        return Vector<ArgumentDescription> {
            { "exception", "WebCore::Exception", nullptr, true },
        };
    case MessageName::DownloadProxy_DecideDestinationWithSuggestedFilename:
        return Vector<ArgumentDescription> {
            { "filename", "String", nullptr, false },
            { "handle", "WebKit::SandboxExtension::Handle", nullptr, false },
            { "allowOverwrite", "bool", "WebKit::AllowOverwrite", false },
        };
#if ENABLE(GPU_PROCESS)
#if ENABLE(MEDIA_STREAM) && PLATFORM(IOS_FAMILY)
    case MessageName::GPUProcessProxy_StatusBarWasTapped:
        return Vector<ArgumentDescription> { };
#endif
#endif
    case MessageName::RemoteWebInspectorUIProxy_Load:
        return Vector<ArgumentDescription> {
            { "content", "String", nullptr, false },
        };
    case MessageName::RemoteWebInspectorUIProxy_PickColorFromScreen:
        return Vector<ArgumentDescription> {
            { "color", "WebCore::Color", nullptr, true },
        };
    case MessageName::WebInspectorUIProxy_Load:
        return Vector<ArgumentDescription> {
            { "content", "String", nullptr, false },
        };
    case MessageName::WebInspectorUIProxy_PickColorFromScreen:
        return Vector<ArgumentDescription> {
            { "color", "WebCore::Color", nullptr, true },
        };
    case MessageName::NetworkProcessProxy_TestProcessIncomingSyncMessagesWhenWaitingForSyncReply:
        return Vector<ArgumentDescription> {
            { "handled", "bool", nullptr, false },
        };
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::NetworkProcessProxy_RequestStorageAccessConfirm:
        return Vector<ArgumentDescription> {
            { "userDidGrantAccess", "bool", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_DeleteWebsiteDataInUIProcessForRegistrableDomains:
        return Vector<ArgumentDescription> {
            { "domainsWithMatchingDataRecords", "HashSet<WebCore::RegistrableDomain>", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_SetDomainsWithCrossPageStorageAccess:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::NetworkProcessProxy_EstablishRemoteWorkerContextConnectionToNetworkProcess:
        return Vector<ArgumentDescription> {
            { "remoteProcessIdentifier", "WebCore::ProcessIdentifier", nullptr, false },
        };
#if ENABLE(APP_BOUND_DOMAINS)
    case MessageName::NetworkProcessProxy_GetAppBoundDomains:
        return Vector<ArgumentDescription> {
            { "appBoundDomains", "HashSet<WebCore::RegistrableDomain>", nullptr, false },
        };
#endif
    case MessageName::NetworkProcessProxy_RequestStorageSpace:
        return Vector<ArgumentDescription> {
            { "newQuota", "uint64_t", nullptr, true },
        };
    case MessageName::NetworkProcessProxy_TriggerBrowsingContextGroupSwitchForNavigation:
        return Vector<ArgumentDescription> {
            { "success", "bool", nullptr, false },
        };
#if ENABLE(APPLE_PAY_REMOTE_UI_USES_SCENE)
    case MessageName::NetworkProcessProxy_GetWindowSceneIdentifierForPaymentPresentation:
        return Vector<ArgumentDescription> {
            { "sceneID", "String", nullptr, false },
        };
#endif
    case MessageName::NetworkProcessProxy_DataTaskReceivedChallenge:
        return Vector<ArgumentDescription> {
            { "disposition", "uint8_t", "WebKit::AuthenticationChallengeDisposition", false },
            { "credential", "WebCore::Credential", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_DataTaskWillPerformHTTPRedirection:
        return Vector<ArgumentDescription> {
            { "allowed", "bool", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_DataTaskDidReceiveResponse:
        return Vector<ArgumentDescription> {
            { "allowed", "bool", nullptr, false },
        };
    case MessageName::NetworkProcessProxy_OpenWindowFromServiceWorker:
        return Vector<ArgumentDescription> {
            { "newPage", "WebCore::PageIdentifier", nullptr, true },
        };
    case MessageName::NetworkProcessProxy_NavigateServiceWorkerClient:
        return Vector<ArgumentDescription> {
            { "page", "WebCore::PageIdentifier", nullptr, true },
            { "frame", "WebCore::FrameIdentifier", nullptr, true },
        };
    case MessageName::NetworkProcessProxy_DeleteWebsiteDataInWebProcessesForOrigin:
        return Vector<ArgumentDescription> { };
    case MessageName::NetworkProcessProxy_ReloadExecutionContextsForOrigin:
        return Vector<ArgumentDescription> { };
    case MessageName::WebUserContentControllerProxy_DidPostMessage:
        return Vector<ArgumentDescription> {
            { "resultValue", "IPC::DataReference", nullptr, false },
            { "errorMessage", "String", nullptr, false },
        };
#if ENABLE(WEBXR)
    case MessageName::PlatformXRSystem_EnumerateImmersiveXRDevices:
        return Vector<ArgumentDescription> {
            { "devicesInfos", "Vector<WebKit::XRDeviceInfo>", nullptr, false },
        };
    case MessageName::PlatformXRSystem_RequestPermissionOnSessionFeatures:
        return Vector<ArgumentDescription> {
            { "userGranted", "Vector<PlatformXR::SessionFeature>", nullptr, true },
        };
    case MessageName::PlatformXRSystem_RequestFrame:
        return Vector<ArgumentDescription> {
            { "frameData", "PlatformXR::Device::FrameData", nullptr, false },
        };
#endif
    case MessageName::WebProcess_FetchWebsiteData:
        return Vector<ArgumentDescription> {
            { "websiteData", "WebKit::WebsiteData", nullptr, false },
        };
    case MessageName::WebProcess_DeleteWebsiteData:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_DeleteWebsiteDataForOrigins:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_DeleteWebsiteDataForOrigin:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_ReloadExecutionContextsForOrigin:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_DeleteAllCookies:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_PrepareToSuspend:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_EstablishRemoteWorkerContextConnectionToNetworkProcess:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_GetActivePagesOriginsForTesting:
        return Vector<ArgumentDescription> {
            { "activeOrigins", "Vector<String>", nullptr, false },
        };
    case MessageName::WebProcess_IsJITEnabled:
        return Vector<ArgumentDescription> {
            { "enabled", "bool", nullptr, false },
        };
    case MessageName::WebProcess_ClearCachedPage:
        return Vector<ArgumentDescription> { };
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::WebProcess_SeedResourceLoadStatisticsForTesting:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_SetThirdPartyCookieBlockingMode:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_SetDomainsWithCrossPageStorageAccess:
        return Vector<ArgumentDescription> { };
    case MessageName::WebProcess_SendResourceLoadStatisticsDataImmediately:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebAutomationSessionProxy_ResolveChildFrameWithOrdinal:
        return Vector<ArgumentDescription> {
            { "errorType", "String", nullptr, true },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
        };
    case MessageName::WebAutomationSessionProxy_ResolveChildFrameWithNodeHandle:
        return Vector<ArgumentDescription> {
            { "errorType", "String", nullptr, true },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
        };
    case MessageName::WebAutomationSessionProxy_ResolveChildFrameWithName:
        return Vector<ArgumentDescription> {
            { "errorType", "String", nullptr, true },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
        };
    case MessageName::WebAutomationSessionProxy_ResolveParentFrame:
        return Vector<ArgumentDescription> {
            { "errorType", "String", nullptr, true },
            { "frameID", "WebCore::FrameIdentifier", nullptr, true },
        };
    case MessageName::WebAutomationSessionProxy_ComputeElementLayout:
        return Vector<ArgumentDescription> {
            { "errorType", "String", nullptr, true },
            { "rect", "WebCore::FloatRect", nullptr, false },
            { "inViewCenterPoint", "WebCore::IntPoint", nullptr, true },
            { "isObscured", "bool", nullptr, false },
        };
    case MessageName::WebAutomationSessionProxy_GetComputedRole:
        return Vector<ArgumentDescription> {
            { "role", "String", nullptr, true },
            { "errorType", "String", nullptr, true },
        };
    case MessageName::WebAutomationSessionProxy_GetComputedLabel:
        return Vector<ArgumentDescription> {
            { "role", "String", nullptr, true },
            { "errorType", "String", nullptr, true },
        };
    case MessageName::WebAutomationSessionProxy_SelectOptionElement:
        return Vector<ArgumentDescription> {
            { "errorType", "String", nullptr, true },
        };
    case MessageName::WebAutomationSessionProxy_SetFilesForInputFileUpload:
        return Vector<ArgumentDescription> {
            { "errorType", "String", nullptr, true },
        };
    case MessageName::WebAutomationSessionProxy_SnapshotRectForScreenshot:
        return Vector<ArgumentDescription> {
            { "errorType", "String", nullptr, true },
            { "rect", "WebCore::IntRect", nullptr, false },
        };
    case MessageName::WebAutomationSessionProxy_GetCookiesForFrame:
        return Vector<ArgumentDescription> {
            { "errorType", "String", nullptr, true },
            { "cookies", "Vector<WebCore::Cookie>", nullptr, false },
        };
    case MessageName::WebAutomationSessionProxy_DeleteCookie:
        return Vector<ArgumentDescription> {
            { "errorType", "String", nullptr, true },
        };
#if ENABLE(FULLSCREEN_API)
#endif
#if ENABLE(GPU_PROCESS)
#if ENABLE(ROUTING_ARBITRATION)
    case MessageName::GPUProcessConnection_BeginRoutingArbitrationWithCategory:
        return Vector<ArgumentDescription> {
            { "error", "uint8_t", "WebKit::AudioSessionRoutingArbitratorProxy::RoutingArbitrationError", false },
            { "defaultRouteChanged", "bool", "WebKit::AudioSessionRoutingArbitratorProxy::DefaultRouteChanged", false },
        };
#endif
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(WEBGL)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
    case MessageName::MediaPlayerPrivateRemote_SendH2Ping:
        return Vector<ArgumentDescription> {
            { "result", "Expected<Seconds, WebCore::ResourceError>", nullptr, false },
        };
#if PLATFORM(IOS_FAMILY)
    case MessageName::MediaPlayerPrivateRemote_GetRawCookies:
        return Vector<ArgumentDescription> {
            { "result", "Vector<WebCore::Cookie>", nullptr, false },
        };
#endif
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
#endif
#if ENABLE(GPU_PROCESS) && USE(AUDIO_SESSION)
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(ENCRYPTED_MEDIA)
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(LEGACY_ENCRYPTED_MEDIA)
#endif
#if ENABLE(GPU_PROCESS)
#endif
#if ENABLE(GPU_PROCESS) && ENABLE(MEDIA_SOURCE)
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateDidReceiveInitializationSegment:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::SourceBufferPrivateClient::ReceiveResult", nullptr, false },
        };
    case MessageName::SourceBufferPrivateRemote_SourceBufferPrivateDurationChanged:
        return Vector<ArgumentDescription> { };
#endif
#if USE(LIBWEBRTC) && PLATFORM(COCOA) && ENABLE(GPU_PROCESS)
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(VIDEO)
#endif
#if PLATFORM(COCOA) && ENABLE(GPU_PROCESS) && ENABLE(MEDIA_STREAM)
#endif
#if ENABLE(MEDIA_SESSION_COORDINATOR)
    case MessageName::RemoteMediaSessionCoordinator_SeekSessionToTime:
        return Vector<ArgumentDescription> {
            { "result", "bool", nullptr, false },
        };
    case MessageName::RemoteMediaSessionCoordinator_PlaySession:
        return Vector<ArgumentDescription> {
            { "result", "bool", nullptr, false },
        };
    case MessageName::RemoteMediaSessionCoordinator_PauseSession:
        return Vector<ArgumentDescription> {
            { "result", "bool", nullptr, false },
        };
    case MessageName::RemoteMediaSessionCoordinator_SetSessionTrack:
        return Vector<ArgumentDescription> {
            { "result", "bool", nullptr, false },
        };
#endif
#if USE(LIBWEBRTC)
#endif
#if ENABLE(WEB_RTC)
#endif
#if ENABLE(WEB_RTC)
#endif
#if USE(LIBWEBRTC)
#endif
#if USE(LIBWEBRTC)
#endif
#if ENABLE(WEB_RTC)
    case MessageName::NetworkProcessConnection_ConnectToRTCDataChannelRemoteSource:
        return Vector<ArgumentDescription> {
            { "result", "bool", nullptr, true },
        };
#endif
#if ENABLE(MEDIA_STREAM)
#endif
    case MessageName::WebSharedWorkerContextManagerConnection_PostConnectEvent:
        return Vector<ArgumentDescription> {
            { "success", "bool", nullptr, false },
        };
    case MessageName::WebSharedWorkerObjectConnection_FetchScriptInClient:
        return Vector<ArgumentDescription> {
            { "fetchResult", "WebCore::WorkerFetchResult", nullptr, false },
            { "initializationData", "WebCore::WorkerInitializationData", nullptr, false },
        };
#if ENABLE(SERVICE_WORKER)
    case MessageName::WebSWClientConnection_SetServiceWorkerClientIsControlled:
        return Vector<ArgumentDescription> {
            { "isSuccess", "bool", nullptr, false },
        };
    case MessageName::WebSWClientConnection_FocusServiceWorkerClient:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::ServiceWorkerClientData", nullptr, true },
        };
#endif
#if ENABLE(SERVICE_WORKER)
    case MessageName::WebSWContextManagerConnection_FirePushEvent:
        return Vector<ArgumentDescription> {
            { "result", "bool", nullptr, false },
        };
    case MessageName::WebSWContextManagerConnection_FireNotificationEvent:
        return Vector<ArgumentDescription> {
            { "result", "bool", nullptr, false },
        };
#endif
    case MessageName::WebBroadcastChannelRegistry_PostMessageToRemote:
        return Vector<ArgumentDescription> { };
#if PLATFORM(COCOA)
    case MessageName::DrawingArea_UpdateGeometry:
        return Vector<ArgumentDescription> { };
    case MessageName::DrawingArea_DispatchAfterEnsuringDrawing:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(IOS_TOUCH_EVENTS)
    case MessageName::EventDispatcher_TouchEvent:
        return Vector<ArgumentDescription> {
            { "handled", "bool", nullptr, false },
        };
#endif
    case MessageName::WebFrame_GetFrameInfo:
        return Vector<ArgumentDescription> {
            { "data", "WebKit::FrameInfoData", nullptr, false },
        };
    case MessageName::WebPage_SetInitialFocus:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_SetActivityState:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_TestProcessIncomingSyncMessagesWhenWaitingForSyncReply:
        return Vector<ArgumentDescription> {
            { "handled", "bool", nullptr, false },
        };
    case MessageName::WebPage_ExecuteEditCommandWithCallback:
        return Vector<ArgumentDescription> { };
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebPage_SelectWithGesture:
        return Vector<ArgumentDescription> {
            { "point", "WebCore::IntPoint", nullptr, false },
            { "gestureType", "uint8_t", "WebKit::GestureType", false },
            { "gestureState", "uint8_t", "WebKit::GestureRecognizerState", false },
            { "flags", "OptionSet<WebKit::SelectionFlags>", nullptr, false },
        };
    case MessageName::WebPage_UpdateSelectionWithTouches:
        return Vector<ArgumentDescription> {
            { "point", "WebCore::IntPoint", nullptr, false },
            { "touch", "uint8_t", "WebKit::SelectionTouch", false },
            { "selectionFlags", "OptionSet<WebKit::SelectionFlags>", nullptr, false },
        };
    case MessageName::WebPage_SelectWithTwoTouches:
        return Vector<ArgumentDescription> {
            { "point", "WebCore::IntPoint", nullptr, false },
            { "gestureType", "uint8_t", "WebKit::GestureType", false },
            { "gestureState", "uint8_t", "WebKit::GestureRecognizerState", false },
            { "selectionFlags", "OptionSet<WebKit::SelectionFlags>", nullptr, false },
        };
    case MessageName::WebPage_ExtendSelection:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_ExtendSelectionForReplacement:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_MoveSelectionByOffset:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_SelectTextWithGranularityAtPoint:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_SelectPositionAtBoundaryWithDirection:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_MoveSelectionAtBoundaryWithDirection:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_SelectPositionAtPoint:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_BeginSelectionInDirection:
        return Vector<ArgumentDescription> {
            { "endIsMoving", "bool", nullptr, false },
        };
    case MessageName::WebPage_UpdateSelectionWithExtentPoint:
        return Vector<ArgumentDescription> {
            { "endIsMoving", "bool", nullptr, false },
        };
    case MessageName::WebPage_UpdateSelectionWithExtentPointAndBoundary:
        return Vector<ArgumentDescription> {
            { "endIsMoving", "bool", nullptr, false },
        };
    case MessageName::WebPage_RequestDictationContext:
        return Vector<ArgumentDescription> {
            { "selectedText", "String", nullptr, false },
            { "textBefore", "String", nullptr, false },
            { "textAfter", "String", nullptr, false },
        };
#endif
#if (PLATFORM(IOS_FAMILY) && ENABLE(REVEAL))
    case MessageName::WebPage_RequestRVItemInCurrentSelectedRange:
        return Vector<ArgumentDescription> {
            { "item", "WebKit::RevealItem", nullptr, false },
        };
    case MessageName::WebPage_PrepareSelectionForContextMenuWithLocationInView:
        return Vector<ArgumentDescription> {
            { "shouldShowMenu", "bool", nullptr, false },
            { "item", "WebKit::RevealItem", nullptr, false },
        };
#endif
#if PLATFORM(IOS_FAMILY)
    case MessageName::WebPage_RequestAutocorrectionData:
        return Vector<ArgumentDescription> {
            { "data", "WebKit::WebAutocorrectionData", nullptr, false },
        };
    case MessageName::WebPage_ApplyAutocorrection:
        return Vector<ArgumentDescription> {
            { "string", "String", nullptr, false },
        };
    case MessageName::WebPage_SyncApplyAutocorrection:
        return Vector<ArgumentDescription> {
            { "autocorrectionApplied", "bool", nullptr, false },
        };
    case MessageName::WebPage_RequestEvasionRectsAboveSelection:
        return Vector<ArgumentDescription> {
            { "rects", "Vector<WebCore::FloatRect>", nullptr, false },
        };
    case MessageName::WebPage_PerformActionOnElement:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_FocusNextFocusedElement:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_GetSelectionContext:
        return Vector<ArgumentDescription> {
            { "selectedText", "String", nullptr, false },
            { "textBefore", "String", nullptr, false },
            { "textAfter", "String", nullptr, false },
        };
    case MessageName::WebPage_GetRectsForGranularityWithSelectionOffset:
        return Vector<ArgumentDescription> {
            { "rect", "Vector<WebCore::SelectionGeometry>", nullptr, false },
        };
    case MessageName::WebPage_GetRectsAtSelectionOffsetWithText:
        return Vector<ArgumentDescription> {
            { "rect", "Vector<WebCore::SelectionGeometry>", nullptr, false },
        };
    case MessageName::WebPage_RequestFocusedElementInformation:
        return Vector<ArgumentDescription> {
            { "info", "WebKit::FocusedElementInformation", nullptr, true },
        };
    case MessageName::WebPage_UpdateSelectionWithDelta:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_RequestDocumentEditingContext:
        return Vector<ArgumentDescription> {
            { "response", "WebKit::DocumentEditingContext", nullptr, false },
        };
    case MessageName::WebPage_InsertTextPlaceholder:
        return Vector<ArgumentDescription> {
            { "placeholder", "WebCore::ElementContext", nullptr, true },
        };
    case MessageName::WebPage_RemoveTextPlaceholder:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_TextInputContextsInRect:
        return Vector<ArgumentDescription> {
            { "contexts", "Vector<WebCore::ElementContext>", nullptr, false },
        };
    case MessageName::WebPage_FocusTextInputContextAndPlaceCaret:
        return Vector<ArgumentDescription> {
            { "success", "bool", nullptr, false },
        };
    case MessageName::WebPage_ClearServiceWorkerEntitlementOverride:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebPage_RequestImageBitmap:
        return Vector<ArgumentDescription> {
            { "image", "WebKit::ShareableBitmapHandle", nullptr, false },
            { "sourceMIMEType", "String", nullptr, false },
        };
    case MessageName::WebPage_GetPDFFirstPageSize:
        return Vector<ArgumentDescription> {
            { "size", "WebCore::FloatSize", nullptr, false },
        };
    case MessageName::WebPage_GetContentsAsString:
        return Vector<ArgumentDescription> {
            { "string", "String", nullptr, false },
        };
#if PLATFORM(COCOA)
    case MessageName::WebPage_GetContentsAsAttributedString:
        return Vector<ArgumentDescription> {
            { "result", "WebCore::AttributedString", nullptr, false },
        };
#endif
#if ENABLE(MHTML)
    case MessageName::WebPage_GetContentsAsMHTMLData:
        return Vector<ArgumentDescription> {
            { "data", "IPC::SharedBufferReference", nullptr, false },
        };
#endif
    case MessageName::WebPage_GetMainResourceDataOfFrame:
        return Vector<ArgumentDescription> {
            { "buffer", "IPC::SharedBufferReference", nullptr, true },
        };
    case MessageName::WebPage_GetResourceDataFromFrame:
        return Vector<ArgumentDescription> {
            { "buffer", "IPC::SharedBufferReference", nullptr, true },
        };
    case MessageName::WebPage_GetRenderTreeExternalRepresentation:
        return Vector<ArgumentDescription> {
            { "string", "String", nullptr, false },
        };
    case MessageName::WebPage_GetSelectionOrContentsAsString:
        return Vector<ArgumentDescription> {
            { "string", "String", nullptr, false },
        };
    case MessageName::WebPage_GetSelectionAsWebArchiveData:
        return Vector<ArgumentDescription> {
            { "data", "IPC::SharedBufferReference", nullptr, true },
        };
    case MessageName::WebPage_GetSourceForFrame:
        return Vector<ArgumentDescription> {
            { "string", "String", nullptr, false },
        };
    case MessageName::WebPage_GetWebArchiveOfFrame:
        return Vector<ArgumentDescription> {
            { "dataReference", "IPC::SharedBufferReference", nullptr, true },
        };
    case MessageName::WebPage_RunJavaScriptInFrameInScriptWorld:
        return Vector<ArgumentDescription> {
            { "resultData", "IPC::DataReference", nullptr, false },
            { "details", "WebCore::ExceptionDetails", nullptr, true },
        };
    case MessageName::WebPage_GetAccessibilityTreeData:
        return Vector<ArgumentDescription> {
            { "dataReference", "IPC::SharedBufferReference", nullptr, true },
        };
    case MessageName::WebPage_ForceRepaint:
        return Vector<ArgumentDescription> { };
#if ENABLE(DATA_DETECTION)
    case MessageName::WebPage_DetectDataInAllFrames:
        return Vector<ArgumentDescription> {
            { "result", "WebKit::DataDetectionResult", nullptr, false },
        };
    case MessageName::WebPage_RemoveDataDetectedLinks:
        return Vector<ArgumentDescription> {
            { "result", "WebKit::DataDetectionResult", nullptr, false },
        };
#endif
    case MessageName::WebPage_Suspend:
        return Vector<ArgumentDescription> {
            { "success", "bool", nullptr, false },
        };
    case MessageName::WebPage_Resume:
        return Vector<ArgumentDescription> {
            { "success", "bool", nullptr, false },
        };
    case MessageName::WebPage_TryClose:
        return Vector<ArgumentDescription> {
            { "shouldClose", "bool", nullptr, false },
        };
    case MessageName::WebPage_ValidateCommand:
        return Vector<ArgumentDescription> {
            { "isEnabled", "bool", nullptr, false },
            { "state", "int32_t", nullptr, false },
        };
    case MessageName::WebPage_RequestFontAttributesAtSelectionStart:
        return Vector<ArgumentDescription> {
            { "attributes", "WebCore::FontAttributes", nullptr, false },
        };
    case MessageName::WebPage_FindString:
        return Vector<ArgumentDescription> {
            { "found", "bool", nullptr, false },
        };
    case MessageName::WebPage_ReplaceMatches:
        return Vector<ArgumentDescription> {
            { "numberOfReplacements", "uint64_t", nullptr, false },
        };
    case MessageName::WebPage_FindRectsForStringMatches:
        return Vector<ArgumentDescription> {
            { "matches", "Vector<WebCore::FloatRect>", nullptr, false },
        };
    case MessageName::WebPage_FindTextRangesForStringMatches:
        return Vector<ArgumentDescription> {
            { "ranges", "Vector<WebKit::WebFoundTextRange>", nullptr, false },
        };
    case MessageName::WebPage_RequestRectForFoundTextRange:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::FloatRect", nullptr, false },
        };
    case MessageName::WebPage_AddLayerForFindOverlay:
        return Vector<ArgumentDescription> {
            { "findLayerID", "WebCore::GraphicsLayer::PlatformLayerID", nullptr, false },
        };
    case MessageName::WebPage_RemoveLayerForFindOverlay:
        return Vector<ArgumentDescription> { };
#if PLATFORM(IOS_FAMILY) && ENABLE(DRAG_SUPPORT)
    case MessageName::WebPage_InsertDroppedImagePlaceholders:
        return Vector<ArgumentDescription> {
            { "dropPreviewRects", "Vector<WebCore::IntRect>", nullptr, false },
            { "textIndicator", "WebCore::TextIndicatorData", nullptr, true },
        };
#endif
#if ENABLE(MEDIA_STREAM)
    case MessageName::WebPage_UserMediaAccessWasGranted:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(ENCRYPTED_MEDIA)
    case MessageName::WebPage_MediaKeySystemWasGranted:
        return Vector<ArgumentDescription> { };
#endif
    case MessageName::WebPage_RequestMediaPlaybackState:
        return Vector<ArgumentDescription> {
            { "state", "uint8_t", "WebKit::MediaPlaybackState", false },
        };
    case MessageName::WebPage_PauseAllMediaPlayback:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_SuspendAllMediaPlayback:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_ResumeAllMediaPlayback:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_IsLayerTreeFrozen:
        return Vector<ArgumentDescription> {
            { "isFrozen", "bool", nullptr, false },
        };
    case MessageName::WebPage_EndPrinting:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_EndPrintingDuringDOMPrintOperation:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_ComputePagesForPrinting:
        return Vector<ArgumentDescription> {
            { "pageRects", "Vector<WebCore::IntRect>", nullptr, false },
            { "totalScaleFactorForPrinting", "double", nullptr, false },
            { "computedPageMargin", "WebCore::RectEdges<float>", nullptr, false },
        };
    case MessageName::WebPage_ComputePagesForPrintingDuringDOMPrintOperation:
        return Vector<ArgumentDescription> {
            { "pageRects", "Vector<WebCore::IntRect>", nullptr, false },
            { "totalScaleFactorForPrinting", "double", nullptr, false },
            { "computedPageMargin", "WebCore::RectEdges<float>", nullptr, false },
        };
#if PLATFORM(COCOA)
    case MessageName::WebPage_DrawRectToImage:
        return Vector<ArgumentDescription> {
            { "image", "WebKit::ShareableBitmapHandle", nullptr, false },
        };
    case MessageName::WebPage_DrawRectToImageDuringDOMPrintOperation:
        return Vector<ArgumentDescription> {
            { "image", "WebKit::ShareableBitmapHandle", nullptr, false },
        };
    case MessageName::WebPage_DrawPagesToPDF:
        return Vector<ArgumentDescription> {
            { "data", "RefPtr<WebCore::SharedBuffer>", nullptr, false },
        };
    case MessageName::WebPage_DrawPagesToPDFDuringDOMPrintOperation:
        return Vector<ArgumentDescription> {
            { "data", "RefPtr<WebCore::SharedBuffer>", nullptr, false },
        };
#endif
#if (PLATFORM(COCOA) && PLATFORM(IOS_FAMILY))
    case MessageName::WebPage_ComputePagesForPrintingiOS:
        return Vector<ArgumentDescription> {
            { "pageCount", "size_t", nullptr, false },
        };
    case MessageName::WebPage_DrawToPDFiOS:
        return Vector<ArgumentDescription> {
            { "data", "RefPtr<WebCore::SharedBuffer>", nullptr, false },
        };
#endif
#if PLATFORM(COCOA)
    case MessageName::WebPage_DrawToPDF:
        return Vector<ArgumentDescription> {
            { "data", "RefPtr<WebCore::SharedBuffer>", nullptr, false },
        };
#endif
#if PLATFORM(GTK)
    case MessageName::WebPage_DrawPagesForPrinting:
        return Vector<ArgumentDescription> {
            { "data", "WebKit::SharedMemory::Handle", nullptr, true },
            { "error", "WebCore::ResourceError", nullptr, false },
        };
    case MessageName::WebPage_DrawPagesForPrintingDuringDOMPrintOperation:
        return Vector<ArgumentDescription> {
            { "data", "WebKit::SharedMemory::Handle", nullptr, true },
            { "error", "WebCore::ResourceError", nullptr, false },
        };
#endif
    case MessageName::WebPage_SetMuted:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_StopMediaCapture:
        return Vector<ArgumentDescription> { };
#if PLATFORM(COCOA)
    case MessageName::WebPage_GetStringSelectionForPasteboard:
        return Vector<ArgumentDescription> {
            { "stringValue", "String", nullptr, false },
        };
    case MessageName::WebPage_GetDataSelectionForPasteboard:
        return Vector<ArgumentDescription> {
            { "buffer", "RefPtr<WebCore::SharedBuffer>", nullptr, false },
        };
    case MessageName::WebPage_ReadSelectionFromPasteboard:
        return Vector<ArgumentDescription> {
            { "result", "bool", nullptr, false },
        };
    case MessageName::WebPage_ShouldDelayWindowOrderingEvent:
        return Vector<ArgumentDescription> {
            { "result", "bool", nullptr, false },
        };
    case MessageName::WebPage_AddDictationAlternative:
        return Vector<ArgumentDescription> {
            { "success", "bool", nullptr, false },
        };
    case MessageName::WebPage_DictationAlternativesAtSelection:
        return Vector<ArgumentDescription> {
            { "contexts", "Vector<WebCore::DictationContext>", nullptr, false },
        };
    case MessageName::WebPage_HasMarkedText:
        return Vector<ArgumentDescription> {
            { "hasMarkedText", "bool", nullptr, false },
        };
    case MessageName::WebPage_GetMarkedRangeAsync:
        return Vector<ArgumentDescription> {
            { "range", "WebKit::EditingRange", nullptr, false },
        };
    case MessageName::WebPage_GetSelectedRangeAsync:
        return Vector<ArgumentDescription> {
            { "range", "WebKit::EditingRange", nullptr, false },
        };
    case MessageName::WebPage_CharacterIndexForPointAsync:
        return Vector<ArgumentDescription> {
            { "location", "uint64_t", nullptr, false },
        };
    case MessageName::WebPage_FirstRectForCharacterRangeAsync:
        return Vector<ArgumentDescription> {
            { "rect", "WebCore::IntRect", nullptr, false },
            { "actualRange", "WebKit::EditingRange", nullptr, false },
        };
#endif
#if PLATFORM(MAC)
    case MessageName::WebPage_AttributedSubstringForCharacterRangeAsync:
        return Vector<ArgumentDescription> {
            { "string", "WebCore::AttributedString", nullptr, false },
            { "range", "WebKit::EditingRange", nullptr, false },
        };
#endif
#if ENABLE(IMAGE_ANALYSIS_ENHANCEMENTS)
    case MessageName::WebPage_ShouldAllowRemoveBackground:
        return Vector<ArgumentDescription> {
            { "result", "bool", nullptr, false },
        };
#endif
    case MessageName::WebPage_GetBytecodeProfile:
        return Vector<ArgumentDescription> {
            { "string", "String", nullptr, false },
        };
    case MessageName::WebPage_GetSamplingProfilerOutput:
        return Vector<ArgumentDescription> {
            { "string", "String", nullptr, false },
        };
    case MessageName::WebPage_TakeSnapshot:
        return Vector<ArgumentDescription> {
            { "image", "WebKit::ShareableBitmapHandle", nullptr, false },
        };
#if PLATFORM(COCOA)
    case MessageName::WebPage_RequestActiveNowPlayingSessionInfo:
        return Vector<ArgumentDescription> {
            { "active", "bool", nullptr, false },
            { "registeredAsNowPlayingApplication", "bool", nullptr, false },
            { "title", "String", nullptr, false },
            { "duration", "double", nullptr, false },
            { "elapsedTime", "double", nullptr, false },
            { "uniqueIdentifier", "uint64_t", nullptr, false },
        };
#endif
    case MessageName::WebPage_DidGetLoadDecisionForIcon:
        return Vector<ArgumentDescription> {
            { "iconData", "IPC::SharedBufferReference", nullptr, false },
        };
    case MessageName::WebPage_URLSchemeTaskWillPerformRedirection:
        return Vector<ArgumentDescription> {
            { "actualRequest", "WebCore::ResourceRequest", nullptr, false },
        };
#if ENABLE(ATTACHMENT_ELEMENT)
    case MessageName::WebPage_InsertAttachment:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_UpdateAttachmentAttributes:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(APPLICATION_MANIFEST)
    case MessageName::WebPage_GetApplicationManifest:
        return Vector<ArgumentDescription> {
            { "manifest", "WebCore::ApplicationManifest", nullptr, true },
        };
#endif
    case MessageName::WebPage_GetTextFragmentMatch:
        return Vector<ArgumentDescription> {
            { "match", "String", nullptr, false },
        };
#if ENABLE(UI_PROCESS_PDF_HUD)
    case MessageName::WebPage_SavePDF:
        return Vector<ArgumentDescription> {
            { "filename", "String", nullptr, false },
            { "url", "URL", nullptr, false },
            { "data", "IPC::DataReference", nullptr, false },
        };
    case MessageName::WebPage_OpenPDFWithPreview:
        return Vector<ArgumentDescription> {
            { "filename", "String", nullptr, false },
            { "frameInfo", "WebKit::FrameInfoData", nullptr, false },
            { "data", "IPC::DataReference", nullptr, false },
            { "uuid", "String", nullptr, false },
        };
#endif
#if ENABLE(TRACKING_PREVENTION)
    case MessageName::WebPage_GetLoadedSubresourceDomains:
        return Vector<ArgumentDescription> {
            { "domains", "Vector<WebCore::RegistrableDomain>", nullptr, false },
        };
#endif
#if PLATFORM(GTK) || PLATFORM(WPE)
    case MessageName::WebPage_SendMessageToWebProcessExtensionWithReply:
        return Vector<ArgumentDescription> {
            { "replyMessage", "WebKit::UserMessage", nullptr, false },
        };
#endif
    case MessageName::WebPage_StartTextManipulations:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_CompleteTextManipulation:
        return Vector<ArgumentDescription> {
            { "allFailed", "bool", nullptr, false },
            { "failures", "Vector<WebCore::TextManipulationController::ManipulationFailure>", nullptr, false },
        };
    case MessageName::WebPage_GetProcessDisplayName:
        return Vector<ArgumentDescription> {
            { "displayName", "String", nullptr, false },
        };
    case MessageName::WebPage_DispatchWheelEventWithoutScrolling:
        return Vector<ArgumentDescription> {
            { "handled", "bool", nullptr, false },
        };
    case MessageName::WebPage_LastNavigationWasAppInitiated:
        return Vector<ArgumentDescription> {
            { "wasAppBound", "bool", nullptr, false },
        };
#if ENABLE(MEDIA_SESSION_COORDINATOR)
    case MessageName::WebPage_CreateMediaSessionCoordinator:
        return Vector<ArgumentDescription> {
            { "success", "bool", nullptr, false },
        };
#endif
#if ENABLE(IMAGE_ANALYSIS)
    case MessageName::WebPage_UpdateWithTextRecognitionResult:
        return Vector<ArgumentDescription> {
            { "result", "uint8_t", "WebKit::TextRecognitionUpdateResult", false },
        };
#endif
    case MessageName::WebPage_NavigateServiceWorkerClient:
        return Vector<ArgumentDescription> {
            { "result", "bool", nullptr, false },
        };
#if ENABLE(ACCESSIBILITY_ANIMATION_CONTROL)
    case MessageName::WebPage_PauseAllAnimations:
        return Vector<ArgumentDescription> { };
    case MessageName::WebPage_PlayAllAnimations:
        return Vector<ArgumentDescription> { };
#endif
#if ENABLE(WEBXR)
#endif
#if ENABLE(MEDIA_STREAM)
    case MessageName::UserMediaCaptureManager_ValidateUserMediaRequestConstraints:
        return Vector<ArgumentDescription> {
            { "invalidConstraint", "String", nullptr, true },
            { "audioDevices", "Vector<WebCore::CaptureDevice>", nullptr, false },
            { "videoDevices", "Vector<WebCore::CaptureDevice>", nullptr, false },
        };
    case MessageName::UserMediaCaptureManager_GetMediaStreamDevices:
        return Vector<ArgumentDescription> {
            { "devices", "Vector<WebCore::CaptureDevice>", nullptr, false },
        };
#endif
    default:
        break;
    }
    return std::nullopt;
}

} // namespace WebKit

#endif // ENABLE(IPC_TESTING_API) || !LOG_DISABLED
